package de.mannheim.uni.expansion.data.dbpedia.adapter.csv;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.zip.GZIPInputStream;

import javax.inject.Inject;

import org.apache.commons.collections4.BidiMap;
import org.apache.commons.collections4.bidimap.DualHashBidiMap;

import com.opencsv.CSVParser;
import com.opencsv.CSVReader;

import de.mannheim.uni.expansion.data.dbpedia.adapter.KnowledgeBaseLoadCache;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ListType;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.CSVSerializer;
import de.mannheim.uni.expansion.utils.ConsumerHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.StringSerializer;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseImpl;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public class DBPediaRepositoryCSVReader {

	private KnowledgeBaseLoadCache cache;
	private Map<ClassId, Integer> tableIndex;

	private ConcurrentHashMap<ClassId, Map<PropertyId, Integer>> propertyIndices;

	private ConcurrentHashMap<ClassId, Map<InstanceId, Integer>> instanceIndices;

	private ConcurrentHashMap<ClassPropertyId, Set<InstanceId>> instancesPerClassProperty;

	private Map<ValueId, QualifiedValueList> values;

	private Map<ClassPropertyId, ColumnDataType> columnDataTypes;
	private Map<ClassId, PropertyId> classKeys;

	private Set<ClassPropertyId> listTypes;
	private EnumMap<ListType, Set<ClassPropertyId>> listTypesMap;

	private Map<String, ColumnDataType> columnDataTypesMapping;

	private final StringSerializer stringSerializer;

	final private CHMHelper.Creator<BidiMap<PropertyId, Integer>> bidiMapCreatorPropertyId = () -> {
		return new DualHashBidiMap<PropertyId, Integer>();
	};

	final private CHMHelper.Creator<Set<InstanceId>> hashSetCreatorInstanceId = () -> {
		return new HashSet<InstanceId>();
	};

	final private CHMHelper.Creator<BidiMap<InstanceId, Integer>> bidiMapCreatorInstanceId = () -> {
		return new DualHashBidiMap<InstanceId, Integer>();
	};

	private final CSVSerializer csvSerializer;
	private final ValueConverter valueConverter;

	@Inject
	public DBPediaRepositoryCSVReader(StringSerializer stringSerializer, CSVSerializer csvSerializer, ValueConverter valueConverter) {
		this.stringSerializer = stringSerializer;
		this.csvSerializer = csvSerializer;
		this.valueConverter = valueConverter;
	}

	synchronized KnowledgeBase read(String folder, boolean newValueFormat) throws InterruptedException, IOException {

		cache = new KnowledgeBaseLoadCache();
		tableIndex = new ConcurrentHashMap<ClassId, Integer>();

		propertyIndices = new ConcurrentHashMap<ClassId, Map<PropertyId, Integer>>();

		instanceIndices = new ConcurrentHashMap<ClassId, Map<InstanceId, Integer>>();

		instancesPerClassProperty = new ConcurrentHashMap<>();

		values = new ConcurrentHashMap<ValueId, QualifiedValueList>(16, 0.75F, 8);

		columnDataTypes = new ConcurrentHashMap<ClassPropertyId, ColumnDataType>();
		classKeys = new ConcurrentHashMap<ClassId, PropertyId>();

		listTypes = Collections.newSetFromMap(new ConcurrentHashMap<>());
		listTypesMap = new EnumMap<>(ListType.class);

		columnDataTypesMapping = new HashMap<String, ColumnDataType>();

		for (ColumnDataType e : ColumnDataType.values()) {
			columnDataTypesMapping.put(e.name(), e);
		}

		if (false) {
			readClassKeys(folder);

			readTableIndices(folder);
		}
		long startTime = System.nanoTime();
		if (newValueFormat) {
			System.out.println("Using NEW Format for reading");
			readValuesNew(folder);
		} else {
			System.out.println("Using OLD Format for reading");
			readValuesOld(folder);
		}

		System.out.println("Read DBPedia Values in second: " + TimeUnit.SECONDS.convert(System.nanoTime() - startTime, TimeUnit.NANOSECONDS));

		if (false) {
			readPropertyIndices(folder);
		}
		startTime = System.nanoTime();
		if (newValueFormat) {
			System.out.println("Using NEW Format for reading");
			readInstanceIndicesNew(folder);
		} else {
			System.out.println("Using OLD Format for reading");
			readInstanceIndices(folder);
		}

		System.out.println("Read DBPedia Instance Indices in second: " + TimeUnit.SECONDS.convert(System.nanoTime() - startTime, TimeUnit.NANOSECONDS));

		if (false) {
			readInstanceIndicesPerClassProperty(folder);
			if (instancesPerClassProperty.size() == 0) {
				instancesPerClassProperty = null;
				System.out.println("Setting Instances Per Class Property To NULL as Files not found!");
			}

			readColumnDataTypes(folder);

			try {
				readListTypes(folder);
			} catch (FileNotFoundException e) {
				listTypes = null;
				listTypesMap = null;
			}
		}
		ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(instanceIndices.entrySet(), (e) -> {
			PersistentParallel6.runAndFinish(e.getValue().keySet(), (instanceId) -> {
				CHMHelper.getFromCHM(setMappings, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e.getKey());
			}, this.getClass());
		}, this.getClass());

		return KnowledgeBaseImpl.create(values, setMappings, new HashMap<>(), new HashMap<>(), new HashMap<>());

		// return new KnowledgeBaseImpl(tableIndex.keySet(), propertyIndices,
		// new HashMap<>(), instanceIndices, values, instancesPerClassProperty,
		// setMappings);

	}

	private void readClassKeys(String folder) throws IOException {
		CSVReader reader = new CSVReader(new FileReader(folder + "classKeys"), ';', '"');
		for (String[] line = reader.readNext(); line != null; line = reader.readNext()) {
			ClassId classId = cache.getFromCache(new ClassId(line[0]));

			PropertyId propertyId = cache.getFromCache(new PropertyId(line[1]));
			classKeys.put(classId, propertyId);
		}
		reader.close();
	}

	private void readListTypes(String folder) throws FileNotFoundException, IOException {

		Set<ClassPropertyId> shortListTypes = new HashSet<>();
		Set<ClassPropertyId> mediumListTypes = new HashSet<>();
		Set<ClassPropertyId> longListTypes = new HashSet<>();

		CSVReader reader = new CSVReader(new FileReader(folder + "listTypes"), ';', '"');
		try {
			for (String[] line = reader.readNext(); line != null; line = reader.readNext()) {
				ClassId classId = cache.getFromCache(new ClassId(line[0]));

				PropertyId propertyId = cache.getFromCache(new PropertyId(line[1]));

				if (line.length > 2) {
					if (ListType.SHORT.toString().equals(line[2])) {
						shortListTypes.add(new ClassPropertyId(classId, propertyId));
					} else if (ListType.MEDIUM.toString().equals(line[2])) {
						mediumListTypes.add(new ClassPropertyId(classId, propertyId));
					} else if (ListType.LONG.toString().equals(line[2])) {
						longListTypes.add(new ClassPropertyId(classId, propertyId));
					} else {
						throw new IllegalStateException("We do not support a list type : " + line[2]);
					}

				}
				listTypes.add(new ClassPropertyId(classId, propertyId));
				listTypesMap.put(ListType.SHORT, shortListTypes);
				listTypesMap.put(ListType.MEDIUM, mediumListTypes);
				listTypesMap.put(ListType.LONG, longListTypes);
			}
		} finally {
			reader.close();
		}
	}

	private void readTableIndices(String folder) throws IOException {
		CSVReader reader = new CSVReader(new FileReader(folder + "tableIndex"), ';', '"');
		for (String[] line = reader.readNext(); line != null; line = reader.readNext()) {
			ClassId classId = cache.getFromCache(new ClassId(line[0]));

			Integer integer = new Integer(line[1]);
			tableIndex.put(classId, integer);
		}
		reader.close();

	}

	static String[] parseCSVLine(String line) {
		ArrayList<String> list = new ArrayList<String>();
		boolean foundBegin = false;
		int begin = 0;
		char[] chars = new char[2];

		int index = line.indexOf('"');
		while (index >= 0) {
			if (!foundBegin) {
				foundBegin = true;
				begin = index + 1;
			} else {
				line.getChars(index - 2, index - 1, chars, 0);
				if (isEscaped(line, chars, index)) {
					list.add(line.substring(begin, index));
					foundBegin = false;
				}
			}

			index = line.indexOf('"', index + 1);

		}

		return list.toArray(new String[list.size()]);
	}

	static boolean isEscaped(String line, char[] chars, int index) {
		line.getChars(index - 2, index - 1, chars, 0);
		if (chars[1] != '\\') {
			return true;
		} else if (chars[0] == '\\') {
			return isEscaped(line, chars, index - 2);
		} else {
			return true;
		}

	}

	private void readValuesNew(String folder) throws FileNotFoundException, IOException, InterruptedException {

		// String bytes =
		// "rO0ABXNyABNqYXZhLnV0aWwuQXJyYXlMaXN0eIHSHZnHYZ0DAAFJAARzaXpleHAAAAACdwQAAAACdAAUV2l0aCBNaW5kICZhbXA7IEhhbmR0AAFceA==";
		// Object o;
		// try {
		// o = stringSerializer.fromString(bytes);
		//
		// StringWriter sw = new StringWriter();
		// CSVWriter writer = new CSVWriter(sw, ';', '"', '\\');
		//
		// String[] line = csvSerializer.getLine(o, 2);
		//
		// line[0] = "CLASS";
		// line[1] = "PROPERTY";
		//
		// writer.writeNext(line);
		//
		// writer.close();
		// System.out.println(sw.toString());
		// int a = 5;
		// } catch (ClassNotFoundException e) {
		// // TODO Auto-generated catch block
		// e.printStackTrace();
		// }

		// CSVParser parser = new CSVParser(';', '"', '\\');

		// ConsumerHelper<String> consumerHelper =
		// ConsumerHelper.withLinkedQueue("Read DBpedia", (finalNextLine) -> {
		//
		// try {
		// String[] finalLine = parseCSVLine(finalNextLine);
		//
		// InstanceId instanceId = cache.getFromCache(new
		// InstanceId(finalLine[0]));
		// PropertyId propertyId = cache.getFromCache(new
		// PropertyId(finalLine[1]));
		// Object value = csvSerializer.getObject(finalLine, 2);
		//
		// values.put(new ValueId(instanceId, propertyId), value);
		// } catch (Exception e1) {
		// // TODO Auto-generated catch block
		// e1.printStackTrace();
		// }
		//
		// } , new String("POISEN"));

		ConsumerHelper<String> producerHelper = ConsumerHelper.withLinkedQueue("DBpedia Read File", (file) -> {
			try {
				BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(file), 204800)));
				String nextLine = null;
				while ((nextLine = br.readLine()) != null) {

					String[] finalLine = parseCSVLine(nextLine);

					InstanceId instanceId = cache.getFromCache(new InstanceId(finalLine[0]));
					PropertyId propertyId = cache.getFromCache(new PropertyId(finalLine[1]));
					Object value = cache.getFromCache(csvSerializer.getObject(finalLine, 2));

					values.put(new ValueId(instanceId, propertyId), valueConverter.getQVL(value));

				}

				br.close();
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

		}, new String("POISEN"));

		producerHelper.start();

		for (int i = 0; i < 32; i++) {

			producerHelper.add(folder + "valuesNew" + i);
		}

		producerHelper.finishCloseAndAwait();

		// consumerHelper.start();
		//
		// consumerHelper.finishCloseAndAwait();
	}

	@Deprecated
	private void readValuesOld(String folder) throws IOException, InterruptedException {
		CSVParser parser = new CSVParser(';', '"');

		BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(folder + "values"), 204800)));

		ConsumerHelper<String> consumerHelper = ConsumerHelper.withLinkedQueue("Read DBpedia", (finalNextLine) -> {

			try {
				String[] finalLine = parser.parseLineMulti(finalNextLine);

				InstanceId instanceId = cache.getFromCache(new InstanceId(finalLine[0]));
				PropertyId propertyId = cache.getFromCache(new PropertyId(finalLine[1]));
				Object value = stringSerializer.fromString(finalLine[2]);

				values.put(new ValueId(instanceId, propertyId), new QualifiedValueList(valueConverter.getAbstractObject(value)));
			} catch (Exception e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}

		}, getThreadSize(), new String("POISEN"));

		consumerHelper.start();

		String nextLine = null;
		while ((nextLine = br.readLine()) != null) {

			consumerHelper.put(nextLine);

		}
		br.close();

		consumerHelper.finishCloseAndAwait();
	}

	private void readPropertyIndices(String folder) throws IOException {
		CSVReader reader = new CSVReader(new FileReader(folder + "propertyIndices"), ';', '"');
		for (String[] line = reader.readNext(); line != null; line = reader.readNext()) {
			ClassId classId = cache.getFromCache(new ClassId(line[0]));
			PropertyId propertyId = cache.getFromCache(new PropertyId(line[1]));
			Integer integer = new Integer(line[2]);
			Map<PropertyId, Integer> properties = CHMHelper.getFromCHM(propertyIndices, classId, () -> new ConcurrentHashMap<>());
			properties.put(propertyId, integer);
		}
		reader.close();

	}

	private void readInstanceIndices(String folder) throws IOException, InterruptedException {
		CSVParser parser = new CSVParser(';', '"');

		ConsumerHelper<String> consumerHelper = ConsumerHelper.withLinkedQueue("Read DBpedia Instance Indices", (finalNextLine) -> {

			try {
				String[] line = parser.parseLineMulti(finalNextLine);

				ClassId classId = cache.getFromCache(new ClassId(line[0]));
				InstanceId instanceId = cache.getFromCache(new InstanceId(line[1]));
				Integer integer = new Integer(line[2]);
				Map<InstanceId, Integer> instances = CHMHelper.getFromCHM(instanceIndices, classId, () -> new ConcurrentHashMap<>());
				// synchronized (instances) {
				instances.put(instanceId, integer);
				// }
			} catch (Exception e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}

		}, getThreadSize(), new String("POISEN"));

		consumerHelper.start();

		BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(folder + "instanceIndices")));

		String nextLine = null;
		while ((nextLine = br.readLine()) != null) {

			consumerHelper.put(nextLine);

		}
		br.close();

		consumerHelper.finishCloseAndAwait();
	}

	private void readInstanceIndicesNew(String folder) throws FileNotFoundException, IOException, InterruptedException {

		ConsumerHelper<String> producerHelper = ConsumerHelper.withLinkedQueue("DBpedia Read Instance Indices", (file) -> {
			try {
				BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(file), 204800)));
				String nextLine = null;
				while ((nextLine = br.readLine()) != null) {

					String[] line = parseCSVLine(nextLine);

					ClassId classId = cache.getFromCache(new ClassId(line[0]));
					InstanceId instanceId = cache.getFromCache(new InstanceId(line[1]));
					Integer integer = new Integer(line[2]);
					Map<InstanceId, Integer> instances = CHMHelper.getFromCHM(instanceIndices, classId, () -> new ConcurrentHashMap<>());
					// synchronized (instances) {
					instances.put(instanceId, integer);
					// }

				}

				br.close();
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

		}, new String("POISEN"));

		producerHelper.start();

		for (int i = 0; i < 32; i++) {

			producerHelper.add(folder + "instanceIndicesNew" + i);
		}

		producerHelper.finishCloseAndAwait();

		// consumerHelper.start();
		//
		// consumerHelper.finishCloseAndAwait();
	}

	private void readInstanceIndicesPerClassProperty(String folder) throws FileNotFoundException, IOException, InterruptedException {

		ConsumerHelper<String> producerHelper = ConsumerHelper.withLinkedQueue("DBpedia Read Instance Indices Per ClassProperty", (file) -> {
			try {
				BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(file), 204800)));
				String nextLine = null;
				while ((nextLine = br.readLine()) != null) {

					String[] line = parseCSVLine(nextLine);

					ClassId classId = cache.getFromCache(new ClassId(line[0]));
					PropertyId propertyId = cache.getFromCache(new PropertyId(line[1]));
					InstanceId instanceId = cache.getFromCache(new InstanceId(line[2]));
					// Integer integer = new Integer(line[3]);
					Set<InstanceId> instances = CHMHelper.getFromCHM(instancesPerClassProperty, new ClassPropertyId(classId, propertyId),
							this.hashSetCreatorInstanceId);
					synchronized (instances) {
						instances.add(instanceId);
					}

				}

				br.close();
			} catch (Exception e) {
				// TODO Auto-generated catch block
				if (e instanceof FileNotFoundException) {

				} else {
					e.printStackTrace();
				}
			}

		}, new String("POISEN"));

		producerHelper.start();

		for (int i = 0; i < 32; i++) {

			producerHelper.add(folder + "instancesPerClassProperty" + i);
		}

		producerHelper.finishCloseAndAwait();

		// consumerHelper.start();
		//
		// consumerHelper.finishCloseAndAwait();
	}

	private void readColumnDataTypes(String folder) throws IOException {
		CSVReader reader = new CSVReader(new FileReader(folder + "columnDataTypes"), ';', '"');
		for (String[] line = reader.readNext(); line != null; line = reader.readNext()) {
			ClassId classId = cache.getFromCache(new ClassId(line[0]));
			PropertyId propertyId = cache.getFromCache(new PropertyId(line[1]));
			ClassPropertyId classPropertyId = new ClassPropertyId(classId, propertyId);

			ColumnDataType columnDataType = columnDataTypesMapping.get(line[2]);

			columnDataTypes.put(classPropertyId, columnDataType);
		}
		reader.close();

	}

	public synchronized Map<ClassPropertyId, ColumnDataType> getColumnDataTypes(String folder) throws IOException {
		this.columnDataTypes = new ConcurrentHashMap<ClassPropertyId, ColumnDataType>();
		cache = new KnowledgeBaseLoadCache();

		columnDataTypesMapping = new HashMap<String, ColumnDataType>();

		for (ColumnDataType e : ColumnDataType.values()) {
			columnDataTypesMapping.put(e.name(), e);
		}

		readColumnDataTypes(folder);
		return columnDataTypes;
	}

	private int getThreadSize() {
		return ConsumerHelper.getAvailableCores() * 3000;
	}
}
