package de.mannheim.uni.expansion.data.ontology.model.repository;

import java.io.Serializable;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;
import com.google.common.collect.ImmutableSet;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ListType;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.QualifierId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseImpl;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class OntologyRepositoryImpl implements OntologyRepository, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	final ImmutableMap<PropertyId, ValueType> propertyDataType;
	final ImmutableSet<PropertyId> propertyTimeDependent;
	final ImmutableMap<ClassPropertyId, Set<ClassId>> propertyRange;
	final ImmutableMap<ClassPropertyId, Set<ClassId>> propertyDomain;
	final ImmutableMap<ClassId, Set<ClassId>> parents;
	final ImmutableMap<ClassId, Set<ClassId>> children;
	final ImmutableMap<PropertyId, Set<PropertyId>> propertyParents;

	public OntologyRepositoryImpl(Map<PropertyId, ValueType> propertyDataType, Set<PropertyId> propertyTimeDependent,
			Map<ClassPropertyId, Set<ClassId>> propertyDomain, Map<ClassPropertyId, Set<ClassId>> propertyRange, Map<ClassId, Set<ClassId>> parents,
			Map<ClassId, Set<ClassId>> children, Map<PropertyId, Set<PropertyId>> propertyParents) {
		this.propertyDataType = ImmutableMap.copyOf(propertyDataType);
		this.propertyTimeDependent = ImmutableSet.copyOf(propertyTimeDependent);

		Builder<ClassPropertyId, Set<ClassId>> builder = ImmutableMap.builder();
		for (Entry<ClassPropertyId, Set<ClassId>> e : propertyRange.entrySet()) {
			builder.put(e.getKey(), ImmutableSet.copyOf(e.getValue()));
		}
		this.propertyRange = builder.build();

		Builder<ClassId, Set<ClassId>> parentsBuilder = ImmutableMap.builder();
		for (Entry<ClassId, Set<ClassId>> e : parents.entrySet()) {
			parentsBuilder.put(e.getKey(), ImmutableSet.copyOf(e.getValue()));
		}
		this.parents = parentsBuilder.build();

		Builder<ClassId, Set<ClassId>> childrenBuilder = ImmutableMap.builder();
		for (Entry<ClassId, Set<ClassId>> e : children.entrySet()) {
			childrenBuilder.put(e.getKey(), ImmutableSet.copyOf(e.getValue()));
		}
		this.children = childrenBuilder.build();

		this.propertyDomain = (ImmutableMap<ClassPropertyId, Set<ClassId>>) ImmuMaker.mapWithSet(propertyDomain);
		this.propertyParents = (ImmutableMap<PropertyId, Set<PropertyId>>) ImmuMaker.mapWithSet(propertyParents);

	}

	@Override
	public boolean isReferenceField(PropertyId propertyId) {
		boolean isReference = propertyDataType.get(propertyId) == ValueType.INSTANCE_REFERENCE;
		return isReference;
	}

	public ImmutableMap<ClassPropertyId, Set<ClassId>> getPropertyRange() {
		return propertyRange;
	}

	@Override
	public ValueType getDataType(PropertyId propertyId) {
		ValueType vt = null;
		if (propertyId instanceof QualifierId) {
			PropertyId qualifierId = ((QualifierId) propertyId).getQualifier();

			if (qualifierId.equals(KnowledgeBaseImpl.endTimeQualifier) || qualifierId.equals(KnowledgeBaseImpl.startTimeQualifier)
					|| qualifierId.equals(KnowledgeBaseImpl.pointInTime)) {
				vt = ValueType.TIME;
			} else {
				vt = propertyDataType.get(qualifierId);
			}

		} else {
			vt = propertyDataType.get(propertyId);
			if (vt == null) {
				if (propertyId.equals(PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY)
						|| propertyId.equals(PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY)) {
					vt = ValueType.STRING;
				}
			}
		}
		if (vt == null) {
			// FIXME:
			System.out.println("Missing Datatype in Ontology for property: " + propertyId.toCombinedString());
			int fdf = 123;
			return ValueType.QUANTITY;
		}
		return vt;
	}

	@Override
	public Set<ClassId> getRange(ClassPropertyId propertyId) {
		return propertyRange.get(propertyId);
	}

	@Override
	public Set<ClassId> getDomain(ClassPropertyId propertyId) {
		return propertyDomain.get(propertyId);
	}

	@Override
	public Set<ClassId> getSubClasses(ClassId classId) {
		Set<ClassId> ret = new HashSet<>();

		Set<ClassId> childrenSet = children.get(classId);
		if (childrenSet != null) {
			ret.addAll(childrenSet);
			for (ClassId child : childrenSet) {
				ret.addAll(getSubClasses(child));
			}
		}

		return ret;
	}

	@Override
	public Set<ClassId> getParentClasses(ClassId classId) {
		Set<ClassId> ret = new HashSet<>();

		Set<ClassId> parentsSet = parents.get(classId);
		if (parentsSet != null) {
			ret.addAll(parentsSet);
			for (ClassId parent : parentsSet) {
				ret.addAll(getParentClasses(parent));
			}
		}

		return ret;
	}

	@Override
	public Set<ClassId> getDirectParentClasses(ClassId classId) {
		Set<ClassId> ret = new HashSet<>();
		Set<ClassId> parentsSet = parents.get(classId);
		if (parentsSet != null) {
			ret.addAll(parentsSet);
		}

		return ret;
	}

	@Override
	public boolean isListType(PropertyId propertyId) {
		return false;
	}

	@Override
	public boolean isListType(ListType listType, PropertyId propertyId) {
		return false;
	}

	@Override
	public boolean isTimeDependent(PropertyId propertyId) {
		return this.propertyTimeDependent.contains(propertyId);
	}

	public ImmutableMap<PropertyId, ValueType> getPropertyDataType() {
		return propertyDataType;
	}

	public ImmutableSet<PropertyId> getPropertyTimeDependent() {
		return propertyTimeDependent;
	}

	public Map<ClassId, Set<ClassId>> getParents() {
		return parents;
	}

	public Map<ClassId, Set<ClassId>> getChildren() {
		return children;
	}

	public Map<PropertyId, Set<PropertyId>> getPropertyParents() {
		return propertyParents;
	}

	@Override
	public Set<PropertyId> getAllPropertyParent(PropertyId propertyId) {
		Set<PropertyId> ret = new HashSet<>();

		Set<PropertyId> parentsSet = propertyParents.get(propertyId);
		if (parentsSet != null) {
			ret.addAll(parentsSet);
			for (PropertyId parent : parentsSet) {
				ret.addAll(getAllPropertyParent(parent));
			}
		}

		return ret;
	}

	@Override
	public Set<PropertyId> getDirectPropertyParent(PropertyId propertyId) {
		if (propertyParents.containsKey(propertyId)) {
			return propertyParents.get(propertyId);
		} else {
			return new HashSet<>();
		}
	}

	public Map<ClassPropertyId, Set<ClassId>> getPropertyDomain() {
		return propertyDomain;
	}

}
