package de.mannheim.uni.expansion.data.tables.model;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSortedMap;
import com.google.common.collect.ImmutableTable;
import com.google.common.collect.ImmutableTable.Builder;
import com.google.common.collect.Table;

import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class AnnotatedTable implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private final ImmutableTable<RowIdx, ColumnIdx, AbstractObject> table;
	private final ImmutableTable<RowIdx, ColumnIdx, String> stringTable;

	private final ImmutableSortedMap<ColumnIdx, ValueType> dataTypes;

	private final ImmutableSortedMap<ColumnIdx, AbstractObject> columnHeaders;

	private final ColumnIdx keyColumn;

	public Map<CellIdx, AbstractObject> getValues() {
		Map<CellIdx, AbstractObject> retMap = new HashMap<>();

		for (Entry<ColumnIdx, Map<RowIdx, AbstractObject>> columnMap : table.columnMap().entrySet()) {

			for (Entry<RowIdx, AbstractObject> row : columnMap.getValue().entrySet()) {
				retMap.put(new CellIdx(columnMap.getKey(), row.getKey()), row.getValue());
			}
		}

		return retMap;
	}

	public Map<CellIdx, String> getStringValues() {
		Map<CellIdx, String> retMap = new HashMap<>();

		for (Entry<ColumnIdx, Map<RowIdx, String>> columnMap : stringTable.columnMap().entrySet()) {

			for (Entry<RowIdx, String> row : columnMap.getValue().entrySet()) {
				retMap.put(new CellIdx(columnMap.getKey(), row.getKey()), row.getValue());
			}
		}

		return retMap;
	}

	private ImmutableTable<RowIdx, ColumnIdx, AbstractObject> tableFromValues(Map<CellIdx, AbstractObject> values) {
		Builder<RowIdx, ColumnIdx, AbstractObject> builder = ImmutableTable.builder();
		for (Entry<CellIdx, AbstractObject> e : values.entrySet()) {
			builder.put(e.getKey().getRowIdx(), e.getKey().getColumnIdx(), e.getValue());
		}

		builder.orderColumnsBy(ColumnIdx::compare);
		builder.orderRowsBy(RowIdx::compare);

		return builder.build();

	}

	private ImmutableTable<RowIdx, ColumnIdx, String> tableFromStrings(Map<CellIdx, String> values) {
		Builder<RowIdx, ColumnIdx, String> builder = ImmutableTable.builder();
		for (Entry<CellIdx, String> e : values.entrySet()) {
			builder.put(e.getKey().getRowIdx(), e.getKey().getColumnIdx(), e.getValue());
		}

		builder.orderColumnsBy(ColumnIdx::compare);
		builder.orderRowsBy(RowIdx::compare);

		return builder.build();

	}

	public AnnotatedTable(Map<CellIdx, AbstractObject> values, Map<CellIdx, String> strings, Map<ColumnIdx, ValueType> dataTypes, ColumnIdx keyColumn,
			Map<ColumnIdx, AbstractObject> columnHeaders) {

		this.table = tableFromValues(values);
		this.stringTable = tableFromStrings(strings);
		com.google.common.collect.ImmutableSortedMap.Builder<ColumnIdx, ValueType> datatypeBuilder = ImmutableSortedMap.naturalOrder();
		datatypeBuilder.putAll(dataTypes);
		this.dataTypes = datatypeBuilder.build();

		this.keyColumn = keyColumn;

		com.google.common.collect.ImmutableSortedMap.Builder<ColumnIdx, AbstractObject> headerBuilder = ImmutableSortedMap.naturalOrder();
		headerBuilder.putAll(columnHeaders);
		this.columnHeaders = headerBuilder.build();
	}

	public AnnotatedTable(Table<RowIdx, ColumnIdx, AbstractObject> table, Table<RowIdx, ColumnIdx, String> stringTable, Map<ColumnIdx, ValueType> dataTypes,
			ColumnIdx keyColumn, Map<ColumnIdx, AbstractObject> columnHeaders) {

		Builder<RowIdx, ColumnIdx, AbstractObject> builder = ImmutableTable.builder();

		builder.putAll(table);

		builder.orderColumnsBy(ColumnIdx::compare);
		builder.orderRowsBy(RowIdx::compare);

		this.table = builder.build();

		Builder<RowIdx, ColumnIdx, String> stringBuilder = ImmutableTable.builder();

		stringBuilder.putAll(stringTable);

		stringBuilder.orderColumnsBy(ColumnIdx::compare);
		stringBuilder.orderRowsBy(RowIdx::compare);

		this.stringTable = stringBuilder.build();

		com.google.common.collect.ImmutableSortedMap.Builder<ColumnIdx, ValueType> datatypeBuilder = ImmutableSortedMap.naturalOrder();
		datatypeBuilder.putAll(dataTypes);
		this.dataTypes = datatypeBuilder.build();
		this.keyColumn = keyColumn;

		com.google.common.collect.ImmutableSortedMap.Builder<ColumnIdx, AbstractObject> headerBuilder = ImmutableSortedMap.naturalOrder();
		headerBuilder.putAll(columnHeaders);
		this.columnHeaders = headerBuilder.build();
	}

	public ImmutableMap<RowIdx, AbstractObject> getColumn(ColumnIdx columnIdx) {
		return table.column(columnIdx);
	}

	public ImmutableMap<RowIdx, AbstractObject> getKeyColumn() {
		return table.column(keyColumn);
	}

	public Set<ColumnIdx> getColumns() {
		return this.getColumnHeaders().keySet();
	}

	public AbstractObject getValue(RowIdx rowIdx, ColumnIdx colIdx) {
		return table.get(rowIdx, colIdx);
	}

	public AbstractObject getKeyValue(RowIdx rowIdx) {
		return table.get(rowIdx, keyColumn);
	}

	public ValueType getDataType(ColumnIdx colIdx) {
		return this.dataTypes.get(colIdx);
	}

	public ImmutableTable<RowIdx, ColumnIdx, AbstractObject> getTable() {
		return table;
	}

	public AbstractObject getColumnHeader(ColumnIdx colIdx) {
		return this.columnHeaders.get(colIdx);
	}

	public ImmutableSortedMap<ColumnIdx, AbstractObject> getColumnHeaders() {
		return columnHeaders;
	}

	public ImmutableSortedMap<ColumnIdx, ValueType> getDataTypes() {
		return this.dataTypes;
	}

	public ColumnIdx getKeyColumnIndex() {
		return keyColumn;
	}

	public List<ColumnIdx> getTimeColumns() {
		List<ColumnIdx> list = new ArrayList<>();

		for (Entry<ColumnIdx, ValueType> e : this.getDataTypes().entrySet()) {
			if (e.getValue() == ValueType.TIME) {
				list.add(e.getKey());
			}
		}

		return list;
	}

	/***
	 * Get the set of Columns that are matched to DBPedia Properties excluding
	 * the KeyColumn
	 * 
	 * @param t
	 * @param metaData
	 * @return
	 */
	public static Map<Integer, ColumnDataType> getDBPediaPropertyColumns(de.mannheim.uni.tableprocessor.model.Table t, AnnotatedTableMetaData metaData) {
		Map<Integer, ColumnDataType> columnsToMatch = new HashMap<Integer, ColumnDataType>();
		for (int i = 0; i < t.getColumns().size(); i++) {
			TableColumn tc = t.getColumns().get(i);
			// Skip keys - key columns are already matched
			if (t.getColumns().get(i).isKey()) {
				continue;
			}

			// Skip if column is not matched to DBpedia
			if (!metaData.getProperties().containsKey(i) || metaData.getProperties().get(i) == null) {
				continue;
			}

			columnsToMatch.put(i, t.getColumns().get(i).getDataType());
		}

		return columnsToMatch;
	}

	public ImmutableTable<RowIdx, ColumnIdx, String> getStringTable() {
		return stringTable;
	}

	//
	// private Map<CellIdx, AbstractObject> values = null;
	//
	// @Override
	// public void write(Kryo kryo, Output output) {
	// this.values = getValues();
	// kryo.writeClassAndObject(output, this);
	// }
	//
	// @Override
	// public void read(Kryo kryo, Input input) {
	// AnnotatedTable data = kryo.readObject(input, AnnotatedTable.class);
	// this.columnHeaders = data.getColumnHeaders();
	// this.dataTypes = data.getDataTypes();
	// this.table = tableFromValues(data.values);
	// this.keyColumn = data.getKeyColumnIndex();
	//
	// }

}
