package de.uni_mannheim.dws.expansion.data.kb.degree;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.zip.GZIPInputStream;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.dbpedia.adapter.KnowledgeBaseLoadCache;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.BatchConsumerProducer;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ConcurrentKryoAdapter;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ConcurrentKryoReader;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ConcurrentKryoWriter;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeNoReferenceKryo;

public class DegreeAdapter {

	private final static String PREFIX = "DEGREE_";

	private final CHMKryoAdapter kryoAdapter;

	@Inject
	public DegreeAdapter(CHMKryoAdapter kryoAdapter) {
		this.kryoAdapter = kryoAdapter;
	}

	public boolean cacheExists(String cacheFolder) {
		File dirFile = new File(cacheFolder);
		if (dirFile.isDirectory()) {
			String[] dirFileList = dirFile.list();
			if (dirFileList.length > 0) {
				for (String f : dirFileList) {
					if (f.startsWith(PREFIX)) {
						return true;
					}
				}

			}
		}
		return false;
	}

	public DegreeRepositoryImpl readFromCache(String cacheFolder) {
		Map<InstanceId, Integer> map = kryoAdapter.readWithReferences(InstanceId.class, Integer.class, cacheFolder, PREFIX);

		return new DegreeRepositoryImpl(map);
	}

	public void writeToCache(DegreeRepositoryImpl degreeRepository, String cacheFolder) {
		kryoAdapter.write(cacheFolder, PREFIX, degreeRepository.getDegreeMap());
	}

	public DegreeRepositoryImpl readFromSource(String source) {

		try {
			KnowledgeBaseLoadCache loadCache = new KnowledgeBaseLoadCache();
			ConcurrentHashMap<InstanceId, Set<ValueId>> map = new ConcurrentHashMap<>();

			Pattern pattern = Pattern.compile("<(.*?)> <(.*?)> <(.*?)> .");

			final BatchConsumerProducer<String> batchConsumerProducer = BatchConsumerProducer.create("Read Indegree", (e) -> {
				Matcher matcher = pattern.matcher(e);

				if (matcher.find()) {
					InstanceId instanceId = loadCache.getFromCache(new InstanceId(matcher.group(1)));
					PropertyId propertyId = loadCache.getFromCache(new PropertyId(matcher.group(2)));
					String s = matcher.group(3);
					if (s.startsWith("http://dbpedia.org/resource/")) {
						InstanceId target = loadCache.getFromCache(new InstanceId(s));
						CHMHelper.getFromCHM(map, target, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(new ValueId(instanceId, propertyId));
					}

				}

			}, 5000, 1);

			BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(source))));
			List<String> list = new ArrayList<>();
			String nextLine = null;
			int sum = 0;
			while ((nextLine = br.readLine()) != null) {
				if (!nextLine.startsWith("#")) {
					list.add(nextLine);
					// if (list.size() == 10000) {
					// batchConsumerProducer.addBatch(list);
					// break;
					// }
					if (list.size() == 200000) {
						sum += 200000;
						batchConsumerProducer.addBatch(list);
						list = new ArrayList<>();
						// break;
					}
					if (sum == 400000) {
						sum = 0;
						System.out.println("read 1000000");
						// break;
					}
				}

			}

			if (!list.isEmpty()) {
				batchConsumerProducer.addBatch(list);
			}

			batchConsumerProducer.finish();

			br.close();

			return new DegreeRepositoryImpl(map.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().size())));
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	public static void main(String[] args) {
		ThreadSafeKryo threadSafeKryo = new ThreadSafeKryo();
		ThreadSafeNoReferenceKryo threadSafeNoReferencesKryo = new ThreadSafeNoReferenceKryo();
		DegreeAdapter adapter = new DegreeAdapter(new CHMKryoAdapter(new ConcurrentKryoAdapter(
				new ConcurrentKryoReader(threadSafeKryo, threadSafeNoReferencesKryo), new ConcurrentKryoWriter(threadSafeKryo, threadSafeNoReferencesKryo))));
		DegreeRepositoryImpl ret;
		if (true) {
			ret = adapter.readFromCache("D:/stuff/degcache/");
		} else {
			ret = adapter.readFromSource("D:/stuff/mappingbased_properties_en.nt.gz");
			adapter.writeToCache(ret, "D:/stuff/degcache/");
		}

		ret.printStatistics();

		try {
			PersistentParallel6.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
