package de.uni_mannheim.dws.expansion.data.kb.repository;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import com.google.common.base.Optional;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.ImmutableBiMap;
import com.google.common.collect.ImmutableBiMap.Builder;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;

import de.mannheim.uni.expansion.data.dbpedia.adapter.KnowledgeBaseLoadCache;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.instances.QualifierId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleArrayMap;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.remote.KnowledgeBaseRemote;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitOfMeasurement;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.Qualifier;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;

public class KnowledgeBaseImpl implements KnowledgeBase, KnowledgeBaseRemote {

	final Set<ClassId> sets;

	final Map<ClassId, BiMap<PropertyId, Integer>> propertyIndices;
	final Map<ClassId, BiMap<PropertyId, Integer>> propertyIndicesWithQualifiers;

	final Map<ClassId, BiMap<InstanceId, Integer>> instanceIndices;

	final Map<ClassId, SimpleArrayMap<InstanceId>> fastInstanceIndices;

	final Map<ValueId, QualifiedValueList> values;

	final Map<ClassPropertyId, Set<InstanceId>> instancesPerClassProperty;

	final Map<ClassPropertyId, Set<InstanceId>> instancesPerQualifier;

	final Map<InstanceId, Set<ClassId>> setMappings;

	final Map<ClassPropertyId, Set<ClassPropertyId>> equalProperties;

	final Map<PropertyId, UnitOfMeasurement> kbUnits;

	final Map<PropertyId, PropertyId> normalizationIndex;

	final Map<String, Set<String>> surfaceForms;
	final Map<String, Set<String>> candidatesSurfaceForms;

	public KnowledgeBaseImpl(Set<ClassId> sets, Map<ClassId, Map<PropertyId, Integer>> propertyIndices,
			Map<ClassId, Map<PropertyId, Integer>> propertyIndicesWithQualifiers, Map<ClassId, Map<InstanceId, Integer>> instanceIndices,
			Map<ValueId, QualifiedValueList> values, Map<ClassPropertyId, Set<InstanceId>> instancesPerClassProperty, Map<InstanceId, Set<ClassId>> setMappings,
			Map<ClassPropertyId, Set<ClassPropertyId>> equalProperties, Map<PropertyId, UnitOfMeasurement> kbUnits,
			Map<PropertyId, PropertyId> normalizationIndex) {
		super();

		KnowledgeBaseLoadCache cache = new KnowledgeBaseLoadCache();

		this.sets = cache.getImmutableSetFromCache(sets);

		Map<ClassId, BiMap<PropertyId, Integer>> propertyIndicesTemp = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(propertyIndices.entrySet(), (e) -> {
			propertyIndicesTemp.put(cache.getFromCache(e.getKey()), cache.getImmutableBiMap(e.getValue()));
		}, this.getClass());
		this.propertyIndices = Collections.unmodifiableMap(propertyIndicesTemp);

		Map<ClassId, BiMap<PropertyId, Integer>> propertyIndicesWithQualifiersTemp = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(propertyIndicesWithQualifiers.entrySet(), (e) -> {
			propertyIndicesWithQualifiersTemp.put(cache.getFromCache(e.getKey()), cache.getImmutableBiMap(e.getValue()));
		}, this.getClass());
		this.propertyIndicesWithQualifiers = Collections.unmodifiableMap(propertyIndicesWithQualifiersTemp);

		Map<ClassId, BiMap<InstanceId, Integer>> instanceIndicesTemp = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(instanceIndices.entrySet(), (e) -> {
			instanceIndicesTemp.put(cache.getFromCache(e.getKey()), cache.getImmutableBiMapInstance(e.getValue()));
		}, this.getClass());
		this.instanceIndices = Collections.unmodifiableMap(instanceIndicesTemp);

		this.values = cache.getImmutableMapValueId(values);

		Map<ClassPropertyId, Set<InstanceId>> instancesPerClassPropertyTemp = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(instancesPerClassProperty.entrySet(), (e) -> {
			instancesPerClassPropertyTemp.put(cache.getFromCache(e.getKey()), cache.getImmutableSetWithoutCache(e.getValue()));
		}, this.getClass());
		this.instancesPerClassProperty = Collections.unmodifiableMap(instancesPerClassPropertyTemp);

		Map<InstanceId, Set<ClassId>> setMappingsTemp = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(setMappings.entrySet(), (e) -> {
			setMappingsTemp.put(cache.getFromCache(e.getKey()), cache.getImmutableSetFromCache(e.getValue()));
		}, this.getClass());
		this.setMappings = ImmutableMap.copyOf(setMappingsTemp);

		ConcurrentHashMap<ClassPropertyId, Set<InstanceId>> instancesPerQualifier = new ConcurrentHashMap<>();

		for (Entry<ClassId, Map<PropertyId, Integer>> e : propertyIndicesWithQualifiers.entrySet()) {

			PersistentParallel6.runAndFinish(e.getValue().entrySet(), (e2) -> {
				if (e2.getKey() instanceof QualifierId) {
					QualifierId qualifierId = (QualifierId) e2.getKey();
					Set<InstanceId> instanceIds = instancesPerClassProperty.get(new ClassPropertyId(e.getKey(), qualifierId.getMain()));
					if (instanceIds == null) {
						int sdf = 123;
					}
					for (InstanceId instanceId : instanceIds) {
						QualifiedValueList qvl = values.get(new ValueId(instanceId, qualifierId.getMain()));
						if (qvl != null && qvl.hasQualifier(qualifierId.getQualifier())) {

							for (QualifiedValueHolder claim : qvl.getClaims()) {
								if (claim.hasQualifier(qualifierId.getQualifier())
										&& claim.getQualifierValue(qualifierId.getQualifier()).getQualifierValue() != null) {
									CHMHelper.getFromCHM(instancesPerQualifier, cache.getFromCache(new ClassPropertyId(e.getKey(), e2.getKey())),
											() -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(cache.getFromCache(instanceId));
									break;
								}
							}

						}
					}

				}

				// else {
				// ClassPropertyId cpid = new ClassPropertyId(e.getKey(),
				// e2.getKey());
				// instancesPerQualifier.put(cpid,
				// instancesPerClassProperty.get(cpid));
				// }
			}, "MAKING INSTANCES PER QUALIEFIER");

		}

		ConcurrentHashMap<ClassPropertyId, Set<InstanceId>> instancesPerQualifierTemp = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(instancesPerQualifier.entrySet(), (e) -> {
			instancesPerQualifierTemp.put(e.getKey(), ImmutableSet.copyOf(e.getValue()));
		}, this.getClass());

		this.instancesPerQualifier = Collections.unmodifiableMap(instancesPerQualifierTemp);

		// Map<ClassId, SimpleArrayMap<InstanceId>> fastInstanceMap = new
		// ConcurrentHashMap<>();
		//
		// PersistentParallel6.runAndFinish(this.instanceIndices.entrySet(),
		// (entry) -> {
		// int size =
		// entry.getValue().values().parallelStream().max(Integer::compareTo).get()
		// + 1;
		// SimpleArrayMap<InstanceId> arrayMap = new SimpleArrayMap<>(size);
		// for (Entry<InstanceId, Integer> innerEntry :
		// entry.getValue().entrySet()) {
		// arrayMap.put(innerEntry.getValue(),
		// cache.getFromCache(innerEntry.getKey()));
		// }
		//
		// fastInstanceMap.put(cache.getFromCache(entry.getKey()), arrayMap);
		// }, this.getClass());

		this.fastInstanceIndices = null;// Collections.unmodifiableMap(fastInstanceMap);

		this.equalProperties = ImmuMaker.mapWithSet(equalProperties);

		this.kbUnits = ImmuMaker.map(kbUnits);

		this.normalizationIndex = ImmuMaker.map(normalizationIndex);

		ConcurrentHashMap<String, Set<String>> surfaceForms = new ConcurrentHashMap<>();
		ConcurrentHashMap<String, Set<String>> candidatesSurfaceForms = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(setMappings.keySet(), (instanceId) -> {
			QualifiedValueList qvl = getQualifiedValueList(new ValueId(instanceId, PropertiesExpansionData.LABEL_PROPERTY));

			if (instanceId.getInstanceId().equals("http://dbpedia.org/resource/Guard_(American_and_Canadian_football)")) {
				int sdf = 123;
			}
			Set<String> labels = new HashSet<>();
			Set<String> candidates = new HashSet<>();
			for (QualifiedValueHolder claim : qvl.getClaims()) {
				if (claim.getValue().getType() == ValueType.STRING) {
					String s = ((StringObject) claim.getValue()).getString();
					s = StringNormalizer.normaliseValue(s, false);
					labels.add(s);
					candidates.addAll(StringNormalizer.tokenise(s, false, false));
				}
			}

			for (String label : labels) {
				CHMHelper.getFromCHM(surfaceForms, label, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).addAll(labels);
				CHMHelper.getFromCHM(candidatesSurfaceForms, label, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).addAll(candidates);

			}

		}, this.getClass());

		ConcurrentHashMap<String, Set<String>> surfaceForms2 = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(surfaceForms.entrySet(), (entry) -> {
			if (!entry.getValue().isEmpty()) {
				surfaceForms2.put(entry.getKey(), Collections.unmodifiableSet(entry.getValue()));
			}
		}, this.getClass());

		ConcurrentHashMap<String, Set<String>> candidatesSurfaceForms2 = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(candidatesSurfaceForms.entrySet(), (entry) -> {
			if (!entry.getValue().isEmpty()) {
				candidatesSurfaceForms2.put(entry.getKey(), Collections.unmodifiableSet(entry.getValue()));
			}
		}, this.getClass());

		int surfaceFormSize = surfaceForms2.size();

		this.surfaceForms = Collections.unmodifiableMap(surfaceForms2);
		this.candidatesSurfaceForms = Collections.unmodifiableMap(candidatesSurfaceForms2);
	}

	public static KnowledgeBaseImpl create(Map<ValueId, QualifiedValueList> values, Map<InstanceId, Set<ClassId>> setMappings,
			Map<ClassPropertyId, Set<ClassPropertyId>> equalProperties, Map<PropertyId, UnitOfMeasurement> kbUnits,
			Map<PropertyId, PropertyId> normalizationIndex) {
		ConcurrentHashMap<ClassId, Set<PropertyId>> propertyIndices = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassId, Set<PropertyId>> withSubProperties = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassId, Set<InstanceId>> instanceIndices = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassPropertyId, Set<InstanceId>> instancesPerClassProperty = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(setMappings.entrySet(), (setEntry) -> {
			InstanceId instanceId = setEntry.getKey();
			for (ClassId setId : setEntry.getValue()) {
				CHMHelper.getFromCHM(instanceIndices, setId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(instanceId);
				CHMHelper.getFromCHM(propertyIndices, setId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>()));
				CHMHelper.getFromCHM(withSubProperties, setId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>()));

			}
		}, KnowledgeBaseImpl.class);

		PersistentParallel6.runAndFinish(values.entrySet(), (e) -> {

			InstanceId instanceId = e.getKey().getInstanceId();
			PropertyId propertyId = e.getKey().getPropertyId();
			Set<ClassId> setIds = setMappings.get(instanceId);
			if (setIds == null) {
				System.out.println(instanceId.toString());
				return;
			}

			for (ClassId setId : setIds) {
				ClassPropertyId classPropertyId = new ClassPropertyId(setId, propertyId);

				CHMHelper.getFromCHM(propertyIndices, setId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(propertyId);
				CHMHelper.getFromCHM(withSubProperties, setId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(propertyId);

				CHMHelper.getFromCHM(instancesPerClassProperty, classPropertyId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(instanceId);

				for (QualifiedValueHolder claim : e.getValue().getClaims()) {

					for (PropertyId qualifier : claim.getQualifierIds()) {
						QualifierId qualifierId = new QualifierId(propertyId, qualifier);
						CHMHelper.getFromCHM(withSubProperties, setId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(qualifierId);
					}

				}

			}

		}, "KnowledgeBase");

		ConcurrentHashMap<ClassId, Map<PropertyId, Integer>> propertyIndicesBidi = new ConcurrentHashMap<>();
		for (Entry<ClassId, Set<PropertyId>> e : propertyIndices.entrySet()) {
			Map<PropertyId, Integer> bidiMap = new HashMap<>();

			Integer i = 0;
			for (PropertyId propertyId : e.getValue()) {
				bidiMap.put(propertyId, i);
				i++;
			}

			propertyIndicesBidi.put(e.getKey(), bidiMap);
		}

		ConcurrentHashMap<ClassId, Map<PropertyId, Integer>> withSubPropertiesBidi = new ConcurrentHashMap<>();
		for (Entry<ClassId, Set<PropertyId>> e : withSubProperties.entrySet()) {
			Map<PropertyId, Integer> bidiMap = new HashMap<>();

			Integer i = 0;
			for (PropertyId propertyId : e.getValue()) {
				bidiMap.put(propertyId, i);
				i++;
			}

			withSubPropertiesBidi.put(e.getKey(), bidiMap);
		}

		ConcurrentHashMap<ClassId, Map<InstanceId, Integer>> instanceIndicesBidi = new ConcurrentHashMap<>();
		for (Entry<ClassId, Set<InstanceId>> e : instanceIndices.entrySet()) {
			Map<InstanceId, Integer> bidiMap = new HashMap<>();

			Integer i = 0;
			for (InstanceId instanceId : e.getValue()) {
				bidiMap.put(instanceId, i);
				i++;
			}

			instanceIndicesBidi.put(e.getKey(), bidiMap);
		}

		return new KnowledgeBaseImpl(instanceIndices.keySet(), propertyIndicesBidi, withSubPropertiesBidi, instanceIndicesBidi, values,
				instancesPerClassProperty, setMappings, equalProperties, kbUnits, normalizationIndex);
	}

	@Override
	public Set<ClassId> getClasses() {
		return sets;
	}

	@Override
	public Set<ClassId> getClasses(InstanceId instanceId) {
		return setMappings.get(instanceId);
	}

	@Override
	public Set<ValueId> getValueIds() {
		return values.keySet();
	}
	// @Override
	// public InstanceId getInstance(ClassId setId, int index) {
	// SimpleArrayMap<InstanceId> indices = fastInstanceIndices.get(setId);
	// if (indices != null) {
	// InstanceId instanceId = indices.get(index);
	// if (instanceId != null) {
	// return instanceId;
	// } else {
	// int size = indices.size();
	// throw new IllegalStateException("This instance does not exist in this
	// set");
	// }
	//
	// } else {
	// throw new IllegalStateException("This set does not exist in the
	// repository");
	// }
	// }

	@Override
	public InstanceId getInstance(ClassId setId, int index) {
		BiMap<InstanceId, Integer> indices = instanceIndices.get(setId);
		if (indices != null) {
			InstanceId instanceId = indices.inverse().get(index);
			if (instanceId != null) {
				return instanceId;
			} else {
				int size = indices.size();
				throw new IllegalStateException("This instance does not exist in this set");
			}

		} else {
			throw new IllegalStateException("This set does not exist in the repository");
		}
	}

	@Override
	public int getInstanceIndex(ClassId setId, InstanceId instanceId) {
		BiMap<InstanceId, Integer> instances = instanceIndices.get(setId);

		if (instances != null) {
			Integer index = instances.get(instanceId);
			if (index != null) {
				return index;
			} else {
				throw new IllegalStateException("This instance does not exist in this set");
			}
		} else {
			throw new IllegalStateException("This set does not exist in the repository");
		}
	}

	@Override
	public BiMap<InstanceId, Integer> getInstances(ClassId setId) {
		BiMap<InstanceId, Integer> map = instanceIndices.get(setId);
		if (map != null) {
			return map;
		} else {
			return HashBiMap.create();
			// throw new IllegalStateException("This set does not exist in the
			// repository");
		}

	}

	@Override
	public Set<InstanceId> getInstances(ClassId setId, PropertyId propertyId) {
		Set<InstanceId> set;
		if (propertyId instanceof QualifierId) {
			set = instancesPerQualifier.get(new ClassPropertyId(setId, propertyId));
		} else {
			set = instancesPerClassProperty.get(new ClassPropertyId(setId, propertyId));
		}

		if (set != null) {
			return set;

		} else {
			return new HashSet<>();
			// throw new IllegalStateException("This set/property combination
			// does not have any instances");
		}

	}

	@Override
	public int getInstancesSize(ClassId setId, PropertyId propertyId) {
		Set<InstanceId> set;
		if (propertyId instanceof QualifierId) {
			set = instancesPerQualifier.get(new ClassPropertyId(setId, propertyId));
		} else {
			set = instancesPerClassProperty.get(new ClassPropertyId(setId, propertyId));
		}

		if (set != null) {
			return set.size();
		} else {
			throw new IllegalStateException("This set/property combination does not have any instances");
		}

	}

	@Override
	public int getNumberOfColumns(ClassId setId) {
		BiMap<PropertyId, Integer> map = propertyIndices.get(setId);

		if (map != null) {
			return map.size();
		} else {
			throw new IllegalStateException("This set does not exist in the repository");
		}

	}

	@Override
	public int getNumberOfInstances() {
		return setMappings.size();
	}

	@Override
	public int getNumberOfInstances(ClassId setId) {
		BiMap<InstanceId, Integer> map = instanceIndices.get(setId);

		if (map != null) {
			return map.size();
		} else {
			throw new IllegalStateException("This set does not exist in the repository");
		}

	}

	@Override
	public BiMap<PropertyId, Integer> getProperties(ClassId setId) {
		BiMap<PropertyId, Integer> map = propertyIndices.get(setId);
		if (map != null) {
			return map;
		} else {
			return HashBiMap.create();
		}
	}

	@Override
	public PropertyId getProperty(ClassId setId, int index) {
		BiMap<PropertyId, Integer> indices = propertyIndices.get(setId);
		if (indices != null) {
			PropertyId propertyId = indices.inverse().get(index);
			if (propertyId != null) {
				return propertyId;
			} else {
				throw new IllegalStateException("This property does not exist in this set");
			}

		} else {
			throw new IllegalStateException("This set does not exist in the repository");
		}
	}

	@Override
	public Integer getPropertyIndex(ClassId setId, PropertyId propertyId) {
		BiMap<PropertyId, Integer> indices = propertyIndices.get(setId);
		if (indices != null) {
			Integer index = indices.get(propertyId);
			if (index != null) {
				return index;
			} else {
				throw new IllegalStateException("This property does not exist in this set or at that index");
			}

		} else {
			throw new IllegalStateException("This set does not exist in the repository");
		}
	}

	@Override
	public int getTotalNumberOfClasses() {
		return this.sets.size();
	}

	@Override
	public int getTotalNumberOfColumns() {
		int count = 0;
		for (Map<PropertyId, Integer> table : propertyIndices.values()) {
			count += table.size();
		}

		return count;
	}

	@Override
	public int getTotalNumberOfRows() {
		int count = 0;
		for (Map<InstanceId, Integer> table : instanceIndices.values()) {
			count += table.size();
		}

		return count;
	}

	@Override
	public QualifiedValueList getQualifiedValueList(InstanceId instanceId, PropertyId propertyId) {
		return getQualifiedValueList(new ValueId(instanceId, propertyId));
	}

	@Override
	public QualifiedValueList getQualifiedValueList(ValueId valueId) {
		if (valueId.getPropertyId() instanceof QualifierId) {
			QualifierId qualifierId = (QualifierId) valueId.getPropertyId();

			ValueId mainValueId = new ValueId(valueId.getInstanceId(), qualifierId.getMain());
			PropertyId qualifierPropertyId = qualifierId.getQualifier();
			if (!this.values.containsKey(mainValueId)) {
				return null;
			} else {

				List<QualifiedValueHolder> claims = new ArrayList<>();
				for (QualifiedValueHolder claim : this.values.get(mainValueId).getClaims()) {
					Qualifier qualifierValue = claim.getQualifierValue(qualifierPropertyId);
					if (qualifierValue != null && qualifierValue.getQualifierValue() != null) {
						claims.add(new QualifiedAbstractObjectHolder(qualifierValue.getQualifierValue()));
					}
				}

				if (claims.isEmpty()) {
					return null;
				} else {
					QualifiedValueList qvl = new QualifiedValueList(claims);
					return qvl;
				}
			}
		} else {
			return this.values.get(valueId);
		}
	}

	public QualifiedValueList getQualifiedValueListWithEqualProperties(ValueId valueId, int recursionCount, int maxRecursion) {
		if (!setMappings.containsKey(valueId.getInstanceId())) {
			return null;
		}
		Set<ValueId> allValueIds = new HashSet<>();

		for (ClassId classId : this.setMappings.get(valueId.getInstanceId())) {
			ClassPropertyId cpid = new ClassPropertyId(classId, valueId.getPropertyId());
			if (equalProperties.containsKey(cpid)) {
				for (ClassPropertyId equalPropertyId : equalProperties.get(cpid)) {
					allValueIds.add(new ValueId(valueId.getInstanceId(), equalPropertyId.getPropertyId()));
				}
			}
		}

		allValueIds.add(valueId);

		List<QualifiedValueHolder> claims = new ArrayList<>();

		for (ValueId vid : allValueIds) {
			QualifiedValueList qvl = getQualifiedValueList(vid);
			if (qvl != null)
				claims.addAll(qvl.getClaims());
		}

		if (recursionCount < maxRecursion) {
			if (valueId.getPropertyId().getFinalPart().equals("isPartOf")) {
				Set<AbstractObject> newClaims = new HashSet<>();
				for (QualifiedValueHolder qvh : claims) {
					if (qvh.getValue().getType() == ValueType.INSTANCE_REFERENCE) {
						QualifiedValueList newQvl = getQualifiedValueListWithEqualProperties(new ValueId((InstanceId) qvh.getValue(), valueId.getPropertyId()),
								recursionCount + 1, maxRecursion);
						if (newQvl != null) {
							newClaims.addAll(newQvl.getClaims().stream().map((e) -> e.getValue()).collect(Collectors.toList()));
						}
					}
				}
				claims.addAll(newClaims.stream().map((e) -> new QualifiedAbstractObjectHolder(e)).collect(Collectors.toList()));
			}
		}
		if (!claims.isEmpty())
			return new QualifiedValueList(claims);
		else
			return null;
	}

	@Override
	public QualifiedValueList getQualifiedValueListWithEqualProperties(ValueId valueId) {
		return getQualifiedValueListWithEqualProperties(valueId, 0, 2);
	}

	@Override
	public boolean hasClass(ClassId setId) {
		return sets.contains(setId);
	}

	@Override
	public boolean hasInstance(ClassId setId, InstanceId instanceId) {
		BiMap<InstanceId, Integer> indices = instanceIndices.get(setId);
		if (indices != null) {
			return indices.containsKey(instanceId);
		} else {
			return false;
			// throw new IllegalStateException("This set does not exist in the
			// repository");
		}
	}

	@Override
	public boolean hasValue(InstanceId instanceId, PropertyId propertyId) {
		return hasValue(new ValueId(instanceId, propertyId));
	}

	@Override
	public boolean hasValue(ValueId valueId) {

		if (valueId.getPropertyId() instanceof QualifierId) {
			ClassId setId = setMappings.get(valueId.getInstanceId()).iterator().next();
			if (setId != null) {
				Set<InstanceId> instances = this.instancesPerQualifier.get(new ClassPropertyId(setId, valueId.getPropertyId()));
				if (instances != null) {
					if (instances.contains(valueId.getInstanceId())) {
						return true;

					}
				}

			}
			return false;
		} else {

			return values.containsKey(valueId);
		}
	}

	@Override
	public boolean hasProperty(ClassId setId, PropertyId propertyId) {
		BiMap<PropertyId, Integer> indices = propertyIndices.get(setId);
		if (indices != null) {
			return indices.containsKey(propertyId);
		} else {
			throw new IllegalStateException("This set does not exist in the repository");
		}
	}

	public Map<ClassId, BiMap<PropertyId, Integer>> getPropertyIndices() {
		return propertyIndices;
	}

	public Map<ClassId, BiMap<InstanceId, Integer>> getInstanceIndices() {
		return instanceIndices;
	}

	public Map<ValueId, QualifiedValueList> getValues() {
		return values;
	}

	public Map<ClassPropertyId, Set<InstanceId>> getInstancesPerClassProperty() {
		return instancesPerClassProperty;
	}

	@Override
	public Set<ClassPropertyId> getClassPropertyIds() {
		return instancesPerClassProperty.keySet();
	}

	@Override
	public Map<InstanceId, Set<ClassId>> getSetMappings() {
		return setMappings;
	}

	private transient SimpleLockableConcurrentCache<ClassId, BiMap<PropertyId, Integer>> cachedPropertiesWithQualifiers = new SimpleLockableConcurrentCache<>();

	public transient static final PropertyId startTimeQualifier = new PropertyId("P580", "StartTimeQualifier");
	public transient static final PropertyId endTimeQualifier = new PropertyId("P582", "EndTimeQualifier");
	public transient static final PropertyId pointInTime = new PropertyId("P585", "PointInTime");

	@Override
	public BiMap<PropertyId, Integer> getPropertiesWithQualifiers(ClassId setId) {

		if (propertyIndicesWithQualifiers.containsKey(setId)) {
			return cachedPropertiesWithQualifiers.getOrCreate(setId, () -> {
				Builder<PropertyId, Integer> builder = ImmutableBiMap.builder();

				int i = 0;
				BiMap<PropertyId, Integer> map = propertyIndicesWithQualifiers.get(setId);
				TreeSet<Integer> set = new TreeSet<>(map.values());
				for (Integer index : set) {
					PropertyId propertyId = map.inverse().get(index);
					if (propertyId instanceof QualifierId) {
						QualifierId qualifierId = (QualifierId) propertyId;
						if (qualifierId.getQualifier().equals(startTimeQualifier) || qualifierId.getQualifier().equals(endTimeQualifier)
								|| qualifierId.getQualifier().equals(pointInTime)) {
							builder.put(propertyId, i);
							i++;
						}
					} else {
						builder.put(propertyId, i);
						i++;
					}
				}
				return builder.build();
			});

		} else {
			throw new IllegalStateException("This set does not exist in the repository");
		}
	}

	@Override
	public List<AbstractObject> getValuesAsList(InstanceId instanceId, PropertyId propertyId) {
		QualifiedValueList qvl = this.getQualifiedValueList(new ValueId(instanceId, propertyId));

		if (qvl != null) {
			List<AbstractObject> ret = new ArrayList<>();
			for (QualifiedValueHolder claim : qvl.getClaims()) {
				ret.add(claim.getValue());
			}
			return ret;
		}
		return null;

	}

	@Override
	public List<AbstractObject> getValuesAsListWithEqualProperties(InstanceId instanceId, PropertyId propertyId) {
		QualifiedValueList qvl = this.getQualifiedValueListWithEqualProperties(new ValueId(instanceId, propertyId));

		if (qvl != null) {
			List<AbstractObject> ret = new ArrayList<>();
			for (QualifiedValueHolder claim : qvl.getClaims()) {
				ret.add(claim.getValue());
			}
			return ret;
		}
		return null;

	}

	@Override
	public Set<String> getLabels(InstanceId instanceId) {
		Set<String> ret = new HashSet<>();

		Consumer<List<AbstractObject>> consumer = (labels) -> {
			if (labels != null) {
				for (AbstractObject label : labels) {
					if (label instanceof StringObject) {
						ret.add(((StringObject) label).getString());
					}
				}
			}
		};

		if (instanceId.getReadableString().startsWith("http://dbpedia.org/resource/")) {
			ret.add(instanceId.getReadableString(instanceId.getReadableString()));
		} else {
			ret.add(instanceId.getReadableString());
		}

		List<AbstractObject> labels = getValuesAsList(instanceId, PropertiesExpansionData.LABEL_PROPERTY);
		consumer.accept(labels);

		// labels = getValuesAsList(instanceId, new
		// PropertyId("http://www.w3.org/2000/01/rdf-schema#label"));
		// consumer.accept(labels);

		labels = getValuesAsList(instanceId, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY);
		consumer.accept(labels);

		labels = getValuesAsList(instanceId, PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY);
		consumer.accept(labels);

		if (instanceId.getInstanceId().equals("http://dbpedia.org/resource/Defensive_tackle")) {
			int sdf = 123;
		}
		return Collections.unmodifiableSet(ret);

	}

	@Override
	public Map<PropertyId, Set<ClassId>> getReversePropertyIndex() {
		Map<PropertyId, Set<ClassId>> ret = new HashMap<>();

		for (Entry<ClassId, BiMap<PropertyId, Integer>> pEntry : this.propertyIndices.entrySet()) {

			for (PropertyId propertyId : pEntry.getValue().keySet()) {
				CHMHelper.getFromMap(ret, propertyId, () -> new HashSet<>()).add(pEntry.getKey());
			}
		}

		return ret;
	}

	@Override
	public Set<InstanceId> getAllInstances() {
		return this.setMappings.keySet();
	}

	@Override
	public Map<ClassPropertyId, Set<ClassPropertyId>> getEqualProperties() {
		return this.equalProperties;
	}

	@Override
	public Map<PropertyId, UnitOfMeasurement> getKbUnits() {
		return kbUnits;
	}

	@Override
	public Map<PropertyId, PropertyId> getNormalizationIndex() {
		return normalizationIndex;
	}

	@Override
	public Optional<Set<String>> getSurfaceForms(String string) {
		Set<String> sf = surfaceForms.get(string);
		if (sf != null)
			return Optional.of(sf);
		else
			return Optional.absent();
	}

	@Override
	public Optional<Set<String>> getCandidatesSurfaceForms(String string) {
		Set<String> sf = candidatesSurfaceForms.get(string);
		if (sf != null)
			return Optional.of(sf);
		else
			return Optional.absent();
	}

}
