package de.uni_mannheim.dws.expansion.data.utils.kryo;

import java.io.ByteArrayOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.zip.GZIPOutputStream;

import javax.inject.Inject;

import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveOutputStream;

import com.esotericsoftware.kryo.Kryo;
import com.esotericsoftware.kryo.io.Output;

import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;

public class ConcurrentKryoWriter {
	private final ThreadSafeNoReferenceKryo threadSafeKryo;

	private final ThreadSafeKryo threadSafeWithReferencesKryo;

	@Inject
	public ConcurrentKryoWriter(ThreadSafeKryo threadSafeWithReferencesKryo, ThreadSafeNoReferenceKryo threadSafeKryo) {
		this.threadSafeWithReferencesKryo = threadSafeWithReferencesKryo;
		this.threadSafeKryo = threadSafeKryo;
	}

	@Deprecated
	public <T> void writeAllWithReferences(T poisen, Iterable<T> data, int size, String folder, String prefix, int groupNum) {

		writeAllInner(poisen, data, size, folder, prefix, groupNum, true);

	}

	public <T> void writeAll(T poisen, Iterable<T> data, int size, String folder, String prefix, int groupNum) {

		writeAllInner(poisen, data, size, folder, prefix, groupNum, false);

	}

	private <T> void writeAllInner(T poisen, Iterable<T> data, int size, String folder, String prefix, int groupNum, boolean withReference) {

		List<Pair<Integer, List<T>>> groups = getGroups(data, size, groupNum);

		PersistentParallel6.runAndFinish(groups, (group) -> {

			Iterable<byte[]> entries = transformIntoByte(group.getSecond(), withReference);

			try {
				writeTarFile(folder, prefix, group.getFirst(), entries);
			} catch (IOException e) {
				e.printStackTrace();
			}

		}, this.getClass(), 12);

	}

	private <T> Iterable<byte[]> transformIntoByte(Iterable<T> entries, boolean withReferences) {
		ConcurrentLinkedQueue<byte[]> queue = new ConcurrentLinkedQueue<>();

		PersistentParallel6.runAndFinishWithDefaultBatchSize(entries, (e) -> {
			Kryo kryo;
			if (withReferences) {
				kryo = threadSafeWithReferencesKryo.getKryo();
			} else {
				kryo = threadSafeKryo.getKryo();
			}

			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			Output output = new Output(baos);
			kryo.writeObject(output, e);
			output.close();
			queue.add(baos.toByteArray());
		}, this.getClass());

		return queue;
	}

	private <T> List<Pair<Integer, List<T>>> getGroups(Iterable<T> entries, int totalSize, int numOfGroups) {
		int perGroup = totalSize / numOfGroups;

		List<Pair<Integer, List<T>>> groups = new ArrayList<>();

		int groupIndex = 0;
		List<T> entriesList = new ArrayList<>();

		for (T entry : entries) {
			if (entriesList.size() == perGroup && groupIndex != numOfGroups - 1) {
				groups.add(new Pair<>(groupIndex, entriesList));
				groupIndex++;
				entriesList = new ArrayList<>();
			}

			entriesList.add(entry);
		}
		groups.add(new Pair<>(groupIndex, entriesList));

		return groups;
	}

	private void writeTarFile(String folder, String prefix, int fileCount, Iterable<byte[]> entries) throws IOException {
		TarArchiveOutputStream tarFile = new TarArchiveOutputStream(new GZIPOutputStream(new FileOutputStream(folder + "/" + prefix + fileCount + ".tar.gz")));

		int entryCount = 0;

		for (byte[] byteEntry : entries) {

			TarArchiveEntry entry = new TarArchiveEntry(String.valueOf(entryCount));
			entry.setSize(byteEntry.length);

			tarFile.putArchiveEntry(entry);

			tarFile.write(byteEntry);

			tarFile.closeArchiveEntry();

			entryCount++;
		}

		tarFile.finish();

		tarFile.close();
	}

	// private class OneWritingInstance<T> {
	// private int fileCount = 0;
	// private int entryCount = 0;
	// private int groupNum = -1;
	// private int perGroup;
	// private TarArchiveOutputStream tarFile;
	// private String folder;
	// private ConsumerHelper<T> consumerHelper;
	// private String prefix;
	//
	// private OneWritingInstance(T poisen, String folder, String prefix, int
	// groupNum, int totalCount) {
	// try {
	// this.folder = folder;
	// this.prefix = prefix;
	// this.groupNum = groupNum;
	// this.perGroup = totalCount / groupNum;
	//
	// tarFile = new TarArchiveOutputStream(new GZIPOutputStream(new
	// FileOutputStream(folder + "/" + prefix + fileCount + ".tar.gz")));
	//
	// consumerHelper = ConsumerHelper.withLinkedQueue("TABLE STORER", 1, (data)
	// -> {
	//
	// try {
	// writeEntry(baos.toByteArray());
	// } catch (Exception e) {
	// // TODO Auto-generated catch block
	// e.printStackTrace();
	// }
	// }, poisen);
	//
	// consumerHelper.start();
	// } catch (IOException e) {
	// throw new RuntimeException(e);
	// }
	//
	// }
	//
	// private synchronized void writeEntry(byte[] bytes) throws IOException {
	//
	// if (entryCount == perGroup && fileCount != (groupNum - 1)) {
	// finishAndClose();
	// fileCount++;
	// tarFile = new TarArchiveOutputStream(new GZIPOutputStream(new
	// FileOutputStream(folder + "/" + prefix + fileCount + ".tar.gz")));
	// entryCount = 0;
	// }
	//
	// TarArchiveEntry entry = new TarArchiveEntry(String.valueOf(entryCount));
	// entry.setSize(bytes.length);
	//
	// tarFile.putArchiveEntry(entry);
	//
	// tarFile.write(bytes);
	//
	// tarFile.closeArchiveEntry();
	//
	// entryCount++;
	// }
	//
	// private synchronized void finishAndClose() throws IOException {
	//
	// tarFile.finish();
	//
	// tarFile.close();
	// }
	//
	// private void add(byte[] data) {
	// queue.add(data);
	// }
	//
	// private void finishCloseAndAwait() throws InterruptedException,
	// IOException {
	// consumerHelper.finishCloseAndAwait();
	// finishAndClose();
	//
	// }
	//
	// }

}
