package de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.mannheim.uni.expansion.tablefusion.core.fusing.BaseFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.FusionScoreThresholdAdapter;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.AbstractClassPair;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.ValueSortedGenericPair;

public class GenericMedianFuser<T extends AbstractObject> implements BaseFuser {

	private final TypeChecker<T> typeChecker;
	private final TypeCaster<T> typeCaster;
	private final ListSorter<T> listSorter;
	private final Averager<T> averager;
	private final Comparer<T> comparer;
	private final Medianizer<T> medianizer;

	public GenericMedianFuser(TypeChecker<T> typeChecker, TypeCaster<T> typeCaster, ListSorter<T> listSorter, Averager<T> averager, Comparer<T> comparer,
			Medianizer<T> medianizer) {
		this.typeCaster = typeCaster;
		this.typeChecker = typeChecker;
		this.listSorter = listSorter;
		this.averager = averager;
		this.comparer = comparer;
		this.medianizer = medianizer;
	}

	@FunctionalInterface
	public interface TypeChecker<T extends AbstractObject> {
		boolean check(Object o);
	}

	@FunctionalInterface
	public interface TypeCaster<T extends AbstractObject> {
		T cast(Object o);
	}

	@FunctionalInterface
	public interface ListSorter<T extends AbstractObject> {
		void sort(List<T> o);
	}

	@FunctionalInterface
	public interface Averager<T extends AbstractObject> {
		T get(T object1, double score1, T object2, double score2);
	}

	@FunctionalInterface
	public interface Comparer<T extends AbstractObject> {
		int compare(T object, T other);
	}

	@FunctionalInterface
	public interface Medianizer<T extends AbstractObject> {
		FusedValue medianize(List<ValueSortedGenericPair<T>> values, Double threshold);
	}

	public static <T extends AbstractObject> FusedValue defaultMedianizer(List<ValueSortedGenericPair<T>> values, Double threshold, Averager<T> averager) {
		Collections.sort(values);
		double sum = values.stream().mapToDouble(ValueSortedGenericPair::getScore).sum();
		double median = sum / 2.0;
		double current = 0.0;

		T value = null;

		ValueSortedGenericPair<T> lastPair = null;
		for (ValueSortedGenericPair<T> p : values) {
			current += Math.abs(p.getScore());

			if (median > current) {
				continue;
			} else {
				if (lastPair == null) {
					value = p.getValue();
					break;
				} else {
					// double diff = p.getScore();
					double s1 = (current) - median;
					double s2 = median - (current - Math.abs(p.getScore()));

					value = averager.get(p.getValue(), s1, lastPair.getValue(), s2);
				}
			}

			lastPair = p;
		}

		if (threshold == null || sum > threshold) {
			return new FusedValue(value, sum, values.size());
		} else {
			return null;
		}

	}

	public FusedValue simpleFuse(List<AbstractObject> list) {
		List<ValueSortedGenericPair<T>> values = new ArrayList<ValueSortedGenericPair<T>>();

		for (Object o : list) {

			if (typeChecker.check(o)) {
				values.add(new Pair(typeCaster.cast(o), 1.0));
			}

		}

		return medianizer.medianize(values, null);

		// if (values.size() > 0) {
		// if (values.size() % 2 != 0) {
		// int idx = (values.size() / 2);
		//
		// return new FusedValue(values.get(idx), (double) values.size());
		//
		// } else {
		// int idx1 = (values.size() / 2);
		// int idx2 = (values.size() / 2) - 1;
		//
		// return new FusedValue(averager.get(values.get(idx1), 0.5,
		// values.get(idx2), 0.5), (double) values.size());
		// }
		// } else {
		// return null;
		// }
	}

	public FusedValue fuse(List<? extends AbstractClassPair> values, Double threshold, FusionScoreThresholdAdapter adapter) {
		if (values.size() == 0) {
			return null;
		} else {
			FusedValue fv;
			if (values.size() == 1) {
				fv = new FusedValue(values.get(0).getValue(), values.get(0).getScore(), 1);

			} else {
				List<ValueSortedGenericPair<T>> valuesSet = new ArrayList<ValueSortedGenericPair<T>>();

				for (AbstractClassPair pair : values) {

					Double probability = pair.getScore();

					AbstractObject o = pair.getValue();

					if (typeChecker.check(o)) {
						valuesSet.add(new Pair(typeCaster.cast(o), probability));
					}

				}
				fv = medianizer.medianize(valuesSet, threshold);
			}

			// if (threshold == null || fv.getSingleScore() > threshold) {
			return fv;
			// } else {
			// return null;
			// }
		}

		// if (threshold == null || sum > threshold) {
		//
		// Collections.sort(pairs);
		//
		// double median = sum / 2.0;
		//
		// double current = 0.0;
		//
		// T value = null;
		//
		// Pair lastPair = null;
		// for (Pair p : pairs) {
		// current += Math.abs(p.getScore());
		//
		// if (median > current) {
		// continue;
		// } else {
		// if (lastPair == null) {
		// value = p.getValue();
		// break;
		// } else {
		// // double diff = p.getScore();
		// double s1 = (current) - median;
		// double s2 = median - (current - Math.abs(p.getScore()));
		//
		// value = averager.get(p.getValue(), s1, lastPair.getValue(), s2);
		// }
		// }
		//
		// lastPair = p;
		// }
		//
		// if (value == null)
		// return null;
		// return new FusedValue(value, sum);
		//
		// }
		// return null;
	}

	private class Pair implements ValueSortedGenericPair<T> {
		private final T value;
		private final Double score;

		public T getValue() {
			return value;
		}

		public Double getScore() {
			return score;
		}

		public Pair(T value, Double score) {
			super();
			this.value = value;
			this.score = score;
		}

		@Override
		public int compareTo(ValueSortedGenericPair<T> other) {
			return comparer.compare(this.value, other.getValue());
		}

		@Override
		public String toString() {
			return "Pair [value=" + value + ", score=" + score + "]";
		}

	}

	@Override
	public boolean checkType(AbstractObject value) {
		return typeChecker.check(value);
	}

	@Override
	public FusedValue clusteredFuse(List<AbstractObject> values) {
		return simpleFuse(values);
	}

}
