package de.uni_mannheim.informatik.dws.attribute.evaluation;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.informatik.dws.attribute.model.AttributeMappings;
import de.uni_mannheim.informatik.dws.attribute.model.PropertyMapping;
import de.uni_mannheim.informatik.dws.attribute.model.TableAttributeMappings;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance.FusionPerformanceBuilder;

public class AttributeMappingsEvaluator {

	@Inject
	public AttributeMappingsEvaluator() {

	}

	public FusionPerformance<AttributeMappingsPerformance> evaluate(AttributeMappings gs, AttributeMappings result, Map<PropertyId, PropertyId> normalization) {

		Set<TableColumnId> gsColumnSet = gs.columnSet();
		Set<TableColumnId> resultColumnSet = result.columnSet();

		int correct = 0;
		int wrong = 0;
		int maxCorrect = gsColumnSet.size();

		Map<ClassId, Integer> maxCorrectClassMap = getMaxCorrectClassMap(gs);
		Map<ClassId, MutableInt> correctClassMap = new HashMap<>();
		Map<ClassId, MutableInt> wrongClassMap = new HashMap<>();

		Map<ClassPropertyId, Integer> maxCorrectMap = getMaxCorrectMap(gs);
		Map<ClassPropertyId, MutableInt> correctMap = new HashMap<>();
		Map<ClassPropertyId, MutableInt> wrongMap = new HashMap<>();

		for (TableColumnId columnId : resultColumnSet) {
			if (!gs.tableSet().contains(columnId.getTableId()))
				continue;
			ClassId classId = result.getClassId(columnId.getTableId());

			PropertyId resultMappingNormalized = result.getMapping(columnId).get().getPropertyId();
			if (normalization.containsKey(resultMappingNormalized))
				resultMappingNormalized = normalization.get(resultMappingNormalized);

			ClassPropertyId cpid = new ClassPropertyId(classId, resultMappingNormalized);

			if (gsColumnSet.contains(columnId)) {

				if (gs.getMapping(columnId).get().getPropertyId().equals(resultMappingNormalized)) {
					CHMHelper.getFromMap(correctMap, cpid, () -> new MutableInt()).increment();
					CHMHelper.getFromMap(correctClassMap, classId, () -> new MutableInt()).increment();
					correct++;
				} else {
					CHMHelper.getFromMap(wrongMap, cpid, () -> new MutableInt()).increment();
					CHMHelper.getFromMap(wrongClassMap, classId, () -> new MutableInt()).increment();
					wrong++;
				}
			} else {
				CHMHelper.getFromMap(wrongMap, cpid, () -> new MutableInt()).increment();
				CHMHelper.getFromMap(wrongClassMap, classId, () -> new MutableInt()).increment();
				wrong++;
			}
		}

		Map<ClassPropertyId, AttributeMappingsPerformance> performanceMap = new HashMap<>();
		Set<ClassPropertyId> mapKeys = new HashSet<>();
		mapKeys.addAll(maxCorrectMap.keySet());
		mapKeys.addAll(correctMap.keySet());
		mapKeys.addAll(wrongMap.keySet());
		for (ClassPropertyId cpid : mapKeys) {
			int c = 0;
			if (correctMap.containsKey(cpid)) {
				c = correctMap.get(cpid).intValue();
			}
			int w = 0;
			if (wrongMap.containsKey(cpid)) {
				w = wrongMap.get(cpid).intValue();
			}
			int mc = 0;
			if (maxCorrectMap.containsKey(cpid)) {
				mc = maxCorrectMap.get(cpid);
			}
			performanceMap.put(cpid, new AttributeMappingsPerformance(c, w, mc));
		}

		Map<ClassId, AttributeMappingsPerformance> performanceClassMap = new HashMap<>();

		for (ClassId classId : maxCorrectClassMap.keySet()) {
			int c = 0;
			if (correctClassMap.containsKey(classId)) {
				c = correctClassMap.get(classId).intValue();
			}
			int w = 0;
			if (wrongClassMap.containsKey(classId)) {
				w = wrongClassMap.get(classId).intValue();
			}
			int mc = 0;
			if (maxCorrectClassMap.containsKey(classId)) {
				mc = maxCorrectClassMap.get(classId);
			}
			performanceClassMap.put(classId, new AttributeMappingsPerformance(c, w, mc));
		}

		FusionPerformanceBuilder<AttributeMappingsPerformance> performanceBuilder = FusionPerformance.builder();
		performanceBuilder.setTotal(new AttributeMappingsPerformance(correct, wrong, maxCorrect));
		performanceBuilder.setPerProperty(performanceMap);
		performanceBuilder.setPerClassId(performanceClassMap);
		return performanceBuilder.get();
	}

	public Map<ClassPropertyId, Integer> getMaxCorrectMap(AttributeMappings gs) {
		Map<ClassPropertyId, MutableInt> ret = new HashMap<>();

		for (Entry<AnnotatedTableId, TableAttributeMappings> entry : gs.getMappings().entrySet()) {
			ClassId classId = gs.getClassId(entry.getKey());
			for (Entry<ColumnIdx, PropertyMapping> mapping : entry.getValue().getMappings().entrySet()) {
				CHMHelper.getFromMap(ret, new ClassPropertyId(classId, mapping.getValue().getPropertyId()), () -> new MutableInt()).increment();
			}
		}

		return ret.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().intValue()));
	}

	public Map<ClassId, Integer> getMaxCorrectClassMap(AttributeMappings gs) {
		Map<ClassId, MutableInt> ret = new HashMap<>();

		for (Entry<AnnotatedTableId, TableAttributeMappings> entry : gs.getMappings().entrySet()) {
			ClassId classId = gs.getClassId(entry.getKey());
			CHMHelper.getFromMap(ret, classId, () -> new MutableInt()).add(entry.getValue().getMappings().size());

		}

		return ret.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().intValue()));
	}

}
