package de.uni_mannheim.informatik.dws.attribute.matcher;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;

public class AttributeMatchingSingleRun {

	public static final String OVERLAP_METHOD = "METHOD_WEIGHT_OVERLAP";
	public static final String DUPLICATE_METHOD = "METHOD_WEIGHT_DUPLICATE";
	public static final String KB_LABEL = "METHOD_KB_LABEL";
	public static final String SECONDARY_LABEL = "METHOD_SECONDARY_LABEL";
	public static final String SECONDARY_DUPLICATE = "METHOD_SECONDARY_DUPLICATE";

	public static final String MAX_OVERLAP_WEIGHT_PER_VALUE = "MAX_OVERLAP_WEIGHT_PER_VALUE";
	public static final String EXTENDED_MAX_SCORE = "EXTENDED_MAX_SCORE";
	public static final String DUPLICATE_MIN_WEIGHT = "DUPLICATE_MIN_WEIGHT";
	public static final String SECONDARY_DUPLICATE_MIN_WEIGHT = "SECONDARY_DUPLICATE_MIN_WEIGHT";

	public static final String getSuffixed(String variable, PropertyId propertyId) {
		return variable; // + "_" + propertyId.getPropertyId();
	}

	public static final String getThreshold(PropertyId propertyId) {
		return "THRESHOLD_" + propertyId.getPropertyId();
	}

	private final Map<PropertyId, Map<String, ParameteredMatcherScore>> scores;

	public AttributeMatchingSingleRun(Map<PropertyId, Map<String, ParameteredMatcherScore>> scores) {
		super();
		this.scores = ImmuMaker.mapWithMap(scores);
	}

	public Map<PropertyId, Map<String, ParameteredMatcherScore>> getScores() {
		return scores;
	}

	public final static Integer convertParameter(Float f) {
		return Math.round(f * 100);
	}

	public final static Integer convertParameter(Double d) {
		return convertParameter(d.floatValue());
	}

	public static class ParameteredMatcherScore {

		private final Set<String> parameters;

		private final PotentialParameteredScore scores;

		public ParameteredMatcherScore(WeightedScore score) {
			this.parameters = null;
			this.scores = new PotentialParameteredScore(score);
		}

		public ParameteredMatcherScore(String parameter1, Map<Integer, WeightedScore> scores) {

			Set<String> parameters = new LinkedHashSet<>();
			parameters.add(parameter1);

			this.parameters = Collections.unmodifiableSet(parameters);

			Map<Integer, PotentialParameteredScore> map = new HashMap<>();
			for (Entry<Integer, WeightedScore> score : scores.entrySet()) {
				map.put(score.getKey(), new PotentialParameteredScore(score.getValue()));
			}

			this.scores = new PotentialParameteredScore(map);
		}

		public ParameteredMatcherScore(String parameter1, String parameter2, Map<Pair<Integer, Integer>, WeightedScore> scores) {

			Set<String> parameters = new LinkedHashSet<>();
			parameters.add(parameter1);
			parameters.add(parameter2);

			this.parameters = Collections.unmodifiableSet(parameters);

			Map<Integer, Map<Integer, PotentialParameteredScore>> map = new HashMap<>();
			for (Entry<Pair<Integer, Integer>, WeightedScore> score : scores.entrySet()) {
				CHMHelper.getFromMap(map, score.getKey().getFirst(), () -> new HashMap<>()).put(score.getKey().getSecond(),
						new PotentialParameteredScore(score.getValue()));
			}

			this.scores = new PotentialParameteredScore(
					map.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new PotentialParameteredScore(e.getValue()))));
		}

		public Optional<Set<String>> getParameters() {
			if (parameters == null)
				return Optional.empty();
			else
				return Optional.of(parameters);
		}

		public WeightedScore getScore(Map<String, Double> valuesMap) {

			if (parameters != null && valuesMap.keySet().equals(parameters)) {
				PotentialParameteredScore currentLevel = scores;
				for (String p : parameters) {
					currentLevel = currentLevel.getSubScores(convertParameter(valuesMap.get(p)));
				}
				return currentLevel.getScore();
			}

			throw new IllegalStateException();
		}

		public WeightedScore getScore() {
			if (parameters == null) {
				return scores.getScore();
			}
			throw new IllegalStateException();
		}
	}

	public static class PotentialParameteredScore {

		private final WeightedScore score;
		private final Map<Integer, PotentialParameteredScore> scores;

		private PotentialParameteredScore(WeightedScore score) {
			this.score = score;
			this.scores = null;
		}

		private PotentialParameteredScore(Map<Integer, PotentialParameteredScore> scores) {
			this.score = null;
			this.scores = scores;
		}

		public WeightedScore getScore() {
			if (score != null) {
				return score;
			}
			throw new IllegalStateException();
		}

		public PotentialParameteredScore getSubScores(Integer value) {
			if (scores != null && scores.containsKey(value)) {
				return scores.get(value);
			}
			throw new IllegalStateException();
		}
	}

}
