package de.uni_mannheim.informatik.dws.expansion.matching;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;

import javax.inject.Named;
import javax.inject.Provider;
import javax.inject.Singleton;

import dagger.Module;
import dagger.Provides;
import de.mannheim.uni.expansion.data.DataNamedProperties;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyAdapter;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyAdapterRunner;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepositoryImpl;
import de.mannheim.uni.expansion.tablefusion.core.matching.impl.reference.MatchingDataBuilder;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.index.IIndex;
import de.mannheim.uni.index.io.DefaultIndex;
import de.mannheim.uni.matching.dbpedia.algorithm.KeyIndex;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreatorsRunner;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseCryoAdapter;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseImpl;
import de.uni_mannheim.dws.expansion.data.kb.repository.remote.KnowledgeBaseRemote;
import de.uni_mannheim.dws.expansion.data.kb.repository.remote.KnowledgeBaseRemoteImpl;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitOfMeasurement;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RemoteRegistry;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;

@Module(library = true, injects = { KnowledgeBase.class, OntologyRepository.class, KeyIndex.class, MatchingDataBuilder.class })
public class KnowledgeBaseModule {

	private final String knowledgeBasefolder;

	private final String ontologyRepositoryFile;

	private final String luceneIndexCache;
	private final RepositoryLoading repositoryLoading;

	private List<KBValueCreator> valueCreatorList;

	private Function<OntologyRepositoryImpl, OntologyRepositoryImpl> ontologyReplacer;

	private Map<ClassPropertyId, Set<ClassPropertyId>> equalProperties = new HashMap<>();

	private Map<PropertyId, UnitOfMeasurement> kbUnits = new HashMap<>();

	private Map<PropertyId, PropertyId> normalizationIndex = new HashMap<>();

	public KnowledgeBaseModule(String knowledgeBasefolder, String ontologyRepositoryFile, String luceneIndexCache) {
		this(knowledgeBasefolder, ontologyRepositoryFile, luceneIndexCache, RepositoryLoading.LOCAL_ONLY);
	}

	public KnowledgeBaseModule(String knowledgeBasefolder, String ontologyRepositoryFile, String luceneIndexCache, RepositoryLoading repositoryLoading) {
		this.knowledgeBasefolder = knowledgeBasefolder;
		this.ontologyRepositoryFile = ontologyRepositoryFile;
		this.luceneIndexCache = luceneIndexCache;
		this.valueCreatorList = null;
		this.repositoryLoading = repositoryLoading;
	}

	public void setValueCreatorList(List<KBValueCreator> valueCreatorList) {
		this.valueCreatorList = valueCreatorList;
	}

	public void setOntologyReplacer(Function<OntologyRepositoryImpl, OntologyRepositoryImpl> ontologyReplacer) {
		this.ontologyReplacer = ontologyReplacer;
	}

	public void setEqualProperties(Map<ClassPropertyId, Set<ClassPropertyId>> equalProperties) {
		this.equalProperties = equalProperties;
	}

	private KnowledgeBaseImpl kb = null;
	private OntologyRepository ontologyRepository = null;

	private List<OntologyAdapter> ontologyAdaptersList = new ArrayList();

	public void setOntologyAdaptersList(List<OntologyAdapter> ontologyAdaptersList) {
		this.ontologyAdaptersList = ontologyAdaptersList;
	}

	private final static String registryKey = "KnowledgeBaseRemoteImpl";

	public static boolean CACHE_KNOWLEDGE_BASE_STATICALLY = false;
	private static KnowledgeBaseImpl kbCached = null;

	@Provides
	@Singleton
	public KnowledgeBase getKnowledgeBase(KBValueCreatorsRunner kbValueCreatorsRunner, KnowledgeBaseCryoAdapter kbAdapter) {
		try {
			if (CACHE_KNOWLEDGE_BASE_STATICALLY && kbCached != null) {
				return kbCached;
			}
			if (kb == null) {
				boolean isInRegistry = true;
				if (repositoryLoading == RepositoryLoading.REMOTE_LOAD_REMOTELY) {
					try {
						Registry registry = RemoteRegistry.getReadingRegistry();
						isInRegistry = Arrays.asList(registry.list()).contains(registryKey);
					} catch (Exception e) {
						isInRegistry = false;
					}
				}

				if (repositoryLoading == RepositoryLoading.LOCAL_ONLY || repositoryLoading == RepositoryLoading.REMOTE_LOAD_LOCALLY || !isInRegistry) {
					// kb = null;
					// attempt to read from cache
					if (kbAdapter.cacheExists(knowledgeBasefolder)) {

						kb = kbValueCreatorsRunner.runReplacers(kbAdapter.read(knowledgeBasefolder, equalProperties, kbUnits, normalizationIndex),
								valueCreatorList);

						if (repositoryLoading == RepositoryLoading.REMOTE_LOAD_LOCALLY) {
							KnowledgeBaseRemote stub = (KnowledgeBaseRemote) UnicastRemoteObject.exportObject(kb, 0);
							Registry registry = RemoteRegistry.getPublishingRegistry();
							// Bind the remote object's stub in the registry

							registry.bind("KnowledgeBaseRemoteImpl", stub);
						}

					} else {
						throw new IllegalStateException("KB FOLDER DOES NOT EXIST");
					}
				} else if (repositoryLoading == RepositoryLoading.REMOTE_LOAD_REMOTELY) {
					try {
						Registry registry = RemoteRegistry.getReadingRegistry();
						KnowledgeBaseRemote stub = (KnowledgeBaseRemote) registry.lookup("KnowledgeBaseRemoteImpl");
						return new KnowledgeBaseRemoteImpl(stub);
					} catch (Exception e) {
						System.err.println("Client exception: " + e.toString());
						e.printStackTrace();
						throw new RuntimeException(e);
					}
				} else {
					throw new IllegalStateException("illegal value for repository loading");
				}
			} else if (this.valueCreatorList != null && !this.valueCreatorList.isEmpty()) {
				kb = kbValueCreatorsRunner.runReplacers(kb, valueCreatorList);
			}

			if (CACHE_KNOWLEDGE_BASE_STATICALLY) {
				kbCached = kb;
			}
			return kb;
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);

		}

	}

	@Provides
	@Singleton
	public OntologyRepository getOntologyRepository(ThreadSafeKryo threadSafeKryo, OntologyAdapterRunner adapterRunner) {

		try {
			if (ontologyRepository == null) {
				Path file = Paths.get(ontologyRepositoryFile);

				if (Files.exists(file)) {
					OntologyRepositoryImpl ontologyRepository = (OntologyRepositoryImpl) threadSafeKryo.read(ontologyRepositoryFile);
					if (ontologyReplacer != null) {
						ontologyRepository = ontologyReplacer.apply(ontologyRepository);
					}

					if (ontologyAdaptersList != null && !ontologyAdaptersList.isEmpty()) {
						ontologyRepository = adapterRunner.run(ontologyRepository, ontologyAdaptersList);
					}

					this.ontologyRepository = ontologyRepository;
				} else {
					throw new IllegalStateException("ONTOLOGY FILER DOES NOT EXIST");
				}

			} else if (ontologyAdaptersList != null && !ontologyAdaptersList.isEmpty()) {
				ontologyRepository = adapterRunner.run((OntologyRepositoryImpl) ontologyRepository, ontologyAdaptersList);
			}

			return this.ontologyRepository;
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			throw new IllegalStateException(e);
		}

	}

	public static boolean CACHE_KEY_INDEX_STATICALLY = false;
	private static KeyIndex keyIndexCache = null;

	@Provides
	@Singleton
	public KeyIndex getKeyIndex(Provider<KnowledgeBase> kbProvider, IndexBuilder indexBuidler) {

		if (CACHE_KEY_INDEX_STATICALLY && keyIndexCache != null) {
			return keyIndexCache;
		} else {
			try {
				Path dir = Paths.get(this.luceneIndexCache);
				MKDIR.createDirUnchecked(new File(this.luceneIndexCache));
				if (Files.exists(dir)) {
					File dirFile = new File(this.luceneIndexCache);

					if (dirFile.isDirectory()) {

						IIndex index;
						if (dirFile.list().length > 0) {
							index = new DefaultIndex(luceneIndexCache);

						} else {

							index = indexBuidler.create(luceneIndexCache, kbProvider.get());

						}

						index.getIndexSearcher();
						KeyIndex keyIndex = new KeyIndex();
						keyIndex.setLuceneIndex(index);
						if (CACHE_KEY_INDEX_STATICALLY) {
							keyIndexCache = keyIndex;
						}
						return keyIndex;
					} else {
						throw new IllegalStateException("LUCENE FOLDER IS NOT A FOLDER");
					}

				} else {
					throw new IllegalStateException("LUCENE FOLDER DOES NOT EXIST");
				}
			} catch (IOException e) {
				e.printStackTrace();
				throw new IllegalStateException(e);
			}
		}

	}

	@Provides
	@Singleton
	@Named(DataNamedProperties.DBPEDIA_MAPPINGS)
	Map<InstanceId, InstanceId> getDbPediaMappings() {
		return null;
	}

	public void setKb(KnowledgeBaseImpl kb) {
		this.kb = kb;
	}

	public void setOntologyRepository(OntologyRepository ontologyRepository) {
		this.ontologyRepository = ontologyRepository;
	}

	public void setKbUnits(Map<PropertyId, UnitOfMeasurement> kbUnits) {
		this.kbUnits = kbUnits;

	}

	public void setNormalizationIndex(Map<PropertyId, PropertyId> normalizationIndex) {
		this.normalizationIndex = normalizationIndex;

	}

}
