package de.uni_mannheim.informatik.dws.expansion.matching.testing;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;

import com.google.common.collect.ImmutableMap;
import com.google.gson.JsonSyntaxException;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.tableprocessor.model.json.AnnotatedTable;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.expansion.matching.testing.cases.MatchingTestCaseInterface;

public class TestingFramework {

	private final Map<String, MatchingTestCaseInterface> testCases;

	private final KnowledgeBase knowledgeBase;
	private final OntologyRepository ontologyRepository;

	public TestingFramework(Map<String, MatchingTestCaseInterface> testCases, KnowledgeBase knowledgeBase, OntologyRepository ontologyRepository) {
		super();
		this.testCases = ImmutableMap.copyOf(testCases);
		this.knowledgeBase = knowledgeBase;
		this.ontologyRepository = ontologyRepository;
	}

	public void createTestCases(Set<ClassId> classIds, String testCasesFolder) {
		try {
			for (Entry<String, MatchingTestCaseInterface> testCase : testCases.entrySet()) {
				String directory = testCasesFolder + testCase.getKey();
				if (MKDIR.createDirBoolean(new File(directory))) {

					for (ClassId setId : classIds) {
						String setDirectory = testCasesFolder + testCase.getKey() + "/" + setId.getFinalPart();
						if (MKDIR.createDirBoolean(new File(setDirectory))) {
							for (AnnotatedTable annotatedTable : testCase.getValue().create(knowledgeBase, ontologyRepository, setId)) {

								annotatedTable.writeJson(new File(setDirectory + "/" + UUID.randomUUID().toString() + ".json"));

							}
						} else {
							throw new RuntimeException("Unable to create Directory!");
						}

					}

				} else {
					throw new RuntimeException("Unable to create Directory!");
				}

			}

		} catch (IOException e) {
			e.printStackTrace();
			throw new RuntimeException(e);
		}
	}

	public void runTestCases(String testCasesFolder) {
		try {
			File folder = new File(testCasesFolder + "/input");

			String outputDirectory = testCasesFolder + "/output";

			File[] listOfFiles = folder.listFiles();

			for (File testCaseFolder : listOfFiles) {
				if (testCaseFolder.isDirectory()) {
					String testCaseName = testCaseFolder.getName();

					if (this.testCases.containsKey(testCaseName)) {
						MatchingTestCaseInterface testCase = this.testCases.get(testCaseName);

						if (!java.nio.file.Files.exists(Paths.get(outputDirectory))) {
							throw new IllegalStateException("The results direcotry does not exist for this test case: " + testCaseName);
						}

						if (!Files.isDirectory(Paths.get(outputDirectory))) {
							throw new IllegalStateException("The results directory is a file for this test case: " + testCaseName);

						}

						List<String> tableNames = new ArrayList<>();
						List<Pair<AnnotatedTable, AnnotatedTable>> generatedAndTestingTables = new ArrayList<>();
						// List<AnnotatedTable> tables = new ArrayList<>();
						// List<AnnotatedTable> generatedTables = new
						// ArrayList<>();

						for (File setDirectory : testCaseFolder.listFiles()) {
							if (setDirectory.isDirectory()) {
								for (File testFile : setDirectory.listFiles()) {

									if (testFile.isFile()) {

										String testName = testFile.getName();
										tableNames.add(testName);
										AnnotatedTable testingTable = AnnotatedTable.fromJson(testFile);

										String resultsFile = outputDirectory + "/" + testName;
										if (!java.nio.file.Files.exists(Paths.get(resultsFile))) {
											System.out.println("This test does not have a result mappings: " + testName);
											continue;
											// throw new
											// IllegalStateException("This test
											// does not have a result mappings:
											// " + testName);
										}
										AnnotatedTable generatedTable = AnnotatedTable.fromJson(new File(resultsFile));

										generatedAndTestingTables.add(new Pair<>(testingTable, generatedTable));

									}

								}
							}
						}

						Map<String, List<Double>> results = testCase.evaluate(knowledgeBase, ontologyRepository, generatedAndTestingTables);

						for (Entry<String, List<Double>> resultEntry : results.entrySet()) {
							double sum = 0;
							double count = 0;

							for (Double score : resultEntry.getValue()) {
								if (score != null) {
									count++;

									sum += score;
								}
							}

							System.out.println();
							System.out.println();
							System.out.println("=========================");
							System.out.println("=========================");
							System.out.println("TestCase: " + testCaseName);
							System.out.println("TestName: " + resultEntry.getKey());
							System.out.println("Success: " + sum / count);
							// + " (" + success / ((double) total) + ")"
							// System.out.println("Failed: " + (total - success)
							// + " (" + (total - success) / ((double) total) +
							// ")");
							System.out.println("Total: " + count);
							System.out.println("=========================");
							// if (success < total) {
							System.out.println("FailedTables: ");
							int tableCount = 0;
							for (Double score : resultEntry.getValue()) {
								// if (Boolean.FALSE.equals(bool)) {
								if (score != null) {
									System.out.println("- " + tableNames.get(tableCount) + " - " + score);
								}
								// }
								tableCount++;
							}
							System.out.println("=========================");
							// }
							System.out.println();
							System.out.println();

						}

					} else {
						throw new IllegalStateException("This Test case does not exist: " + testCaseName);
					}
				}
			}

		} catch (JsonSyntaxException | IOException e) {
			throw new RuntimeException(e);
		}

	}

}
