package de.uni_mannheim.informatik.dws.matching;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.mannheim.uni.index.dbpedia.DBPediaInstanceIndex;
import de.mannheim.uni.index.dbpedia.DBpediaIndexEntry;
import de.mannheim.uni.matching.dbpedia.algorithm.KeyIndex;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.matching.IndexBuilder;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;

public class LabelMatcher {
	private final DBPediaInstanceIndex idx;
	private final SimilarityCalculator sc;

	private final KnowledgeBase knowledgeBase;

	public static enum MatchStrictness {
		STRICT, NON_STRICT;
	}

	@Inject
	public LabelMatcher(KnowledgeBase knowledgeBase, KeyIndex index, SimilarityCalculator sc) {
		this.knowledgeBase = knowledgeBase;
		this.idx = new DBPediaInstanceIndex(index.getLuceneIndex(), DBpediaIndexEntry.LABEL_FIELD);
		this.idx.setMaxEditDistance(2);
		this.sc = sc;

	}

	public Optional<Pair<InstanceId, Double>> getMaxMatch(String label, ClassId classId, MatchStrictness strictness) {
		Set<String> labels = new HashSet<>();
		labels.add(label);
		Set<ClassId> classIds = new HashSet<>();
		classIds.add(classId);
		return getMaxMatch(labels, classIds, strictness, true);
	}

	public Optional<Pair<InstanceId, Double>> getMaxMatch(String label, Set<ClassId> classIds, MatchStrictness strictness) {
		Set<String> labels = new HashSet<>();
		labels.add(label);
		return getMaxMatch(labels, classIds, strictness, true);

	}

	public Optional<Pair<InstanceId, Double>> getMaxMatch(Set<String> labelsInput, ClassId classId, MatchStrictness strictness) {
		Set<ClassId> classIds = new HashSet<>();
		classIds.add(classId);
		return getMaxMatch(labelsInput, classIds, strictness, true);
	}

	public Optional<Pair<InstanceId, Double>> getMaxMatch(Set<String> labelsInput, Set<ClassId> classIds, MatchStrictness strictness,
			boolean removeAfterCommaAndRemoveBrackets) {

		SimilarityMeasure sm;
		if (strictness == MatchStrictness.STRICT) {
			sm = sc.getStrictSimilarity();
		} else {
			sm = sc.getMeasure(ValueType.STRING);
		}

		double maxSimilarity = 0.0;
		InstanceId maxMatch = null;

		for (String labelEntry : labelsInput) {
			List<DBpediaIndexEntry> matches = idx.search(labelEntry);
			for (DBpediaIndexEntry entry : matches) {
				InstanceId instanceId = new InstanceId((String) entry.getUri(), (String) entry.getUri());
				if (classIds != null) {

					Set<ClassId> classes = knowledgeBase.getClasses(instanceId);

					if (classes != null && classOverlapExists(classes, classIds)) {
						Set<String> labels = knowledgeBase.getLabels(instanceId);
						for (String kBLabel : labels) {
							double similarity = sm.get(new StringObject(labelEntry), new StringObject(kBLabel));
							if (maxMatch == null || similarity > maxSimilarity) {
								maxMatch = instanceId;
								maxSimilarity = similarity;
							}
						}
					}
				} else {
					Set<String> labels = knowledgeBase.getLabels(instanceId);
					labels = StringNormalizer.removeAfterCommaAndRemoveBrackets(labels);
					for (String kBLabel : labels) {
						double similarity = sm.get(new StringObject(labelEntry), new StringObject(kBLabel));
						if (maxMatch == null || similarity > maxSimilarity) {
							maxMatch = instanceId;
							maxSimilarity = similarity;
						}
					}
				}

			}
		}
		if (maxMatch != null && maxSimilarity > sm.getEquivalenceThreshold()) {
			return Optional.of(new Pair<>(maxMatch, maxSimilarity));
		}
		return Optional.empty();
	}

	public Optional<Map<InstanceId, Tri<Double, Double, Double>>> getMatchesFull(Set<String> labelsInput, ClassId classId, MatchStrictness strictness) {
		Set<ClassId> classIds = new HashSet<>();
		classIds.add(classId);
		return getMatchesFull(labelsInput, classIds, strictness, true);
	}

	public Optional<Map<InstanceId, Tri<Double, Double, Double>>> getMatchesFull(Set<String> labelsInput, Set<ClassId> classIds, MatchStrictness strictness) {

		return getMatchesFull(labelsInput, classIds, strictness, true);
	}

	public static final int MAX_STRING_LENGTH = 200;

	public Optional<Map<InstanceId, Double>> getMatches(Set<String> labelsInput, ClassId classId, MatchStrictness strictness) {
		Set<ClassId> classIds = new HashSet<>();
		classIds.add(classId);
		return getMatches(labelsInput, classIds, strictness, true);
	}

	public Optional<Map<InstanceId, Double>> getMatches(Set<String> labelsInput, Set<ClassId> classIds, MatchStrictness strictness) {

		return getMatches(labelsInput, classIds, strictness, true);
	}

	public Optional<Map<InstanceId, Double>> getMatches(Set<String> labelsInput, Set<ClassId> classIds, MatchStrictness strictness,
			boolean removeAfterCommaAndRemoveBrackets) {

		Optional<Map<InstanceId, Tri<Double, Double, Double>>> matches = getMatchesFull(labelsInput, classIds, strictness, true);
		if (matches.isPresent()) {
			return Optional.of(matches.get().entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getA())));
		}
		return Optional.empty();

	}

	public Optional<Map<InstanceId, Tri<Double, Double, Double>>> getMatchesFull(Set<String> labelsInput, Set<ClassId> classIds, MatchStrictness strictness,
			boolean removeAfterCommaAndRemoveBrackets) {
		double classIdsSize = classIds.size();

		SimilarityMeasure sm;
		if (strictness == MatchStrictness.STRICT) {
			sm = sc.getStrictSimilarity();
		} else {
			sm = sc.getMeasure(ValueType.STRING);
		}

		SimilarityMeasure strictSM = sc.getStrictSimilarity();
		;
		double threshold = sm.getEquivalenceThreshold();
		Map<InstanceId, Tri<Double, Double, Double>> ret = new HashMap<>();

		for (String labelEntry : labelsInput) {

			if (labelEntry.length() > MAX_STRING_LENGTH) {
				System.err.println("INPUT LABEL EXCEEDS MAX STRING LENGTH, LENGTH IS: " + labelEntry.length());
				continue;
			}

			StringObject soLabelEntry = new StringObject(labelEntry);
			List<DBpediaIndexEntry> matches = idx.search(labelEntry);
			if (matches.isEmpty()) {
				int sdf = 123;
			}

			Set<String> alreadyExists = new HashSet<>();
			for (Iterator<DBpediaIndexEntry> it = matches.iterator(); it.hasNext();) {
				String uri = it.next().getUri();
				if (!alreadyExists.contains(uri)) {
					alreadyExists.add(uri);
				} else {
					it.remove();
				}
			}

			if (labelEntry.equals("onofre marimon")) {
				int sdf = 123;
			}
			for (DBpediaIndexEntry entry : matches) {
				InstanceId instanceId = new InstanceId((String) entry.getUri(), (String) entry.getUri());

				if (entry.getUri().equals("http://dbpedia.org/resource/Aliens_&_Rainbows")) {
					int sdf = 123;
				}

				if (entry.getUri().equals("http://dbpedia.org/resource/New_Delhi")) {
					int sdf = 123;
				}

				if (entry.getUri().equals("http://dbpedia.org/resource/Boulogne-Billancourt")) {
					int sdf = 123;
				}

				if (entry.getUri().equals("http://dbpedia.org/resource/Heilbronn_(district)")) {
					int sdf = 123;
				}

				int classOverlap = getClassOverlap(instanceId, classIds);

				if (classOverlap > 0) {
					double maxSimilarity = 0.0;
					double maxNormalSimilarity = 0.0;
					InstanceId maxMatch = null;
					Set<String> labels = getLabels(instanceId, removeAfterCommaAndRemoveBrackets, labelEntry);
					String mainLabel = instanceId.getReadableString();

					if (mainLabel.startsWith("http://dbpedia.org/resource/")) {
						mainLabel = instanceId.getReadableString(instanceId.getReadableString());
					}

					mainLabel = StringNormalizer.normaliseValue(mainLabel, false);
					double mainLabelSimilarity = sm.get(soLabelEntry, new StringObject(mainLabel));
					labels.add(mainLabel);
					for (String kBLabel : labels) {
						StringObject soKBLabel = new StringObject(kBLabel);
						double similarity = sm.get(soLabelEntry, soKBLabel);
						double strictSimilarity = strictSM.get(soLabelEntry, soKBLabel);
						if (similarity <= sm.getEquivalenceThreshold()) {
							int sdf = 123;
						}
						double penalty = 0.0;

						// double penalty = (mainLabel.equals(kBLabel) ? 0.0 :
						// -0.1);

						if (similarity > sm.getEquivalenceThreshold() / 2.0 && (maxMatch == null || strictSimilarity + penalty > maxSimilarity)) {
							maxMatch = instanceId;
							maxSimilarity = strictSimilarity + penalty;
						}
						if (similarity + penalty > maxNormalSimilarity) {
							maxNormalSimilarity = similarity + penalty;
						}
					}

					double classSimilarity = getClassSimilarity(classOverlap, classIdsSize);
					if (maxMatch != null) {
						double sim = (maxSimilarity * 2.0 + classSimilarity) / 3.0;
						sim = maxSimilarity * 1.25 + classSimilarity * 0.75;
						if (!ret.containsKey(maxMatch))
							ret.put(maxMatch, new Tri<>(mainLabelSimilarity, maxNormalSimilarity, classSimilarity));
						else {
							if (ret.get(maxMatch).getB() < maxNormalSimilarity) {
								Tri<Double, Double, Double> tri = ret.get(maxMatch);

								ret.put(maxMatch, new Tri<>(tri.getA(), maxNormalSimilarity, tri.getC()));
							}
							if (ret.get(maxMatch).getA() < mainLabelSimilarity) {
								Tri<Double, Double, Double> tri = ret.get(maxMatch);

								ret.put(maxMatch, new Tri<>(mainLabelSimilarity, tri.getB(), tri.getC()));
							}
						}
					}
				}

			}
		}
		if (!ret.isEmpty()) {
			return Optional.of(ret);
		}
		return Optional.empty();
	}

	public double getLabelSim(MatchStrictness strictness, Set<String> labelsInput, InstanceId instanceId, boolean removeAfterCommaAndRemoveBrackets,
			boolean mainLabelFlag) {

		SimilarityMeasure sm;
		if (strictness == MatchStrictness.STRICT) {
			sm = sc.getStrictSimilarity();
		} else {
			sm = sc.getMeasure(ValueType.STRING);
		}

		double maxSimilarity = 0.0;
		InstanceId maxMatch = null;

		String mainLabel = instanceId.getReadableString();

		if (mainLabel.startsWith("http://dbpedia.org/resource/")) {
			mainLabel = instanceId.getReadableString(instanceId.getReadableString());
		}

		mainLabel = StringNormalizer.normaliseValue(mainLabel, false);
		SimilarityMeasure strictSM = sc.getStrictSimilarity();

		for (String labelEntry : labelsInput) {

			if (labelEntry.length() > MAX_STRING_LENGTH) {
				System.err.println("INPUT LABEL EXCEEDS MAX STRING LENGTH, LENGTH IS: " + labelEntry.length());
				continue;
			}

			StringObject soLabelEntry = new StringObject(labelEntry);

			if (mainLabelFlag) {

				double similarity = sm.get(soLabelEntry, new StringObject(mainLabel));

				if (similarity > maxSimilarity) {
					maxSimilarity = similarity;
				}
			} else {
				Set<String> labels = getLabels(instanceId, removeAfterCommaAndRemoveBrackets, labelEntry);

				for (String kBLabel : labels) {
					StringObject soKBLabel = new StringObject(kBLabel);
					double similarity = sm.get(soLabelEntry, soKBLabel);
					double strictSimilarity = strictSM.get(soLabelEntry, soKBLabel);

					if (similarity > maxSimilarity) {
						maxSimilarity = similarity;
					}

					if (similarity <= sm.getEquivalenceThreshold()) {
						int sdf = 123;
					}
					// if (similarity > sm.getEquivalenceThreshold() / 2.0 &&
					// (maxMatch == null || strictSimilarity > maxSimilarity)) {
					// maxMatch = instanceId;
					// maxSimilarity = strictSimilarity;
					// }
				}
			}

		}
		return maxSimilarity;
	}

	private Set<String> getLabels(InstanceId instanceId, boolean removeAfterCommaAndRemoveBrackets, String labelEntry) {
		Set<String> labels = new HashSet<>(knowledgeBase.getLabels(instanceId));
		if (removeAfterCommaAndRemoveBrackets) {
			labels.addAll(StringNormalizer.removeAfterCommaAndRemoveBrackets(labels));
		}
		for (Iterator<String> it = labels.iterator(); it.hasNext();) {
			int length = it.next().length();
			if (length > MAX_STRING_LENGTH) {

				System.err.println("KB LABEL EXCEEDS MAX STRING LENGTH, LENGTH IS: " + length);
				it.remove();

			}
		}
		Set<String> moreLabels = new HashSet<>();
		for (String label : labels) {

			Optional<String> concatenatedLabel = IndexBuilder.concatenate(label);
			if (concatenatedLabel.isPresent()) {
				moreLabels.add(concatenatedLabel.get());

				Optional<String> deAccented = IndexBuilder.deAccent(concatenatedLabel.get());
				if (deAccented.isPresent()) {
					moreLabels.add(deAccented.get());

				}
			}

			Optional<String> changedLabel = IndexBuilder.lettersCombiner(StringNormalizer.normaliseValue(label, false));
			if (changedLabel.isPresent()) {
				moreLabels.add(changedLabel.get());

				Optional<String> deAccented = IndexBuilder.deAccent(changedLabel.get());
				if (deAccented.isPresent()) {
					moreLabels.add(deAccented.get());

				}
			}

			Optional<String> deAccented = IndexBuilder.deAccent(label);
			if (deAccented.isPresent()) {
				moreLabels.add(deAccented.get());

			}
		}
		labels.addAll(moreLabels);
		return labels;
	}

	public int getClassOverlap(InstanceId instanceId, Set<ClassId> classIds) {
		Set<ClassId> classes = knowledgeBase.getClasses(instanceId);
		if (classes == null) {
			classes = new HashSet<>();
		}
		int classOverlap = classOverlap(classes, classIds);
		return classOverlap;
	}

	public double getClassSimiliarity(InstanceId instanceId, Set<ClassId> classIds) {
		double classIdsSize = classIds.size();
		return getClassSimilarity(getClassOverlap(instanceId, classIds), classIdsSize);
	}

	public double getClassSimilarity(int classOverlap, double classIdsSize) {
		double classSimilarity = classOverlap / classIdsSize;

		return classSimilarity;
	}

	public boolean classOverlapExists(Set<ClassId> set1, Set<ClassId> set2) {
		if (set1 == null || set2 == null) {
			int sdf = 123;
		}
		for (ClassId classId : set1) {
			if (set2.contains(classId)) {
				return true;
			}
		}
		return false;
	}

	public int classOverlap(Set<ClassId> set1, Set<ClassId> set2) {
		int overlap = 0;
		for (ClassId classId : set1) {
			if (set2.contains(classId)) {
				overlap++;
			}
		}
		return overlap;
	}

}
