package de.uni_mannheim.informatik.dws.matching;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.mannheim.uni.expansion.data.dbpedia.PropertiesDBpediaTableFilter;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.index.io.DefaultIndex;
import de.mannheim.uni.matching.dbpedia.Preprocessing;
import de.mannheim.uni.matching.dbpedia.algorithm.InstanceMatchingTask;
import de.mannheim.uni.matching.dbpedia.algorithm.KeyIndex;
import de.mannheim.uni.matching.dbpedia.algorithm.WebTableToDBpediaComponentProcess;
import de.mannheim.uni.matching.dbpedia.algorithm.WebtableToDBpediaMatchingProcess;
import de.mannheim.uni.matching.dbpedia.model.EvaluationParameters;
import de.mannheim.uni.matching.dbpedia.model.MatchingData;
import de.mannheim.uni.matching.dbpedia.model.MatchingParameters;
import de.mannheim.uni.matching.dbpedia.model.MatchingResult;
import de.mannheim.uni.matching.process.Configuration;
import de.mannheim.uni.similarity.functions.QuadraticSimilarityDecorator;
import de.mannheim.uni.similarity.functions.date.WeightedDatePartSimilarity;
import de.mannheim.uni.similarity.functions.numeric.DeviationSimilarity;
import de.mannheim.uni.similarity.functions.numeric.NormalisedNumericSimilarity;
import de.mannheim.uni.similarity.functions.string.GeneralisedStringJaccard;
import de.mannheim.uni.similarity.functions.string.JaccardOnNGramsSimilarity;
import de.mannheim.uni.similarity.functions.string.LevenshteinSimilarity;
import de.mannheim.uni.units.UnitParser;
import de.mannheim.uni.utils.concurrent.Parallel;
import de.mannheim.uni.utils.timer.Timer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseAdapter;

/**
 *
 * @author domi
 */
public class Match {

	public static void main(String[] args) throws IOException {
		TableCacheSetter.setCaches();

		String dataHomeTemp = null;

		if (args.length > 0 && args[0] != null && !args[0].equals("")) {
			dataHomeTemp = args[0];
		}

		String kbName = null;

		if (args.length > 1 && args[1] != null && !args[1].equals("")) {
			kbName = args[1];
		}

		String testName = null;

		if (args.length > 2 && args[2] != null && !args[2].equals("")) {
			testName = args[2];
		}

		String dataHome = dataHomeTemp;

		String kbHome = dataHome + "/kbs/" + kbName + "/";
		String kbPath = kbHome + "/kb/";
		String ontologyRepositoryFile = kbHome + "/ontology.kryo";
		String luceneIndexCache = kbHome + "/lucene/";

		String tablePath = dataHome + "/tests/" + testName + "/input/";
		String outputFolder = dataHome + "/tests/" + testName + "/output/";
		String config = dataHome + "/conf.config";

		String emptyFile = dataHome + "/empty.txt";
		// String unitPath = dataHome + "/empty.txt";

		UnitParser.unitPath = dataHome + "/Units/Convertible";

		UnitParser.readInUnits();

		List<KBValueCreator> valueCreatorList = new ArrayList<>();
		valueCreatorList.add(new PropertiesDBpediaTableFilter());

		KnowledgeBaseAdapter adapter = new KnowledgeBaseAdapter(kbPath, ontologyRepositoryFile, luceneIndexCache, valueCreatorList);

		Set<ClassId> allowedClasses = new HashSet<>();
		KnowledgeBase kb = adapter.getKnowledgeBase();

		for (ClassId setId : kb.getClasses()) {
			if (kb.getProperties(setId).size() > 1) {
				allowedClasses.add(setId);
			}
		}

		MatchingData matchingData = adapter.getMatchingData(allowedClasses);

		KeyIndex index = adapter.getLuceneIndex();

		MatchFull.setSystemOut();

		run(matchingData, index, luceneIndexCache, config, tablePath, outputFolder, emptyFile);

		try {
			PersistentParallel6.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public static void run(MatchingData matchingData, KeyIndex index, String luceneIndexCache, String config, String tablePath, String outputFolder,
			String emptyFile) {
		try {
			final InstanceMatchingTask m = new InstanceMatchingTask();

			Timer mainTimer = new Timer("Matching");
			String version = "";

			MatchingParameters loadDbpParams = new MatchingParameters();
			MatchingParameters params = new MatchingParameters();
			EvaluationParameters eval = new EvaluationParameters();

			m.setMatchingParameters(loadDbpParams);
			m.setData(matchingData.shallowCopy());

			if (new File(luceneIndexCache).exists()) {
				m.setLuceneIndex(new DefaultIndex(luceneIndexCache));
			}
			Preprocessing.loadSurfaceForms(emptyFile);
			Preprocessing.loadRedirects(emptyFile);
			m.getData().setEquivalentProperties(new HashMap<>());

			m.setMatchingParameters(params);

			Map<String, Object> aliases = new HashMap<>();
			aliases.put("QuadraticDeviation", new QuadraticSimilarityDecorator<Double>(new DeviationSimilarity()));
			aliases.put("QuadraticNormalized", new QuadraticSimilarityDecorator<Double>(new NormalisedNumericSimilarity()));
			aliases.put("GeneralizedLevenshtein", new GeneralisedStringJaccard(new LevenshteinSimilarity(), 0.5, 0.5));
			aliases.put("Jaccard2Grams", new JaccardOnNGramsSimilarity(2));
			aliases.put("Jaccard3Grams", new JaccardOnNGramsSimilarity(3));
			aliases.put("Jaccard4Grams", new JaccardOnNGramsSimilarity(4));
			aliases.put("Jaccard5Grams", new JaccardOnNGramsSimilarity(5));
			aliases.put("WeightedDate", new WeightedDatePartSimilarity(1, 3, 5));
			aliases.put("lodtable", WebtableToDBpediaMatchingProcess.tableType.lodtable);
			aliases.put("webtable", WebtableToDBpediaMatchingProcess.tableType.webtable);
			aliases.put("jsonWebTable", WebtableToDBpediaMatchingProcess.tableType.jsonWebTable);
			Configuration conf = null;
			try {
				conf = Configuration.readConfiguration(config, m.getParameters(), aliases);
			} catch (ClassNotFoundException | InstantiationException | IllegalAccessException e1) {
				e1.printStackTrace();
				return;
			}

			if (((double) conf.getValue(InstanceMatchingTask.PAR_ABSTRACT_WEIGHT) > 0.0)) {
				m.getData().initializeAbstracts();
			}

			System.out.println("wikidata classes: " + m.getData().getDbpediaTables().size());
			System.out.println("wikidata cols: " + m.getData().getDbpediaColSet().size());
			System.out.println("wikidata rows: " + m.getData().getDbpediaRowSet().size());

			Integer numThreads = (Integer) conf.getValue(WebtableToDBpediaMatchingProcess.PAR_MAX_PARALLEL);
			if (numThreads > 0) {
				Parallel.SetDefaultNumProcessors(numThreads);
			}

			List<File> files = new ArrayList<>(getAllFiles(new File(tablePath)));

			final Configuration configFinal = conf;
			try {
				// new
				// Parallel<File>().foreach(Arrays.asList(new
				// File(web).listFiles()), new Consumer<File>()
				// {

				PersistentParallel6.runAndFinish(files, (parameter) -> {
					try {
						WebTableToDBpediaComponentProcess p = new WebTableToDBpediaComponentProcess();
						p.setMatchingParameters(loadDbpParams);
						p.setData(m.getData().shallowCopy());
						p.setMatchingParameters(params);
						p.setEvaluationParameters(eval);
						p.setLuceneIndex(index.getLuceneIndex());

						Configuration c = configFinal.clone();
						c.getConfig().put(p.PAR_WEBTABLE, parameter.getAbsolutePath());
						p.run(c);
						MatchingResult r = p.getResult();
						if (r.getInstanceMappings() == null) {

							System.out.println("NO MATCH FOR YOU!");
							System.out.println(p.getLogger().getMatchingLog().toString());
						} else {
							r.write(WebtableToDBpediaMatchingProcess.tableType.jsonWebTable, outputFolder);
						}

					} catch (Exception e) {
						System.err.println(String.format("Matching Table %s failed!", parameter.getName()));
						e.printStackTrace();
						int sdf = 123;
					}
				}, "Match");

			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

		} catch (Exception e) {
			e.printStackTrace();

		}

		int ddfdf = 123;
	}

	private static List<File> getAllFiles(File dir) {

		List<File> ret = new ArrayList<>();

		for (File file : dir.listFiles()) {
			if (file.isDirectory()) {
				ret.addAll(getAllFiles(file));
			} else {
				ret.add(file);
			}
		}

		return ret;
	}

}
