package de.uni_mannheim.dws.expansion.data.ontology.model.objects;

import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import de.mannheim.uni.expansion.utils.ThreadSafeLocalCalendar;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class TimeObject extends AbstractObject implements Comparable<TimeObject> {

	// month starts at 1
	private final long year;
	private final int month;
	private final int day;
	private final int hours;
	private final int minutes;
	private final int seconds;
	private final int minutesOffset;

	private final static ThreadSafeLocalCalendar threadSafeCalendar = new ThreadSafeLocalCalendar();

	private final Precision precision;

	public TimeObject(long year, int month, int day, int hours, int minutes, int seconds, int minutesOffset, Precision precision) {
		this.year = year;
		this.month = month;
		this.day = day;
		this.hours = hours;
		this.minutes = minutes;
		this.seconds = seconds;
		this.minutesOffset = minutesOffset;
		this.precision = precision;
	}

	public static TimeObject fromDate(Date date) {
		Calendar calendar = threadSafeCalendar.getCalendar();
		calendar.setTime(date);

		int year = calendar.get(Calendar.YEAR);
		int month = calendar.get(Calendar.MONTH);
		int day = calendar.get(Calendar.DAY_OF_MONTH);
		int hourOfDay = calendar.get(Calendar.HOUR_OF_DAY);
		int minute = calendar.get(Calendar.MINUTE);
		int second = calendar.get(Calendar.SECOND);

		Precision precision = null;

		if (month == 0 && day == 1 && hourOfDay == 0 && minute == 0 && second == 0) {
			precision = Precision.T09_years;
		} else if (day == 1 && hourOfDay == 0 && minute == 0 && second == 0) {
			precision = Precision.T10_months;
		} else {
			precision = Precision.T11_days;
		}

		return new TimeObject(year, month + 1, day, hourOfDay, minute, second, 0, precision);
	}

	public long getYear() {
		return year;
	}

	public int getMonth() {
		return month;
	}

	public int getDay() {
		return day;
	}

	public int getHours() {
		return hours;
	}

	public int getMinutes() {
		return minutes;
	}

	public int getSeconds() {
		return seconds;
	}

	public int getMinutesOffset() {
		return minutesOffset;
	}

	public Precision getPrecision() {
		return precision;
	}

	public enum Precision {

		T00_1_Gigayear(0),
		T01_100_Megayears(1),
		T02_10_Megayears(2),
		T03_Megayear(3),
		T04_100_Kiloyears(4),
		T05_10_Kiloyears(5),
		T06_Kiloyear(6),
		T07_100_years(7),
		T08_10_years(8),
		T09_years(9),
		T10_months(10),
		T11_days(11),
		T12_hours(12),
		T13_minutes(13),
		T14_seconds(14);

		private Precision(int position) {
			this.position = position;
		}

		private int position;

		public int getPosition() {
			return position;
		}

		private static Map<Integer, Precision> map = new HashMap<>();

		static {
			for (Precision p : Precision.values()) {
				map.put(p.getPosition(), p);
			}
		}

		public static Precision getPrecision(int position) {
			return map.get(position);
		}
	}

	@Override
	public String toString() {
		return this.readableString();
	}

	@Override
	public String readableString() {

		switch (precision) {
		case T00_1_Gigayear:
			return "" + year;

		case T01_100_Megayears:
			return "" + year;

		case T02_10_Megayears:
			return "" + year;

		case T03_Megayear:
			return "" + year;

		case T04_100_Kiloyears:
			return "" + year;

		case T05_10_Kiloyears:
			return "" + year;

		case T06_Kiloyear:
			return "" + year;

		case T07_100_years:
			return "" + year;

		case T08_10_years:
			return "" + year;

		case T09_years:
			return "" + year;

		case T10_months:
			return "" + year + "-" + month;

		case T11_days:
			return "" + year + "-" + month + "-" + day;

		case T12_hours:
			return "" + year + "-" + month + "-" + day + "T" + hours;

		case T13_minutes:
			return "" + year + "-" + month + "-" + day + "T" + hours + ":" + minutes;

		case T14_seconds:
			return "" + year + "-" + month + "-" + day + "T" + hours + ":" + minutes + ":" + seconds;
		default:
			return "";

		}

	}

	@Override
	public int compareTo(TimeObject o) {
		return TimeObject.compare(this, o);
	}

	// FIXME: fix this. It should sort probably
	// Use methods below
	public static int compare(TimeObject o1, TimeObject o2) {
		int compare = Long.compare(o1.getYear(), o2.getYear());
		if (compare != 0) {
			return compare;
		}

		Precision maxPrecision = null;
		if (o1.getPrecision().getPosition() > o2.getPrecision().getPosition()) {
			maxPrecision = o1.getPrecision();
		} else {
			maxPrecision = o2.getPrecision();
		}

		if (maxPrecision.getPosition() >= Precision.T10_months.getPosition()) {

			compare = Integer.compare(o1.getMonth(), o2.getMonth());
			if (compare != 0) {
				return compare;
			}

			if (maxPrecision.getPosition() > Precision.T11_days.getPosition()) {
				compare = Integer.compare(o1.getDay(), o2.getDay());
				if (compare != 0) {
					return compare;
				}

				if (maxPrecision.getPosition() > Precision.T12_hours.getPosition()) {

					compare = Integer.compare(o1.getHours(), o2.getHours());
					if (compare != 0) {
						return compare;
					}

					if (maxPrecision.getPosition() > Precision.T13_minutes.getPosition()) {
						compare = Integer.compare(o1.getMinutes(), o2.getMinutes());
						if (compare != 0) {
							return compare;
						}

						if (maxPrecision.getPosition() > Precision.T14_seconds.getPosition()) {

							compare = Integer.compare(o1.getSeconds(), o2.getSeconds());
							if (compare != 0) {
								return compare;
							}
						}
					}
				}

			}

		}

		return 0;
	}

	public static TimeObject averager(TimeObject o1, double score1, TimeObject o2, double score2) {
		TimeObject higher = getHigherScored(o1, score1, o2, score2);

		// Why not choose common precision?
		// If they are equal, we can choose higher anyway
		// If they are unequal, we choose higher as fallback
		Precision precision = higher.getPrecision();

		long year = 0;
		int month = 0;
		int day = 0;
		int hours = 0;
		int minutes = 0;
		int seconds = 0;
		int minutesOffset = 0;

		if (o1.getYear() == o2.getYear()) {
			year = o1.getYear();
		}

		if (precision.getPosition() >= Precision.T10_months.getPosition()) {
			if (o1.getMonth() == o2.getMonth()) {
				month = o1.getMonth();
			} else {
				precision = Precision.T09_years;
			}
		}

		if (precision.getPosition() >= Precision.T11_days.getPosition()) {
			if (o1.getDay() == o2.getDay()) {
				day = o1.getDay();
			} else {
				precision = Precision.T10_months;
			}
		}

		if (precision.getPosition() >= Precision.T12_hours.getPosition()) {
			if (o1.getHours() == o2.getHours()) {
				hours = o1.getHours();
			} else {
				precision = Precision.T11_days;
			}
		}

		if (precision.getPosition() >= Precision.T13_minutes.getPosition()) {
			if (o1.getMinutes() == o2.getMinutes()) {
				minutes = o1.getMinutes();
			} else {
				precision = Precision.T12_hours;
			}
		}

		if (precision.getPosition() >= Precision.T14_seconds.getPosition()) {
			if (o1.getSeconds() == o2.getSeconds()) {
				seconds = o1.getSeconds();
			} else {
				precision = Precision.T13_minutes;
			}
		}

		if (o1.getMinutesOffset() == o2.getMinutesOffset()) {
			minutesOffset = o1.getMinutesOffset();
		}

		return new TimeObject(year, month, day, hours, minutes, seconds, minutesOffset, precision);

	}

	/**
	 * Returns the higher score object with preference to the first object given
	 * 
	 * @param o1
	 * @param score1
	 * @param o2
	 * @param score2
	 * @return
	 */
	private static TimeObject getHigherScored(TimeObject o1, double score1, TimeObject o2, double score2) {
		if (score1 >= score2) {
			return o1;
		} else {
			return o2;
		}
	}

	public static Precision getHighestPrecision(Collection<TimeObject> values) {
		Precision highestPrecision = null;

		for (TimeObject e : values) {
			if (highestPrecision == null || e.getPrecision().getPosition() > highestPrecision.getPosition()) {
				highestPrecision = e.getPrecision();
			}
		}

		return highestPrecision;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + day;
		result = prime * result + hours;
		result = prime * result + minutes;
		result = prime * result + minutesOffset;
		result = prime * result + month;
		result = prime * result + ((precision == null) ? 0 : precision.hashCode());
		result = prime * result + seconds;
		result = prime * result + (int) (year ^ (year >>> 32));
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		TimeObject other = (TimeObject) obj;
		if (day != other.day)
			return false;
		if (hours != other.hours)
			return false;
		if (minutes != other.minutes)
			return false;
		if (minutesOffset != other.minutesOffset)
			return false;
		if (month != other.month)
			return false;
		if (precision != other.precision)
			return false;
		if (seconds != other.seconds)
			return false;
		if (year != other.year)
			return false;
		return true;
	}

	public boolean isYearDate() {
		return this.getPrecision().getPosition() >= Precision.T00_1_Gigayear.getPosition()
				&& this.getPrecision().getPosition() < Precision.T10_months.getPosition();
	}

	public boolean isMonthDate() {
		return this.getPrecision().getPosition() == Precision.T10_months.getPosition();
	}

	public Date getDate() {
		// TODO Auto-generated method stub
		return new Date();
	}

	public static TimeObject fromYear(Integer year) {
		return new TimeObject(year, 0, 0, 0, 0, 0, 0, Precision.T09_years);
	}

	// public boolean isDayDate() {
	// return this.getPrecision().getPosition() >=
	// Precision.T11_days.getPosition();
	// }

	public static Integer getYear(AbstractObject ao) {
		if (ao instanceof TimeObject) {
			TimeObject timeObject = (TimeObject) ao;
			int year = (int) timeObject.getYear();
			return year;
		}
		return null;
	}

	@Override
	public ValueType getType() {
		return ValueType.TIME;
	}

}
