package expansion.set.clustering.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.rowclusterers.ClusterSimilarities;

public class RowCluster {
	private final ClassId classId;
	private final List<TableRowId> rows;
	private final Map<TableRowId, Integer> rowsIndex;
	private final RowClusterSimilarities similarities;

	private final RowClusterUUID uuid;
	
	

	public RowCluster( RowClusterUUID uuid, ClassId classId, List<TableRowId> rows, RowClusterSimilarities similarities) {
		this.uuid = uuid;
		this.classId = classId;
		this.rows = Collections.unmodifiableList(new ArrayList<>(rows));
		Map<TableRowId, Integer> rowsIndex = new HashMap<>();
		for (ListEntry<TableRowId> entry : ListIterable.iterable(this.rows)) {
			rowsIndex.put(entry.getValue(), entry.getIndex());
		}
		this.rowsIndex = Collections.unmodifiableMap(rowsIndex);
		this.similarities = similarities;
	}

	public RowCluster(ClassId classId, List<TableRowId> rows, RowClusterSimilarities similarities) {
		this.uuid = new RowClusterUUID();
		this.classId = classId;
		this.rows = Collections.unmodifiableList(new ArrayList<>(rows));
		Map<TableRowId, Integer> rowsIndex = new HashMap<>();
		for (ListEntry<TableRowId> entry : ListIterable.iterable(this.rows)) {
			rowsIndex.put(entry.getValue(), entry.getIndex());
		}
		this.rowsIndex = Collections.unmodifiableMap(rowsIndex);
		this.similarities = similarities;
	}

	public List<TableRowId> getRows() {
		return rows;
	}

	public ClassId getClassId() {
		return classId;
	}

	public RowClusterSimilarities getSimilarities() {
		return similarities;
	}

	public ClusterSimilaritiesWrapper getSimilaritiesWrapper() {
		return new ClusterSimilaritiesWrapper();
	}

	public RowClusterUUID getUuid() {
		return uuid;
	}

	public Integer getIndex(TableRowId tableRowId) {
		return this.rowsIndex.get(tableRowId);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((uuid == null) ? 0 : uuid.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		RowCluster other = (RowCluster) obj;
		if (uuid == null) {
			if (other.uuid != null)
				return false;
		} else if (!uuid.equals(other.uuid))
			return false;
		return true;
	}

	public static class ClusterEntryKey {

		@Override
		public String toString() {
			return "[e1=" + e1 + ", e2=" + e2 + "]";
		}

		private final int e1;
		private final int e2;

		public ClusterEntryKey(int e1, int e2) {
			if (e1 == e2) {
				throw new IllegalStateException("A Similarity Key can not be between the same entry");
			}

			if (Integer.compare(e1, e2) == -1) {
				this.e1 = e1;
				this.e2 = e2;
			} else {
				this.e1 = e2;
				this.e2 = e1;
			}

		}

		public int getE1() {
			return e1;
		}


		public int getE2() {
			return e2;
		}



		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + e1;
			result = prime * result + e2;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			ClusterEntryKey other = (ClusterEntryKey) obj;
			if (e1 != other.e1)
				return false;
			if (e2 != other.e2)
				return false;
			return true;
		}

	}

	public class ClusterSimilaritiesWrapper implements ClusterSimilarities {

		@Override
		public double get(ClusterEntryKey key) {
			return similarities.get(rows, classId, key);
		}

		@Override
		public Set<Integer> regionQuery(Integer point, double eps) {
			TableRowId tableRowId = rows.get(point);
			Set<TableRowId> candidates = similarities.regionQuery(RowCluster.this, tableRowId, eps);
			return candidates.stream().map((e) -> rowsIndex.get(e)).collect(Collectors.toSet());

		}

	}

}
