package expansion.set.creation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.adapters.columnmatches.ColumnMatchesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.SynchronizedMatchedValuesMapMerger;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitterFactoryRepository;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionRun;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfigFilter;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionDataType;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionPreproccesser;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionRunner;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.ScorerRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.tablerow.AnnotatedQualifiedTableRow;
import expansion.set.tablerow.TableRowCreator;

public class EntityCreator {

	private final AnnotatedTablesRepository tableRepository;
	private final KnowledgeBase knowledgeBase;
	private final FusionRunner fusionRunner;
	private final ScorerRepository scorerRepository;
	private final OntologyRepository ontologyRepository;
	private final LabelMatcher labelMatcher;

	private final ColumnMatchesRepository columnMatchesRepository;
	private final PropertiesConfigFilter propertiesConfigFilter;
	private final SplitterFactoryRepository splitterFactoryRepository;
	private final TableRowCreator tableRowCreator;

	@Inject
	public EntityCreator(AnnotatedTablesRepository tableRepository, KnowledgeBase knowledgeBase, FusionRunner fusionRunner, ScorerRepository scorerRepository,
			ColumnMatchesRepository columnMatchesRepository, OntologyRepository ontologyRepository, LabelMatcher labelMatcher,
			PropertiesConfigFilter propertiesConfigFilter, SplitterFactoryRepository splitterFactoryRepository, TableRowCreator tableRowCreator) {
		super();
		this.tableRepository = tableRepository;
		this.knowledgeBase = knowledgeBase;
		this.fusionRunner = fusionRunner;
		this.scorerRepository = scorerRepository;
		this.columnMatchesRepository = columnMatchesRepository;
		this.ontologyRepository = ontologyRepository;
		this.labelMatcher = labelMatcher;
		this.propertiesConfigFilter = propertiesConfigFilter;
		this.splitterFactoryRepository = splitterFactoryRepository;
		this.tableRowCreator = tableRowCreator;
	}

	public List<Entity> createEntites(SplitGroupedValuesSets sets, List<ThresholdContainer> thresholdContainers, FusionConfig fusionConfig,
			StrategiesConfig strategiesConfig, SetExpansionStrategiesConfig setExpansionStrategiesConfig, List<RowCluster> rowClusters,
			ClustererInput clustererInput) {
		List<FusionDataType> dataTypes = strategiesConfig.getDataTypes();

		List<FusionPreproccesser> preprocessors = strategiesConfig.getFusionPreprocessers();

		ConcurrentHashMap<InstanceId, ClassId> classes = new ConcurrentHashMap<>();
		ConcurrentHashMap<InstanceId, Map<String, Integer>> labels = new ConcurrentHashMap<>();
		SynchronizedMatchedValuesMapMerger<QualifiedTripleId> testingSetCreator = new SynchronizedMatchedValuesMapMerger<>();
		
		Map<InstanceId, RowClusterUUID> mapping = new ConcurrentHashMap<>();
		Map<InstanceId, RowCluster> index = new ConcurrentHashMap<>();

		List<Integer> range = IntStream.rangeClosed(0, rowClusters.size() - 1).boxed().collect(Collectors.toList());

		PersistentParallel6.runAndFinish(ListIterable.iterable(rowClusters), (rowCluster) -> {
			int row = rowCluster.getIndex();
			RowCluster cluster = rowCluster.getValue();
			 
			InstanceId instanceId = new InstanceId("INSTANCE_" + row, "INSTANCE_" + row);
			mapping.put(instanceId, cluster.getUuid());
			index.put(instanceId, cluster);
			
			classes.put(instanceId, cluster.getClassId());
			ConcurrentHashMap<String, AtomicInteger> labelMap = new ConcurrentHashMap<>();

			PersistentParallel6.runAndFinish(cluster.getRows(), (tableRowId) -> {
				AnnotatedQualifiedTableRow annotatedTableRow = tableRowCreator.getTableRow(tableRowId, clustererInput.getRowValuesSets());
				if (annotatedTableRow.getLabel() != null) {
					CHMHelper.getFromCHM(labelMap, StringNormalizer.normaliseValue(annotatedTableRow.getLabel().trim(), false), () -> new AtomicInteger()).getAndIncrement();
				}

				annotatedTableRow.getValues().entrySet().stream().forEach((e) -> {

					testingSetCreator.merge(new QualifiedTripleId(cluster.getClassId(), instanceId, e.getKey()), e.getValue());

				});
			}, this.getClass(), cluster.getRows().size() > 500); 
			
			Map<String, Integer> labelMapValue = labelMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return e.getValue().intValue();
			}));
			
			if(labelMapValue == null) {
				int sdf = 123;
			}

			labels.put(instanceId, labelMapValue);

		} , this.getClass());

		Map<QualifiedTripleId, MatchedValues> testingSetSplitted = testingSetCreator.getCleanMap();
		;

		// Map<TripleId, MatchedValues> testingSet =
		// testingSetCreator.getCleanMap();
		// Map<TripleId, MatchedValues> testingSet =
		// testingSetCreator.getCleanMap().entrySet().stream().filter((e) -> {
		// return e.getValue().getValues().size() >=
		// fusionConfig.getMinSource();
		// }).collect(Collectors.toMap(Entry::getKey, Entry::getValue));


		// for (int i = 0; i <
		// setExpansionStrategiesConfig.getFusionSetups().size(); i++) {
		// Pair<FusionStrategy, PropertiesConfig> fusionSetup =
		// setExpansionStrategiesConfig.getFusionSetups().get(i);
		//
		// Map<TripleId, MatchedValues> filteredTestingSet =
		// testingSet.entrySet().stream().filter((e) -> {
		// return propertiesConfigFilter.include(fusionSetup.getSecond(),
		// e.getKey());
		// }).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
		//
		// GroupedValues filteredTestingSetGV = new
		// GroupedValues(filteredTestingSet);
		//
		// Scorer scoresGetter =
		// scorerRepository.get(fusionSetup.getFirst().getScoringMethod());
		// if (scoresGetter instanceof SplitterProvider) {
		// System.out.println("RESPLITING");
		// FilteredSplitter filteredSplitter = ((SplitterProvider)
		// scoresGetter).getSplitter(sets);
		// Map<QualifiedTripleId, MatchedValues> splitValues =
		// filteredSplitter.split(filteredTestingSetGV,
		// fusionSetup.getSecond()).getValues();
		//
		// testingSetSplitted.putAll(splitValues);
		// } else {
		//
		// Splitter splitter =
		// this.splitterFactoryRepository.get(NonTimeDependentSplitterFactory.class).getSplitter(filteredTestingSetGV,
		// filteredTestingSetGV, filteredTestingSetGV);
		// testingSetSplitted.putAll(splitter.split(filteredTestingSetGV).getValues());
		// }
		//
		// }

		Map<QualifiedTripleId, MatchedValues> testingSetSplittedMap = testingSetSplitted.entrySet().stream().filter((e) -> {
			return e.getValue().getValues().size() >= fusionConfig.getMinSource();
		}).collect(Collectors.toMap((e) -> e.getKey(), (e) -> e.getValue()));

		testingSetSplittedMap = testingSetSplitted;
		ConcurrentHashMap<QualifiedTripleId, MatchedValues> fullSet = new ConcurrentHashMap<>(sets.getLearningSet().getValues());

		fullSet.putAll(testingSetSplittedMap);

//		System.out.println("Before MinSource Filter: " + testingSetSplitted.size());
//		System.out.println("After MinSource Filter: " + testingSetSplittedMap.size());
//
//		System.out.println("BeforeSplitting: " + testingSet.size());
//		System.out.println("FullSetSiz: " + fullSet.size());

		// CREATE SPLIT_GROUPED_VALUES FROM ROWCLUSTER-LIST

		SplitGroupedValuesSets setsNew = new SplitGroupedValuesSets(new SplitGroupedValues(fullSet), sets.getLearningSet(),
				new SplitGroupedValues(testingSetSplittedMap));

		Map<QualifiedTripleId, FusedValue> fusedValues = new HashMap<>();

		for (int i = 0; i < setExpansionStrategiesConfig.getFusionSetups().size(); i++) {
			Pair<FusionStrategy, PropertiesConfig> fusionSetup = setExpansionStrategiesConfig.getFusionSetups().get(i);

			Scorer scorer = scorerRepository.get(fusionSetup.getFirst().getScoringMethod());
			ScoresContainer probabilities = scorer.getScores(sets);

			Map<FusionRun, Map<QualifiedTripleId, FusedValue>> fusedValuesMap = fusionRunner.run(dataTypes, preprocessors, fusionSetup.getFirst(),
					setsNew.getTestingSet(), probabilities, thresholdContainers.get(i), fusionConfig, fusionSetup.getSecond());

			fusedValues.putAll(CHMHelper.flattenMap(fusedValuesMap));

		}

		ConcurrentHashMap<InstanceId, ConcurrentHashMap<QualifiedTripleId, FusedValue>> entities = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(fusedValues.entrySet(), (e) -> {

			CHMHelper.getFromCHM(entities, e.getKey().getTripleId().getInstanceId(), () -> new ConcurrentHashMap<QualifiedTripleId, FusedValue>())
					.put(e.getKey(), e.getValue());

		} , this.getClass());

		System.out.println("FusedValues: " + entities.entrySet().stream().mapToInt((e) -> {
			return e.getValue().size();
		}).sum());

		List<Entity> entitiesList = new ArrayList<>();

		for(Entry<InstanceId, RowClusterUUID> entityEntry : mapping.entrySet()) {
			Map<QualifiedPropertyId, FusedValue> values;
			if(entities.containsKey(entityEntry.getKey())) {
			  values = entities.get(entityEntry.getKey()).entrySet().stream().collect(Collectors.toMap((e) -> {
				return e.getKey().getQualifiedPropertyId();
			} , Entry::getValue));
			}
			else {
				values = new HashMap<>();
			}
			
			entitiesList.add(
					new Entity(entityEntry.getValue(), classes.get(entityEntry.getKey()), labels.get(entityEntry.getKey()), values, index.get(entityEntry.getKey()) ));
		}
		
//		for (Entry<InstanceId, ConcurrentHashMap<QualifiedTripleId, FusedValue>> entry : entities.entrySet()) {
//			entitiesList.add(
//					new Entity(mapping.get(entry.getKey()), classes.get(entry.getKey()), labels.get(entry.getKey()), entry.getValue().entrySet().stream().collect(Collectors.toMap((e) -> {
//						return e.getKey().getQualifiedPropertyId();
//					} , Entry::getValue))));
//		}


		return entitiesList;

	}






}
