package expansion.set.distant_supervision;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.ConfigInterface;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.UuidStringId;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.goldstandard.GoldStandardClustersSetPerClass;

public class DistantSupervisionInput implements ConfigInterface<DistantSupervisionInput> {

	private Map<ClassId, Set<ClassId>> classesWithParents;
	private SetExpansionStrategiesConfig setExpansionStrategiesConfig;

	private List<EntityMatchingRun> entityMatchingRuns;
	private Map<ClassId, TableIdsSet> tableIdsPerClass;
	private Map<ClassId, TableRowIdsSet> fullRowIdsPerClass;
	private Map<ClassId, Set<PropertyId>> nonTailProperties;
	private FusionConfig fusionConfig;
	private StrategiesConfig strategiesConfig;
	private String thresholdsFolder;
	private GoldStandardClustersSetPerClass goldStandardClustersSet;
	private Map<ClassId, Set<TableRowId>> wsRowIdsSet;

	private DistantSupervisionInput() {

	}

	public Map<ClassId, Set<ClassId>> getClassesWithParents() {
		return classesWithParents;
	}

	public SetExpansionStrategiesConfig getSetExpansionStrategiesConfig() {
		return setExpansionStrategiesConfig;
	}


	public List<EntityMatchingRun> getEntityMatchingRuns() {
		return entityMatchingRuns;
	}

	public Map<ClassId, TableIdsSet> getTableIdsPerClass() {
		return tableIdsPerClass;
	}


	public Map<ClassId, Set<PropertyId>> getNonTailProperties() {
		return nonTailProperties;
	}

	public void setNonTailProperties(Map<ClassId, Set<PropertyId>> nonTailProperties) {
		this.nonTailProperties = nonTailProperties;
	}

	public GoldStandardClustersSetPerClass getGoldStandardClustersSet() {
		return goldStandardClustersSet;
	}

	@Override
	public DistantSupervisionInput copy() {
		DistantSupervisionInput copy = new DistantSupervisionInput();
		copy.classesWithParents = this.classesWithParents;
		copy.setExpansionStrategiesConfig = setExpansionStrategiesConfig;
		copy.entityMatchingRuns = entityMatchingRuns;

		copy.tableIdsPerClass = tableIdsPerClass;
		copy.fullRowIdsPerClass = fullRowIdsPerClass;
		copy.nonTailProperties = nonTailProperties;
		copy.fusionConfig = fusionConfig;
		copy.strategiesConfig = strategiesConfig;
		copy.thresholdsFolder = thresholdsFolder;
		copy.goldStandardClustersSet = goldStandardClustersSet;

		copy.wsRowIdsSet = wsRowIdsSet;
		return copy;
	}

	@Override
	public DistantSupervisionInputBuilder rebuilder() {
		return new DistantSupervisionInputBuilder(this);
	}

	public static DistantSupervisionInputBuilder builder() {
		return new DistantSupervisionInputBuilder();

	}

	public static class DistantSupervisionInputBuilder implements ConfigBuilderInterface<DistantSupervisionInput> {

		private final DistantSupervisionInput input;

		public DistantSupervisionInputBuilder() {
			this.input = new DistantSupervisionInput();
		}

		public DistantSupervisionInputBuilder(DistantSupervisionInput input) {
			this.input = input.copy();
		}

		public DistantSupervisionInputBuilder setClassesWithParents(Map<ClassId, Set<ClassId>> classesWithParents) {
			input.classesWithParents = ImmuMaker.mapWithSet(classesWithParents);
			return this;
		}

		public DistantSupervisionInputBuilder setSetExpansionStrategiesConfig(SetExpansionStrategiesConfig setExpansionStrategiesConfig) {
			input.setExpansionStrategiesConfig = setExpansionStrategiesConfig;
			return this;
		}

		public DistantSupervisionInputBuilder setEntityMatchingRuns(List<EntityMatchingRun> entityMatchingRuns) {
			input.entityMatchingRuns = Collections.unmodifiableList(new ArrayList<>(entityMatchingRuns));
			return this;
		}

		@Override
		public DistantSupervisionInput build() {
			return input.copy();
		}

		public DistantSupervisionInputBuilder setTableIdsPerClass(Map<ClassId, TableIdsSet> tableIds) {
			input.tableIdsPerClass = ImmuMaker.map(tableIds);
			return this;
		}

		public DistantSupervisionInputBuilder setFullTableRowIdSet(Map<ClassId, TableRowIdsSet> fullTableRowIds) {
			input.fullRowIdsPerClass = ImmuMaker.map(fullTableRowIds);
			return this;
		}

		public DistantSupervisionInputBuilder setFusionConfig(FusionConfig fusionConfig) {
			input.fusionConfig = fusionConfig;
			return this;
		}

		public DistantSupervisionInputBuilder setStrategiesConfig(StrategiesConfig strategiesConfig) {
			input.strategiesConfig = strategiesConfig;
			return this;
		}

		public DistantSupervisionInputBuilder setThresholdsFolder(String thresholdsFolder) {
			input.thresholdsFolder = thresholdsFolder;
			return this;
		}

		public DistantSupervisionInputBuilder setNonTailProperties(Map<ClassId, Set<PropertyId>> nonTailProperties) {
			input.nonTailProperties = ImmuMaker.mapWithSet(nonTailProperties);
			return this;
		}

		public DistantSupervisionInputBuilder setGoldStandardClustersSet(GoldStandardClustersSetPerClass goldStandardClustersSet) {
			input.goldStandardClustersSet = goldStandardClustersSet;
			return this;
		}

		public DistantSupervisionInputBuilder setWsRowIds(Map<ClassId, Set<TableRowId>> wsRowIdsSet) {
			input.wsRowIdsSet = ImmuMaker.mapWithSet(wsRowIdsSet);
			return this;
		}

	}

	public TableIdsSet getTableIdsSet(ClassId classId) {
		return tableIdsPerClass.get(classId);
	}

	public TableRowIdsSet getFullRowIdsSet(ClassId classId) {
		return fullRowIdsPerClass.get(classId);
	}

	public Set<TableRowId> getWsRowIdsSet(ClassId classId) {
		return wsRowIdsSet.get(classId);
	}

	public TableRowIdsSet getFullRowIdsSetAllClasses() {
		Set<TableRowId> rowIds = new HashSet<>();
		for (TableRowIdsSet entry : fullRowIdsPerClass.values()) {
			rowIds.addAll(entry.getRowIds());
		}

		return new TableRowIdsSet(rowIds, new UuidStringId("ALL"));
	}

	public FusionConfig getFusionConfig() {
		return fusionConfig;
	}

	public StrategiesConfig getStrategiesConfig() {
		return strategiesConfig;
	}

	public String getThresholdsFolder() {
		return thresholdsFolder;
	}

}
