package expansion.set.distant_supervision.matching_rules;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.NumericSimilarity;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import expansion.set.clustering.model.Block;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.clustersplitters.minScoreSplitter.WekaLearnableRowClusterSimilarities;

public class MatchingRulesWrapperCreator {

	@Inject
	protected NumericSimilarity numericSimilarity;

	@Inject
	protected SimilarityCalculator similarityCalculator;

	@Inject
	public MatchingRulesWrapperCreator() {

	}


	public RowClusterSimilarities getSim(RowClusterSimilarities sim, WekaLearnableRowClusterSimilarities fullSim) {

		// double recordDateThreshold =
		// numericSimilarity.getThreshold(single_releaseDate, 1.0);

		return new MR_ClusterSimilarities(sim, fullSim);

	}

	public class MR_ClusterSimilarities implements RowClusterSimilarities {

		private final RowClusterSimilarities sim;
		private final WekaLearnableRowClusterSimilarities fullSim;
		private final double labelEquivalenceThreshold;
		private final Set<TableRowIdPair> positivePairs = Collections.newSetFromMap(new ConcurrentHashMap<>());
		private final Set<TableRowIdPair> negativePairs = Collections.newSetFromMap(new ConcurrentHashMap<>());

		public MR_ClusterSimilarities(RowClusterSimilarities sim, WekaLearnableRowClusterSimilarities fullSim) {
			super();
			this.sim = sim;
			this.fullSim = fullSim;
			this.labelEquivalenceThreshold = similarityCalculator.getMeasure(ValueType.STRING).getEquivalenceThreshold();

		}

		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {

			RecordFeature fv = fullSim.getFeatureVector(rows.get(key.getE1()), rows.get(key.getE2()), classId);
			Optional<Double> score = getMatchingRulesScore(classId, fv.getFeatures(), labelEquivalenceThreshold);

			if (CRAZY) {
				if (score.isPresent()) {
					return score.get();
				} else {
					return 0.0;
				}
			}

			double simScore = sim.get(rows, classId, key);
			if (score.isPresent()) {
				TableRowIdPair pair = new TableRowIdPair(rows.get(key.getE1()), rows.get(key.getE2()));
				if (score.get() < 0.0) {
					negativePairs.add(pair);
					if (simScore < score.get()) {
						return simScore;
					} else {
						return (score.get() + simScore) / 2.0;
					}
				} else {
					positivePairs.add(pair);
					if (simScore > score.get()) {
						return simScore;
					} else {
						return (score.get() + simScore) / 2.0;
					}
				}

				// return (score.get() + simScore) / 2.0;
			} else {
				return simScore;
			}
		}

		public List<WeightedScore> getList(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			// return sim.getList(rows, classId, key);
			throw new UnsupportedOperationException();
		}

		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			return sim.regionQuery(cluster, point, eps);
		}

		public String getName() {
			return sim.getName();
		}

		public Set<Integer> getBlocks(TableRowId rowId) {
			return sim.getBlocks(rowId);
		}

		public boolean supportsBlocking() {
			return sim.supportsBlocking();
		}

		public List<Set<Integer>> getSplitBlocks(TableRowId tableRowId) {
			return sim.getSplitBlocks(tableRowId);
		}

		public List<Pair<Set<Integer>, Integer>> getSplitBlocksWithMin(TableRowId tableRowId) {
			return sim.getSplitBlocksWithMin(tableRowId);
		}

		public List<Tri<Set<Integer>, Integer, Double>> getSplitBlocksWithScores(TableRowId tableRowId) {
			return sim.getSplitBlocksWithScores(tableRowId);
		}

		public List<Pair<Set<Block>, Integer>> getSplitBlocksWithScoresAndMinWithMin(TableRowId tableRowId) {
			return sim.getSplitBlocksWithScoresAndMinWithMin(tableRowId);
		}

		public int getBlockMin(TableRowId tableRowId) {
			return sim.getBlockMin(tableRowId);
		}

		public int getBlockerSizes() {
			return sim.getBlockerSizes();
		}

		public void clean() {
			positivePairs.clear();
			negativePairs.clear();

		}

		public int getPositivePairsSize() {
			return positivePairs.size();
		}

		public int getNegativePairsSize() {
			return negativePairs.size();
		}
	}

	public static final boolean CRAZY = false;

	public ND_MR_SCORE_ADAPTER getNdScoreGetter(ClassId classId) {
		double labelEquivalenceThreshold = similarityCalculator.getMeasure(ValueType.STRING).getEquivalenceThreshold();

		return new ND_MR_SCORE_ADAPTER_CLASS(classId, labelEquivalenceThreshold);
	}

	public class ND_MR_SCORE_ADAPTER_CLASS implements ND_MR_SCORE_ADAPTER {

		private final ClassId classId;
		private final double labelEquivalenceThreshold;

		private final AtomicInteger positivePairs = new AtomicInteger(0);
		private final AtomicInteger negativePairs = new AtomicInteger(0);

		public ND_MR_SCORE_ADAPTER_CLASS(ClassId classId, double labelEquivalenceThreshold) {
			super();
			this.classId = classId;
			this.labelEquivalenceThreshold = labelEquivalenceThreshold;
		}

		@Override
		public Double get(Double simScore, RecordFeature fv) {
			Optional<Double> score = getMatchingRulesScore(classId, fv.getFeatures(), labelEquivalenceThreshold);

			if (CRAZY) {
				if (score.isPresent()) {
					return score.get();
				} else {
					return 0.0;
				}
			}
			// double simScore = sim.get(rows, classId, key);
			if (score.isPresent()) {
				if (score.get() < 0.0) {
					negativePairs.incrementAndGet();
					if (simScore < score.get()) {
						return simScore;
					} else {
						return (score.get() + simScore) / 2.0;
					}
				} else {
					positivePairs.incrementAndGet();
					if (simScore > score.get()) {
						return simScore;
					} else {
						return (score.get() + simScore) / 2.0;
					}
				}

			} else {
				return simScore;
			}
		}

		public void clean() {
			positivePairs.set(0);
			negativePairs.set(0);

		}

		public int getPositivePairsSize() {
			return positivePairs.intValue();
		}

		public int getNegativePairsSize() {
			return negativePairs.intValue();
		}

	}

	public static interface ND_MR_SCORE_ADAPTER {
		Double get(Double initial, RecordFeature rf);

		public void clean();

		public int getPositivePairsSize();

		public int getNegativePairsSize();
	}

	protected Optional<Double> getMatchingRulesScore(ClassId classId, Map<String, Double> features, double labelEquivalenceThreshold) {
		switch (classId.getFinalPart()) {
		case "GridironFootballPlayer":
			return getMatchingRulesScoreGF(features, labelEquivalenceThreshold);
		case "Single":
			return getMatchingRulesScoreSong(features, labelEquivalenceThreshold);
		case "Settlement":
			return getMatchingRulesScoreSettlement(features, labelEquivalenceThreshold);
		}

		throw new IllegalStateException();
	}

	private Optional<Double> getMatchingRulesScoreGF(Map<String, Double> features, double labelEquivalenceThreshold) {


		String ATT_draftYear_NO_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/draftYear::NO_MATCH";
		String IMP_draftYear_NO_MATCH = "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/draftYear::NO_MATCH";

		String ATT_draftPick_NO_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/draftPick::NO_MATCH";
		// String IMP_draftPick_NO_MATCH =
		// "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/draftPick::NO_MATCH";

		String ATT_college_NO_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/college::NO_MATCH";
		String IMP_college_NO_MATCH = "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/college::NO_MATCH";

		String ATT_draftYear_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/draftYear::MATCH";
		String IMP_draftYear_MATCH = "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/draftYear::MATCH";

		String ATT_draftPick_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/draftPick::MATCH";
		// String IMP_draftPick_MATCH =
		// "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/draftPick::MATCH";

		String ATT_birthDate_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/birthDate::MATCH";
		// String IMP_birthDate_MATCH =
		// "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/birthDate::MATCH";

		String LABEL = "LABEL";

		Double negativeMatch = null;

		if (features.containsKey(ATT_draftYear_NO_MATCH) && features.get(ATT_draftYear_NO_MATCH) > 0.0) {
			double noMatch = features.get(ATT_draftYear_NO_MATCH);
			if (negativeMatch == null || noMatch > negativeMatch) {
				negativeMatch = noMatch;
			}
		}

		if (features.containsKey(IMP_draftYear_NO_MATCH) && features.get(IMP_draftYear_NO_MATCH) > 0.0) {
			double noMatch = features.get(IMP_draftYear_NO_MATCH);
			if (negativeMatch == null || noMatch > negativeMatch) {
				negativeMatch = noMatch;
			}
		}

		if (features.containsKey(ATT_draftPick_NO_MATCH) && features.get(ATT_draftPick_NO_MATCH) > 0.0) {
			double noMatch = features.get(ATT_draftPick_NO_MATCH);
			if (negativeMatch == null || noMatch > negativeMatch) {
				negativeMatch = noMatch;
			}
		}

		// if (features.containsKey(ATT_college_NO_MATCH) &&
		// features.get(ATT_college_NO_MATCH) > 0.0) {
		// double noMatch = features.get(ATT_college_NO_MATCH);
		// if (negativeMatch == null || noMatch > negativeMatch) {
		// negativeMatch = noMatch;
		// }
		// }

		// if (features.containsKey(IMP_college_NO_MATCH) &&
		// features.get(IMP_college_NO_MATCH) > 0.0) {
		// double noMatch = features.get(IMP_college_NO_MATCH);
		// if (negativeMatch == null || noMatch > negativeMatch) {
		// negativeMatch = noMatch;
		// }
		// }

		Double positiveMatch = null;

		if (features.containsKey(IMP_draftYear_MATCH) && features.get(IMP_draftYear_MATCH) > 0.0) {
			double match = features.get(IMP_draftYear_MATCH);

			if (features.containsKey(ATT_draftPick_MATCH) && features.get(ATT_draftPick_MATCH) > 0.0) {
				double match2 = features.get(ATT_draftPick_MATCH) * match;
				if (positiveMatch == null || match2 > positiveMatch) {
					positiveMatch = match2;
				}
			}

			// if (features.containsKey(IMP_draftPick_MATCH) &&
			// features.get(IMP_draftPick_MATCH) > 0.0 ) {
			// double match2 = features.get(IMP_draftPick_MATCH) * match;
			// if(positiveMatch == null || match2 > positiveMatch) {
			// positiveMatch = match2;
			// }
			// }
			//

		}

		if (features.containsKey(ATT_draftYear_MATCH) && features.get(ATT_draftYear_MATCH) > 0.0) {
			double match = features.get(ATT_draftYear_MATCH);

			if (features.containsKey(ATT_draftPick_MATCH) && features.get(ATT_draftPick_MATCH) > 0.0) {
				double match2 = features.get(ATT_draftPick_MATCH) * match;
				if (positiveMatch == null || match2 > positiveMatch) {
					positiveMatch = match2;
				}
			}

			// if (features.containsKey(IMP_draftPick_MATCH) &&
			// features.get(IMP_draftPick_MATCH) > 0.0 ) {
			// double match2 = features.get(IMP_draftPick_MATCH) * match;
			// if(positiveMatch == null || match2 > positiveMatch) {
			// positiveMatch = match2;
			// }
			// }

		}

		if (features.containsKey(LABEL) && features.get(LABEL) > labelEquivalenceThreshold) {
			double match = features.get(LABEL);

			if (features.containsKey(ATT_birthDate_MATCH) && features.get(ATT_birthDate_MATCH) > 0.0) {
				double match2 = features.get(ATT_birthDate_MATCH) * match;
				if (positiveMatch == null || match2 > positiveMatch) {
					positiveMatch = match2;
				}
			}

		}

		if (negativeMatch != null && negativeMatch > 0.0 && (NEGATIVE_PREFERRED || positiveMatch == null || negativeMatch >= positiveMatch)) {
			if (MAX_LIMIT) {
				return Optional.of(Math.max(-1.0, -1.0 * negativeMatch));
			} else {
				return Optional.of(-1.0 * negativeMatch);
			}
		} else if (positiveMatch != null && positiveMatch > 0.0) {
			if (MAX_LIMIT) {
				return Optional.of(Math.min(1.0, positiveMatch));
			} else {
				return Optional.of(positiveMatch);
			}
		} else {
			return Optional.empty();
		}
	}

	public static final ClassId single = new ClassId("http://dbpedia.org/ontology/Single");
	public static final PropertyId releaseDate = new PropertyId("http://dbpedia.org/ontology/releaseDate");
	public static final ClassPropertyId single_releaseDate = new ClassPropertyId(single, releaseDate);

	private Optional<Double> getMatchingRulesScoreSong(Map<String, Double> features, double labelEquivalenceThreshold) {

		String ATT_musicalArtist_NO_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/musicalArtist::NO_MATCH";
		String IMP_musicalArtist_NO_MATCH = "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/musicalArtist::NO_MATCH";
		
		String ATT_musicalArtist_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/musicalArtist::MATCH";
		String IMP_musicalArtist_MATCH = "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/musicalArtist::MATCH";
		

		String ATT_album_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/album::MATCH";
		String IMP_album_MATCH = "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/album::MATCH";
		
		

		String ATT_releaseData_NO_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/releaseDate::NO_MATCH";
 		
		String LABEL = "LABEL";
		
		Double negativeMatch = null;
		
		if (features.containsKey(ATT_musicalArtist_NO_MATCH) && features.get(ATT_musicalArtist_NO_MATCH) > 0.0) {
			double noMatch = features.get(ATT_musicalArtist_NO_MATCH);
			if (negativeMatch == null || noMatch > negativeMatch) {
				negativeMatch = noMatch;
			}
		}

		if (features.containsKey(IMP_musicalArtist_NO_MATCH) && features.get(IMP_musicalArtist_NO_MATCH) > 0.0) {
			double noMatch = features.get(IMP_musicalArtist_NO_MATCH);
			if (negativeMatch == null || noMatch > negativeMatch) {
				negativeMatch = noMatch;
			}
		}
		
		if (features.containsKey(ATT_releaseData_NO_MATCH) && features.get(ATT_releaseData_NO_MATCH) > 0.0) {
			double noMatch = features.get(ATT_releaseData_NO_MATCH);
			if (negativeMatch == null || noMatch > negativeMatch) {
				negativeMatch = noMatch;
			}
		}
		

		Double positiveMatch = null;
		
		if (features.containsKey(LABEL) && features.get(LABEL) > 0.99) {
			double match = features.get(LABEL);

			if (features.containsKey(ATT_musicalArtist_MATCH) && features.get(ATT_musicalArtist_MATCH) > 0.0) {
				double match2 = features.get(ATT_musicalArtist_MATCH) * match;
				if (positiveMatch == null || match2 > positiveMatch) {
					positiveMatch = match2;
				}
			}
			
			if (features.containsKey(IMP_musicalArtist_MATCH) && features.get(IMP_musicalArtist_MATCH) > 0.0) {
				double match2 = features.get(IMP_musicalArtist_MATCH) * match;
				if (positiveMatch == null || match2 > positiveMatch) {
					positiveMatch = match2;
				}
			}

		}
		
		
		if (features.containsKey(LABEL) && features.get(LABEL) > labelEquivalenceThreshold) {
			double match = features.get(LABEL);

			if (features.containsKey(ATT_musicalArtist_MATCH) && features.get(ATT_musicalArtist_MATCH) > 0.0) {
				double match2 = features.get(ATT_musicalArtist_MATCH) * match;
				
				if (features.containsKey(ATT_album_MATCH) && features.get(ATT_album_MATCH) > 0.0) {
					double match3 = features.get(ATT_album_MATCH) * match2;
				
					if (positiveMatch == null || match3 > positiveMatch) {
						positiveMatch = match3;
					}
				}
				
				if (features.containsKey(IMP_album_MATCH) && features.get(IMP_album_MATCH) > 0.0) {
					double match3 = features.get(IMP_album_MATCH) * match2;
				
					if (positiveMatch == null || match3 > positiveMatch) {
						positiveMatch = match3;
					}
				}
			}
			
			if (features.containsKey(IMP_musicalArtist_MATCH) && features.get(IMP_musicalArtist_MATCH) > 0.0) {
				double match2 = features.get(IMP_musicalArtist_MATCH) * match;
				
				if (features.containsKey(ATT_album_MATCH) && features.get(ATT_album_MATCH) > 0.0) {
					double match3 = features.get(ATT_album_MATCH) * match2;
				
					if (positiveMatch == null || match3 > positiveMatch) {
						positiveMatch = match3;
					}
				}
				
				if (features.containsKey(IMP_album_MATCH) && features.get(IMP_album_MATCH) > 0.0) {
					double match3 = features.get(IMP_album_MATCH) * match2;
				
					if (positiveMatch == null || match3 > positiveMatch) {
						positiveMatch = match3;
					}
				}
			}

		}

	

		
		if (negativeMatch != null && negativeMatch > 0.0 && (NEGATIVE_PREFERRED || positiveMatch == null || negativeMatch >= positiveMatch)) {
			if (MAX_LIMIT) {
				return Optional.of(Math.max(-1.0, -1.0 * negativeMatch));
			} else {
				return Optional.of(-1.0 * negativeMatch);
			}
		} else if (positiveMatch != null && positiveMatch > 0.0) {
			if (MAX_LIMIT) {
				return Optional.of(Math.min(1.0, positiveMatch));
			} else {
				return Optional.of(positiveMatch);
			}
		} else {
			return Optional.empty();
		}
	}

	public static boolean COUNTRY_ATT_NO_MATCH = true;
	public static boolean COUNTRY_IMPL_NO_MATCH = true;
	public static boolean POSTAL_CODE_COUNTRY_MATCH = true;
	public static boolean LABEL_ISPARTOF_MATCH = true;
	public static boolean LABEL_POSTAL_CODE_MATCH = true;

	private Optional<Double> getMatchingRulesScoreSettlement(Map<String, Double> features, double labelEquivalenceThreshold) {

		String ATT_country_NO_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/country::NO_MATCH";
		String IMP_country_NO_MATCH = "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/country::NO_MATCH";

		String ATT_country_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/country::MATCH";
		String IMP_country_MATCH = "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/country::MATCH";

		String ATT_isPartOf_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/isPartOf::MATCH";
		String IMP_isPartOf_MATCH = "IMPLICIT_ATTRIBUTES::http://dbpedia.org/ontology/isPartOf::MATCH";

		String ATT_postalCode_MATCH = "ATTRIBUTE::http://dbpedia.org/ontology/postalCode::MATCH";

		String LABEL = "LABEL";

		Double negativeMatch = null;

		if (COUNTRY_ATT_NO_MATCH) {
			if (features.containsKey(ATT_country_NO_MATCH) && features.get(ATT_country_NO_MATCH) > 0.50) {
				double noMatch = features.get(ATT_country_NO_MATCH);
				if (features.containsKey(IMP_country_NO_MATCH) && features.get(IMP_country_NO_MATCH) > 0.90) {

					double noMatch2 = features.get(IMP_country_NO_MATCH);
					if (features.containsKey(IMP_country_MATCH) && features.get(IMP_country_MATCH) < 0.01) {
						if (negativeMatch == null || noMatch > negativeMatch) {
							negativeMatch = noMatch * noMatch2;
						}
					}
				}

			}
		}
		// if (COUNTRY_ATT_NO_MATCH) {
		// if (features.containsKey(ATT_country_NO_MATCH) &&
		// features.get(ATT_country_NO_MATCH) > 0.50) {
		// double noMatch = features.get(ATT_country_NO_MATCH);
		// // if (features.containsKey(IMP_country_NO_MATCH) &&
		// // features.get(IMP_country_NO_MATCH) > 0.90) {
		//
		// // double noMatch2 = features.get(IMP_country_NO_MATCH);
		// if (!features.containsKey(IMP_country_MATCH) ||
		// features.get(IMP_country_MATCH) < features.get(ATT_country_NO_MATCH))
		// {
		// if (negativeMatch == null || noMatch > negativeMatch) {
		// // negativeMatch = noMatch * noMatch2;
		// negativeMatch = noMatch;
		//
		// }
		// }
		// // }
		//
		// }
		// }
		//
		// if (COUNTRY_IMPL_NO_MATCH) {
		//
		// if (features.containsKey(IMP_country_NO_MATCH) &&
		// features.get(IMP_country_NO_MATCH) > 0.90) {
		//
		// double noMatch2 = features.get(IMP_country_NO_MATCH);
		// if (!features.containsKey(IMP_country_MATCH) ||
		// features.get(IMP_country_MATCH) < features.get(IMP_country_NO_MATCH))
		// {
		// if (negativeMatch == null || noMatch2 > negativeMatch) {
		// negativeMatch = noMatch2;
		// }
		// }
		// }
		//
		// }


		Double positiveMatch = null;

		if (POSTAL_CODE_COUNTRY_MATCH) {
			if (features.containsKey(ATT_postalCode_MATCH) && features.get(ATT_postalCode_MATCH) > 0.0) {
				double match = features.get(ATT_postalCode_MATCH);

				if (features.containsKey(ATT_country_MATCH) && features.get(ATT_country_MATCH) > 0.0) {
					double match2 = features.get(ATT_country_MATCH) * match;
					if (positiveMatch == null || match2 > positiveMatch) {
						positiveMatch = match2;
					}
				}

				if (features.containsKey(IMP_country_MATCH) && features.get(IMP_country_MATCH) > 0.0) {
					double match2 = features.get(IMP_country_MATCH) * match;
					if (positiveMatch == null || match2 > positiveMatch) {
						positiveMatch = match2;
					}
				}

			}
		}
		if (LABEL_ISPARTOF_MATCH) {

			if (features.containsKey(LABEL) && features.get(LABEL) > 0.99) {
				double match = features.get(LABEL);



				if (features.containsKey(ATT_isPartOf_MATCH) && features.get(ATT_isPartOf_MATCH) > 0.50) {
					double match3 = features.get(ATT_isPartOf_MATCH);

						if (positiveMatch == null || match3 > positiveMatch) {
							positiveMatch = match3;
						}
					}



			}

		}


		if (LABEL_POSTAL_CODE_MATCH) {
			if (features.containsKey(LABEL) && features.get(LABEL) > labelEquivalenceThreshold) {
				double match = features.get(LABEL);

				if (features.containsKey(ATT_postalCode_MATCH) && features.get(ATT_postalCode_MATCH) > 0.0) {
					double match2 = features.get(ATT_postalCode_MATCH) * match;
					if (positiveMatch == null || match2 > positiveMatch) {
						positiveMatch = match2;
					}
				}

			}
		}

		// if (negativeMatch != null && negativeMatch > 0.0 && positiveMatch !=
		// null & positiveMatch > 0.0) {
		// return Optional.empty();
		// } else
		if (negativeMatch != null && negativeMatch > 0.0 && (NEGATIVE_PREFERRED || positiveMatch == null || negativeMatch >= positiveMatch)) {
			if (MAX_LIMIT) {
				return Optional.of(Math.max(-1.0, -1.0 * negativeMatch));
			} else {
				return Optional.of(-1.0 * negativeMatch);
			}
		} else if (positiveMatch != null && positiveMatch > 0.0) {
			if (MAX_LIMIT) {
				return Optional.of(Math.min(1.0, positiveMatch));
			} else {
				return Optional.of(positiveMatch);
			}
		} else {
			return Optional.empty();
		}
	}

	public static final boolean NEGATIVE_PREFERRED = false;
	public static final boolean MAX_LIMIT = false;

	public double getAdaptedScore(double normalizedScored, double targetThreshold) {
		if (normalizedScored > 0.0) {
			return normalizedScored * (1.0 - targetThreshold) + 0.5;
		} else {
			return (1.0 + normalizedScored) * 0.5;
		}
	}
}
