package expansion.set.entitymatching;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Tri;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.creation.Entity;
import expansion.set.entitymatching.candidateselection.CandidateSelection;
import expansion.set.entitymatching.candidateselection.CandidateSelectionConfig;
import expansion.set.entitymatching.candidateselection.CandidateSelectionRepository;
import expansion.set.entitymatching.candidateselection.CandidateSelectionResults;
import expansion.set.entitymatching.candidateselection.CandidateSelectionStrategy;
import expansion.set.entitymatching.matchers.EntityMatcher;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;
import expansion.set.entitymatching.matchers.EntityMatcherRepository;
import expansion.set.entitymatching.matchers.EntityMatcherStrategy;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.run.EntityMatchingRun;

public class EntityMatchingRunner {

	@Inject
	protected CandidateSelectionRepository candidateSelectionRepository;

	@Inject
	protected EntityMatcherRepository entityMatcherRepository;

	// @Inject
	// protected EntityMatchingWeightLearner learner;

	@Inject
	public EntityMatchingRunner() {
	}


	public EntityMatcherFullResults fullRun(EntityMatcherConfig emConfig, EntityMatchingRun run, List<Entity> entities,
			Map<ClassId, Set<ClassId>> allParents) {
		return fullRun(emConfig, run, entities, allParents, null);
	}

	public EntityMatcherFullResults fullRun(EntityMatcherConfig emConfig, EntityMatchingRun run, List<Entity> entities, Map<ClassId, Set<ClassId>> allParents,
			List<CandidateSelectionStrategy> additionalStrategies) {
		
		
		// additionalStrategies = null;

		Map<CandidateSelectionStrategy, CandidateSelection> csList = new HashMap<>();
		for (CandidateSelectionStrategy csStrategy : run.getCandidateSelection()) {
			csList.put(csStrategy, candidateSelectionRepository.get(csStrategy.getClazz()));
		}
		if (additionalStrategies != null) {
			for (CandidateSelectionStrategy csStrategy : additionalStrategies) {
				csList.put(csStrategy, candidateSelectionRepository.get(csStrategy.getClazz()));
			}
		}

		Map<EntityMatcherStrategy, EntityMatcher> emList = new HashMap<>();
		for (EntityMatcherStrategy emStrategy : run.getEntityMatcher()) {
			emList.put(emStrategy, entityMatcherRepository.get(emStrategy.getClazz()));
		}

		CandidateSelectionConfig csConfig = new CandidateSelectionConfig(allParents);

		ConcurrentLinkedQueue<Tri<Entity, InstanceId, Map<CandidateSelectionStrategy, CandidateSelectionResults>>> secondQueue = new ConcurrentLinkedQueue<>();

		PersistentParallel6.runAndFinish(ListIterable.iterable(entities), (entityEntry) -> {

			Integer index = entityEntry.getIndex();
			Entity entity = entityEntry.getValue();

			Map<InstanceId, Map<CandidateSelectionStrategy, CandidateSelectionResults>> csResultsMap = runCS(entity, csConfig, csList);
			
			for (Entry<InstanceId, Map<CandidateSelectionStrategy, CandidateSelectionResults>> entry : csResultsMap.entrySet()) {
				secondQueue.add(new Tri<>(entity, entry.getKey(), entry.getValue()));
			}
		}, this.getClass());

		Set<InstanceId> candidatesSet = secondQueue.stream().map((e) -> e.getB()).collect(Collectors.toSet());

		for (EntityMatcher em : emList.values()) {
			em.prepare(emConfig, entities, candidatesSet);
		}
		ConcurrentHashMap<RowClusterUUID, Map<InstanceId, EntityMatcherSingleResults>> ret = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(secondQueue, (secondEntry) -> {

			Entity entity = secondEntry.getA();
			InstanceId istanceId = secondEntry.getB();
			Map<CandidateSelectionStrategy, CandidateSelectionResults> csResults = secondEntry.getC();

			Map<InstanceId, EntityMatcherSingleResults> singleResults = CHMHelper.getFromCHM(ret, entity.getRowClusterUUID(), () -> new ConcurrentHashMap<>());

			EntityMatcherSingleResults instanceResults = getEntityMatcherScores(entity, istanceId, csResults, emConfig, emList);
			singleResults.put(istanceId, instanceResults);

		}, this.getClass());

		for (Entity e : entities) {
			if (!ret.containsKey(e.getRowClusterUUID())) {
				ret.put(e.getRowClusterUUID(), new ConcurrentHashMap<>());
			}
		}

		for (EntityMatcher em : emList.values()) {
			em.clean();
		}

		return new EntityMatcherFullResults(emConfig, entities, ret);


	}

	private EntityMatcherSingleResults getEntityMatcherScores(Entity entity, InstanceId instanceId,
			Map<CandidateSelectionStrategy, CandidateSelectionResults> value,
			EntityMatcherConfig emConfig, Map<EntityMatcherStrategy, EntityMatcher> emList) {
		Map<EntityMatcherStrategy, List<WeightedScore>> scores = new LinkedHashMap<>();

		for (Entry<EntityMatcherStrategy, EntityMatcher> emStrategy : emList.entrySet()) {
			scores.put(emStrategy.getKey(), emStrategy.getValue().run(emConfig, entity, instanceId, value));
		}

		return new EntityMatcherSingleResults(scores);
	}

	private Map<InstanceId, Map<CandidateSelectionStrategy, CandidateSelectionResults>> runCS(Entity entity, CandidateSelectionConfig csConfig,
			Map<CandidateSelectionStrategy, CandidateSelection> csList) {
		Map<InstanceId, Map<CandidateSelectionStrategy, CandidateSelectionResults>> csResultsMap = new HashMap<>();
		for (Entry<CandidateSelectionStrategy, CandidateSelection> cs : csList.entrySet()) {
			Optional<Map<InstanceId, CandidateSelectionResults>> csRet = cs.getValue().run(entity, csConfig);
			if (csRet.isPresent()) {
				for (Entry<InstanceId, CandidateSelectionResults> csEntry : csRet.get().entrySet()) {
					CHMHelper.getFromMap(csResultsMap, csEntry.getKey(), () -> new HashMap<>()).put(cs.getKey(), csEntry.getValue());
				}
			}
		}

		return csResultsMap;
	}

	public EntityMatcherWeightedResults weightedRun(EntityMatcherConfig emConfig, EntityMatchingRun run, List<Entity> entities,
			Map<ClassId, Set<ClassId>> allParents,
			EM_LearnedModelRepository weights) {
		// run = EntityMatchingRunRepository.RF_CS_LABEL_EM_ALL;
		return weights.get(run.getOptimizableClass()).apply(fullRun(emConfig, run, entities, allParents));

	}

	public EntityMatcherWeightedResults weightedRun(EntityMatchingRun run, EM_LearnedModelRepository weights, EntityMatcherFullResults fullRun) {
		// run = EntityMatchingRunRepository.RF_CS_LABEL_EM_ALL;
		return weights.get(run.getOptimizableClass()).apply(fullRun);

	}
}
