package expansion.set.entitymatching.matchers.strategies;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Named;

import org.apache.commons.lang.mutable.MutableDouble;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.PropertyValuePair;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import expansion.set.SetExpansionProperties;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.creation.Entity;
import expansion.set.entitymatching.candidateselection.CandidateSelectionResults;
import expansion.set.entitymatching.candidateselection.CandidateSelectionStrategy;
import expansion.set.entitymatching.matchers.EntityMatcher;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;
import expansion.set.implicitAttributes.ImplicitAttributeFinder;
import expansion.set.implicitAttributes.ImplicitAttributes;

public class EntityMatcher_ImplicitAttribute implements EntityMatcher {

	@Inject
	protected ImplicitAttributeFinder finder;

	@Inject
	protected KnowledgeBase knowledgeBase;

	@Inject
	protected ValueEvaluator valueEvaluator;

	@Inject
	protected CHMKryoAdapter kryoAdapter;

	@Inject
	@Named(SetExpansionProperties.tablesCachesFolder)
	protected String cachesFolder;

	@Override
	public List<WeightedScore> run(EntityMatcherConfig config, Entity entity, InstanceId instanceId,
			Map<CandidateSelectionStrategy, CandidateSelectionResults> csResults) {
		List<WeightedScore> ret = new ArrayList<>();

		Map<PropertyValuePair, Double> aggregatedScores = aggregatedCache.get(entity.getRowClusterUUID());
		for(Entry<PropertyValuePair, Double> entry : aggregatedScores.entrySet()) {

			ValueId vid = new ValueId(instanceId, entry.getKey().getPropertyId());
			String readableString = entry.getKey().getPropertyId().readableString();

			Optional<Boolean> correct = valueEvaluator.correct(vid, new QualifiedAbstractObjectHolder(entry.getKey().getAbstractObject(), new ArrayList<>()));
			Optional<Double> score = valueEvaluator.score(vid, entry.getKey().getAbstractObject());

			if (correct.isPresent()) {
				if (correct.get()) {

					if (entry.getKey().getAbstractObject().getType() == ValueType.TIME) {
						ret.add(new WeightedScore(score.get(), entry.getValue() * score.get(), readableString, true));
					} else {
						ret.add(new WeightedScore(1.0, entry.getValue(), readableString, true));
					}

				} else {
					ret.add(new WeightedScore(0.0, entry.getValue(), readableString, false));

				}
			}
		}
		 
		return ret;
	}

	private Map<PropertyValuePair, Double> getAggregatedScores(RowCluster rowCluster, Map<ClassId, Set<ClassId>> allParents, Set<PropertyId> nonTailProperties) {

			List<TableRowId> rowIds = rowCluster.getRows();

			Map<PropertyValuePair, MutableDouble> sum = new HashMap<>();

			for(TableRowId rowId : rowIds) {
				ImplicitAttributes scores = finder.get(rowId.getTableId(), allParents, nonTailProperties);
				for (Entry<PropertyValuePair, Double> entry : scores.getImplicitAttributes().entrySet()) {
					
					CHMHelper.getFromMap(sum, entry.getKey(), () -> new MutableDouble()).add(entry.getValue());
				}
			}
			
			double threshold = 0.0; // 0.5 * rowIds.size();
			
			Map<PropertyValuePair, Double> ret = sum.entrySet().stream().filter((e) -> e.getValue().doubleValue() > threshold)
					.collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().doubleValue() / rowIds.size()));
			return ret;

	}

	private final static SimpleLockableConcurrentCache<RowClusterUUID, Map<PropertyValuePair, Double>> aggregatedCache = new SimpleLockableConcurrentCache<>();

	private final static SimpleLockableConcurrentCache<UuidStringId, Map<AnnotatedTableId, ImplicitAttributes>> tableCache = new SimpleLockableConcurrentCache<>();


	@Override
	public void prepare(EntityMatcherConfig config, List<Entity> entities, Set<InstanceId> candidatesSet) {

		String cacheFolder = cachesFolder + "/ImplicitAttributes/";
		MKDIR.createDirUnchecked(new File(cacheFolder));

		Map<AnnotatedTableId, ImplicitAttributes> tableScores = tableCache.getOrCreate(new UuidStringId(config.getClustererInput().getClassId().getFinalPart()),
				() -> {
					String PREFIX = config.getClustererInput().getClassId().getFinalPart() + "_";
			if (kryoAdapter.cacheExists(cacheFolder, PREFIX)) {
				return kryoAdapter.read(AnnotatedTableId.class, ImplicitAttributes.class, cacheFolder, PREFIX);
			} else {
				Map<AnnotatedTableId, ImplicitAttributes> tableScoresInner = new ConcurrentHashMap<>();
				Set<AnnotatedTableId> tableIds = config.getClustererInput().getTableIds().getTableIds();

				PersistentParallel6.runAndFinish(tableIds, (tableId) -> {
					tableScoresInner.put(tableId, finder.find(tableId, config.getAllParents(), config.getNonTailProperties(), true));
				}, this.getClass());

				kryoAdapter.write(cacheFolder, PREFIX, tableScoresInner);
				return tableScoresInner;

			}
		});



		ConcurrentHashMap<TableRowId, RowClusterUUID> firstStep = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(entities, (e) -> {
			for (TableRowId rowId : e.getRowCluster().getRows()) {
				firstStep.put(rowId, e.getRowClusterUUID());
			}
		}, this.getClass());

		ConcurrentHashMap<RowClusterUUID, ConcurrentHashMap<PropertyValuePair, Collection<Double>>> intermediary = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(firstStep.entrySet(), (row) -> {
			ImplicitAttributes scores = tableScores.get(row.getKey().getTableId());
			ConcurrentHashMap<PropertyValuePair, Collection<Double>> innerMap = CHMHelper.getFromCHM(intermediary, row.getValue(),
					() -> new ConcurrentHashMap<>());

			for (Entry<PropertyValuePair, Double> entry : scores.getImplicitAttributes().entrySet()) {

				CHMHelper.getFromMap(innerMap, entry.getKey(), () -> new ConcurrentLinkedQueue<>()).add(entry.getValue());
			}

		}, this.getClass());

		// Map<RowClusterUUID, Map<PropertyValuePair, Double>> cache = new
		// ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(intermediary.entrySet(), (cluster) -> {
			aggregatedCache.add(cluster.getKey(), cluster.getValue().entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				double sum = 0.0;
				int count = 0;
				for (Double score : e.getValue()) {
					sum += score;
					count++;
				}
				return sum / count;
			})));
		}, this.getClass());


		// aggregatedCache.add(config.getClustererInput().getUuid(), cache);

	}

	@Override
	public void clean() {
		aggregatedCache.clear();
	}

}
