package expansion.set.entitymatching.modelLearning.weightsGeneric;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.nio.channels.IllegalSelectorException;
import java.util.ArrayList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.Function;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.entitymatching.EntityMatcherEquationResult;
import expansion.set.entitymatching.EntityMatcherFullResults;
import expansion.set.entitymatching.EntityMatcherSingleResults;
import expansion.set.entitymatching.EntityMatcherWeightedResults;
import expansion.set.entitymatching.modelLearning.EntityMatchingModel;
import expansion.set.entitymatching.modelLearning.simpleWeights.SimpleEntityMatcherEquation;
import expansion.set.entitymatching.modelLearning.weights.EntityMatcherEquation;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.evaluation.matching.ScoredEntity;
import weka.core.SerializationHelper;

public class EM_WeightsGenericModel extends EntityMatchingModel {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private final EntityMatchingRun run;
	private final WeightsRepository weights;
	private transient EM_WeightsGenericEquation equation;
	private final EquationType equationType;

	private static enum EquationType {
		COMPLEX, SIMPLE;
	}

	public void startup() {
		if (equation == null) {
			if (equationType == EquationType.SIMPLE) {
				equation = new SimpleEntityMatcherEquation();
			} else if (equationType == EquationType.COMPLEX) {
				equation = new EntityMatcherEquation();
			}
		}
	}

	public EM_WeightsGenericModel(byte[] input) {
		ByteArrayInputStream stream = new ByteArrayInputStream(input);
		try {
			EM_WeightsGenericModel readModel = (EM_WeightsGenericModel) weka.core.SerializationHelper.read(stream);

			this.run = readModel.run;
			this.weights = readModel.weights;
			this.equation = readModel.equation;
			if (equation instanceof SimpleEntityMatcherEquation) {
				equationType = EquationType.SIMPLE;
			} else if (equation instanceof EntityMatcherEquation) {
				equationType = EquationType.COMPLEX;
			} else {
				throw new IllegalStateException();
			}

		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	public EM_WeightsGenericModel(EntityMatchingRun run, WeightsRepository weights, EM_WeightsGenericEquation equation) {
		this.run = run;
		this.weights = weights;
		this.equation = equation;
		if (equation instanceof SimpleEntityMatcherEquation) {
			equationType = EquationType.SIMPLE;
		} else if (equation instanceof EntityMatcherEquation) {
			equationType = EquationType.COMPLEX;
		} else {
			throw new IllegalSelectorException();
		}
	}

	@Override
	public EntityMatcherWeightedResults apply(EntityMatcherFullResults fullRun) {

		ConcurrentHashMap<InstanceId, ConcurrentLinkedQueue<ScoredEntity>> mappingForExistingClusters = new ConcurrentHashMap<>();
		ConcurrentHashMap<InstanceId, ConcurrentLinkedQueue<ScoredEntity>> unsure = new ConcurrentHashMap<>();
		ConcurrentLinkedQueue<ScoredEntity> newEntities = new ConcurrentLinkedQueue<>();

		PersistentParallel6.runAndFinish(fullRun.getEntities().entrySet(), (entityEntry) -> {
			Map<InstanceId, EntityMatcherSingleResults> candidates = fullRun.getResults().get(entityEntry.getKey());

			Tri<EntityMatcherEquationResult, InstanceId, Double> eqResult = equation.runEquation(candidates, weights, run.getOptimizableClass(),
					run.getEntityMatcher());

			switch (eqResult.getA()) {
			case BAD:
				CHMHelper.getFromCHM(unsure, eqResult.getB(), () -> new ConcurrentLinkedQueue<>())
						.add(new ScoredEntity(entityEntry.getValue(), eqResult.getC()));
				break;
			case EXISTING:
				CHMHelper.getFromCHM(mappingForExistingClusters, eqResult.getB(), () -> new ConcurrentLinkedQueue<>())
						.add(new ScoredEntity(entityEntry.getValue(), eqResult.getC()));
				break;
			case NEW:
				newEntities.add(new ScoredEntity(entityEntry.getValue(), eqResult.getC()));
				break;
			default:
				throw new IllegalStateException();

			}
		}, this.getClass());

		return new EntityMatcherWeightedResults(
				mappingForExistingClusters.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new ArrayList<>(e.getValue()))),

				unsure.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new ArrayList<>(e.getValue()))), new ArrayList<>(newEntities));
	}

	@Override
	public byte[] outputModelFile() {
		ByteArrayOutputStream stream = new ByteArrayOutputStream();

		try {

			SerializationHelper.write(stream, this);

			return stream.toByteArray();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	public WeightsRepository getWeights() {
		return weights;
	}

	@Override
	public Function<InstanceId, Double> getOtherModelSupplier(Map<InstanceId, EntityMatcherSingleResults> candidates) {
		Map<InstanceId, Double> precomputedMap = equation.getNormalized(candidates, weights, run.getOptimizableClass(), run.getEntityMatcher());

		return precomputedMap::get;
	}

	@Override
	public String toString() {
		return "EM_WeightsGenericModel [weights=" + weights + "]";
	}

	@Override
	public Map<Pair<RowClusterUUID, InstanceId>, Pair<Boolean, Double>> getFullResults(EntityMatcherFullResults fullRun) {
		throw new UnsupportedOperationException();
	}

}
