package expansion.set.entitymatching.modelLearning.weka;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.inject.Inject;
import javax.inject.Provider;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.DataSetData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearningFeatures;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumDouble;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;
import de.uni_mannheim.informatik.dws.expansion.utils.learning.weka.WekaLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.learning.weka.WekaLearnerCreator;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities.PairLearningData;
import expansion.set.entitymatching.EntityMatcherFullResults;
import expansion.set.entitymatching.EntityMatcherSingleResults;
import expansion.set.entitymatching.matchers.EntityMatcherStrategy;
import expansion.set.entitymatching.modelLearning.EntityMatchingModel;
import expansion.set.entitymatching.modelLearning.EntityMatchingModelLearnerRepository;
import expansion.set.entitymatching.modelLearning.EntityMatchingModelLearnerRunner;
import expansion.set.entitymatching.modelLearning.EntityMatchingUpsamplingHelper;
import expansion.set.entitymatching.modelLearning.simpleWeights.SimpleEntityMatchingWeightLearner;
import expansion.set.entitymatching.modelLearning.weightsGeneric.EntityMatchingGenericWeightsOptimizier;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.entitymatching.run.EntityMatchingRunRepository;

public class EntityMatcherWekaLearner implements EntityMatchingModelLearnerRunner {
	public static final boolean DEFAULT_UPSAMPLE = false;
	private final Set<EntityMatcherStrategy> withConfidence;
	private final String classifierName;
	private final String[] paramters;
	private final boolean numeric;
	private final Class<? extends EntityMatchingModelLearnerRunner> otherModel;
	private final boolean rangeDescretization;
	private final WeightsRepository fixedWeights;
	private final boolean extraProperties;
	private final boolean upsample;

	@Inject
	protected SimpleEntityMatchingWeightLearner simpleLearner;

	@Inject
	protected EntityMatchingGenericWeightsOptimizier generic;

	@Inject
	protected Provider<EntityMatchingModelLearnerRepository> repository;

	public EntityMatcherWekaLearner(Set<EntityMatcherStrategy> withConfidence, String classifierName, String[] paramters, boolean upsample) {
		this(withConfidence, classifierName, paramters, true, null, upsample);
	}

	public EntityMatcherWekaLearner(Set<EntityMatcherStrategy> withConfidence, String classifierName, String[] paramters, boolean numeric, boolean upsample) {
		this(withConfidence, classifierName, paramters, numeric, null, upsample);
	}

	public EntityMatcherWekaLearner(Set<EntityMatcherStrategy> withConfidence, String classifierName, String[] paramters, boolean numeric,
			Class<? extends EntityMatchingModelLearnerRunner> otherModel, boolean upsample) {
		this(withConfidence, classifierName, paramters, numeric, otherModel, false, upsample);
	}

	public EntityMatcherWekaLearner(Set<EntityMatcherStrategy> withConfidence, String classifierName, String[] paramters, boolean numeric,
			Class<? extends EntityMatchingModelLearnerRunner> otherModel, boolean rangeDescretization, boolean upsample) {
		this(withConfidence, classifierName, paramters, numeric, otherModel, rangeDescretization, upsample, false);

	}

	public EntityMatcherWekaLearner(Set<EntityMatcherStrategy> withConfidence, String classifierName, String[] paramters, boolean numeric,
			Class<? extends EntityMatchingModelLearnerRunner> otherModel, boolean rangeDescretization, boolean upsample, boolean extraProperties) {
		super();
		this.withConfidence = Collections.unmodifiableSet(new HashSet<>(withConfidence));
		this.classifierName = classifierName;
		this.paramters = Arrays.copyOf(paramters, paramters.length);
		this.numeric = numeric;
		this.otherModel = otherModel;
		this.rangeDescretization = rangeDescretization;
		this.fixedWeights = null;
		this.extraProperties = extraProperties;
		this.upsample = upsample;

	}

	public EntityMatcherWekaLearner(Set<EntityMatcherStrategy> withConfidence, String classifierName, String[] paramters, boolean numeric,
			Class<? extends EntityMatchingModelLearnerRunner> otherModel, boolean rangeDescretization, boolean upsample, WeightsRepository fixedWeights) {
		super();
		this.withConfidence = Collections.unmodifiableSet(new HashSet<>(withConfidence));
		this.classifierName = classifierName;
		this.paramters = Arrays.copyOf(paramters, paramters.length);
		this.numeric = numeric;
		this.otherModel = otherModel;
		this.rangeDescretization = rangeDescretization;
		this.fixedWeights = fixedWeights;
		this.extraProperties = false;
		this.upsample = upsample;
	}

	public EntityMatcherWekaLearner(Set<EntityMatcherStrategy> withConfidence, String classifierName, String[] paramters, boolean numeric,
			Class<? extends EntityMatchingModelLearnerRunner> otherModel, boolean rangeDescretization, boolean upsample, WeightsRepository fixedWeights,
			boolean extraProperties) {
		super();
		this.withConfidence = Collections.unmodifiableSet(new HashSet<>(withConfidence));
		this.classifierName = classifierName;
		this.paramters = Arrays.copyOf(paramters, paramters.length);
		this.numeric = numeric;
		this.otherModel = otherModel;
		this.rangeDescretization = rangeDescretization;
		this.fixedWeights = fixedWeights;
		this.extraProperties = extraProperties;
		this.upsample = upsample;
	}

	public EntityMatcherWekaLearner(Set<EntityMatcherStrategy> withConfidence, String classifierName, String[] paramters, boolean numeric,
			Class<? extends EntityMatchingModelLearnerRunner> otherModel, boolean rangeDescretization, boolean upsample, WeightsRepository fixedWeights,
			boolean extraProperties,
			SimpleEntityMatchingWeightLearner simpleLearner, EntityMatchingGenericWeightsOptimizier generic,
			Provider<EntityMatchingModelLearnerRepository> repository) {
		super();
		this.withConfidence = Collections.unmodifiableSet(new HashSet<>(withConfidence));
		this.classifierName = classifierName;
		this.paramters = Arrays.copyOf(paramters, paramters.length);
		this.numeric = numeric;
		this.otherModel = otherModel;
		this.rangeDescretization = rangeDescretization;
		this.fixedWeights = fixedWeights;
		this.extraProperties = extraProperties;
		this.simpleLearner = simpleLearner;
		this.generic = generic;
		this.repository = repository;
		this.upsample = upsample;
	}

	private DataSetData getDataset(EntityMatchingRun run, boolean numeric2, Set<PropertyId> propertyIds) {



		List<String> featuresList = new ArrayList<>();
		for (EntityMatcherStrategy sim : run.getEntityMatcher()) {

			if (withConfidence.contains(sim)) {
				if (extraProperties) {
					for (PropertyId propertyId : propertyIds) {
						featuresList.add(sim.getName() + "::" + propertyId.readableString() + "::MATCH");
						featuresList.add(sim.getName() + "::" + propertyId.readableString() + "::NO_MATCH");
					}
				}

				else {
					featuresList.add(sim.getName());
					// if
					// (!sim.equals(EntityMatcherStrategies.IMPLICIT_ATTRIBUTE))
					// {
						featuresList.add(sim.getName() + "::CONFIDENCE");
					// }
				}

				// if (sim.equals(EntityMatcherStrategies.IMPLICIT_ATTRIBUTE)) {
				// featuresList.add(sim.getName() + "::SUM");
				// }
			}
			else {
				featuresList.add(sim.getName());
			}
		}

		if (numeric) {
			Map<String, Double> labels = new HashMap<>();
			labels.put(LearnableRowClusterSimilarities.getLabel(true), 1.0);
			labels.put(LearnableRowClusterSimilarities.getLabel(false), -1.0);
			return new DataSetData(run.getName(), labels, new LearningFeatures(featuresList));
		} else {
			List<String> classLabels = new ArrayList<>();
			classLabels.add(LearnableRowClusterSimilarities.getLabel(true));
			classLabels.add(LearnableRowClusterSimilarities.getLabel(false));
			return new DataSetData(run.getName(), classLabels, new LearningFeatures(featuresList));
		}
	}

	@Override
	public EntityMatchingModel learnModel(EntityMatchingRun run, Set<InstanceId> newEntities,
			Map<RowClusterUUID, Optional<InstanceId>> gsMapped,
			EntityMatcherFullResults fullRun) {
		// String[] paramters = new String[] { "-U", "-attribute-importance",
		// "-I", "3000", "-num-slots", "100", "-O", "-P", "85" };
		DataSetData datasetData = getDataset(run, numeric, fullRun.getEmConfig().getNonTailProperties());
		
		Map<String, Double> labels = new HashMap<>();
		labels.put(LearnableRowClusterSimilarities.getLabel(true), 1.0);
		labels.put(LearnableRowClusterSimilarities.getLabel(false), -1.0);
		
		
		
		
		EntityMatchingUpsamplingHelper upsamplingHelper = new EntityMatchingUpsamplingHelper(newEntities, gsMapped);
		
		ConcurrentAccumulator<AccumDouble> matchClass = new ConcurrentAccumulator<>(AccumDouble.empty());
		ConcurrentAccumulator<AccumDouble> noMatchClass = new ConcurrentAccumulator<>(AccumDouble.empty());

		PersistentParallel6.runAndFinish(fullRun.getResults().entrySet(), (e) -> {
			boolean existing = false;
			if (gsMapped.get(e.getKey()).isPresent() && !newEntities.contains(gsMapped.get(e.getKey()).get())) {
				existing = true;
			}
			double weight;
			if (existing) {
				weight = upsamplingHelper.getUpsampleExisting();
			} else {
				weight = upsamplingHelper.getUpsampleNew();
			}
			weight = 1.0;

			for (Entry<InstanceId, EntityMatcherSingleResults> innerEntry : e.getValue().entrySet()) {

				if (gsMapped.containsKey(e.getKey())) {
					Optional<InstanceId> opt = gsMapped.get(e.getKey());
					if (opt.isPresent() && opt.get().equals(innerEntry.getKey())) {
						matchClass.add(new AccumDouble(weight));
					} else {
						noMatchClass.add(new AccumDouble(weight));

					}
				} else {
					throw new IllegalStateException();
				}
			}
		}, this.getClass(), PersistentParallel6.AlreadyParalell.isCurrentThreadParallel());

		double matchClassSum = matchClass.get().get();
		double noMatchClassSum = noMatchClass.get().get();

		double max = matchClassSum;
		if (noMatchClassSum > max) {
			max = noMatchClassSum;
		}

		double upsampleMatch = max / matchClassSum;
		double upsampleNoMatch = max / noMatchClassSum;
		// System.out.println("Upsample Match: " + upsampleMatch + ", Upsample
		// No Match: " + upsampleNoMatch);

		ConcurrentLinkedQueue<RecordData> recordData = new ConcurrentLinkedQueue<>();

		PersistentParallel6.runAndFinish(fullRun.getResults().entrySet(), (e) ->  {
			boolean existing = false;
			if (gsMapped.get(e.getKey()).isPresent() && !newEntities.contains(gsMapped.get(e.getKey()).get())) {
				existing = true;
			}
			
			// double weight;
			// if (existing) {
			// weight = upsamplingHelper.getUpsampleExisting();
			//// weight = 4.0;
			// } else {
			// weight = upsamplingHelper.getUpsampleNew();
			//// weight = 1.0;
			// }
			// if (!upsample) {
			// weight = 1.0;
			// }
			Map<InstanceId, RecordFeature> featureVectors = getFeatureVectors(run.getEntityMatcher(), e.getValue(), datasetData, withConfidence,
					rangeDescretization, (extraProperties ? fullRun.getEmConfig().getNonTailProperties() : null));

			for(Entry<InstanceId, EntityMatcherSingleResults> innerEntry : e.getValue().entrySet()) {
				
				boolean match;
				double weight;
				String label;
				if(gsMapped.containsKey(e.getKey())) {
					Optional<InstanceId> opt = gsMapped.get(e.getKey());
					if (opt.isPresent() && opt.get().equals(innerEntry.getKey())) {
						label = LearnableRowClusterSimilarities.getLabel(true);
						match = true;
						weight = 1.0;
					}
					else {
						label = LearnableRowClusterSimilarities.getLabel(false);
						match = false;
						weight = 1.0;
					}
				}
				else {
					throw new IllegalStateException();
				}
				
				
				RecordFeature rf = featureVectors.get(innerEntry.getKey());
				// int rounded = (int) Math.round(weight);

				RecordData pld = new RecordData(label, weight, rf);
				// for (int i = 0; i < rounded; i++) {
					recordData.add(pld);
				// }
				
				
			}
			
		}, this.getClass(), PersistentParallel6.AlreadyParalell.isCurrentThreadParallel());
		
		int positivePairsSize = 0;
		int negativePairsSize = 0;

		for (RecordData v : recordData) {
			if (v.getLabel().equals(LearnableRowClusterSimilarities.MATCH_CLASS)) {
				positivePairsSize++;
			} else {
				negativePairsSize++;
			}
		}

		if (false) {
		System.out.println("For Class: " + fullRun.getEmConfig().getClustererInput().getClassId() + ", ForMethod:" + run.getName() + " Positive Pairs "
				+ positivePairsSize + ", Negative Pairs: " + negativePairsSize);
		}

		WekaLearnerCreator creator = new WekaLearnerCreator(classifierName, Arrays.copyOf(paramters, paramters.length));
		WekaLearnedModel model = creator.learn(new TrainingData(datasetData, recordData), (e) -> {
			return 0.0;
		});

		EntityMatchingModel learnedModel = null;
		
		if (otherModel != null) {
			learnedModel = repository.get().get(otherModel).learnModel(run, newEntities, gsMapped, fullRun);

		}


		EntityMatchingWekaModel modelTemp = new EntityMatchingWekaModel(run, datasetData, withConfidence, numeric,
				model, null, learnedModel, rangeDescretization, (extraProperties ? fullRun.getEmConfig().getNonTailProperties() : null));
		WeightsRepository finalWeights = fixedWeights;
		if (finalWeights == null) {
			finalWeights = generic.run(gsMapped, newEntities, (weights) -> {
				return modelTemp.get(weights, fullRun);
			}, (weights) -> {
				double newThreshold = weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_NEW", 0.375));
				double matchThreshold = weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH", 0.40));
				// double match = (newThreshold + ((1.0 - newThreshold) *
				// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH",
				// 0.50))));
				// double ret = Math.abs((0.0 - newThreshold)) / 100.0 +
				// Math.abs((1.0 - match)) / 100.0;
				// double ret2 = (1.0 -
				// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_NEW",
				// 0.00))) / 400.0
				// + (1.0 -
				// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH",
				// 0.00))) / 400.0;
				// double ret3 = -1.0 * ((newThreshold +
				// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH",
				// 0.50))) / 200.0);
				// // return 0.0;
				// double ret4 = (1.0 -
				// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH",
				// 0.75))) / 100.0
				// +
				// (weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_NEW",
				// 0.50))) / 100.0;

				// if(matchThreshold < 0.25) {
				// return matchThreshold * 0.001;
				// }
				// else {
				// return 0.0;
				// }

				double matchThresholdFinal = (newThreshold + ((1.0 - newThreshold) * matchThreshold)) * 2.0 - 1.0;

				double penalize = 0.0;
				if (newThreshold > 0.45) {
					penalize += (newThreshold - 0.45 * 0.0001) * 2;
				}
				if (matchThresholdFinal < 0.10) {
					penalize += (-1.0 - 0.10) * 0.0001;
				}
				// double matchPenalize = 0.0;

				return 0.0;
			}, true);
		}

		if (run == EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL) {
			System.out.println(model.toString());
			System.out.println(learnedModel.toString());
			System.out.println(finalWeights.toString());
		}

		if (run == EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL) {
			System.out.println(model.toString());
			// System.out.println(learnedModel.toString());
			System.out.println(finalWeights.toString());
		}
		return new EntityMatchingWekaModel(run, datasetData, withConfidence, numeric, model, finalWeights, learnedModel, rangeDescretization,
				(extraProperties ? fullRun.getEmConfig().getNonTailProperties() : null));
	}

	public EntityMatchingModel getCustomModel(EntityMatchingRun run, WekaLearnedModel model, WeightsRepository weights, EntityMatchingModel secondaryModel,
			Set<PropertyId> propertyIds) {
		DataSetData datasetData = getDataset(run, numeric, null);

		return new EntityMatchingWekaModel(run, datasetData, withConfidence, numeric, model, weights, secondaryModel, rangeDescretization, propertyIds);

	}
	static Map<InstanceId, RecordFeature> getFeatureVectors(List<EntityMatcherStrategy> emStrategies, Map<InstanceId, EntityMatcherSingleResults> value,
			DataSetData datasetData, Set<EntityMatcherStrategy> withConfidence, boolean rangeDescretize, Set<PropertyId> propertyIds) {
		Map<EntityMatcherStrategy, Double> diffRankMap = new HashMap<>();
		Map<EntityMatcherStrategy, Double> minRankMap = new HashMap<>();

		for (EntityMatcherStrategy emStrategy : emStrategies) {

			if (emStrategy.isRanked()) {
				Double maxRank = null;
				Double minRank = null;
				Double average = null;
				int count = 0;
				for (Entry<InstanceId, EntityMatcherSingleResults> instanceEntry : value.entrySet()) {
					for (WeightedScore ws : instanceEntry.getValue().getScores().get(emStrategy)) {
						if (average == null) {
							average = ws.getScore();
							count++;
						} else {
							average += ws.getScore();
							count++;
						}
						if (maxRank == null || ws.getScore() > maxRank) {
							maxRank = ws.getScore();
						}
						if (minRank == null || ws.getScore() < minRank) {
							minRank = ws.getScore();
						}
					}
				}
				if (average != null && minRank != null) {
					double diff = (average / count) - minRank;
					if (diff > 0.01) {
						diffRankMap.put(emStrategy, diff);
						minRankMap.put(emStrategy, minRank);
					}

				}
			}
		}
		Map<InstanceId, RecordFeature> ret = new HashMap<>();
		for (Entry<InstanceId, EntityMatcherSingleResults> entry : value.entrySet()) {
			Map<String, Double> map = new HashMap<>();

			for (Entry<EntityMatcherStrategy, List<WeightedScore>> feature : entry.getValue().getScores().entrySet()) {



				List<WeightedScore> simList = feature.getValue();

				if (propertyIds != null && withConfidence.contains(feature.getKey())) {
					for (WeightedScore ws : simList) {
						if (ws.isMatch()) {
							map.put(feature.getKey().getName() + "::" + ws.getSuffix().get() + "::MATCH", ws.getWeight());
						} else {
							map.put(feature.getKey().getName() + "::" + ws.getSuffix().get() + "::NO_MATCH", ws.getWeight());
						}
					}
				}
				else {
					double confidence = 0.0;
					double sum = 0.0;
					double count = 0.0;
					for (WeightedScore ws : simList) {
						confidence += ws.getWeight();
						count += 1.0;

						if (feature.getKey().isRanked()) {
							if (diffRankMap.containsKey(feature.getKey())) {
								sum += Math.round(10 * Math.min(1.0, (ws.getScore() - minRankMap.get(feature.getKey())) / diffRankMap.get(feature.getKey())))
										/ 10.0;
								int sdf = 123;
							} else {
								sum += 1.0 * ws.getWeight();
							}
						} else {
							sum += ws.getScore() * ws.getWeight();
						}
					}
					// if (confidence > 0.0) {
					if (confidence > 0.0) {

						double penalization = 1.0;
						if (confidence < 1.0) {
							penalization = confidence;
						}
						{
							// if
							// (!feature.getKey().equals(EntityMatcherStrategies.IMPLICIT_ATTRIBUTE))
							// {
							map.put(feature.getKey().getName(), (sum / count));
							// } else {
							// map.put(feature.getKey().getName(), (sum));
							// }

							if (withConfidence.contains(feature.getKey())) {
								// confidence = Math.round(confidence);
								// if (confidence > 5.0) {
								// confidence = 5.0;
								// }
								// if
								// (!feature.getKey().equals(EntityMatcherStrategies.IMPLICIT_ATTRIBUTE))
								// {
								map.put(feature.getKey().getName() + "::CONFIDENCE", confidence);
								// }
								// if
								// (feature.getKey().equals(EntityMatcherStrategies.IMPLICIT_ATTRIBUTE))
								// {
								// map.put(feature.getKey().getName() + "::SUM",
								// sum);
								// }
							}

						}

					} else {
						map.put(feature.getKey().getName(), 0.0);
						if (withConfidence.contains(feature.getKey())) {
							// if
							// (!feature.getKey().equals(EntityMatcherStrategies.IMPLICIT_ATTRIBUTE))
							// {
								map.put(feature.getKey().getName() + "::CONFIDENCE", confidence);
							// }

							// if
							// (feature.getKey().equals(EntityMatcherStrategies.IMPLICIT_ATTRIBUTE))
							// {
							// map.put(feature.getKey().getName() + "::SUM",
							// 0.0);
							// }
							// map.put(feature.getKey().getName() + "::SUM",
							// 0.0);
						}
					}
				}

			}
			if (rangeDescretize) {
				map = RecordFeature.descretizeMap(map, 8);
			}

			for (String feature : datasetData.getFeatures().getFeatures().values()) {
				if (!map.containsKey(feature)) {
					map.put(feature, 0.0);
				}
			}

			ret.put(entry.getKey(), new RecordFeature(map, datasetData));
		}

		return ret;
	}

	private final static boolean REPLICATION = false;
	@Override
	public EntityMatchingModel learnModel(EntityMatchingRun run, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData> pairs,
			EntityMatcherFullResults fullRun
	 ,
	 Set<InstanceId> newEntities,
	 Map<RowClusterUUID, Optional<InstanceId>> gsMapped
	) {

		DataSetData datasetData = getDataset(run, numeric, fullRun.getEmConfig().getNonTailProperties());

		ConcurrentLinkedQueue<RecordData> recordData = new ConcurrentLinkedQueue<>();
		double positive = 0.0;
		double negative = 0.0;

		Map<RowClusterUUID, Map<InstanceId, PairLearningData>> index = new HashMap<>();
		for (Entry<Pair<RowClusterUUID, InstanceId>, PairLearningData> entry : pairs.entrySet()) {
			CHMHelper.getFromMap(index, entry.getKey().getFirst(), () -> new HashMap<>()).put(entry.getKey().getSecond(), entry.getValue());
			if (entry.getValue().isMatch()) {
				positive += entry.getValue().getWeight();
			} else {
				if (entry.getValue().getWeight() > 1.0) {
					int sdf = 123;
				}
				negative += entry.getValue().getWeight();
			}
		}
		double max = Math.max(positive, negative);
		double positiveUpsample = max / positive;
		double negativeUpsample = max / negative;

		PersistentParallel6.runAndFinish(index.entrySet(), (e) -> {

			Map<InstanceId, EntityMatcherSingleResults> restults = fullRun.getResults().get(e.getKey());

			Map<InstanceId, RecordFeature> featureVectors = getFeatureVectors(run.getEntityMatcher(), restults, datasetData, withConfidence,
					rangeDescretization, (extraProperties ? fullRun.getEmConfig().getNonTailProperties() : null));

			for (Entry<InstanceId, PairLearningData> innerEntry : e.getValue().entrySet()) {

				String label = LearnableRowClusterSimilarities.getLabel(innerEntry.getValue().isMatch());
				double weight = innerEntry.getValue().getWeight();
				double upsample = (innerEntry.getValue().isMatch() ? positiveUpsample : negativeUpsample);
				upsample = 1.0;

				if (REPLICATION) {
					int upsampleRounded = (int) Math.round(upsample);
					for (int i = 0; i < upsampleRounded; i++) {
						RecordFeature rf = featureVectors.get(innerEntry.getKey());

						RecordData pld = new RecordData(label, weight, rf);

						recordData.add(pld);
					}

				} else {
					RecordFeature rf = featureVectors.get(innerEntry.getKey());

					RecordData pld = new RecordData(label, weight * upsample, rf);

					recordData.add(pld);
				}

			}

		}, this.getClass());

		WekaLearnerCreator creator = new WekaLearnerCreator(classifierName, Arrays.copyOf(paramters, paramters.length));
		WekaLearnedModel model = creator.learn(new TrainingData(datasetData, recordData), (e) -> {
			return 0.0;
		});
		EntityMatchingModel learnedModel = null;

		if (otherModel != null) {
			throw new IllegalStateException();
		}

		EntityMatchingWekaModel modelTemp = new EntityMatchingWekaModel(run, datasetData, withConfidence, numeric, model, null, learnedModel,
				rangeDescretization, (extraProperties ? fullRun.getEmConfig().getNonTailProperties() : null));
		WeightsRepository finalWeights = fixedWeights; 
		if (finalWeights == null) {
			finalWeights = generic.run(gsMapped, newEntities, (weights) -> {
				return modelTemp.get(weights, fullRun);
			}, (weights) -> {
				if (false) {
				double newThreshold = weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_NEW", 0.375));
				double matchThreshold = weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH", 0.40));
				// double match = (newThreshold + ((1.0 - newThreshold) *
				// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH",
				// 0.50))));
				// double ret = Math.abs((0.0 - newThreshold)) / 100.0 +
				// Math.abs((1.0 - match)) / 100.0;
				// double ret2 = (1.0 -
				// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_NEW",
				// 0.00))) / 400.0
				// + (1.0 -
				// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH",
				// 0.00))) / 400.0;
				// double ret3 = -1.0 * ((newThreshold +
				// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH",
				// 0.50))) / 200.0);
				// // return 0.0;
				// double ret4 = (1.0 -
				// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH",
				// 0.75))) / 100.0
				// +
				// (weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_NEW",
				// 0.50))) / 100.0;

				// if(matchThreshold < 0.25) {
				// return matchThreshold * 0.001;
				// }
				// else {
				// return 0.0;
				// }

				double matchThresholdFinal = (newThreshold + ((1.0 - newThreshold) * matchThreshold)) * 2.0 - 1.0;

				double penalize = 0.0;
				if (newThreshold > 0.45) {
					penalize += (newThreshold - 0.45 * 0.0001) * 2;
				}
				if (matchThresholdFinal < 0.10) {
					penalize += (-1.0 - 0.10) * 0.0001;
				}
				// double matchPenalize = 0.0;
				}
				return 0.0;
			}, false);
		}

		


		return new EntityMatchingWekaModel(run, datasetData, withConfidence, numeric, model, finalWeights, learnedModel, rangeDescretization,
				(extraProperties ? fullRun.getEmConfig().getNonTailProperties() : null));

	}

	// public static RecordFeature getFeatureVector(EntityMatcherSingleResults
	// value, DataSetData datasetData, Set<EntityMatcherStrategy>
	// withConfidence) {
	//
	// Map<String, Double> map = new HashMap<>();
	//
	// for (Entry<EntityMatcherStrategy, List<WeightedScore>> feature :
	// value.getScores().entrySet()) {
	//
	// if (withConfidence.contains(feature.getKey())) {
	//
	// }
	//
	// List<WeightedScore> simList = feature.getValue();
	// double confidence = 0.0;
	// double sum = 0.0;
	// double count = 0.0;
	// for (WeightedScore ws : simList) {
	// confidence += ws.getWeight();
	// count++;
	// sum += ws.getScore();
	// }
	// // if (confidence > 0.0) {
	// if (count > 0.0) {
	// map.put(feature.getKey().getName(), sum / count);
	// if (withConfidence.contains(feature.getKey())) {
	// map.put(feature.getKey().getName() + "::CONFIDENCE", confidence);
	// }
	// } else {
	// map.put(feature.getKey().getName(), 0.0);
	// if (withConfidence.contains(feature.getKey())) {
	// map.put(feature.getKey().getName() + "::CONFIDENCE", 0.0);
	// }
	// }
	// }
	// return new RecordFeature(map, datasetData);
	// }

}
