package expansion.set.evaluation;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.RandomStringUtils;
import org.apache.commons.lang.StringUtils;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.model.ClusterSimilaritiesImplNormalized;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.creation.Entity;
import expansion.set.evaluation.matching.ScoredEntity;
import expansion.set.strategies.rowclusterers.GreedyClustering;

public class CaseCreator {

	private final ValueEvaluator valueEvaluator;
	private final KnowledgeBase knowledgeBase;
	private final TimestampQualifiersAdapter adapter;
	private final AnnotatedTablesRepository tableRepository;
	
	
	 
	@Inject
	public CaseCreator(ValueEvaluator valueEvaluator, KnowledgeBase knowledgeBase, TimestampQualifiersAdapter adapter,
			AnnotatedTablesRepository tableRepository) {
		super();
		this.valueEvaluator = valueEvaluator;
		this.knowledgeBase = knowledgeBase;
		this.adapter = adapter;
		this.tableRepository = tableRepository;
	}

	public void createForNewEntitiesByRowsSize(String caseFolder, String fileName, Collection<ScoredEntity> entities, int topK) {
 
		entities = entities.stream()
				.sorted((e1, e2) -> Integer.compare(e2.getEntity().getRowCluster().getRows().size(), e1.getEntity().getRowCluster().getRows().size())).limit(20)
				.collect(Collectors.toList());
		
		create(caseFolder, fileName, entities);
	}
	
	public void createForNewEntitiesByValuesSizeTopK(String caseFolder, String fileName, Collection<ScoredEntity> entities, int topK) {
		List<ScoredEntity> list = new ArrayList<>(entities);
		list.sort((e1, e2) -> Integer.compare(e2.getEntity().getValues().size(), e1.getEntity().getValues().size()));
		list = list.subList(0, Math.min(topK, list.size()));
 		
		create(caseFolder, fileName,  list);
	}
	
	public void createForNewEntitiesByValuesSizeKperSize(String caseFolder, String fileName, Collection<ScoredEntity> entities, int maxPerSize) {
		Map<Integer, List<ScoredEntity>> map = entities.stream().filter((e) -> {
			return e.getEntity().getLabels().size() > 0 && e.getEntity().getValues().size() > 0;
		}).collect(
					Collectors.groupingBy(
						(e) -> (Integer) e.getEntity().getValues().size(), () -> new HashMap<>(), Collectors.toList())
					);
		List<Integer> sortedKeys = map.keySet().stream().sorted((e1, e2) -> Integer.compare(e2, e1) ).collect(Collectors.toList());
 		
		double largestSize = map.entrySet().stream().filter((e) -> e.getKey() != 0).mapToInt((e) -> e.getValue().size()).max().getAsInt();

		Map<Integer, Double> extractionSize = map.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			return new Double(Math.min(e.getValue().size() / largestSize, 1.0));
		}));

		// if (extractionSize.get(0) > 0.5) {
		// extractionSize.put(0, 0.5);
		// }

		System.out.println(extractionSize.toString());

		List<ScoredEntity> list = new ArrayList<>();
		for(Integer key : sortedKeys) {
			List<ScoredEntity> innerList = new ArrayList<>(map.get(key));
			Collections.shuffle(innerList);
			list.addAll(innerList.subList(0, Math.min(Math.max(1, Math.round((float) (maxPerSize * extractionSize.get(key)))), innerList.size())));
		}
			
			
		
 
		create(caseFolder, fileName,  list);
	}

	public void createWithErrors(String caseFolder, String fileName, List<Entity> entities, Map<InstanceId, Integer> mapping,
			Map<InstanceId, Set<TableRowId>> fullGs, Function<TableRowIdPair, Map<String, Double>> scoreGetter) {
		List<Case> cases = new ArrayList<>();

		Set<RowClusterUUID> added = new HashSet<>();
		for (Entry<InstanceId, Set<TableRowId>> entry : fullGs.entrySet()) {
			// int size = entity.getValues().size();

			String mainLabel = entry.getKey().getReadableString();

			// Optional<Entry<String, Integer>> mainLabelOptional =
			// entity.getLabels().entrySet().stream().sorted((e1, e2) -> {
			// return Integer.compare(e2.getValue(),
			// e1.getValue());}).findFirst();
			//
			// String mainLabel = "";
			// if(mainLabelOptional.isPresent())
			// mainLabelOptional.get().getKey();

			InstanceId instanceId = entry.getKey();
			String caseFileName = mainLabel.replaceAll("[^A-Za-z0-9 ]", "") + "-" + RandomStringUtils.randomAlphanumeric(4);
			;

			Case caseObject;
			if (mapping.containsKey(entry.getKey())) {
				Entity entity = entities.get(mapping.get(entry.getKey()));
				added.add(entity.getRowClusterUUID());
				caseObject = new Case(instanceId, entity.getLabels(), caseFileName);

			} else {
				Map<String, Integer> labels = new HashMap<>();
				caseObject = new Case(instanceId, labels, caseFileName);

			}

			if (mapping.containsKey(entry.getKey())) {
				Entity entity = entities.get(mapping.get(entry.getKey()));

				for (TableRowId rowId : entity.getRowCluster().getRows()) {
					if (entry.getValue().contains(rowId)) {
						caseObject.addCorrectRow(rowId);

					} else {
						caseObject.addBadRow(rowId);

					}

				}

				for (TableRowId rowId : entry.getValue()) {
					if (!entity.getRowCluster().getRows().contains(rowId)) {
						caseObject.addMissingRow(rowId);
					}
				}
			} else {
				for (TableRowId rowId : entry.getValue()) {

					caseObject.addMissingRow(rowId);

				}
			}
			

			caseObject.addScoreGetter(scoreGetter);

//			int i = 0;
//			for(TableRowId rowId : entity.getRowCluster().getRows()) {
//				
//				caseObject.addTables(rowId.getTableId(), tableRepository.getTable(rowId.getTableId()), rowId.getRowIdx());
//				i++;
//				if(i == 20) {
//					break;
//				}
//			}
			
			// if(false) {
			//
			// for(Entry<QualifiedPropertyId, FusedValue> values :
			// entity.getValues().entrySet()) {
			// QualifiedTripleId qtid = new QualifiedTripleId(new
			// TripleId(entity.getClassId(), values.getKey().getPropertyId(),
			// instanceId),
			// values.getKey().getQualifiers());
			//
			//
			// Integer contributors = values.getValue().getSingleContributors();
			// if(contributors == null)
			// contributors = 0;
			// caseObject.addValue(Optional.empty(), qtid, values.getValue(),
			// new HashSet<>(), contributors);
			//
			//
			//
			// }
			// }
			
			cases.add(caseObject);
 
			

		}
		
		
		for( Entity entity : entities) {
			if(!added.contains(entity.getRowClusterUUID())) {
				String mainLabel = entity.getMainLabel();
				if (mainLabel == null)
					mainLabel = "";
				String caseFileName = mainLabel.replaceAll("[^A-Za-z0-9 ]", "") + "-" + RandomStringUtils.randomAlphanumeric(4);

				Case caseObject;
				
				caseObject = new Case(entity, caseFileName);

				for (TableRowId rowId : entity.getRowCluster().getRows()) {

					caseObject.addBadRow(rowId);

				}
				caseObject.addScoreGetter(scoreGetter);
				cases.add(caseObject);

				
			}
		}
		
		
		

		try {
			Case.writeToFolder(fileName, caseFolder, cases);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}

	public void create(String caseFolder, String fileName, Collection<ScoredEntity> entities) {
		// List<Entity> shuffled = new ArrayList<>(entities);
		// Collections.shuffle(shuffled);
		
		List<Case> cases = new ArrayList<>();
		for (ScoredEntity scoredEntity : entities) {
			Entity entity = scoredEntity.getEntity();
			int size = entity.getValues().size();
			Optional<Entry<String, Integer>> mainLabelOptional = entity.getLabels().entrySet().stream().sorted((e1, e2) -> { return Integer.compare(e2.getValue(), e1.getValue());}).findFirst();
			String mainLabel = "";
			if(mainLabelOptional.isPresent())
				mainLabelOptional.get().getKey();

			InstanceId instanceId =  new InstanceId(mainLabel, mainLabel);
			String caseFileName = StringUtils.leftPad(String.valueOf(size), 4, "0") +  "-" + mainLabel.replaceAll("[^A-Za-z0-9 ]", "") + "-" + RandomStringUtils.randomAlphanumeric(4); ;

			Case caseObject = new Case(instanceId, entity.getLabels(),  caseFileName);
			for (TableRowId rowId : entity.getRowCluster().getRows()) {

				caseObject.addTableRow(rowId);

			}

//			int i = 0;
//			for(TableRowId rowId : entity.getRowCluster().getRows()) {
//				
//				caseObject.addTables(rowId.getTableId(), tableRepository.getTable(rowId.getTableId()), rowId.getRowIdx());
//				i++;
//				if(i == 20) {
//					break;
//				}
//			}
			
			
			for(Entry<QualifiedPropertyId, FusedValue> values : entity.getValues().entrySet()) {
				QualifiedTripleId qtid = new QualifiedTripleId(new TripleId(entity.getClassId(), values.getKey().getPropertyId(), instanceId),
						values.getKey().getQualifiers());
	 
			 
					 Integer contributors = values.getValue().getSingleContributors();
					 if(contributors == null)
						 contributors = 0;
						caseObject.addValue(Optional.empty(), qtid, values.getValue(), new HashSet<>(), contributors);
						
	 
				 
			}
			
			cases.add(caseObject);
 
			
			
		}
		
		try {
			Case.writeToFolder(fileName, caseFolder, cases);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public void create(String caseFolder, Map<InstanceId, List<ScoredEntity>> mappedEntities, ClusterList clusters, FusionPerformance<Integer> maxCorrect, SplitGroupedValues originalTestingSet) {
//		ConcurrentHashMap<QualifiedTripleId, FusedValue> fusedValues = new ConcurrentHashMap<>();
		
		ConcurrentHashMap<RowClusterUUID, RowCluster> clusterMap = new ConcurrentHashMap<>();
		
		PersistentParallel6.runAndFinish(clusters.getClusterList(), (cluster) -> {
			clusterMap.put(cluster.getUuid(), cluster);
		}, this.getClass());

//		FusionPerformanceCreator<AccumInteger> accumulator = FusionPerformance.accumulator(new AccumInteger(0));

		PersistentParallel6.runAndFinish(mappedEntities.entrySet(), (mappings) -> {

			Integer maxIndex = null;
			double maxScore = 0.0;

			for (ListEntry<ScoredEntity> entry : ListIterable.iterable(mappings.getValue())) {
				if (maxIndex == null || entry.getValue().getScore() > maxScore) {
					maxIndex = entry.getIndex();
					maxScore = entry.getValue().getScore();
				}
			}

 
			ScoredEntity max = mappings.getValue().get(maxIndex);
			
			writeCase(caseFolder, mappings.getKey(), max,originalTestingSet, clusterMap.get(max.getEntity().getRowClusterUUID()));

//			for (ListEntry<ScoredEntity> entry : ListIterable.iterable(mappings.getValue())) {
//				if (entry.getIndex() != maxIndex) {
//					for (Entry<QualifiedPropertyId, FusedValue> values : entry.getValue().getEntity().getValues().entrySet()) {
//						QualifiedTripleId qtid = new QualifiedTripleId(
//								new TripleId(max.getEntity().getClassId(), values.getKey().getPropertyId(), mappings.getKey()),
//								values.getKey().getQualifiers());
//						if (originalTestingSet.containsKey(qtid)) {
//							accumulator.addWithYear(entry.getValue().getEntity().getClassId(), values.getKey(), ontologyRepository, new AccumInteger(1));
//						}
//					}
//				}
//			}

		} , this.getClass());
		
	}
	
	
	private void writeCase(String baseFolder, InstanceId instanceId, ScoredEntity max, SplitGroupedValues originalTestingSet, RowCluster rowCluster ) {
		boolean include = false;
		
//		if(false || true)
//			return;
		
		Set<String> labels = knowledgeBase.getLabels(instanceId);
		
		Case caseObject = new Case( instanceId,  max.getEntity().getLabels(), max.getEntity().getRowClusterUUID().toString());
		
//		List<Pair<QualifiedTripleId, FusedValue>> correctValues = new ArrayList<>();
//		List<Pair<QualifiedTripleId, FusedValue>> incorrectValues = new ArrayList<>();
		
		for(Entry<QualifiedPropertyId, FusedValue> values : max.getEntity().getValues().entrySet()) {
			QualifiedTripleId qtid = new QualifiedTripleId(new TripleId(max.getEntity().getClassId(), values.getKey().getPropertyId(), instanceId),
					values.getKey().getQualifiers());
			if (originalTestingSet.containsKey(qtid)) {
				
				
				
				Optional<Boolean> correct = valueEvaluator.correct(qtid, values.getValue());
				
				if(correct.isPresent()) {
					
					
					if(!correct.get()) {
						include = true;
						
					}
					
					
					QualifiedValueList qvl = knowledgeBase.getQualifiedValueList(qtid.getTripleId().getValueId());
					Set<AbstractObject> correctValues = adapter.getForQualifiers(qvl, qtid.getQualifiers());
				 
					 Integer contributors = values.getValue().getSingleContributors();
					 if(contributors == null)
						 contributors = 0;
					 
					caseObject.addValue(Optional.of(correct.get()), qtid, values.getValue(), correctValues, contributors);
					
				}
			}
			else {
				int adsd = 122;
			}
		}
		
		for(TableRowId rowId : rowCluster.getRows()) {
			
			caseObject.addTables(rowId.getTableId(), tableRepository.getTable(rowId.getTableId()), rowId.getRowIdx());
			
		}
		
		
		try {
			if(include) {
			caseObject.writeToFolder(baseFolder);
			
			int df = 123;
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		 
		int sdf = 13;
		
	}
	
	
	public static class Case {
		
//		private final Set<String> labels;
//		private final InstanceId instanceId;
		
		private String header = "";
		private String values = "";
		private String tables = "";
		private final String uuid; 
		private Set<TableRowId> rowIds = new HashSet<>();
		private Set<TableRowId> correctRowIds = new HashSet<>();
		private Set<TableRowId> missingRowIds = new HashSet<>();
		private Set<TableRowId> badRowIds = new HashSet<>();

		private Map<AnnotatedTableId, AnnotatedTable> tablesMap = new HashMap<>();
		private Map<AnnotatedTableId, Set<RowIdx>> rowIdx = new HashMap<>();
		private Function<TableRowIdPair, Map<String, Double>> scoreGetter;
		private Function<TableRowId, Set<TableRowId>> blockingCandidatesGetter;
		
		public Case(Entity entity, String uuid) {

			List<Pair<String, Integer>> list = entity.getLabels().entrySet().stream().sorted((e1, e2) -> Integer.compare(e2.getValue(), e1.getValue()))
					.map((e) -> new Pair<>(e.getKey(), e.getValue())).collect(Collectors.toList());

			Collections.sort(list, (e1, e2) -> Integer.compare(e2.getSecond(), e1.getSecond()));
			header += "<hr />";
			header += "<h1>" + entity.getMainLabel() + "</h1>";

			header += "<h2>Labels</h2><ul>";

			for (Pair<String, Integer> label : list) {
				header += "<li>" + label.getFirst() + " - " + label.getSecond() + "</li>";
			}
			header += "</ul>";

			// values += "<h2>Values</h2>";

			this.uuid = uuid;

			// this.instanceId = instanceId;
			// this.labels = labels;
		}

		public Case(InstanceId instanceId, Map<String, Integer> map, String uuid) {
			
			List<Pair<String, Integer>> list = map.entrySet().stream().sorted(  (e1, e2) -> Integer.compare(e2.getValue(), e1.getValue())).map((e) -> new Pair<>(e.getKey(), e.getValue())).collect(Collectors.toList());
 

			Collections.sort(list, (e1, e2) -> Integer.compare(e2.getSecond(), e1.getSecond()));
			header += "<hr />";
			header += "<h1>" + instanceId.getReadableString() + " - "+ instanceId.getInstanceId() + "</h1>";
			
			header += "<h2>Labels</h2><ul>";
			
			for(Pair<String, Integer> label : list) {
				header += "<li>" + label.getFirst() + " - " + label.getSecond() + "</li>";
			}
			header += "</ul>";
			
//			values += "<h2>Values</h2>";
			
			this.uuid = uuid;
			 
//			this.instanceId = instanceId;
//			this.labels = labels;
		}
		

		public void addScoreGetter(Function<TableRowIdPair, Map<String, Double>> scoreGetter) {
			this.scoreGetter = scoreGetter;

		}

		public void addTables(AnnotatedTableId tableId, AnnotatedTable annotatedTable, RowIdx rowIdx) {
			
			if(!this.tablesMap.containsKey(tableId)) {
				this.tablesMap.put(tableId, annotatedTable);
			}
			
			CHMHelper.getFromMap(this.rowIdx, tableId, () -> new HashSet<>()).add(rowIdx);
		}
		
		public void addCorrectRow(TableRowId rowId) {

			this.correctRowIds.add(rowId);

		}

		public void addMissingRow(TableRowId rowId) {
			this.missingRowIds.add(rowId);

		}

		public void addBadRow(TableRowId rowId) {

			this.badRowIds.add(rowId);

		}

		public void addTableRow(TableRowId rowId) {

			this.rowIds.add(rowId);

		}

		public void addValue(Optional<Boolean> correct, QualifiedTripleId qtid, FusedValue fusedValue, Set<AbstractObject> correctValues, Integer contributors) {
			values += "<tr>";
			
			if(correct.isPresent())
				values += "<td style=\"font-weight: bold;\">" + (correct.get() ? "<span style=\"color: red;\">Yes</span>" : "<span style=\"color: green;\">No</span>" ) + "</td>";
			else
				values += "<td style=\"font-weight: bold;\"></td>";
			values += "<td>" + qtid.getQualifiedPropertyId().getPropertyId().getReadableString() + "</td>";
			values += "<td>" + qtid.getQualifiedPropertyId().getQualifiers().toString() + "</td>";
			values += "<td>" + fusedValue.getSingle().toString() + "</td>";
			values += "<td>" + correctValues.toString() + "</td>";
			values += "<td>" + contributors + "</td>";
			
			values += "</tr>";
		}
		
		private String createValues() {
			String ret = "<h2>Values</h2>";
			
			ret += "<table><tr>";
			ret += "<th>Correct</th>";
			ret += "<th>Property</th>";
			ret += "<th>Timestamp</th>";
			ret += "<th>FusedValue</th>";
			ret += "<th>CorrectValues</th>";
			ret += "<th>Contributors</th>";
			ret += "</tr>";
			
			ret += values;
			
			
			ret += "</table>";
			
			return ret;
		}
		
		private String createRows() {
			String ret = "";
			if (!rowIds.isEmpty()) {
				ret += "<h2>Rows</h2>";

				ret += "<ul>";

				for (TableRowId rowId : rowIds) {
					ret += "<li>" + rowId.getTableId().getTableName() + ", " + rowId.getRowIdx().getRowIdx() + "</li>";
				}

				ret += "</ul>";
			}

			if (!correctRowIds.isEmpty()) {
				ret += "<h2>Correct Rows</h2>";

				ret += "<ul>";

				for (TableRowId rowId : correctRowIds) {
					ret += "<li style=\"color: green;\">" + rowId.getTableId().getTableName() + ", " + rowId.getRowIdx().getRowIdx() + "<br />";
					if (scoreGetter != null && (correctRowIds.size() > 1 || !badRowIds.isEmpty() || !missingRowIds.isEmpty())) {
						ret += comparisonTable(rowId, correctRowIds, badRowIds, missingRowIds);
					}

					if (blockingCandidatesGetter != null) {
						ret += "<ul>";

						Set<TableRowId> blockingIds = blockingCandidatesGetter.apply(rowId);
						for (TableRowId correctId : correctRowIds) {
							if (!correctId.equals(rowId)) {
								if (!blockingIds.contains(correctId)) {
									ret += "<li style=\"color: orange;\">" + correctId.getTableId().getTableName() + ", " + correctId.getRowIdx().getRowIdx()
											+ "</li>";
								}
							}

						}

						ret += "</ul>";


					}
					ret += "</li>";

				}

				ret += "</ul>";
			}

			if (!badRowIds.isEmpty()) {
				ret += "<h2>Bad Rows</h2>";

				ret += "<ul>";

				for (TableRowId rowId : badRowIds) {
					ret += "<li style=\"color: red;\">" + rowId.getTableId().getTableName() + ", " + rowId.getRowIdx().getRowIdx() + "<br />";

					if (badRowIds.size() > 1 && scoreGetter != null && correctRowIds.size() == 0 && missingRowIds.size() == 0) {
						ret += comparisonTable(rowId, new HashSet<>(), badRowIds, new HashSet<>());
					}

					ret += "</li>";
				}

				ret += "</ul>";
			}

			if (!missingRowIds.isEmpty()) {
				ret += "<h2>Missing Rows</h2>";

				ret += "<ul>";

				for (TableRowId rowId : missingRowIds) {
					ret += "<li style=\"color: orange;\">" + rowId.getTableId().getTableName() + ", " + rowId.getRowIdx().getRowIdx() + "<br />";

					if (scoreGetter != null && correctRowIds.size() == 0 && missingRowIds.size() > 1) {
						ret += comparisonTable(rowId, new HashSet<>(), new HashSet<>(), missingRowIds);
					}

					ret += "</li>";
				}

				ret += "</ul>";
			}

			return ret;

		}

		private String comparisonTable(TableRowId rowId, Set<TableRowId> correctRowIds, Set<TableRowId> badRowIds, Set<TableRowId> missingRowIds) {

			String ret = "";

			ret += "<table cellspacing=\"0\" border=\"2\" cellpadding=\"5\">";
			ret += "<tr>";
			Set<String> headRowMap = null;
			if (!badRowIds.isEmpty()) {
				// headRowMap = scoreGetter.apply(new TableRowIdPair(rowId,
				// badRowIds.iterator().next())).keySet();
				Iterator<TableRowId> it = badRowIds.iterator();
				TableRowId badRowId = it.next();
				if (badRowId.equals(rowId)) {
					badRowId = it.next();
				}
				headRowMap = scoreGetter.apply(new TableRowIdPair(rowId, badRowId)).keySet();
			} else if (!missingRowIds.isEmpty()) {
				Iterator<TableRowId> it = missingRowIds.iterator();
				TableRowId missingId = it.next();
				if (missingId.equals(rowId)) {
					missingId = it.next();
				}
				headRowMap = scoreGetter.apply(new TableRowIdPair(rowId, missingId)).keySet();

			} else if (!correctRowIds.isEmpty()) {
				Iterator<TableRowId> it = correctRowIds.iterator();
				TableRowId correctId = it.next();
				if (correctId.equals(rowId)) {
					correctId = it.next();
				}
				headRowMap = scoreGetter.apply(new TableRowIdPair(rowId, correctId)).keySet();

			}

			ret += "<th>Row</th>";
			for (String string : headRowMap) {
				if (string.equals("AA_LABEL_PHI"))
					string = "PHI";
				ret += "<th>" + string + "</th>";

			}

			ret += "</tr>";
			for (TableRowId correctId : correctRowIds) {
				if (!correctId.equals(rowId)) {
					Map<String, Double> score = scoreGetter.apply(new TableRowIdPair(rowId, correctId));
					ret += "<tr style=\"color: green;\">";
					ret += "<td>" + correctId.getTableId().getTableName() + ", " + correctId.getRowIdx().getRowIdx() + "</td>";
					for (String string : headRowMap) {
						Double d = score.get(string);
						ret += "<td>" + DecimalFormatter.format(d) + "</td>";

					}
					ret += "</tr>";
				}
			}

			for (TableRowId badRowId : badRowIds) {
				if (!badRowId.equals(rowId)) {
					Map<String, Double> score = scoreGetter.apply(new TableRowIdPair(rowId, badRowId));
					ret += "<tr style=\"color: red;\">";
					ret += "<td>" + badRowId.getTableId().getTableName() + ", " + badRowId.getRowIdx().getRowIdx() + "</td>";
					for (String string : headRowMap) {
						Double d = score.get(string);
						ret += "<td>" + DecimalFormatter.format(d) + "</td>";

					}
					ret += "</tr>";
				}
			}
			for (TableRowId missingBadRowId : missingRowIds) {
				if (!missingBadRowId.equals(rowId)) {
					Map<String, Double> score = scoreGetter.apply(new TableRowIdPair(rowId, missingBadRowId));
					ret += "<tr style=\"color: orange;\">";
					ret += "<td>" + missingBadRowId.getTableId().getTableName() + ", " + missingBadRowId.getRowIdx().getRowIdx() + "</td>";
					for (String string : headRowMap) {
						Double d = score.get(string);
						ret += "<td>" + DecimalFormatter.format(d) + "</td>";

					}
					ret += "</tr>";
				}

			}
			ret += "</table>";

			return ret;
		}

		private String createTables() {
			String ret = "<h2>Tables</h2>";
			
			for(  Entry<AnnotatedTableId, AnnotatedTable> tableEntry : this.tablesMap.entrySet()) {
				
				Set<RowIdx> rowIds = this.rowIdx.get(tableEntry.getKey());
				
				int columnSize = tableEntry.getValue().getColumns().size();
				
				int rowSize = tableEntry.getValue().getValues().size() / columnSize;
				
				
				ret += "<h3>" + tableEntry.getKey().getTableName() + "</h3>";
				ret += "<table>";
				
				for(int i = 0; i < rowSize; i++) {
					

					RowIdx rowIdx = new RowIdx(i);
					
					if(rowIds.contains(rowIdx)) {
						
					
						ret += "<tr style=\"background-color: #eeeeee;\">";
					}
					else {
						ret += "<tr>";
					}
					
					
					
					for(int c = 0; c < columnSize; c++) {
						AbstractObject v = tableEntry.getValue().getValue(rowIdx, new ColumnIdx(c));
						String str= "";
						if(v != null)
							str = v.toString();
						if(i == 0) {
							ret += "<th>" + str  + "</th>";
						}
						else {
							ret += "<td>" + str  + "</td>";
						}
						
					}
					
					ret += "</tr>";
				}
				
				ret += "</table>";
			}
			
			return ret;
		}
		
		private String getHtml() {
			StringBuilder sb = new StringBuilder( "<html><head><title>" + this.uuid + "</title></head><body>");
			
			sb.append(header);
			
			sb.append(createValues());
			sb.append(createTables());
			sb.append(createRows());
			sb.append("</body></html>");
			
			return sb.toString();
		}
		
		private String getInnerHtml() {
			StringBuilder sb = new StringBuilder(  );
			
			sb.append(header);
			
			sb.append(createValues());
			sb.append(createTables()); 
			sb.append(createRows());
			
			return sb.toString();
		}
		
		public void writeToFolder(String folder) throws IOException {
			String html =  getHtml();
			
			FileUtils.writeStringToFile(new File(folder + "/" + uuid.toString() + ".html"), html);
			 
		}
		
		public static void writeToFolder(String uuid, String folder, Collection<Case> cases) throws IOException {
			StringBuilder sb = new StringBuilder( "<html><head><title>" + uuid + "</title></head><body>");
			
			for(Case caze : cases) {
				sb.append(caze.getInnerHtml());
			}
			
			
			sb.append("</body></html>");
			
		 
			
			FileUtils.writeStringToFile(new File(folder + "/" + uuid + ".html"), sb.toString());
		}

		public void setBlockingCandidatesGetter(Function<TableRowId, Set<TableRowId>> blockingCandidatesGetter) {
			this.blockingCandidatesGetter = blockingCandidatesGetter;

		}
	}


	public void createForBlockers(String caseFolder, String fileName, ClassId classId, Map<InstanceId, Set<TableRowId>> fullGs,
			RowClusterSimilarities sim) {
		Map<Integer, TableRowId> index = new LinkedHashMap<>();
		Map<TableRowId, Integer> reverseIndex = new LinkedHashMap<>();
		for (Entry<InstanceId, Set<TableRowId>> entry : fullGs.entrySet()) {
			for(TableRowId rowId : entry.getValue()) {
				int i = index.size();
				index.put(i, rowId);
				reverseIndex.put(rowId, i);
			}
		}
		RowCluster rc = new RowCluster(classId, new ArrayList<>( reverseIndex.keySet()), sim);
		ClusterSimilaritiesImplNormalized clusterSim = new ClusterSimilaritiesImplNormalized(rc, sim, false);
		
		Map<UUID, TableRowId> uuidMapping = new HashMap<>();
		Map<Integer, Set<UUID>> blocks = new HashMap<>();
		for (Entry<Integer, TableRowId> entry : index.entrySet()) {
			UUID uuid = UUID.randomUUID();
			uuidMapping.put(uuid, entry.getValue());
			for (Tri<Set<Integer>, Integer, Double> rowBlocks : clusterSim.getSplitBlocksWithScores(entry.getKey())) {
				for (Integer block : rowBlocks.getA()) {
					CHMHelper.getFromMap(blocks, block, () -> new HashSet<>()).add(uuid);
				}
			}

		}

		Set<TableRowId> rowIdSet = new HashSet<>();
		rowIdSet.add(new TableRowId("1346876860840_1346953393351_1858.arc4944083447630949323#2865484_1_8078897309408309233", 12));
		rowIdSet.add(new TableRowId("1346823845675_1346865484759_551.arc5679908987767700891#21791343_2_1487834260780300373", 5));
		rowIdSet.add(new TableRowId("1346876860445_1346962229057_555.arc690512194785518291#28168330_0_3483009587671943125", 3));
		rowIdSet.add(new TableRowId("1346981172239_1346996411405_686.arc5220620847597988200#80178618_0_9094416607926131718", 11));
		rowIdSet.add(new TableRowId("1346876860454_1346971822823_987.arc2109482148997547632#85409501_0_5286798261832614426", 3));
		rowIdSet.add(new TableRowId("1346823846150_1346831690408_45.arc5994122906536653366#78012523_2_6448012290109416123", 5));
		rowIdSet.add(new TableRowId("1346876860779_1346958288804_595.arc7823972158741128416#56229781_0_2382996291996445910", 4));
		rowIdSet.add(new TableRowId("1346876860828_1346952706663_3429.arc5921241954962074354#20517249_1_9167235913169124029", 12));
		rowIdSet.add(new TableRowId("1346981172155_1347001440797_1392.arc9090411764390287749#66457071_11_4481519565885296706", 3));
		rowIdSet.add(new TableRowId("1346981172264_1347018377126_3653.arc5782822234765809404#30042358_0_8283684810574430634", 10));
		rowIdSet.add(new TableRowId("1346981172234_1347006187111_1817.arc3336533571957179335#77829756_0_5604638202298041456", 142));
		rowIdSet.add(new TableRowId("1346981172264_1347018690685_2692.arc5694695746453878202#5401079_0_644130583662511127", 5));
		rowIdSet.add(new TableRowId("1346981172142_1347012114824_828.arc6735235393062592958#85134424_0_1746425791170541054", 1));
		rowIdSet.add(new TableRowId("1346981172250_1346985391828_2366.arc2516789212534356496#82069893_0_103809824742711773", 6));


		Function<TableRowId, Set<TableRowId>> getBlockingCandidates = (rowId) -> {


			Collection<TableRowId> ret = GreedyClustering.getClustersFromBlockMap(reverseIndex.get(rowId), clusterSim, blocks, uuidMapping);
			if (rowIdSet.contains(rowId)) {
				int sdf = 123;
			}
			return new HashSet<>(ret);
		};

		List<Case> cases = new ArrayList<>();

		for (Entry<InstanceId, Set<TableRowId>> entry : fullGs.entrySet()) {
			// int size = entity.getValues().size();

			String mainLabel = entry.getKey().getReadableString();

			// Optional<Entry<String, Integer>> mainLabelOptional =
			// entity.getLabels().entrySet().stream().sorted((e1, e2) -> {
			// return Integer.compare(e2.getValue(),
			// e1.getValue());}).findFirst();
			//
			// String mainLabel = "";
			// if(mainLabelOptional.isPresent())
			// mainLabelOptional.get().getKey();

			InstanceId instanceId = entry.getKey();
			String caseFileName = mainLabel.replaceAll("[^A-Za-z0-9 ]", "") + "-" + RandomStringUtils.randomAlphanumeric(4);
			;

			Map<String, Integer> labels = new HashMap<>();
			Case caseObject = new Case(instanceId, labels, caseFileName);

			for (TableRowId rowId : entry.getValue()) {

				caseObject.addCorrectRow(rowId);


			}
			caseObject.setBlockingCandidatesGetter(getBlockingCandidates);

			 
 
			
			cases.add(caseObject);
 
			

		}
		

		try {
			Case.writeToFolder(fileName, caseFolder, cases);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}

}
