package expansion.set.evaluation.crossvalidation;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.profiling.GenericLabelRowIndex;
import expansion.set.profiling.GenericLabelRowIndex.IndexId;
import expansion.set.strategies.rowclusterers.unary.ClusterCombiner;

public class EntitySetsCreatorHard {
	private final KnowledgeBase knowledgeBase;
	private final SimilarityCalculator similarityCalculator;
	private final AnnotatedTablesRepository tableRepository;
	@Inject
	public EntitySetsCreatorHard(KnowledgeBase knowledgeBase, SimilarityCalculator similarityCalculator, AnnotatedTablesRepository tableRepository) {
		this.knowledgeBase = knowledgeBase;
		this.similarityCalculator = similarityCalculator;
		this.tableRepository = tableRepository;
	}

//	private final Comparator<InstanceId> byTimestamp = Comparator.comparing(InstanceId::getInstanceId);

	
	private Collection<Set<InstanceId>> getHardInstances(Set<InstanceId> instances, boolean removeAfterCommaAndRemoveBrackets) throws IOException {
		
		ConcurrentHashMap<String, Set<IndexId>> preIndex = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(instances, (instance) -> {
				
			
				Set<String> labels = knowledgeBase.getLabels(instance);
				if(removeAfterCommaAndRemoveBrackets) {
					labels = StringNormalizer.removeAfterCommaAndRemoveBrackets(labels);
				}
				
				for(String string : labels) {
					CHMHelper.getFromCHM(preIndex, string, () -> new HashSet<>()).add(new IndexId(instance.toCombinedString()));
				}
			
			}, this.getClass());
			
		
		GenericLabelRowIndex index = new GenericLabelRowIndex(preIndex);
		
		ConcurrentLinkedQueue<Set<InstanceId>> individualClusters = new ConcurrentLinkedQueue<>();
		
		Set<InstanceId> testSet = Collections.newSetFromMap(new ConcurrentHashMap< >());
		SimilarityMeasure sm = similarityCalculator.getMeasure(ValueType.STRING);
		PersistentParallel6.runAndFinish(instances, (instance) -> {
			try {
				Set<InstanceId> cluster = new HashSet<>();
				cluster.add(instance);
				
				boolean conflict = false;
				Set<String> labels = knowledgeBase.getLabels(instance);
				if(removeAfterCommaAndRemoveBrackets) {
					labels = StringNormalizer.removeAfterCommaAndRemoveBrackets(labels);
				}
				for(String string : labels) {
					
					Map<String, Set<IndexId>> matches = index.getMatches(string);
				 
					StringObject target = new StringObject(string);
					for(Entry<String, Set<IndexId>> matchEntry : matches.entrySet()) {
						 if(matchEntry.getValue().size() > 1 || !instance.equals(InstanceId.fromCombinedString(matchEntry.getValue().iterator().next().getIndexId()))) { 
							if(sm.isEqual(target, new StringObject(matchEntry.getKey()))) {
								conflict = true;
								for(IndexId indexId : matchEntry.getValue()) {
									cluster.add(InstanceId.fromCombinedString(indexId.getIndexId()));
									testSet.add(instance);
									testSet.add(InstanceId.fromCombinedString(indexId.getIndexId()));
								}
							}
						 }
					}
					
				 
					
				}
				
				if(cluster.size() > 1) {
					individualClusters.add(cluster);
				}
				
				 
				
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		}, this.getClass());
		
		
		ClusterCombiner<InstanceId> cc = new ClusterCombiner<>();
		cc.add(individualClusters);
		
		int size = testSet.size();
		
		return cc.get();
	}
	

	private final Comparator<InstanceId> byInstanceId = Comparator.comparing(InstanceId::getInstanceId);
	
	private final Supplier<TreeSet<InstanceId>> supplier = () -> new TreeSet<InstanceId>(byInstanceId);

	public Map<Integer, EntitySets> splitHard(Map<ClassId, Set<InstanceId>> instanceIdSet, int setSize) throws IOException {
		return split(instanceIdSet, setSize, true);
	}
	
	public Map<Integer, EntitySets> splitEasy(Map<ClassId, Set<InstanceId>> instanceIdSet, int setSize) throws IOException {
		return split(instanceIdSet, setSize, false);
	}
	private Map<Integer, EntitySets> split(Map<ClassId, Set<InstanceId>> instanceIdSet, int setSize, boolean hard) throws IOException {


		

//		Map<ClassId, TreeSet<InstanceId>> classes = groupedValues.entrySet().stream().collect(Collectors.groupingBy((e) -> {
//			return e.getKey().getTripleId().getClassId();
//		} , Collectors.mapping((e) -> {
//			return e.getKey().getTripleId().getInstanceId();
//		} , Collectors.toCollection(supplier))));
		
		Map<ClassId, TreeSet<InstanceId>> classes = instanceIdSet.entrySet().stream().collect(Collectors.toMap((e) -> {
			return e.getKey();
		} , (e) -> {
			TreeSet<InstanceId> ts = new TreeSet<>(byInstanceId);
			ts.addAll(e.getValue());
			return  ts;
		}));

 
		Map<Integer, Set<InstanceId>> learning = new HashMap<>();
		Map<Integer, Set<InstanceId>> testing = new HashMap<>();

		for(Entry<ClassId, TreeSet<InstanceId>> entry : classes.entrySet()) {
			Map<Integer, EntitySets> splitted = splitInner(entry.getValue(), setSize, hard);
			for (Entry<Integer, EntitySets> splittedEntry : splitted.entrySet()) {
				CHMHelper.getFromMap(learning, splittedEntry.getKey(), () -> new HashSet<>()).addAll(splittedEntry.getValue().getLearningSet());
				CHMHelper.getFromMap(testing, splittedEntry.getKey(), () -> new HashSet<>()).addAll(splittedEntry.getValue().getTestingSet());
			}

		}

		Map<Integer, EntitySets> ret = new HashMap<>();
		for (Integer key : learning.keySet()) {
			ret.put(key, new EntitySets(learning.get(key), testing.get(key)));
		}
		
		return ret;

	}

//	public Map<Integer, EntitySets> split(Collection<InstanceId> entities, int setSize) {
//
//		TreeSet<InstanceId> entityIds = new TreeSet<>(byTimestamp);
//		entityIds.addAll(entities);
//
//		return splitInner(entityIds, setSize);
//	}

	private Map<Integer, EntitySets> splitInner(TreeSet<InstanceId> entityIds, int setSize, boolean hard) throws IOException {

		
		
		  Comparator<TreeSet<InstanceId>> byString = Comparator.comparing(Set::toString);
		
		  Supplier<TreeSet<TreeSet<InstanceId>>> byStringSupplier = () -> new TreeSet<TreeSet<InstanceId>>(byString);
		
		
		Collection<Set<InstanceId>> hardInstances = getHardInstances(entityIds, true);
		
		Set<InstanceId> hardInstancesFull = hardInstances.stream().flatMap((e) -> e.stream()).collect(Collectors.toCollection(supplier));
		
		Set<InstanceId> nonHardInstances = entityIds.stream().filter((e) -> !hardInstancesFull.contains(e)).collect(Collectors.toCollection(supplier));
		
		TreeSet<TreeSet<InstanceId>> instanceClusters = hardInstances.parallelStream().map((e) -> {
		TreeSet<InstanceId> newTreeSet = new TreeSet<>(byInstanceId);
		newTreeSet.addAll(e);
		return newTreeSet;}).collect(Collectors.toCollection(byStringSupplier));
		
//		int total = instanceClusters.stream().mapToInt((e) -> e.size()).sum();
		
		
		
		Map<Integer, List<InstanceId>> learningSet = new HashMap<>();
		Map<Integer, List<InstanceId>> testingSet = new HashMap<>();

		if(hard) {
			for (int i = 0; i < setSize; i++) {
				learningSet.put(i, new ArrayList<>(nonHardInstances));
				testingSet.put(i, new ArrayList<>());
			}
	
	
			int i = 0;
			for (TreeSet<InstanceId> instanceId : instanceClusters) {
				for (int set = 0; set < setSize; set++) {
					if (set == i) {
						testingSet.get(set).addAll(instanceId.stream().filter((e) -> {return this.tableRepository.hasInstance(e);}).collect(Collectors.toSet()));
					} else {
						learningSet.get(set).addAll(instanceId);
					}
				}
	
				i++;
				if (i == setSize) {
					i = 0;
				}
			}
		}
		else {
			
			for (int i = 0; i < setSize; i++) {
				learningSet.put(i, new ArrayList<>());
				testingSet.put(i, new ArrayList<>());
			}
	
			
			int i = 0;
			for (InstanceId instanceId : nonHardInstances) {
				for (int set = 0; set < setSize; set++) {
					if (set == i) {
						if(this.tableRepository.hasInstance(instanceId)) {
							testingSet.get(set).add(instanceId);
						}						
					} else {
						learningSet.get(set).add(instanceId);
					}
				}
	
				i++;
				if (i == setSize) {
					i = 0;
				}
			}
		}

		Map<Integer, EntitySets> ret = new HashMap<>();

		for (int s = 0; s < setSize; s++) {
			ret.put(s, new EntitySets(learningSet.get(s), testingSet.get(s)));
		}

		return ret;
	}

}
