package expansion.set.evaluation.goldstandard;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Tri;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.creation.Entity;

public class ClusterToGoldStandardMapper {

	private final AnnotatedTablesRepository tableRepository;
	
	
	@Inject
	public ClusterToGoldStandardMapper(AnnotatedTablesRepository tableRepository) {
		super();
		this.tableRepository = tableRepository;
	}



	public GoldStandardMapping map(GoldStandardClusters gs, List<Entity> entities) {
		
		Map<InstanceId, Set<TableRowId>> gsMap = gs.getTestingSetMappingClusters();
		
		Map<RowClusterUUID, Entity> entitiesMap = entities.parallelStream().collect(Collectors.toMap(Entity::getRowClusterUUID, Function.identity()));
		
//		ConcurrentHashMap<InstanceId, Map<RowClusterUUID, Integer>> countMap = new ConcurrentHashMap<>();
		
		ConcurrentLinkedQueue<Tri<Double, InstanceId, RowClusterUUID>> queue = new ConcurrentLinkedQueue<>();
		
		PersistentParallel6.runAndFinish(entitiesMap.entrySet(), (e) -> {
			Map<InstanceId, MutableInt> count = new HashMap<>();
			for(TableRowId rowId : e.getValue().getRowCluster().getRows()) {
				Optional<InstanceId> instanceId = tableRepository.getMatchedInstanceId(rowId);
				
				if(instanceId.isPresent()) {
					CHMHelper.getFromMap(count, instanceId.get(), () -> new MutableInt()).increment();
				}
				else {
					throw new IllegalStateException("I expect all candidate rows to be of instance in testing set");
				}
			}
			
			for(Entry<InstanceId, MutableInt> countEntry : count.entrySet()) {
//				CHMHelper.getFromCHM(countMap, countEntry.getKey(), () -> new ConcurrentHashMap<>()).put(e.getKey(), countEntry.getValue().intValue());
				double clusterSize = countEntry.getValue().doubleValue();
				double gsSize = gsMap.get(countEntry.getKey()).size();
				double score = 0.0;
				// JACCARD SIM
				if(clusterSize > gsSize) {
					score = gsSize / clusterSize;
				}
				else {
					score = clusterSize / gsSize;
				}
				queue.add(new Tri<>(score, countEntry.getKey(), e.getKey()));
			}
		}, this.getClass());
		

		
		List<Tri<Double, InstanceId, RowClusterUUID>> scores = new ArrayList<>(queue);
		
		Collections.sort(scores, (a,b) -> {
			int compare = Double.compare(a.getA(), b.getA());
			if(compare == 0) {
				compare = Integer.compare(gsMap.get(a.getB()).size(), gsMap.get(b.getB()).size());
			}
			return compare;
			
		});
		
		
		Set<RowClusterUUID> unmappedClusters = new HashSet<>();
		
		unmappedClusters.addAll(entitiesMap.keySet());
		
		
		Map<InstanceId, Entity> mapping = new HashMap<>();
		for(Tri<Double, InstanceId, RowClusterUUID> tri : scores) {
			if(!mapping.containsKey(tri.getB()) ) {
				mapping.put(tri.getB(), entitiesMap.get(tri.getC()));
				unmappedClusters.remove(tri.getC());
			}
			
			if(gsMap.size() == mapping.size()) {
				break;
			}
		}
		
		List<Entity> unmappedEntities = unmappedClusters.stream().map((e) -> {return entitiesMap.get(e);}).collect(Collectors.toList());
		
		return new GoldStandardMapping(mapping, unmappedEntities);
		
	}
	
}
