package expansion.set.evaluation.goldstandard;

import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public class GoldStandardClustersSet {

	private final GoldStandardData full;
	private final Map<Integer, GoldStandardData> sets;

	private final List<Set<InstanceId>> conflictingClusters;
	private final Map<InstanceId, Integer> conflictingClustersIndex;

	public GoldStandardClustersSet(GoldStandardData full, List<GoldStandardData> sets, List<Set<InstanceId>> conflictingClusters) {
		super();
		this.full = full;
		Map<Integer, GoldStandardData> setsMap = new TreeMap<>();

		for (ListEntry<GoldStandardData> data : ListIterable.iterable(sets)) {
			setsMap.put(data.getIndex(), data.getValue());
		}

		this.sets = Collections.unmodifiableMap(setsMap);

		this.conflictingClusters = Collections.unmodifiableList(new ArrayList<>(conflictingClusters));
		Map<InstanceId, Integer> conflictingSetsIndex = new HashMap<>();
		for (ListEntry<Set<InstanceId>> entry : ListIterable.iterable(conflictingClusters)) {
			for (InstanceId instanceId : entry.getValue()) {
				conflictingSetsIndex.put(instanceId, entry.getIndex());
			}
		}

		this.conflictingClustersIndex = Collections.unmodifiableMap(conflictingSetsIndex);
	}



	public GoldStandardData getFull() {
		return full;
	}

	public GoldStandardData getSet(int key) {
		return sets.get(key);
	}

	public static String statistics(GoldStandardClustersSetPerClass input) throws IOException {


		StringWriter sw = new StringWriter();

		CSVWriter csvWriter = new CSVWriter(sw);

		int columns = 1 + 6 * (1 + input.getSetCount());
		{
			String[] line = new String[columns];
			int i = 0;
			line[i++] = "ClassId";

			line[i++] = "";
			line[i++] = "TotalClusters";
			line[i++] = "TotalNew";
			line[i++] = "TotalCollisionGroups";
			line[i++] = "TotalCollisionClusters";
			line[i++] = "TotalRows";

			for (int j = 0; j < input.getSetCount(); j++) {
				line[i++] = "";
				line[i++] = "" + j + "_Clusters";
				line[i++] = "" + j + "_New";
				line[i++] = "" + j + "_CollisionGroups";
				line[i++] = "" + j + "_CollisionClusters";
				line[i++] = "" + j + "_Rows";
			}



			csvWriter.writeNext(line);
		}

		for (Entry<ClassId, GoldStandardClustersSet> entry : input.getPerClass().entrySet()) {

			GoldStandardClustersSet set = entry.getValue();
			ClassId classId = entry.getKey();
			String[] line = new String[columns];
			int i = 0;
			line[i++] = classId.getFinalPart();

			line[i++] = "";
			line[i++] = String.valueOf(set.getFull().getClusters().getTestingSetMappingClusters().size());
			line[i++] = String.valueOf(set.getFull().getNewInstances().size());
			line[i++] = String.valueOf(set.conflictingClusters.size());
			line[i++] = String.valueOf(set.conflictingClusters.stream().mapToInt((e) -> e.size()).sum());
			line[i++] = String.valueOf(set.getFull().getClusters().getIndex().size());

			for (int j = 0; j < input.getSetCount(); j++) {
				Set<InstanceId> instances = set.getSet(j).getClusters().getTestingSetMappingClusters().keySet();

				line[i++] = "";
				line[i++] = String.valueOf(set.getSet(j).getClusters().getTestingSetMappingClusters().size());
				line[i++] = String.valueOf(set.getSet(j).getNewInstances().size());
				line[i++] = String.valueOf(set.conflictingClusters.stream().filter((e) -> {
					for (InstanceId instanceId : e) {
						if (instances.contains(instanceId))
							return true;
					}
					return false;
				}).count());
				line[i++] = String.valueOf(set.conflictingClusters.stream().filter((e) -> {
					for (InstanceId instanceId : e) {
						if (instances.contains(instanceId))
							return true;
					}
					return false;
				}).mapToInt((e) -> e.size()).sum());
				line[i++] = String.valueOf(set.getSet(j).getClusters().getIndex().size());

			}


			csvWriter.writeNext(line);
		}

		csvWriter.close();
		String output = sw.toString();
		System.out.println(output);
		return output;
	}


	public Map<InstanceId, Integer> getConflictingClustersIndex() {
		return conflictingClustersIndex;
	}

	public static class GoldStandardData {

		private final GoldStandardClusters clusters;
		private final Set<InstanceId> newInstances;
		private final Map<QualifiedTripleId, QualifiedValueList> factsMap;

		public GoldStandardData(GoldStandardClusters clusters, Set<InstanceId> newInstances, Map<QualifiedTripleId, QualifiedValueList> factsMap) {
			super();
			this.clusters = clusters;
			this.newInstances = Collections.unmodifiableSet(new HashSet<>(newInstances));
			this.factsMap = Collections.unmodifiableMap(new HashMap<>(factsMap));
		}

		public GoldStandardClusters getClusters() {
			return clusters;
		}

		public Set<InstanceId> getNewInstances() {
			return newInstances;
		}

		public Map<QualifiedTripleId, QualifiedValueList> getFactsMap() {
			return factsMap;
		}

	}

	public int getSetCount() {
		return sets.size();
	}

	public Set<Integer> getSetsKeys() {
		return sets.keySet();
	}

}
