package expansion.set.exportimport;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.commons.compress.compressors.gzip.GzipCompressorOutputStream;
import org.apache.commons.lang.RandomStringUtils;

import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SparseArray;
import de.mannheim.uni.expansion.utils.containers.Tri;

public class OptionalClusterCache {

	private final Map<TableRowId, Integer> mapping;
	private final Map<Integer, Set<Integer>> pairs;

	private final OptionalClusterCacheMap[] cache;
	private final String cacheFolder;

	// private final ExecutorService executorService =
	// Executors.newCachedThreadPool();
	// private final Object[] locks;

	public OptionalClusterCache(Map<TableRowId, Integer> mapping, Map<Integer, Set<Integer>> pairs, OptionalClusterCacheMap[] cache, String cacheFolder) {
		super();
		this.mapping = mapping;
		this.pairs = pairs;
		this.cache = cache;
		this.cacheFolder = cacheFolder;
		// this.locks = new Object[cache.length];
		// for (int i = 0; i < cache.length; i++) {
		// locks[i] = new Object();
		// }
	}

	public Double get(TableRowId rowId1, TableRowId rowId2) {
		Integer id1 = mapping.get(rowId1);
		Integer id2 = mapping.get(rowId2);
		if (id1 < id2) {
			return get(id1, id2);
		} else {
			return get(id2, id1);
		}

	}

	private Double get(Integer id1, Integer id2) {
		OptionalClusterCacheMap innerMap = cache[id1];
		if (innerMap != null) {
			Double innerVal = innerMap.get(id2);
			if (innerVal != null) {
				return innerVal;
			}
		}
		Set<Integer> innerFullMap = pairs.get(id1);
		if (innerFullMap != null && innerFullMap.contains(id2)) {
			return -1.0;
		}
		return null;
	}

	public static class OptionalClusterCacheMap extends SparseArray<Double> {

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;


	}

	public void add(TableRowId rowId1, TableRowId rowId2, Double value) {
		Integer id1 = mapping.get(rowId1);
		Integer id2 = mapping.get(rowId2);
		if (id1 < id2) {
			add(id1, id2, value);
		} else {
			add(id2, id1, value);
		}

	}

	public void addPublic(Integer id1, Integer id2, Double value) {
		if (id1 < id2) {
			add(id1, id2, value);
		} else {
			add(id2, id1, value);
		}
	}

	private void add(Integer id1, Integer id2, Double value) {
		CHMHelper.getFromCHM(updateQueue, id1, () -> new ConcurrentHashMap<>()).put(id2, value);
	}

	private ConcurrentHashMap<Integer, ConcurrentHashMap<Integer, Double>> updateQueue = new ConcurrentHashMap<>();

	public static NumberFormat formatter = new DecimalFormat("#0.0000");

	public void writeUpdates() {
		writeUpdates(true);
	}

	public void writeUpdates(boolean store) {
		if (!updateQueue.isEmpty()) {

			// ConcurrentHashMap<Integer, ConcurrentHashMap<Integer, Double>>
			// updateQueueCopy = updateQueue;
			// updateQueue = new ConcurrentHashMap<>();

			ConcurrentLinkedQueue<Tri<Integer, Integer, Double>> queueFinal = new ConcurrentLinkedQueue<>();
			AtomicInteger doneComputing = new AtomicInteger(1);
			CountDownLatch doneWritingFile = new CountDownLatch(1);

			try {

				MKDIR.createDirUnchecked(new File(cacheFolder));

				String file = cacheFolder + "/extra_" + new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()) + "-"
						+ RandomStringUtils.randomAlphanumeric(4);

				PersistentParallel6.LIMIT_OBSERVER_RUN = true;

				if (store) {

				@SuppressWarnings("resource")
				BufferedWriter outPairs = new BufferedWriter(
						new OutputStreamWriter(new GzipCompressorOutputStream(new FileOutputStream(new File(file + ".gz"))), StandardCharsets.UTF_8));

				Runnable runA = new Runnable() {
					public void run() {
						try {
							Tri<Integer, Integer, Double> entry;
							boolean lastRun = false;
							while (true) {
								entry = queueFinal.poll();
								if (entry == null) {
									if (lastRun) {
										break;
									} else {
										java.util.concurrent.TimeUnit.SECONDS.sleep(5);
										if (doneComputing.intValue() == 0) {
											lastRun = true;
										}
									}
								} else {

									outPairs.write(
											String.valueOf(entry.getA()) + " " + String.valueOf(entry.getB()) + " " + formatter.format(entry.getC()) + "\n");

								}

							}
							outPairs.close();
							doneWritingFile.countDown();
						} catch (Exception e) {
							e.printStackTrace();
							throw new IllegalStateException(e);
						}
					}

				};

				Thread threadA = new Thread(runA, "threadA");
				threadA.start();

				}

			} catch (Exception e) {
				e.printStackTrace();
				throw new IllegalStateException(e);
			}

			PersistentParallel6.runAndFinish(updateQueue.entrySet(), (updateSet) -> {
				Integer id1 = updateSet.getKey();
				OptionalClusterCacheMap newSparseArray;
				if (cache[id1] == null) {

					newSparseArray = new OptionalClusterCacheMap();

				} else {
					newSparseArray = (OptionalClusterCacheMap) cache[id1].clone();

				}

				for (Entry<Integer, Double> entry : updateSet.getValue().entrySet()) {

					newSparseArray.append(entry.getKey(), entry.getValue());
					if (store) {
						queueFinal.add(new Tri<>(id1, entry.getKey(), entry.getValue()));
					}

				}

				cache[id1] = newSparseArray;

			}, this.getClass(), 16);

			updateQueue.clear();

			doneComputing.decrementAndGet();

			if (store) {

				try {
					doneWritingFile.await();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			// write(outputCacheFile, mapping, queueFinal);
			PersistentParallel6.LIMIT_OBSERVER_RUN = false;

		}

	}

}
