package expansion.set.learning;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableDouble;
import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.learning.RowValuesGetter;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.learning.SplitRowValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.DoubleIteraterable.DoubleEntry;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterer;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities.PairLearningData;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.evaluation.SetExpansionRunner;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.crossvalidation.EntitySetGroupedValuesSplitter;
import expansion.set.evaluation.crossvalidation.EntitySets;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;
import expansion.set.evaluation.goldstandard.oktie.OktieClusterEvaluator;
import expansion.set.evaluation.goldstandard.oktie.OktieClusterEvaluator.WeightGetter;
import expansion.set.evaluation.goldstandard.oktie.OktiePerformance;
import expansion.set.evaluation.sim.SimEvaluator;
import expansion.set.implicitAttributes.ImplicitAttributeFinder;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.SetExpansionStrategy;
import expansion.set.strategies.clustersplitters.minScoreSplitter.CombinedModel;
import expansion.set.strategies.rowclusterers.GreedyClusteringSplitter;
import expansion.set.strategies.rowclusterers.RowClustererRepository;

public class ClusteringModelLearnerRunner {

	@Inject
	protected RowValuesGetter rowValuesGetter;

	@Inject
	protected RowClustererRepository rowClustererRepository;

	@Inject
	protected AnnotatedTablesRepository tablesRepository;

	@Inject
	protected SetExpansionRunner setExpansionRunner;

	@Inject
	protected EntitySetGroupedValuesSplitter entitySetGroupedValuesSplitter;

	@Inject
	protected GreedyClusteringSplitter greedyClusteringSplitter;

	@Inject
	protected OktieClusterEvaluator oktieClusterEvaluator;

	@Inject
	protected SimEvaluator simEvaluator;

	@Inject
	protected WeightsOptimizer weightsParallelOptimizer;

	@Inject
	protected ImplicitAttributeFinder implicitAttributeFinder;


	public Map<ClassId, LearnedModelRepository> getAllWeights(Map<ClassId, LearnedModelRepository> repositoryMap, Map<ClassId, Set<TableRowId>> input,
			Map<ClassId, TableRowIdsSet> fullInput,
			FusionConfig fusionConfig,
			SetExpansionStrategiesConfig setExpansionStrategiesConfig, StrategiesConfig strategiesConfig, String thresholdsFolder,
			Map<ClassId, GoldStandardClusters> goldStandard, Map<ClassId, Set<InstanceId>> newEntities, Map<ClassId, Map<InstanceId, Integer>> conflictingSets,
			Map<ClassId, Set<ClassId>> allParents, Map<ClassId, Set<PropertyId>> nonTailProperties, TableRowIdsSet fullInputAllClasses)
			throws IOException {
		
		rowClustererRepository.init();
		implicitAttributeFinder.prepare(nonTailProperties);

		Map<ClassId, LearnedModelRepository> ret = new LinkedHashMap<>(repositoryMap);
		
		TableRowIdsSet rowIdsSetAllClasses;
		if (fullInputAllClasses == null) {
			Set<TableRowId> rowIdsAllClasses = new HashSet<>();
			for (TableRowIdsSet rowIds : fullInput.values()) {
				rowIdsAllClasses.addAll(rowIds.getRowIds());
			}
			rowIdsSetAllClasses = new TableRowIdsSet(rowIdsAllClasses, new UuidStringId("ALL"));

		} else {
			rowIdsSetAllClasses = fullInputAllClasses;
		}
		for(Entry<ClassId, Set<TableRowId>> inputEntry : input.entrySet()) {
			ClassId classId = inputEntry.getKey();
			Set<TableRowId> candidates = inputEntry.getValue();
			TableRowIdsSet fullCandidates = fullInput.get(classId);
			


			// List<SetExpansionStrategy> clusteringStrategies =
			// setExpansionStrategiesConfig.getStrategies().stream()
			// .filter((e) ->
			// !weightsRepository.contains(e.getOptimizableClassCluster())).collect(Collectors.toList());

			List<SetExpansionStrategy> similarityThresholdStrategies = setExpansionStrategiesConfig.getStrategies().stream()
					.filter((e) -> {
						return !ret.containsKey(classId) || !ret.get(classId).contains(e.getOptimizableClassSimilarityThreshold());
					}).collect(Collectors.toList());

			LearnedModelRepository weightsRepository = ret.get(classId);
			if (weightsRepository == null) {
				weightsRepository = new LearnedModelRepository();
			}

			if (!similarityThresholdStrategies.isEmpty()) {

				TableIdsSet tableIdsSet = new TableIdsSet(this.tablesRepository.getAllTableIds(classId), new UuidStringId(classId.getFinalPart()));

				Map<ClassId, Set<InstanceId>> instancesWithMatches = setExpansionRunner.getInstancesWithMatches(classId, tableIdsSet.getTableIds(), 1);
				if (instancesWithMatches.get(classId).isEmpty()) {
					System.out.println("No Instances with MinSource found for: " + classId.getClassId());
					throw new IllegalStateException();
				}

				EntitySets entitySets = new EntitySets(instancesWithMatches.get(classId), instancesWithMatches.get(classId));
				SplitGroupedValues splitGroupedValues = setExpansionRunner.getSplitGroupedValues(fusionConfig, tableIdsSet);
				SplitGroupedValuesSets splitGroupedValuesSets = entitySetGroupedValuesSplitter.split(splitGroupedValues, entitySets);
				SplitRowValues splitRowValues = rowValuesGetter.getSplitRowValues(tablesRepository, tableIdsSet);
				ClustererInput clustererInput = new ClustererInput(splitGroupedValuesSets, candidates, fullCandidates, tableIdsSet, null, classId,
						splitRowValues, allParents, nonTailProperties, rowIdsSetAllClasses);

				List<ThresholdContainer> thresholdContainers = setExpansionRunner.getThresholdContainers(setExpansionStrategiesConfig, thresholdsFolder,
						classId, fusionConfig, strategiesConfig, splitGroupedValues);



				LearnedModelRepository repository = runForThresholds(null, classId, similarityThresholdStrategies, clustererInput, setExpansionStrategiesConfig,
						thresholdContainers, goldStandard.get(classId), newEntities.get(classId), conflictingSets.get(classId), fusionConfig);

				// repository =
				// repository.combineWith(runForStrategies(repository,
				// clusteringStrategies, clustererInput,
				// setExpansionStrategiesConfig,
				// thresholdContainers, goldStandard.get(classId),
				// newEntities.get(classId), conflictingSets.get(classId)));
				



				weightsRepository = weightsRepository.combineWith(repository);
			}

			ret.put(classId, weightsRepository);


		}

		return ret;
	}

	// WeightsRepositoryImpl runForStrategies(WeightsRepositoryImpl
	// weightsRepositoryFromRMSE, List<SetExpansionStrategy> strategies,
	// ClustererInput clustererInput,
	// SetExpansionStrategiesConfig setExpansionStrategiesConfig,
	// List<ThresholdContainer> thresholdContainers, GoldStandardClusters
	// goldStandard,
	// Set<InstanceId> newEntities, Map<InstanceId, Integer> conflictingSets)
	// throws IOException {
	//
	// WeightsRepositoryImpl repository;
	// if (weightsRepositoryFromRMSE != null) {
	// repository = weightsRepositoryFromRMSE;
	// } else {
	// repository = new WeightsRepositoryImpl();
	// }
	//
	// Map<InstanceId, Double> upsampledWeights =
	// simEvaluator.getUpsampledClusterSet(clustererInput, goldStandard,
	// newEntities, conflictingSets);
	//
	// WeightGetter<InstanceId> weightsGetter = (rowId) -> {
	// return upsampledWeights.get(rowId);
	// };
	//
	// for (SetExpansionStrategy strategy : strategies) {
	// final WeightsRepositoryImpl finalRepository = repository;
	// WeightsRepository weightsRepository =
	// weightsParallelOptimizer.optimize((weightsRepositoryInput) -> {
	// RowClusteringConfig rowClusteringConfig = new
	// RowClusteringConfig(setExpansionStrategiesConfig, thresholdContainers,
	// finalRepository);
	//
	// rowClustererRepository.init();
	// RowClusterer clusterer =
	// rowClustererRepository.get(strategy.getClusterer());
	// RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig,
	// clustererInput);
	//
	// rowClusteringConfig = new
	// RowClusteringConfig(setExpansionStrategiesConfig, thresholdContainers,
	// weightsRepositoryInput);
	//
	// ClusterList clusters = greedyClusteringSplitter.clusterRows(sim,
	// rowClusteringConfig, clustererInput,
	// strategy.getOptimizableClassCluster());
	//
	// List<Set<TableRowId>> clustersAsSet =
	// clusters.getClusterList().stream().map((e) -> new
	// HashSet<>(e.getRows())).collect(Collectors.toList());
	// OktiePerformance clusteringPerformance =
	// oktieClusterEvaluator.oktieEvaluateWithGsMap(goldStandard.getTestingSetMappingClusters(),
	// clustersAsSet, newEntities, weightsGetter);
	//
	// return clusteringPerformance.getFMeasureMacro();
	// });
	//
	// repository = repository.combineWith(weightsRepository);
	//
	// }
	//
	// return repository;
	// }

	public LearnedModelRepository runForThresholds(LearnedModelRepository weightsRepositoryInput, ClassId classId, List<SetExpansionStrategy> strategies,
			ClustererInput clustererInput, SetExpansionStrategiesConfig setExpansionStrategiesConfig, List<ThresholdContainer> thresholdContainers,
			GoldStandardClusters goldStandard, Set<InstanceId> newEntities, Map<InstanceId, Integer> conflictingSets, FusionConfig fusionConfig)
			throws IOException {

		// List<Tri<TableRowIdPair, Boolean, Double>> evaluationSet =
		// simEvaluator.getUpsampledRMSESet(clustererInput, goldStandard,
		// newEntities, conflictingSets,
		// new Random(classId.getClassId().hashCode()));.

		// Map<TableRowIdPair, PairLearningData> evaluationSet =
		// simEvaluator.getUpsampledRowPairSet(clustererInput, goldStandard,
		// newEntities,
		// conflictingSets).entrySet().stream().collect(Collectors.toMap(Entry::getKey,
		// (e) -> new PairLearningData(e.getValue().getFirst(),
		// e.getValue().getSecond())));

		Map<OptimizableClass, LearnedModel> repository = new HashMap<>();
		if (weightsRepositoryInput != null) {
			repository = weightsRepositoryInput.getMap();
		}

		Map<InstanceId, Double> upsampledWeights = simEvaluator.getUpsampledClusterSet(clustererInput, goldStandard, newEntities, conflictingSets);

		// Map<TableRowIdPair, PairLearningData> upsampledRowIdPairsOld =
		// getUpsampledRowIdPairs(clustererInput.getCandidates(),
		// goldStandard.getIndex(), upsampledWeights, conflictingSets);

		Map<TableRowIdPair, PairLearningData> upsampledRowIdPairs = getUpsampledRowIdPairs(clustererInput.getCandidates(), goldStandard.getIndex(),
				upsampledWeights, conflictingSets);

		Map<TableRowIdPair, PairLearningData> notUpsampledRowIdPairs = getNonUpsampledRowIdPairs(clustererInput.getCandidates(), goldStandard.getIndex(),
				conflictingSets);

		WeightGetter<InstanceId> weightsGetter = (rowId) -> {
			return 1.0;
		};
		// WeightGetter<InstanceId> weightsGetter = (rowId) -> {
		// return upsampledWeights.get(rowId);
		// // return 1.0;
		// };
		
		EvaluationMethod evaluationMethod = getEvaluationMethod();

		RowClusteringConfig rowClusteringConfig = new RowClusteringConfig(setExpansionStrategiesConfig, thresholdContainers, fusionConfig);

		for (SetExpansionStrategy strategy : strategies) {
			if (strategy == SetExpansionStrategiesRepository.EXTERNAL)
				continue;
			if (strategy == SetExpansionStrategiesRepository.GS_CLUSTERER)
				continue;
			if (strategy.getName().startsWith("GS_CLUSTERER"))
				continue;
			if (strategy == SetExpansionStrategiesRepository.FUSION_TESTER)
				continue;
			RowClusterer clusterer = strategy.getClustererImpl();
			if (clusterer == null)
				clusterer = rowClustererRepository.get(strategy.getClusterer());

			RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);
			if (sim instanceof LearnableRowClusterSimilarities) {
				Map<TableRowIdPair, PairLearningData> rowIdPairs;

				if (true || (strategy == SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW
						|| strategy == SetExpansionStrategiesRepository.COMBINED_RF_BASE)) {
					rowIdPairs = notUpsampledRowIdPairs;
				} else {
					rowIdPairs = upsampledRowIdPairs;
				}

				int positivePairsSize = 0;
				int negativePairsSize = 0;

				for (PairLearningData v : rowIdPairs.values()) {
					if (v.isMatch()) {
						positivePairsSize++;
					} else {
						negativePairsSize++;
					}
				}

				System.out.println("For Class: " + classId.getFinalPart() + "ForMethod:" + strategy.getName() + " Positive Pairs " + positivePairsSize
						+ ", Negative Pairs: " + negativePairsSize);

				LearnedModel learnedModel = learn(classId, (LearnableRowClusterSimilarities) sim, strategy, rowIdPairs, evaluationMethod, clusterer,
						setExpansionStrategiesConfig, thresholdContainers, clustererInput, goldStandard, newEntities, weightsGetter, fusionConfig);
				repository.put(strategy.getOptimizableClassCluster(), learnedModel);
				
				// if
				// (strategy.equals(SetExpansionStrategiesRepository.COMBINED_LMT))
				// {
				//
				// System.out.println(classId.getClassId());
				// System.out.println(learnedModel.toString());
				// }

			}




			// Map<OptimizableWeightKey, Double> thresholdMap = new HashMap<>();
			// thresholdMap.put(strategy.getOptimizableClassSimilarityThreshold().getWeight("SIMILARITY_THRESHOLD",
			// 0.5).getKey(), 0.5);
			// WeightsRepositoryImpl threshold = new
			// WeightsRepositoryImpl(thresholdMap);

			// WeightsRepository weightsRepository =
			// weightsParallelOptimizer.optimize((weightsRepositoryToOptimize)
			// -> {
			// RowClusteringConfig rowClusteringConfig = new
			// RowClusteringConfig(setExpansionStrategiesConfig,
			// thresholdContainers,
			// weightsRepositoryToOptimize);
			//
			// RowClusterSimilarities sim =
			// clusterer.clusterRows(rowClusteringConfig, clustererInput);
			//
			// double rmse =
			// simEvaluator.evaluateFMeasure(weightsRepositoryToOptimize,
			// strategy.getOptimizableClassCluster(), clustererInput, sim,
			// weightsGetter, goldStandard);
			// // SimPerformance simPerformance =
			// // simEvaluator.evaluateSim(threshold,
			// // strategy.getOptimizableClassSimilarityThreshold(), sim,
			// // rowClusteringConfig, clustererInput, goldStandard,
			// // );
			//
			// return rmse;
			// });

			
			// repository = repository.combineWith(threshold);

		}

		return new LearnedModelRepository(repository);
	}

	public EvaluationMethod getEvaluationMethod() {
		return (predictions) -> {
			double rmse = 0.0;
			// double sumPositive = 0.0;
			// double sumNegative = 0.0;
			// for (EvaluationPrediction prediction : predictions) {
			// if
			// (prediction.getLabel().equals(LearnableRowClusterSimilarities.MATCH_CLASS))
			// {
			// sumPositive += prediction.getWeight();
			// if (prediction.getWeight() > 1.01 || prediction.getWeight() <
			// 0.99) {
			// int sdf = 123;
			// }
			// } else {
			// if (prediction.getWeight() > 1.01 || prediction.getWeight() <
			// 0.99) {
			// int sdf = 123;
			// }
			// sumNegative += prediction.getWeight();
			// }
			// }
			// double maxClass = sumPositive;
			// if (sumNegative > sumPositive) {
			// maxClass = sumNegative;
			// }
			// double upSamplePositive = maxClass / sumPositive;
			// double upSampleNegative = maxClass / sumNegative;
			// upSamplePositive = 1.0;
			// upSampleNegative = 1.0;

			// double maxCorrectInt = 0;
			// double recallCorrectInt = 0;
			double correct = 0.0;
			// double recallCorrect = 0.0;
			double wrong = 0;
			double maxCorrect = 0;
			// double weights = 0.0;

			for (EvaluationPrediction prediction : predictions) {

				if (prediction.getLabel().equals(LearnableRowClusterSimilarities.MATCH_CLASS)) {
					// double weight = prediction.getWeight() *
					// upSamplePositive;
					// weights += weight;
					//
					// maxCorrect += prediction.getWeight();
					maxCorrect++;
					if (prediction.getPredictedLabel().equals(LearnableRowClusterSimilarities.MATCH_CLASS)) {
						// correct += prediction.getWeight() * upSamplePositive;
						// recallCorrect += prediction.getWeight();
						// recallCorrectInt++;
						correct++;
					} else {

						rmse++;
					}
				} else {
					// double weight = prediction.getWeight() *
					// upSampleNegative;
					// weights += weight;
					if (prediction.getPredictedLabel().equals(LearnableRowClusterSimilarities.MATCH_CLASS)) {
						wrong++;


						rmse++;


					}
				}
			}

			double precision = correct / (correct + wrong);
			if (Double.isNaN(precision)) {
				precision = 0.0;
			}
			// double recallNonUpsampled = recallCorrectInt / maxCorrectInt;

			double recall = correct / maxCorrect;
			if (Double.isNaN(recall)) {
				recall = 0.0;
			}

			double fMeasure = (2.0 * precision * recall) / (precision + recall);
			if (Double.isNaN(fMeasure)) {
				fMeasure = 0.0;
			}

			// return fMeasure;
			double rmseFinal = Math.sqrt(rmse / predictions.size());
			return fMeasure;
		};
	}

	private Map<TableRowIdPair, PairLearningData> getNonUpsampledRowIdPairs(Set<TableRowId> candidates, Map<TableRowId, InstanceId> index,
			Map<InstanceId, Integer> conflictingSets) {

		double correct = 0.0;
		double incorrectWithoutConflict = 0.0;
		double incorrectWithConflict = 0.0;
		for (DoubleEntry<TableRowId> doubleEntry : DoubleIteraterable.iterable(candidates)) {
			TableRowIdPair pairId = new TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond());
			if (index.get(doubleEntry.getFirst()).equals(index.get(doubleEntry.getSecond()))) {

				correct++;

			} else {

				Integer set1 = conflictingSets.get(index.get(doubleEntry.getFirst()));
				Integer set2 = conflictingSets.get(index.get(doubleEntry.getSecond()));

				if (set1 != null && set1.equals(set2)) {
					incorrectWithConflict++;
				} else {
					incorrectWithoutConflict++;
				}

			}
		}

		double max = 0.0;
		if (incorrectWithoutConflict > incorrectWithConflict) {
			max = incorrectWithoutConflict;
		} else {
			max = incorrectWithConflict;
		}

		double incorrectWithConflictMultipliyer = 1.0;// (max /
														// incorrectWithConflict);
		double incorrectWithoutConflictMultipliyer = 1.0;// (max /
															// incorrectWithoutConflict);

		double sumIncorrect = incorrectWithConflict * incorrectWithConflictMultipliyer + incorrectWithoutConflict * incorrectWithoutConflictMultipliyer;

		double maxAll = 0.0;
		if (sumIncorrect > correct) {
			maxAll = sumIncorrect;
		} else {
			maxAll = correct;
		}

		double incorrectMultiplyer = maxAll / sumIncorrect;

		double correctMultiplyer = maxAll / correct;

		Map<TableRowIdPair, PairLearningData> ret = new HashMap<>();

		for (DoubleEntry<TableRowId> doubleEntry : DoubleIteraterable.iterable(candidates)) {
			TableRowIdPair pairId = new TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond());
			if (index.get(doubleEntry.getFirst()).equals(index.get(doubleEntry.getSecond()))) {

				ret.put(pairId, new PairLearningData(true, 1.0));

			} else {

				Integer set1 = conflictingSets.get(index.get(doubleEntry.getFirst()));
				Integer set2 = conflictingSets.get(index.get(doubleEntry.getSecond()));

				if (set1 != null && set1.equals(set2)) {
					ret.put(pairId, new PairLearningData(false, 1.0));
				} else {
					ret.put(pairId, new PairLearningData(false, 1.0));

				}

			}
		}

		return ret;
	}

	private Map<TableRowIdPair, PairLearningData> getUpsampledRowIdPairs(Set<TableRowId> candidates, Map<TableRowId, InstanceId> index,
			Map<InstanceId, Double> upsampledWeights) {

		Map<TableRowIdPair, PairLearningData> ret = new HashMap<>();

		for (DoubleEntry<TableRowId> doubleEntry : DoubleIteraterable.iterable(candidates)) {
			TableRowIdPair pairId = new TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond());
			if (index.get(doubleEntry.getFirst()).equals(index.get(doubleEntry.getSecond()))) {
				double weight = upsampledWeights.get(index.get(pairId.getId1()));
				if (weight > 1.1) {
					int sdf = 213;
				}
				ret.put(pairId, new PairLearningData(true, weight));
			} else {

				ret.put(pairId, new PairLearningData(false, 1.0));
			}
		}

		return ret;
	}

	private Map<TableRowIdPair, PairLearningData> getUpsampledRowIdPairs(Set<TableRowId> candidates, Map<TableRowId, InstanceId> index,
			Map<InstanceId, Double> upsampledWeights, Map<InstanceId, Integer> conflictingSets) {

		Map<InstanceId, MutableInt> countPositive = new HashMap<>();
		Map<InstanceId, MutableInt> rowCount = new HashMap<>();

		for (InstanceId instanceId : index.values()) {
			CHMHelper.getFromMap(rowCount, instanceId, () -> new MutableInt()).increment();
		}

		int incorrect = 0;
		double incorrectConflict = 0;

		int correct = 0;
		double correctConflict = 0;
		for (DoubleEntry<TableRowId> doubleEntry : DoubleIteraterable.iterable(candidates)) {

			InstanceId i1 = index.get(doubleEntry.getFirst());
			InstanceId i2 = index.get(doubleEntry.getSecond());

			if (i1.equals(i2)) {
				CHMHelper.getFromMap(countPositive, i1, () -> new MutableInt()).increment();

				Integer set1 = conflictingSets.get(index.get(doubleEntry.getFirst()));
				Integer set2 = conflictingSets.get(index.get(doubleEntry.getSecond()));

				if (set1 != null && set1.equals(set2)) {
					correctConflict++;
					// incorrect++;
				} else {
					correct++;
				}
			}
			else {
				Integer set1 = conflictingSets.get(index.get(doubleEntry.getFirst()));
				Integer set2 = conflictingSets.get(index.get(doubleEntry.getSecond()));

				if (set1 != null && set1.equals(set2)) {
					incorrectConflict++;
					// incorrect++;
				} else {
					incorrect++;
				}
			}
		}

		double max = countPositive.values().stream().mapToInt(MutableInt::intValue).max().getAsInt();

		Map<InstanceId, Double> intanceWeights = countPositive.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, (e) -> max / e.getValue().intValue()));

		double newMax = rowCount.values().stream().mapToInt(MutableInt::intValue).max().getAsInt();

		Map<InstanceId, Double> newIntanceWeights = rowCount.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, (e) -> newMax / e.getValue().intValue()));

		// for (InstanceId instanceId : upsampledWeights.keySet()) {
		// if (conflictingSets.containsKey(instanceId)) {
		//
		// incorrectConflict++;
		// } else {
		// incorrect++;
		// }
		//
		// }
		double umpsampleIncorrectConflict = 1.0;
		if (incorrectConflict < incorrect)
			umpsampleIncorrectConflict = incorrect / incorrectConflict;

		double upsampleCorrectConflict = 1.0;
		if (correctConflict < correct)
			upsampleCorrectConflict = correct / correctConflict;

		Map<TableRowIdPair, PairLearningData> ret = new HashMap<>();

		for (DoubleEntry<TableRowId> doubleEntry : DoubleIteraterable.iterable(candidates)) {
			TableRowIdPair pairId = new TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond());
			if (index.get(doubleEntry.getFirst()).equals(index.get(doubleEntry.getSecond()))) {
				InstanceId instanceId = index.get(doubleEntry.getFirst());
				if (false) {
					double weight = upsampledWeights.get(index.get(pairId.getId1()));
					if (weight > 1.1) {
						int sdf = 213;
					}
					ret.put(pairId, new PairLearningData(true, weight));
				} else {
					Integer set1 = conflictingSets.get(index.get(doubleEntry.getFirst()));
					Integer set2 = conflictingSets.get(index.get(doubleEntry.getSecond()));
					double weight = upsampledWeights.get(index.get(pairId.getId1()));
					double instanceWeight = intanceWeights.get(instanceId);
					weight = newIntanceWeights.get(instanceId);
					weight = 1.0;
					if (set1 != null && set1.equals(set2)) {
						double weight1 = upsampledWeights.get(index.get(pairId.getId1()));
						double weight2 = upsampledWeights.get(index.get(pairId.getId2()));

						// double weight = Math.max(weight1, weight2);
						weight = upsampleCorrectConflict;
						// weight = 1.0;
						ret.put(pairId, new PairLearningData(true, weight));
					} else {

						ret.put(pairId, new PairLearningData(true, weight));
					}
				}
			} else {


				Integer set1 = conflictingSets.get(index.get(doubleEntry.getFirst()));
				Integer set2 = conflictingSets.get(index.get(doubleEntry.getSecond()));
				double weight1 = newIntanceWeights.get(index.get(pairId.getId1()));
				double weight2 = newIntanceWeights.get(index.get(pairId.getId2()));
				double weight = 1.0;// Math.max(weight1, weight2);

				if (set1 != null && set1.equals(set2)) {



					weight = umpsampleIncorrectConflict;
					ret.put(pairId, new PairLearningData(false, weight));
				} else {

					ret.put(pairId, new PairLearningData(false, weight));
				}
			}
		}

		Map<String, MutableDouble> classCount = new HashMap<>();

		for (Entry<TableRowIdPair, PairLearningData> pair : ret.entrySet()) {
			CHMHelper.getFromMap(classCount, pair.getValue().getLabel(), () -> new MutableDouble()).add(pair.getValue().getWeight());
		}

		String maxClass = null;
		double classMax = -1;
		for (Entry<String, MutableDouble> entry : classCount.entrySet()) {
			if (maxClass == null || entry.getValue().intValue() > classMax) {
				classMax = entry.getValue().doubleValue();
				maxClass = entry.getKey();
			}
		}
		double maxDouble = (double) classMax;

		Map<String, Double> replicationMap = new HashMap<>();
		for (Entry<String, MutableDouble> entry : classCount.entrySet()) {
			replicationMap.put(entry.getKey(), maxDouble / entry.getValue().doubleValue());
		}

		ret = ret.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			double upsampling = replicationMap.get(e.getValue().getLabel());
			return new PairLearningData(e.getValue().isMatch(), e.getValue().getWeight() * upsampling);
		}));

		return ret;
	}

	public LearnedModel learn(ClassId classId, LearnableRowClusterSimilarities learnableClusterer, SetExpansionStrategy strategy,
			Map<TableRowIdPair, PairLearningData> rowIdPairs, EvaluationMethod evaluationMethod, RowClusterer clusterer,
			SetExpansionStrategiesConfig setExpansionStrategiesConfig, List<ThresholdContainer> thresholdContainers, ClustererInput clustererInput,
			GoldStandardClusters goldStandard, Set<InstanceId> newEntities, WeightGetter<InstanceId> weightsGetter, FusionConfig fusionConfig) {

		LearnedModel learnedModel = learnableClusterer.learn(rowIdPairs, classId, evaluationMethod);

		RowClusterer finalClusterer = clusterer;
		if (learnedModel instanceof CombinedModel) {
			CombinedModel casted = (CombinedModel) learnedModel;

			WeightsRepository learnedWeights;
			if (goldStandard == null) {

				Map<OptimizableWeightKey, Double> combinedWeightsMap = new HashMap<>();
				combinedWeightsMap.put(strategy.getOptimizableClassCluster().getKey("aWeight"), 0.0);

				learnedWeights = new WeightsRepositoryImpl(combinedWeightsMap);

				if (false && true) {
					learnedWeights = weightsParallelOptimizer.optimize((weights) -> {

						Map<OptimizableClass, LearnedModel> repositorySecond = new HashMap<>();
						CombinedModel adaptedModel = new CombinedModel(casted, weights);

						repositorySecond.put(strategy.getOptimizableClassCluster(), adaptedModel);
						boolean disableBlocking = true;
						RowClusteringConfig rowClusteringConfigSecond = new RowClusteringConfig(setExpansionStrategiesConfig, thresholdContainers,
								new LearnedModelRepository(repositorySecond), disableBlocking, fusionConfig);

						RowClusterSimilarities simSecond = finalClusterer.clusterRows(rowClusteringConfigSecond, clustererInput);

						List<EvaluationPrediction> predictions = new ArrayList<>();
						for (Entry<TableRowIdPair, PairLearningData> entry : rowIdPairs.entrySet()) {
							List<TableRowId> rows = new ArrayList<>();
							rows.add(entry.getKey().getId1());
							rows.add(entry.getKey().getId2());
							List<WeightedScore> score = simSecond.getList(rows, classId, new ClusterEntryKey(0, 1));
							if (score.isEmpty() || score.size() > 1) {
								throw new IllegalStateException();
							}
							double s = score.get(0).getScore();
							if (s > 0.0) {
								predictions.add(new EvaluationPrediction(entry.getValue().getLabel(), entry.getValue().getWeight(),
										new RecordPrediction(LearnableRowClusterSimilarities.getLabel(true), s)));
							} else {
								predictions.add(new EvaluationPrediction(entry.getValue().getLabel(), entry.getValue().getWeight(),
										new RecordPrediction(LearnableRowClusterSimilarities.getLabel(false), s)));
							}
						}

						double performance = evaluationMethod.performance(predictions);

						return performance;
					});
				}
			} else {
				learnedWeights = weightsParallelOptimizer.optimize((weights) -> {

					Map<OptimizableClass, LearnedModel> repositorySecond = new HashMap<>();
					CombinedModel adaptedModel = new CombinedModel(casted, weights);
					repositorySecond.put(strategy.getOptimizableClassCluster(), adaptedModel);
					boolean disableBlocking = true;
					RowClusteringConfig rowClusteringConfigSecond = new RowClusteringConfig(setExpansionStrategiesConfig, thresholdContainers,
							new LearnedModelRepository(repositorySecond), disableBlocking, fusionConfig);

					RowClusterSimilarities simSecond = finalClusterer.clusterRows(rowClusteringConfigSecond, clustererInput);

					ClusterList clusters = greedyClusteringSplitter.clusterRows(simSecond, rowClusteringConfigSecond, clustererInput,
							strategy.getOptimizableClassCluster(), disableBlocking);

					List<Set<TableRowId>> clustersAsSet = clusters.getClusterList().stream().map((e) -> new HashSet<>(e.getRows()))
							.collect(Collectors.toList());
					OktiePerformance clusteringPerformance = oktieClusterEvaluator.oktieEvaluateWithGsMap(goldStandard.getTestingSetMappingClusters(),
							clustersAsSet, newEntities, weightsGetter);

					double fmeasure = clusteringPerformance.getFMeasureMacro();
					// double penalty = -1.0 *
					// (adaptedModel.getNegativeConfidenceThreshold().get()
					// +
					// adaptedModel.getPositiveConfidenceThreshold().get())
					// * 0.001;
					double penalty = 0.0;
					// double penalty2 =
					// adaptedModel.getVeryLowPairThreshold().get() * 0.04;
					// double penalty3 =
					// adaptedModel.getPositiveConfidenceThreshold().get() *
					// 0.01;
					// double penalty4 =
					// adaptedModel.getNegativeConfidenceThreshold().get() *
					// 0.01;
					return fmeasure - penalty;
				});
			}

			// System.out.println(classId.getFinalPart());
			// System.out.println(learnedWeights.toString());

			learnedModel = new CombinedModel(casted, learnedWeights);
			if (strategy == SetExpansionStrategiesRepository.COMBINED_RF_BASE
					// || strategy ==
					// SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi_Attribute_Implicit
					|| strategy == SetExpansionStrategiesRepository.RF_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE
					|| strategy == SetExpansionStrategiesRepository.WA_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE
					|| strategy == SetExpansionStrategiesRepository.ET_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE) {
				System.out.println("=============================");
				System.out.println(classId.getFinalPart());
				System.out.println(learnedModel.toString());
			}
		}
		else if (strategy == SetExpansionStrategiesRepository.COMBINED_RF_BASE
				// || strategy ==
				// SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi_Attribute_Implicit
				|| strategy == SetExpansionStrategiesRepository.RF_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE
				|| strategy == SetExpansionStrategiesRepository.WA_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE
				|| strategy == SetExpansionStrategiesRepository.ET_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE) {
			System.out.println("=============================");
			System.out.println(classId.getFinalPart());
			System.out.println(learnedModel.toString());
		}
		if (false && strategy == SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW && learnedModel instanceof MT_WekaLearnedModel) {
			MT_WekaLearnedModel casted = (MT_WekaLearnedModel) learnedModel;

			WeightsRepository learnedWeights = weightsParallelOptimizer.optimize((weights) -> {

				Map<OptimizableClass, LearnedModel> repositorySecond = new HashMap<>();
				MT_WekaLearnedModel adaptedModel = new MT_WekaLearnedModel(casted, weights);
				repositorySecond.put(strategy.getOptimizableClassCluster(), adaptedModel);
				boolean disableBlocking = true;
				RowClusteringConfig rowClusteringConfigSecond = new RowClusteringConfig(setExpansionStrategiesConfig, thresholdContainers,
						new LearnedModelRepository(repositorySecond), disableBlocking, fusionConfig);

				RowClusterSimilarities simSecond = finalClusterer.clusterRows(rowClusteringConfigSecond, clustererInput);

				ClusterList clusters = greedyClusteringSplitter.clusterRows(simSecond, rowClusteringConfigSecond, clustererInput,
						strategy.getOptimizableClassCluster(), disableBlocking);

				List<Set<TableRowId>> clustersAsSet = clusters.getClusterList().stream().map((e) -> new HashSet<>(e.getRows())).collect(Collectors.toList());
				OktiePerformance clusteringPerformance = oktieClusterEvaluator.oktieEvaluateWithGsMap(goldStandard.getTestingSetMappingClusters(),
						clustersAsSet, newEntities, weightsGetter);

				double fmeasure = clusteringPerformance.getFMeasureMacro();
				double penalty = -1.0 * (adaptedModel.getNegativeConfidenceThreshold().get() + adaptedModel.getPositiveConfidenceThreshold().get()) * 0.001;
				// double penalty2 =
				// adaptedModel.getVeryLowPairThreshold().get() * 0.04;
				// double penalty3 =
				// adaptedModel.getPositiveConfidenceThreshold().get() *
				// 0.01;
				// double penalty4 =
				// adaptedModel.getNegativeConfidenceThreshold().get() *
				// 0.01;
				return fmeasure - penalty;
			});
			System.out.println(classId.getFinalPart());
			System.out.println(learnedWeights.toString());
			learnedModel = new MT_WekaLearnedModel(casted, learnedWeights);
		}

		return learnedModel;

	}
}
