package expansion.set.learning;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import javax.inject.Inject;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class EqualSetSplitter {

	@Inject
	protected WeightsOptimizer weightsParallelOptimizer;

	@Inject
	public EqualSetSplitter() {

	}

	public Map<Integer, Set<Integer>> getSetNumberPriority(List<Integer> setSizes, int denominator) {
		double sum = setSizes.stream().mapToInt(Integer::intValue).sum();

		double targetPerSet = sum / denominator;
		double targetCountPerSet = setSizes.size() / denominator;

		double step = 1.0 / 4;

		TreeMap<Double, Integer> floorMap = new TreeMap<>();
		for (int i = 0; i < denominator; i++) {
			floorMap.put(step * i, i);
		}

		WeightsRepository finalWeights = weightsParallelOptimizer.optimize((weights) -> {

			Map<Integer, Set<Integer>> distribution = getDistribution(denominator, floorMap, weights, setSizes.size());

			double performance = getPerformanceNumberPriority(denominator, distribution, targetCountPerSet, targetPerSet, setSizes);
			return performance;
		});

		Map<Integer, Set<Integer>> finalDistribution = getDistribution(denominator, floorMap, finalWeights, setSizes.size());
		System.out.println(getPerformanceNumberPriority(denominator, finalDistribution, targetCountPerSet, targetPerSet, setSizes));

		return finalDistribution;
	}

	public Map<Integer, Set<Integer>> get(List<Integer> setCount, List<Integer> setSizes, int denominator) {
		double sizesSum = setSizes.stream().mapToInt(Integer::intValue).sum();
		double sumInnerCount = setCount.stream().mapToInt(Integer::intValue).sum();
		double sumCount = setSizes.size();

		double targetSizesPerSet = sizesSum / denominator;
		double targetInnerCountPerSet = sumInnerCount / denominator;
		double targetCountPerSet = sumCount / denominator;

		double step = 1.0 / 4;

		TreeMap<Double, Integer> floorMap = new TreeMap<>();
		for (int i = 0; i < denominator; i++) {
			floorMap.put(step * i, i);
		}

		WeightsRepository finalWeights = weightsParallelOptimizer.optimize((weights) -> {

			Map<Integer, Set<Integer>> distribution = getDistribution(denominator, floorMap, weights, setSizes.size());

			double performance = getPerformanceWithSizeInnerCountAndCount(denominator, distribution, targetCountPerSet, targetInnerCountPerSet,
					targetSizesPerSet, setCount, setSizes);
			return performance;
		});

		Map<Integer, Set<Integer>> finalDistribution = getDistribution(denominator, floorMap, finalWeights, setSizes.size());
		System.out.println(getPerformanceWithSizeInnerCountAndCount(denominator, finalDistribution, targetCountPerSet, targetInnerCountPerSet,
				targetSizesPerSet, setCount, setSizes));
		return finalDistribution;
	}

	public Map<Integer, Set<Integer>> get(List<Integer> setSizes, int denominator) {
		double sum = setSizes.stream().mapToInt(Integer::intValue).sum();

		double targetPerSet = sum / denominator;

		double step = 1.0 / 4;

		TreeMap<Double, Integer> floorMap = new TreeMap<>();
		for (int i = 0; i < denominator; i++) {
			floorMap.put(step * i, i);
		}

		WeightsRepository finalWeights = weightsParallelOptimizer.optimize((weights) -> {

			Map<Integer, Set<Integer>> distribution = getDistribution(denominator, floorMap, weights, setSizes.size());

			double performance = getPerformance(denominator, distribution, targetPerSet, setSizes);
			return performance;
		});

		Map<Integer, Set<Integer>> finalDistribution = getDistribution(denominator, floorMap, finalWeights, setSizes.size());
		System.out.println(getPerformance(denominator, finalDistribution, targetPerSet, setSizes));
		return finalDistribution;
	}

	public OptimizableClass optimizableClass = new OptimizableClass("GOLD_STANDARD");

	public Map<Integer, Set<Integer>> getDistribution(int denominator, TreeMap<Double, Integer> floorMap, WeightsRepository weights, int conflictSets) {
		Map<Integer, Set<Integer>> ret = new HashMap<>();

		for (int i = 0; i < denominator; i++) {
			ret.put(i, new HashSet<>());
		}
		for (int i = 0; i < conflictSets; i++) {
			double weight = weights.getOrDefault(optimizableClass.getWeight(String.valueOf(i), 0.5));
			Integer targetSet = floorMap.floorEntry(weight).getValue();

			CHMHelper.getFromMap(ret, targetSet, () -> new HashSet<>()).add(i);
		}

		return ret;

	}

	public double getPerformanceWithSizeInnerCountAndCount(int denominator, Map<Integer, Set<Integer>> distribution, double targetCountPerSet,
			double targetInnerCountPerSet, double targetSetSizePerSet, List<Integer> conflictingSetsCount, List<Integer> conflictSetSizes) {
		double count = 1.0;
		double sizePerSet = 1.0;
		double innerCount = 1.0;

		double weight = 1.0 / denominator;

		for (int i = 0; i < denominator; i++) {
			double sum = distribution.get(i).stream().mapToInt((e) -> conflictSetSizes.get(e)).sum();
			if (sum > targetSetSizePerSet) {
				sizePerSet -= weight * (1.0 - (targetSetSizePerSet / sum));
			} else {
				sizePerSet -= weight * (1.0 - (sum / targetSetSizePerSet));

			}

			double innerCountSum = distribution.get(i).stream().mapToInt((e) -> conflictingSetsCount.get(e)).sum();
			if (innerCountSum > targetInnerCountPerSet) {
				innerCount -= weight * (1.0 - (targetInnerCountPerSet / innerCountSum));
			} else {
				innerCount -= weight * (1.0 - (innerCountSum / targetInnerCountPerSet));

			}

			double size = distribution.get(i).size();
			if (size > targetCountPerSet) {
				count -= weight * (1.0 - (targetCountPerSet / size));
			} else {
				sizePerSet -= weight * (1.0 - (size / targetCountPerSet));
			}
		}

		return count * sizePerSet * innerCount;
	}

	public double getPerformanceNumberPriority(int denominator, Map<Integer, Set<Integer>> distribution, double targetCountPerSet, double targetPerSet,
			List<Integer> conflictSetSizes) {
		double count = 1.0;
		double sizePerSet = 1.0;

		double weight = 1.0 / denominator;

		for (int i = 0; i < denominator; i++) {
			double sum = distribution.get(i).stream().mapToInt((e) -> conflictSetSizes.get(e)).sum();
			if (sum > targetPerSet) {
				sizePerSet -= weight * (1.0 - (targetPerSet / sum));
			} else {
				sizePerSet -= weight * (1.0 - (sum / targetPerSet));

			}

			double size = distribution.get(i).size();
			if (size > targetCountPerSet) {
				count -= weight * (1.0 - (targetCountPerSet / size));
			} else {
				sizePerSet -= weight * (1.0 - (size / targetCountPerSet));
			}
		}

		return count * sizePerSet;
	}

	public double getPerformance(int denominator, Map<Integer, Set<Integer>> distribution, double targetPerSet, List<Integer> conflictSetSizes) {
		double ret = 1.0;

		double weight = 1.0 / denominator;

		for (int i = 0; i < denominator; i++) {
			double sum = distribution.get(i).stream().mapToInt((e) -> conflictSetSizes.get(e)).sum();
			if (sum > targetPerSet) {
				ret -= weight * (1.0 - (targetPerSet / sum));
			} else {
				ret -= weight * (1.0 - (sum / targetPerSet));

			}
		}

		return ret;
	}

	public Map<Integer, Set<Integer>> getSetNumberPriority(List<Integer> setSizes, int denominator, Map<Integer, Pair<Integer, Integer>> existingCountAndSize) {
		double sum = setSizes.stream().mapToInt(Integer::intValue).sum();

		double targetPerSet = (sum + existingCountAndSize.values().stream().mapToInt((e) -> e.getSecond()).sum()) / denominator;
		double targetCountPerSet = (setSizes.size() + existingCountAndSize.values().stream().mapToInt((e) -> e.getFirst()).sum()) / denominator;

		double step = 1.0 / 4;

		TreeMap<Double, Integer> floorMap = new TreeMap<>();
		for (int i = 0; i < denominator; i++) {
			floorMap.put(step * i, i);
		}

		WeightsRepository finalWeights = weightsParallelOptimizer.optimize((weights) -> {

			Map<Integer, Set<Integer>> distribution = getDistribution(denominator, floorMap, weights, setSizes.size());

			double performance = getPerformanceNumberPriority(denominator, distribution, targetCountPerSet, targetPerSet, setSizes, existingCountAndSize);
			return performance;
		});

		Map<Integer, Set<Integer>> finalDistribution = getDistribution(denominator, floorMap, finalWeights, setSizes.size());
		System.out.println(getPerformanceNumberPriority(denominator, finalDistribution, targetCountPerSet, targetPerSet, setSizes, existingCountAndSize));

		return finalDistribution;
	}

	private double getPerformanceNumberPriority(int denominator, Map<Integer, Set<Integer>> distribution, double targetCountPerSet, double targetPerSet,
			List<Integer> conflictSetSizes, Map<Integer, Pair<Integer, Integer>> existingCountAndSize) {

		double count = 1.0;
		double sizePerSet = 1.0;

		double weight = 1.0 / denominator;

		for (int i = 0; i < denominator; i++) {
			double sum = distribution.get(i).stream().mapToInt((e) -> conflictSetSizes.get(e)).sum() + existingCountAndSize.get(i).getSecond();
			if (sum > targetPerSet) {
				sizePerSet -= weight * (1.0 - (targetPerSet / sum));
			} else {
				sizePerSet -= weight * (1.0 - (sum / targetPerSet));

			}

			double size = distribution.get(i).size() + existingCountAndSize.get(i).getFirst();
			if (size > targetCountPerSet) {
				count -= weight * (1.0 - (targetCountPerSet / size));
			} else {
				sizePerSet -= weight * (1.0 - (size / targetCountPerSet));
			}
		}

		return count * sizePerSet;
	}
}
