package expansion.set.profiling;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;
import java.util.stream.DoubleStream;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableDouble;
import org.apache.commons.lang.mutable.MutableInt;
import org.apache.commons.math3.util.CombinatoricsUtils;

import com.google.common.primitives.Ints;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Quad;
import de.mannheim.uni.expansion.utils.containers.Tri;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.creation.Entity;
import expansion.set.tablerow.AnnotatedTableRow;
import expansion.set.tablerow.TableRowCreator;

public class ClusterListEvaluator {

	
	@Inject
	protected AnnotatedTablesRepository tableRepository;
	
	@Inject
	protected TableRowCreator tableRowCreator;
	
	public ClusterListEvaluator() {
	
	}
	
	public List<Integer> blockSizes(ClusterList clusters) {
		return 	clusters.getClusterList().stream().map((e) -> e.getRows().size()).collect(Collectors.toList());
	}
	
	public List<Integer> blockSizes(Collection<Entity> entities) {
		return 	entities.stream().map((e) -> e.getRowCluster().getRows().size()).collect(Collectors.toList());
	}
	
	public List<Integer> entityValueSizes(Collection<Entity> entities) {
		return 	entities.stream().map((e) -> e.getValues().size()).collect(Collectors.toList());
	}
	
	public List<Integer> distinctInstancesPerBlock(ClusterList clusters) {
		return 	clusters.getClusterList().stream().map((e) ->  {
			int size = (int) e.getRows().stream().map(tableRepository::getMatchedInstanceId).filter(Optional::isPresent).map(Optional::get).distinct().count();
			return size;
			
		}).collect(Collectors.toList());
	}
	
	
	public Tri<Integer, Integer, Integer> getTotalCorrectAndMaxCorrect(Set<InstanceId> relevantSet, Map<InstanceId, RowCluster> clusterWithMaxInstanceCount, Set<TableRowId> candidates) {
		int TPandFP = clusterWithMaxInstanceCount.entrySet().stream().filter((e) -> relevantSet.contains(e.getKey())).mapToInt((e) -> e.getValue().getRows().size()).sum();
		
		int TP = clusterWithMaxInstanceCount.entrySet().stream().filter((e) -> relevantSet.contains(e.getKey())).mapToInt((e) -> countInstanceOccurence(e.getValue().getRows(), e.getKey())).sum();
		
		int maxCorrect = Ints.checkedCast(candidates.stream().filter((e) -> { 
			Optional<InstanceId> match = tableRepository.getMatchedInstanceId(e);
			if(match.isPresent() && relevantSet.contains(match.get()))
				return true;
			else
				return false;
		}).count());
		 
		return new Tri<>(TPandFP, TP, maxCorrect);
	}
	
	public Quad<Double,Integer, Double, Integer> getAveragePrecisionAndRecall(Set<InstanceId> relevantSet, Map<InstanceId, RowCluster> clusterWithMaxInstanceCount, Set<TableRowId> candidates) {
		MutableDouble precison  = new MutableDouble();
		MutableDouble recall  = new MutableDouble();
		MutableInt recallCount = new MutableInt();
		MutableInt precisionCount = new MutableInt();
		
		 Map<InstanceId, Long> maxCorrectMap = candidates.stream().map((e) -> { 
				Optional<InstanceId> match = tableRepository.getMatchedInstanceId(e);
				return match;
//				if(match.isPresent() && relevantSet.contains(match.get()))
//					return true;
//				else
//					return false;
			}).filter(Optional::isPresent).collect(Collectors.groupingBy(Optional::get, () -> new HashMap<>(), Collectors.counting()));
		 
		 
		for(InstanceId instanceId : relevantSet) {
			RowCluster cluster = clusterWithMaxInstanceCount.get(instanceId);
			double tp = 0.0;
			if(cluster != null) {
				precisionCount.increment();
				
				double tpAndFP = cluster.getRows().size();
				 tp = countInstanceOccurence(cluster.getRows(), instanceId);
				precison.add(tp / tpAndFP); 
			}
			if(!maxCorrectMap.containsKey(instanceId)) {
				int sdf = 123;
			}
			else {
				double maxCorrect = maxCorrectMap.get(instanceId);
				recall.add(tp / maxCorrect); 
				recallCount.increment();
			}
		}
		
	 
		 
		return new Quad<>(precison.doubleValue() , precisionCount.intValue(), recall.doubleValue(), recallCount.intValue());
	}
 
	
	
	public double getAveragePurity(ClusterList clusters) {
		int countAllRows = clusters.getClusterList().stream().mapToInt((e) -> e.getRows().size()).sum();
		
		double purity = clusters.getClusterList().stream().map((e) ->  {
			Map< InstanceId , Integer> map = rowCountMap(e.getRows());
			
			double count = e.getRows().size();
			int maxSize = 0;
			for(Entry<InstanceId, Integer> entry : map.entrySet()) {
				
				if(entry.getValue() > maxSize) {
					maxSize = entry.getValue();
				}
			}
			if(maxSize == 0)
				return Double.NaN;
			else
				return (maxSize / count) * (count / countAllRows);
			
		}).filter((e) -> !e.isNaN()).mapToDouble(Double::doubleValue).sum();
		return purity;
	}
	
	
	public Map<InstanceId, Integer> clusterPerInstanceMap(ClusterList clusters) {
		Map<InstanceId, Long> duplicateMap = clusters.getClusterList().stream().map(RowCluster::getRows).map(this::getMaxInstanceId).filter(Optional::isPresent).collect(
				Collectors.groupingBy(
						Optional::get, ()->new HashMap<>(), Collectors.counting())
				);
		 
		
		return duplicateMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().intValue()));
		
	}
	
	public Map<InstanceId, List<RowCluster>> clusterPerInstance(ClusterList clusters) {
		Map<InstanceId, List<RowCluster>> ret = new HashMap<>();
		
		for(RowCluster c : clusters.getClusterList()) {
			Optional<InstanceId> instanceId = this.getMaxInstanceId(c.getRows());
			
			if(instanceId.isPresent())
				CHMHelper.getFromMap(ret, instanceId.get(), () -> new ArrayList<>()).add(c);
		}
	 
		 
		
		return ret;
		
	}
	
	public Map<InstanceId,  RowCluster> clusterWithMaxInstanceCount(Map<InstanceId, List<RowCluster>> input) {
		Map<InstanceId, RowCluster> ret = new HashMap<>();
		
		for(Entry<InstanceId, List<RowCluster>> entry : input.entrySet()) {
			Integer maxCount = null;
			RowCluster maxCluster = null;
			for(RowCluster rc : entry.getValue()) {
				Integer count = countInstanceOccurence(rc.getRows(), entry.getKey());
				if(maxCount == null || count > maxCount) {
					maxCount = count;
					maxCluster = rc;
				}
			}
			ret.put(entry.getKey(), maxCluster);
		}
		
		return ret;
	}

	public double getAverage(List<Integer> list) {
		
		return list.stream().mapToDouble(Integer::doubleValue).average().getAsDouble();

	}

	public double getMedian(List<Integer> list) {
		DoubleStream sorted = list.stream().mapToDouble(Integer::doubleValue).sorted();
		double median = list.size()%2 == 0?
				sorted.skip(list.size()/2-1).limit(2).average().getAsDouble():        
					sorted.skip(list.size()/2).findFirst().getAsDouble();
				return median;
	}
	
//	public double getMedianFromUnsortedStream(List<Integer> sizes) {
//		
//	}
//	
//	
	public Map<InstanceId, List<Pair<Entity, Double>>> majorityMapping(List<Entity> entities) {
		
		
		Map<InstanceId, List<Pair<Entity, Double>>> ret = new HashMap<>();
		
		for(Entity e : entities) {
			Optional<Pair<InstanceId, Double>> instanceId = getMaxInstanceIdWithScore(e.getRowCluster().getRows());
			if(instanceId.isPresent()) {
				CHMHelper.getFromMap(ret, instanceId.get().getFirst(), () -> new ArrayList<>()).add(new Pair<>(e, instanceId.get().getSecond()));
			}
			
		}
		
		ret.forEach((e1, e2) -> {
			Collections.sort(e2, (a,b) -> {
				return Double.compare(b.getSecond(), a.getSecond());
			});
		});
		return ret; 
		
		
	 
		
	}
	
	public Map<InstanceId, Entity> maxMajorityMappings(List<Entity> entities) {
		
		
		Map<InstanceId, List<Pair<Entity, Integer>>> ret = new HashMap<>();
		
		for(Entity e : entities) {
			Optional<InstanceId> instanceId = getMaxInstanceId(e.getRowCluster().getRows());
			if(instanceId.isPresent()) {
				int count = countInstanceOccurence(e.getRowCluster().getRows(), instanceId.get());
				CHMHelper.getFromMap(ret, instanceId.get(), () -> new ArrayList<>()).add(new Pair<>(e, count));
			}
			
		}
		
		Map<InstanceId, Entity> ret2 = new HashMap<>();
		
		for(Entry<InstanceId, List<Pair<Entity, Integer>>> e : ret.entrySet()) {
			Entity maxEntity = null;
			Integer maxCount = null;
			for(Pair<Entity, Integer> e2 : e.getValue()) {
				if(maxCount == null || e2.getSecond() > maxCount) {
					maxEntity = e2.getFirst();
					maxCount = e2.getSecond();
				}
			}
			ret2.put(e.getKey(), maxEntity);
			
		}
 
		return ret2; 
		
		
	 
		
	}
	
	
	public Optional< Pair<InstanceId, Double> > getMaxInstanceIdWithScore(List<TableRowId> rows) {
		Map< InstanceId , Integer> map = rowCountMap(rows);
		
		double totalSize = rows.size();
		int maxSize = 0;
		InstanceId maxInstanceId = null;
		for(Entry<InstanceId, Integer> entry : map.entrySet()) {
			
			if(maxInstanceId == null || entry.getValue() > maxSize) {
				maxSize = entry.getValue();
				maxInstanceId = entry.getKey();
			}
		}
		double score = maxSize / totalSize;
		if(maxInstanceId == null || score <= 0.5)
			return Optional.empty();
		
		return Optional.of( new Pair<>( maxInstanceId, score));
	}
	public Integer countInstanceOccurence(List<TableRowId> rows, InstanceId instanceId) {
		return rowCountMap(rows).getOrDefault(instanceId, 0);
	}
	
	public Optional< InstanceId > getMaxInstanceId(List<TableRowId> rows) {
		Optional<Pair<InstanceId, Double>> withScore = getMaxInstanceIdWithScore(rows);
		
		 
		if(!withScore.isPresent())
			return Optional.empty();
		
		return Optional.of( withScore.get().getFirst());
	}
	
	public Map<InstanceId, Integer> rowCountMap(List<TableRowId> rows) {
		return rows.stream().map(tableRepository::getMatchedInstanceId).filter(Optional::isPresent).collect(
				Collectors.groupingBy(
						Optional::get, ()->new HashMap<>(), Collectors.counting())
				).entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) ->  Ints.checkedCast(e.getValue()) ));
	}

	public Quad<Double, Integer, Double, Integer> getPairWiseClusterComparison(Set<InstanceId> testingSet, ClusterList clusterList,
			Set<TableRowId> candidates) {

		Map<RowClusterUUID, Set<TableRowId>> clusters = new HashMap<>();
		Map<TableRowId, Integer> goldstandardCulsterMapping = new HashMap<>();
		Map<Integer, Set<TableRowId>> goldStandard = new HashMap<>();
		Map<InstanceId, Integer> goldStandardIndexMapping = new HashMap<>();
		
		int i = 0;
		for(InstanceId instanceId : testingSet) {
			goldStandardIndexMapping.put(instanceId, i);
			i++;
		}
		
		
		for(TableRowId candidate : candidates) {
			Optional<InstanceId> match = tableRepository.getMatchedInstanceId(candidate);
			if(match.isPresent() && goldStandardIndexMapping.containsKey(match.get())) {
				Integer gsIndex = goldStandardIndexMapping.get(match.get());
				CHMHelper.getFromMap(goldStandard, gsIndex, () -> new HashSet<>()).add(candidate);
				goldstandardCulsterMapping.put(candidate, gsIndex);
			}
		}
		
		
		for(RowCluster rc : clusterList.getClusterList()) {
			clusters.put(rc.getUuid(), new HashSet<>(rc.getRows()));
 
		}
		
		
	return getPairWiseClusterComparison(clusters,   goldStandard.values());		
	}

	public Quad<Double, Integer, Double, Integer> getPairWiseClusterComparison(Map<RowClusterUUID, Set<TableRowId>> clusters,
			  Collection<Set<TableRowId>> goldStandard ) {
		
		Set<TableRowId> relevantRows = new HashSet<>();
		for(Set<TableRowId> trSet : goldStandard) {
			relevantRows.addAll(trSet);
		}
		
		Map<TableRowId, RowClusterUUID> clustering = new HashMap<>();
		
		for(Entry<RowClusterUUID, Set<TableRowId>> rc : clusters.entrySet()) {
			 
			for(TableRowId rowId : rc.getValue()) {
				clustering.put(rowId, rc.getKey());
			}
		}
		
		AtomicLong maxRecallTotal = new AtomicLong();
		AtomicLong positivesTotal = new AtomicLong();
		AtomicLong truePositivesTotal = new AtomicLong();
		
		PersistentParallel6.runAndFinish(clusters.values(), (cluster) -> {
			
 
			int relevantPoints = 0;
			int irrelevantPoints = 0;
			for(TableRowId rowId : cluster) {
				if(relevantRows.contains(rowId)) {
					relevantPoints++;
				}
				else {
					irrelevantPoints++;
				}
			}
			
			long positives = relevantPoints <= 1 ? 0 :CombinatoricsUtils.binomialCoefficient(relevantPoints, 2);
			positives += relevantPoints * irrelevantPoints;
			
			positivesTotal.addAndGet(positives);
			
		}, this.getClass());
		
		
		PersistentParallel6.runAndFinish(goldStandard , (gsEntry) -> {
			long maxRecall = gsEntry.size() == 1 ? 0 : CombinatoricsUtils.binomialCoefficient(gsEntry.size(), 2);;
			long truePositives = 0;
					
			Set<RowClusterUUID> relevantClusters = new HashSet<>();
			for(TableRowId rowId : gsEntry ) {
				relevantClusters.add(clustering.get(rowId));
			}
			
			for(RowClusterUUID rc : relevantClusters) {
				Set<TableRowId> cluster = clusters.get(rc);
 if(cluster == null)
	 continue;
				long innerTruePositives = 0;
				int correctIncluster = 0; 
				for(TableRowId rowId : cluster) {
					if(gsEntry.contains(rowId)) {
						correctIncluster++;
					}
			 
				}

				innerTruePositives +=  correctIncluster == 1 ? 0 :CombinatoricsUtils.binomialCoefficient(correctIncluster, 2);


				truePositives += innerTruePositives;
				
			}
			
			maxRecallTotal.addAndGet(maxRecall);
			truePositivesTotal.addAndGet(truePositives);
			
			
			
		}, this.getClass());
		
		
		double doubleTruePositives = truePositivesTotal.doubleValue();
		return new Quad<Double, Integer, Double, Integer>(doubleTruePositives / positivesTotal.get(), 1, doubleTruePositives / maxRecallTotal.get(), 1);

	}

	public List<Integer> rawValueSizes(ClusterList clusters) {
		ConcurrentLinkedQueue<Integer> list = new ConcurrentLinkedQueue<>();
		PersistentParallel6.runAndFinish(clusters.getClusterList(), (e) -> {
				int count = 0;
				for(TableRowId rowId : e.getRows()) {
					AnnotatedTableRow row = tableRowCreator.getTableRow(rowId);
					int innerCount = row.getValues().entrySet().stream().mapToInt((f) -> f.getValue().getValues().size()).sum();
					count += innerCount;
				}
				list.add(count);
			}
		 , this.getClass());
		 
		return new ArrayList<>(list);
	}


}
