package expansion.set.profiling;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.InstanceIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.FusionPerformanceSynchronizedCreator;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.generic.GenericResultsWriterAccumRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumAverage;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumMedian;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AcumReference;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable.ImmutableAccumulatorListBuild;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnListBuilder;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult.ResultBuilder;

public class SetExpansionSuitabilityProfiler {

	private final AnnotatedTablesRepository tableRepository;
	private final OntologyRepository ontologyRepository;
	private final KnowledgeBase knowledgeBase;

	@Inject
	public SetExpansionSuitabilityProfiler(AnnotatedTablesRepository tableRepository, OntologyRepository ontologyRepository, KnowledgeBase knowledgeBase) {
		this.tableRepository= tableRepository;
		this.ontologyRepository = ontologyRepository;
		this.knowledgeBase = knowledgeBase;
	}

	public ProfilingResult measureAndStoreSuitability(Set<ClassId> classes, boolean withSubclasses, String outputFolder, String outputPrefix) {

		if(classes == null)
			classes = knowledgeBase.getClasses();
		
		ImmutableAccumulatorListBuild builder = AccumulatorListImmutable.builder();

		AcumReference<AccumInteger> TOTAL_MAPPINGS = builder.add(AccumInteger.empty());
		AcumReference<AccumInteger> TOTAL_INSTANCES = builder.add(AccumInteger.empty());
		

		AcumReference<AccumInteger> UNMAPPED_ROWS = builder.add(AccumInteger.empty());

		AcumReference<AccumAverage> AVERAGE_INSTANCES_PER_MAPPING = builder.add(AccumAverage.empty());
		AcumReference<AccumMedian> MEDIAN_INSTANCES_PER_MAPPING = builder.add(AccumMedian.empty());

		AcumReference<AccumAverage> AVERAGE_MAPPING_PER_INSTANCE = builder.add(AccumAverage.empty());
		AcumReference<AccumMedian> MEDIAN_MAPPING_PER_INSTANCE = builder.add(AccumMedian.empty());

		ColumnListBuilder columnListBuilder = ColumnList.builderAccumRepository();
		columnListBuilder.addColumn("TotalMappings", (repository) -> {
			return String.valueOf(repository.get(TOTAL_MAPPINGS).get());
		});
		columnListBuilder.addColumn("UNMAPPED_ROWS", (repository) -> {
			return String.valueOf(repository.get(TOTAL_INSTANCES).get());
		});
		columnListBuilder.addColumn("TotalInstances", (repository) -> {
			return String.valueOf(repository.get(TOTAL_INSTANCES).get());
		});
		columnListBuilder.addColumn("Average Instances Per Mapping", (repository) -> {
			return String.valueOf(repository.get(AVERAGE_INSTANCES_PER_MAPPING).get());
		});
		columnListBuilder.addColumn("Median Instances Per Mapping", (repository) -> {
			return String.valueOf(repository.get(MEDIAN_INSTANCES_PER_MAPPING).get());
		});

		columnListBuilder.addColumn("Average Mappings per Instance", (repository) -> {
			return String.valueOf(repository.get(AVERAGE_MAPPING_PER_INSTANCE).get());
		});
		columnListBuilder.addColumn("Median Mappings per Instance", (repository) -> {
			return String.valueOf(repository.get(MEDIAN_MAPPING_PER_INSTANCE).get());
		});

		ColumnList columnList = columnListBuilder.build();
		AccumulatorListImmutable resultsType = builder.getResultsType();

		FusionPerformanceSynchronizedCreator<GenericAccumaltableResult> creatorSingle = new FusionPerformanceSynchronizedCreator<>(
				GenericAccumaltableResult.empty(resultsType));

		PersistentParallel6.runAndFinish(classes, (classId) -> {
			try {

				Set<ClassId> allClasses = new HashSet<>();
				allClasses.add(classId);
				allClasses.addAll(ontologyRepository.getSubClasses(classId));

				Set<AnnotatedTableId> tables = new HashSet<>();
				for (ClassId subClassId : allClasses) {
					tables.addAll(tableRepository.getAllTableIds(subClassId));
				}
				ConcurrentHashMap<InstanceId, AtomicInteger> instanceCount = new ConcurrentHashMap<>();

				ConcurrentHashMap<PropertyId, ConcurrentHashMap<InstanceId, AtomicInteger>> propertyInstanceCount = new ConcurrentHashMap<>();
				ConcurrentHashMap<PropertyId, ConcurrentLinkedQueue<Integer>> propertyCount = new ConcurrentHashMap<>();
				ConcurrentHashMap<PropertyId, ConcurrentLinkedQueue<Integer>> propertyUnmappedCount = new ConcurrentHashMap<>();

				PersistentParallel6.runAndFinish(tables, (tid) -> {
					AnnotatedTableMetaData md = tableRepository.getMetaData(tid);
				 
					int instanceSize = (int) md.getInstancesScores().keySet().stream().map((e) -> e.getInstanceId()).filter((e) -> knowledgeBase.hasInstance(classId, e)).count();
					
					ResultBuilder valueBuilder = GenericAccumaltableResult.builder(resultsType);
					valueBuilder.setValue(TOTAL_MAPPINGS, new AccumInteger(1));

					valueBuilder.setValue(UNMAPPED_ROWS, new AccumInteger(md.getNumberOfRows()   - instanceSize));

					valueBuilder.setValue(AVERAGE_INSTANCES_PER_MAPPING, new AccumAverage((double) instanceSize));
					valueBuilder.setValue(MEDIAN_INSTANCES_PER_MAPPING, new AccumMedian((double) instanceSize));

					creatorSingle.addPerClassId(classId, valueBuilder.build());

					for (PropertyIndexComposite composite : md.getAllPropertyScores().keySet()) {
						PropertyId propertyId = composite.getPropertyId();
						CHMHelper.getFromMap(propertyCount, propertyId, () -> new ConcurrentLinkedQueue<>()).add(instanceSize);
						
						CHMHelper.getFromMap(propertyUnmappedCount, propertyId, () -> new ConcurrentLinkedQueue<>()).add(md.getNumberOfRows()   - instanceSize);
						 

					}

					for (InstanceIndexComposite composite : md.getInstancesScores().keySet()) {
						InstanceId instanceId = composite.getInstanceId();
						if(knowledgeBase.hasInstance(classId, instanceId)) {
							CHMHelper.getFromMap(instanceCount, instanceId, () -> new AtomicInteger()).incrementAndGet();
							for (PropertyIndexComposite propertyComposite : md.getAllPropertyScores().keySet()) {
								PropertyId propertyId = propertyComposite.getPropertyId();
								ConcurrentHashMap<InstanceId, AtomicInteger> innerMap = CHMHelper.getFromCHM(propertyInstanceCount, propertyId,
										() -> new ConcurrentHashMap<>());

								CHMHelper.getFromMap(innerMap, instanceId, () -> new AtomicInteger()).incrementAndGet();
							}
						}
					

					}

				}, this.getClass());

				for (Entry<InstanceId, AtomicInteger> entry : instanceCount.entrySet()) {

					int tablesSize = entry.getValue().intValue();
					ResultBuilder valueBuilder = GenericAccumaltableResult.builder(resultsType);
					valueBuilder.setValue(TOTAL_INSTANCES, new AccumInteger(1));

					valueBuilder.setValue(AVERAGE_MAPPING_PER_INSTANCE, new AccumAverage((double) tablesSize));
					valueBuilder.setValue(MEDIAN_MAPPING_PER_INSTANCE, new AccumMedian((double) tablesSize));

					creatorSingle.addPerClassId(classId, valueBuilder.build());
				}

				PersistentParallel6.runAndFinish(propertyCount.entrySet(), (propertyEntry) -> {
					try {

						PropertyId propertyId = propertyEntry.getKey();
						ClassPropertyId cpid = new ClassPropertyId(classId, propertyId);
						for (Integer instancesCountPerMapping : propertyEntry.getValue()) {
							ResultBuilder valueBuilder = GenericAccumaltableResult.builder(resultsType);
							valueBuilder.setValue(TOTAL_MAPPINGS, new AccumInteger(1));
							
						

							valueBuilder.setValue(AVERAGE_INSTANCES_PER_MAPPING, new AccumAverage((double) instancesCountPerMapping));
							valueBuilder.setValue(MEDIAN_INSTANCES_PER_MAPPING, new AccumMedian((double) instancesCountPerMapping));

							creatorSingle.addPerProperty(cpid, valueBuilder.build());
						}

						if(propertyUnmappedCount.containsKey(propertyEntry.getKey())) {
							for (Integer unmappedCount : propertyUnmappedCount.get(propertyEntry.getKey())) {
								ResultBuilder valueBuilder = GenericAccumaltableResult.builder(resultsType);
								valueBuilder.setValue(UNMAPPED_ROWS, new AccumInteger(unmappedCount));
								creatorSingle.addPerProperty(cpid, valueBuilder.build());
							}
						}
						
						for (Entry<InstanceId, AtomicInteger> instanceEntry : propertyInstanceCount.get(propertyId).entrySet()) {
							int propertySize = instanceEntry.getValue().intValue();
							ResultBuilder valueBuilder = GenericAccumaltableResult.builder(resultsType);
							valueBuilder.setValue(TOTAL_INSTANCES, new AccumInteger(1));

							valueBuilder.setValue(AVERAGE_MAPPING_PER_INSTANCE, new AccumAverage((double) propertySize));
							valueBuilder.setValue(MEDIAN_MAPPING_PER_INSTANCE, new AccumMedian((double) (propertySize)));

							creatorSingle.addPerProperty(cpid, valueBuilder.build());
						}

					} catch (Exception e) {
						e.printStackTrace();
						int sdf = 123;
					}

				}, this.getClass());

			} catch (Exception e) {
				e.printStackTrace();
				int sdf = 123;
			}

		}, this.getClass());

		GenericResultsWriterAccumRepository writer = new GenericResultsWriterAccumRepository(columnList);

		FusionPerformance<GenericAccumaltableResult> results = creatorSingle.get();
		writer.writeAccumRepository("SetExpansionProfiling", ontologyRepository,results );
		

		try {
			if (outputFolder != null && outputPrefix != null) {
				MKDIR.createDirUnchecked(new File(outputFolder));
				writer.writeToFolder(outputFolder, outputPrefix);

			}
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		
		return new ProfilingResult(TOTAL_MAPPINGS, TOTAL_INSTANCES, UNMAPPED_ROWS, AVERAGE_INSTANCES_PER_MAPPING, MEDIAN_INSTANCES_PER_MAPPING, AVERAGE_MAPPING_PER_INSTANCE, MEDIAN_MAPPING_PER_INSTANCE, results);
	}
	
	
	public static class ProfilingResult {
		public final AcumReference<AccumInteger> TOTAL_MAPPINGS;
		public final AcumReference<AccumInteger> TOTAL_INSTANCES;
		

		public final AcumReference<AccumInteger> UNMAPPED_ROWS;

		public final AcumReference<AccumAverage> AVERAGE_INSTANCES_PER_MAPPING;
		public final AcumReference<AccumMedian> MEDIAN_INSTANCES_PER_MAPPING;

		public final AcumReference<AccumAverage> AVERAGE_MAPPING_PER_INSTANCE;
		public final AcumReference<AccumMedian> MEDIAN_MAPPING_PER_INSTANCE;
		public final FusionPerformance<GenericAccumaltableResult> results;
		public ProfilingResult(AcumReference<AccumInteger> TOTAL_MAPPINGS, AcumReference<AccumInteger> TOTAL_INSTANCES,
				AcumReference<AccumInteger> UNMAPPED_ROWS, AcumReference<AccumAverage> AVERAGE_INSTANCES_PER_MAPPING,
				AcumReference<AccumMedian> MEDIAN_INSTANCES_PER_MAPPING, AcumReference<AccumAverage> AVERAGE_MAPPING_PER_INSTANCE,
				AcumReference<AccumMedian> MEDIAN_MAPPING_PER_INSTANCE, FusionPerformance<GenericAccumaltableResult> results) {
			super();
			this.TOTAL_MAPPINGS = TOTAL_MAPPINGS;
			this.TOTAL_INSTANCES = TOTAL_INSTANCES;
			this.UNMAPPED_ROWS = UNMAPPED_ROWS;
			this.AVERAGE_INSTANCES_PER_MAPPING = AVERAGE_INSTANCES_PER_MAPPING;
			this.MEDIAN_INSTANCES_PER_MAPPING = MEDIAN_INSTANCES_PER_MAPPING;
			this.AVERAGE_MAPPING_PER_INSTANCE = AVERAGE_MAPPING_PER_INSTANCE;
			this.MEDIAN_MAPPING_PER_INSTANCE = MEDIAN_MAPPING_PER_INSTANCE;
			this.results = results;
		}
		 
		
		
		
	}
	
}
