package expansion.set.profiling;

import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumAverage;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;

public class TableSetProfiler {
	private final AnnotatedTablesRepository tablesRepository;

	@Inject
	public TableSetProfiler(AnnotatedTablesRepository tablesRepository) {
		this.tablesRepository = tablesRepository;
	}

	public TableProfiling getStatistics(Set<AnnotatedTableId> set) {
		return new TableProfiling(set);
	}
	
	
	public   class TableProfiling {
		private final double averageColumns;
		private final double averageRows;
		private final double averageMappedColumns;
		private final double averageMappedInstances;

		private final int totalUnmappedRows;
		private final int totalUnmappedValues;
		
		
		private final int totalMappedColumns;
		private final int totalMappedInstances;
		
		private final int totalMappedValues;
		
		
		private final List<Integer> distributionOfMappedValuesPerInstance;
		

		private final List<Integer> distributionOfMappedRowsPerInstance;
		private final int totalUniqueInstancesMapped;
		
		public TableProfiling(Set<AnnotatedTableId> set) {
			
			ConcurrentAccumulator<AccumAverage> averageColumns = new ConcurrentAccumulator<>(AccumAverage.empty());
			ConcurrentAccumulator<AccumAverage> averageRows = new ConcurrentAccumulator<>(AccumAverage.empty());
			ConcurrentAccumulator<AccumAverage> averageMappedColumns = new ConcurrentAccumulator<>(AccumAverage.empty());
			ConcurrentAccumulator<AccumAverage> averageMappedInstances = new ConcurrentAccumulator<>(AccumAverage.empty());
			

			ConcurrentAccumulator<AccumInteger> totalUnmappedRows = new ConcurrentAccumulator<>(AccumInteger.empty());
			ConcurrentAccumulator<AccumInteger> totalUnmappedValues = new ConcurrentAccumulator<>(AccumInteger.empty());
			
			ConcurrentAccumulator<AccumInteger> totalMappedColumns = new ConcurrentAccumulator<>(AccumInteger.empty());
			ConcurrentAccumulator<AccumInteger> totalMappedInstances = new ConcurrentAccumulator<>(AccumInteger.empty());
			ConcurrentAccumulator<AccumInteger> totalMappedValues = new ConcurrentAccumulator<>(AccumInteger.empty());
			
			ConcurrentHashMap<InstanceId, AtomicInteger> instanceMap = new ConcurrentHashMap<>();
			

			ConcurrentHashMap<InstanceId, AtomicInteger> instanceRowMap = new ConcurrentHashMap<>();
			
			PersistentParallel6.runAndFinish(set, (tid) -> {
				AnnotatedTable data = tablesRepository.getTable(tid);
				AnnotatedTableMetaData md = tablesRepository.getMetaData(tid);
				averageColumns.add(new AccumAverage((double) data.getColumns().size()));
				averageRows.add(new AccumAverage((double) md.getNumberOfRows()));
				
				int mappedRowsSize = md.getInstances().size();
				int rowSize =  md.getNumberOfRows();
				int unamppedRowsSize = rowSize - mappedRowsSize;
				
				int propertySize = md.getProperties().size();
				
				totalUnmappedRows.add(new AccumInteger(unamppedRowsSize));
				
				totalUnmappedValues.add(new AccumInteger(propertySize * unamppedRowsSize));
				
//				int mappedRows = md.getInstances().entrySet().stream().filter((e) -> e.getValue().
				
				
				averageMappedColumns.add(new AccumAverage((double) propertySize));
				averageMappedInstances.add(new AccumAverage((double) mappedRowsSize));
				
				totalMappedColumns.add(new AccumInteger(propertySize));
				totalMappedInstances.add(new AccumInteger(mappedRowsSize));
				
				totalMappedValues.add(new AccumInteger(propertySize * mappedRowsSize));
				
				for(InstanceId instanceId : md.getInstances().values()) {
					CHMHelper.getFromCHM(instanceMap, instanceId, () -> new AtomicInteger()).addAndGet(propertySize);
				}
				
				for(InstanceId instanceId : md.getInstances().values()) {
					CHMHelper.getFromCHM(instanceRowMap, instanceId, () -> new AtomicInteger()).addAndGet(1);
				}
				
				
			}, this.getClass());
			this.averageColumns = averageColumns.get().get();
			this.averageRows = averageRows.get().get();
			this.averageMappedColumns = averageMappedColumns.get().get();
			this.averageMappedInstances = averageMappedInstances.get().get();
			this.totalUnmappedRows = totalUnmappedRows.get().get();
			this.totalUnmappedValues = totalUnmappedValues.get().get();

			this.totalMappedColumns = totalMappedColumns.get().get();
			this.totalMappedInstances = totalMappedInstances.get().get();
			this.totalMappedValues = totalMappedValues.get().get();
			
			this.totalUniqueInstancesMapped = instanceMap.size();
			
			for(Entry<InstanceId, AtomicInteger> instanceEntry : instanceMap.entrySet()) {
				if(instanceEntry.getValue().intValue() > 1000) {
					System.out.println(instanceEntry);
					int sdf = 123;
				}
			}
			
			distributionOfMappedValuesPerInstance = instanceMap.values().stream().map(AtomicInteger::intValue).collect(Collectors.toList());
			

			distributionOfMappedRowsPerInstance = instanceRowMap.values().stream().map(AtomicInteger::intValue).collect(Collectors.toList());
		}

		public double getAverageColumns() {
			return averageColumns;
		}

		public double getAverageRows() {
			return averageRows;
		}

		public double getAverageMappedColumns() {
			return averageMappedColumns;
		}

		public double getAverageMappedInstances() {
			return averageMappedInstances;
		}

		public int getTotalMappedColumns() {
			return totalMappedColumns;
		}

		public int getTotalMappedInstances() {
			return totalMappedInstances;
		}

		public int getTotalMappedValues() {
			return totalMappedValues;
		}

		public List<Integer> getDistributionOfMappedValuesPerInstance() {
			return distributionOfMappedValuesPerInstance;
		}

		public int getTotalUnmappedRows() {
			return totalUnmappedRows;
		}

		public int getTotalUnmappedValues() {
			return totalUnmappedValues;
		}

		public List<Integer> getDistributionOfMappedRowsPerInstance() {
			return distributionOfMappedRowsPerInstance;
		}

		public int getTotalUniqueInstancesMapped() {
			return totalUniqueInstancesMapped;
		}
		
		
		
		
	}
}
