package expansion.set.strategies.clustersplitters.attributes;

import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.blockers.Mappings;
import expansion.set.strategies.clustersplitters.attributes.AttributeValuesCreator.AttributeValueContainer;

public class AttributeBasedSplitter {


	@Inject
	protected AttributeScoreGetter attributeScoreGetter;

	private final SimilarityCalculator similarityCalculator;
	private final AttributeValuesCreator attributeValuesCreator;
	private final OntologyRepository ontologyRepository;

	public enum AttributeMethod {
		KBT_THRESHOLDED_AND_SCORED, DIRECT, KBT_SCORED, KBT_THRESHOLDED, KBT_POSITIVE_SCORED,
		DIRECT_WITH_NUMERIC_RANGE, DIRECT_WITH_SURFACE, DIRECT_FULL;
	}

	// SimpleLockableConcurrentCache<UUID, AttributeCachedSimilarities>

	// public RowClusterSimilarities cluster(RowClusteringConfig config,
	// ClustererInput input, List<RowCluster> rowCluster) {
	//
	// }

	private static SimpleLockableConcurrentCache<UUID, AttributeCachedSimilarities> cache = new SimpleLockableConcurrentCache<>();
	private static SimpleLockableConcurrentCache<UuidStringId, AttributeMappings> mappingsCache = new SimpleLockableConcurrentCache<>();

	@Inject
	public AttributeBasedSplitter(SimilarityCalculator similarityCalculator, AttributeValuesCreator attributeValuesCreator,
			OntologyRepository ontologyRepository) {
		super();
		this.similarityCalculator = similarityCalculator;
		this.attributeValuesCreator = attributeValuesCreator;
		this.ontologyRepository = ontologyRepository;
	}

	public RowClusterSimilarities get(RowClusteringConfig config, ClustererInput input,  AttributeMethod attributeMethod) {

		
		return new AttributeSimilarties(attributeMethod, getPrivate(config, input));
	}
	


	public AttributeCachedSimilarities getPrivate(RowClusteringConfig config, ClustererInput input) {

		return cache.getOrCreate(input.getRowValuesSets().getUUID(), () -> {
			return getPrivateInner(config, input);
		});
	}
	
	public AttributeMappings getMappings(RowClusteringConfig config, ClustererInput input ) {
		return mappingsCache.getOrCreate(input.getFullCandidatesAllClasses().getId(), () -> {
			AttributeValueContainer attributeValueContainer = attributeValuesCreator.get(config, input);
			
			
			Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> kbtValues = attributeValueContainer.getKbtValues(); 
			Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> directValues = attributeValueContainer.getDirectValues();

			ConcurrentHashMap<QualifiedPropertyId, Set<TableRowId>> kbtPropertyBlocking = new ConcurrentHashMap<>();
			ConcurrentHashMap<Pair<QualifiedPropertyId, AbstractObject>, Set<TableRowId>> kbtPropertyValueBlocking = new ConcurrentHashMap<>();
			
			ConcurrentHashMap<QualifiedPropertyId, Set<TableRowId>> directPropertyBlocking = new ConcurrentHashMap<>();
			ConcurrentHashMap<Pair<QualifiedPropertyId, AbstractObject>, Set<TableRowId>> directPropertyValueBlocking = new ConcurrentHashMap<>();

			
			PersistentParallel6.runAndFinish(input.getCandidates(), (tableRowIdPair) -> {

				if (kbtValues.containsKey(tableRowIdPair)) {

					for (Entry<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> entry : kbtValues.get(tableRowIdPair).entrySet()) {
											
						for(AbstractObject innerEntry : entry.getValue().values()) {
							CHMHelper.getFromCHM(kbtPropertyBlocking,entry.getKey(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowIdPair);
							CHMHelper.getFromCHM(kbtPropertyValueBlocking, new Pair<>(entry.getKey(), innerEntry), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowIdPair);

						}
					

					}
					
				 
				}

				if (directValues.containsKey(tableRowIdPair)) {

					for (Entry<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> entry : directValues.get(tableRowIdPair).entrySet()) {
						
						for(AbstractObject innerEntry : entry.getValue().values()) {
							CHMHelper.getFromCHM(directPropertyBlocking,entry.getKey(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowIdPair);
							CHMHelper.getFromCHM(directPropertyValueBlocking, new Pair<>(entry.getKey(), innerEntry), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowIdPair);
						}
					}
				}
				

			} , this.getClass());
			
			return new AttributeMappings(new Mappings(kbtPropertyBlocking.values()), new Mappings(kbtPropertyValueBlocking.values()), new Mappings(directPropertyBlocking.values()), new Mappings(directPropertyValueBlocking.values()));
		});
	}

	private AttributeCachedSimilarities getPrivateInner(RowClusteringConfig config, ClustererInput input) {
		AttributeValueContainer attributeValueContainer = attributeValuesCreator.get(config, input);
		
		
		// ConcurrentLinkedQueue<Pair<TableRowId, RowCluster>> allRowIds = new
		// ConcurrentLinkedQueue<>();

		// PersistentParallel6.runAndFinish(clusterList.getClusterList(),
		// (cluster) -> {
		//
		// for (TableRowId tableRow : cluster.getRows()) {
		// allRowIds.add(new Pair<>(tableRow, cluster));
		// }
		// } , this.getClass());

		Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> kbtValues = attributeValueContainer.getKbtValues();
		Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, Double>>> kbtScores = attributeValueContainer.getKbtScores();
		Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> directValues = attributeValueContainer.getDirectValues();

		ConcurrentHashMap<QualifiedPropertyId, Set<TableRowId>> kbtMappings = new ConcurrentHashMap<>();

		ConcurrentHashMap<QualifiedPropertyId, Set<TableRowId>> directMappings = new ConcurrentHashMap<>();
		
//		ConcurrentHashMap<Pair<QualifiedPropertyId, AbstractObject>, Set<TableRowId>> kbtPropertyBlocking = new ConcurrentHashMap<>();

 

		PersistentParallel6.runAndFinish(input.getFullCandidates().getRowIds(), (tableRowId) -> {

			if (kbtValues.containsKey(tableRowId)) {

				for (QualifiedPropertyId entry : kbtValues.get(tableRowId).keySet()) {
					CHMHelper.getFromCHM(kbtMappings, entry, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
					
					
//					CHMHelper.getFromCHM(kbtPropertyBlocking, new Pair<>(entry.getKey(), entry.getValue()), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowIdPair.getFirst());


				}
				
			 
			}

			if (directValues.containsKey(tableRowId)) {

				for (QualifiedPropertyId entry : directValues.get(tableRowId).keySet()) {
					CHMHelper.getFromCHM(directMappings, entry, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);

				}
			}

		} , this.getClass());

		return new AttributeCachedSimilarities(kbtMappings, kbtValues, kbtScores, directMappings, directValues, attributeValueContainer.getThresholdContainer(),
				config.isWithCache());
	}




 
	
	public static class AttributeScores {
		public final EnumMap<AttributeMethod, List<WeightedScore>> map;
		private final static List<WeightedScore> emptyList = Collections.unmodifiableList(new ArrayList<>());

//		public AttributeScores() {
//			map = new EnumMap<>(AttributeMethod.class);
//			for (AttributeMethod method : AttributeMethod.values()) {
//				map.put(method, 1.0);
//			}
//		}

		public AttributeScores(EnumMap<AttributeMethod, List<WeightedScore>> map) {
			this.map = map;
		}

		public List<WeightedScore> get(AttributeMethod attributeMethod) {
			return map.getOrDefault(attributeMethod, emptyList);
		}

	}

	public class AttributeCachedSimilarities {
		private final ConcurrentHashMap<TableRowId, ConcurrentHashMap<TableRowId, AttributeScores>> cache = new ConcurrentHashMap<>();

		// private final List<Pair<FusionStrategy, PropertiesConfig>>
		// fusionSetups;
		// private final List<ThresholdContainer> thresholdContainers;
		// private final List<ScoresContainer> scoresContainer;
		// private final List<Splitter> splitter;


		private final Map<QualifiedPropertyId, Set<TableRowId>> kbtMappings;
		private final Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId,AbstractObject>>> kbtValues;

		private final Map<QualifiedPropertyId, Set<TableRowId>> directMappings;
		private final Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId,AbstractObject>>> directValues;
		private final Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId,Double>>> kbtScores;

		private final ThresholdContainer thresholdContainer;
		private final boolean withCache;
		private AttributeCachedSimilarities(
				Map<QualifiedPropertyId, Set<TableRowId>> kbtMappings,
				Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> kbtValues,
				Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, Double>>> kbtScores, Map<QualifiedPropertyId, Set<TableRowId>> directMappings,
				Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> directValues, ThresholdContainer thresholdContainer,
				boolean withCache) {
			this.withCache = withCache;
			this.kbtMappings = kbtMappings;
			this.kbtValues = kbtValues;
			this.kbtScores = kbtScores;
			this.directMappings = directMappings;
			this.directValues = directValues;
			this.thresholdContainer = thresholdContainer;
		}


		public Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> getDirectValues() {
			return directValues;
		}

		public Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, Double>>> getScores() {
			return kbtScores;
		}

		public AttributeScores get(ClassId classId, TableRowIdPair pair) {
			TableRowId first = pair.getId1();
			TableRowId second = pair.getId2();

			int compare = TableRowId.compare(first, second);
			TableRowId firstKey;
			TableRowId secondKey;
			if (compare < 0) {
				firstKey = first;
				secondKey = second;
			} else {

				firstKey = second;
				secondKey = first;
			}

			if (withCache) {
				ConcurrentHashMap<TableRowId, AttributeScores> innerCache = CHMHelper.getFromCHM(cache, firstKey, () -> new ConcurrentHashMap<>());

				AttributeScores score = CHMHelper.getFromCHM(innerCache, secondKey, () -> getScoreInner(classId, pair));

				return score;
			} else {
				return getScoreInner(classId, pair);
			}

		}

		private AttributeScores getScoreInner(ClassId classId, TableRowIdPair tableIdPair) {
			
			
			
			// TableRowId first = tableIdPair.getId1();
			// TableRowId second = tableIdPair.getId2();
			//
			// AnnotatedTableMetaData mdFirst =
			// tableRepository.getMetaData(first.getTableId());
			// AnnotatedTableMetaData mdSecond =
			// tableRepository.getMetaData(second.getTableId());
			// //
			// // int firstIndex = doubleEntry.getFirst().getIndex();
			// // int secondIndex = doubleEntry.getSecond().getIndex();
			//
			// AnnotatedTableRow rowFirst = tableRowCreator.getTableRow(first);
			// AnnotatedTableRow rowSecond =
			// tableRowCreator.getTableRow(second);
			//
			//
			//
			// Map<QualifiedPropertyId, MatchedValues> firstValues = kbtValues.
			// Map<QualifiedPropertyId, MatchedValues> splittedValuesSecond =
			// split.apply(rowSecond.getValues());
			//
			//
			//
			// double correctKBT = 0;
			// int totalKBT = 0;
			//
			// double correctDirect = 0;
			// int totalDirect = 0;
			// for (Entry<QualifiedPropertyId, MatchedValues> firstEntry :
			// splittedValuesFirst.entrySet()) {
			// if (splittedValuesSecond.containsKey(firstEntry.getKey())) {
			// MatchedValues firstValue = firstEntry.getValue();
			// MatchedValues secondValue =
			// splittedValuesSecond.get(firstEntry.getKey());
			//
			// int scoresIndex = -1;
			// for (ListEntry<Pair<FusionStrategy, PropertiesConfig>>
			// setupListEntry : ListIterable
			// .iterable(fusionSetups)) {
			// if
			// (propertiesConfigFilter.include(setupListEntry.getValue().getSecond(),
			// firstEntry.getKey().getPropertyId())) {
			// scoresIndex = setupListEntry.getIndex();
			// }
			// }
			//
			// if (scoresIndex == -1) {
			// throw new IllegalStateException();
			// }
			//
			// double threshold =
			// this.thresholdContainers.get(scoresIndex).getScore(mdFirst.getDbPediaClass(),
			// firstEntry.getKey().getPropertyId(),
			// null);
			//
			// TripleId tripleId = new TripleId(mdFirst.getDbPediaClass(),
			// firstEntry.getKey().getPropertyId(), null);
			// AbstractObject firstAbstractObject = getValue(tripleId,
			// firstEntry.getKey().getQualifiers(), firstValue,
			// scoresContainer.get(scoresIndex),
			// threshold);
			// AbstractObject secondAbstractObject = getValue(tripleId,
			// firstEntry.getKey().getQualifiers(), secondValue,
			// scoresContainer.get(scoresIndex),
			// threshold);
			//
			// if (firstAbstractObject != null && secondAbstractObject != null)
			// {
			// totalKBT++;
			// if (similarityCalculator.isEqual(firstAbstractObject,
			// secondAbstractObject)) {
			// correctKBT++;
			// }
			// }
			//
			// firstAbstractObject = getValue(tripleId,
			// firstEntry.getKey().getQualifiers(), firstValue,
			// scoresContainerConstant, 0.0);
			// secondAbstractObject = getValue(tripleId,
			// firstEntry.getKey().getQualifiers(), secondValue,
			// scoresContainerConstant, 0.0);
			//
			// if (firstAbstractObject != null && secondAbstractObject != null)
			// {
			// totalDirect++;
			// if (similarityCalculator.isEqual(firstAbstractObject,
			// secondAbstractObject)) {
			// correctDirect++;
			// }
			// }
			//
			// }
			// }
			//
			// double direct = 0.0;
			// if (totalDirect == 0) {
			// direct = 0.0;
			// } else {
			// direct = correctDirect / totalDirect;
			// }
			// double kbt = 0.0;
			// if (totalKBT == 0) {
			// kbt = 0.0;
			// } else {
			// kbt = correctKBT / correctKBT;
			// }
 
			boolean includeAll = false;

			EnumMap<AttributeMethod, List<WeightedScore>> map = new EnumMap<>(AttributeMethod.class);

			if (includeAll) {
				List<WeightedScore> direct = attributeScoreGetter.getScores(directValues.get(tableIdPair.getId1()), directValues.get(tableIdPair.getId2()),
						false, false, classId);
				List<WeightedScore> directWithSurfaceForms = attributeScoreGetter.getScores(directValues.get(tableIdPair.getId1()),
						directValues.get(tableIdPair.getId2()), true,
						false, classId);

				List<WeightedScore> directWithNumericRange = attributeScoreGetter.getScores(directValues.get(tableIdPair.getId1()),
						directValues.get(tableIdPair.getId2()), false,
						true, classId);

				if (direct != null)
					map.put(AttributeMethod.DIRECT, direct);
				if (directWithSurfaceForms != null)
					map.put(AttributeMethod.DIRECT_WITH_SURFACE, directWithSurfaceForms);
				if (directWithNumericRange != null)
					map.put(AttributeMethod.DIRECT_WITH_NUMERIC_RANGE, directWithNumericRange);
			}
			List<WeightedScore> directFull = attributeScoreGetter.getScores(directValues.get(tableIdPair.getId1()), directValues.get(tableIdPair.getId2()),
					true, true, classId, kbtScores.get(tableIdPair.getId1()), kbtScores.get(tableIdPair.getId2()));

			if (includeAll) {
			List<WeightedScore> kbtNotThresholdedAndScored = getScores(directValues.get(tableIdPair.getId1()), directValues.get(tableIdPair.getId2()),
					kbtScores.get(tableIdPair.getId1()), kbtScores.get(tableIdPair.getId2()), true, false, null);
			List<WeightedScore> kbtThresholdedAndScored = getScores(kbtValues.get(tableIdPair.getId1()), kbtValues.get(tableIdPair.getId2()),
					kbtScores.get(tableIdPair.getId1()), kbtScores.get(tableIdPair.getId2()), true, false, null);
			List<WeightedScore> kbThresholdedUnScored = getScores(kbtValues.get(tableIdPair.getId1()), kbtValues.get(tableIdPair.getId2()),
					kbtScores.get(tableIdPair.getId1()), kbtScores.get(tableIdPair.getId2()), false, false, null);

			List<WeightedScore> kbtPositiveScored = getScores(directValues.get(tableIdPair.getId1()), directValues.get(tableIdPair.getId2()),
					kbtScores.get(tableIdPair.getId1()), kbtScores.get(tableIdPair.getId2()), true, true, classId);
			

			if(kbtNotThresholdedAndScored != null)
				map.put(AttributeMethod.KBT_SCORED, kbtNotThresholdedAndScored);
			if(kbtThresholdedAndScored != null)
				map.put(AttributeMethod.KBT_THRESHOLDED_AND_SCORED, kbtThresholdedAndScored);
			if(kbThresholdedUnScored != null)
				map.put(AttributeMethod.KBT_THRESHOLDED, kbThresholdedUnScored);
			
			if(kbtPositiveScored != null)
				map.put(AttributeMethod.KBT_POSITIVE_SCORED, kbtPositiveScored);
			
			}

			if (directFull != null)
				map.put(AttributeMethod.DIRECT_FULL, directFull);

			AttributeScores scores = new AttributeScores(map);



			return scores;
		}

		

		List<WeightedScore> getScores(Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> firstValues, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> secondValues, Map<QualifiedPropertyId, Map<TableColumnId, Double>> firstScores, Map<QualifiedPropertyId,Map<TableColumnId,  Double>> secondScores, boolean weighted, boolean positiveScore, ClassId classId) {

			if (firstValues == null || secondValues == null) {
				return null;
			}
			
			List<WeightedScore> weightedScore = new ArrayList<>();

			 
			for (Entry<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> firstEntry : firstValues.entrySet()) {
				
				
//				Double maxSim = null;
//				Double weight = null;
				for(Entry<TableColumnId, AbstractObject> innerFirstEntry : firstEntry.getValue().entrySet()) {
					if (secondValues.containsKey(firstEntry.getKey())) {
						

						for(Entry<TableColumnId, AbstractObject> innerSecondValue : secondValues.get(firstEntry.getKey()).entrySet()) {
					 
							double sim = similarityCalculator.getSimilarity(innerFirstEntry.getValue(), innerSecondValue.getValue());
							double thisWeight = 1.0;
							if(weighted) {
								if(positiveScore) {
									double threshold = thresholdContainer.getScore(classId, firstEntry.getKey().getPropertyId(), null);
									thisWeight = ( 1.0  + (firstScores.get(firstEntry.getKey()).get(innerFirstEntry.getKey()) - threshold )) * ( 1.0 + ( secondScores.get(firstEntry.getKey()).get(innerSecondValue.getKey()) - threshold ));
								}
								else {
									thisWeight = firstScores.get(firstEntry.getKey()).get(innerFirstEntry.getKey()) * secondScores.get(firstEntry.getKey()).get(innerSecondValue.getKey());
								}
							}
							
							if (similarityCalculator.isEqual(innerFirstEntry.getValue(), innerSecondValue.getValue())) {
								weightedScore.add(new WeightedScore(1.0, thisWeight)); 
								
							}
							else  {
								weightedScore.add(new WeightedScore(0.0, thisWeight)); 
							}
						}
						
					}
				}
				
				 
				
				
//				if (secondValues.containsKey(firstEntry.getKey())) {
//					total++;
//					if (similarityCalculator.isEqual(firstEntry.getValue(), secondValues.get(firstEntry.getKey()))) {
//						correct += 1.0 * firstScores.get(firstEntry.getKey()) * secondScores.get(firstEntry.getKey());
//					}
//					else {
//						correct -= 1.0 * firstScores.get(firstEntry.getKey()) * secondScores.get(firstEntry.getKey());
//					}
//				}
			}

			if(!weightedScore.isEmpty())
				return weightedScore;
			else
				return null;
		}

		public Set<TableRowId> query(AttributeMethod attributeMethod, RowClusterUUID uuid, TableRowId tableRowId) {
			
			if (directValues.containsKey(tableRowId))
				return query(directMappings, directValues.get(tableRowId).keySet(), tableRowId);
			else
				return new HashSet<>();

			// switch(attributeMethod) {
			// case DIRECT:
			// if (directValues.containsKey(tableRowId))
			// return query(directMappings.get(uuid),
			// directValues.get(tableRowId).keySet(), tableRowId);
			// else
			// return new HashSet<>();
			//
			// case KBT:
			// if (directValues.containsKey(tableRowId) &&
			// !kbtValues.containsKey(tableRowId)) {
			// int fsdf = 123;
			// }
			// if (kbtValues.containsKey(tableRowId))
			// return query(kbtMappings.get(uuid),
			// kbtValues.get(tableRowId).keySet(), tableRowId);
			// else
			// return new HashSet<>();
			//
			// }
			// throw new IllegalStateException();
		}

		private Set<TableRowId> query(Map<QualifiedPropertyId, Set<TableRowId>> mappings, Set<QualifiedPropertyId> qpids, TableRowId tableRowId) {

			Set<TableRowId> candidates = new HashSet<>();
			if (mappings != null && qpids != null) {
				for (QualifiedPropertyId qpid : qpids) {
					Set<TableRowId> mapping = mappings.get(qpid);
					if (mapping != null) {
						candidates.addAll(mapping);
					}

				}
			}
			candidates.remove(tableRowId);
			return candidates;
		}

	}

	public static class AttributeSimilarties implements RowClusterSimilarities {
		private final AttributeMethod attributeMethod;
		private final AttributeCachedSimilarities scores;

		private final Map<TableRowId, Set<Integer>> rowToBlocks;

		private AttributeSimilarties(AttributeMethod attributeMethod, AttributeCachedSimilarities scores) {
			super();
			this.attributeMethod = attributeMethod;
			this.scores = scores;

			ConcurrentHashMap<Pair<QualifiedPropertyId, AbstractObject>, Set<TableRowId>> preBlock = new ConcurrentHashMap<>();

			PersistentParallel6.runAndFinish(scores.getDirectValues().entrySet(), (e) -> {
				for (Entry<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> innerEntry : e.getValue().entrySet()) {

					for (Entry<TableColumnId, AbstractObject> innerInnerEntry : innerEntry.getValue().entrySet()) {
						CHMHelper.getFromCHM(preBlock, new Pair<>(innerEntry.getKey(), innerInnerEntry.getValue()),
								() -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e.getKey());
					}
				}
			}, this.getClass(), scores.getDirectValues().size() < 2000);

			ConcurrentHashMap<TableRowId, Set<Integer>> rowToBlocks = new ConcurrentHashMap<>();

			AtomicInteger ai = new AtomicInteger(0);

			PersistentParallel6.runAndFinish(preBlock.entrySet(), (e) -> {
				int blockKey = ai.getAndIncrement();

				for (TableRowId rowId : e.getValue()) {
					CHMHelper.getFromCHM(rowToBlocks, rowId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(blockKey);

				}
			}, this.getClass(), preBlock.size() < 400);

			this.rowToBlocks = rowToBlocks;
		}

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			TableRowId id1 = rows.get(key.getE1());
			TableRowId id2 = rows.get(key.getE2());
			return get(classId, id1, id2);
		}

		private double get(ClassId classId, TableRowId id1, TableRowId id2) {
			throw new UnsupportedOperationException(); 
//			if (id1.equals(id2)) {
//				return 1.0;
//			}
//			TableRowIdPair tableIdPair = new TableRowIdPair(id1, id2);
//			AttributeScores weights = scores.get(classId, tableIdPair);
//			if (weights == null)
//				return 0.0;
//			else {
//					return weights.get(attributeMethod);
//			}
		}
		
		// public List<Tri<Set<Integer>, Integer, Double>>
		// getSplitBlocksWithScores(TableRowId tableRowId) {
		// List<Tri<Set<Integer>, Integer, Double>> list = new ArrayList<>();
		// Set<Integer> blocks = getBlocks(tableRowId);
		//
		// if (blocks == null)
		// blocks = new HashSet<>();
		// // 0.70
		// list.add(new Tri<>(blocks, getBlockMin(tableRowId), 0.00));
		// return list;
		// }

		// @Override
		// public Set<Integer> getBlocks(TableRowId rowId) {
		// return rowToBlocks.get(rowId);
		// }
		//
		// @Override
		// public int getBlockMin(TableRowId tableRowId) {
		// return 1;
		// }
		//
		// @Override
		// public boolean supportsBlocking() {
		// return rowToBlocks != null;
		// }

		@Override
		public   List<WeightedScore> getList(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			TableRowId id1 = rows.get(key.getE1());
			TableRowId id2 = rows.get(key.getE2());
			return getList(classId, id1, id2);
			
		}
		
		private List<WeightedScore> getList(ClassId classId, TableRowId id1, TableRowId id2) {
//			List<WeightedScore> list = new ArrayList<>();
			if (id1.equals(id2)) {
				throw new IllegalStateException();
			}
			TableRowIdPair tableIdPair = new TableRowIdPair(id1, id2);
			AttributeScores weights = scores.get(classId, tableIdPair);

			if (id1.equals(new TableRowId("1346981172186_1346994902741_1908.arc7330478708359062489#28035184_0_7160201567823219801", 9))
					&& id2.equals(new TableRowId("1346876860614_1346890684800_1332.arc5282133196239291515#912032_0_1527817512187283215", 15))) {
				int sdf = 13;
			}

			if (id2.equals(new TableRowId("1346981172186_1346994902741_1908.arc7330478708359062489#28035184_0_7160201567823219801", 9))
					&& id1.equals(new TableRowId("1346876860614_1346890684800_1332.arc5282133196239291515#912032_0_1527817512187283215", 15))) {
				int sdf = 13;
			}

			if (weights == null)
				throw new IllegalStateException();
			else {
					return weights.get(attributeMethod);
			}
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {

			Set<TableRowId> candidates = scores.query(attributeMethod, cluster.getUuid(), point);
			
			Set<TableRowId> ret = new HashSet<>();
			Set<TableRowId> illegal = new HashSet<>();
			for(TableRowId candidate : candidates) {
				
				double score = get(cluster.getClassId(), point, candidate);
				if ((1 - score) < eps) {
					ret.add(candidate);
				}
 else {
					illegal.add(candidate);
				}
			}

			if (eps < 0.2) {
				return ret;
			} else {
				HashSet<TableRowId> legal = new HashSet<>(cluster.getRows());
				legal.removeAll(illegal);

				return legal;
			}
			// else {
			// return legal;
			// }
			

			
			
			
		}

		@Override
		public String getName() {
			return "ATTRIBUTE";
		}

	}
	
	public static class AttributeMappings {
		private final Mappings kbtProperty;
		private final Mappings kbtPropertyValue;
		private final Mappings directProperty;
		private final Mappings directPropertyValue;
		public AttributeMappings(Mappings kbtProperty, Mappings kbtPropertyValue, Mappings directProperty, Mappings directPropertyValue) {
			super();
			this.kbtProperty = kbtProperty;
			this.kbtPropertyValue = kbtPropertyValue;
			this.directProperty = directProperty;
			this.directPropertyValue = directPropertyValue;
		}
		public Mappings getKbtProperty() {
			return kbtProperty;
		}
		public Mappings getKbtPropertyValue() {
			return kbtPropertyValue;
		}
		public Mappings getDirectProperty() {
			return directProperty;
		}
		public Mappings getDirectPropertyValue() {
			return directPropertyValue;
		}
		
		
		
	}

}
