package expansion.set.strategies.clustersplitters.unaryattributes;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.core.fusing.EquivalenceClusterer;
import de.mannheim.uni.expansion.tablefusion.core.fusing.FuserRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;

public class UnaryAttributeAdapter {

	private final KnowledgeBase knowledgeBase;
	private final ThreadSafeKryo kryo;
	private final EquivalenceClusterer cluster;
	private final SimilarityCalculator similarityCalculator;
	private final FuserRepository fuserRepository;
	private final OntologyRepository ontologyRepository;
	private final TimestampQualifiersAdapter adapter;

	@Inject
	public UnaryAttributeAdapter(KnowledgeBase knowledgeBase, ThreadSafeKryo kryo, EquivalenceClusterer cluster, SimilarityCalculator similarityCalculator,
			FuserRepository fuserRepository, OntologyRepository ontologyRepository, TimestampQualifiersAdapter adapter) {
		super();
		this.knowledgeBase = knowledgeBase;
		this.kryo = kryo;
		this.cluster = cluster;
		this.similarityCalculator = similarityCalculator;
		this.fuserRepository = fuserRepository;
		this.ontologyRepository = ontologyRepository;
		this.adapter = adapter;
	}

	public UnaryAttributes getUnaryAttributes(String folder) throws IOException {
		Optional<UnaryAttributes> containerOptional = read(folder);
		if (containerOptional.isPresent()) {
			return containerOptional.get();
		} else {

			UnaryAttributes attributes = create();
			this.write(attributes, folder);

			return attributes;
		}
	}

	private UnaryAttributes create() {
		String target = "_TARGET";
		int targetLength = "_TARGET".length();
		Set<ClassId> classes = knowledgeBase.getClasses();
		classes = classes.stream().filter((e) -> {

			String classIdString = e.getClassId();
			int length = classIdString.length();
			if (length <= targetLength)
				return true;
			String subString = classIdString.substring(length - targetLength, length);

			return !subString.equals(target);
		}).collect(Collectors.toSet());
		List<UnaryAttribute> primary = new ArrayList<>();
		List<UnaryAttribute> secondary = new ArrayList<>();

		for (ClassId classId : classes) {
			for (PropertyId propertyId : knowledgeBase.getProperties(classId).keySet()) {
				if (propertyId.equals(PropertiesExpansionData.LABEL_PROPERTY)) {
					continue;
				}

				primary.add(new UnaryAttribute(classId, propertyId));


				for (PropertyId propertyId2 : knowledgeBase.getProperties(classId).keySet()) {
					if (propertyId2.equals(PropertiesExpansionData.LABEL_PROPERTY)) {
						continue;
					}
					if (!propertyId.equals(propertyId2)) {

						secondary.add(new UnaryAttribute(classId, propertyId, propertyId2));
					}
				}
			}

		}
		Set<UnaryAttribute> attributes = create(primary);
		Set<PropertyId> filterId = attributes.stream().map((u) -> u.getProperties()).flatMap((e) -> {
			return e.stream();
		}).collect(Collectors.toSet());

		secondary = secondary.stream().filter((p) -> {
			for (PropertyId propertyId : p.getProperties()) {
				if (filterId.contains(propertyId))
					return false;
			}
			return true;
		}).collect(Collectors.toList());

		Set<UnaryAttribute> secondaryAttributes = create(secondary);

		Set<UnaryAttribute> finalAttributes = new HashSet<>();
		finalAttributes.addAll(attributes);
		finalAttributes.addAll(secondaryAttributes);
		
		for (UnaryAttribute unaryAttribute : finalAttributes) {
			System.out.println(unaryAttribute.toString());
		}

		return new UnaryAttributes(new ArrayList<>(finalAttributes));
	}

	public Set<UnaryAttribute> create(List<UnaryAttribute> allClassPropertyIds) {
		Set<UnaryAttribute> attributes = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(allClassPropertyIds, (unaryAttribute) -> {

			List<PropertyId> propertyIds = new ArrayList<>(unaryAttribute.getProperties());
			List<Set<InstanceId>> instanceSets = new ArrayList<>();

			for (PropertyId propertyId : unaryAttribute.getProperties()) {
				instanceSets.add(knowledgeBase.getInstances(unaryAttribute.getClassId(), propertyId));
			}

			Set<InstanceId> startingSet = instanceSets.get(0);

			Set<InstanceId> instanceOverlap = new HashSet<>();

			for (InstanceId instanceId : startingSet) {
				boolean notFound = false;
				for (int i = 1; i < instanceSets.size(); i++) {
					if (!instanceSets.get(i).contains(instanceId)) {
						notFound = true;
						break;
					}
				}
				if (!notFound) {
					instanceOverlap.add(instanceId);
				}
			}

			List<Map<Integer, Pair<Set<AbstractQualifier>, AbstractObject>>> clusters = new ArrayList<>();
			// for (ListEntry<PropertyId> propertyEntry :
			// ListIterable.iterable(propertyIds)) {
			//
			// }
			

			int total = propertyIds.size();
			int i = 0;
			for (InstanceId instanceId : instanceOverlap) {
				if (i >= 4000)
					break;

				Map<Integer, List<Pair<Set<AbstractQualifier>, AbstractObject>>> valueList = new HashMap<>();

				for (ListEntry<PropertyId> propertyEntry : ListIterable.iterable(propertyIds)) {
					List<Pair<Set<AbstractQualifier>, AbstractObject>> valuesList = new ArrayList<>();
					ValueType valueType = ontologyRepository.getDataType(propertyEntry.getValue());
					// SimilarityMeasure measure =
					// similarityCalculator.getMeasure(valueType);
					// BaseFuser baseFuser =
					// fuserRepository.getBaseFuser(valueType);

					
					QualifiedValueList qvl = knowledgeBase.getQualifiedValueList(instanceId, propertyEntry.getValue());

					for (QualifiedValueHolder claim : qvl.getClaims()) {
						Set<AbstractQualifier> qualifiers = new HashSet<>(adapter.getQualifierFromQVH(claim));
						if (!qualifiers.isEmpty()) {
							for (AbstractQualifier aq : qualifiers) {
								Set<AbstractQualifier> q = new HashSet<>();
								q.add(aq);
								valuesList.add(new Pair<>(q, claim.getValue()));
							}
						} else {
							valuesList.add(new Pair<>(qualifiers, claim.getValue()));
						}
						 
					}

					if (valuesList.size() == 0) {
						Set<AbstractQualifier> qualifiers = new HashSet<>();
						valuesList.add(new Pair<>(qualifiers, null));
					}
					valueList.put(propertyEntry.getIndex(), valuesList);
					
//					Optional<List<AbstractObject>> values = knowledgeBase.getValuesAsList(instanceId, propertyEntry.getValue());
//					if (values.isPresent() && values.get().size() == 1) {
//						AbstractObject ao = values.get().get(0);
//						
//						valueMap.put(propertyEntry.getIndex(), ao);
//						
//					}
				}
				
				List<Map<Integer, Pair<Set<AbstractQualifier>, AbstractObject>>> totalValues = new ArrayList<>();
				int sizes[] = new int[valueList.size()];

				// int setSize = 1;
				for (Entry<Integer, List<Pair<Set<AbstractQualifier>, AbstractObject>>> entry : valueList.entrySet()) {
					// setSize *= entry.getValue().size();
					sizes[entry.getKey()] = entry.getValue().size();
				}
				
				int currentIndeces[] = new int[valueList.size()];
				
				// int s = 0;
				while (true) {

					Map<Integer, Pair<Set<AbstractQualifier>, AbstractObject>> valueMap = new HashMap<>();
					
					for (int cI = 0; cI < currentIndeces.length; cI++) {
						Pair<Set<AbstractQualifier>, AbstractObject> value = valueList.get(cI).get(currentIndeces[cI]);
					
						valueMap.put(cI, value);
					
					}

					totalValues.add(valueMap);

					boolean continueFor = incrementIndeces(currentIndeces, sizes, sizes.length - 1);

					if (!continueFor) {
						break;
					}
					// s++;
				}
				
				i += totalValues.size();

				for (Map<Integer, Pair<Set<AbstractQualifier>, AbstractObject>> value : totalValues) {
					boolean found = false;
					for (Map<Integer, Pair<Set<AbstractQualifier>, AbstractObject>> cluster : clusters) {
						int correct = 0;
						for (int p = 0; p < total; p++) {
							Pair<Set<AbstractQualifier>, AbstractObject> cc = cluster.get(p);
							if (value.get(p).getFirst().equals(cc.getFirst()) && similarityCalculator.isEqual(cc.getSecond(), value.get(p).getSecond())) {
								correct++;
							}
						}
						if (correct == total) {
							found = true;
							break;
						}
					}

					if (!found) {
						clusters.add(value);
					}
				}



				//
			}


			double percent = clusters.size() / ((double) i);
			if (percent > 0.999999999) {

				attributes.add(unaryAttribute);

			}

		} , this.getClass());

		return attributes;

	}

	public static boolean incrementIndeces(int[] currentIndeces, int[] sizes, int current) {
		currentIndeces[current]++;
		if (currentIndeces[current] == sizes[current]) {
			if (current == 0) {
				return false;
			} else {
				currentIndeces[current] = 0;
				return incrementIndeces(currentIndeces, sizes, current - 1);
			}
		}
		return true;

	}

	public void write(UnaryAttributes container, String file) throws IOException {
		this.kryo.write(container, file);
	}

	public Optional<UnaryAttributes> read(String file) throws IOException {
		Path path = Paths.get(file);
		if (Files.exists(path)) {
			Object properties = kryo.read(file);
			if (properties instanceof UnaryAttributes) {
				return Optional.of((UnaryAttributes) properties);
			}
		}

		return Optional.empty();

	}

	public static class UnaryAttributes {
		private final List<UnaryAttribute> attributes;

		private UnaryAttributes(List<UnaryAttribute> attributes) {
			super();
			this.attributes = attributes;
		}

		public List<UnaryAttribute> getAttributes() {
			return attributes;
		}

	}



}
