package expansion.set.strategies.implicit;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;

import javax.inject.Inject;
import javax.inject.Named;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.PropertyValuePair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import expansion.set.SetExpansionProperties;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.implicitAttributes.ImplicitAttributeFinder;
import expansion.set.implicitAttributes.ImplicitAttributes;
import expansion.set.strategies.clustersplitters.attributes.AttributeBasedSplitter;
import expansion.set.strategies.clustersplitters.attributes.AttributeBasedSplitter.AttributeCachedSimilarities;
import expansion.set.strategies.clustersplitters.attributes.AttributeScoreGetter;

public class ImplicitClustererSimilarityCreator {

	@Inject
	protected AttributeScoreGetter attributeScoreGetter;

	@Inject
	protected ImplicitAttributeFinder finder;

	@Inject
	protected SimilarityCalculator sc;

	@Inject
	@Named(SetExpansionProperties.tablesCachesFolder)
	protected String cachesFolder;

	@Inject
	protected CHMKryoAdapter kryoAdapter;

	@Inject
	public ImplicitClustererSimilarityCreator() {

	}

	@Inject
	protected AttributeBasedSplitter attributeSplitter;

	public ImplicitClustererSimilarities create(ClustererInput input, RowClusteringConfig config) {

		return new ImplicitClustererSimilarities(input.getAllParents(), input.getNonTailProperties(), attributeSplitter.getPrivate(config, input),
				input.getTableIds(), input.getClassId(), input.getUuid(), input.getFullCandidatesAllClasses(), config.isWithCache(), false);
	}

	public ImplicitClustererSimilarities createWithDirectCompared(ClustererInput input, RowClusteringConfig config) {

		return new ImplicitClustererSimilarities(input.getAllParents(), input.getNonTailProperties(), attributeSplitter.getPrivate(config, input),
				input.getTableIds(), input.getClassId(), input.getUuid(), input.getFullCandidatesAllClasses(), config.isWithCache(), true);
	}

	private final static SimpleLockableConcurrentCache<UuidStringId, Map<TableRowIdPair, List<WeightedScore>>> cache = new SimpleLockableConcurrentCache<>();

	private final static SimpleLockableConcurrentCache<UuidStringId, Map<AnnotatedTableId, ImplicitAttributes>> tableCache = new SimpleLockableConcurrentCache<>();

	public class ImplicitClustererSimilarities implements RowClusterSimilarities {

		private final Map<ClassId, Set<ClassId>> allParents;
		private final Map<ClassId, Set<PropertyId>> nonTailProperties;
		private final AttributeCachedSimilarities attributeCachedSimilarities;
		private final UUID clusterUUID;
		private final Map<AnnotatedTableId, ImplicitAttributes> tableScores;
		private final Map<TableRowIdPair, List<WeightedScore>> cachedScores;
		private final boolean directCompared;

		public ImplicitClustererSimilarities(Map<ClassId, Set<ClassId>> allParents, Map<ClassId, Set<PropertyId>> nonTailProperties,
				AttributeCachedSimilarities attributeCachedSimilarities, TableIdsSet tableIdsSet, ClassId classId, UUID uuid,
				TableRowIdsSet rowIdsSetFullClasses,
				boolean withCache, boolean directCompared) {
			this.allParents = allParents;
			this.nonTailProperties = nonTailProperties;
			this.attributeCachedSimilarities = attributeCachedSimilarities;
			this.clusterUUID = uuid;
			this.directCompared = directCompared;

			String cacheFolder = cachesFolder + "/ImplicitAttributes/";
			MKDIR.createDirUnchecked(new File(cacheFolder));

			// tableIdsSet.getId()
			tableScores = tableCache.getOrCreate(new UuidStringId("ALL"), () -> {

				Map<AnnotatedTableId, ImplicitAttributes> ret = new HashMap<>();

				Function<ClassId, Map<AnnotatedTableId, ImplicitAttributes>> perClassGetter = (oClassId) -> {
					String PREFIX = oClassId.getFinalPart() + "_";
					if (kryoAdapter.cacheExists(cacheFolder, PREFIX)) {
						return kryoAdapter.read(AnnotatedTableId.class, ImplicitAttributes.class, cacheFolder, PREFIX);
					} else {
						Map<AnnotatedTableId, ImplicitAttributes> tableScoresInner = new ConcurrentHashMap<>();
						Set<AnnotatedTableId> tableIds = tableIdsSet.getTableIds();

						PersistentParallel6.runAndFinish(tableIds, (tableId) -> {
							tableScoresInner.put(tableId, finder.find(tableId, allParents, nonTailProperties.get(classId), true));
						}, this.getClass());

						kryoAdapter.write(cacheFolder, PREFIX, tableScoresInner);
						return tableScoresInner;

					}

				};

				for (ClassId oClassId : allParents.keySet()) {
					ret.putAll(perClassGetter.apply(oClassId));
				}

				return ret;

			});

			if (withCache) {
				this.cachedScores = cache.getOrCreate(new UuidStringId("ALL"), () -> {
					ConcurrentHashMap<TableRowIdPair, List<WeightedScore>> map = new ConcurrentHashMap<>();
					;
					// long size =
					// DoubleIteraterable.getResultingSize(rowIdsSet.getRowIds().size());
					// PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(rowIdsSet.getRowIds()),
					// (pair) -> {
					// map.put(new TableRowIdPair(pair.getFirst(),
					// pair.getSecond()), get(pair.getFirst(), pair.getSecond(),
					// classId));
					// }, size, this.getClass());

					return map;
				});

			} else {
				this.cachedScores = null;
			}

		}

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			throw new UnsupportedOperationException();
		}

		public List<WeightedScore> getList(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			TableRowId rowId1 = rows.get(key.getE1());
			TableRowId rowId2 = rows.get(key.getE2());

			if (cachedScores != null) {
				TableRowIdPair rowIdPair = new TableRowIdPair(rowId1, rowId2);
				List<WeightedScore> ret = cachedScores.get(rowIdPair);
				if (ret == null) {
					ret = get(rowId1, rowId2, classId);
					cachedScores.put(rowIdPair, ret);
				}
				return ret;
			} else {
				return get(rowId1, rowId2, classId);
			}

		}

		public List<WeightedScore> get(TableRowId rowId1, TableRowId rowId2, ClassId classId) {

			if (rowId1.getTableId().equals(rowId2.getTableId()))
				return new ArrayList<>();

			Map<PropertyId, List<Pair<AbstractObject, Double>>> directValuesMap1 = getMapDirect(attributeCachedSimilarities.getDirectValues().get(rowId1),
					attributeCachedSimilarities.getScores().get(rowId1));
			Map<PropertyId, List<Pair<AbstractObject, Double>>> directValuesMap2 = getMapDirect(attributeCachedSimilarities.getDirectValues().get(rowId2),
					attributeCachedSimilarities.getScores().get(rowId2));

			ImplicitAttributes impl1 = tableScores.get(rowId1.getTableId());
			Map<PropertyId, List<Pair<AbstractObject, Double>>> implMap1 = getMapImpl(impl1.getImplicitAttributes());

			ImplicitAttributes impl2 = tableScores.get(rowId2.getTableId());
			Map<PropertyId, List<Pair<AbstractObject, Double>>> implMap2 = getMapImpl(impl2.getImplicitAttributes());

			if (true) {
				for (Entry<PropertyId, List<Pair<AbstractObject, Double>>> entry : implMap1.entrySet()) {
					if (implMap2.containsKey(entry.getKey()) || directValuesMap2.containsKey(entry.getKey())) {
						CHMHelper.getFromMap(directValuesMap1, entry.getKey(), () -> new ArrayList<>()).addAll(entry.getValue());
					}

				}

				for (Entry<PropertyId, List<Pair<AbstractObject, Double>>> entry : implMap2.entrySet()) {
					if (implMap1.containsKey(entry.getKey()) || directValuesMap1.containsKey(entry.getKey())) {
						CHMHelper.getFromMap(directValuesMap2, entry.getKey(), () -> new ArrayList<>()).addAll(entry.getValue());
					}

				}
			}

			Map<PropertyId, List<Pair<AbstractObject, Double>>> allMap2 = getMap(impl2.getImplicitAttributes(),
					attributeCachedSimilarities.getDirectValues().get(rowId2));

			List<TableRowIdPair> checkList = new ArrayList<>();

			checkList.add(new TableRowIdPair(new TableRowId("1346981172268_1346988871447_406.arc6119387513471376336#61205074_1_8717651146880790460", 12),
					new TableRowId("1346876860791_1346883222123_3489.arc4630073647238401030#12621564_1_6951226935139685323", 36)));

			checkList.add(new TableRowIdPair(new TableRowId("1346876860877_1346943073684_968.arc5922259307483110426#60114799_0_7760305389767059486", 23),
					new TableRowId("1346981172186_1346995588787_3253.arc906365098896261468#74748754_0_7781564339096028037", 22)));

			checkList.add(new TableRowIdPair(new TableRowId("1346876860611_1346927708927_1410.arc4002823674111578405#45998718_1_4411650333576785701", 1),
					new TableRowId("1346876860782_1346907601469_1653.arc8897894755961220520#18044892_0_1892400934803920236", 21)));

			checkList.add(new TableRowIdPair(new TableRowId("1346876860835_1346937147483_3336.arc1027594771575188191#87448106_0_6559554806554942204", 3),
					new TableRowId("1350433107059_1350464532277_262.arc3274150641837087721#45894422_0_4048022465851316720", 2)));

			checkList.add(new TableRowIdPair(new TableRowId("1346823846125_1346828766087_3133.arc8434669523913793224#16630613_1_442920039925902077", 19),
					new TableRowId("1346876860445_1346962354404_1053.arc3561445202962273515#50446515_26_5175382634929694534", 15)));

			checkList.add(new TableRowIdPair(new TableRowId("1350433107054_1350486965107_799.arc7111710114070000786#21907331_0_5050878971533094232", 11),
					new TableRowId("1346981172186_1346994903857_1044.arc4636266937237321018#46166619_0_591080969917940788", 28)));

			// checkList.add(new TableRowIdPair(new TableRowId("", ),new
			// TableRowId("", )));
			// checkList.add(new TableRowIdPair(new TableRowId("", ),new
			// TableRowId("", )));

			for (TableRowIdPair pair : checkList) {
				if ((pair.getId1().equals(rowId1) && pair.getId2().equals(rowId2)) || (pair.getId1().equals(rowId2) && pair.getId2().equals(rowId1))) {
					int sdf = 123;
				}
			}


			List<WeightedScore> list = new ArrayList<>();

			if (false) {
				List<WeightedScore> ll = attributeScoreGetter.getScores2(implMap1, directValuesMap2, true, true, classId);
				if (ll != null)
					list.addAll(ll);
				ll = attributeScoreGetter.getScores2(implMap1, implMap2, true, true, classId);
				if (ll != null)
					list.addAll(ll);
				ll = attributeScoreGetter.getScores2(directValuesMap1, implMap2, true, true, classId);
				if (ll != null)
					list.addAll(ll);

			} else {
				if (false && !directCompared) {
					LinkedHashMap<Pair<String, Boolean>, WeightedScore> weights = new LinkedHashMap<>();

					list = attributeScoreGetter.getScores2(implMap1, directValuesMap2, true, true, classId);

					if (list != null) {
						for (WeightedScore ws : list) {
							Pair<String, Boolean> pair = new Pair<>(ws.getSuffix().get(), ws.isMatch());
							if (weights.containsKey(pair)) {
								if (weights.get(pair).getWeight() < ws.getWeight()) {
									weights.put(pair, ws);
								}
							} else {
								weights.put(pair, ws);
							}
						}
					}

					list = attributeScoreGetter.getScores2(directValuesMap1, implMap2, true, true, classId);
					if (list != null) {
						for (WeightedScore ws : list) {
							Pair<String, Boolean> pair = new Pair<>(ws.getSuffix().get(), ws.isMatch());
							if (weights.containsKey(pair)) {
								if (weights.get(pair).getWeight() < ws.getWeight()) {
									weights.put(pair, ws);
								}
							} else {
								weights.put(pair, ws);
							}

						}
					}

					if (weights.isEmpty()) {
						list = null;
					} else {
						list = new ArrayList<>(weights.values());
					}

				} else {
					LinkedHashMap<Pair<String, Boolean>, WeightedScore> weights = new LinkedHashMap<>();

					list = attributeScoreGetter.getScores2(directValuesMap1, directValuesMap2, true, true, classId);
					if (false) {
						if (list != null) {
							for (WeightedScore ws : list) {
								Pair<String, Boolean> pair = new Pair<>(ws.getSuffix().get(), ws.isMatch());
								if (weights.containsKey(pair)) {
									if (weights.get(pair).getWeight() < ws.getWeight()) {
										weights.put(pair, ws);
									}
								} else {
								weights.put(pair, ws);
							}
						}
					}

						if (weights.isEmpty()) {
							list = null;
						} else {
							list = new ArrayList<>(weights.values());
						}
					}
				}
			}



			// double threshold = 0.7;
			// Set<PropertyId> propertyIds = new HashSet<>();
			// propertyIds.addAll(implMap1.keySet());
			// propertyIds.addAll(implMap2.keySet());
			// for (PropertyId propertyId : propertyIds) {
			//
			// if (implMap1.containsKey(propertyId) &&
			// allMap2.containsKey(propertyId)) {
			// for (Pair<AbstractObject, Double> first :
			// implMap1.get(propertyId)) {
			// if (first.getSecond() > threshold) {
			// for (Pair<AbstractObject, Double> second :
			// allMap2.get(propertyId)) {
			//
			// if (sc.isEqual(new ClassPropertyId(classId, propertyId),
			// first.getFirst(), second.getFirst())) {
			// list.add(new WeightedScore(1.0, Math.min(first.getSecond(),
			// second.getSecond())));
			// } else {
			// list.add(new WeightedScore(0.0, Math.min(first.getSecond(),
			// second.getSecond())));
			// }
			//
			// }
			// }
			//
			// }
			// }
			//
			// if (implMap2.containsKey(propertyId) &&
			// directValuesMap1.containsKey(propertyId)) {
			// for (Pair<AbstractObject, Double> first :
			// implMap2.get(propertyId)) {
			// if (first.getSecond() > threshold) {
			// for (Pair<AbstractObject, Double> second :
			// directValuesMap1.get(propertyId)) {
			//
			// if (sc.isEqual(new ClassPropertyId(classId, propertyId),
			// first.getFirst(), second.getFirst())) {
			// list.add(new WeightedScore(1.0, Math.min(first.getSecond(),
			// second.getSecond())));
			// } else {
			// list.add(new WeightedScore(0.0, Math.min(first.getSecond(),
			// second.getSecond())));
			// }
			//
			// }
			// }
			//
			// }
			// }
			//
			// }
			if (list == null)
				list = new ArrayList<>();
			return list;
		}

		private Map<PropertyId, List<Pair<AbstractObject, Double>>> getMapDirect(Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> map,
				Map<QualifiedPropertyId, Map<TableColumnId, Double>> map2) {
			Map<PropertyId, List<Pair<AbstractObject, Double>>> ret = new HashMap<>();
			if (map != null) {
				for (Entry<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> entry : map.entrySet()) {
					List<Pair<AbstractObject, Double>> innerMap = CHMHelper.getFromMap(ret, entry.getKey().getPropertyId(), () -> new ArrayList<>());

					for (Entry<TableColumnId, AbstractObject> innerEntry : entry.getValue().entrySet()) {
						double weight = 1.0;
						if (AttributeScoreGetter.SCORING_MATCHING_SCORES) {
							weight = map2.get(entry.getKey()).get(innerEntry.getKey());
						}
						innerMap.add(new Pair<>(innerEntry.getValue(), weight));
					}
				}
			}
			return ret;
		}
		private Map<PropertyId, List<Pair<AbstractObject, Double>>> getMapImpl(Map<PropertyValuePair, Double> implicitAttributes) {
			Map<PropertyId, List<Pair<AbstractObject, Double>>> ret = new HashMap<>();

			for (Entry<PropertyValuePair, Double> entry : implicitAttributes.entrySet()) {
				List<Pair<AbstractObject, Double>> innerMap = CHMHelper.getFromMap(ret, entry.getKey().getPropertyId(), () -> new ArrayList<>());

				innerMap.add(new Pair<>(entry.getKey().getAbstractObject(), entry.getValue()));
			}

			return ret;
		}

		private Map<PropertyId, List<Pair<AbstractObject, Double>>> getMap(Map<PropertyValuePair, Double> implicitAttributes,
				Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> map) {
			Map<PropertyId, List<Pair<AbstractObject, Double>>> ret = new HashMap<>();

			if (map != null) {
				for (Entry<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> entry : map.entrySet()) {
					List<Pair<AbstractObject, Double>> innerMap = CHMHelper.getFromMap(ret, entry.getKey().getPropertyId(), () -> new ArrayList<>());

					for (AbstractObject innerEntry : entry.getValue().values()) {
						innerMap.add(new Pair<>(innerEntry, 1.0));
					}
				}
			}

			for (Entry<PropertyValuePair, Double> entry : implicitAttributes.entrySet()) {
				List<Pair<AbstractObject, Double>> innerMap = CHMHelper.getFromMap(ret, entry.getKey().getPropertyId(), () -> new ArrayList<>());

				innerMap.add(new Pair<>(entry.getKey().getAbstractObject(), entry.getValue()));
			}

			return ret;
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return "IMPLICIT_ATTRIBUTES";
		}

	}
}
