package expansion.set.strategies.rowclusterers;

import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StoredField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.index.io.InMemoryIndex;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.informatik.dws.expansion.utils.MutableReference;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.strategies.rowclusterers.label.LabelRowIndexClusterMapCreator;

public class RowLabelMapper {

	private final SimilarityCalculator similarityCalculator;
	private final AnnotatedTablesRepository tablesRepository;

	@Inject
	public RowLabelMapper(SimilarityCalculator similarityCalculator,AnnotatedTablesRepository tablesRepository) {
		super();
		this.similarityCalculator = similarityCalculator;
		this.tablesRepository = tablesRepository;
	}

	// private static SimpleLockableConcurrentCache<Boolean, Map<ClassId,
	// Map<String, InstanceId>>> cache = new SimpleLockableConcurrentCache<>();
	//
	// public Map<ClassId, Map<String, InstanceId>> getMappingForAll() throws
	// IOException {
	// MutableReference<IOException> exception = new
	// MutableReference<IOException>(null);
	// Map<ClassId, Map<String, InstanceId>> ret =
	// cache.getOrCreate(Boolean.TRUE, () -> {
	// try {
	// Set<AnnotatedTableId> tableIds = tablesRepository.getAllTableIds();
	//
	// return getMapping(tableIds);
	// } catch (IOException e) {
	// exception.set(e);
	// }
	// return null;
	// });
	//
	// if (ret == null) {
	// throw exception.get();
	// }
	//
	// return ret;
	// }

	private final static SimpleLockableConcurrentCache<UuidStringId, Map<ClassId, Map<String, InstanceId>>> cacheUUID = new SimpleLockableConcurrentCache<>();

	public Map<ClassId, Map<String, InstanceId>> getMapping(TableIdsSet tables) throws IOException {
		MutableReference<IOException> exception = new MutableReference<IOException>(null);
		Map<ClassId, Map<String, InstanceId>> ret = cacheUUID.getOrCreate(tables.getId(), () -> {
			try {
				return getMapping(tables.getTableIds());
			} catch (IOException e) {
				exception.set(e);
			}
			return null;
		});

		if (ret == null) {
			throw exception.get();
		}

		return ret;
	}

	private Map<ClassId, Map<String, InstanceId>> getMapping(Set<AnnotatedTableId> tableIds) throws IOException {

		ConcurrentHashMap<ClassId, Set<String>> allString = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(tableIds, (tableId) -> {

			CHMHelper
					.getFromCHM(allString, tablesRepository.getMetaData(tableId).getDbPediaClass(),
							() -> Collections.newSetFromMap(new ConcurrentHashMap<>()))
					.addAll(this.tablesRepository.getLabels(tableId));


		}, this.getClass());

		Map<ClassId, Set<String>> stringLists = new HashMap<>(allString);

		Map<ClassId, Map<String, InstanceId>> mappings = new HashMap<>();
		int nameIndex = 0;
		for (Entry<ClassId, Set<String>> entry : stringLists.entrySet()) {
			mappings.put(entry.getKey(), getStringMapping(entry.getValue(), nameIndex));
			nameIndex++;
		}

		return mappings;
	}

	public Map<String, InstanceId> createClusters(Map<String, String> map, int nameIndex) {
		List<Set<String>> clusters = new ArrayList<>();
		HashMap<String, Integer> indexMap = new HashMap<>();

		for (Entry<String, String> entry : map.entrySet()) {
			createClusters(clusters, indexMap, map, new HashSet<>(), entry.getKey(), entry.getValue());
		}
		ConcurrentHashMap<String, InstanceId> ret = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(ListIterable.iterable(clusters), (entry) -> {
			InstanceId instanceId = InstanceId.simple("INSTANCE_" + nameIndex + "_" + entry.getIndex());
			for (String s : entry.getValue()) {
				ret.put(s, instanceId);
			}
		}, this.getClass());

		return ret;
	}

	public void createClusters(List<Set<String>> clusters, HashMap<String, Integer> indexMap, Map<String, String> map,
			Set<String> chain, String key, String value) {
		if (chain.contains(key) || value == null) {
			int index = clusters.size();
			Set<String> newSet = new HashSet<>();
			newSet.add(key);
			clusters.add(newSet);
			indexMap.put(key, index);
		} else if (indexMap.containsKey(value)) {
			Integer i = indexMap.get(value);
			clusters.get(i).add(key);
			indexMap.put(key, i);
		} else {
			chain.add(key);
			createClusters(clusters, indexMap, map, chain, value, map.get(value));
			Integer i = indexMap.get(value);
			clusters.get(i).add(key);
			indexMap.put(key, i);
		}
	}

	public Map<ClassId, Map<String, InstanceId>> getStringMaping(Map<ClassId, Set<String>> strings) throws IOException {
		Map<ClassId, Map<String, InstanceId>> ret = new HashMap<>();

		int nameIndex = 0;
		for (Entry<ClassId, Set<String>> entry : strings.entrySet()) {
			ret.put(entry.getKey(), getStringMapping(entry.getValue(), nameIndex));
			nameIndex++;
		}

		return ret;
	}

	private Map<String, InstanceId> getStringMapping(Set<String> stringSet, int nameIndex) throws IOException {
		List<String> stringList = new ArrayList<>(stringSet);
		LabelRowIndex index = new LabelRowIndex(stringList, 2);

		ConcurrentHashMap<String, String> map = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(stringList, (string) -> {
			StringObject so = new StringObject(string);
			Set<Integer> matches;
			try {
				matches = index.getMatches(string);

				double maxSim = 0.0;
				String maxMatch = null;
				for (Integer i : matches) {
					double sim = this.similarityCalculator.matchingStringSimilarity.get(so, new StringObject(stringList.get(i)));
					if (sim > this.similarityCalculator.matchingStringSimilarity.getEquivalenceThreshold() && (maxMatch == null || sim > maxSim)) {
						maxSim = sim;
						maxMatch = stringList.get(i);
					}
				}
				if (maxMatch == null) {
					map.put(string, string);
				} else {
					map.put(string, maxMatch);
				}

			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

		} , this.getClass());

		return createClusters(map, nameIndex);
	}
	
	public static class LabelRowIndex {

		public static final String INDEX = "index";
		public static final String LABEL = "label";

		private final InMemoryIndex idx = new InMemoryIndex();

		// private final QueryParser queryParser = new
		// QueryParser(Version.LUCENE_46, LABEL, new
		// StandardAnalyzer(Version.LUCENE_46));

		private final ThreadLocal<QueryParser> queryParserLocal = new ThreadLocal<>();
		private final int maxEditDistance;

		private QueryParser getQueryParser() {
			QueryParser qp = queryParserLocal.get();
			if (qp == null) {
				qp = new QueryParser( LABEL, new StandardAnalyzer());
				queryParserLocal.set(qp);
			}
			return qp;
		}

		public LabelRowIndex(List<String> strings, int maxEditDistance) throws IOException {
this.maxEditDistance = maxEditDistance;
			IndexWriter w = idx.getIndexWriter();

			PersistentParallel6.runAndFinish(ListIterable.iterable(strings), (e) -> {
				String label = e.getValue();
				String labelClean = StringNormalizer.normaliseValue(label, false);
				// normalises the value to improve lookup results
				label = LabelRowIndexClusterMapCreator.normalise(labelClean, true, true);

				Document doc = createDocument(label, e.getIndex());

				try {
					w.addDocument(doc);
				} catch (Exception e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}

			} , this.getClass());

			idx.closeIndexWriter();
		}

		Set<Integer> getMatches(String label) throws IOException, ParseException {
			try {
				String labelClean = StringNormalizer.normaliseValue(label, false);
				// normalises the value to improve lookup results
				label = LabelRowIndexClusterMapCreator.normalise(labelClean, true, true);
				label = QueryParser.escape(label);
				if(label.length() > 0 ) {
					
					 List<String> tokens = StringNormalizer.tokenise(label, false);
		                
		                StringBuilder sb = new StringBuilder();
		                
		                for(String token : tokens) {
		                    sb.append(token);
		                    
		                    if(maxEditDistance>0) {
		                        sb.append("~");
		                        sb.append(maxEditDistance);
		                    }
		                    sb.append(" ");
		                }
		                
		                
		                label = sb.toString();
					org.apache.lucene.search.Query q = getQueryParser().parse(label);
					IndexSearcher indexSearcher = this.idx.getIndexSearcher();
					// new TermQuery(new Term(LABEL, search));
					TopDocs results = indexSearcher.search(q, 30);
					Set<Integer> instanceIds = new HashSet<>();
					for (ScoreDoc doc : results.scoreDocs) {
						int index = fromDocument(indexSearcher.doc(doc.doc));

						instanceIds.add(index);
					}

					return instanceIds;
				}
				
			} catch (Exception e) {
				e.printStackTrace();
				System.out.println("CAUGHT EXCEPTION");
			}
			return new HashSet<>();

		}

		public int fromDocument(Document doc) {
			int index = doc.getField(INDEX).numericValue().intValue();
			return index;

		}

		public Document createDocument(String label, int index) {
			Document doc = new Document();

			doc.add(new StoredField(INDEX, index));
			doc.add(new TextField(LABEL, label, Field.Store.YES));

			return doc;
		}
	}

}
