package expansion.set.strategies.rowclusterers;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.uni_mannheim.informatik.dws.expansion.utils.MutableReference;

public class TableSetsCreator {

	private final AnnotatedTablesRepository tablesRepository;
	private final RowLabelMapper rowLabelMapper;

	@Inject
	public TableSetsCreator(AnnotatedTablesRepository tablesRepository, RowLabelMapper rowLabelMapper) {
		super();
		this.tablesRepository = tablesRepository;
		this.rowLabelMapper = rowLabelMapper;
	}

	// private static SimpleLockableConcurrentCache<Boolean, Map<ClassId,
	// TableSets>> cache = new SimpleLockableConcurrentCache<>();
	//
	// public Map<ClassId, TableSets> createForAll() throws IOException {
	// MutableReference<IOException> exception = new
	// MutableReference<IOException>(null);
	// Map<ClassId, TableSets> ret = cache.getOrCreate(Boolean.TRUE, () -> {
	// try {
	// Set<AnnotatedTableId> tableIds = tablesRepository.getAllTableIds();
	// Map<ClassId, Map<String, InstanceId>> stringMappings =
	// rowLabelMapper.getMappingForAll();
	// return create(tableIds, stringMappings);
	// } catch (IOException e) {
	// exception.set(e);
	// return null;
	// }
	// });
	//
	// if (ret == null) {
	// throw exception.get();
	// }
	// return ret;
	//
	// }

	private final static SimpleLockableConcurrentCache<UuidStringId, Map<ClassId, TableSets>> cacheUUID = new SimpleLockableConcurrentCache<>();

	public Map<ClassId, TableSets> create(TableIdsSet input) throws IOException {

		return getNoCache(input);

	}

	private Map<ClassId, TableSets> getNoCache(TableIdsSet input) throws IOException {
		Map<ClassId, Map<String, InstanceId>> stringMappings = rowLabelMapper.getMapping(input);
		return create(input.getTableIds(), stringMappings);
	}

	private Map<ClassId, TableSets> getWithCache(TableIdsSet input) throws IOException {
		MutableReference<IOException> exception = new MutableReference<IOException>(null);
		Map<ClassId, TableSets> ret = cacheUUID.getOrCreate(input.getId(), () -> {
			try {
				Map<ClassId, Map<String, InstanceId>> stringMappings = rowLabelMapper.getMapping(input);
				return create(input.getTableIds(), stringMappings);
			} catch (IOException e) {
				exception.set(e);
			}
			return null;
		});

		if (ret == null) {
			throw exception.get();
		}

		return ret;
	}

	public Map<ClassId, TableSets> create(Set<AnnotatedTableId> tableIds, Map<ClassId, Map<String, InstanceId>> stringMappings) {
		
		Map<ClassId, Set<AnnotatedTableId>> tableIdsMap = tableIds.parallelStream().collect(
				Collectors.groupingBy(
						(e) -> { return this.tablesRepository.getMetaData(e).getDbPediaClass();},
						Collectors.mapping(
								(e) -> {
									return e;
								},
								Collectors.toSet()
						)
				));
		
		AtomicInteger nameCounter = new AtomicInteger(0);
		Map<ClassId, TableSets> ret = new HashMap<>();
		for(Entry<ClassId, Set<AnnotatedTableId>> entry : tableIdsMap.entrySet()) {
			ret.put(entry.getKey(), createSet(entry.getValue(), stringMappings.get(entry.getKey()), nameCounter));
		}
		

		return ret;
		 

	}

	private TableSets createSet(Set<AnnotatedTableId> tableIds, Map<String, InstanceId> stringMapping, AtomicInteger nameCounter) {

		ConcurrentHashMap<Integer, Map<Set<InstanceId>, TableSetId>> hashCodeMapping = new ConcurrentHashMap<>();
		// ConcurrentHashMap<Pair<InstanceId, TableSetId>, AtomicInteger>
		// mappingCounter = new ConcurrentHashMap<Pair<InstanceId, TableSetId>,
		// AtomicInteger>();
		ConcurrentHashMap<TableSetId, AtomicInteger> setWeights = new ConcurrentHashMap<>();

		ConcurrentHashMap<InstanceId, Set<TableSetId>> clusterMapping = new ConcurrentHashMap<>();
		ConcurrentHashMap<AnnotatedTableId, TableSetId> tableMapping = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(tableIds, (tableId) -> {
			List<String> labels = this.tablesRepository.getLabels(tableId);
			Set<InstanceId> mappedLabels = labels.stream().map((e) -> stringMapping.get(e)).collect(Collectors.toSet());
			

			Integer hashCode = mappedLabels.hashCode();
			 Map<Set<InstanceId>, TableSetId> hashCodeInnerMap = CHMHelper.getFromCHM(hashCodeMapping, hashCode, () -> new HashMap<>());
			TableSetId tableSetId = null;
			boolean add = true;
			 synchronized(hashCodeInnerMap) {

				if (!hashCodeInnerMap.isEmpty()) {

					for (Entry<Set<InstanceId>, TableSetId> entry : hashCodeInnerMap.entrySet()) {
						if (mappedLabels.equals(entry.getKey())) {
							tableSetId = entry.getValue();
							add = false;
							break;
						}
					}
				}

				if (tableSetId == null) {
					tableSetId = new TableSetId("Set_" + nameCounter.getAndIncrement());
					hashCodeInnerMap.put(mappedLabels, tableSetId);
				}
			 }

			CHMHelper.getFromCHM(setWeights, tableSetId, () -> new AtomicInteger()).incrementAndGet();

			tableMapping.put(tableId, tableSetId);

			if (add) {
			for (InstanceId label : mappedLabels) {
					CHMHelper.getFromCHM(clusterMapping, label, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableSetId);
			}
			}
			
		}, this.getClass());


		ConcurrentHashMap<TableSetId, Set<InstanceId>> retClusters = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(clusterMapping.entrySet(), (e) -> {
			for (TableSetId setId : e.getValue()) {
				CHMHelper.getFromCHM(retClusters, setId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e.getKey());
			}
			// List<WeightedTableSetId> list =
			// CHMHelper.getFromCHM(clusterMapping, e.getKey().getFirst(), () ->
			// new ArrayList<>());
			// synchronized (list) {
			// list.add(new WeightedTableSetId(e.getKey().getSecond(),
			// e.getValue().get()));
			// }
		}, this.getClass());

		Map<TableSetId, Integer> setWeightsInteger = setWeights.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().get()));

		return new TableSets(retClusters, clusterMapping, setWeightsInteger, tableMapping);
	}

	private TableSets createSetNoCombination(Set<AnnotatedTableId> tableIds, Map<String, InstanceId> stringMapping) {
		AtomicInteger nameCounter = new AtomicInteger(0);

		// ConcurrentHashMap<Integer, Map<Set<InstanceId>, TableSetId>>
		// hashCodeMapping = new ConcurrentHashMap<>();
		// ConcurrentHashMap<Pair<InstanceId, TableSetId>, AtomicInteger>
		// mappingCounter = new ConcurrentHashMap<Pair<InstanceId, TableSetId>,
		// AtomicInteger>();

		ConcurrentHashMap<InstanceId, List<WeightedTableSetId>> clusterMapping = new ConcurrentHashMap<>();
		ConcurrentHashMap<TableSetId, Set<InstanceId>> retClusters = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(tableIds, (tableId) -> {
			List<String> labels = this.tablesRepository.getLabels(tableId);
			Set<InstanceId> mappedLabels = labels.stream().map((e) -> stringMapping.get(e)).collect(Collectors.toSet());

			// Integer hashCode = mappedLabels.hashCode();
			// Map<Set<InstanceId>, TableSetId> hashCodeInnerMap =
			// CHMHelper.getFromCHM(hashCodeMapping, hashCode, () -> new
			// HashMap<>());
			TableSetId tableSetId = new TableSetId("Set_" + nameCounter.getAndIncrement());

			retClusters.put(tableSetId, new HashSet<>(mappedLabels));

			for (InstanceId instanceId : mappedLabels) {

				List<WeightedTableSetId> list = CHMHelper.getFromCHM(clusterMapping, instanceId, () -> new ArrayList<>());
				synchronized (list) {
					list.add(new WeightedTableSetId(tableSetId, 1));

				}
			}

		} , this.getClass());



		return null;
	}

	public static class TableSetId implements Serializable {

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		private final String tableSetId;
		private transient Integer hashCode;

		public TableSetId(String tableSetId) {
			super();
			this.tableSetId = tableSetId;
		}

		@Override
		public int hashCode() {
			if (hashCode == null) {
				final int prime = 31;
				int result = 1;
				result = prime * result + ((tableSetId == null) ? 0 : tableSetId.hashCode());
				hashCode = result;
			}
			return hashCode;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			if (this.hashCode() != obj.hashCode())
				return false;
			TableSetId other = (TableSetId) obj;
			if (tableSetId == null) {
				if (other.tableSetId != null)
					return false;
			} else if (!tableSetId.equals(other.tableSetId))
				return false;
			return true;
		}

		public String getTableSetId() {
			return tableSetId;
		}

		@Override
		public String toString() {
			return tableSetId;
		}

		
	}
	public static class TableSets {
		private final Map<TableSetId, Set<InstanceId>> clusters;
		private final Map<InstanceId, Set<TableSetId>> clusterMappings;
		private final Map<TableSetId, Integer> setWeights;
		private final Map<AnnotatedTableId, TableSetId> tableMappings;

		public TableSets(Map<TableSetId, Set<InstanceId>> clusters, Map<InstanceId, Set<TableSetId>> clusterMappings, Map<TableSetId, Integer> clusterWeights,
				Map<AnnotatedTableId, TableSetId> tableMappings) {
			super();
			this.clusters = Collections.unmodifiableMap(clusters.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return Collections.unmodifiableSet(e.getValue());
			})));
			this.clusterMappings = Collections.unmodifiableMap(clusterMappings.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return Collections.unmodifiableSet(e.getValue());
			})));
			this.setWeights = Collections.unmodifiableMap(clusterWeights);
			this.tableMappings = Collections.unmodifiableMap(tableMappings);
		}


		public Map<TableSetId, Set<InstanceId>> getClusters() {
			return clusters;
		}

		public Map<InstanceId, Set<TableSetId>> getClusterMappings() {
			return clusterMappings;
		}

		public Map<TableSetId, Integer> getSetWeights() {
			return setWeights;
		}

		public Map<AnnotatedTableId, TableSetId> getTableMappings() {
			return tableMappings;
		}

	}

	public static class WeightedTableSetId {
		private final TableSetId tableSetId;
		private final int weight;

		public WeightedTableSetId(TableSetId tableSetId, int weight) {
			super();
			this.tableSetId = tableSetId;
			this.weight = weight;
		}


		public TableSetId getTableSetId() {
			return tableSetId;
		}

		public int getWeight() {
			return weight;
		}

	}
}
