package de.mannheim.uni.expansion.tablefusion.core.similarity.measures;

import java.util.Calendar;
import java.util.Date;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.similarity.AbstractSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityMeasures;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThreshold;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepository;
import de.mannheim.uni.expansion.utils.ThreadSafeLocalCalendar;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject.Precision;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class DateSimilarity extends AbstractSimilarity {

	private final ThreadSafeLocalCalendar threadSafeCalendar;

	private final NumericSimilarity numericSimilarity;

	private final SimilarityThreshold similarityThreshold;

	@Inject
	public DateSimilarity(SimilarityThresholdsRepository similarityThresholdsRepository, ThreadSafeLocalCalendar threadSafeCalendar,
			NumericSimilarity numericSimilarity) {
		super();
		this.similarityThreshold = similarityThresholdsRepository.get(SimilarityMeasures.DATE);
		this.numericSimilarity = numericSimilarity;
		this.threadSafeCalendar = threadSafeCalendar;

	}

	private boolean isYearDate(Date date) {
		Calendar calendar = threadSafeCalendar.getCalendar();
		calendar.setTime(date);

		int month = calendar.get(Calendar.MONTH);
		int day = calendar.get(Calendar.DAY_OF_MONTH);
		int hourOfDay = calendar.get(Calendar.HOUR_OF_DAY);
		int minute = calendar.get(Calendar.MINUTE);
		int second = calendar.get(Calendar.SECOND);
		if (month == 0 && day == 1 && hourOfDay == 0 && minute == 0 && second == 0) {
			return true;
		}

		return false;

	}

	private boolean isMonthDate(Date date) {
		Calendar calendar = threadSafeCalendar.getCalendar();
		calendar.setTime(date);

		int day = calendar.get(Calendar.DAY_OF_MONTH);
		int hourOfDay = calendar.get(Calendar.HOUR_OF_DAY);
		int minute = calendar.get(Calendar.MINUTE);
		int second = calendar.get(Calendar.SECOND);
		if (day == 1 && hourOfDay == 0 && minute == 0 && second == 0) {
			return true;
		}

		return false;

	}

	private boolean sameMonth(Date dateA, Date dateB) {
		Calendar calendar = threadSafeCalendar.getCalendar();
		calendar.setTime(dateA);

		int yearA = calendar.get(Calendar.YEAR);
		int monthA = calendar.get(Calendar.MONTH);

		calendar.setTime(dateB);

		int yearB = calendar.get(Calendar.YEAR);
		int monthB = calendar.get(Calendar.MONTH);

		if (yearA == yearB && monthA == monthB) {
			return true;
		}

		return false;
	}

	private boolean sameYear(Date dateA, Date dateB) {
		Calendar calendar = threadSafeCalendar.getCalendar();
		calendar.setTime(dateA);

		int yearA = calendar.get(Calendar.YEAR);

		calendar.setTime(dateB);

		int yearB = calendar.get(Calendar.YEAR);

		if (yearA == yearB) {
			return true;
		}

		return false;
	}

	@Override
	public double normalSimilarity(AbstractObject a, AbstractObject b) {

		if (a instanceof QuantityObject) {
			a = convertQuantityToTime((QuantityObject) a);
		} else if (b instanceof QuantityObject) {
			b = convertQuantityToTime((QuantityObject) b);
		}

		if (a == null || b == null) {
			return 0.0;
		} else if (!(a instanceof TimeObject) || !(b instanceof TimeObject)) {
			return 0.0;
		} else {
			TimeObject dateA = (TimeObject) a;
			TimeObject dateB = (TimeObject) b;

			if (dateA.isYearDate() || dateB.isYearDate()) {

				if (dateA.getYear() == dateB.getYear()) {
					return 1.0;
				} else if ((Math.abs(dateA.getYear() - dateB.getYear()) == 1)) {
					return 0.25;
				} else {
					return 0.0;
				}

			} else if (dateA.isMonthDate() || dateB.isMonthDate()) {
				if (dateA.getYear() == dateB.getYear() && dateA.getMonth() == dateB.getMonth()) {
					return 1.0;
				} else if ((Math.abs(dateA.getYear() - dateB.getYear()) == 1)) {
					return 0.25;
				} else {
					return 0.0;
				}
			} else {

				if (dateA.getYear() == dateB.getYear() && dateA.getMonth() == dateB.getMonth() && dateA.getDay() == dateB.getDay()) {
					return 1.0;
				} else if ((Math.abs(dateA.getYear() - dateB.getYear()) == 1)) {
					return 0.25;
				} else if (dateA.getYear() == dateB.getYear() && dateA.getMonth() == dateB.getDay() && dateA.getDay() == dateB.getMonth()) {
					return 0.75;
				} else {
					return 0.0;
				}

				//
				// long days1 = dateA.getDay();
				// long days2 = dateB.getDay();
				//
				// double v1 = (double) days1;
				// double v2 = (double) days2;
				//
				// return numericSimilarity.get(v1, v2);
			}

		}
	}

	private AbstractObject convertQuantityToTime(QuantityObject b) {
		if (b.isExactInt()) {
			if (String.valueOf(b.getExactInt()).length() == 4) {
				return new TimeObject(b.getExactInt(), 0, 0, 0, 0, 0, 0, Precision.T09_years);
			}
		} else {
			int intValue = new Double(b.getAmount()).intValue();
			if (String.valueOf(intValue).length() == 4) {
				return new TimeObject(intValue, 0, 0, 0, 0, 0, 0, Precision.T09_years);
			}
		}

		return b;
	}

	@Override
	public double getQuickSimilarity(AbstractObject a, AbstractObject b) {
		return normalSimilarity(a, b);
	}

	@Override
	protected ValueType getColumnDataType() {
		return ValueType.TIME;
	}

	@Override
	public double getEquivalenceThreshold() {
		return similarityThreshold.get();
	}

}
