package de.mannheim.uni.expansion.tablefusion.core.similarity.measures;

import java.util.HashMap;
import java.util.Optional;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.similarity.AbstractSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.numericRange.NumericRangeRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityMeasures;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThreshold;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepository;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class NumericSimilarity extends AbstractSimilarity {

	private final SimilarityThreshold similarityThreshold;
	private final NumericRangeRepository numericRangeRepository;

	@Inject
	public NumericSimilarity(SimilarityThresholdsRepository similarityThresholdsRepository, NumericRangeRepository numericRangeRepository) {

		this.similarityThreshold = similarityThresholdsRepository.get(SimilarityMeasures.NUMERIC);
		this.numericRangeRepository = numericRangeRepository;
	}

	public NumericSimilarity(SimilarityThreshold similarityThreshold) {

		this.similarityThreshold = similarityThreshold;
		this.numericRangeRepository = new NumericRangeRepository(new HashMap<>());
	}

	@Override
	public double normalSimilarity(AbstractObject a, AbstractObject b) {
		if (a == null || b == null) {
			return 0.0;
		} else if (!(a instanceof QuantityObject) || !(b instanceof QuantityObject)) {
			return 0.0;
		} else {
			double v1 = ((QuantityObject) a).getValue();
			double v2 = ((QuantityObject) b).getValue();

			return get(v1, v2);

		}
	}

	@Override
	public double normalSimilarity(ClassPropertyId classPropertyId, AbstractObject a, AbstractObject b) {
		Optional<Double> range = numericRangeRepository.getRange(classPropertyId);
		if (!range.isPresent()) {
			return normalSimilarity(a, b);
		} else {
			if (a == null || b == null) {
				return 0.0;
			} else if (!(a instanceof QuantityObject) || !(b instanceof QuantityObject)) {
				return 0.0;
			} else {
				double v1 = ((QuantityObject) a).getValue();
				double v2 = ((QuantityObject) b).getValue();

				return get(v1, v2, range.get());

			}
		}
	}

	public double getThreshold(ClassPropertyId classPropertyId, Double diff) {
		Optional<Double> range = numericRangeRepository.getRange(classPropertyId);
		if (!range.isPresent()) {
			throw new IllegalStateException();
		} else {
			double threshold = 1.0 - diff / range.get() - 0.0001;

			return threshold;
		}

	}

	private double cast(Object o) {
		if (o instanceof Double) {
			return (Double) o;
		} else if (o instanceof Integer) {
			return ((Integer) o).doubleValue();
		} else if (o instanceof String) {
			return Double.parseDouble((String) o);
		}
		throw new ClassCastException();
	}

	double getDiff(double v1, double v2) {
		double diff = 0.0;

		if (v1 < v2) {
			diff = v2 - v1;
		} else {
			diff = v1 - v2;
		}
		return diff;
	}

	double get(double v1, double v2, double range) {
		double diff = getDiff(v1, v2);

		double sim = 1.0 - diff / range;

		if (sim < 0.0) {
			sim = 0.0;
		} else if (sim > 1.0) {
			sim = 1.0;
		}

		return sim;
	}

	double get(double v1, double v2) {
		double diff = getDiff(v1, v2);

		double sim = 0.0;

		double abs1 = Math.abs(v1);
		double abs2 = Math.abs(v2);

		if (abs1 > abs2) {
			if (abs1 < 0.0001 && abs1 > -0.0001) {
				sim = 1.0;
			} else {
				sim = 1.0 - (diff / abs1);
			}
		} else {
			if (abs2 < 0.0001 && abs2 > -0.0001) {
				sim = 1.0;
			} else {
				sim = 1.0 - (diff / abs2);
			}
		}

		if (sim < 0.0) {
			sim = 0.0;
		}

		return sim;
	}

	@Override
	public double getQuickSimilarity(AbstractObject a, AbstractObject b) {
		return normalSimilarity(a, b);
	}

	@Override
	protected ValueType getColumnDataType() {
		return ValueType.QUANTITY;
	}

	@Override
	public double getEquivalenceThreshold() {
		return similarityThreshold.get();
	}
}
