package de.mannheim.uni.expansion.tablefusion.core.similarity.measures;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import com.wcohen.ss.Levenstein;

import de.mannheim.uni.expansion.tablefusion.core.similarity.AbstractSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityMeasures;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThreshold;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepository;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class StringSimilarityMixed extends AbstractSimilarity {

	private final SimilarityThreshold similarityThreshold;

	@Inject
	public StringSimilarityMixed(SimilarityThresholdsRepository similarityThresholdsRepository) {

		this.similarityThreshold = similarityThresholdsRepository.get(SimilarityMeasures.STRING_MIXED);
	}

	private final ThreadLocal<Levenstein> levenSteins = new ThreadLocal<>();

	private static final ConcurrentHashMap<StringPair, Double> levensteinCache = new ConcurrentHashMap<>();

	public double normalSimilarity(AbstractObject a, AbstractObject b) {

		if (a == null || b == null) {
			return 0.0;
		} else if (a.equals(b)) {
			return 1.0;
		} else if (!(a instanceof StringObject) || !(b instanceof StringObject)) {
			return 0.0;
		} else {
			String a2 = getCleanString(((StringObject) a).getString());
			String b2 = getCleanString(((StringObject) b).getString());

			if (a2.equals(b2)) {
				return 1.0;
			} else {
				// double fullSim = getLevenstein(a2, b2);
				// return getLevenstein((String) a, (String)b);

				List<String> first = Arrays.asList(a2.split(" ")).stream().map(String::trim).filter((e) -> {
					return e.length() > 0;
				}).collect(Collectors.toList());
				List<String> second = Arrays.asList(b2.split(" ")).stream().map(String::trim).filter((e) -> {
					return e.length() > 0;
				}).collect(Collectors.toList());

				double sim = getStringOverlap(first, second);
				// if (fullSim > sim)
				// sim = fullSim;
				return sim;
			}
		}

	}

	private double getLevenstein(String first, String second) {
		StringPair sp = new StringPair(first, second);
		Double sim = levensteinCache.get(sp);
		if (sim == null) {

			Levenstein l = levenSteins.get();
			if (l == null) {
				l = new Levenstein();
				levenSteins.set(l);
			}

			double score = l.score(first, second);
			score = score / Math.max(first.length(), second.length());
			if (score < 0) {
				score = score * -1;
			}
			score = 1 - score;

			sim = score;
			levensteinCache.put(sp, sim);
		}
		return sim;
	}

	public String getCleanString(String e) {

		e = e.replaceAll("\\([^\\(]*\\)", "");
		e = e.replaceAll("[^ \\pL\\d]+", " ");
		e = e.trim();
		e = e.toLowerCase();

		return e;
	}

	@Override
	public double getQuickSimilarity(AbstractObject a, AbstractObject b) {
		return normalSimilarity(a, b);
	}

	// public List<Object> tokenize(String string) {
	// string.split(" ");
	// }

	@Override
	protected ValueType getColumnDataType() {
		return ValueType.STRING;
	}

	private double getStringOverlap(List<String> aList, List<String> bList) {

		double intersection = 0;
		int largerSize = 0;
		int smallerSize = 0;

		if (aList.size() > bList.size()) {
			largerSize = aList.size();
		} else {
			largerSize = bList.size();
		}

		if (aList.size() < bList.size()) {
			smallerSize = aList.size();
		} else {
			smallerSize = bList.size();
		}

		Double[][] matrix = new Double[aList.size()][bList.size()];

		for (int a = 0; a < aList.size(); a++) {
			for (int b = 0; b < bList.size(); b++) {
				matrix[a][b] = getLevenstein(aList.get(a), bList.get(b));
			}
		}

		Set<Integer> aIndeces = new HashSet<>();
		for (int a = 0; a < aList.size(); a++) {
			aIndeces.add(a);
		}

		Set<Integer> bIndeces = new HashSet<>();
		for (int b = 0; b < bList.size(); b++) {
			bIndeces.add(b);
		}

		Double max = null;
		int aMax = -1;
		int bMax = -1;

		outerloop: while (true) {
			max = null;
			aMax = -1;
			bMax = -1;

			for (Iterator<Integer> aIt = aIndeces.iterator(); aIt.hasNext();) {
				int a = aIt.next();

				for (Iterator<Integer> bIt = bIndeces.iterator(); bIt.hasNext();) {
					int b = bIt.next();

					Double sim = matrix[a][b];
					if (sim > 0.999999) {
						intersection++;
						aIt.remove();
						bIt.remove();
						continue outerloop;
					} else if ((max == null || sim > max) && sim > 0.95) {
						max = sim;
						aMax = a;
						bMax = b;
					}

				}

			}

			if (max != null) {
				intersection += max;
				aIndeces.remove(aMax);
				bIndeces.remove(bMax);
			} else {
				break;
			}

		}

		double overlap = ((double) intersection) / ((double) smallerSize);
		double jaccard = ((double) intersection) / ((double) largerSize);

		double sim = (overlap + jaccard) / 2;
		return sim;
	}

	public static class StringPair {
		private final String a;
		private final String b;

		public StringPair(String A, String B) {
			int compare = A.compareTo(B);
			if (compare < 0) {
				a = A;
				b = B;
			} else {
				b = A;
				a = B;
			}
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((a == null) ? 0 : a.hashCode());
			result = prime * result + ((b == null) ? 0 : b.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			StringPair other = (StringPair) obj;
			if (a == null) {
				if (other.a != null)
					return false;
			} else if (!a.equals(other.a))
				return false;
			if (b == null) {
				if (other.b != null)
					return false;
			} else if (!b.equals(other.b))
				return false;
			return true;
		}

	}

	@Override
	public double getEquivalenceThreshold() {
		return similarityThreshold.get();
	}

}
