package de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching;

import java.util.HashSet;
import java.util.Set;

import javax.inject.Inject;

import com.google.common.base.Optional;

import de.mannheim.uni.expansion.tablefusion.core.similarity.AbstractSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityMeasures;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThreshold;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepository;
import de.mannheim.uni.similarity.functions.string.GeneralisedStringJaccard;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;

public class MatchingStringSimilarityWithSurfaceForms extends AbstractSimilarity {

	private final GeneralisedStringJaccard sim;

	private final SimilarityThreshold similarityThreshold;

	private final SurfaceFormsGetterContainer surfaceForms;

	public static boolean NO_SURFACE_FORMS = false;

	@Inject
	public MatchingStringSimilarityWithSurfaceForms(SimilarityThresholdsRepository similarityThresholdsRepository, SurfaceFormsGetterContainer surfaceForms) {
		sim = new GeneralisedStringJaccard(new LevelSteinSimilarityWithCache(), 0.5, 0.0);
		this.similarityThreshold = similarityThresholdsRepository.get(SimilarityMeasures.MATCHING_STRING);
		this.surfaceForms = surfaceForms;
	}

	// public MatchingStringSimilarityWithSurfaceForms(double outer, double
	// inner) {
	// sim = new GeneralisedStringJaccard(new LevelSteinSimilarityWithCache(),
	// inner, outer);
	// this.similarityThreshold = new SimilarityThresholdImpl(outer);
	// }

	@Override
	public double getQuickSimilarity(AbstractObject a, AbstractObject b) {
		return normalSimilarity(a, b);
	}

	@Override
	protected double normalSimilarity(AbstractObject a, AbstractObject b) {

		if (a == null || b == null) {
			return 0.0;
		} else if (a.equals(b)) {
			return 1.0;
		} else if (!(a instanceof StringObject) || !(b instanceof StringObject)) {
			return 0.0;
		} else {
			String a2 = ((StringObject) a).getString();
			String b2 = ((StringObject) b).getString();
			if (a2.equals(b2)) {
				return 1.0;
			}

			if (false && surfaceForms.candidatesIsPresent()) {

				boolean atLeastOneOverlap = false;
				Optional<Set<String>> candidatesAOptional = surfaceForms.getCandidatesSurfaceForms().getSurfaceForms(a2);
				if (candidatesAOptional.isPresent() && candidatesAOptional.get().contains(b2)) {
					atLeastOneOverlap = true;
				}

				if (!atLeastOneOverlap) {
					Optional<Set<String>> candidatesBOptional = surfaceForms.getCandidatesSurfaceForms().getSurfaceForms(b2);
					if (candidatesBOptional.isPresent() && candidatesBOptional.get().contains(a2)) {
						atLeastOneOverlap = true;
					}

				}

				if (!atLeastOneOverlap) {
					return 0.0;
				}
			}

			Set<String> setA;
			Optional<Set<String>> sfA = surfaceForms.get().getSurfaceForms(StringNormalizer.normaliseValue(a2, false));
			if (sfA.isPresent() && !NO_SURFACE_FORMS) {
				setA = sfA.get();
			} else {
				setA = new HashSet<>();
				setA.add(a2);
			}

			Set<String> setB;
			Optional<Set<String>> sfB = surfaceForms.get().getSurfaceForms(StringNormalizer.normaliseValue(b2, false));
			if (sfB.isPresent() && !NO_SURFACE_FORMS) {
				setB = sfB.get();
			} else {
				setB = new HashSet<>();
				setB.add(b2);
			}

			double max = 0.0;

			if (setA.size() < setB.size()) {

				double innerMax = compare(setA, b2);
				if (innerMax > max) {
					max = innerMax;
				}
				if (innerMax <= 0.99) {
					innerMax = compare(setB, a2);
					if (innerMax > max) {
						max = innerMax;
					}
				}
			} else {
				double innerMax = compare(setB, a2);
				if (innerMax > max) {
					max = innerMax;
				}
				if (innerMax <= 0.99) {
					innerMax = compare(setA, b2);
					if (innerMax > max) {
						max = innerMax;
					}
				}
			}
			return max;

		}

	}

	private boolean atLeastOneOverlap(Set<String> candidatesA, Set<String> candidatesB) {
		for (String sA : candidatesA) {
			if (candidatesB.contains(sA))
				return true;
		}
		return false;
	}

	double compare(Set<String> setA, Set<String> setB) {
		double max = 0.0;
		for (String aString : setA) {
			if (setB.contains(aString)) {
				max = 1.0;
				break;
			} else {
				for (String bString : setB) {
					double similarity = 0.0;

					similarity = sim.calculate(aString, bString);

					if (similarity > max) {
						max = similarity;

						if (similarity > 0.99) {
							break;
						}
					}
				}
			}
		}
		return max;
	}

	double compare(Set<String> setA, String b) {

		if (setA.contains(b)) {
			return 1.0;
		}

		double max = 0.0;
		for (String aString : setA) {
			double similarity = 0.0;

			similarity = sim.calculate(aString, b);

			if (similarity > max) {
				max = similarity;

				if (similarity > 0.99) {
					break;
				}
			}
		}
		return max;
	}

	@Override
	protected ValueType getColumnDataType() {
		return ValueType.STRING;
	}

	@Override
	public double getEquivalenceThreshold() {
		return similarityThreshold.get();
	}

	@FunctionalInterface
	public static interface SurfaceFormsGetter {

		Optional<Set<String>> getSurfaceForms(String label);
	}

	public static class SurfaceFormsGetterContainer {
		private SurfaceFormsGetter surfaceFormsGetter;
		private SurfaceFormsGetter candidatesSurfaceFormsgetter;

		public boolean isPresent() {
			if (this.surfaceFormsGetter == null)
				return false;
			else
				return true;
		}

		public boolean candidatesIsPresent() {
			if (this.candidatesSurfaceFormsgetter == null)
				return false;
			else
				return true;
		}

		public void setSurfaceFormsGetter(SurfaceFormsGetter surfaceFormsGetter) {
			this.surfaceFormsGetter = surfaceFormsGetter;
		}

		public SurfaceFormsGetter get() {
			return surfaceFormsGetter;
		}

		public void setCandidatesSurfaceFroms(SurfaceFormsGetter candidatesSurfaceFormsgetter) {
			this.candidatesSurfaceFormsgetter = candidatesSurfaceFormsgetter;

		}

		public SurfaceFormsGetter getCandidatesSurfaceForms() {
			return candidatesSurfaceFormsgetter;
		}

	}

}
