package de.mannheim.uni.expansion.tablefusion.core.similarity.numericRange;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableSet;
import java.util.Optional;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.QuartileData;
import de.mannheim.uni.expansion.utils.QuartileData.Quartile;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.CategoricalIntObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class NumericRangeRepository {

	private final Map<ClassPropertyId, Double> rangeMap;
	private final Map<ClassPropertyId, Double> minMap;
	private final Map<ClassPropertyId, Double> maxMap;
	private final Map<ClassPropertyId, Double> extendedMinMap;
	private final Map<ClassPropertyId, Double> extendedMaxMap;

	public NumericRangeRepository(Map<ClassPropertyId, Collection<AbstractObject>> values) {

		Map<ClassPropertyId, Double> rangeMap = new HashMap<>();
		Map<ClassPropertyId, Double> minMap = new HashMap<>();
		Map<ClassPropertyId, Double> maxMap = new HashMap<>();
		Map<ClassPropertyId, Double> extendedMinMap = new HashMap<>();
		Map<ClassPropertyId, Double> extendedMaxMap = new HashMap<>();
		for (Entry<ClassPropertyId, Collection<AbstractObject>> entry : values.entrySet()) {

			List<Double> sortedValues = entry.getValue().stream().map((e) -> {
				if (e.getType() == ValueType.QUANTITY) {
					return ((QuantityObject) e).getValue();
				} else if (e.getType() == ValueType.CATEGORICAL_INT) {
					return Double.valueOf(((CategoricalIntObject) e).getValue());
				} else {
					throw new IllegalStateException();
				}
			}).sorted().collect(Collectors.toList());

			List<Integer> list = new ArrayList<>();
			Map<Integer, Number> map = new HashMap<>();
			for (ListEntry<Double> le : ListIterable.iterable(sortedValues)) {
				list.add(le.getIndex());
				map.put(le.getIndex(), le.getValue());
			}

			QuartileData<Integer> qd = new QuartileData<>(list, map);
			Map<Quartile, NavigableSet<Integer>> perQuartile = qd.getSelectedPerQuartile();

			double q1Average = perQuartile.get(Quartile.Q1).stream().mapToDouble((e) -> map.get(e).doubleValue()).average().getAsDouble();
			double q4Average = perQuartile.get(Quartile.Q4).stream().mapToDouble((e) -> map.get(e).doubleValue()).average().getAsDouble();
			double range = q4Average - q1Average;
			rangeMap.put(entry.getKey(), range);

			// http://www.mathwords.com/o/outlier.htm
			double q1Value = qd.getQ1Value();
			double q3Value = qd.getQ3Value();
			double IQR = q3Value - q1Value;
			Double minValue = null;
			Double maxValue = null;
			for (AbstractObject ao : entry.getValue()) {
				Double value = null;
				if (ao.getType() == ValueType.QUANTITY) {
					value = ((QuantityObject) ao).getValue();
				} else if (ao.getType() == ValueType.CATEGORICAL_INT) {
					value = Double.valueOf(((CategoricalIntObject) ao).getValue());
				}

				if (value != null) {
					double min = value - 0.1;
					double max = value + 0.1;
					if (value > (q1Average - 1.5 * range) && (minValue == null || min < minValue)) {
						minValue = min;
					}
					if (value < (q4Average + 1.5 * range) && (maxValue == null || max > maxValue)) {
						maxValue = max;
					}
				}

			}

			minMap.put(entry.getKey(), minValue);
			maxMap.put(entry.getKey(), maxValue);
			extendedMinMap.put(entry.getKey(), q1Average - 3 * range);
			extendedMaxMap.put(entry.getKey(), q4Average + 3 * range);
		}
		this.rangeMap = Collections.unmodifiableMap(rangeMap);
		this.minMap = Collections.unmodifiableMap(minMap);
		this.maxMap = Collections.unmodifiableMap(maxMap);
		this.extendedMinMap = Collections.unmodifiableMap(extendedMinMap);
		this.extendedMaxMap = Collections.unmodifiableMap(extendedMaxMap);
	}

	public Optional<Double> getRange(ClassPropertyId classPropertyId) {
		Double range = rangeMap.get(classPropertyId);
		if (range != null) {
			return Optional.of(range);
		} else {
			return Optional.empty();
		}

	}

	public Optional<Pair<Double, Double>> getMinMax(ClassPropertyId cpid) {
		if (minMap.containsKey(cpid)) {
			return Optional.of(new Pair<>(minMap.get(cpid), maxMap.get(cpid)));
		}
		return Optional.empty();
	}

	public Optional<Pair<Double, Double>> getExtendedMinMax(ClassPropertyId cpid) {
		if (extendedMinMap.containsKey(cpid)) {
			return Optional.of(new Pair<>(extendedMinMap.get(cpid), extendedMaxMap.get(cpid)));
		}
		return Optional.empty();
	}
}
