package de.mannheim.uni.expansion.tablefusion.modules;

import java.io.IOException;
import java.net.JarURLConnection;
import java.util.HashMap;
import java.util.Map;

import javax.inject.Named;
import javax.inject.Singleton;

import dagger.Module;
import dagger.Provides;
import de.mannheim.uni.expansion.columnmatching.CMProperties;
import de.mannheim.uni.expansion.tablefusion.adapters.columnmatches.ColumnMatchesReader;
import de.mannheim.uni.expansion.tablefusion.adapters.columnmatches.ColumnMatchesRepository;
import de.mannheim.uni.expansion.tablefusion.adapters.columnmatches.ColumnMatchesWriter;
import de.mannheim.uni.matching.dbpedia.Preprocessing;
import de.mannheim.uni.matching.dbpedia.algorithm.CandidateSelectionComponent;
import de.mannheim.uni.matching.dbpedia.algorithm.WebTableToDBpediaComponentProcess;
import de.mannheim.uni.matching.dbpedia.algorithm.WebtableToDBpediaMatchingProcess;
import de.mannheim.uni.matching.dbpedia.model.MatchingParameters;
import de.mannheim.uni.matching.process.Configuration;
import de.mannheim.uni.similarity.functions.QuadraticSimilarityDecorator;
import de.mannheim.uni.similarity.functions.date.WeightedDatePartSimilarity;
import de.mannheim.uni.similarity.functions.numeric.DeviationSimilarity;
import de.mannheim.uni.similarity.functions.numeric.NormalisedNumericSimilarity;
import de.mannheim.uni.similarity.functions.string.GeneralisedStringJaccard;
import de.mannheim.uni.similarity.functions.string.JaccardOnNGramsSimilarity;
import de.mannheim.uni.similarity.functions.string.LevenshteinSimilarity;
import de.mannheim.uni.utils.timer.Timer;
import de.uni_mannheim.informatik.dws.matching.TableCacheSetter;

@Module(injects = { ColumnMatchesWriter.class }, library = true, complete = false)
public class MatchingModule {

	private final boolean useUnits;

	private final String columnMappingsFolder;
	private final String candidateSelectionConfig;
	private final String matchingConfig;
	private final String emptyFile;

	private final String surfaceForms;
	private final String redirects;
	// private final String equivalentProperties;



	public MatchingModule(boolean useUnits, String columnMappingsFolder, String candidateSelectionConfig, String matchingConfig,
 String emptyFile) {
		super();
		this.useUnits = useUnits;
		this.columnMappingsFolder = columnMappingsFolder;
		this.candidateSelectionConfig = candidateSelectionConfig;
		this.matchingConfig = matchingConfig;
		this.emptyFile = emptyFile;
		this.surfaceForms = emptyFile;
		this.redirects = emptyFile;
		TableCacheSetter.setCaches();
	}

	@Provides
	@Singleton
	MatchingParameters loadMatchingParameters(Configuration conf) {
		MatchingParameters mp = new MatchingParameters();
		mp.setUseUnitDetection(useUnits);
		mp.setMaxSpanningCells((Integer) conf.getValue(WebTableToDBpediaComponentProcess.PAR_SPANNING_CELL_THRESHOLD));
		mp.setTableType((de.mannheim.uni.matching.dbpedia.algorithm.WebtableToDBpediaMatchingProcess.tableType) conf
				.getValue(WebTableToDBpediaComponentProcess.PAR_TABLE_TYPE));
		// mp.setFastJoinDelta(1); // = edit distance threshold
		// mp.setFastJoinTau(1); // = jaccard threshold
		// mp.setCollectMatchingInfo(true);
		return mp;
	}

	@Provides
	@Named(CMProperties.COLUMN_MATCHES_FOLDER)
	String getColumnMatchesFolder() {
		return this.columnMappingsFolder;
	}

	@Provides
	@Singleton
	ColumnMatchesRepository getColumnMatchesRepository(ColumnMatchesReader reader) {
		return reader.read();
	}

	@Provides
	@Singleton
	// @Named(CMProperties.CANDIDATE_SELECTION_CONFIG)
	Configuration getCandidateSelectionConfig() {

		try {
			Preprocessing.loadSurfaceForms(emptyFile);

			Preprocessing.loadRedirects(emptyFile);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		Map<String, Object> aliases = new HashMap<>();

		aliases.put("QuadraticDeviation", new QuadraticSimilarityDecorator<Double>(new DeviationSimilarity()));
		aliases.put("QuadraticNormalized", new QuadraticSimilarityDecorator<Double>(new NormalisedNumericSimilarity()));
		aliases.put("GeneralizedLevenshtein", new GeneralisedStringJaccard(new LevenshteinSimilarity(), 0.5, 0.5));
		aliases.put("Jaccard2Grams", new JaccardOnNGramsSimilarity(2));
		aliases.put("Jaccard3Grams", new JaccardOnNGramsSimilarity(3));
		aliases.put("Jaccard4Grams", new JaccardOnNGramsSimilarity(4));
		aliases.put("Jaccard5Grams", new JaccardOnNGramsSimilarity(5));
		aliases.put("WeightedDate", new WeightedDatePartSimilarity(1, 3, 5));
		aliases.put("lodtable", WebtableToDBpediaMatchingProcess.tableType.lodtable);
		aliases.put("webtable", WebtableToDBpediaMatchingProcess.tableType.webtable);
		aliases.put("jsonWebTable", WebtableToDBpediaMatchingProcess.tableType.jsonWebTable);

		Configuration conf = null;

		try {

			conf = Configuration.readConfiguration(candidateSelectionConfig, CandidateSelectionComponent.getParams(), aliases);
			return conf;

		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | IOException e1) {

			e1.printStackTrace();
			return null;

		}
	}

	// @Provides
	// @Singleton
	// @Named(CMProperties.MATCHING_CONFIG)
	// Configuration getMatchingConfig() {
	//
	// WebtableToDBpediaMatchingProcess m = new
	// WebtableToDBpediaMatchingProcess();
	//
	// Map<String, Object> aliases = new HashMap<>();
	// aliases.put("QuadraticDeviation", new
	// QuadraticSimilarityDecorator<Double>(new DeviationSimilarity()));
	// aliases.put("QuadraticNormalized", new
	// QuadraticSimilarityDecorator<Double>(new NormalisedNumericSimilarity()));
	// aliases.put("GeneralizedLevenshtein", new GeneralisedStringJaccard(new
	// LevenshteinSimilarity(), 0.5, 0.5));
	// aliases.put("Jaccard2Grams", new JaccardOnNGramsSimilarity(2));
	// aliases.put("Jaccard3Grams", new JaccardOnNGramsSimilarity(3));
	// aliases.put("Jaccard4Grams", new JaccardOnNGramsSimilarity(4));
	// aliases.put("Jaccard5Grams", new JaccardOnNGramsSimilarity(5));
	// aliases.put("WeightedDate", new WeightedDatePartSimilarity(1, 3, 5));
	// aliases.put("lodtable",
	// WebtableToDBpediaMatchingProcess.tableType.lodtable);
	// aliases.put("webtable",
	// WebtableToDBpediaMatchingProcess.tableType.webtable);
	// aliases.put("jsonWebTable",
	// WebtableToDBpediaMatchingProcess.tableType.jsonWebTable);
	//
	// Configuration conf = null;
	//
	// try {
	//
	// conf = Configuration.readConfiguration(matchingConfig, m.getParameters(),
	// aliases);
	// return conf;
	//
	// } catch (ClassNotFoundException | InstantiationException |
	// IllegalAccessException | IOException e1) {
	//
	// e1.printStackTrace();
	// return null;
	//
	// }
	// }

	// @Named(CMProperties.REPOSITORY_KEYINDEX)
	// @Provides
	// @Singleton
	// KeyIndex getRepositoryKeyIndex() {
	// KeyIndex keyIndex = new RepositoryKeyIndex();
	// keyIndex.setLuceneIndex(new DefaultIndex(keyIndexPath));
	// return keyIndex;
	// }
	//
	// @Named(CMProperties.KEYINDEX)
	// @Provides
	// @Singleton
	// KeyIndex getLuceIndex() {
	// KeyIndex keyIndex = new KeyIndex();
	// keyIndex.setLuceneIndex(new DefaultIndex(keyIndexPath));
	// return keyIndex;
	// }

	@Provides
	@Singleton
	Timer getParentTimer() {
		return Timer.getNamed("Matching v3 build #" + getTime(MatchingModule.class), null);
	}

	public static Long getTime(Class<?> cl) {
		try {
			String rn = cl.getName().replace('.', '/') + ".class";
			JarURLConnection j = (JarURLConnection) ClassLoader.getSystemResource(rn).openConnection();
			return j.getJarFile().getEntry("META-INF/MANIFEST.MF").getTime();
		} catch (Exception e) {
			return null;
		}
	}

	@Provides
	@Named(CMProperties.EMPTY_FILE)
	String getEmptyFile() {
		return this.emptyFile;
	}

	@Provides
	@Named(CMProperties.SURFACE_FORMS)
	String getSurfaceForms() {
		return this.surfaceForms;
	}

	@Provides
	@Named(CMProperties.REDIRECTS)
	String getRedirects() {
		return this.redirects;
	}


}
