package de.mannheim.uni.expansion.tablefusion.modules;

import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Arrays;
import java.util.List;

import javax.inject.Named;
import javax.inject.Provider;
import javax.inject.Singleton;

import dagger.Module;
import dagger.Provides;
import de.mannheim.uni.expansion.data.DataNamedProperties;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.FileEnding;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableParser;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableParserCSV;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableParserJSON;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesGzReader;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesKryoReader;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesReader;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesSimpleReader;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesWriter;
import de.mannheim.uni.expansion.data.tables.reader.TableConverter;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepositoryImpl;
import de.mannheim.uni.expansion.data.tables.repository.MagnitudeFixer;
import de.mannheim.uni.expansion.data.tables.repository.memory.AnnotatedTablesRepositoryImplMemoryFactory;
import de.mannheim.uni.expansion.data.tables.repository.remote.AnnotatedTablesRepositoryRemote;
import de.mannheim.uni.expansion.tablefusion.profiling.distribution.AnnotatedTablesProfilerRunner;
import de.mannheim.uni.tableprocessor.ColumnTypeGuesser;
import de.mannheim.uni.tableprocessor.IO.ConvertFileToTable;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeNoReferenceKryo;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RemoteRegistry;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;

@Module(library = true, complete = false, injects = { AnnotatedTablesRepository.class, AnnotatedTablesWriter.class, AnnotatedTableParserJSON.class,
		AnnotatedTablesProfilerRunner.class })
public class TablesModule {

	private final String mappingFolder;
	private final String cacheInputFolder;
	private final String cacheOutputFolder;
	private RepositoryLoading repositoryLoading;
	private boolean loadFromCache;
	private final TableParser tableType;
	private final FileEnding tablesFileEnding;
	private final String badAnnotatedTablesFile;
	private boolean filter;
	private KnowledgeBase classFilter;
	private List<AnnotatedTableFilter> filters;
	private TableLoading tableLoading;

	public static enum TableLoading {
		RAW, KRYO;
	}

	public static enum TableParser {
		JSON_TABLE, WEB_TABLE, LOD_TABLE;
	}

	public static TablesModule withFiltersAndCache(String mappingFolder, List<AnnotatedTableFilter> filters, String cacheFolder) {
		return new TablesModule(mappingFolder, cacheFolder, cacheFolder, RepositoryLoading.LOCAL_ONLY, false, TableParser.JSON_TABLE, FileEnding.NO_URL, null,
 false,
 filters, TableLoading.KRYO);
	}

	public static TablesModule withCache(String mappingFolder, String cacheFolder, boolean loadFromCache) {
		return new TablesModule(mappingFolder, cacheFolder, cacheFolder, RepositoryLoading.LOCAL_ONLY, loadFromCache, TableParser.JSON_TABLE, FileEnding.NO_URL,
				null, false, null, TableLoading.KRYO);
	}

	public static TablesModule csvWithCache(String mappingFolder, String cacheFolder, boolean loadFromCache) {
		return new TablesModule(mappingFolder, cacheFolder, cacheFolder, RepositoryLoading.LOCAL_ONLY, loadFromCache, TableParser.WEB_TABLE, FileEnding.NO_URL,
				null, false, null, TableLoading.KRYO);
	}

	public static TablesModule withCache(String mappingFolder, String cacheInputFolder, String cacheOutputFolder, boolean loadFromCache) {
		return new TablesModule(mappingFolder, cacheInputFolder, cacheOutputFolder, RepositoryLoading.LOCAL_ONLY, loadFromCache, TableParser.JSON_TABLE,
				FileEnding.NO_URL, null, false,
				null,
				TableLoading.KRYO);
	}

	// public static TablesModule withFiltersAndCache(String mappingFolder,
	// ArrayList<AnnotatedTableFilter> filters, String cacheInputFolder,
	// String cacheOutputFolder) {
	// return new TablesModule(mappingFolder, cacheInputFolder,
	// cacheOutputFolder, RepositoryLoading.LOCAL_ONLY, false,
	// TableParser.JSON_TABLE,
	// FileEnding.NO_URL, null, true, filters);
	// }

	public TablesModule(String mappingFolder, String cacheInputFolder, String cacheOutputFolder, RepositoryLoading repositoryLoading, boolean loadFromCache,
 TableParser tableType,
 FileEnding tablesFileEnding, String badAnnotatedTablesFile, boolean filter, List<AnnotatedTableFilter> filters,
			TableLoading tableLoading) {
		super();
		this.mappingFolder = mappingFolder;
		this.cacheInputFolder = cacheInputFolder;
		this.cacheOutputFolder = cacheOutputFolder;
		this.repositoryLoading = repositoryLoading;
		this.loadFromCache = loadFromCache;
		this.tableType = tableType;
		this.tablesFileEnding = tablesFileEnding;
		this.badAnnotatedTablesFile = badAnnotatedTablesFile;
		this.filter = filter;
		this.filters = filters;
		this.tableLoading = tableLoading;
	}

	public void setRepositoryLoading(RepositoryLoading repositoryLoading) {
		this.repositoryLoading = repositoryLoading;
	}

	public void setLoadFromCache(boolean loadFromCache) {
		this.loadFromCache = loadFromCache;
	}

	public void setFilter(boolean filter) {
		this.filter = filter;
	}

	public void setClassFilter(KnowledgeBase dbpedia) {
		this.classFilter = dbpedia;
	}

	private final static String registryKey = "AnnotatedTablesRepositoryRemote";

	// private AnnotatedTablesRepository beforeValueAdaptionRepository = null;
	//
	// public void setBeforeValueAdaptionRepository(AnnotatedTablesRepository
	// beforeValueAdaptionRepository) {
	// this.beforeValueAdaptionRepository = beforeValueAdaptionRepository;
	// };
	//
	// @Named(AttributeMatcher.BEFORE_VALUE_ADAPTION)
	// Provider<AnnotatedTablesRepository> getBeforeValueAdaption() {
	// return new Provider<AnnotatedTablesRepository>() {
	//
	// @Override
	// public AnnotatedTablesRepository get() {
	// return beforeValueAdaptionRepository;
	// }
	//
	// };
	// }

	@Provides
	@Singleton
	public AnnotatedTablesRepository loadAnnotatedTables(AnnotatedTablesReader annotatedTablesReader, MagnitudeFixer magnitudeFixer) {

		try {

			boolean isInRegistry = true;
			if (repositoryLoading == RepositoryLoading.REMOTE_LOAD_REMOTELY) {
				Registry registry = RemoteRegistry.getReadingRegistry();
				isInRegistry = Arrays.asList(registry.list()).contains(registryKey);

			}

		// AnnotatedTablesRepositoryImplMemoryFactory factory = new
		// AnnotatedTablesRepositoryImplMemoryFactory(reader);
		// return factory.getRepository();
			if (repositoryLoading == RepositoryLoading.LOCAL_ONLY || repositoryLoading == RepositoryLoading.REMOTE_LOAD_LOCALLY || !isInRegistry) {

				AnnotatedTablesRepositoryImplMemoryFactory repositoryFactory = new AnnotatedTablesRepositoryImplMemoryFactory(annotatedTablesReader, filter,
						classFilter, magnitudeFixer, filters);
				AnnotatedTablesRepository annotatedTablesRepository = repositoryFactory.getRepository();

				if (repositoryLoading == RepositoryLoading.REMOTE_LOAD_LOCALLY) {
					Registry registry = RemoteRegistry.getPublishingRegistry();

					AnnotatedTablesRepositoryRemote annotatedTablesRepositoryRemote = (AnnotatedTablesRepositoryRemote) annotatedTablesRepository;
					AnnotatedTablesRepositoryRemote stubAnno = (AnnotatedTablesRepositoryRemote) UnicastRemoteObject
							.exportObject(annotatedTablesRepositoryRemote, 0);
					registry.bind("AnnotatedTablesRepositoryRemote", stubAnno);

				}

				return annotatedTablesRepository;
			} else if (repositoryLoading == RepositoryLoading.REMOTE_LOAD_REMOTELY) {
			try {
					Registry registry = RemoteRegistry.getReadingRegistry();
				AnnotatedTablesRepositoryRemote stub = (AnnotatedTablesRepositoryRemote) registry.lookup("AnnotatedTablesRepositoryRemote");
				return new AnnotatedTablesRepositoryImpl(stub);
			} catch (Exception e) {
				System.err.println("Client exception: " + e.toString());
				e.printStackTrace();
					throw new RuntimeException(e);
			}

		}
			else {
				throw new IllegalStateException("illegal value for repository loading");
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

	}


	@Provides
	@Singleton
	public AnnotatedTablesReader getReader(Provider<AnnotatedTablesKryoReader> kryoReader, Provider<AnnotatedTablesGzReader> gzReader) {
		if (loadFromCache) {
			return kryoReader.get();
		} else {
			return gzReader.get();
		}
	}

	@Provides
	@Singleton
	AnnotatedTablesKryoReader getKryoReader(TableConverter tableConverter, ThreadSafeNoReferenceKryo threadSafeKryo) {
		return new AnnotatedTablesKryoReader(cacheInputFolder, tableConverter, true, threadSafeKryo);
	}

	@Provides
	@Singleton
	AnnotatedTablesGzReader getGZReader(AnnotatedTableParser parser) {
		return new AnnotatedTablesGzReader(mappingFolder, tablesFileEnding, parser, badAnnotatedTablesFile);
	}

	@Provides
	@Singleton
	AnnotatedTablesSimpleReader getSimpleReader(AnnotatedTableParser parser) {
		return new AnnotatedTablesSimpleReader(mappingFolder, this.tablesFileEnding, parser);
	}

	@Provides
	@Named(DataNamedProperties.TABLES_MAPPING_FOLDER)
	String getTablesMappingFolder() {
		return this.mappingFolder;
	}

	@Provides
	@Named(DataNamedProperties.CACHE_INPUT_FOLDER)
	String getCacheInputFolder() {
		return this.cacheInputFolder;
	}

	@Provides
	@Named(DataNamedProperties.CACHE_OUTPUT_FOLDER)
	String getCacheOutputFolder() {
		return this.cacheOutputFolder;
	}


	@Provides
	AnnotatedTableParser getParser(AnnotatedTableParserJSON annotatedTableParserJSON, AnnotatedTableParserCSV annotatedTableParserCSV) {

		switch (this.tableType) {
		case JSON_TABLE:
			return annotatedTableParserJSON;
		// case LOD_TABLE:
		// return annotatedTableParserCSV;
		case WEB_TABLE:
			return annotatedTableParserCSV;

		}
		throw new IllegalStateException("TABLE TYPE UNIMPLEMENTED / YOU CAN NOT INJECT A PARSER WITHOUT A TABLETYPE");
	}

	@Provides
	ConvertFileToTable getConvertFileToToable() {
		return new ConvertFileToTable();
	}

	@Provides
	ColumnTypeGuesser getColumnTypeGuesser() {
		return new ColumnTypeGuesser();
	}

	@Provides
	@Named(DataNamedProperties.BAD_ANNOTATED_TABLES_FILE)
	String getBadAnnotatedTablesFile() {
		return this.badAnnotatedTablesFile;
	}

	@Provides
	@Named(DataNamedProperties.TABLES_MAPPING_ENDING)
	FileEnding getTabledFileEnding() {
		return this.tablesFileEnding;
	}

	@Provides
	@Named(DataNamedProperties.TABLES_TYPE)
	TableParser getTableType() {
		return this.tableType;
	}

	public void setFilters(List<AnnotatedTableFilter> filters) {
		this.filters = filters;

	}

}
