package de.mannheim.uni.expansion.tablefusion.adapters.regressionweights;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.inject.Inject;
import javax.inject.Named;

import au.com.bytecode.opencsv.CSVReader;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.tablefusion.NamedProperties;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;

public class RegressionWeightsReader {

	private final String outputFolder;

	@Inject
	public RegressionWeightsReader(@Named(NamedProperties.WEIGHTS_FOLDER) String outputFolder) {
		this.outputFolder = outputFolder;
	}

	public RegressionWeightsRepository read() throws IOException {
		Map<String, RegressionWeights> map = new HashMap<>();

		Set<String> names = getPrefixesFromFolder(outputFolder);

		for (String name : names) {
			File generalFile = new File(outputFolder + "/weights" + name + ".csv");
			RegressionWeightsContainer generalWeight = readGeneralFile(generalFile);

			Map<ClassPropertyId, RegressionWeightsContainer> propertyWeights = new HashMap<>();
			for (File classPropertyFile : new File(outputFolder + "/weights" + name).listFiles()) {
				ClassPropertyScores propertyWeight = readClassPropertyFile(classPropertyFile);
				// // TODO: bad fix
				// if (propertyWeight != null)
				propertyWeights.put(propertyWeight.getClassPropertyId(), propertyWeight.getRegressionWeightsContainer());
			}

			RegressionWeights regressionWeights = new RegressionWeights(generalWeight, propertyWeights);

			map.put(name, regressionWeights);
		}

		return new RegressionWeightsRepositoryImpl(map);
	}

	private Set<String> getPrefixesFromFolder(String outputFolder) {
		Set<String> ret = new HashSet<>();
		String[] names = new File(outputFolder).list();
		for (String name : names) {
			int length = name.length();
			int weightsLength = "weights".length();
			if (".csv".equals(name.substring(length - 4, length))) {

				ret.add(name.substring(weightsLength, length - 4));
			} else {
				ret.add(name.substring(weightsLength, length));
			}

		}

		return ret;

	}

	private RegressionWeightsContainer readGeneralFile(File file) throws IOException {
		CSVReader reader = new CSVReader(new InputStreamReader(new FileInputStream(file)), ';', '"');

		EnumMap<SimilarityType, RegressionWeightsScores> map = new EnumMap<>(SimilarityType.class);

		for (String[] line : reader.readAll()) {
			SimilarityType similarityType = SimilarityType.getFromString(line[0]);
			RegressionWeightsScores scores = parseLine(line[1], line[2], line[3], line[4], line[5]);
			map.put(similarityType, scores);
		}

		reader.close();
		return new RegressionWeightsContainer(map);
	}

	private ClassPropertyScores readClassPropertyFile(File file) throws IOException {
		CSVReader reader = new CSVReader(new InputStreamReader(new FileInputStream(file)), ';', '"');

		EnumMap<SimilarityType, RegressionWeightsScores> map = new EnumMap<>(SimilarityType.class);
		ClassPropertyId classPropertyId = null;

		for (String[] line : reader.readAll()) {

			if (classPropertyId == null) {
				ClassId classId = new ClassId(line[0]);
				PropertyId propertyId = new PropertyId(line[1]);
				classPropertyId = new ClassPropertyId(classId, propertyId);
			}

			SimilarityType similarityType = SimilarityType.getFromString(line[2]);
			RegressionWeightsScores scores = parseLine(line[3], line[4], line[5], line[6], line[7]);

			// // TODO: bad fix
			// if (scores == null)
			// return null;
			map.put(similarityType, scores);
		}

		reader.close();
		return new ClassPropertyScores(classPropertyId, new RegressionWeightsContainer(map));
	}

	private RegressionWeightsScores parseLine(String instanceScoreString, String columnScoreString, String tableScoreString, String hostScoreString,
			String interceptString) {
		// // TODO: Bad Fix. FIX IT
		// if (instanceScoreString.equals("0.0") &&
		// columnScoreString.equals("0.0") && tableScoreString.equals("0.0") &&
		// hostScoreString.equals("0.0"))
		// return null;

		Double instanceScore = Double.parseDouble(instanceScoreString);
		Double columnScore = Double.parseDouble(columnScoreString);
		Double tableScore = Double.parseDouble(tableScoreString);
		Double hostScore = Double.parseDouble(hostScoreString);
		Double intercept = Double.parseDouble(interceptString);

		return new RegressionWeightsScores(instanceScore, columnScore, tableScore, hostScore, intercept);
	}

	private static class ClassPropertyScores {
		private final ClassPropertyId classPropertyId;
		private final RegressionWeightsContainer regressionWeightsContainer;

		private ClassPropertyScores(ClassPropertyId classPropertyId, RegressionWeightsContainer regressionWeightsContainer) {
			super();
			this.classPropertyId = classPropertyId;
			this.regressionWeightsContainer = regressionWeightsContainer;
		}

		public ClassPropertyId getClassPropertyId() {
			return classPropertyId;
		}

		public RegressionWeightsContainer getRegressionWeightsContainer() {
			return regressionWeightsContainer;
		}

	}

}
