package de.mannheim.uni.expansion.tablefusion.analysis;

import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PerformancePR;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.ResultsCategory;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;

public class AccuracyAverager {
	private final OntologyRepository ontologyRepository;

	@Inject
	public AccuracyAverager(OntologyRepository ontologyRepository) {
		this.ontologyRepository = ontologyRepository;
	}

	public <E> void writeAllTo(String folderPath, Map<E, FusionPerformance<PerformancePR>> ret) throws IOException {

		FileWriter fw = new FileWriter(folderPath + "/AVERAGE_TOTAL.csv", true);
		fw.write(averageTotal(ret));
		fw.close();

		fw = new FileWriter(folderPath + "/AVERAGE_CLASS_ID.csv", true);
		fw.write(averageClassId(ret));
		fw.close();

		fw = new FileWriter(folderPath + "/AVERAGE_TIME_DEPENDENCE.csv", true);
		fw.write(averageTimeDependence(ret));
		fw.close();

		fw = new FileWriter(folderPath + "/AVERAGE_DATA_TYPE.csv", true);
		fw.write(averageDataType(ret));
		fw.close();

	}

	public <E> String averageTotal(Map<E, FusionPerformance<PerformancePR>> ret) throws IOException {

		return average(ret, ResultsCategory.TOTAL.getPrefix(), (f, e) -> {
			return ResultsCategory.getNormalPrefix(f.toString());
		} , (e) -> {
			return Boolean.TRUE;
		});

	}

	public <E> String averageClassId(Map<E, FusionPerformance<PerformancePR>> ret) throws IOException {

		return average(ret, ResultsCategory.CLASS_ID.getPrefix(), (f, e) -> {
			return ResultsCategory.getClassIdPrefix(f.toString(), e);
		} , (e) -> {
			return e.getClassId();
		});

	}

	public <E> String averageTimeDependence(Map<E, FusionPerformance<PerformancePR>> ret) throws IOException {

		return average(ret, ResultsCategory.TIME_DEPENDENCE.getPrefix(), (f, e) -> {
			return ResultsCategory.getTimeDepedence(f.toString(), e);
		} , (e) -> {
			return ontologyRepository.isTimeDependent(e.getPropertyId());
		});

	}

	public <E> String averageDataType(Map<E, FusionPerformance<PerformancePR>> ret) throws IOException {

		return average(ret, ResultsCategory.DATA_TYPE.getPrefix(), (f, e) -> {
			return ResultsCategory.getDataTypePrefix(f.toString(), e);
		} , (e) -> {
			return ontologyRepository.getDataType(e.getPropertyId());
		});

	}

	public <E, F> String average(Map<E, FusionPerformance<PerformancePR>> ret, String[] headerPrefix, BiFunction<E, F, String[]> linePrefix,
			Function<ClassPropertyId, F> mapper) throws IOException {
		StringWriter sw = new StringWriter();
		CSVWriter writer = new CSVWriter(sw);

		int length = headerPrefix.length + 8;
		{
			String[] headerLine = new String[length];

			for (int i = 0; i < headerPrefix.length; i++) {
				headerLine[i] = headerPrefix[i];
			}
			headerLine[headerPrefix.length] = "Precision";
			headerLine[headerPrefix.length + 1] = "Recall";
			headerLine[headerPrefix.length + 2] = "F-Measure";
			headerLine[headerPrefix.length + 3] = "AF-Measure";
			headerLine[headerPrefix.length + 4] = "VarPREC";
			headerLine[headerPrefix.length + 5] = "VarREC";
			headerLine[headerPrefix.length + 6] = "VarFM";
			headerLine[headerPrefix.length + 7] = "VarAFM";
			writer.writeNext(headerLine);

		}
		for (Entry<E, FusionPerformance<PerformancePR>> entry : ret.entrySet()) {
			Map<ClassPropertyId, PerformancePR> performance = entry.getValue().getPerProperty();

			Map<F, List<PerformancePR>> splitMap = performance.entrySet().stream().collect(Collectors.groupingBy((e) -> {
				return mapper.apply(e.getKey());
			} , Collectors.mapping((e) -> {
				return e.getValue();
			} , Collectors.toList())));

			for (Entry<F, List<PerformancePR>> splitEntry : splitMap.entrySet()) {
				int count = splitEntry.getValue().size();
				double precision = 0.0;
				double recall = 0.0;
				double fMeasure = 0.0;
				double aFmeasure = 0.0;

				for (PerformancePR ppr : splitEntry.getValue()) {
					precision += ppr.getPrecision();
					recall += ppr.getRecall();
					fMeasure += ppr.getfMeasure();
					aFmeasure += ppr.getaFMeasure();
				}

				precision = precision / count;
				recall = recall / count;
				fMeasure = fMeasure / count;
				aFmeasure = aFmeasure / count;

				double varPrecision = 0.0;
				double varRecall = 0.0;
				double varFMeasure = 0.0;
				double varAFmeasure = 0.0;

				for (Entry<ClassPropertyId, PerformancePR> cp : performance.entrySet()) {
					varPrecision += Math.pow(cp.getValue().getPrecision() - precision, 2);
					varRecall += Math.pow(cp.getValue().getRecall() - recall, 2);
					varFMeasure += Math.pow(cp.getValue().getfMeasure() - fMeasure, 2);
					varAFmeasure += Math.pow(cp.getValue().getaFMeasure() - aFmeasure, 2);
				}

				String[] line = new String[length];
				String[] prefix = linePrefix.apply(entry.getKey(), splitEntry.getKey());
				for (int i = 0; i < prefix.length; i++) {
					line[i] = prefix[i];
				}

				line[prefix.length] = String.valueOf(precision);
				line[prefix.length + 1] = String.valueOf(recall);
				line[prefix.length + 2] = String.valueOf(fMeasure);
				line[prefix.length + 3] = String.valueOf(aFmeasure);
				line[prefix.length + 4] = String.valueOf(Math.sqrt(varPrecision / count));
				line[prefix.length + 5] = String.valueOf(Math.sqrt(varRecall / count));
				line[prefix.length + 6] = String.valueOf(Math.sqrt(varFMeasure / count));
				line[prefix.length + 7] = String.valueOf(Math.sqrt(varAFmeasure / count));
				writer.writeNext(line);
			}

		}

		writer.close();
		return sw.toString();
	}
}
