package de.mannheim.uni.expansion.tablefusion.analysis;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleIdContainer;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.AbstractGroupledValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfigFilter;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.Qualifier;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.FilteredValues;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;

public class MaxCorrectFinder {

	private final GroupedValuesRepository groupedValuesRepository;
	private final OntologyRepository ontologyRepository;
	private final ValueEvaluator valueEvaluator;
	private final PropertiesConfigFilter propertiesConfigFilter;

	// private final PersistentParallel3<Entry<TripleId, MatchedValues>>
	// getMaxCorrectValues = new PersistentParallel3<Map.Entry<TripleId,
	// MatchedValues>>(
	// "MaxCorrectValuesTotal");
	// private final PersistentParallel3<Entry<TripleId, MatchedValues>>
	// getMaxCorrectValuePerDataType = new
	// PersistentParallel3<Map.Entry<TripleId, MatchedValues>>(
	// "MaxCorrectValuesDataType");
	//
	// private final PersistentParallel3<Entry<TripleId, MatchedValues>>
	// getMaxCorrectValuePerClass = new PersistentParallel3<Map.Entry<TripleId,
	// MatchedValues>>(
	// "MaxCorrectValuesClass");
	// private final PersistentParallel3<Entry<TripleId, MatchedValues>>
	// getMaxCorrectValuePerClassPropertyId = new
	// PersistentParallel3<Map.Entry<TripleId, MatchedValues>>(
	// "MaxCorrectValuesClassPropertyId");

	private final KnowledgeBase knowledgeBase;
	private final AnnotatedTablesRepository tablesRepository;
	private final SimilarityCalculator similarityCalculator;
	private final TimestampQualifiersAdapter timestampQualifiersAdapter;

	@Inject
	public MaxCorrectFinder(GroupedValuesRepository groupedValuesRepository, OntologyRepository ontologyRepository, ValueEvaluator valueEvaluator,
			KnowledgeBase knowledgeBase, AnnotatedTablesRepository tablesRepository, SimilarityCalculator similarityCalculator,
			TimestampQualifiersAdapter timestampQualifiersAdapter, PropertiesConfigFilter propertiesConfigFilter) {
		this.groupedValuesRepository = groupedValuesRepository;
		this.ontologyRepository = ontologyRepository;
		this.valueEvaluator = valueEvaluator;
		this.knowledgeBase = knowledgeBase;
		this.tablesRepository = tablesRepository;
		this.similarityCalculator = similarityCalculator;
		this.timestampQualifiersAdapter = timestampQualifiersAdapter;
		this.propertiesConfigFilter = propertiesConfigFilter;
	}

	public <E extends TripleIdContainer> FusionPerformance<Integer> getMaxCorrect(AbstractGroupledValues<E> groupedValues, FilteredValues filteredValues,
			PropertiesConfig propertiesConfig, FusionConfig fusionConfig) {

		Integer max = getMaxCorrectValues(groupedValues, filteredValues, propertiesConfig, fusionConfig);
		Map<ValueType, Integer> maxPerDataType = getMaxCorrectValuePerDataType(groupedValues, filteredValues, propertiesConfig, fusionConfig);
		Map<Boolean, Integer> maxPerTimeDependence = getMaxCorrectValuePerTimeDependence(groupedValues, filteredValues, propertiesConfig, fusionConfig);
		Map<ClassId, Integer> maxPerClassId = getMaxCorrectValuePerClass(groupedValues, filteredValues, propertiesConfig, fusionConfig);
		Map<ClassPropertyId, Integer> maxPerClassPropertyId = getMaxCorrectValuePerClassPropertyId(groupedValues, filteredValues, propertiesConfig,
				fusionConfig);
		Map<TimestampPointInTime, Integer> maxPerYear = getMaxCorrectValuePerYear(groupedValues, filteredValues, propertiesConfig, fusionConfig);
		Map<Pair<ClassPropertyId, TimestampPointInTime>, Integer> maxPerYearAndClassProperty = getMaxCorrectValuePerClassPropertyAndYear(groupedValues,
				filteredValues, propertiesConfig, fusionConfig);

		// maxPerClassPropertyId.put(new
		// ClassPropertyId(WikidataClasses.wikidataCountry, Stats.population),
		// 15);
		// maxPerClassPropertyId.put(new
		// ClassPropertyId(WikidataClasses.wikidataCountry,
		// Stats.populationDensity), 60);
		// maxPerClassPropertyId.put(new
		// ClassPropertyId(WikidataClasses.wikidataFootball, Stats.teams), 400);

		return new FusionPerformance<Integer>(max, maxPerTimeDependence, maxPerDataType, maxPerClassId, maxPerClassPropertyId, maxPerYear,
				maxPerYearAndClassProperty);
		// return new Quad<>(max, maxPerDataType, maxPerClassId,
		// maxPerClassPropertyId);
	}

	private <E extends TripleIdContainer> Map<Pair<ClassPropertyId, TimestampPointInTime>, Integer> getMaxCorrectValuePerClassPropertyAndYear(
			AbstractGroupledValues<E> groupedValues, FilteredValues filteredValues, PropertiesConfig propertiesConfig, FusionConfig fusionConfig) {
		ConcurrentHashMap<Pair<ClassPropertyId, TimestampPointInTime>, AtomicInteger> map = new ConcurrentHashMap<>();

		try {
			PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {
				if (!(e.getValue().getValues().size() >= fusionConfig.getMinSource()))
					return;
				if (!this.propertiesConfigFilter.include(propertiesConfig, e.getKey()))
					return;
				if (e.getKey() instanceof QualifiedTripleId) {
					QualifiedTripleId qtid = (QualifiedTripleId) e.getKey();
					TimestampPointInTime tipt = null;
					for (AbstractQualifier qualifier : qtid.getQualifiers()) {
						if (qualifier instanceof TimestampPointInTime) {
							tipt = (TimestampPointInTime) qualifier;
							break;
						}
					}
					if (tipt != null) {
						ClassPropertyId cpid = new ClassPropertyId(e.getKey().tripleId());
						AtomicInteger maxCorrect = CHMHelper.getFromCHM(map, new Pair<>(cpid, tipt), () -> new AtomicInteger());
						maxCorrect.addAndGet(correctCountGeneric(e.getKey(), e.getValue().getValues()));
					}

				}

			}, "MaxCorrect");
		} catch (Exception e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		return map.entrySet().parallelStream().collect(Collectors.toMap(Entry::getKey, e -> e.getValue().get()));
	}

	// public void printMaxCorrectValuesInCorpus(boolean fuseLists) {
	//
	// GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();
	//
	// int maxCorrect = getMaxCorrectValues(groupedValues);
	//
	// System.out.println("Max Correct Values in Corpus: " + maxCorrect);
	//
	// System.out.println("Max Correct Values for DataType");
	// for (Entry<ValueType, Integer> e :
	// getMaxCorrectValuePerDataType(groupedValues, fuseLists).entrySet()) {
	// System.out.println(e.getKey().toString() + ": " + e.getValue());
	// }
	// }

	// @Deprecated
	// public int getMaxCorrectValues(GroupedValues groupedValues) {
	// return getMaxCorrectValues(groupedValues, null,
	// PropertiesConfig.defaultConfig(), null);
	// }

	public <E extends TripleIdContainer> int getMaxCorrectValues(AbstractGroupledValues<E> groupedValues, FilteredValues filteredValues,
			PropertiesConfig propertiesConfig, FusionConfig fusionConfig) {

		AtomicInteger maxCorrect = new AtomicInteger();

		try {
			PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {
				if (!this.propertiesConfigFilter.include(propertiesConfig, e.getKey()))
					return;
				if (!this.propertiesConfigFilter.include(propertiesConfig, e.getKey()))
					return;

				if (!(e.getValue().getValues().size() >= fusionConfig.getMinSource()))
					return;

				maxCorrect.addAndGet(correctCountGeneric(e.getKey(), e.getValue().getValues()));
				// ColumnDataType cdt =
				// ontologyRepository.getDataType(e.getKey().getClassId(),
				// e.getKey().getPropertyId());

			}, "MaxCorrect");
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return maxCorrect.get();
	}

	// @Deprecated
	// public Map<ValueType, Integer>
	// getMaxCorrectValuePerDataType(GroupedValues groupedValues, boolean
	// fuseLists) {
	//
	// return getMaxCorrectValuePerDataType(groupedValues, null,
	// PropertiesConfig.defaultConfig());
	// }

	public <E extends TripleIdContainer> Map<Boolean, Integer> getMaxCorrectValuePerTimeDependence(AbstractGroupledValues<E> groupedValues,
			FilteredValues filteredValues, PropertiesConfig propertiesConfig, FusionConfig fusionConfig) {
		ConcurrentHashMap<Boolean, AtomicInteger> timeDependences = new ConcurrentHashMap<>();

		try {
			PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {
				if (!this.propertiesConfigFilter.include(propertiesConfig, e.getKey()))
					return;
				if (!(e.getValue().getValues().size() >= fusionConfig.getMinSource()))
					return;

				TripleId tid = e.getKey().tripleId();
				boolean timeDependence = ontologyRepository.isTimeDependent(tid.getPropertyId());

				AtomicInteger maxCorrectPerTimeDependence = CHMHelper.getFromCHM(timeDependences, timeDependence, () -> new AtomicInteger());
				maxCorrectPerTimeDependence.addAndGet(correctCountGeneric(e.getKey(), e.getValue().getValues()));

			}, "MaxCorrect");
		} catch (Exception e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		return timeDependences.entrySet().parallelStream().collect(Collectors.toMap(Entry::getKey, e -> e.getValue().get()));
	}

	public <E extends TripleIdContainer> Set<E> getCorrectTripleIds(AbstractGroupledValues<E> groupedValues, FilteredValues filteredValues,
			PropertiesConfig propertiesConfig, FusionConfig fusionConfig) {
		Set<E> ret = Collections.newSetFromMap(new ConcurrentHashMap<>());

		try {
			PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {
				if (!this.propertiesConfigFilter.include(propertiesConfig, e.getKey()))
					return;
				if (!(e.getValue().getValues().size() >= fusionConfig.getMinSource()))
					return;

				int correctCount = correctCountGeneric(e.getKey(), e.getValue().getValues());
				if (correctCount > 0) {
					ret.add(e.getKey());
				}

			}, "MaxCorrect");
		} catch (Exception e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		return ret;
	}

	public <E extends TripleIdContainer> Map<ValueType, Integer> getMaxCorrectValuePerDataType(AbstractGroupledValues<E> groupedValues,
			FilteredValues filteredValues, PropertiesConfig propertiesConfig, FusionConfig fusionConfig) {
		ConcurrentHashMap<ValueType, AtomicInteger> dataTypes = new ConcurrentHashMap<>();

		try {
			PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {

				if (!(e.getValue().getValues().size() >= fusionConfig.getMinSource()))
					return;

				if (!this.propertiesConfigFilter.include(propertiesConfig, e.getKey()))
					return;
				TripleId tid = e.getKey().tripleId();
				ValueType cdt = ontologyRepository.getDataType(tid.getPropertyId());

				AtomicInteger maxCorrectPerDataType = CHMHelper.getFromCHM(dataTypes, cdt, () -> new AtomicInteger());
				maxCorrectPerDataType.addAndGet(correctCountGeneric(e.getKey(), e.getValue().getValues()));

			}, "MaxCorrect");
		} catch (Exception e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		return dataTypes.entrySet().parallelStream().collect(Collectors.toMap(Entry::getKey, e -> e.getValue().get()));
	}

	public <E extends TripleIdContainer> Map<ClassId, Integer> getMaxCorrectValuePerClass(AbstractGroupledValues<E> groupedValues,
			FilteredValues filteredValues, PropertiesConfig propertiesConfig, FusionConfig fusionConfig) {
		ConcurrentHashMap<ClassId, AtomicInteger> map = new ConcurrentHashMap<>();

		try {
			PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {
				if (!(e.getValue().getValues().size() >= fusionConfig.getMinSource()))
					return;
				if (!this.propertiesConfigFilter.include(propertiesConfig, e.getKey()))
					return;
				TripleId tid = e.getKey().tripleId();
				AtomicInteger maxCorrect = CHMHelper.getFromCHM(map, tid.getClassId(), () -> new AtomicInteger());
				maxCorrect.addAndGet(correctCountGeneric(e.getKey(), e.getValue().getValues()));

			}, "MaxCorrect");
		} catch (Exception e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		return map.entrySet().parallelStream().collect(Collectors.toMap(Entry::getKey, e -> e.getValue().get()));
	}

	public <E extends TripleIdContainer> Map<TimestampPointInTime, Integer> getMaxCorrectValuePerYear(AbstractGroupledValues<E> groupedValues,
			FilteredValues filteredValues, PropertiesConfig propertiesConfig, FusionConfig fusionConfig) {
		ConcurrentHashMap<TimestampPointInTime, AtomicInteger> map = new ConcurrentHashMap<>();

		try {
			PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {
				if (!(e.getValue().getValues().size() >= fusionConfig.getMinSource()))
					return;
				if (!this.propertiesConfigFilter.include(propertiesConfig, e.getKey()))
					return;
				if (e.getKey() instanceof QualifiedTripleId) {
					QualifiedTripleId qtid = (QualifiedTripleId) e.getKey();
					TimestampPointInTime tipt = null;
					for (AbstractQualifier qualifier : qtid.getQualifiers()) {
						if (qualifier instanceof TimestampPointInTime) {
							tipt = (TimestampPointInTime) qualifier;
							break;
						}
					}
					if (tipt != null) {
						AtomicInteger maxCorrect = CHMHelper.getFromCHM(map, tipt, () -> new AtomicInteger());
						maxCorrect.addAndGet(correctCountGeneric(e.getKey(), e.getValue().getValues()));
					}

				}

			}, "MaxCorrect");
		} catch (Exception e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		return map.entrySet().parallelStream().collect(Collectors.toMap(Entry::getKey, e -> e.getValue().get()));
	}

	public <E extends TripleIdContainer> Map<ClassPropertyId, Integer> getMaxCorrectValuePerClassPropertyId(AbstractGroupledValues<E> groupedValues,
			FilteredValues filteredValues, PropertiesConfig propertiesConfig, FusionConfig fusionConfig) {
		ConcurrentHashMap<ClassPropertyId, AtomicInteger> map = new ConcurrentHashMap<>();

		try {
			PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {
				if (!this.propertiesConfigFilter.include(propertiesConfig, e.getKey()))
					return;
				if (!(e.getValue().getValues().size() >= fusionConfig.getMinSource()))
					return;
				TripleId tid = e.getKey().tripleId();
				AtomicInteger maxCorrect = CHMHelper.getFromCHM(map, new ClassPropertyId(tid), () -> new AtomicInteger());
				maxCorrect.addAndGet(correctCountGeneric(e.getKey(), e.getValue().getValues()));

			}, "MaxCorrect");
		} catch (Exception e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		return map.entrySet().parallelStream().collect(Collectors.toMap(Entry::getKey, e -> e.getValue().get()));
	}

	// private <E> boolean correct(E tid, List<SingleValue> values, boolean
	// fuseLists, FilteredValues filteredValues) {
	// List<AbstractObject> combinedValues =
	// values.stream().map(SingleValue::getValue).collect(Collectors.toList());
	// if (tid instanceof TripleId) {
	// return valueEvaluator.maxCorrect((TripleId) tid, combinedValues,
	// fuseLists, filteredValues);
	// }
	//
	// return valueEvaluator.maxCorrect((QualifiedTripleId) tid, combinedValues,
	// fuseLists, filteredValues);
	//
	// }

	private int correctCountGeneric(TripleIdContainer tid, List<SingleValue> values) {
		if (tid instanceof TripleId) {
			return correctCount((TripleId) tid, values);
		} else if (tid instanceof QualifiedTripleId) {
			return correctCount((QualifiedTripleId) tid, values);

		}

		throw new IllegalStateException();

	}

	private int correctCount(QualifiedTripleId tid, List<SingleValue> values) {
		QualifiedValueList qvl = this.knowledgeBase.getQualifiedValueListWithEqualProperties(tid.tripleId().getValueId());
		Set<AbstractObject> correctValues = this.timestampQualifiersAdapter.getForQualifiers(qvl, tid.getQualifiers());

		Set<PropertyId> propertyIdSet = new HashSet<>();
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/previousWork"));
		propertyIdSet.add(new PropertyId("http://dbpedia.org/ontology/producer"));
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/recordLabel"));
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/Work/runtime"));
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/runtime"));
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/genre"));

		if (propertyIdSet.contains(tid.getQualifiedPropertyId().getPropertyId())) {
			List<Pair<AbstractObject, AbstractObject>> list = values.stream().map((e) -> {
				return new Pair<>(tablesRepository.getValue(e.getId().getTableId(), e.getId().getRowIdx(), e.getId().getColumnIdx()), e.getValue());
			}).collect(Collectors.toList());

			List<AbstractObject> labelList = values.stream().map((e) -> {
				return tablesRepository.getKeyValue(e.getId().getTableId(), e.getId().getRowIdx());
			}).collect(Collectors.toList());

			List<AbstractObject> columnHeaders = values.stream().map((e) -> {
				return tablesRepository.getTable(e.getId().getTableId()).getColumnHeaders().get(e.getId().getColumnIdx());
			}).collect(Collectors.toList());
			int sdf = 123;
		}

		for (AbstractObject correctValue : correctValues) {
			for (SingleValue sv : values) {
				if (similarityCalculator.isEqual(correctValue, sv.getValue())) {
					return 1;
				}
			}
		}

		return 0;
	}

	private int correctCount(TripleId tid, List<SingleValue> values) {
		int correctCount = 0;
		QualifiedValueList qvl = this.knowledgeBase.getQualifiedValueListWithEqualProperties(tid.getValueId());

		// Set<PropertyId> propertyIdSet = new HashSet<>();
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/previousWork"));
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/producer"));
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/recordLabel"));
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/Work/runtime"));
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/runtime"));
		// propertyIdSet.add(new
		// PropertyId("http://dbpedia.org/ontology/genre"));
		//
		// if (propertyIdSet.contains(tid.getPropertyId())) {
		// int sdf = 123;
		// }
		if (qvl != null) {
			for (QualifiedValueHolder claim : qvl.getClaims()) {

				for (SingleValue sv : values) {
					this.similarityCalculator.isEqual(claim.getValue(), sv.getValue());

					Map<TimestampIdentifier, Integer> tableYears = this.tablesRepository.getMetaData(sv.getId().getTableId()).getTableYears();
					List<ColumnIdx> timeColumns = this.tablesRepository.getTable(sv.getId().getTableId()).getTimeColumns();

					boolean correct = true;
					if (claim.hasQualifiers()) {
						for (Qualifier qualifier : claim.getQualifiers()) {

							if (qualifier.getQualifierValue() instanceof TimeObject) {

								Integer qualifierYear = TimeObject.getYear(qualifier.getQualifierValue());
								boolean qualifierCorrect = false;

								for (Entry<TimestampIdentifier, Integer> ty : tableYears.entrySet()) {
									if (ty.getValue().equals(qualifierYear)) {
										qualifierCorrect = true;
										break;
									}
								}

								if (!qualifierCorrect) {
									for (ColumnIdx columnIdx : timeColumns) {
										AbstractObject ao = this.tablesRepository.getTable(sv.getId().getTableId()).getValue(sv.getId().getRowIdx(), columnIdx);
										Integer columnYear = TimeObject.getYear(ao);
										if (columnYear != null && columnYear.equals(qualifierYear)) {
											qualifierCorrect = true;
											break;
										}
									}
								}

								if (!qualifierCorrect) {
									correct = false;
								}

							}
						}
					}
					if (correct) {
						correctCount++;
					}
				}
			}
		}
		return correctCount;
	}

	@Deprecated
	public FusionPerformance<Integer> getMaxCorrect(GroupedValues groupedValues, Object object, PropertiesConfig defaultConfig) {
		throw new UnsupportedOperationException();
	}

	@Deprecated
	public FusionPerformance<Integer> getMaxCorrect(SplitGroupedValues sgvTesting, Object object, PropertiesConfig propertiesConfig) {
		throw new UnsupportedOperationException();
	}

}
