package de.mannheim.uni.expansion.tablefusion.analysis.stores;

import java.util.Collections;
import java.util.Map;
import java.util.OptionalDouble;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;

public class InputProfileStoreCreator {
	private final AtomicInteger totalTables = new AtomicInteger();
	private final ConcurrentHashMap<ClassId, AtomicInteger> classCount = new ConcurrentHashMap<ClassId, AtomicInteger>();
	private final ConcurrentHashMap<PropertyId, AtomicInteger> propertyCount = new ConcurrentHashMap<PropertyId, AtomicInteger>();
	private final ConcurrentHashMap<InstanceId, AtomicInteger> instanceCount = new ConcurrentHashMap<InstanceId, AtomicInteger>();

	private final ConcurrentHashMap<ClassPropertyId, AtomicInteger> compositeTableCount = new ConcurrentHashMap<ClassPropertyId, AtomicInteger>();
	private final ConcurrentHashMap<ClassPropertyId, AtomicInteger> compositeValueCount = new ConcurrentHashMap<ClassPropertyId, AtomicInteger>();
	private final ConcurrentHashMap<ClassPropertyId, ConcurrentHashMap<InstanceId, AtomicInteger>> compositeInstanceCount = new ConcurrentHashMap<ClassPropertyId, ConcurrentHashMap<InstanceId, AtomicInteger>>();

	public <E> void increment(ConcurrentHashMap<E, AtomicInteger> countMap, E key, int value) {
		AtomicInteger ai = countMap.get(key);
		if (ai != null) {
			ai.addAndGet(value);
		} else {
			ai = new AtomicInteger(value);
			AtomicInteger ret = countMap.putIfAbsent(key, ai);
			if (ret != null) {
				countMap.get(key).addAndGet(value);
			}
		}
	}

	public void increment(ClassId dbpediaClass) {
		increment(classCount, dbpediaClass, 1);
	}

	public void increment(PropertyId dbpediaProperty) {
		increment(propertyCount, dbpediaProperty, 1);
	}

	public void increment(InstanceId dbpediaInstnace) {
		increment(instanceCount, dbpediaInstnace, 1);
	}

	public void incrementTableCount(ClassPropertyId compositeKey) {
		increment(compositeTableCount, compositeKey, 1);
	}

	public void incrementValueCount(ClassPropertyId compositeKey, int value) {
		increment(compositeValueCount, compositeKey, value);
	}

	public void incrementCompositeInstanceValueCount(ClassPropertyId compositeKey, InstanceId instanceId) {
		ConcurrentHashMap<InstanceId, AtomicInteger> ai = compositeInstanceCount.get(compositeKey);
		if (ai == null) {
			ai = new ConcurrentHashMap<InstanceId, AtomicInteger>();
			ConcurrentHashMap<InstanceId, AtomicInteger> ret = compositeInstanceCount.putIfAbsent(compositeKey, ai);
			if (ret != null) {
				ai = compositeInstanceCount.get(compositeKey);
			}
		}

		increment(ai, instanceId, 1);
	}

	public void incrementTotalTables() {
		totalTables.incrementAndGet();

	}

	public InputProfileStore get() {
		return new InputProfileStore();
	}

	public class InputProfileStore {

		private final Map<ClassId, Integer> classCount;
		private final Map<PropertyId, Integer> propertyCount;
		private final Map<InstanceId, Integer> instanceCount;
		private final Map<ClassPropertyId, Integer> compositeTableCount;
		private final Map<ClassPropertyId, Integer> compositeValueCount;
		private Map<ClassPropertyId, Map<InstanceId, Integer>> compositeInstanceCount;

		private InputProfileStore() {
			this.classCount = convertToInt(InputProfileStoreCreator.this.classCount);
			this.propertyCount = convertToInt(InputProfileStoreCreator.this.propertyCount);
			this.instanceCount = convertToInt(InputProfileStoreCreator.this.instanceCount);
			this.compositeTableCount = convertToInt(InputProfileStoreCreator.this.compositeTableCount);
			this.compositeValueCount = convertToInt(InputProfileStoreCreator.this.compositeValueCount);
			this.compositeInstanceCount = InputProfileStoreCreator.this.compositeInstanceCount.entrySet().stream()
					.collect(Collectors.toMap(Map.Entry::getKey, (e) -> {
						return convertToInt(e.getValue());
					}));
		}

		public int getTotalTables() {
			return totalTables.get();
		}

		public Map<ClassId, Integer> getClassCount() {
			return Collections.unmodifiableMap(classCount);
		}

		public Map<PropertyId, Integer> getPropertyCount() {
			return Collections.unmodifiableMap(propertyCount);
		}

		public Map<InstanceId, Integer> getInstanceCount() {
			return Collections.unmodifiableMap(instanceCount);
		}

		public Map<ClassPropertyId, Integer> getCompositeTableCount() {
			return Collections.unmodifiableMap(compositeTableCount);
		}

		public int getCompositeTableCount(ClassPropertyId classPropertyId) {
			return compositeTableCount.get(classPropertyId);
		}

		public Map<ClassPropertyId, Integer> getCompositeValueCount() {
			return Collections.unmodifiableMap(compositeValueCount);
		}

		public int getCompositeValueCount(ClassPropertyId classPropertyId) {
			return compositeValueCount.get(classPropertyId);
		}

		public Set<ClassPropertyId> getClassPropertyIds() {
			return Collections.unmodifiableSet(compositeInstanceCount.keySet());
		}

		public Map<InstanceId, Integer> getCompositeInstanceCount(ClassPropertyId classPropertyId) {
			return Collections.unmodifiableMap(compositeInstanceCount.get(classPropertyId));
		}

		private <T> Map<T, Integer> convertToInt(Map<T, AtomicInteger> map) {
			return map.entrySet().stream().collect(Collectors.toMap(Map.Entry::getKey, (e) -> {
				return e.getValue().get();
			}));
		}

		public OptionalDouble getAverageNumberOfInstances(ClassPropertyId key) {
			return compositeInstanceCount.get(key).entrySet().stream().mapToInt(d -> {
				return d.getValue();
			}).average();
		}

		public OptionalDouble getMedianNumberOfInstances(ClassPropertyId key) {

			int[] valueList = compositeInstanceCount.get(key).entrySet().stream().mapToInt(d -> {
				return d.getValue();
			}).sorted().toArray();
			if (valueList.length > 0) {
				double median;
				if (valueList.length % 2 == 0) {
					int idx1 = valueList.length / 2;
					int idx2 = idx1 - 1;
					median = ((double) valueList[idx1] + valueList[idx2]) / 2.0;
				} else {
					int idx1 = valueList.length / 2;
					median = valueList[idx1];
				}

				return OptionalDouble.of(median);
			} else {
				return OptionalDouble.empty();
			}

		}

		public int getCompositeInstanceCountSize(ClassPropertyId key) {
			return compositeInstanceCount.get(key).size();
		}

	}
}
