package de.mannheim.uni.expansion.tablefusion.analysis.writer.generic;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;
import java.util.Map.Entry;
import java.util.function.Function;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.AbstractResultsWriter;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.ResultsCategory;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnAcumValueGeneric;

public class GenericResultsWriter<E> extends AbstractResultsWriter {
	private final ColumnList<E> columnList;

	public GenericResultsWriter(ColumnList<E> columnList) {
		super(getHeaderLine(columnList.getColumnHeaders()));
		this.columnList = columnList;

	}

	private static String[] getHeaderLine(List<String> headers) {
		final String[] headerLine;

		headerLine = new String[headers.size()];
		for (int i = 0; i < headers.size(); i++) {
			headerLine[i] = headers.get(i);
		}

		return headerLine;
	}

	public void writeTotal(String name, FusionPerformance<E> performance) {
		writeTotal(name, performance, Function.identity());
	}

	public <X> void writeTotal(String name, FusionPerformance<X> performance, Function<X, E> converter) {
		String[] prefix = ResultsCategory.getNormalPrefix(name);
		write(ResultsCategory.TOTAL, prefix, converter.apply(performance.getTotal()));
	}

	public void writeTotal(String name, E data) {
		String[] prefix = ResultsCategory.getNormalPrefix(name);
		write(ResultsCategory.TOTAL, prefix, data);
	}

	public void writeTotalWithLearner(String run, String learner, E data) {
		String[] prefix = ResultsCategory.getTotalWithLearnerPrefix(run, learner);
		write(ResultsCategory.TOTAL_WITH_LEARNER, prefix, data);
	}

	public void writeClassId(String name, FusionPerformance<E> performance) {
		writeClassId(name, performance, Function.identity());
	}

	public <X> void writeClassId(String name, FusionPerformance<X> performance, Function<X, E> converter) {
		for (Entry<ClassId, X> e : performance.getPerClassId().entrySet()) {
			String[] prefix = ResultsCategory.getClassIdPrefix(name, e.getKey());
			write(ResultsCategory.CLASS_ID, prefix, converter.apply(e.getValue()));
		}
	}

	public void writeClassId(String name, ClassId classId, E data) {

		String[] prefix = ResultsCategory.getClassIdPrefix(name, classId);
		write(ResultsCategory.CLASS_ID, prefix, data);

	}

	public void writeClassIdWithLearner(String run, String learner, ClassId classId, E data) {

		String[] prefix = ResultsCategory.getClassIdWithLearnerPrefix(run, learner, classId);
		write(ResultsCategory.CLASS_ID_WITH_LEARNER, prefix, data);

	}

	public void writeClassPropertyId(String name, ClassId classId, PropertyId propertyId, E data) {

		String[] prefix = ResultsCategory.getClassPropertyIdSimplePrefix(name, new ClassPropertyId(classId, propertyId));
		write(ResultsCategory.CLASS_PROPERTY_ID_SIMPLE, prefix, data);

	}

	public void writeClassIdNoSet(ClassId classId, E data) {

		String[] prefix = ResultsCategory.getClassIdNoSetPrefix(classId);
		write(ResultsCategory.CLASS_ID_NO_SET, prefix, data);

	}

	// public void writeClassIdAndSet(String name,
	// FusionPerformance<GenericAccumaltableResult> performance) {
	// for (Entry<Pair<ClassId, String>, GenericAccumaltableResult> e :
	// performance.getPerClassAndSet().entrySet()) {
	// String[] prefix = ResultsCategory.getClassIdAndSetPrefix(name,
	// e.getKey().getFirst(), e.getKey().getSecond());
	// write(ResultsCategory.CLASS_ID_AND_SET, prefix, e.getValue());
	// }
	// }

	public <X> void write(String name, OntologyRepository ontologyRepository, FusionPerformance<E> performance) {
		write(name, ontologyRepository, performance, Function.identity());
	}

	public <X> void write(String name, OntologyRepository ontologyRepository, FusionPerformance<X> performance, Function<X, E> converter) {
		{
			String[] prefix = ResultsCategory.getNormalPrefix(name);
			write(ResultsCategory.TOTAL, prefix, converter.apply(performance.getTotal()));
		}

		for (Entry<Boolean, X> e : performance.getPerTimeDependence().entrySet()) {
			String[] prefix = ResultsCategory.getTimeDepedence(name, e.getKey());
			write(ResultsCategory.TIME_DEPENDENCE, prefix, converter.apply(e.getValue()));
		}

		for (Entry<ValueType, X> e : performance.getPerDataType().entrySet()) {
			String[] prefix = ResultsCategory.getDataTypePrefix(name, e.getKey());
			write(ResultsCategory.DATA_TYPE, prefix, converter.apply(e.getValue()));
		}

		for (Entry<ClassId, X> e : performance.getPerClassId().entrySet()) {
			String[] prefix = ResultsCategory.getClassIdPrefix(name, e.getKey());
			write(ResultsCategory.CLASS_ID, prefix, converter.apply(e.getValue()));
		}

		for (Entry<ClassPropertyId, X> e : performance.getPerProperty().entrySet()) {
			ValueType cdt = ontologyRepository.getDataType(e.getKey().getPropertyId());
			String[] prefix = ResultsCategory.getClassPropertyIdPrefix(name, ontologyRepository.isTimeDependent(e.getKey().getPropertyId()), cdt, e.getKey());
			write(ResultsCategory.CLASS_PROPERTY_ID, prefix, converter.apply(e.getValue()));
		}

	}

	private void write(ResultsCategory rc, String[] prefix, E result) {
		List<ColumnAcumValueGeneric<E>> valueGetters = this.columnList.getColumnValueGetters();

		String[] line = new String[this.columnList.getColumnSize()];
		for (int i = 0; i < this.columnList.getColumnSize(); i++) {
			line[i] = valueGetters.get(i).getValue(result);

		}

		write(rc, prefix, line);
	}

	@Override
	protected String[] getHeaderLine() {
		return null;
	}

	public void writeToFolder(String folder, String fileNamePrefix) throws IOException {

		for (ResultsCategory rc : ResultsCategory.values()) {
			String val = this.getAll(rc);

			FileWriter fw = new FileWriter(new File(folder + "/" + fileNamePrefix + "_" + rc.name() + ".csv"));
			fw.write(val);
			fw.close();
		}
	}

	public void writeToFolder(ResultsCategory rc, String folder, String fileNamePrefix) throws IOException {

		String val = this.getAll(rc);

		FileWriter fw = new FileWriter(new File(folder + "/" + fileNamePrefix + "_" + rc.name() + ".csv"));
		fw.write(val);
		fw.close();
	}

	// public void writeToFiles(String outputFolder, String outputPrefix) throws
	// IOException {
	//
	// MKDIR.createDirUnchecked(new File(outputFolder));
	//
	// String output = getAll(ResultsCategory.TOTAL);
	// FileUtils.writeStringToFile(new File(outputFolder + "/" + outputPrefix +
	// "_TOTAL.csv"), output);
	//
	// output = getAll(ResultsCategory.TIME_DEPENDENCE);
	// FileUtils.writeStringToFile(new File(outputFolder + "/" + outputPrefix +
	// "_TIME_DEPENDENCE.csv"), output);
	//
	// output = getAll(ResultsCategory.DATA_TYPE);
	// FileUtils.writeStringToFile(new File(outputFolder + "/" + outputPrefix +
	// "_DATA_TYPE.csv"), output);
	//
	// output = getAll(ResultsCategory.CLASS_ID);
	// FileUtils.writeStringToFile(new File(outputFolder + "/" + outputPrefix +
	// "_CLASS.csv"), output);
	//
	// output = getAll(ResultsCategory.CLASS_PROPERTY_ID);
	// FileUtils.writeStringToFile(new File(outputFolder + "/" + outputPrefix +
	// "_CLASS_PROPERTY.csv"), output);
	//
	// }

	// public static ResultBuilder builder(ResultsType resultsType) {
	// return new ResultBuilder(resultsType);
	// }
	//
	// public static class ResultBuilder {
	// private final HashMap<Integer, String> headers = new HashMap<>();
	// private final ResultsType resultsType = resultsType;
	//
	// private ResultBuilder(ResultsType resultsType) {
	// this.resultsType
	// }
	//
	// public void setColumndHeader(Integer key, String value) {
	// headers.put(key, value);
	// }
	//
	// public GenericResultsWriter build() {
	// return new GenericResultsWriter(headers);
	// }
	// }
	//
	// @Override
	// protected String[] getHeaderLine() {
	// // TODO Auto-generated method stub
	// return null;
	// }

}
