package de.mannheim.uni.expansion.tablefusion.core.scoring;

import java.util.EnumMap;
import java.util.List;
import java.util.Map;

import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;

public class Probability {

	private final EnumMap<SimilarityType, Accuracy> map;

	public Probability(Probability[] probabilities, Double[] weights) {
		map = new EnumMap<>(SimilarityType.class);
		for (SimilarityType type : SimilarityType.values()) {

			double appliedWeights = 0.0;
			double accuracy = 0.0;
			int total = 0;

			for (int i = 0; i < probabilities.length; i++) {
				if (probabilities[i] != null) {
					accuracy += probabilities[i].get(type).getAccuracy() * weights[i];
					total += probabilities[i].get(type).getTotal() * weights[i];
					appliedWeights += weights[i];
				}
			}

			accuracy = accuracy * (1.0 / (1.0 - appliedWeights));
			total = (int) (total * (1.0 / (1.0 - appliedWeights)));
			map.put(type, new Accuracy(accuracy, total));
		}
	}

	public Probability(Probability[] probabilities, Map<SimilarityType, Double[]> weights) {

		this(probabilities, weights, null, 0.0);
	}

	public Probability(Probability[] probabilities, Map<SimilarityType, Double[]> weights, Map<SimilarityType, Double> intercept, Double correction) {
		map = new EnumMap<>(SimilarityType.class);
		for (SimilarityType type : SimilarityType.values()) {

			double accuracy = 0.0;
			int total = 0;

			for (int i = 0; i < probabilities.length; i++) {
				Double probabilityDouble = probabilities[i].get(type).getAccuracy();
				Double weightsDouble = weights.get(type)[i];
				accuracy += probabilityDouble * weightsDouble;
				total += probabilities[i].get(type).getTotal() * weightsDouble;
			}

			if (intercept != null)
				accuracy += intercept.get(type);
			if (correction != null) {
				accuracy += correction;
			}

			map.put(type, new Accuracy(accuracy, total));
		}
	}

	public Probability(Map<SimilarityType, Accuracy> accuracies) {
		map = new EnumMap<>(accuracies);
	}

	public Probability(Accuracy accuracy) {
		map = new EnumMap<>(SimilarityType.class);
		for (SimilarityType type : SimilarityType.values()) {
			map.put(type, accuracy);
		}
	}

	public Probability(List<Probability> values) {
		EnumMap<SimilarityType, Accuracy> newMap = new EnumMap<>(SimilarityType.class);
		for (SimilarityType type : SimilarityType.values()) {
			double sum = 0.0;
			int totalSum = 0;
			int count = 0;
			for (Probability value : values) {
				Accuracy v = value.getOr(type, new Accuracy(0.0, 0));
				sum += v.getAccuracy();
				totalSum += v.getTotal();
				count++;
			}

			newMap.put(type, new Accuracy(sum / count, totalSum / count));
		}
		map = newMap;
	}

	public static Probability sumUp(List<Probability> values) {
		EnumMap<SimilarityType, Accuracy> newMap = new EnumMap<>(SimilarityType.class);
		for (SimilarityType type : SimilarityType.values()) {
			double sum = 0.0;
			int totalSum = 0;
			int count = 0;
			for (Probability value : values) {
				Accuracy v = value.getOr(type, new Accuracy(0.0, 0));
				sum += v.getAccuracy();
				totalSum += v.getTotal();
				count++;
			}

			newMap.put(type, new Accuracy(sum, totalSum / count));
		}

		return new Probability(newMap);
	}

	public Probability(Probability[] values, double[] weights, double intercept) {
		EnumMap<SimilarityType, Accuracy> newMap = new EnumMap<>(SimilarityType.class);
		for (SimilarityType type : SimilarityType.values()) {
			double probabilityWeighted = 0.0;
			double totalWeighted = 0;

			for (int i = 0; i < values.length; i++) {
				Probability value = values[i];
				if (value != null) {
					Accuracy v = value.getOr(type, new Accuracy(0.0, 0));
					probabilityWeighted += v.getAccuracy() * weights[i];
					totalWeighted += v.getTotal() * weights[i];
				}
			}

			newMap.put(type, new Accuracy(probabilityWeighted + intercept, (int) totalWeighted));
		}
		map = newMap;
	}

	public Accuracy get(SimilarityType type) {
		return map.get(type);
	}

	public Accuracy getOr(SimilarityType type, Accuracy alternative) {
		Accuracy value = this.get(type);
		if (value == null || Double.isNaN(value.getAccuracy()))
			return alternative;
		else
			return value;
	}

	public Probability multiply(Probability other) {
		Map<SimilarityType, Accuracy> newMap = new EnumMap<>(SimilarityType.class);
		for (SimilarityType type : SimilarityType.values()) {
			Accuracy d1 = this.getOr(type, new Accuracy(1.0, 0));

			Accuracy d2 = other.getOr(type, new Accuracy(1.0, 0));

			newMap.put(type, new Accuracy(d1.getAccuracy() * d2.getAccuracy(), (d1.getTotal() + d2.getTotal()) / 2));
		}
		return new Probability(newMap);
	}

	// public Probability weightedMultiply(Probability other) {
	// Map<SimilarityType, Accuracy> newMap = new
	// EnumMap<>(SimilarityType.class);
	// for (SimilarityType type : SimilarityType.values()) {
	// Accuracy d1 = this.getOr(type, new Accuracy(0.0, 0));
	//
	// Accuracy d2 = other.getOr(type, new Accuracy(0.0, 0));
	//
	// newMap.put(type, new Accuracy(d1.getAccuracy() * d2.getAccuracy(),
	// (d1.getTotal() + d2.getTotal()) / 2));
	// }
	// return new Probability(newMap);
	// }

	public Probability simpleWeightedMultiply(Probability other) {
		Map<SimilarityType, Accuracy> newMap = new EnumMap<>(SimilarityType.class);
		for (SimilarityType type : SimilarityType.values()) {
			Accuracy ac1 = this.getOr(type, new Accuracy(0.0, 0));

			Accuracy ac2 = other.getOr(type, new Accuracy(0.0, 0));

			double d1 = ac1.getAccuracy();
			double d2 = ac2.getAccuracy();

			newMap.put(type, new Accuracy(d1 * 0.75 * d2 + 0.25 * d1, (ac1.getTotal() + ac2.getTotal()) / 2));
		}
		return new Probability(newMap);
	}

	public Probability resize(Probability other) {
		Map<SimilarityType, Accuracy> newMap = new EnumMap<>(SimilarityType.class);
		for (SimilarityType type : SimilarityType.values()) {
			Accuracy ac1 = this.getOr(type, new Accuracy(0.0, 0));

			Accuracy ac2 = other.getOr(type, new Accuracy(0.0, 0));

			double d1 = ac1.getAccuracy();
			double d2 = ac2.getAccuracy();

			double dNew = (d1 > 0.5 ? (1.0 - d1) * (d2 - 0.5) : d1 * (d2 - 0.5)) + d1;

			newMap.put(type, new Accuracy(dNew, (ac1.getTotal() + ac2.getTotal()) / 2));
		}
		return new Probability(newMap);
	}

	@Override
	public String toString() {
		return "Probability [map=" + map + "]";
	}

	public Probability weightedMultiply(Double mainWeight, Probability[] secondaryProbability, Double[] secondaryWeights) {
		Map<SimilarityType, Accuracy> newMap = new EnumMap<>(SimilarityType.class);
		for (SimilarityType type : SimilarityType.values()) {

			double newAcc = 0.0;
			double newTotal = 0.0;

			Accuracy thisAc = this.getOr(type, new Accuracy(0.0, 0));
			newAcc += mainWeight * thisAc.getAccuracy();
			newTotal += mainWeight * thisAc.getTotal();

			for (int i = 0; i < secondaryProbability.length; i++) {
				if (secondaryProbability[i] == null) {
					newAcc += secondaryWeights[i] * thisAc.getAccuracy();
					newTotal += secondaryWeights[i] * thisAc.getTotal();
				} else {
					Accuracy otherAcc = secondaryProbability[i].getOr(type, new Accuracy(0.0, 0));
					newAcc += secondaryWeights[i] * otherAcc.getAccuracy() * thisAc.getAccuracy();
					newTotal += secondaryWeights[i] * otherAcc.getTotal();
				}
			}

			newMap.put(type, new Accuracy(newAcc, (int) newTotal));
		}
		return new Probability(newMap);
	}

}
