package de.mannheim.uni.expansion.tablefusion.core.scoring;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableDouble;
import org.apache.commons.lang.mutable.MutableInt;

import com.google.common.collect.ImmutableMap;

import de.mannheim.uni.expansion.data.ontology.model.PropertyIdContainer;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.fusing.BaseFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.FuserRepository;
import de.mannheim.uni.expansion.tablefusion.core.fusing.ScoredEquivalenceClusterer;
import de.mannheim.uni.expansion.tablefusion.core.fusing.ScoredEquivalenceClusterer.Cluster;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.NonQualifiedColumnScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.QualifiedColumnScores;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleConcurrentCache;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;

public class QualifiedKbtCalculator {

	private static SimpleConcurrentCache<UUID, QualifiedColumnScores> cache = new SimpleConcurrentCache<>();
	private static SimpleConcurrentCache<UUID, NonQualifiedColumnScores> nonQualifiedCache = new SimpleConcurrentCache<>();

	private final AnnotatedTablesRepository repository;
	private final KnowledgeBase knowledgeBase;
	private final SimilarityCalculator similarityCalculator;
	private final ProbabilitiesCalculator kbtScorer;
	private final TimestampQualifiersAdapter qualifierAdapter;
	private final ScoredEquivalenceClusterer equivalenceClusterer;
	private final FuserRepository fuserRepository;

	@Inject
	public QualifiedKbtCalculator(AnnotatedTablesRepository tableRepository, KnowledgeBase knowledgeBase, SimilarityCalculator similarityCalculator,
			ProbabilitiesCalculator kbtScorer, TimestampQualifiersAdapter qualifierAdapter, ScoredEquivalenceClusterer equivalenceClusterer,
			FuserRepository fuserRepository) {
		super();
		this.repository = tableRepository;
		this.knowledgeBase = knowledgeBase;
		this.similarityCalculator = similarityCalculator;
		this.kbtScorer = kbtScorer;
		this.qualifierAdapter = qualifierAdapter;
		this.equivalenceClusterer = equivalenceClusterer;
		this.fuserRepository = fuserRepository;
	}

	public QualifiedColumnScores byColumnIdQualified(SplitGroupedValues learningSet) {
		if (cache.hasCacheFor(learningSet.getUUID())) {
			return cache.get(learningSet.getUUID());
		} else {

			ConcurrentHashMap<TableColumnId, Map<QualifiedPropertyId, Probability>> probabilties = new ConcurrentHashMap<>();
			Map<TableColumnId, Map<QualifiedTripleId, SingleValue>> grouped = groupByColumnId(learningSet);

			PersistentParallel4.runAndFinish(grouped.entrySet(), (e) -> {

				ConcurrentHashMap<QualifiedPropertyId, AtomicInteger> total = new ConcurrentHashMap<>();
				ConcurrentHashMap<QualifiedPropertyId, AtomicInteger> correct = new ConcurrentHashMap<>();

				MutableInt totalOverlap = new MutableInt(0);
				int sdf = 1;
				PersistentParallel6.runAndFinish(e.getValue().entrySet(), (sv) -> {
					QualifiedValueList a = knowledgeBase.getQualifiedValueListWithEqualProperties(sv.getKey().getTripleId().getValueId());

					if (a != null) {
						Set<AbstractObject> objects = qualifierAdapter.getForQualifiers(a, sv.getKey().getQualifiers());

						if (objects != null && !objects.isEmpty()) {
							CHMHelper.getFromCHM(total, sv.getKey().getQualifiedPropertyId(), () -> new AtomicInteger()).incrementAndGet();
							AtomicInteger correctInt = CHMHelper.getFromCHM(correct, sv.getKey().getQualifiedPropertyId(), () -> new AtomicInteger());

							for (AbstractObject ao : objects) {
								if (similarityCalculator.isEqual(ao, sv.getValue().getValue())) {
									correctInt.incrementAndGet();
									break;
								}
							}
						}

					}
				}, this.getClass(), e.getValue().size() < 1000000);

				for (Entry<QualifiedPropertyId, AtomicInteger> totalEntry : total.entrySet()) {
					Accuracy acc = Accuracy.fromCorrectAndTotal(correct.get(totalEntry.getKey()).intValue(), totalEntry.getValue().intValue());
					Probability prob = new Probability(acc);
					CHMHelper.getFromCHM(probabilties, e.getKey(), () -> new ConcurrentHashMap<>()).put(totalEntry.getKey(), prob);
				}

			}, "QualifiedKBT");

			QualifiedColumnScores scores = new QualifiedColumnScores(probabilties);
			cache.add(learningSet.getUUID(), scores);

			return scores;
		}
	}

	public NonQualifiedColumnScores byColumnIdNonQualified(SplitGroupedValues learningSet) {
		if (nonQualifiedCache.hasCacheFor(learningSet.getUUID())) {
			return nonQualifiedCache.get(learningSet.getUUID());
		} else {

			ConcurrentHashMap<TableColumnId, Map<PropertyId, Probability>> probabilties = new ConcurrentHashMap<>();
			Map<TableColumnId, Map<QualifiedTripleId, SingleValue>> grouped = groupByColumnId(learningSet);

			PersistentParallel4.runAndFinish(grouped.entrySet(), (e) -> {

				ConcurrentHashMap<PropertyId, AtomicInteger> total = new ConcurrentHashMap<>();
				ConcurrentHashMap<PropertyId, AtomicInteger> correct = new ConcurrentHashMap<>();

				MutableInt totalOverlap = new MutableInt(0);
				int sdf = 1;

				PersistentParallel6.runAndFinish(e.getValue().entrySet(), (sv) -> {
					QualifiedValueList a = knowledgeBase.getQualifiedValueListWithEqualProperties(sv.getKey().getTripleId().getValueId());

					if (a != null && a.size() > 0) {
						CHMHelper.getFromCHM(total, sv.getKey().getQualifiedPropertyId().getPropertyId(), () -> new AtomicInteger()).incrementAndGet();
						AtomicInteger correctInt = CHMHelper.getFromMap(correct, sv.getKey().getQualifiedPropertyId().getPropertyId(),
								() -> new AtomicInteger());

						for (QualifiedValueHolder claim : a.getClaims()) {
							AbstractObject ao = claim.getValue();
							if (similarityCalculator.isEqual(ao, sv.getValue().getValue())) {
								correctInt.incrementAndGet();
								break;
							}
						}
					}

				}, this.getClass().getName() + "_A", e.getValue().size() < 1000000);

				for (Entry<PropertyId, AtomicInteger> totalEntry : total.entrySet()) {
					Accuracy acc = Accuracy.fromCorrectAndTotal(correct.get(totalEntry.getKey()).intValue(), totalEntry.getValue().intValue());
					Probability prob = new Probability(acc);
					CHMHelper.getFromCHM(probabilties, e.getKey(), () -> new ConcurrentHashMap<>()).put(totalEntry.getKey(), prob);
				}

			}, "QualifiedKBT");

			NonQualifiedColumnScores scores = new NonQualifiedColumnScores(probabilties);
			nonQualifiedCache.add(learningSet.getUUID(), scores);

			return scores;
		}
	}

	private static SimpleConcurrentCache<SplitGroupedValues, QualifiedColumnScores> cachePropgated = new SimpleConcurrentCache<>();
	private static SimpleConcurrentCache<SplitGroupedValues, NonQualifiedColumnScores> nonQualifiedCachePropgated = new SimpleConcurrentCache<>();

	public QualifiedColumnScores getQualifiedPropogated(SplitGroupedValues learningSet) {
		if (cachePropgated.hasCacheFor(learningSet)) {
			return cachePropgated.get(learningSet);
		} else {

			QualifiedColumnScores scores = byColumnIdQualified(learningSet);

			Map<TableColumnId, Map<QualifiedPropertyId, Probability>> newFullScores = getPropgation(learningSet, scores.getFullScores(), (e) -> {
				return e;
			});
			QualifiedColumnScores newScores = new QualifiedColumnScores(newFullScores);
			cachePropgated.add(learningSet, newScores);

			return scores;
		}

	}

	public NonQualifiedColumnScores getNonQualifiedPropogated(SplitGroupedValues learningSet) {
		if (nonQualifiedCachePropgated.hasCacheFor(learningSet)) {
			return nonQualifiedCachePropgated.get(learningSet);
		} else {

			NonQualifiedColumnScores scores = byColumnIdNonQualified(learningSet);

			Map<TableColumnId, Map<PropertyId, Probability>> newFullScores = getPropgation(learningSet, scores.getFullScores(), (e) -> {
				return e.getPropertyId();
			});

			NonQualifiedColumnScores newScores = new NonQualifiedColumnScores(newFullScores);
			nonQualifiedCachePropgated.add(learningSet, newScores);

			return newScores;
		}

	}

	private <E extends PropertyIdContainer> Map<TableColumnId, Map<E, Probability>> getPropgation(SplitGroupedValues groupedValues,
			Map<TableColumnId, Map<E, Probability>> fullScores, Function<QualifiedPropertyId, E> getPropertyId) {

		Map<TableColumnId, Map<QualifiedTripleId, SingleValue>> grouped = groupByColumnId(groupedValues);

		// Map<E, Set<TableColumnId>> mapping = getPropertyToColumn(fullScores);

		Map<TableColumnId, Map<E, Probability>> currentScores = fullScores;

		for (int i = 0; i < 3; i++) {
			ConcurrentHashMap<QualifiedTripleId, List<Cluster>> clusterMap = new ConcurrentHashMap<>();

			ConcurrentHashMap<TableColumnId, Map<E, Probability>> probabilties = new ConcurrentHashMap<>();
			Map<TableColumnId, Map<E, Probability>> currentScoresFinal = currentScores;

			PersistentParallel6.runAndFinish(groupedValues.entrySet(), (e) -> {
				SimilarityMeasure sm = similarityCalculator.getMeasure(e.getValue().getDataType());
				BaseFuser fuser = this.fuserRepository.getBaseFuser(e.getValue().getDataType());
				List<Cluster> clusterList = new ArrayList<>();
				for (SingleValue sv : e.getValue().getValues()) {
					Map<E, Probability> columnScores = currentScoresFinal.get(sv.getId().getTableColumnId());
					if (columnScores != null) {
						Probability score = columnScores.get(getPropertyId.apply(e.getKey().getQualifiedPropertyId()));
						if (score != null) {
							this.equivalenceClusterer.addValue(clusterList, sm, fuser, sv.getValue(), score.get(SimilarityType.DIRECT).getAccuracy());
						}
					}
				}

				clusterMap.put(e.getKey(), clusterList);

			}, this.getClass().getName() + "_B");

			PersistentParallel6.runAndFinish(grouped.entrySet(), (e) -> {

				Map<E, MutableInt> total = new HashMap<>();
				Map<E, MutableDouble> score = new HashMap<>();

				MutableInt totalOverlap = new MutableInt(0);
				int sdf = 1;
				for (Entry<QualifiedTripleId, SingleValue> sv : e.getValue().entrySet()) {
					List<Cluster> clusterList = clusterMap.get(sv.getKey());

					for (Cluster cluster : clusterList) {
						if (similarityCalculator.isEqual(cluster.getValue(), sv.getValue().getValue())) {
							CHMHelper.getFromMap(total, getPropertyId.apply(sv.getKey().getQualifiedPropertyId()), () -> new MutableInt()).increment();
							CHMHelper.getFromMap(score, getPropertyId.apply(sv.getKey().getQualifiedPropertyId()), () -> new MutableDouble())
									.add(cluster.getScore());
						}
					}

				}

				for (Entry<E, MutableInt> totalEntry : total.entrySet()) {
					Accuracy acc = Accuracy.fromCorrectAndTotal(score.get(totalEntry.getKey()).doubleValue(), totalEntry.getValue().intValue());
					Probability prob = new Probability(acc);
					CHMHelper.getFromCHM(probabilties, e.getKey(), () -> new ConcurrentHashMap<>()).put(totalEntry.getKey(), prob);
				}

			}, "QualifiedKBT");

			currentScores = probabilties;

		}

		return currentScores;
	}

	private <E extends PropertyIdContainer> ConcurrentHashMap<TableColumnId, Map<E, Probability>> getPropgationOld(
			Map<TableColumnId, Map<E, Probability>> fullScores) {

		Map<E, Set<TableColumnId>> mapping = getPropertyToColumn(fullScores);

		// Map<TableColumnId, Map<QualifiedTripleId, SingleValue>> grouped =
		// groupByColumnId(learningSet);

		ConcurrentHashMap<TableColumnId, Map<E, Probability>> newFullScores = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(fullScores.entrySet(), (e) -> {
			ImmutableMap<RowIdx, AbstractObject> column = getColumnFromRepository(repository, e.getKey());

			for (E property : e.getValue().keySet()) {

				double total = 0;
				double correct = 0;
				for (Entry<RowIdx, AbstractObject> value : column.entrySet()) {
					InstanceId instanceId = repository.getMetaData(e.getKey().getTableId()).getInstance(value.getKey().getRowIdx().intValue());
					if (instanceId != null) {
						for (TableColumnId columnId : mapping.get(property)) {

							Optional<RowIdx> row = repository.getMetaData(columnId.getTableId()).getRowIdx(instanceId);
							if (row.isPresent()) {
								total += 1.0;

								AbstractObject secondary = getColumnFromRepository(repository, columnId).get(row.get());

								if (similarityCalculator.isEqual(value.getValue(), secondary)) {
									correct += fullScores.get(columnId).get(property).get(SimilarityType.DIRECT).getAccuracy();
								} else {
									correct -= fullScores.get(columnId).get(property).get(SimilarityType.DIRECT).getAccuracy();
								}

							}

						}
					}
				}
				Accuracy acc = null;
				if (correct < 0.0 || total == 0.0) {
					acc = new Accuracy(0.0, 0);
				}
				acc = Accuracy.fromCorrectAndTotal(correct, total);

				CHMHelper.getFromCHM(newFullScores, e.getKey(), () -> new ConcurrentHashMap<>()).put(property, new Probability(acc));
			}

		}, "PROPGATETION");

		return newFullScores;

	}

	private ImmutableMap<RowIdx, AbstractObject> getColumnFromRepository(AnnotatedTablesRepository repository, TableColumnId columnId) {
		return repository.getTable(columnId.getTableId()).getColumn(columnId.getColumnIdx());
	}

	private <E extends PropertyIdContainer> Map<E, Set<TableColumnId>> getPropertyToColumn(Map<TableColumnId, Map<E, Probability>> fullScores) {
		ConcurrentHashMap<E, Set<TableColumnId>> mapping = new ConcurrentHashMap<>();

		PersistentParallel4.runAndFinish(fullScores.entrySet(), (e) -> {
			for (E property : e.getValue().keySet()) {
				CHMHelper.getFromCHM(mapping, property, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e.getKey());
			}
		}, "SOMETHING");

		return mapping;
	}

	// public NonQualifiedColumnScores
	// getNonQualifiedPropogated(SplitGroupedValues learningSet) {
	//
	// }

	private Map<TableColumnId, Map<QualifiedTripleId, SingleValue>> groupByColumnId(SplitGroupedValues learningSet) {
		ConcurrentHashMap<TableColumnId, Map<QualifiedTripleId, SingleValue>> groupedValues = new ConcurrentHashMap<TableColumnId, Map<QualifiedTripleId, SingleValue>>();

		PersistentParallel4.runAndFinish(learningSet.entrySet(), (e) -> {

			PersistentParallel6.runAndFinish(e.getValue().getValues(), (sv) -> {
				ConcurrentHashMap<QualifiedTripleId, SingleValue> tableColumnIdMap = (ConcurrentHashMap<QualifiedTripleId, SingleValue>) CHMHelper
						.getFromCHM(groupedValues, new TableColumnId(sv.getId().getTableId(), sv.getId().getColumnIdx()), () -> new ConcurrentHashMap<>());

				tableColumnIdMap.put(e.getKey(), sv);
			}, this.getClass(), e.getValue().getValues().size() < 10000);

		}, "KBTCalculator - Group Values by TableColumnId");

		return groupedValues;
	}

	// public TimestampScores getTimestampScores(GroupedValues learningSet) {
	// if (cache.hasCacheFor(learningSet)) {
	// return cache.get(learningSet);
	// } else {
	//
	// Map<TableColumnId, Map<TripleId, SingleValue>> grouped =
	// groupByColumnId(learningSet);
	//
	// PersistentParallel4.runAndFinish(grouped.entrySet(), (e) -> {
	//
	// Map<TimestampIdentifier, Map<PropertyId, MutableInt>> total = new
	// HashMap<>();
	// Map<TimestampIdentifier, Map<PropertyId, MutableInt>> correct = new
	// HashMap<>();
	//
	// AnnotatedTable table = repository.getTable(e.getKey().getTableId());
	//
	// AnnotatedTableMetaData am =
	// repository.getMetaData(e.getKey().getTableId());
	//
	// List<ColumnIdx> timeColumns = table.getTimeColumns();
	// Map<TimestampIdentifier, Integer> tableYears =
	// repository.getMetaData(e.getKey().getTableId()).getTableYears();
	//
	// MutableInt totalOverlap = new MutableInt(0);
	// int sdf = 1;
	// for (Entry<TripleId, SingleValue> sv : e.getValue().entrySet()) {
	// QualifiedValueList a =
	// knowledgeBase.getQualifiedValueList(sv.getKey().getValueId());
	//
	// if (a != null) {
	// totalOverlap.increment();
	// Map<Qualifier, List<Integer>> qualifiers = a.getQualifiers();
	// for (Entry<Qualifier, List<Integer>> qEntry : qualifiers.entrySet()) {
	// Qualifier q = qEntry.getKey();
	//
	// PropertyId qualifierProperty = q.getQualifier();
	// AbstractObject qualifierValue = q.getQualifierValue();
	//
	// Integer year = TimeObject.getYear(qualifierValue);
	//
	// if (year != null) {
	//
	// Map<TimestampIdentifier, Integer> tableYearsTemp = new
	// HashMap<>(tableYears);
	//
	// tableYearsTemp.put(TimestampIdentifier.createPropagatedIdentifier(year),
	// year);
	//
	// for (ColumnIdx columnIdx : timeColumns) {
	// AbstractObject ao = table.getValue(sv.getValue().getId().getRowIdx(),
	// columnIdx);
	// Integer columnYear = TimeObject.getYear(ao);
	// if (columnYear != null) {
	// tableYearsTemp.put(TimestampIdentifier.createCascadedIdentifier(CascadedType.COLUMN_VALUE,
	// columnIdx.getColumnIdx(), 0),
	// columnYear);
	// }
	// }
	//
	// for (Entry<TimestampIdentifier, Integer> ty : tableYearsTemp.entrySet())
	// {
	// Map<PropertyId, MutableInt> totalMap = CHMHelper.getFromMap(total,
	// ty.getKey(), () -> new HashMap<>());
	// Map<PropertyId, MutableInt> correctMap = CHMHelper.getFromMap(correct,
	// ty.getKey(), () -> new HashMap<>());
	//
	// if (ty.getValue().equals(year)) {
	//
	// for (Integer qualifiedValueHolderIndex : qEntry.getValue()) {
	// QualifiedValueHolder qvh = a.getClaim(qualifiedValueHolderIndex);
	// CHMHelper.getFromMap(totalMap, qualifierProperty, () -> new
	// MutableInt()).increment();
	// MutableInt correctInt = CHMHelper.getFromMap(correctMap,
	// qualifierProperty, () -> new MutableInt());
	//
	// if (similarityCalculator.isEqual(qvh.getValue(),
	// sv.getValue().getValue())) {
	// correctInt.increment();
	//
	// }
	// }
	//
	// }
	// }
	// }
	//
	// }
	// }
	// }
	//
	// Map<Pair<TimestampIdentifier, PropertyId>, Accuracy> accuracies = new
	// HashMap<>();
	//
	// for (Entry<TimestampIdentifier, Map<PropertyId, MutableInt>> entry :
	// total.entrySet()) {
	//
	// for (Entry<PropertyId, MutableInt> innerEntry :
	// entry.getValue().entrySet()) {
	// Accuracy acc =
	// Accuracy.fromCorrectAndTotal(correct.get(entry.getKey()).get(innerEntry.getKey()).intValue(),
	// innerEntry.getValue().intValue());
	// accuracies.put(new Pair<>(entry.getKey(), innerEntry.getKey()), acc);
	//
	// }
	//
	// }
	//
	// Set<TimestampIdentifier> identifierSet = new HashSet<>();
	// Set<PropertyId> qualifierSet = new HashSet<>();
	//
	// Map<Pair<TimestampIdentifier, PropertyId>, Accuracy> harmonicMean =
	// accuracies.entrySet().stream()
	// .collect(Collectors.toMap(Entry::getKey, (acc) -> {
	// Accuracy acc1 = acc.getValue();
	// Accuracy acc2 = Accuracy.fromCorrectAndTotal(acc1.getCorrect(),
	// totalOverlap.intValue());
	//
	// return Accuracy.harmonicMean(acc1, acc2, acc1.getTotal());
	// }));
	//
	// LinkedHashMap<Pair<TimestampIdentifier, PropertyId>, Accuracy>
	// sortedAccuracies = MapSorter.sortHashMapByValuesD(harmonicMean);
	//
	// Map<TimestampIdentifier, Pair<PropertyId, Accuracy>> mappings = new
	// HashMap<>();
	// for (Entry<Pair<TimestampIdentifier, PropertyId>, Accuracy> entry :
	// sortedAccuracies.entrySet()) {
	// if (!identifierSet.contains(entry.getKey().getFirst()) &&
	// !qualifierSet.contains(entry.getKey().getSecond())) {
	// mappings.put(entry.getKey().getFirst(), new
	// Pair<>(entry.getKey().getSecond(), entry.getValue()));
	//
	// identifierSet.add(entry.getKey().getFirst());
	// qualifierSet.add(entry.getKey().getSecond());
	// }
	// }
	//
	// if (!mappings.isEmpty())
	// probabilties.put(e.getKey(), mappings);
	//
	// } , "TimeStampCalculator");
	//
	// TimestampScores scores = new TimestampScores(probabilties);
	// cache.add(learningSet, scores);
	//
	// return scores;
	// }
	// }
	//
	//
	//
	// public ScoresContainer getScores(GroupedValues learningSet, GroupedValues
	// testingSet) {
	//
	// TimestampScores timestampScores = this.getTimestampScores(learningSet);
	//
	// return new TimestampScoresContainer(timestampScores,
	// kbtScorer.byProperty(learningSet));
	// }
	//
	// private class TimestampScoresContainer implements ScoresContainer {
	//
	// private final TimestampScores timestampScores;
	// private final ColumnScores columnScores;
	//
	// public TimestampScoresContainer(TimestampScores timestampScores,
	// ColumnScores columnScores) {
	// this.timestampScores = timestampScores;
	// this.columnScores = columnScores;
	// }
	//
	// @Override
	// public TableScores getTableScores(ClassId classId, PropertyId propertyId,
	// InstanceId instanceId) {
	// return new TableScoresInterface(this, classId, propertyId, instanceId);
	// }
	//
	// @Override
	// public Probability getScore(ClassId classId, PropertyId propertyId,
	// InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx,
	// RowIdx rowIdx) {
	// return columnScores.getScore(classId, propertyId, instanceId, tableId,
	// columnIdx, rowIdx);
	// }
	//
	// @Override
	// public boolean containsKey(ClassId classId, PropertyId propertyId,
	// InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx,
	// RowIdx rowIdx) {
	// return columnScores.containsKey(classId, propertyId, instanceId, tableId,
	// columnIdx, rowIdx);
	// }
	//
	// public Probability getScore(TripleId tripleId, CellId cellId, Qualifier
	// qualifier) {
	// TimestampIdentifier tsi = timestampScores.getMappings(new
	// TableColumnId(cellId)).get(qualifier.getQualifier());
	// if (tsi != null) {
	//
	// Integer tableYear;
	// if (TimestampType.PROPAGATED.equals(tsi.getType())) {
	// tableYear = tsi.getPropogatedYear();
	// } else {
	// tableYear = repository.getTimeStamp(cellId, tsi);
	//
	// }
	//
	// if (TimeObject.getYear(qualifier.getQualifierValue()).equals(tableYear))
	// {
	// Accuracy acc = timestampScores.getScore(new TableColumnId(cellId),
	// qualifier.getQualifier(), tsi);
	// return new Probability(acc);
	// }
	//
	// }
	//
	// return columnScores.getScore(tripleId, cellId, qualifier);
	// }
	//
	// public boolean containsKey(TripleId tripleId, CellId cellId, Qualifier
	// qualifier) {
	// boolean ret = getScore(tripleId, cellId, qualifier) != null;
	// if (ret)
	// return true;
	// else
	// return columnScores.containsKey(tripleId, cellId, qualifier);
	//
	// }
	//
	// }

}
