package de.mannheim.uni.expansion.tablefusion.core.scoring;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.QualifierId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.fusing.BaseFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.FuserRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.scoring.YearEquivalenceClusterer.Cluster;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.timestamps.TimestampScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.NumericSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.ThreadSafeLocalCalendar;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseImpl;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;

public class TimeStampsScoringCalculator {

	public static enum TS_METHODS {
		CELL_YEAR,
		MAX_COLUMN_YEAR,
		MAX_TABLE_TILE,
		MAX_PAGE_TITLE,
		TABLE_YEAR_HIERARCHY,
		MOST_COMMON_TABLE_TEXT_YEAR,
		GET_HIERARICAL_YEAR,
		GET_FULL_HIERARICAL_YEAR;
	}

	public static class TS_METHOD extends Pair<TS_METHODS, Function<CellId, Integer>> {

		public TS_METHOD(TS_METHODS first, Function<CellId, Integer> second) {
			super(first, second);
		}

	}

	int minYear = 1950;
	int maxYear = 2016;
	double maxDiff = maxYear - minYear + 1.0;
	public final TS_METHOD CELL_YEAR = new TS_METHOD(TS_METHODS.CELL_YEAR, this::getCellYear);
	public final TS_METHOD MAX_COLUMN_YEAR = new TS_METHOD(TS_METHODS.MAX_COLUMN_YEAR, this::getMaxColumnYear);
	public final TS_METHOD MAX_TABLE_TILE = new TS_METHOD(TS_METHODS.MAX_TABLE_TILE, this::getMaxTableTitleYear);
	public final TS_METHOD MAX_PAGE_TITLE = new TS_METHOD(TS_METHODS.MAX_PAGE_TITLE, this::getMaxPageTitleYear);
	public final TS_METHOD TABLE_YEAR_HIERARCHY = new TS_METHOD(TS_METHODS.TABLE_YEAR_HIERARCHY, this::getTableYearHiearchy);
	public final TS_METHOD MOST_COMMON_TABLE_TEXT_YEAR = new TS_METHOD(TS_METHODS.MOST_COMMON_TABLE_TEXT_YEAR, this::getMostCommonTableTextYear);
	public final TS_METHOD GET_HIERARICAL_YEAR = new TS_METHOD(TS_METHODS.GET_HIERARICAL_YEAR, this::getHiearicalYear);
	public final TS_METHOD GET_FULL_HIERARICAL_YEAR = new TS_METHOD(TS_METHODS.GET_FULL_HIERARICAL_YEAR, this::getFullHiearicalYear);

	public List<TS_METHOD> getAllMethods() {
		return Arrays.asList(CELL_YEAR, MAX_COLUMN_YEAR, MAX_TABLE_TILE, MAX_PAGE_TITLE, TABLE_YEAR_HIERARCHY, MOST_COMMON_TABLE_TEXT_YEAR, GET_HIERARICAL_YEAR,
				GET_FULL_HIERARICAL_YEAR);
	}

	private final AnnotatedTablesRepository annotatedTablesRepository;
	private final NumericSimilarity numericSimilarity;
	private final KnowledgeBase dbpediaRepository;
	private final ThreadSafeLocalCalendar calendar;
	private final YearEquivalenceClusterer yearEquivalenceClusterer;
	private final FuserRepository fuserRepository;
	private final SimilarityCalculator similarityCalculator;
	private final TimestampQualifiersAdapter timestampQualifiersAdapter;
	private final OntologyRepository ontologyRepository;

	@Inject
	public TimeStampsScoringCalculator(AnnotatedTablesRepository annotatedTablesRepository, NumericSimilarity numericSimilarity,
			KnowledgeBase dbpediaRepository, ThreadSafeLocalCalendar calendar, YearEquivalenceClusterer yearEquivalenceClusterer,
			FuserRepository fuserRepository, SimilarityCalculator similarityCalculator, TimestampQualifiersAdapter timestampQualifiersAdapter,
			OntologyRepository ontologyRepository) {

		this.annotatedTablesRepository = annotatedTablesRepository;
		this.numericSimilarity = numericSimilarity;
		this.dbpediaRepository = dbpediaRepository;
		this.calendar = calendar;
		this.yearEquivalenceClusterer = yearEquivalenceClusterer;
		this.fuserRepository = fuserRepository;
		this.similarityCalculator = similarityCalculator;
		this.timestampQualifiersAdapter = timestampQualifiersAdapter;
		this.ontologyRepository = ontologyRepository;

	}

	private boolean yearInRange(Integer year) {
		if (year != null && year >= minYear && year <= maxYear)
			return true;
		else
			return false;
	}

	public Integer getMaxColumnYear(CellId aid) {

		AnnotatedTableMetaData md = annotatedTablesRepository.getMetaData(aid.getTableId());

		Integer year = null;

		if (md.getPropertiesMaxYear() != null) {
			Integer iYear = md.getPropertiesMaxYear().get(aid.getColumnIdx().getColumnIdx());

			if (yearInRange(year)) {
				if (year == null || iYear > year) {
					year = iYear;
				}
			}
		}

		return year;
	}

	public Integer getCellYear(CellId aid) {

		PropertyId activeYearsStartYear = new PropertyId("http://dbpedia.org/ontology/activeYearsStartYear");

		AnnotatedTableMetaData md = annotatedTablesRepository.getMetaData(aid.getTableId());
		AnnotatedTable td = annotatedTablesRepository.getTable(aid.getTableId());

		Integer year = null;
		for (int i = 0; i < td.getColumns().size(); i++) {
			if (td.getDataType(new ColumnIdx(i)) == ValueType.TIME) {
				PropertyId propertyId = md.getProperty(i);
				boolean addYear = false;
				if (propertyId == null)
					addYear = true;
				if (propertyId != null) {
					if (propertyId instanceof QualifierId) {
						QualifierId qualifierId = (QualifierId) propertyId;
						if (qualifierId.getQualifier().equals(KnowledgeBaseImpl.startTimeQualifier)) {
							addYear = true;
						} else if (qualifierId.getQualifier().equals(KnowledgeBaseImpl.endTimeQualifier)) {
							addYear = true;
						} else if (qualifierId.getQualifier().equals(KnowledgeBaseImpl.pointInTime)) {
							addYear = true;
						}
					}
				}
				if (addYear) {
					AbstractObject oYear = td.getValue(aid.getRowIdx(), new ColumnIdx(i));
					if (oYear instanceof TimeObject) {
						Integer iYear = TimeObject.getYear(oYear);
						if (iYear != null) {

							if (year == null || iYear > year) {
								year = iYear;
							}
						}
					}
				}
			}

		}

		return year;
	}

	public Integer getMaxTableTitleYear(CellId aid) {
		AnnotatedTableMetaData md = annotatedTablesRepository.getMetaData(aid.getTableId());

		Integer year = md.getMaxYearInTableTitle();
		if (yearInRange(year)) {
			return year;
		} else {
			return null;
		}

	}

	public Integer getMaxPageTitleYear(CellId aid) {
		AnnotatedTableMetaData md = annotatedTablesRepository.getMetaData(aid.getTableId());

		Integer year = md.getMaxYearInPageTitle();
		if (yearInRange(year)) {
			return year;
		} else {
			return null;
		}

	}

	public Integer getTableYearHiearchy(CellId aid) {
		Integer tableTitleYear = getMaxTableTitleYear(aid);
		if (tableTitleYear != null) {
			return tableTitleYear;
		} else {
			return getMostCommonTableTextYear(aid);
		}
	}

	public Integer getMostCommonTableTextYear(CellId aid) {
		AnnotatedTableMetaData md = annotatedTablesRepository.getMetaData(aid.getTableId());

		Map<Integer, MutableInt> yearCount = new HashMap<>();

		Consumer<List<Integer>> addYears = (list) -> {
			if (list != null) {
				for (Integer year : list) {
					if (yearInRange(year)) {
						CHMHelper.getFromMap(yearCount, year, () -> new MutableInt(0)).add(1);
					}

				}
			}
		};

		addYears.accept(md.getYearsAfterText());
		addYears.accept(md.getYearsBeforeText());

		if (!yearCount.isEmpty()) {
			Integer year = yearCount.entrySet().stream().max((e1, e2) -> {
				if (e1.getValue().intValue() > e2.getValue().intValue()) {
					return 1;
				} else if (e1.getValue().intValue() < e2.getValue().intValue()) {
					return -1;
				} else if (e1.getKey() > e2.getKey()) {
					return 1;
				} else
					return -1;
			}).get().getKey();

			return year;
		} else {
			return null;
		}
	}

	public Integer getHiearicalYear(CellId aid) {
		Integer columnYear = getMaxColumnYear(aid);
		if (columnYear != null)
			return columnYear;
		else {
			Integer cellYear = getCellYear(aid);
			if (cellYear != null)
				return cellYear;
			else {
				Integer tableYear = getMaxTableTitleYear(aid);
				if (tableYear != null) {
					return tableYear;
				} else {
					Integer pageTitle = getMaxPageTitleYear(aid);
					return pageTitle;
				}
			}
		}
	}

	public Integer getFullHiearicalYear(CellId aid) {
		Integer columnYear = getMaxColumnYear(aid);
		if (columnYear != null)
			return columnYear;
		else {
			Integer cellYear = getCellYear(aid);
			if (cellYear != null)
				return cellYear;
			else {
				Integer tableYear = getMaxTableTitleYear(aid);
				if (tableYear != null) {
					return tableYear;
				} else {
					Integer pageTitle = getMaxPageTitleYear(aid);
					if (pageTitle != null) {
						return pageTitle;
					} else {
						return getMostCommonTableTextYear(aid);
					}
				}
			}
		}
	}

	// public ColumnScores getColumnYearsScores(GroupedValues groupedValues) {
	//
	// return genericYearScorer(groupedValues, this::getMaxColumnYear);
	//
	// }
	//
	// public ColumnScores getTableYears(GroupedValues groupedValues) {
	//
	// return genericYearScorer(groupedValues, this::getTableYearHiearchy);
	//
	// }
	//
	// public ColumnScores getMaxTableTitleYear(GroupedValues groupedValues) {
	//
	// return genericYearScorer(groupedValues, this::getMaxTableTitleYear);
	//
	// }
	//
	// public ColumnScores getHierarichalYears(GroupedValues groupedValues) {
	// return genericYearScorer(groupedValues, this::getHiearicalYear);
	// }
	//
	// public ColumnScores getFullHierarichalYears(GroupedValues groupedValues)
	// {
	// return genericYearScorer(groupedValues, this::getFullHiearicalYear);
	// }

	private final static String sync = new String("SYNCHRONIZE");
	private static String cached = "";

	private static SimpleLockableConcurrentCache<UUID, TimestampScoresContainer> cache = new SimpleLockableConcurrentCache<>();

	public ScoresContainer genericYearScorer(SplitGroupedValues groupedValues, TS_METHOD tsMethod) {
		return genericYearScorer(groupedValues, tsMethod, false);
	}

	public ScoresContainer genericYearScorer(SplitGroupedValues groupedValues, TS_METHOD tsMethod, boolean propagate) {

		if (!cache.hasCacheFor(groupedValues.getUUID())) {
			Optional<java.util.function.Consumer<TimestampScoresContainer>> lock = cache.getLock(groupedValues.getUUID());
			if (lock.isPresent()) {

				lock.get().accept(generateCache(groupedValues));

			}
		}

		return cache.poll(groupedValues.getUUID()).get(tsMethod.getFirst(), propagate);

	}

	private TimestampScoresContainer generateCache(SplitGroupedValues groupedValues) {
		ConcurrentHashMap<Pair<QualifiedTripleId, CellId>, Map<Pair<TS_METHODS, Boolean>, Double>> values = new ConcurrentHashMap<>();

		Set<Pair<TS_METHOD, Boolean>> set = new HashSet<>();
		for (TS_METHOD tsMethod : this.getAllMethods()) {
			set.add(new Pair<>(tsMethod, true));
			set.add(new Pair<>(tsMethod, false));
		}

		PersistentParallel6.runAndFinish(set, (e) -> {
			genericYearScorerCalc(values, groupedValues, e.getFirst(), e.getSecond());
		}, this.getClass());

		return new TimestampScoresContainer(values);

	}

	public static ConcurrentHashMap<ClassPropertyId, MutableInt> original = new ConcurrentHashMap<>();
	public static ConcurrentHashMap<ClassPropertyId, MutableInt> propogated = new ConcurrentHashMap<>();

	public void genericYearScorerCalc(ConcurrentHashMap<Pair<QualifiedTripleId, CellId>, Map<Pair<TS_METHODS, Boolean>, Double>> targetMap,
			SplitGroupedValues groupedValues, TS_METHOD tsMethod, boolean propagate) {

		// Tri<SplitGroupedValues, String, Boolean> cacheKey = new
		// Tri<>(groupedValues, tsMethod.getFirst(), propagate);
		// // cache = new SimpleConcurrentCache<>();
		// if (cache.hasCacheFor(cacheKey)) {
		// return cache.get(cacheKey);
		// } else {

		Function<CellId, Integer> yearGetter = tsMethod.getSecond();
		// ConcurrentHashMap<QualifiedTripleId, Map<CellId, Probability>>
		// valueScores = new ConcurrentHashMap<>();

		PersistentParallel4.runAndFinish(groupedValues.entrySet(), (gv) -> {

			if (!ontologyRepository.isTimeDependent(gv.getKey().getQualifiedPropertyId().getPropertyId()))
				return;

			Set<TableColumnId> propertyToTableMapping = new HashSet<>();

			for (SingleValue sv : gv.getValue().getValues()) {
				propertyToTableMapping.add(sv.getId().getTableColumnId());
			}

			SimilarityMeasure measure = similarityCalculator.getMeasure(gv.getValue().getDataType());
			BaseFuser baseFuser = fuserRepository.getBaseFuser(gv.getValue().getDataType());

			QualifiedTripleId tid = gv.getKey();
			Integer kbYear = timestampQualifiersAdapter.getFromQualifiers(tid.getQualifiers());
			if (kbYear == null) {
				return;
			}
			SortedMap<Integer, Set<CellId>> sortedByYear = new TreeMap<>();
			List<Cluster> clusters = new ArrayList<>();
			// Map<CellId, Probability> subCHMMap = new HashMap<>();
			List<SingleValue> noScoreSV = new ArrayList<>();
			for (SingleValue sv : gv.getValue().getValues()) {
				CellId aid = sv.getId();
				Integer year = yearGetter.apply(aid);

				if (year != null) {

					CHMHelper.getFromMap(sortedByYear, year, () -> new HashSet<>()).add(aid);

					if (propagate) {
						clusters = yearEquivalenceClusterer.addValue(clusters, measure, baseFuser, sv.getValue(), year);
					}
				} else {
					noScoreSV.add(sv);
				}

			}

			if (propagate) {

				CHMHelper.getFromCHM(original, tid.getTripleId().getClassPropertyId(), () -> new MutableInt(0))
						.add(sortedByYear.entrySet().stream().map((e) -> {
							return e.getValue().size();
						}).mapToInt(Integer::intValue).sum());

				for (Iterator<SingleValue> it = noScoreSV.iterator(); it.hasNext();) {
					SingleValue sv = it.next();
					Integer year = yearEquivalenceClusterer.getYear(clusters, measure, sv.getValue());
					if (year != null) {
						CHMHelper.getFromMap(sortedByYear, year, () -> new HashSet<>()).add(sv.getId());
						it.remove();
						CHMHelper.getFromCHM(propogated, tid.getTripleId().getClassPropertyId(), () -> new MutableInt(0)).add(1);
					}
				}

			}

			Set<CellId> noScore = noScoreSV.stream().map((e) -> {
				return e.getId();
			}).collect(Collectors.toSet());

			int divider = sortedByYear.size() + (noScore.isEmpty() ? 0 : 1);
			double step = 1.0 / divider;

			int i = 1;
			double lowestScore = 1.0;
			for (Entry<Integer, Set<CellId>> e : sortedByYear.entrySet()) {

				Double score = getYearSimilarity(kbYear, e.getKey());

				if (score != null) {
					// if (e.getKey() >= 2012) {
					for (CellId aid : e.getValue()) {
						// subCHMMap.put(aid, new Probability(new
						// Accuracy(score,
						// 1)));

						ConcurrentHashMap<Pair<TS_METHODS, Boolean>, Double> innerMap = (ConcurrentHashMap<Pair<TS_METHODS, Boolean>, Double>) CHMHelper
								.getFromCHM(targetMap, new Pair<>(tid, aid), () -> new ConcurrentHashMap<>());
						innerMap.put(new Pair<>(tsMethod.getFirst(), propagate), score);
					}

					if (score < lowestScore) {
						lowestScore = score;
					}
				}
				// } else {
				// for (TableColumnId aid : e.getValue()) {
				// subCHMMap.put(aid, new Probability(new Accuracy(0.0,
				// 1)));
				// }
				// }

				i++;

			}

			// if (sortedByYear.isEmpty()) {
			// for (CellId aid : noScore) {
			// subCHMMap.put(aid, new Probability(new Accuracy(0.0, 1)));
			// }
			// } else {
			// for (CellId aid : noScore) {
			// subCHMMap.put(aid, new Probability(new Accuracy(0.0, 1)));
			// }
			// }

			// int dfdf = 123123;
			// if (!subCHMMap.isEmpty()) {
			// valueScores.put(tid, subCHMMap);
			//
			// }

		}, "ColumnYearsScores");

		// QualifiedTripleIdScores cs = new
		// QualifiedTripleIdScores(valueScores);
		// cache.add(cacheKey, cs);
		//
		// return cs;
		// }

	}

	// private static PropertyId yearProperty = new
	// PropertyId("http://uni-mannheim.de/year");

	// public int getKBYear(ValueId vid) {
	// Object o = dbpediaRepository.getQualifiedValueList(valueId).getValue(new
	// ValueId(vid.getInstanceId(), yearProperty));
	// if (o != null) {
	// if (o instanceof Date) {
	// calendar.getCalendar().setTime((Date) o);
	// int ret = calendar.getCalendar().get(Calendar.YEAR);
	//
	// return ret;
	// }
	// }
	//
	// throw new IllegalStateException();
	// }

	public Double getYearSimilarity(int kbYear, int matchedYear) {
		if (matchedYear == kbYear)
			return 1.0;
		else {

			int diff = Math.abs(kbYear - matchedYear);
			if (diff < 4) {
				return 1.0 - (diff / 4.0);
			}
		}

		return null;
	}

	@Deprecated
	public ScoresContainer genericYearScorer(GroupedValues fullSet, TS_METHOD gET_HIERARICAL_YEAR2) {
		throw new UnsupportedOperationException();
	}

	@Deprecated
	public ScoresContainer genericYearScorer(GroupedValues fullSet, TS_METHOD gET_HIERARICAL_YEAR2, boolean propagate) {
		throw new UnsupportedOperationException();
	}
}
