package de.mannheim.uni.expansion.tablefusion.core.scoring;

import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.OptionalInt;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleIdContainer;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier.CascadedType;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier.TimestampType;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.matching.AbstractGroupledValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScoresInterface;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MapSorter;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded.DefaultIntegerScorable;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded.DefaultIntegerScorable.DoubleScorable;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded.IntegerScorable;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleConcurrentCache;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Quad;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;

public class TimestampScorerCleanedUp {

	private static SimpleConcurrentCache<CacheKey, TimestampScores> newCache = new SimpleConcurrentCache<>();

	private final AnnotatedTablesRepository repository;
	private final KnowledgeBase knowledgeBase;
	private final SimilarityCalculator similarityCalculator;
	private final TimestampQualifiersAdapter qualifierAdapter;

	@Inject
	public TimestampScorerCleanedUp(AnnotatedTablesRepository tableRepository, KnowledgeBase knowledgeBase, SimilarityCalculator similarityCalculator,
			TimestampQualifiersAdapter qualifierAdapter) {
		super();
		this.repository = tableRepository;
		this.knowledgeBase = knowledgeBase;
		this.similarityCalculator = similarityCalculator;
		this.qualifierAdapter = qualifierAdapter;
	}

	public TimestampScores getTimestampScores(SplitGroupedValues learningSet, boolean withTimestamps, boolean propagate, boolean rowCell) {
		CacheKey cacheKey = new CacheKey(learningSet.getUUID(), withTimestamps, propagate, rowCell);
		if (newCache.hasCacheFor(cacheKey)) {
			return newCache.get(cacheKey);
		} else {

			ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> probabilties = new ConcurrentHashMap<>();
			ConcurrentHashMap<TableColumnId, Map<TimestampIdentifier, Accuracy>> fullMappings = new ConcurrentHashMap<>();
			ConcurrentHashMap<ClassPropertyId, Accuracy> averageMaxScores = new ConcurrentHashMap<>();
			ConcurrentHashMap<ClassPropertyId, ConcurrentLinkedQueue<Accuracy>> maxScores = new ConcurrentHashMap<>();
			Map<TableColumnId, Map<QualifiedTripleId, SingleValue>> grouped = groupByColumnId(learningSet);

			Set<TableColumnId> needPropogated = Collections.newSetFromMap(new ConcurrentHashMap<>());

			ConcurrentHashMap<TableColumnId, Set<TimestampIdentifier>> countTsMap = new ConcurrentHashMap<>();

			PersistentParallel4.runAndFinish(grouped.entrySet(), (e) -> {
				boolean needsPropogation = false;
				Map<TimestampIdentifier, MutableInt> total = new HashMap<>();
				Map<TimestampIdentifier, MutableInt> correct = new HashMap<>();

				AnnotatedTable table = repository.getTable(e.getKey().getTableId());

				AnnotatedTableMetaData am = repository.getMetaData(e.getKey().getTableId());

				// System.out.println("Values in group: " + e.getValue().size()
				// + ", ROWS in Table: " + am.getInstances().size());

				List<ColumnIdx> timeColumns = table.getTimeColumns();
				Map<TimestampIdentifier, Integer> tableYears = repository.getMetaData(e.getKey().getTableId()).getTableYears();

				if (tableYears.isEmpty() && timeColumns.isEmpty()) {
					needsPropogation = true;
				}
				Set<Pair<PropertyId, InstanceId>> totalCount = new HashSet<>();

				// for (QualifiedTripleId vvv : e.getValue().keySet()) {
				//
				// Pair<PropertyId, InstanceId> pair = new
				// Pair<>(vvv.tripleId().getPropertyId(),
				// vvv.tripleId().getInstanceId());
				// if (!totalCount.contains(pair)) {
				// totalCount.add(pair);
				// }
				// }
				// int totalOverlap = e.getValue().size();

				int sdf = 1;
				for (Entry<QualifiedTripleId, SingleValue> sv : e.getValue().entrySet()) {
					QualifiedValueList a = knowledgeBase.getQualifiedValueListWithEqualProperties(sv.getKey().tripleId().getValueId());
					if (a != null) {

						Set<AbstractObject> values = this.qualifierAdapter.getForQualifiers(a, sv.getKey().getQualifiers());

						if (!values.isEmpty()) {

							Integer year = null;
							for (AbstractQualifier aq : sv.getKey().getQualifiers()) {
								if (aq instanceof TimestampPointInTime) {
									year = ((TimestampPointInTime) aq).getYear();
								}
							}

							if (year != null) {

								Map<TimestampIdentifier, Integer> tableYearsTemp = new HashMap<>();
								if (withTimestamps)
									tableYearsTemp.putAll(tableYears);

								tableYearsTemp.put(TimestampIdentifier.createPropagatedIdentifier(year), year);

								if (rowCell) {
									for (ColumnIdx columnIdx : timeColumns) {
										AbstractObject ao = table.getValue(sv.getValue().getId().getRowIdx(), columnIdx);
										Integer columnYear = TimeObject.getYear(ao);
										if (columnYear != null) {
											tableYearsTemp.put(
													TimestampIdentifier.createCascadedIdentifier(CascadedType.COLUMN_VALUE, columnIdx.getColumnIdx(), 0),
													columnYear);
										}
									}
								}
								for (Entry<TimestampIdentifier, Integer> ty : tableYearsTemp.entrySet()) {

									if (ty.getValue().equals(year)) {
										MutableInt totalMap = CHMHelper.getFromMap(total, ty.getKey(), () -> new MutableInt());
										MutableInt correctMap = CHMHelper.getFromMap(correct, ty.getKey(), () -> new MutableInt());

										totalMap.increment();

										for (AbstractObject ao : values) {

											if (similarityCalculator.isEqual(ao, sv.getValue().getValue())) {
												correctMap.increment();
												break;
											}
										}

									}
								}
							}
						}
					}
				}

				OptionalInt totalOverlapOptional = total.entrySet().stream().map((totalEntry) -> {
					return totalEntry.getValue();
				}).mapToInt(MutableInt::intValue).max();

				Map<TimestampIdentifier, Accuracy> accuracies = new HashMap<>();

				for (Entry<TimestampIdentifier, MutableInt> entry : total.entrySet()) {
					MutableInt totalInt = entry.getValue();
					MutableInt correctInt = correct.get(entry.getKey());
					if (correctInt.intValue() > 0) {

						Accuracy acc = Accuracy.fromCorrectAndTotal(correctInt.intValue(), totalInt.intValue());

						accuracies.put(entry.getKey(), acc);
						if (entry.getKey().getType() == TimestampIdentifier.TimestampType.PROPAGATED) {
							CHMHelper.getFromCHM(fullMappings, e.getKey(), () -> new HashMap<>()).put(entry.getKey(), acc);
						}
					}

				}
				LinkedHashMap<TimestampIdentifier, Accuracy> sortedAccuracies;
				if (false && totalOverlapOptional.isPresent()) {
					final int totalOverlap = totalOverlapOptional.getAsInt();
					Map<TimestampIdentifier, Accuracy> harmonicMean = accuracies.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (acc) -> {
						Accuracy acc1 = acc.getValue();
						Accuracy acc2 = Accuracy.fromCorrectAndTotal(acc1.getCorrect(), totalOverlap);
						Accuracy harmonic = Accuracy.harmonicMean(acc1, acc2, acc1.getTotal());

						// CHMHelper.getFromCHM(fullMappings, e.getKey(), () ->
						// new HashMap<>()).put(acc.getKey(), harmonic);

						return harmonic;
						// return acc1;
					}));

					sortedAccuracies = MapSorter.sortHashMapByValuesD(harmonicMean);
				} else {
					sortedAccuracies = MapSorter.sortHashMapByValuesD(accuracies);
				}

				Map<Integer, Set<TimestampIdentifier>> tableYearsReverse = new HashMap<>();
				for (Entry<TimestampIdentifier, Integer> ty : tableYears.entrySet()) {
					CHMHelper.getFromMap(tableYearsReverse, ty.getValue(), () -> new HashSet<>()).add(ty.getKey());
				}
				Set<TimestampIdentifier> tsSet = new HashSet<>();

				LinkedHashMap<TimestampIdentifier, Accuracy> mappings = new LinkedHashMap<>();
				for (Entry<TimestampIdentifier, Accuracy> entry : sortedAccuracies.entrySet()) {
					if (!withTimestamps || entry.getKey().getType() != TimestampIdentifier.TimestampType.PROPAGATED) {
						Accuracy acc = accuracies.get(entry.getKey());
						probabilties.put(e.getKey(), new Pair<>(entry.getKey(), acc));

						if (withTimestamps) {
							tsSet.addAll(tableYearsReverse.get(tableYears.get(entry.getKey())));
						}

						mappings.put(entry.getKey(), acc);

						// Set<QualifiedPropertyId> propertyIds =
						// e.getValue().keySet().stream().map((qtid) -> {
						// return qtid.getQualifiedPropertyId();
						// }).collect(Collectors.toSet());

						Set<ClassPropertyId> propertyIds = e.getValue().keySet().stream().map((qtid) -> {
							return qtid.getTripleId().getClassPropertyId();
						}).collect(Collectors.toSet());
						for (ClassPropertyId cpid : propertyIds) {

							CHMHelper.getFromCHM(maxScores, cpid, () -> new ConcurrentLinkedQueue<>()).add(acc);

							// MutableReference<Accuracy> reference =
							// CHMHelper.getFromCHM(maxScores,
							// qualifiedTripleId.tripleId(),
							// () -> new MutableReference<Accuracy>(null));
							// synchronized (reference) {
							// if (reference.get() != null &&
							// reference.get().getAccuracy() <
							// acc.getAccuracy()) {
							// reference.set(acc);
							// } else if (reference.get() == null) {
							// reference.set(acc);
							// }
							// }
						}

						break;
					}

				}

				if (!tsSet.isEmpty()) {
					countTsMap.put(e.getKey(), tsSet);
				}

				if (needsPropogation && propagate && withTimestamps) {
					if (!mappings.isEmpty())
						throw new IllegalStateException();
					else {
						needPropogated.add(e.getKey());
					}
				}

				// for (Entry<TimestampIdentifier, Accuracy> entry :
				// sortedAccuracies.entrySet()) {
				// if (entry.getKey().getType() !=
				// TimestampIdentifier.TimestampType.PROPAGATED) {
				// Accuracy acc = accuracies.get(entry.getKey());
				// probabilties.put(e.getKey(), new Pair<>(entry.getKey(),
				// acc));
				//
				// Set<QualifiedPropertyId> propertyIds =
				// e.getValue().keySet().stream().map((qtid) -> {
				// return qtid.getQualifiedPropertyId();
				// }).collect(Collectors.toSet());
				// for (QualifiedPropertyId qualifiedPropertyId :
				// propertyIds) {
				// MutableReference<Accuracy> reference =
				// CHMHelper.getFromCHM(maxScores,
				// qualifiedTripleId.tripleId(),
				// () -> new MutableReference<Accuracy>(null));
				// synchronized (reference) {
				// if (reference.get() != null &&
				// reference.get().getAccuracy() < acc.getAccuracy()) {
				// reference.set(acc);
				// } else if (reference.get() == null) {
				// reference.set(acc);
				// }
				// }
				// }
				//
				// break;
				// }
				//
				// }

			}, "TimeStampCalculator");

			for (Entry<ClassPropertyId, ConcurrentLinkedQueue<Accuracy>> me : maxScores.entrySet()) {
				double acc = 0.0;
				int total = 0;
				int count = 0;

				for (Accuracy a : me.getValue()) {
					acc += a.getAccuracy();
					total += a.getTotal();
					count++;
				}

				averageMaxScores.put(me.getKey(), new Accuracy(acc / count, total / count));
			}

			ConcurrentHashMap<TableColumnId, LinkedHashMap<TimestampIdentifier, Accuracy>> fullMappingsSorted = new ConcurrentHashMap<>();
			PersistentParallel6.runAndFinish(fullMappings.entrySet(), (e) -> {
				fullMappingsSorted.put(e.getKey(), MapSorter.sortHashMapByValuesD(e.getValue()));
			}, this.getClass());

			if (propagate && !needPropogated.isEmpty()) {

				ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> propagatedProbabilities = runPropogatedJaqqardOneSidedOtsu(needPropogated,
						grouped, learningSet, probabilties, fullMappingsSorted, grouped.keySet());

				probabilties.putAll(propagatedProbabilities);
			}

			Map<ClassPropertyId, Map<String, MutableInt>> tsCount = new HashMap<>();
			for (Entry<TableColumnId, Set<TimestampIdentifier>> entry : countTsMap.entrySet()) {
				for (TimestampIdentifier tsid : entry.getValue()) {

					String key = null;
					switch (tsid.getType()) {
					case SIMPLE:
						key = tsid.getSimpleType().name();
						break;
					case CASCADED:
						key = tsid.getCascadedType().name();
						break;
					case PROPAGATED:
						key = "PROPAGATED";
						break;

					}

					AnnotatedTableMetaData metadata = repository.getMetaData(entry.getKey().getTableId());
					ClassId classId = metadata.getDbPediaClass();
					PropertyId propertyId = metadata.getProperties().get(entry.getKey().getColumnIdx().getColumnIdx());

					Map<String, MutableInt> innerMap = CHMHelper.getFromMap(tsCount, new ClassPropertyId(classId, propertyId), () -> new HashMap<>());

					CHMHelper.getFromMap(innerMap, key, () -> new MutableInt()).increment();
				}
			}

			TimestampScores scores = new TimestampScores(probabilties, fullMappingsSorted, averageMaxScores, tsCount);
			newCache.add(cacheKey, scores);

			return scores;
		}
	}

	private ConcurrentHashMap<TableColumnId, Map<InstanceId, AbstractObject>> getColumnByInstances(Set<TableColumnId> tableColumnIds) {
		ConcurrentHashMap<TableColumnId, Map<InstanceId, AbstractObject>> groupedTableColumnsFull = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(tableColumnIds, (columnId) -> {
			// PropertyId propertyId =
			// repository.getMetaData(columnId.getTableId()).getProperty(columnId.getColumnIdx().getColumnIdx());
			// ClassId classId =
			// repository.getMetaData(columnId.getTableId()).getDbPediaClass();
			// ClassPropertyId cpid = new ClassPropertyId(classId, propertyId);
			// Integer year = repository.getTimeStamp(columnId,
			// columnIdEntry.getValue().getFirst());

			Map<RowIdx, AbstractObject> column = repository.getColumn(columnId.getTableId(), columnId.getColumnIdx());
			Map<Integer, InstanceId> instances = repository.getMetaData(columnId.getTableId()).getInstances();

			Map<InstanceId, AbstractObject> retMap = new HashMap<>();
			for (Entry<Integer, InstanceId> instanceId : instances.entrySet()) {
				retMap.put(instanceId.getValue(), column.get(new RowIdx(instanceId.getKey())));
			}

			groupedTableColumnsFull.put(columnId, retMap);

		}, this.getClass());

		return groupedTableColumnsFull;
	}

	private ConcurrentHashMap<ClassPropertyId, Collection<Tri<TableColumnId, Integer, Accuracy>>> getColumnsByClassPropertyId(
			ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> probabilties) {

		ConcurrentHashMap<ClassPropertyId, Collection<Tri<TableColumnId, Integer, Accuracy>>> groupedTableColumns = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(probabilties.entrySet(), (columnIdEntry) -> {
			TableColumnId columnId = columnIdEntry.getKey();
			PropertyId propertyId = repository.getMetaData(columnId.getTableId()).getProperty(columnId.getColumnIdx().getColumnIdx());
			ClassId classId = repository.getMetaData(columnId.getTableId()).getDbPediaClass();
			ClassPropertyId cpid = new ClassPropertyId(classId, propertyId);
			Integer year = repository.getTimeStamp(columnId, columnIdEntry.getValue().getFirst());
			if (year != null) {
				CHMHelper.getFromCHM(groupedTableColumns, cpid, () -> new ConcurrentLinkedQueue<>())
						.add(new Tri<>(columnId, year, columnIdEntry.getValue().getSecond()));
			}

		}, this.getClass());

		return groupedTableColumns;
	}

	private ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> runPropogatedJaqqardOneSidedOtsu(Set<TableColumnId> needPropogated,
			Map<TableColumnId, Map<QualifiedTripleId, SingleValue>> grouped, SplitGroupedValues learningSet,
			ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> probabilties,
			ConcurrentHashMap<TableColumnId, LinkedHashMap<TimestampIdentifier, Accuracy>> fullMappingsSorted, Set<TableColumnId> tableColumnIds) {
		ConcurrentHashMap<ClassPropertyId, Collection<Tri<TableColumnId, Integer, Accuracy>>> groupedTableColumns = getColumnsByClassPropertyId(probabilties);
		ConcurrentHashMap<TableColumnId, Map<InstanceId, AbstractObject>> groupedTableColumnsFull = getColumnByInstances(tableColumnIds);

		ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> propagatedProbabilities = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(needPropogated, (columnId) -> {

			PropertyId propertyId = repository.getMetaData(columnId.getTableId()).getProperty(columnId.getColumnIdx().getColumnIdx());
			ClassId classId = repository.getMetaData(columnId.getTableId()).getDbPediaClass();
			ClassPropertyId cpid = new ClassPropertyId(classId, propertyId);
			double maxOverlap = 0;
			Integer maxYear = null;
			Map<RowIdx, AbstractObject> column = repository.getColumn(columnId.getTableId(), columnId.getColumnIdx());
			Map<Integer, InstanceId> instances = repository.getMetaData(columnId.getTableId()).getInstances();

			Map<InstanceId, AbstractObject> instanceMap = new HashMap<>();

			for (Entry<Integer, InstanceId> instanceEntry : instances.entrySet()) {
				instanceMap.put(instanceEntry.getValue(), column.get(new RowIdx(instanceEntry.getKey())));
			}

			Map<Integer, Double> overlapMap = new HashMap<>();
			for (Tri<TableColumnId, Integer, Accuracy> candidateColumnId : groupedTableColumns.get(cpid)) {

				Map<InstanceId, AbstractObject> candidateValues = groupedTableColumnsFull.get(candidateColumnId.getA());

				double correct = 0;
				for (Entry<InstanceId, AbstractObject> entry : instanceMap.entrySet()) {
					AbstractObject candidate = candidateValues.get(entry.getKey());

					if (this.similarityCalculator.isEqual(entry.getValue(), candidate)) {
						correct++;
					}
				}

				if (correct > 0 && correct * 2 >= instanceMap.size()) {
					double overlap = (correct / instanceMap.size()) * candidateColumnId.getC().getAccuracy();

					overlapMap.put(candidateColumnId.getB(), overlap);
				}
				// if (overlap > maxOverlap) {
				// maxOverlap = overlap;
				// maxYear = candidateColumnId.getB();
				// }

			}

			if (!overlapMap.isEmpty()) {
				new DefaultIntegerScorable(new DoubleScorable(0.0));

				List<OtsuYearDoubleEntry> otsuYears = overlapMap.entrySet().stream().map((e) -> {
					return new OtsuYearDoubleEntry(e.getKey(), e.getValue());
				}).collect(Collectors.toList());

				List<OtsuYearDoubleEntry> otsuYearsNew = otsu.threshold(otsuYears, (e) -> {
					return e;
				});

				LinkedHashMap<TimestampIdentifier, Accuracy> innerMap = fullMappingsSorted.get(columnId);
				if (innerMap != null) {

					Accuracy maxAcc = null;
					TimestampIdentifier maxTsid = null;
					for (OtsuYearDoubleEntry oy : otsuYears) {
						TimestampIdentifier tsid = TimestampIdentifier.createPropagatedIdentifier(oy.getYear());
						Accuracy acc = innerMap.get(tsid);
						if (acc != null) {
							if (maxAcc == null || acc.getAccuracy() > maxAcc.getAccuracy()) {
								maxAcc = acc;
								maxTsid = tsid;
							}
						}
					}

					if (maxAcc != null) {
						propagatedProbabilities.put(columnId, new Pair<>(maxTsid, maxAcc));
					}

				}
			}

		}, this.getClass());

		return propagatedProbabilities;
	}

	private ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> runPropogatedJaqqardOneSided(Set<TableColumnId> needPropogated,
			Map<TableColumnId, Map<QualifiedTripleId, SingleValue>> grouped, SplitGroupedValues learningSet,
			ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> probabilties,
			ConcurrentHashMap<TableColumnId, LinkedHashMap<TimestampIdentifier, Accuracy>> fullMappingsSorted, Set<TableColumnId> tableColumnIds) {
		ConcurrentHashMap<ClassPropertyId, Collection<Tri<TableColumnId, Integer, Accuracy>>> groupedTableColumns = getColumnsByClassPropertyId(probabilties);
		ConcurrentHashMap<TableColumnId, Map<InstanceId, AbstractObject>> groupedTableColumnsFull = getColumnByInstances(tableColumnIds);

		ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> propagatedProbabilities = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(needPropogated, (columnId) -> {

			PropertyId propertyId = repository.getMetaData(columnId.getTableId()).getProperty(columnId.getColumnIdx().getColumnIdx());
			ClassId classId = repository.getMetaData(columnId.getTableId()).getDbPediaClass();
			ClassPropertyId cpid = new ClassPropertyId(classId, propertyId);
			double maxOverlap = 0;
			Integer maxYear = null;
			Map<RowIdx, AbstractObject> column = repository.getColumn(columnId.getTableId(), columnId.getColumnIdx());
			Map<Integer, InstanceId> instances = repository.getMetaData(columnId.getTableId()).getInstances();

			Map<InstanceId, AbstractObject> instanceMap = new HashMap<>();

			for (Entry<Integer, InstanceId> instanceEntry : instances.entrySet()) {
				instanceMap.put(instanceEntry.getValue(), column.get(new RowIdx(instanceEntry.getKey())));
			}

			for (Tri<TableColumnId, Integer, Accuracy> candidateColumnId : groupedTableColumns.get(cpid)) {

				Map<InstanceId, AbstractObject> candidateValues = groupedTableColumnsFull.get(candidateColumnId.getA());

				double correct = 0;
				for (Entry<InstanceId, AbstractObject> entry : instanceMap.entrySet()) {
					AbstractObject candidate = candidateValues.get(entry.getKey());

					if (this.similarityCalculator.isEqual(entry.getValue(), candidate)) {
						correct++;
					}
				}

				if (correct > 0 && correct * 2 >= instanceMap.size()) {
					double overlap = (correct / instanceMap.size()) * candidateColumnId.getC().getAccuracy();
					if (overlap > maxOverlap) {
						maxOverlap = overlap;
						maxYear = candidateColumnId.getB();
					}
				}

			}

			if (maxYear != null) {
				TimestampIdentifier tsid = TimestampIdentifier.createPropagatedIdentifier(maxYear);
				LinkedHashMap<TimestampIdentifier, Accuracy> innerMap = fullMappingsSorted.get(columnId);
				if (innerMap == null) {
					// System.out.println("Expected to have key. FIX THIS.");
				} else {
					Accuracy acc = innerMap.get(tsid);
					if (acc == null) {
						// System.out.println("Expected to have ACC. FIX
						// THIS.");

					} else {
						propagatedProbabilities.put(columnId, new Pair<>(tsid, acc));
					}
				}

			}

		}, this.getClass());

		return propagatedProbabilities;
	}

	private final OtsuThresholderRounded otsu = new OtsuThresholderRounded();

	private static class OtsuYearEntry implements IntegerScorable {

		private final int score;
		private final int year;

		public OtsuYearEntry(int year, int score) {
			this.year = year;
			this.score = score;
		}

		@Override
		public int compareTo(IntegerScorable o) {
			return Integer.compare(this.score, o.getScore());
		}

		@Override
		public Integer getScore() {
			return score;
		}

		public int getYear() {
			return year;
		}

	}

	private static class OtsuYearDoubleEntry implements IntegerScorable {

		private final int year;
		private final double score;

		public OtsuYearDoubleEntry(int year, double score) {
			this.score = score;
			this.year = year;
		}

		public int getYear() {
			return year;
		}

		@Override
		public int compareTo(IntegerScorable o) {
			return Integer.compare(this.getScore(), o.getScore());
		}

		@Override
		public Integer getScore() {
			return (int) Math.round(score * 10000);
		}

		@Override
		public String toString() {
			return "OtsuYearDoubleEntry [year=" + year + ", score=" + score + "]";
		}

	}

	private ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> runPropogatedOld(Set<TableColumnId> needPropogated,
			Map<TableColumnId, Map<QualifiedTripleId, SingleValue>> grouped, SplitGroupedValues learningSet,
			ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> probabilties,
			ConcurrentHashMap<TableColumnId, LinkedHashMap<TimestampIdentifier, Accuracy>> fullMappingsSorted, Set<TableColumnId> columnIdsSet) {

		ConcurrentHashMap<TableColumnId, Map<InstanceId, AbstractObject>> groupedTableColumnsFull = getColumnByInstances(columnIdsSet);

		ConcurrentHashMap<TripleId, Collection<Quad<TableColumnId, Integer, Accuracy, AbstractObject>>> groupedTableColumnsValues = getColumnsByTripleId(
				probabilties, groupedTableColumnsFull);

		ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> propagatedProbabilities = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(needPropogated, (columnId) -> {

			Map<Integer, MutableInt> totalYearCount = new HashMap<>();

			PropertyId propertyId = repository.getMetaData(columnId.getTableId()).getProperty(columnId.getColumnIdx().getColumnIdx());
			ClassId classId = repository.getMetaData(columnId.getTableId()).getDbPediaClass();
			Map<InstanceId, AbstractObject> instanceMap = groupedTableColumnsFull.get(columnId);

			if (instanceMap == null) {

				Map<RowIdx, AbstractObject> column = repository.getColumn(columnId.getTableId(), columnId.getColumnIdx());
				Map<Integer, InstanceId> instances = repository.getMetaData(columnId.getTableId()).getInstances();
				Map<InstanceId, AbstractObject> retMap = new HashMap<>();
				for (Entry<Integer, InstanceId> instanceId : instances.entrySet()) {
					retMap.put(instanceId.getValue(), column.get(new RowIdx(instanceId.getKey())));
				}

				LinkedHashMap<TimestampIdentifier, Accuracy> valueTest = fullMappingsSorted.get(columnId);
				int sdf = 123;
			}
			for (Entry<InstanceId, AbstractObject> value : instanceMap.entrySet()) {
				Map<Integer, MutableInt> yearCount = new HashMap<>();

				Collection<Quad<TableColumnId, Integer, Accuracy, AbstractObject>> candidates = groupedTableColumnsValues
						.get(new TripleId(classId, propertyId, value.getKey()));

				if (candidates != null) {
					for (Quad<TableColumnId, Integer, Accuracy, AbstractObject> candidate : candidates) {
						if (similarityCalculator.isEqual(value.getValue(), candidate.getD())) {
							CHMHelper.getFromMap(yearCount, candidate.getB(), () -> new MutableInt()).increment();
						}

					}
				} else {
					int sdfsdf = 123;
				}

				if (!yearCount.isEmpty()) {

					List<OtsuYearEntry> otsuYears = yearCount.entrySet().stream().map((e) -> {
						return new OtsuYearEntry(e.getKey(), e.getValue().intValue());
					}).collect(Collectors.toList());

					otsuYears = otsu.threshold(otsuYears, (e) -> {
						return e;
					});

					for (OtsuYearEntry oy : otsuYears) {
						CHMHelper.getFromMap(totalYearCount, oy.getYear(), () -> new MutableInt()).increment();
					}

					// int maxCount = 0;
					// Integer maxYear = null;
					//
					// for (Entry<Integer, MutableInt> yearEntry :
					// yearCount.entrySet()) {
					// if (yearEntry.getValue().intValue() > maxCount || maxYear
					// == null) {
					// maxYear = yearEntry.getKey();
					// maxCount = yearEntry.getValue().intValue();
					// }
					// }

				}
			}

			if (!totalYearCount.isEmpty()) {

				List<OtsuYearEntry> otsuYears = totalYearCount.entrySet().stream().map((e) -> {
					return new OtsuYearEntry(e.getKey(), e.getValue().intValue());
				}).collect(Collectors.toList());

				otsuYears = otsu.threshold(otsuYears, (e) -> {
					return e;
				});

				LinkedHashMap<TimestampIdentifier, Accuracy> innerMap = fullMappingsSorted.get(columnId);
				if (innerMap != null) {

					Accuracy maxAcc = null;
					TimestampIdentifier maxTsid = null;
					for (OtsuYearEntry oy : otsuYears) {
						TimestampIdentifier tsid = TimestampIdentifier.createPropagatedIdentifier(oy.getYear());
						Accuracy acc = innerMap.get(tsid);
						if (acc != null) {
							if (maxAcc == null || acc.getAccuracy() > maxAcc.getAccuracy()) {
								maxAcc = acc;
								maxTsid = tsid;
							}
						}
					}

					if (maxAcc != null) {
						propagatedProbabilities.put(columnId, new Pair<>(maxTsid, maxAcc));
					}

				}

				// int maxCount = 0;
				// Integer maxYear = null;
				//
				// for (Entry<Integer, MutableInt> yearEntry :
				// totalYearCount.entrySet()) {
				// if (yearEntry.getValue().intValue() > maxCount || maxYear ==
				// null) {
				// maxYear = yearEntry.getKey();
				// maxCount = yearEntry.getValue().intValue();
				// }
				// }

				// TimestampIdentifier tsid =
				// TimestampIdentifier.createPropagatedIdentifier(maxYear);
				// LinkedHashMap<TimestampIdentifier, Accuracy> innerMap =
				// fullMappingsSorted.get(columnId);
				// if (innerMap == null) {
				// // System.out.println("Expected to have key. FIX THIS.");
				// } else {
				// Accuracy acc = innerMap.get(tsid);
				// if (acc == null) {
				// // System.out.println("Expected to have ACC. FIX
				// // THIS.");
				//
				// } else {
				// propagatedProbabilities.put(columnId, new Pair<>(tsid, acc));
				// }
				// }

			}

			// for (Entry<QualifiedTripleId, SingleValue> sv : gv.entrySet()) {
			// for (SingleValue otherSv :
			// learningSet.get(sv.getKey()).getValues()) {
			// Pair<TimestampIdentifier, Accuracy> mappedTs =
			// probabilties.get(otherSv.getId().getTableColumnId());
			// if (mappedTs != null) {
			// if (similarityCalculator.isEqual(sv.getValue().getValue(),
			// otherSv.getValue())) {
			// Integer year = repository.getTimeStamp(otherSv.getId(),
			// mappedTs.getFirst());
			// Pair<QualifiedTripleId, Integer> key = new Pair<>(sv.getKey(),
			// year);
			// if (yearScores.containsKey(key)) {
			// MutableDouble md = CHMHelper.getFromMap(yearScores, new
			// Pair<>(sv.getKey(), year), () -> new MutableDouble());
			// if (md.doubleValue() < mappedTs.getSecond().getAccuracy()) {
			// md.setValue(mappedTs.getSecond().getAccuracy());
			// }
			// } else {
			// CHMHelper.getFromMap(yearCount, year, () -> new
			// MutableInt()).increment();
			// CHMHelper.getFromMap(yearScores, new Pair<>(sv.getKey(), year),
			// () -> new MutableDouble())
			// .setValue(mappedTs.getSecond().getAccuracy());
			// }
			//
			// }
			//
			// }
			// }
			// }
			//
			// double maxCount = 0.0;
			// Integer maxYear = null;
			//
			// for (Entry<Integer, MutableInt> yearC : yearCount.entrySet()) {
			// MutableInt count = yearC.getValue();
			// if (count != null) {
			// if (count.intValue() > maxCount) {
			// maxCount = count.intValue();
			// maxYear = yearC.getKey();
			// }
			// }
			//
			// }
			//
			// if (maxYear != null) {
			// double score = 0.0;
			// int count = 0;
			// for (Entry<Pair<QualifiedTripleId, Integer>, MutableDouble>
			// yearScore : yearScores.entrySet()) {
			// if (yearScore.getKey().getSecond() == maxYear) {
			// score += yearScore.getValue().doubleValue();
			// count++;
			// }
			// }
			// if (count > 0)
			// propagatedProbabilities.put(columnId, new
			// Pair<>(TimestampIdentifier.createPropagatedIdentifier(maxYear),
			// Accuracy.simple(score / count)));
			// }

		}, this.getClass());

		return propagatedProbabilities;

	}

	private ConcurrentHashMap<TripleId, Collection<Quad<TableColumnId, Integer, Accuracy, AbstractObject>>> getColumnsByTripleId(
			ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>> probabilties,
			ConcurrentHashMap<TableColumnId, Map<InstanceId, AbstractObject>> groupedTableColumnsFull) {

		ConcurrentHashMap<TripleId, Collection<Quad<TableColumnId, Integer, Accuracy, AbstractObject>>> groupedTableColumns = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(probabilties.entrySet(), (columnIdEntry) -> {
			TableColumnId columnId = columnIdEntry.getKey();
			PropertyId propertyId = repository.getMetaData(columnId.getTableId()).getProperty(columnId.getColumnIdx().getColumnIdx());
			ClassId classId = repository.getMetaData(columnId.getTableId()).getDbPediaClass();
			ClassPropertyId cpid = new ClassPropertyId(classId, propertyId);
			Integer year = repository.getTimeStamp(columnId, columnIdEntry.getValue().getFirst());

			Map<InstanceId, AbstractObject> map = groupedTableColumnsFull.get(columnIdEntry.getKey());

			if (year != null) {
				for (Entry<InstanceId, AbstractObject> entry : map.entrySet()) {
					CHMHelper.getFromCHM(groupedTableColumns, new TripleId(classId, propertyId, entry.getKey()), () -> new ConcurrentLinkedQueue<>())
							.add(new Quad<>(columnId, year, columnIdEntry.getValue().getSecond(), entry.getValue()));
				}
			}

		}, this.getClass());

		return groupedTableColumns;
	}

	private <E extends TripleIdContainer> Map<TableColumnId, Map<E, SingleValue>> groupByColumnId(AbstractGroupledValues<E> learningSet) {
		ConcurrentHashMap<TableColumnId, Map<E, SingleValue>> groupedValues = new ConcurrentHashMap<TableColumnId, Map<E, SingleValue>>();

		PersistentParallel4.runAndFinish(learningSet.entrySet(), (e) -> {

			for (SingleValue sv : e.getValue().getValues()) {
				Map<E, SingleValue> tableColumnIdMap = CHMHelper.getFromCHM(groupedValues,
						new TableColumnId(sv.getId().getTableId(), sv.getId().getColumnIdx()), () -> new ConcurrentHashMap<>());

				tableColumnIdMap.put(e.getKey(), sv);
			}
		}, "KBTCalculator - Group Values by TableColumnId");

		return groupedValues;
	}

	private TimestampScoresContainer getScores(SplitGroupedValues learningSet, SplitGroupedValues testingSet, ScoresContainer other, MaxDiffsContainer maxDiffs,
			boolean withTimestamps, boolean propagate, boolean withRowCell) {

		TimestampScores timestampScores = this.getTimestampScores(learningSet, withTimestamps, propagate, withRowCell);

		return new TimestampScoresContainer(timestampScores, other, testingSet, maxDiffs);
	}

	public TimestampScoresContainer getScoresWithoutRowCell(SplitGroupedValues learningSet, SplitGroupedValues testingSet, ScoresContainer other,
			MaxDiffsContainer maxDiffs, boolean propagate) {

		return this.getScores(learningSet, testingSet, other, maxDiffs, true, propagate, false);
		//
		// TimestampScores timestampScores =
		// this.getTimestampScores(learningSet, true, propagate);
		//
		// return new TimestampScoresContainer(timestampScores, other,
		// testingSet, maxDiffs);
	}

	public TimestampScoresContainer getScores(SplitGroupedValues learningSet, SplitGroupedValues testingSet, ScoresContainer other, MaxDiffsContainer maxDiffs,
			boolean propagate) {

		return this.getScores(learningSet, testingSet, other, maxDiffs, true, propagate, true);
		//
		// TimestampScores timestampScores =
		// this.getTimestampScores(learningSet, true, propagate);
		//
		// return new TimestampScoresContainer(timestampScores, other,
		// testingSet, maxDiffs);
	}

	public TimestampScoresContainer getScoresWithoutTimestamps(SplitGroupedValues learningSet, SplitGroupedValues testingSet, ScoresContainer other,
			MaxDiffsContainer maxDiffs) {
		return this.getScores(learningSet, testingSet, other, maxDiffs, false, false, true);

	}

	@FunctionalInterface
	public static interface MaxDiffsContainer {
		Optional<Integer> getMaxDiff(ClassPropertyId classPropertyId);
	}

	public static boolean kbtAsBackup = false;

	public class TimestampScoresContainer implements ScoresContainer {

		private final TimestampScores timestampScores;
		private final ScoresContainer kbtScores;
		private final SplitGroupedValues testingSet;
		private final MaxDiffsContainer maxDiffs;

		private TimestampScoresContainer(TimestampScores timestampScores, ScoresContainer scoresContainer, SplitGroupedValues testingSet,
				MaxDiffsContainer maxDiffs) {
			this.timestampScores = timestampScores;
			this.kbtScores = scoresContainer;
			this.testingSet = testingSet;
			this.maxDiffs = maxDiffs;
		}

		public Map<ClassPropertyId, Map<String, MutableInt>> evaluateMappingsCSV() throws IOException {

			return timestampScores.getTsCount();

		}

		@Override
		public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {
			return new TableScoresInterface(this, classId, propertyId, instanceId);
		}

		@Override
		public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx,
				RowIdx rowIdx) {
			return kbtScores.getScore(classId, propertyId, instanceId, tableId, columnIdx, rowIdx);
		}

		@Override
		public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx,
				RowIdx rowIdx) {
			return kbtScores.containsKey(classId, propertyId, instanceId, tableId, columnIdx, rowIdx);
		}

		public Probability getScore(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {

			Integer qualifiedYear = null;
			for (AbstractQualifier aq : qualifiers) {
				if (aq instanceof TimestampPointInTime) {
					qualifiedYear = ((TimestampPointInTime) aq).getYear();
				}
			}
			Double kbtScore = getKbtScore(tripleId, cellId, qualifiers);
			double accuracy = 0.0;
			int total = 0;
			if (qualifiedYear != null) {

				TimestampIdentifier tsi = timestampScores.getMapping(new TableColumnId(cellId));
				if (tsi != null) {
					Integer tableYear = null;
					if (TimestampType.PROPAGATED.equals(tsi.getType())) {
						tableYear = tsi.getPropogatedYear();
					} else {
						tableYear = repository.getTimeStamp(cellId, tsi);
					}

					if (tableYear != null) {
						Accuracy acc = timestampScores.getScore(new TableColumnId(cellId));
						int diff = Math.abs(tableYear.intValue() - qualifiedYear);
						Optional<Double> weight = getWeight(tripleId.getClassPropertyId(), diff);
						if (weight.isPresent()) {
							total += acc.getTotal();
							accuracy += acc.getAccuracy() * weight.get();
						}
						// switch (diff) {
						// case 0:
						//
						// total += acc.getTotal();
						// accuracy += acc.getAccuracy();
						// break;
						// case 1:
						// total += acc.getTotal();
						// accuracy += acc.getAccuracy() * 0.75;
						// break;
						// case 2:
						// total += acc.getTotal();
						// accuracy += acc.getAccuracy() * 0.5;
						// break;
						// case 3:
						// total += acc.getTotal();
						// accuracy += acc.getAccuracy() * 0.25;
						// break;
						// }

						// if (kbtScore != null && kbtScore >
						// this.getMaxMappedScore(tripleId, qualifiers)) {
						// accuracy = (accuracy + kbtScore) / 2;
						// }

					}

				} else if (kbtAsBackup && kbtScore != null && (!hasScore(tripleId, qualifiers))) {
					accuracy = kbtScore;
				}
			}

			return new Probability(new Accuracy(accuracy, total));

			// Integer qualifiedYear = null;
			// for (AbstractQualifier aq : qualifiers) {
			// if (aq instanceof TimestampPointInTime) {
			// qualifiedYear = ((TimestampPointInTime) aq).getYear();
			// }
			// }
			// // if (otherScore != null) {
			// // accuracy +=
			// otherScore.get(SimilarityType.DIRECT).getAccuracy();
			// // total += otherScore.get(SimilarityType.DIRECT).getTotal();
			// // }
			// Map<TimestampIdentifier, Accuracy> mappings =
			// timestampScores.getMappings(new TableColumnId(cellId));
			// if (mappings != null) {
			//
			// int fffsdf = 123;
			//
			// for (Entry<TimestampIdentifier, Accuracy> tsiEntry :
			// mappings.entrySet()) {
			// if (tsiEntry.getValue().getAccuracy() > 0.6) {
			// int fsdf = 12;
			// }
			// TimestampIdentifier tsi = tsiEntry.getKey();
			// if (tsi != null) {
			//
			// Integer tableYear;
			// if (TimestampType.PROPAGATED.equals(tsi.getType())) {
			// tableYear = tsi.getPropogatedYear();
			// if (qualifiedYear != null && !qualifiedYear.equals(tableYear)) {
			// tableYear = null;
			// }
			// } else {
			// tableYear = repository.getTimeStamp(cellId, tsi);
			//
			// }
			//
			// if (tableYear != null) {
			// for (AbstractQualifier qualifier : qualifiers) {
			// if (qualifier instanceof TimestampPointInTime) {
			// TimestampPointInTime tpit = (TimestampPointInTime) qualifier;
			// Accuracy acc = tsiEntry.getValue();
			// int diff = Math.abs(tableYear.intValue() - tpit.getYear());
			// switch (diff) {
			// case 0:
			//
			// total += acc.getTotal();
			// accuracy += acc.getAccuracy();
			// break;
			// // case 1:
			// // total += acc.getTotal();
			// // accuracy += acc.getAccuracy() * 0.75;
			// // break;
			// // case 2:
			// // total += acc.getTotal();
			// // accuracy += acc.getAccuracy() * 0.5;
			// // break;
			// // case 3:
			// // total += acc.getTotal();
			// // accuracy += acc.getAccuracy() * 0.25;
			// // break;
			// }
			// }
			// }
			// }
			//
			// }
			// }
			//
			// }
			//
			// if (accuracy > 0.0) {
			// return new Probability(new Accuracy(accuracy, total));
			// }
			//
			// return KBT;
		}

		// private Double getKbtScore(TripleId tripleId, CellId cellId,
		// Set<AbstractQualifier> qualifiers) {
		// Integer qualifiedYear = null;
		// for (AbstractQualifier aq : qualifiers) {
		// if (aq instanceof TimestampPointInTime) {
		// qualifiedYear = ((TimestampPointInTime) aq).getYear();
		// }
		// }
		//
		// if (qualifiedYear != null) {
		// ImmutableMap<TimestampIdentifier, Accuracy> mappings =
		// this.timestampScores.getSortedMappings(cellId.getTableColumnId());
		//
		// if (mappings != null) {
		// for (Entry<TimestampIdentifier, Accuracy> entry :
		// mappings.entrySet()) {
		// if (entry.getKey().getPropogatedYear().equals(qualifiedYear)) {
		// return entry.getValue().getAccuracy();
		// }
		// }
		// }
		//
		// }
		//
		// return null;
		// }

		private Optional<Double> getWeight(ClassPropertyId classPropertyId, int diff) {
			if (diff == 0) {
				return Optional.of(1.0);
			}

			Optional<Integer> maxDiff = maxDiffs.getMaxDiff(classPropertyId);

			if (!maxDiff.isPresent() || diff > maxDiff.get()) {
				return Optional.empty();
			} else {
				return Optional.of(1.0 - diff * 0.1);
			}

		}

		private Double getKbtScore(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {

			Probability kbtScore = this.kbtScores.getScore(tripleId, cellId, qualifiers);
			if (kbtScore != null) {
				return kbtScore.get(SimilarityType.DIRECT).getAccuracy();
			}

			return null;
		}

		private boolean hasScore(TripleId tripleId, Set<AbstractQualifier> qualifiers) {
			MatchedValues v = this.testingSet.get(new QualifiedTripleId(tripleId, qualifiers));
			if (v != null) {
				for (SingleValue sv : v.getValues()) {
					if (timestampScores.getMapping(new TableColumnId(sv.getId())) != null)
						return true;
				}
			}
			return false;
		}

		ConcurrentHashMap<QualifiedTripleId, Double> cacheMap = new ConcurrentHashMap<>();

		private double getMaxMappedScore(TripleId tripleId, Set<AbstractQualifier> qualifiers) {
			QualifiedTripleId qtid = new QualifiedTripleId(tripleId, qualifiers);
			if (cacheMap.containsKey(qtid))
				return cacheMap.get(qtid);

			MatchedValues v = this.testingSet.get(new QualifiedTripleId(tripleId, qualifiers));
			double maxScore = 0.0;
			if (v != null) {
				for (SingleValue sv : v.getValues()) {
					TableColumnId tcid = new TableColumnId(sv.getId());
					if (timestampScores.getMapping(tcid) != null) {

						double acc = this.timestampScores.getScore(tcid).getAccuracy();
						if (acc > maxScore) {

							TimestampIdentifier mapping = timestampScores.getMapping(new TableColumnId(sv.getId()));
							Integer tableYear = repository.getTimeStamp(sv.getId(), mapping);
							if (tableYear != null) {
								for (AbstractQualifier q : qualifiers) {
									if (q instanceof TimestampPointInTime) {
										Integer year = ((TimestampPointInTime) q).getYear();
										if (tableYear.equals(year)) {
											maxScore = acc;
										}
									}
								}
							}

						}

					}
				}
			}

			cacheMap.put(qtid, maxScore);
			return maxScore;
		}

		public boolean containsKey(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
			boolean ret = getScore(tripleId, cellId, qualifiers) != null;
			if (ret)
				return true;
			else
				return false;

		}

		public TimestampScores getTimestampScores() {
			return timestampScores;
		}

	}

	private static class CacheKey {
		private final UUID uuid;
		private final boolean withTimestamps;
		private final boolean propagte;
		private final boolean rowCell;

		private CacheKey(UUID uuid, boolean withTimestamps, boolean propagte, boolean rowCell) {
			super();
			this.uuid = uuid;
			this.withTimestamps = withTimestamps;
			this.propagte = propagte;
			this.rowCell = rowCell;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + (propagte ? 1231 : 1237);
			result = prime * result + (rowCell ? 1231 : 1237);
			result = prime * result + ((uuid == null) ? 0 : uuid.hashCode());
			result = prime * result + (withTimestamps ? 1231 : 1237);
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			CacheKey other = (CacheKey) obj;
			if (propagte != other.propagte)
				return false;
			if (rowCell != other.rowCell)
				return false;
			if (uuid == null) {
				if (other.uuid != null)
					return false;
			} else if (!uuid.equals(other.uuid))
				return false;
			if (withTimestamps != other.withTimestamps)
				return false;
			return true;
		}

	}
}
