package de.mannheim.uni.expansion.tablefusion.core.splitting;

import java.util.HashSet;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Consumer;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier.TimestampType;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.matching.SynchronizedMatchedValuesMapMerger;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimestampScorerCleanedUp.MaxDiffsContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimestampScores;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfigFilter;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;

public class TimestampMappingSplitter implements FilteredSplitter {

	private final Map<TableColumnId, Pair<TimestampIdentifier, Accuracy>> mappings;
	private final AnnotatedTablesRepository tableRepository;
	private final UUID scoresUUID;
	private final PropertiesConfigFilter propertiesConfigFilter;
	private final MaxDiffsContainer maxDiffsContainer;

	public TimestampMappingSplitter(TimestampScores timestampScores, AnnotatedTablesRepository tableRepository, PropertiesConfigFilter propertiesConfigFilter,
			MaxDiffsContainer maxDiffsContainer) {
		this.scoresUUID = timestampScores.getUUID();
		this.mappings = timestampScores.getMappings();
		this.tableRepository = tableRepository;
		this.propertiesConfigFilter = propertiesConfigFilter;
		this.maxDiffsContainer = maxDiffsContainer;

	}

	private static final SimpleLockableConcurrentCache<CacheKey, SplitGroupedValues> cache = new SimpleLockableConcurrentCache<>();

	@Override
	public SplitGroupedValues split(GroupedValues groupedValues, Set<QualifiedTripleId> allowedTripleIds) {
		return generateCache(groupedValues, allowedTripleIds);

		// CacheKey cacheKey = new CacheKey(scoresUUID,
		// groupedValues.getUUID());
		//
		// return cache.getOrCreate(cacheKey, () -> {
		// return generateCache(groupedValues, allowedTripleIds);
		// });

		// if (cache.hasCacheFor(cacheKey)) {
		// return cache.get(cacheKey);
		// } else {
		//
		// Optional<Consumer<SplitGroupedValues>> lock =
		// cache.getLock(cacheKey);
		// if (lock.isPresent()) {
		//
		// lock.get().accept(generateCache(groupedValues, allowedTripleIds));
		//
		// }
		//
		// return cache.poll(cacheKey);
		// }

	}

	@Override
	public Set<QualifiedPropertyId> split(ClassId classId, PropertyId propertyId, CellId cellId) {
		Set<QualifiedPropertyId> ret = new HashSet<>();
		Pair<TimestampIdentifier, Accuracy> pair = mappings.get(cellId.getTableColumnId());
		if (pair != null) {
			Integer tableYear = null;
			if (TimestampType.PROPAGATED.equals(pair.getFirst().getType())) {
				tableYear = pair.getFirst().getPropogatedYear();
			} else {
				tableYear = tableRepository.getTimeStamp(cellId, pair.getFirst());
			}
			if (tableYear != null) {

				Consumer<Integer> adder = (yearInt) -> {
					HashSet<AbstractQualifier> qualifierSet = new HashSet<>();
					qualifierSet.add(new TimestampPointInTime(yearInt));
					QualifiedPropertyId qualifierId = new QualifiedPropertyId(propertyId, qualifierSet);
					ret.add(qualifierId);

				};

				adder.accept(tableYear);

				Optional<Integer> maxDiff = this.maxDiffsContainer.getMaxDiff(new ClassPropertyId(classId, propertyId));
				if (maxDiff.isPresent())
					for (int i = 1; i <= maxDiff.get(); i++) {
						adder.accept(tableYear + i);
						adder.accept(tableYear - i);
					}

			}
		}

		return ret;
	}

	public SplitGroupedValues generateCache(GroupedValues groupedValues, Set<QualifiedTripleId> allowedTripleIds) {
		SynchronizedMatchedValuesMapMerger<QualifiedTripleId> merger = new SynchronizedMatchedValuesMapMerger<>();

		PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {

			for (SingleValue sv : e.getValue().getValues()) {

				Pair<TimestampIdentifier, Accuracy> pair = mappings.get(sv.getId().getTableColumnId());
				if (pair != null) {
					Integer tableYear = null;
					if (TimestampType.PROPAGATED.equals(pair.getFirst().getType())) {
						tableYear = pair.getFirst().getPropogatedYear();
					} else {
						tableYear = tableRepository.getTimeStamp(sv.getId(), pair.getFirst());
					}
					if (tableYear != null) {

						Consumer<Integer> adder = (yearInt) -> {
							HashSet<AbstractQualifier> qualifierSet = new HashSet<>();
							qualifierSet.add(new TimestampPointInTime(yearInt));
							QualifiedTripleId qualifierId = new QualifiedTripleId(e.getKey(), qualifierSet);
							if (allowedTripleIds == null || allowedTripleIds.contains(qualifierId)) {
								merger.merge(qualifierId, sv, e.getValue().getDataType());
							}
						};

						adder.accept(tableYear);

						Optional<Integer> maxDiff = this.maxDiffsContainer.getMaxDiff(e.getKey().getClassPropertyId());
						if (maxDiff.isPresent())
							for (int i = 1; i <= maxDiff.get(); i++) {
								adder.accept(tableYear + i);
								adder.accept(tableYear - i);
							}

					}
				}
			}

		} , "TIMESTAMPS PLITTER SPLITTER");

		SplitGroupedValues spgv = new SplitGroupedValues(merger.getCleanMap());
		return spgv;
	}

	private static class CacheKey extends Pair<UUID, UUID> {

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		public CacheKey(UUID timestampScores, UUID splitGroupedValues) {
			super(timestampScores, splitGroupedValues);
		}

	}

	@Override
	public SplitGroupedValues split(SplitGroupedValues sgvTesting) {

		SynchronizedMatchedValuesMapMerger<QualifiedTripleId> merger = new SynchronizedMatchedValuesMapMerger<>();

		PersistentParallel4.runAndFinish(sgvTesting.entrySet(), (e) -> {

			for (SingleValue sv : e.getValue().getValues()) {

				Pair<TimestampIdentifier, Accuracy> pair = mappings.get(sv.getId().getTableColumnId());
				if (pair != null) {
					Integer tableYear = null;
					if (TimestampType.PROPAGATED.equals(pair.getFirst().getType())) {
						tableYear = pair.getFirst().getPropogatedYear();
					} else {
						tableYear = tableRepository.getTimeStamp(sv.getId(), pair.getFirst());
					}
					if (tableYear != null) {
						HashSet<AbstractQualifier> qualifierSet = new HashSet<>();
						qualifierSet.add(new TimestampPointInTime(tableYear));
						QualifiedTripleId qualifierId = new QualifiedTripleId(e.getKey().tripleId(), qualifierSet);
						// if (allowedTripleIds.contains(qualifierId)) {
						merger.merge(qualifierId, sv, e.getValue().getDataType());
						// }
					}
				}
			}

		} , "TIMESTAMPS PLITTER SPLITTER");

		SplitGroupedValues spgv = new SplitGroupedValues(merger.getCleanMap());
		return spgv;
	}

	@Override
	public SplitGroupedValues split(GroupedValues groupedValues) {
		return generateCache(groupedValues, null);

	}

	@Override
	public SplitGroupedValues split(GroupedValues groupedValues, PropertiesConfig propertiesConfig) {
		SplitGroupedValues map = generateCache(groupedValues, null);

		ConcurrentHashMap<QualifiedTripleId, MatchedValues> finalMap = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(map.entrySet(), (e) -> {
			if (propertiesConfigFilter.include(propertiesConfig, e.getKey())) {
				finalMap.put(e.getKey(), e.getValue());
			}
		} , this.getClass());

		return new SplitGroupedValues(finalMap);
	}

}
