package de.mannheim.uni.expansion.tablefusion.learning;

import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.SynchronizedMatchedValuesMapMerger;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleConcurrentCache;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;

public class LearningSetCreatorStratified {

	@Inject
	public LearningSetCreatorStratified() {

	}

	private final static SimpleConcurrentCache<CacheKey, Map<Integer, SplitGroupedValuesSets>> cache = new SimpleConcurrentCache<>();

	private DecimalFormat dfmt = new DecimalFormat("0.00");

	public Map<Integer, SplitGroupedValuesSets> createByTableProperty(int numberOfSets, SplitGroupedValues splitGroupedValues) {
		CacheKey cacheKey = new CacheKey(splitGroupedValues.getUUID(), numberOfSets);

		if (cache.hasCacheFor(cacheKey)) {
			return cache.get(cacheKey);
		} else {
			double percent = 1.0 / numberOfSets;
			Map<Integer, SplitGroupedValuesSets> createdSets = new HashMap<Integer, SplitGroupedValuesSets>();

			ConcurrentAccumulator<AccumInteger> totalCount = new ConcurrentAccumulator<AccumInteger>(new AccumInteger(0));

			ConcurrentHashMap<InstanceId, AtomicInteger> instanceSize = new ConcurrentHashMap<>();
			ConcurrentHashMap<QualifiedPropertyId, AtomicInteger> propertySize = new ConcurrentHashMap<>();

			ConcurrentHashMap<InstanceId, Integer> instancesMax = new ConcurrentHashMap<>();
			ConcurrentHashMap<QualifiedPropertyId, Integer> propertiesMax = new ConcurrentHashMap<>();

			PersistentParallel6.runAndFinish(splitGroupedValues.entrySet(), (e) -> {
				CHMHelper.getFromCHM(propertySize, e.getKey().getQualifiedPropertyId(), () -> new AtomicInteger()).addAndGet(e.getValue().getValues().size());
				CHMHelper.getFromCHM(instanceSize, e.getKey().getTripleId().getInstanceId(), () -> new AtomicInteger())
						.addAndGet(e.getValue().getValues().size());
				totalCount.add(new AccumInteger(e.getValue().getValues().size()));
			} , LearningSetCreatorStratified.class);

			PersistentParallel6.runAndFinish(propertySize.entrySet(), (e) -> {
				propertiesMax.put(e.getKey(), (int) (percent * e.getValue().get()));
			} , LearningSetCreatorStratified.class);

			PersistentParallel6.runAndFinish(instanceSize.entrySet(), (e) -> {
				instancesMax.put(e.getKey(), (int) (percent * e.getValue().get()));
			} , LearningSetCreatorStratified.class);

			int totalCountFinal = totalCount.get().get();
			int totalGroups = splitGroupedValues.size();
			System.out.println("Total Values: " + totalCountFinal);
			System.out.println("Total Groups: " + totalGroups);
			TreeMap<QualifiedTripleId, MatchedValues> splitGroupedValuesSorted = new TreeMap<>(splitGroupedValues.getValues());
			for (int i = 0; i < numberOfSets; i++) {

				ConcurrentAccumulator<AccumInteger> inTestingSet = new ConcurrentAccumulator<AccumInteger>(new AccumInteger(0));
				ConcurrentAccumulator<AccumInteger> groupsInTestingSet = new ConcurrentAccumulator<AccumInteger>(new AccumInteger(0));

				ConcurrentHashMap<InstanceId, AtomicInteger> instancesDone = new ConcurrentHashMap<>();
				ConcurrentHashMap<QualifiedPropertyId, AtomicInteger> propertiesDone = new ConcurrentHashMap<>();

				SynchronizedMatchedValuesMapMerger<QualifiedTripleId> learningSetMerger = new SynchronizedMatchedValuesMapMerger<>();
				SynchronizedMatchedValuesMapMerger<QualifiedTripleId> testingSetMerger = new SynchronizedMatchedValuesMapMerger<>();

				int s = 0;
				for (Entry<QualifiedTripleId, MatchedValues> vwi : splitGroupedValuesSorted.entrySet()) {

					if (s == i) {
						int size = vwi.getValue().getValues().size();
						int instanceCurrent = CHMHelper.getFromCHM(instancesDone, vwi.getKey().getTripleId().getInstanceId(), () -> new AtomicInteger())
								.getAndAdd(size);
						int propertyCurrent = CHMHelper.getFromCHM(propertiesDone, vwi.getKey().getQualifiedPropertyId(), () -> new AtomicInteger())
								.getAndAdd(size);

						int instancesMaxCount = instancesMax.get(vwi.getKey().getTripleId().getInstanceId());
						int propertiesMaxCount = propertiesMax.get(vwi.getKey().getQualifiedPropertyId());
						if ((instanceCurrent <= instancesMaxCount && propertyCurrent <= propertiesMaxCount)) {

							testingSetMerger.merge(vwi.getKey(), vwi.getValue());
							inTestingSet.add(new AccumInteger(size));
							groupsInTestingSet.add(new AccumInteger(1));
						} else {
							learningSetMerger.merge(vwi.getKey(), vwi.getValue());
						}
					} else {
						learningSetMerger.merge(vwi.getKey(), vwi.getValue());
					}

					s++;
					if (s == numberOfSets) {
						s = 0;
					}
				}

				int inTestingSetFinal = inTestingSet.get().get();
				int groupsInTestingSetFinal = groupsInTestingSet.get().get();
				double testingSetPercent = ((double) inTestingSetFinal) / totalCountFinal;
				double groupsInTestingSetPercent = ((double) groupsInTestingSetFinal) / totalGroups;
				System.out.println("=================");
				System.out.println("Finished One Set.");
				System.out.println("Values in testing set: " + inTestingSetFinal);
				System.out.println("Percent of total values in set: " + dfmt.format(testingSetPercent));
				System.out.println("Grioups in testing set: " + groupsInTestingSetFinal);
				System.out.println("Percent of total groups in set: " + dfmt.format(groupsInTestingSetPercent));

				createdSets.put(i, new SplitGroupedValuesSets(splitGroupedValues, new SplitGroupedValues(learningSetMerger.getCleanMap()),
						new SplitGroupedValues(testingSetMerger.getCleanMap())));

			}

			cache.add(cacheKey, createdSets);
			return createdSets;
		}
	}

	private static class CacheKey {
		private final UUID uuid;
		private final int numberOfSets;

		private CacheKey(UUID uuid, int numberOfSets) {
			super();
			this.uuid = uuid;
			this.numberOfSets = numberOfSets;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + numberOfSets;
			result = prime * result + ((uuid == null) ? 0 : uuid.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			CacheKey other = (CacheKey) obj;
			if (numberOfSets != other.numberOfSets)
				return false;
			if (uuid == null) {
				if (other.uuid != null)
					return false;
			} else if (!uuid.equals(other.uuid))
				return false;
			return true;
		}

	}
}
