package de.mannheim.uni.expansion.tablefusion.profiling;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.instances.QualifierId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.FusionPerformanceSynchronizedCreator;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.generic.GenericResultsWriterAccumRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.utils.DomainUtils;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumAverage;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumMedian;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AcumReference;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable.ImmutableAccumulatorListBuild;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnListBuilder;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult.ResultBuilder;

public class PldProfiler {

	private final AnnotatedTablesRepository tablesRepository;
	private final OntologyRepository ontologyRepository;

	@Inject
	public PldProfiler(AnnotatedTablesRepository tablesRepository, OntologyRepository ontologyRepository) {
		this.tablesRepository = tablesRepository;
		this.ontologyRepository = ontologyRepository;
	}

	// private static final Pair<Integer, String> TOTAL_MAPPED_VALUES = new
	// Pair<>(0, "TotalMappedValues");
	// private static final Pair<Integer, String> ONE_VALUE_PER_PLD = new
	// Pair<>(1, "OneValuePerPLD");
	// private static final Pair<Integer, String> AVERAGE_PER_PLD = new
	// Pair<>(2, "AveragePerPLD");
	// private static final Pair<Integer, String> MEDIAN_PER_PLD = new Pair<>(3,
	// "MedianPerPLD");

	public void profileByPld(String outputFolder, String outputPrefix) {

		ImmutableAccumulatorListBuild builder = AccumulatorListImmutable.builder();

		AcumReference<AccumInteger> TOTAL_MAPPED_VALUES = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> ONE_VALUE_PER_PLD = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumAverage> AVERAGE_PER_PLD = builder.addAcummulator(AccumAverage.class, new AccumAverage());
		AcumReference<AccumMedian> MEDIAN_PER_PLD = builder.addAcummulator(AccumMedian.class, new AccumMedian());

		ColumnListBuilder columnListBuilder = ColumnList.builderAccumRepository();
		columnListBuilder.addColumn("TotalMappedValues", (repository) -> {
			return String.valueOf(repository.get(TOTAL_MAPPED_VALUES).get());
		});

		columnListBuilder.addColumn("OneValuePerPLD", (repository) -> {
			return String.valueOf(repository.get(ONE_VALUE_PER_PLD).get());
		});

		columnListBuilder.addColumn("AveragePerPLD", (repository) -> {
			return String.valueOf(repository.get(AVERAGE_PER_PLD).getValue());
		});

		columnListBuilder.addColumn("MEDIAN_PER_PLD", (repository) -> {
			return String.valueOf(repository.get(MEDIAN_PER_PLD).getValue());
		});

		ColumnList columnList = columnListBuilder.build();
		AccumulatorListImmutable resultsType = builder.getResultsType();

		FusionPerformanceSynchronizedCreator<GenericAccumaltableResult> creatorSingle = new FusionPerformanceSynchronizedCreator<>(
				GenericAccumaltableResult.empty(resultsType));

		ConcurrentHashMap<TripleId, ConcurrentHashMap<String, AtomicInteger>> pldCount = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(tablesRepository.getAllTableIds(), (tid) -> {

			AnnotatedTableMetaData md = this.tablesRepository.getMetaData(tid);
			AnnotatedTable td = this.tablesRepository.getTable(tid);
			String source = md.getSource().getSource();
			String PLD = DomainUtils.getPayLevelDomainFromWholeURL(source);
			if (PLD != null) {
				for (Entry<Integer, List<PropertyId>> propertyIdEntry : md.getAllProperties().entrySet()) {
					PropertyId propertyId = propertyIdEntry.getValue().get(0);
					if (!propertyId.equals(new PropertyId("ID"))) {

						if (!(propertyId instanceof QualifierId)) {

							for (Entry<Integer, List<InstanceId>> instanceIdEntry : md.getAllInstances().entrySet()) {
								InstanceId instanceId = instanceIdEntry.getValue().get(0);
								ConcurrentHashMap<String, AtomicInteger> pldMap = CHMHelper.getFromCHM(pldCount,
										new TripleId(md.getDbPediaClass(), propertyId, instanceId), () -> new ConcurrentHashMap<>());

								if (td.getValue(new RowIdx(instanceIdEntry.getKey()), new ColumnIdx(propertyIdEntry.getKey())) != null) {

									CHMHelper.getFromCHM(pldMap, PLD, () -> new AtomicInteger()).incrementAndGet();

								}
							}
						}
					}
				}
			}

		} , this.getClass());

		PersistentParallel6.runAndFinish(pldCount.entrySet(), (entry) -> {

			PersistentParallel6.runAndFinish(entry.getValue().entrySet(), (pldEntry) -> {

				ResultBuilder resultBuilder = GenericAccumaltableResult.builder(resultsType);

				resultBuilder.setValue(TOTAL_MAPPED_VALUES, new AccumInteger(pldEntry.getValue().get()));
				resultBuilder.setValue(ONE_VALUE_PER_PLD, new AccumInteger(1));
				resultBuilder.setValue(AVERAGE_PER_PLD, new AccumAverage(pldEntry.getValue().doubleValue()));
				resultBuilder.setValue(MEDIAN_PER_PLD, new AccumMedian(pldEntry.getValue().doubleValue()));

				GenericAccumaltableResult result = resultBuilder.build();

				creatorSingle.addForAll(entry.getKey().getClassPropertyId(), ontologyRepository, result);

			} , this.getClass());

		} , this.getClass());

		GenericResultsWriterAccumRepository writer = new GenericResultsWriterAccumRepository(columnList);
		writer.writeAccumRepository("PLDProfiling", ontologyRepository, creatorSingle.get());

		try {
			if (outputFolder != null && outputPrefix != null) {
				MKDIR.createDirUnchecked(new File(outputFolder));
				writer.writeToFolder(outputFolder, outputPrefix);

			}
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		// Set<String> uniquePLDs = new HashSet<>();
		// for (SingleValue sv : e.getValue().getValues()) {
		// String source =
		// annotatedTablesRepository.getMetaData(sv.getId().getTableId()).getSource().getSource();
		//
		// String PLD = DomainUtils.getPayLevelDomainFromWholeURL(source);
		// uniquePLDs.add(PLD);
		//
		// }

	}
}
