package de.mannheim.uni.expansion.tablefusion.profiling;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.util.Collections;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.inject.Inject;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.splitting.Splitter;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitterFactory;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitterFactoryRepository;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;

public class PropertiesQualityChecker {

	private final GroupedValuesRepository groupedValuesRepository;
	private final AnnotatedTablesRepository tableRepository;
	private final KnowledgeBase knowledgeBase;
	private final SimilarityCalculator similarityCalculator;
	private final SplitterFactoryRepository splitterFactoriesRepository;
	private final TimestampQualifiersAdapter timestampQualifiersAdapter;
	private final OntologyRepository ontologyRepository;

	@Inject
	public PropertiesQualityChecker(GroupedValuesRepository groupedValuesRepository, AnnotatedTablesRepository tableRepository, KnowledgeBase knowledgeBase,
			SimilarityCalculator similarityCalculator, SplitterFactoryRepository splitterFactoriesRepository,
			TimestampQualifiersAdapter timestampQualifiersAdapter, OntologyRepository ontologyRepository) {
		this.groupedValuesRepository = groupedValuesRepository;
		this.tableRepository = tableRepository;
		this.knowledgeBase = knowledgeBase;
		this.similarityCalculator = similarityCalculator;
		this.splitterFactoriesRepository = splitterFactoriesRepository;
		this.timestampQualifiersAdapter = timestampQualifiersAdapter;
		this.ontologyRepository = ontologyRepository;
	}

	public void run(FusionConfig fusionConfig) throws IOException {
		run(fusionConfig, null, null);

	}

	public void run(FusionConfig fusionConfig, String outputFolder, String outputPrefix) throws IOException {

		GroupedValues groupedValues = groupedValuesRepository.computeGroupedValuesDirect(fusionConfig);

		ConcurrentHashMap<ClassPropertyId, AtomicInteger> timeseriesCount = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassPropertyId, AtomicInteger> totalCount = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassPropertyId, AtomicInteger> pldsValueCount = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassPropertyId, AtomicInteger> overlapMap = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassPropertyId, Set<String>> pldsCount = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(groupedValues.entrySet(), (e) -> {
			CHMHelper.getFromCHM(timeseriesCount, e.getKey().getClassPropertyId(), () -> new AtomicInteger()).incrementAndGet();
			CHMHelper.getFromCHM(totalCount, e.getKey().getClassPropertyId(), () -> new AtomicInteger()).addAndGet(e.getValue().getValues().size());

			Set<String> plds = e.getValue().getValues().stream().map((sv) -> {
				return tableRepository.getMetaData(sv.getId().getTableId()).getSource().getPLD();
			}).filter((sv) -> {
				return sv != null;
			}).distinct().collect(Collectors.toSet());

			long pldCount = plds.size();

			CHMHelper.getFromCHM(pldsCount, e.getKey().getClassPropertyId(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).addAll(plds);

			CHMHelper.getFromCHM(pldsValueCount, e.getKey().getClassPropertyId(), () -> new AtomicInteger()).addAndGet((int) pldCount);
			int overlap = 0;
			for (SingleValue sv : e.getValue().getValues()) {

				QualifiedValueList qvl = knowledgeBase.getQualifiedValueListWithEqualProperties(e.getKey().getValueId());
				if (qvl != null) {
					for (QualifiedValueHolder qvh : qvl.getClaims()) {

						if (similarityCalculator.isEqual(qvh.getValue(), sv.getValue())) {
							overlap++;
							break;
						}
					}

				}
			}

			CHMHelper.getFromCHM(overlapMap, e.getKey().getClassPropertyId(), () -> new AtomicInteger()).addAndGet(overlap);
			// if (pldCount >= fusionConfig.getMinSource()) {
			// CHMHelper.getFromCHM(count, e.getKey().getClassPropertyId(), ()
			// -> new AtomicInteger()).incrementAndGet();
			// }
		}, this.getClass());

		SplitterFactory splitterFactory = this.splitterFactoriesRepository.get(fusionConfig.getSplitterFactory());

		Splitter splitter = splitterFactory.getSplitter(groupedValues, groupedValues, groupedValues);

		SplitGroupedValues splitGroupedValues = splitter.split(groupedValues);

		ConcurrentHashMap<ClassPropertyId, AtomicInteger> maxCorrect = new ConcurrentHashMap<>();

		ConcurrentHashMap<ClassPropertyId, Set<Pair<ValueId, Integer>>> mappedValueCount = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassPropertyId, AtomicInteger> mappedValueCountInteger = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(splitGroupedValues.entrySet(), (e) -> {

			CHMHelper.getFromCHM(mappedValueCountInteger, e.getKey().tripleId().getClassPropertyId(), () -> new AtomicInteger()).incrementAndGet();

			QualifiedValueList qvl = knowledgeBase.getQualifiedValueListWithEqualProperties(e.getKey().tripleId().getValueId());

			AtomicInteger ai = CHMHelper.getFromCHM(maxCorrect, e.getKey().tripleId().getClassPropertyId(), () -> new AtomicInteger(0));
			if (qvl != null) {
				Map<Integer, AbstractObject> correctValues = this.timestampQualifiersAdapter.getForQualifiersMap(qvl, e.getKey().getQualifiers());

				Set<Pair<ValueId, Integer>> set = CHMHelper.getFromCHM(mappedValueCount, e.getKey().tripleId().getClassPropertyId(),
						() -> Collections.newSetFromMap(new ConcurrentHashMap<>()));

				for (Integer index : correctValues.keySet()) {
					set.add(new Pair<>(e.getKey().tripleId().getValueId(), index));
				}

				boolean correct = false;
				for (Entry<Integer, AbstractObject> correctValueEntry : correctValues.entrySet()) {
					for (SingleValue sv : e.getValue().getValues()) {
						if (similarityCalculator.isEqual(correctValueEntry.getValue(), sv.getValue())) {
							correct = true;
							break;
						}
					}
					if (correct) {
						break;
					}
				}
				if (correct) {
					ai.incrementAndGet();
				}
			}

		}, this.getClass());

		StringWriter sw = new StringWriter();
		CSVWriter writer = new CSVWriter(sw);
		int length = 11;
		{
			String[] headerLine = new String[length];

			headerLine[0] = "ClassId";
			headerLine[1] = "PropertyId";
			headerLine[2] = "PropertyId";
			headerLine[3] = "TimeDependent";
			headerLine[4] = "PLDsCount";
			headerLine[5] = "CountTSAboveMinSource";
			headerLine[6] = "NumMatched";
			headerLine[7] = "OnePerPLD";
			headerLine[8] = "Overlap";
			headerLine[9] = "MaxCorrectQualified";
			headerLine[10] = "MappedValueCount";
			writer.writeNext(headerLine);
		}
		for (Entry<ClassPropertyId, AtomicInteger> entry : timeseriesCount.entrySet()) {
			String[] line = new String[length];
			line[0] = entry.getKey().getClassId().toString();
			line[1] = entry.getKey().getPropertyId().getPropertyId();
			line[2] = entry.getKey().getPropertyId().readableString();
			line[3] = (this.ontologyRepository.isTimeDependent(entry.getKey().getPropertyId()) ? "YES" : "NO");
			line[4] = String.valueOf(pldsCount.get(entry.getKey()).size());
			line[5] = String.valueOf(entry.getValue().intValue());
			line[6] = String.valueOf(totalCount.get(entry.getKey()).intValue());
			line[7] = String.valueOf(pldsValueCount.get(entry.getKey()).intValue());
			line[8] = String.valueOf(overlapMap.get(entry.getKey()).intValue());
			line[9] = String.valueOf(maxCorrect.getOrDefault(entry.getKey(), new AtomicInteger(0)).intValue());
			line[10] = String.valueOf(mappedValueCountInteger.getOrDefault(entry.getKey(), new AtomicInteger(0)).intValue());
			writer.writeNext(line);
		}
		writer.close();
		System.out.println(sw.toString());

		if (outputFolder != null && outputPrefix != null) {
			MKDIR.createDirUnchecked(new File(outputFolder));

			FileWriter fw = new FileWriter(new File(outputFolder + "/" + outputPrefix + "_quality.csv"));
			fw.write(sw.toString());
			fw.close();

		}

	}
}
