package de.mannheim.uni.expansion.tablefusion.profiling;

import java.io.File;
import java.io.FileWriter;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.MapSorter;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.utils.DomainUtils;

public class TopKpld {

	private final GroupedValuesRepository groupedValuesRepository;
	private final OntologyRepository ontologyRepository;
	private final AnnotatedTablesRepository tableRepository;

	@Inject
	public TopKpld(GroupedValuesRepository groupedValuesRepository, OntologyRepository ontologyRepository, AnnotatedTablesRepository tableRepository) {
		this.groupedValuesRepository = groupedValuesRepository;
		this.ontologyRepository = ontologyRepository;
		this.tableRepository = tableRepository;
	}

	public void profileByPld(String outputFolder, String outputPrefix) {
		MKDIR.createDirUnchecked(new File(outputFolder));
		ConcurrentHashMap<ClassPropertyId, ConcurrentHashMap<String, AtomicInteger>> pldTotalCount = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassPropertyId, ConcurrentHashMap<String, AtomicInteger>> pldOnePerValueTotalCount = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassPropertyId, AtomicInteger> total = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassPropertyId, AtomicInteger> onePerValue = new ConcurrentHashMap<>();

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();

		PersistentParallel6.runAndFinish(groupedValues.entrySet(), (group) -> {
			Map<String, MutableInt> pldMap = new HashMap<>();

			for (SingleValue sv : group.getValue().getValues()) {
				AnnotatedTableMetaData md = this.tableRepository.getMetaData(sv.getId().getTableId());
				String source = md.getSource().getSource();
				String PLD = DomainUtils.getPayLevelDomainFromWholeURL(source);

				if (PLD != null) {
					CHMHelper.getFromMap(pldMap, PLD, () -> new MutableInt()).add(1);
				}

			}

			ConcurrentHashMap<String, AtomicInteger> totalMap = CHMHelper.getFromCHM(pldTotalCount, group.getKey().getClassPropertyId(),
					() -> new ConcurrentHashMap<String, AtomicInteger>());

			ConcurrentHashMap<String, AtomicInteger> onePerPLDMap = CHMHelper.getFromCHM(pldOnePerValueTotalCount, group.getKey().getClassPropertyId(),
					() -> new ConcurrentHashMap<String, AtomicInteger>());

			int totalValueCount = 0;
			int onePerValueCount = 0;
			for (Entry<String, MutableInt> pldEntry : pldMap.entrySet()) {
				CHMHelper.getFromCHM(totalMap, pldEntry.getKey(), () -> new AtomicInteger()).addAndGet(pldEntry.getValue().intValue());
				totalValueCount += pldEntry.getValue().intValue();

				CHMHelper.getFromCHM(onePerPLDMap, pldEntry.getKey(), () -> new AtomicInteger()).addAndGet(1);
				onePerValueCount += 1;

			}

			CHMHelper.getFromCHM(total, group.getKey().getClassPropertyId(), () -> new AtomicInteger()).addAndGet(totalValueCount);
			CHMHelper.getFromCHM(onePerValue, group.getKey().getClassPropertyId(), () -> new AtomicInteger()).addAndGet(onePerValueCount);

		} , this.getClass());

		int length = 9;
		String[] header = new String[length];
		header[0] = "PLD";
		header[1] = "TotalValues";
		header[2] = "Proportion";
		header[3] = "";
		header[4] = "PLD";
		header[5] = "TotalGroups";
		header[6] = "Proportion";
		header[7] = "TotalValues";
		header[8] = "Proportion";

		PersistentParallel6.runAndFinish(total.keySet(), (cpid) -> {
			Map<String, Integer> pldTotalMap = pldTotalCount.get(cpid).entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return e.getValue().intValue();
			}));

			Map<String, Integer> pldOnePerValueMap = pldOnePerValueTotalCount.get(cpid).entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return e.getValue().intValue();
			}));
			LinkedHashMap<String, Integer> pldTotalMapSorted = MapSorter.sortHashMapByValuesD(pldTotalMap);
			LinkedHashMap<String, Integer> pldOnePerValueMapSorted = MapSorter.sortHashMapByValuesD(pldOnePerValueMap);

			int totalCount = total.get(cpid).intValue();
			int onePerValueCount = onePerValue.get(cpid).intValue();

			try {
				CSVWriter writer = new CSVWriter(new FileWriter(new File(
						outputFolder + "/" + outputPrefix + "_" + cpid.getClassId().getClassId() + "_" + cpid.getPropertyId().getReadableString() + ".csv")));
				try {
					writer.writeNext(header);

					Iterator<Entry<String, Integer>> pldTotalIterator = pldTotalMapSorted.entrySet().iterator();
					Iterator<Entry<String, Integer>> pldOnePerValueIterator = pldOnePerValueMapSorted.entrySet().iterator();

					for (int i = 0; i < 10; i++) {
						String[] line = new String[length];
						if (pldTotalIterator.hasNext()) {
							Entry<String, Integer> entry = pldTotalIterator.next();
							line[0] = entry.getKey();
							line[1] = String.valueOf(entry.getValue());
							line[2] = String.valueOf(entry.getValue().doubleValue() / totalCount);
						} else {
							line[0] = "";
							line[1] = "";
							line[2] = "";
						}

						line[3] = "";

						if (pldOnePerValueIterator.hasNext()) {
							Entry<String, Integer> entry = pldOnePerValueIterator.next();
							line[4] = entry.getKey();
							line[5] = String.valueOf(entry.getValue());
							line[6] = String.valueOf(entry.getValue().doubleValue() / onePerValueCount);
							Integer totalEntry = pldTotalMap.get(entry.getKey());
							if (totalEntry != null) {
								line[7] = String.valueOf(totalEntry);
								line[8] = String.valueOf(totalEntry.doubleValue() / totalCount);
							} else {
								line[7] = "";
								line[8] = "";
							}
						} else {
							line[4] = "";
							line[5] = "";
							line[6] = "";
						}

						writer.writeNext(line);
					}
				} finally {
					writer.close();
				}

			} catch (Exception e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}

		} , this.getClass());

	}
}
