package de.mannheim.uni.expansion.tablefusion.profiling.distribution;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Map.Entry;
import java.util.OptionalDouble;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import com.opencsv.CSVWriter;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.instances.QualifierId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.PersistentParallel6;

public class AnnotatedTablesProfiler {

	private final AnnotatedTablesRepository repository;
	private final String resultsFolder;

	public AnnotatedTablesProfiler(AnnotatedTablesRepository repository, String resultsFolder) {
		this.resultsFolder = resultsFolder;
		this.repository = repository;
	}

	private final AtomicInteger totalTables = new AtomicInteger();
	private final ConcurrentHashMap<ClassId, AtomicInteger> classCount = new ConcurrentHashMap<ClassId, AtomicInteger>();
	private final ConcurrentHashMap<PropertyId, AtomicInteger> propertyCount = new ConcurrentHashMap<PropertyId, AtomicInteger>();
	private final ConcurrentHashMap<InstanceId, AtomicInteger> instanceCount = new ConcurrentHashMap<InstanceId, AtomicInteger>();

	private final ConcurrentHashMap<ClassPropertyId, AtomicInteger> compositeTableCount = new ConcurrentHashMap<ClassPropertyId, AtomicInteger>();
	private final ConcurrentHashMap<ClassPropertyId, AtomicInteger> compositeValueCount = new ConcurrentHashMap<ClassPropertyId, AtomicInteger>();
	private final ConcurrentHashMap<ClassPropertyId, ConcurrentHashMap<InstanceId, AtomicInteger>> compositeInstanceCount = new ConcurrentHashMap<ClassPropertyId, ConcurrentHashMap<InstanceId, AtomicInteger>>();

	private <E> void increment(ConcurrentHashMap<E, AtomicInteger> countMap, E key, int value) {
		AtomicInteger ai = countMap.get(key);
		if (ai != null) {
			ai.addAndGet(value);
		} else {
			ai = new AtomicInteger(value);
			AtomicInteger ret = countMap.putIfAbsent(key, ai);
			if (ret != null) {
				countMap.get(key).addAndGet(value);
			}
		}
	}

	private void increment(ClassId dbpediaClass) {
		increment(classCount, dbpediaClass, 1);
	}

	private void increment(PropertyId dbpediaProperty) {
		increment(propertyCount, dbpediaProperty, 1);
	}

	private void increment(InstanceId dbpediaInstnace) {
		increment(instanceCount, dbpediaInstnace, 1);
	}

	private void incrementTableCount(ClassPropertyId compositeKey) {
		increment(compositeTableCount, compositeKey, 1);
	}

	private void incrementValueCount(ClassPropertyId compositeKey, int value) {
		increment(compositeValueCount, compositeKey, value);
	}

	private void incrementCompositeInstanceValueCount(ClassPropertyId compositeKey, InstanceId instanceId) {
		ConcurrentHashMap<InstanceId, AtomicInteger> ai = compositeInstanceCount.get(compositeKey);
		if (ai == null) {
			ai = new ConcurrentHashMap<InstanceId, AtomicInteger>();
			ConcurrentHashMap<InstanceId, AtomicInteger> ret = compositeInstanceCount.putIfAbsent(compositeKey, ai);
			if (ret != null) {
				ai = compositeInstanceCount.get(compositeKey);
			}
		}

		increment(ai, instanceId, 1);
	}

	public void run() {

		PersistentParallel6.runAndFinish(repository.getAllTableIds(), (tid) -> {
			AnnotatedTableMetaData data = repository.getMetaData(tid);

			totalTables.incrementAndGet();

			if (data.getDbPediaClass() != null)
				increment(data.getDbPediaClass());

			int instanceCount = 0;
			for (InstanceId dbpediaInstnace : data.getInstances().values()) {
				if (dbpediaInstnace != null) {
					increment(dbpediaInstnace);
					instanceCount++;
				}
			}

			for (PropertyId dbpediaProperty : data.getProperties().values()) {
				if (dbpediaProperty != null && !(dbpediaProperty instanceof QualifierId)) {
					increment(dbpediaProperty);

					incrementTableCount(new ClassPropertyId(data.getDbPediaClass(), dbpediaProperty));
					incrementValueCount(new ClassPropertyId(data.getDbPediaClass(), dbpediaProperty), instanceCount);

					for (InstanceId dbpediaInstnace : data.getInstances().values()) {
						if (dbpediaInstnace != null) {
							incrementCompositeInstanceValueCount(new ClassPropertyId(data.getDbPediaClass(), dbpediaProperty), dbpediaInstnace);
						}

					}

				}
			}
		} , this.getClass());
		// repository.runForEachTable(new Consumer<AnnotatedTablesReaderData>()
		// {
		//
		// @Override
		// public void accept(AnnotatedTablesReaderData fullData) {
		//
		//
		//
		// }
		// });

		try {
			CSVWriter writerClasses = new CSVWriter(new FileWriter(new File(resultsFolder + "/" + "classes.csv")));

			for (Entry<ClassId, AtomicInteger> e : classCount.entrySet()) {
				String[] line = new String[2];
				line[0] = e.getKey().toString();
				line[1] = e.getValue().toString();
				writerClasses.writeNext(line);
			}

			writerClasses.close();

			CSVWriter writerProperties = new CSVWriter(new FileWriter(new File(resultsFolder + "/" + "properties.csv")));

			for (Entry<PropertyId, AtomicInteger> e : propertyCount.entrySet()) {
				String[] line = new String[2];
				line[0] = e.getKey().toString();
				line[1] = e.getValue().toString();
				writerProperties.writeNext(line);
			}

			writerProperties.close();

			CSVWriter writerInstances = new CSVWriter(new FileWriter(new File(resultsFolder + "/" + "instances.csv")));

			for (Entry<InstanceId, AtomicInteger> e : instanceCount.entrySet()) {
				String[] line = new String[2];
				line[0] = e.getKey().getInstanceId();
				line[1] = e.getValue().toString();
				writerInstances.writeNext(line);
			}

			writerInstances.close();

			CSVWriter compositeWriter = new CSVWriter(new FileWriter(new File(resultsFolder + "/" + "composite.csv")));

			for (Entry<ClassPropertyId, AtomicInteger> e : compositeTableCount.entrySet()) {
				String[] line = new String[7];
				line[0] = e.getKey().getClassId().toString();
				line[1] = e.getKey().getPropertyId().readableString();
				line[2] = e.getValue().toString();
				line[3] = compositeValueCount.get(e.getKey()).toString();

				line[4] = String.valueOf(compositeInstanceCount.get(e.getKey()).size());

				OptionalDouble optionalAverage = compositeInstanceCount.get(e.getKey()).entrySet().stream().mapToInt(d -> {
					return d.getValue().get();
				}).average();
				if (optionalAverage.isPresent()) {
					line[5] = String.valueOf(optionalAverage.getAsDouble());
				} else {
					line[5] = "";
				}

				int[] valueList = compositeInstanceCount.get(e.getKey()).entrySet().stream().mapToInt(d -> {
					return d.getValue().get();
				}).sorted().toArray();
				if (valueList.length > 0) {
					double median;
					if (valueList.length % 2 == 0) {
						int idx1 = valueList.length / 2;
						int idx2 = idx1 - 1;
						median = ((double) valueList[idx1] + valueList[idx2]) / 2.0;
					} else {
						int idx1 = valueList.length / 2;
						median = valueList[idx1];
					}

					line[6] = String.valueOf(median);
				} else {
					line[6] = "";
				}

				compositeWriter.writeNext(line);
			}

			compositeWriter.close();

		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		System.out.println();
	}

	// public static void main(String[] args) {
	// TableConverter tableConverter = new TableConverter(new ValueConverter());
	//
	// AnnotatedTablesProfiler pat = new
	// AnnotatedTablesProfiler("C:/Users/yaser/SampleRuns/Wikidata/profiling",
	// "C:/Users/yaser/SampleRuns/Wikidata/tables/temp", tableConverter);
	// pat.run();
	// }
}
