package de.mannheim.uni.expansion.tablefusion.profiling.timestamps;

import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;

public class TimeStampDistributionProfiler {
	private final AnnotatedTablesRepository tableRepository;
	private final KnowledgeBase knowledgeBase;
	private final GroupedValuesRepository groupedValuesRepository;
	private final SimilarityCalculator similarityCalculator;
	private final OntologyRepository ontologyRepository;
	private final TimestampQualifiersAdapter timestampQualifiersAdapter;

	@Inject
	public TimeStampDistributionProfiler(AnnotatedTablesRepository tableRepository, KnowledgeBase knowledgeBase,
			GroupedValuesRepository groupedValuesRepository, SimilarityCalculator similarityCalculator, OntologyRepository ontologyRepository,
			TimestampQualifiersAdapter timestampQualifiersAdapter) {
		this.tableRepository = tableRepository;
		this.knowledgeBase = knowledgeBase;
		this.groupedValuesRepository = groupedValuesRepository;
		this.similarityCalculator = similarityCalculator;
		this.ontologyRepository = ontologyRepository;
		this.timestampQualifiersAdapter = timestampQualifiersAdapter;
	}

	public void profileKB(String resultsFile) throws IOException {
		profileKB(resultsFile, this.knowledgeBase.getClassPropertyIds());
	}

	public void profileKB(String resultsFile, Set<ClassPropertyId> classPropertyIdSet) throws IOException {

		Map<ClassPropertyId, Map<TimestampPointInTime, MutableInt>> countMap = new HashMap<>();

		for (ClassPropertyId cpid : classPropertyIdSet) {

			Map<TimestampPointInTime, MutableInt> count = new HashMap<>();

			Set<InstanceId> instances = this.knowledgeBase.getInstances(cpid.getClassId(), cpid.getPropertyId());
			for (InstanceId instanceId : instances) {

				QualifiedValueList qvl = this.knowledgeBase.getQualifiedValueList(new ValueId(instanceId, cpid.getPropertyId()));
				Map<TimestampPointInTime, List<Integer>> qualifierMap = timestampQualifiersAdapter.getQualifierMapFromQVL(qvl);
				for (Entry<TimestampPointInTime, List<Integer>> q : qualifierMap.entrySet()) {
					CHMHelper.getFromMap(count, q.getKey(), () -> new MutableInt()).add(1);

				}

			}

			countMap.put(cpid, count);

		}

		CSVWriter csvWriter = new CSVWriter(new FileWriter(resultsFile));

		int length = 3;

		String[] headerLine = new String[length];
		headerLine[0] = "ClassPropertyId";
		headerLine[1] = "Year";
		headerLine[2] = "Count";

		csvWriter.writeNext(headerLine);

		for (Entry<ClassPropertyId, Map<TimestampPointInTime, MutableInt>> cm : countMap.entrySet()) {
			for (Entry<TimestampPointInTime, MutableInt> ce : cm.getValue().entrySet()) {
				String[] line = new String[length];
				line[0] = cm.getKey().toString();
				line[1] = String.valueOf(ce.getKey().getYear());
				line[2] = String.valueOf(ce.getValue().intValue());

				csvWriter.writeNext(line);
			}
		}

		csvWriter.close();

	}

	public void profileTables(String resultsFile) throws IOException {

		ConcurrentHashMap<ClassId, ConcurrentHashMap<TimestampPointInTime, AtomicInteger>> countMap = new ConcurrentHashMap<>();

		for (AnnotatedTableId tableId : this.tableRepository.getAllTableIds()) {

			AnnotatedTableMetaData md = this.tableRepository.getMetaData(tableId);
			AnnotatedTable td = this.tableRepository.getTable(tableId);

			for (Entry<TimestampIdentifier, Integer> e : md.getTableYears().entrySet()) {
				ConcurrentHashMap<TimestampPointInTime, AtomicInteger> innerMap = CHMHelper.getFromCHM(countMap, md.getDbPediaClass(),
						() -> new ConcurrentHashMap<>());
				CHMHelper.getFromCHM(innerMap, new TimestampPointInTime(e.getValue()), () -> new AtomicInteger()).incrementAndGet();
			}

			for (Entry<ColumnIdx, ValueType> e : td.getDataTypes().entrySet()) {
				if (e.getValue().equals(ValueType.TIME)) {

					ConcurrentHashMap<TimestampPointInTime, AtomicInteger> innerMap = CHMHelper.getFromCHM(countMap, md.getDbPediaClass(),
							() -> new ConcurrentHashMap<>());
					for (Entry<RowIdx, AbstractObject> v : td.getColumn(e.getKey()).entrySet()) {

						Integer year = TimeObject.getYear(v.getValue());
						if (year != null) {
							CHMHelper.getFromCHM(innerMap, new TimestampPointInTime(year), () -> new AtomicInteger()).incrementAndGet();
						}
					}

				}
			}
		}

		CSVWriter csvWriter = new CSVWriter(new FileWriter(resultsFile));

		int length = 3;

		String[] headerLine = new String[length];
		headerLine[0] = "Class";
		headerLine[1] = "Year";
		headerLine[2] = "Count";

		csvWriter.writeNext(headerLine);

		for (Entry<ClassId, ConcurrentHashMap<TimestampPointInTime, AtomicInteger>> cm : countMap.entrySet()) {
			for (Entry<TimestampPointInTime, AtomicInteger> ce : cm.getValue().entrySet()) {
				String[] line = new String[length];
				line[0] = cm.getKey().toString();
				line[1] = String.valueOf(ce.getKey().getYear());
				line[2] = String.valueOf(ce.getValue().intValue());

				csvWriter.writeNext(line);
			}
		}

		csvWriter.close();

	}

}
