package de.mannheim.uni.expansion.tablefusion.services;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.InstanceIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.adapters.columnmatches.ColumnMatchesRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.matching.IndexBuilder;

public class AnnotatedTableInstanceScorer {

	private final KnowledgeBase kb;
	private final SimilarityCalculator sc;
	private final ColumnMatchesRepository columnMatchesRepository;

	private final OntologyRepository ontologyRepository;

	@Inject
	public AnnotatedTableInstanceScorer(KnowledgeBase kb, SimilarityCalculator sc, OntologyRepository ontologyRepository,
			ColumnMatchesRepository columnMatchesRepository) {
		this.kb = kb;
		this.sc = sc;
		this.columnMatchesRepository = columnMatchesRepository;
		this.ontologyRepository = ontologyRepository;
	}

	public Map<InstanceIndexComposite, Double> getScores(AnnotatedTablesReaderData data) {

		Map<InstanceIndexComposite, Double> ret = new HashMap<>();

		Map<Integer, List<InstanceId>> instances = new HashMap<>();
		Map<InstanceIndexComposite, Double> scores = new HashMap<>();

		ColumnIdx keyColumn = new ColumnIdx(data.getMetaData().getKeyColumn());
		// we start with name, then the other columns
		double scoreCount = 1.0;
		scoreCount += data.getMetaData().getProperties().size();

		for (Entry<Integer, InstanceId> e : data.getMetaData().getInstances().entrySet()) {
			double scoreSum = 0.0;
			RowIdx rowIdx = new RowIdx(e.getKey());
			InstanceId instanceId = e.getValue();
			AbstractObject name = data.getTableData().getValue(rowIdx, keyColumn);
			Set<AbstractObject> names = new HashSet<>();
			names.add(name);
			if (name != null && name.getType().equals(ValueType.STRING)) {

				StringObject stringName = (StringObject) name;

				Optional<String> combined = IndexBuilder.lettersCombiner(stringName.getString());
				if (combined.isPresent()) {
					names.add(new StringObject(combined.get()));

					Optional<String> deAccented = IndexBuilder.deAccent(combined.get());
					if (deAccented.isPresent()) {
						names.add(new StringObject(deAccented.get()));
					}
				}

				Optional<String> deAccented = IndexBuilder.deAccent(stringName.getString());
				if (deAccented.isPresent()) {
					names.add(new StringObject(deAccented.get()));
				}

			}

			QualifiedValueList labelQvl = kb.getQualifiedValueList(instanceId, PropertiesExpansionData.LABEL_PROPERTY);
			if (labelQvl == null || labelQvl.size() == 0) {
				System.out.println("Mapped Instance doesn't have a label.");
			} else {

				double maxSim = 0.0;
				for (QualifiedValueHolder labelQVH : labelQvl.getClaims()) {
					AbstractObject label = labelQVH.getValue();

					for (AbstractObject nameAO : names) {
						double sim = sc.getSimilarity(nameAO, label);
						if (sim > maxSim) {
							maxSim = sim;
						}
					}
				}

				scoreSum += maxSim;

			}

			for (Entry<Integer, PropertyId> propertyEntry : data.getMetaData().getProperties().entrySet()) {
				PropertyId propertyId = propertyEntry.getValue();
				ColumnIdx columnIdx = new ColumnIdx(propertyEntry.getKey());
				QualifiedValueList qvl = kb.getQualifiedValueListWithEqualProperties(new ValueId(instanceId, propertyId));
				AbstractObject ao = data.getTableData().getValue(rowIdx, columnIdx);

				AbstractObject ao2 = null;
				if (ontologyRepository.isReferenceField(propertyEntry.getValue())) {
					ao2 = columnMatchesRepository.getMatch(new CellId(data.getId(), columnIdx, rowIdx));
				}

				if (ao != null) {
					if (qvl != null) {
						double maxScore = 0.0;
						for (QualifiedValueHolder claim : qvl.getClaims()) {

							if (ao2 != null) {
								double score = sc.getSimilarity(ao2, claim.getValue());
								if (score > maxScore)
									maxScore = score;
							}

							{
								double score = sc.getSimilarity(ao, claim.getValue());
								if (score > maxScore)
									maxScore = score;
							}

							// if (sc.isEqual(ao, claim.getValue())) {
							// scoreSum++;
							// break;
							// }
						}

						scoreSum += maxScore;
					}
				}
			}

			double score = scoreSum / scoreCount;

			InstanceIndexComposite index = new InstanceIndexComposite(e.getKey(), instanceId);

			ret.put(index, score);

		}

		return ret;

	}

}
