package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies;

import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.RegressionScoresClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel3;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import weka.classifiers.functions.LinearRegression;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.FastVector;
import weka.core.Instances;

public class RegressionScorerClassProperty {

	private final ValueEvaluator valueEvaluator;

	@Inject
	public RegressionScorerClassProperty(ValueEvaluator valueEvaluator) {
		this.valueEvaluator = valueEvaluator;
	}

	public RegressionScoresClassPropertyId get(GroupedValues groupedValues, Map<String, ScoresContainer> scoresContainers) {

		String[] names = new String[scoresContainers.size()];
		ScoresContainer[] containers = new ScoresContainer[scoresContainers.size()];
		FastVector fv = new FastVector(scoresContainers.size());
		fv.addElement(new Attribute("Score"));
		{
			int i = 0;
			for (Entry<String, ScoresContainer> sc : scoresContainers.entrySet()) {
				names[i] = sc.getKey();
				containers[i] = sc.getValue();

				fv.addElement(new Attribute(sc.getKey()));
				i++;
			}
		}

		ConcurrentHashMap<ClassPropertyId, ConcurrentLinkedQueue<Pair<Boolean, double[]>>> listMap = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(groupedValues.entrySet(), (gv) -> {
			for (SingleValue sv : gv.getValue().getValues()) {

				Optional<Boolean> correct = valueEvaluator.correct(gv.getKey().getValueId(), FusedValue.fromObject(sv.getValue()));
				if (correct.isPresent()) {
					double[] scores = new double[scoresContainers.size()];
					for (int i = 0; i < containers.length; i++) {
						Probability probability = containers[i].getScore(gv.getKey().getClassId(), gv.getKey().getPropertyId(), gv.getKey().getInstanceId(),
								sv.getId());
						if (probability != null)
							scores[i] = probability.get(SimilarityType.DIRECT).getAccuracy();
						else
							scores[i] = 0.0;
					}
					CHMHelper.getFromCHM(listMap, new ClassPropertyId(gv.getKey()), () -> new ConcurrentLinkedQueue<>())
							.add(new Pair<Boolean, double[]>(correct.get(), scores));

				}

			}
		}, "RegressionScorer");

		int length = listMap.size();

		ConcurrentHashMap<ClassPropertyId, double[]> weightsMap = new ConcurrentHashMap<>();
		ConcurrentHashMap<ClassPropertyId, Double> interceptsMap = new ConcurrentHashMap<>();

		PersistentParallel3.runAndFinish(listMap.entrySet(), (entry) -> {
			ConcurrentLinkedQueue<Pair<Boolean, double[]>> list = entry.getValue();

			LinearRegression regression = new LinearRegression();

			Instances data = new Instances("Regression", fv, length);

			// double[] y = new double[length];
			// double[][] x = new double[length][];

			for (Pair<Boolean, double[]> l : list) {
				double[] z = new double[scoresContainers.size() + 1];

				if (l.getFirst()) {
					z[0] = 1.0;
				} else {
					z[0] = 0.0;
				}

				int vp = 1;
				for (double v : l.getSecond()) {

					z[vp] = v;
					vp++;
				}

				data.add(new DenseInstance(1.0, z));
			}

			data.setClassIndex(0);

			// RealMatrix matrix = new Array2DRowRealMatrix(x);
			//
			// RealMatrix vt = new SingularValueDecomposition(matrix).getVT();
			//
			// x = matrix.multiply(vt).getData();

			// matrix = new Array2DRowRealMatrix(x);
			//
			// double[][] cov = new
			// Covariance(matrix.transpose().getData())..getCovarianceMatrix().getData();

			try {
				regression.buildClassifier(data);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

			double[] weightsWithIntercept = regression.coefficients();

			double[] weights = new double[containers.length];

			int j2 = 0;
			for (int j = 1; j < weightsWithIntercept.length - 1; j++) {

				weights[j2] = weightsWithIntercept[j];

				j2++;
			}

			weightsMap.put(entry.getKey(), weights);
			interceptsMap.put(entry.getKey(), weightsWithIntercept[weightsWithIntercept.length - 1]);

		}, "ClassPropertyRegression");

		return new RegressionScoresClassPropertyId(names, containers, weightsMap, interceptsMap);
	}
}
