package de.mannheim.uni.expansion.tablefusion.services.grouping;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.impl.BasicGroupMatcher;
import de.mannheim.uni.expansion.tablefusion.core.matching.impl.ColumnMatchesGrouper;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.CHMHelper.Creator;
import de.mannheim.uni.expansion.utils.MapSorter;
import de.mannheim.uni.expansion.utils.PersistentParallel;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class GroupedValuesRepository {

	private final BasicMatcherService matcherService;

	private final ColumnMatchesGrouper columnMatchesGrouper;
	private final BasicGroupMatcher groupMatcher;
	private final OntologyRepository ontologyRepository;

	private final KnowledgeBase knowledgeBase;

	@Inject
	public GroupedValuesRepository(BasicMatcherService matcherService, ColumnMatchesGrouper columnMatchesGrouper, BasicGroupMatcher groupMatcher,
			OntologyRepository ontologyRepository, KnowledgeBase knowledgeBase) {
		this.matcherService = matcherService;
		this.columnMatchesGrouper = columnMatchesGrouper;
		this.groupMatcher = groupMatcher;
		this.ontologyRepository = ontologyRepository;
		this.knowledgeBase = knowledgeBase;
	}

	private static GroupedValues groupedValues;

	public GroupedValues getGroupedValues() {

		if (groupedValues == null) {
			computeGroupedValues(null);
		}
		return groupedValues;

	}

	public GroupedValues getGroupedValues(FusionConfig fusionConfig) {

		if (groupedValues == null) {
			computeGroupedValues(fusionConfig);
		}
		return groupedValues;

	}

	public Map<ClassPropertyId, Integer> getCount(int minCount, boolean groupTriples) {
		ConcurrentHashMap<ClassPropertyId, AtomicInteger> countsPerClass = new ConcurrentHashMap<>();

		Creator<AtomicInteger> aiCreator = () -> {
			return new AtomicInteger();
		};

		PersistentParallel6.runAndFinish(getGroupedValues(null).entrySet(), (e) -> {
			int count = e.getValue().getValues().size();
			if (count >= minCount) {
				ClassPropertyId classPropertyId = new ClassPropertyId(e.getKey().getClassId(), e.getKey().getPropertyId());

				AtomicInteger countPerClass = CHMHelper.getFromCHM(countsPerClass, classPropertyId, aiCreator);
				if (groupTriples) {
					countPerClass.incrementAndGet();
				} else {
					countPerClass.addAndGet(count);
				}
			}

		}, "Calculate Class Property Statistics");

		Map<ClassPropertyId, Integer> countFinal = new HashMap<>();

		for (Entry<ClassPropertyId, AtomicInteger> e : countsPerClass.entrySet()) {
			countFinal.put(e.getKey(), e.getValue().get());
		}
		return countFinal;
	}

	public Map<ClassPropertyId, Integer> getSourceCount(int minCount) {
		return getCount(minCount, false);
	}

	public Map<ClassPropertyId, Double> getStatistics(int minCount) {

		Map<ClassPropertyId, Integer> countsPerClass = getCount(minCount, true);

		int countTotal = 0;

		for (Entry<ClassPropertyId, Integer> e : countsPerClass.entrySet()) {
			countTotal += e.getValue();
		}

		Map<ClassPropertyId, Double> statistics = new HashMap<>();
		for (Entry<ClassPropertyId, Integer> e : countsPerClass.entrySet()) {

			double countPerClass = e.getValue();

			statistics.put(e.getKey(), countPerClass / countTotal);
		}

		return statistics;

	}

	public Map<ClassPropertyId, LinkedHashMap<InstanceId, Integer>> getSourceSorted(int minCount) {
		ConcurrentHashMap<ClassPropertyId, ConcurrentHashMap<InstanceId, Integer>> countsPerClass = new ConcurrentHashMap<>();

		Creator<ConcurrentHashMap<InstanceId, Integer>> mapCreator = () -> {
			return new ConcurrentHashMap<InstanceId, Integer>();
		};

		PersistentParallel6.runAndFinish(getGroupedValues(null).entrySet(), (e) -> {
			int count = e.getValue().getValues().size();
			if (count >= minCount) {
				ClassPropertyId classPropertyId = new ClassPropertyId(e.getKey().getClassId(), e.getKey().getPropertyId());

				ConcurrentHashMap<InstanceId, Integer> countPerClass = CHMHelper.getFromCHM(countsPerClass, classPropertyId, mapCreator);
				countPerClass.put(e.getKey().getInstanceId(), count);
			}
		}, "Sort Grouped Values by Count");

		ConcurrentHashMap<ClassPropertyId, LinkedHashMap<InstanceId, Integer>> sortedMap = new ConcurrentHashMap<>();

		PersistentParallel.runAndFinish(countsPerClass.entrySet(), (e) -> {

			sortedMap.put(e.getKey(), MapSorter.sortHashMapByValuesD(e.getValue()));

		}, "Sort Grouped Values by Count");

		return sortedMap;
	}

	private synchronized void computeGroupedValues(FusionConfig fusionConfig) {
		if (groupedValues == null) {
			Map<TripleId, MatchedValues> groupedValues = computeGroupedValuesNoCache(fusionConfig);
			//
			// groupedValues = groupedValues.entrySet().stream().filter((e) -> {
			// return e.getValue().getValues().size() > 5;
			// }).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

			this.groupedValues = new GroupedValues(groupedValues);
		}
	}

	private Map<TripleId, MatchedValues> computeGroupedValuesNoCache(FusionConfig fusionConfig) {
		MatcherOptionsBuilder matcherOptions;
		Map<TripleId, MatchedValues> groupedValues;

		matcherOptions = new MatcherOptionsBuilder()
				.withNewMatcher(groupMatcher, ValueType.QUANTITY, ValueType.STRING, ValueType.TIME, ValueType.URL, ValueType.CATEGORICAL_INT)
				.withNewReferenceMatcher(columnMatchesGrouper);
		groupedValues = matcherService.matchAndGroup(matcherOptions, fusionConfig);

		return groupedValues;
	}

	public GroupedValues computeGroupedValuesDirect(FusionConfig fusionConfig) {
		return new GroupedValues(computeGroupedValuesNoCache(fusionConfig));
	}

	public Map<Integer, Integer> getDistribution(FusionConfig fusionConfig) {
		GroupedValues groupedValues = getGroupedValues(fusionConfig);

		ConcurrentHashMap<Integer, AtomicInteger> map = new ConcurrentHashMap<Integer, AtomicInteger>();

		PersistentParallel6.runAndFinish(groupedValues.entrySet(), (e) -> {
			Integer count = e.getValue().getValues().size();
			AtomicInteger ai = CHMHelper.getFromCHM(map, count, () -> new AtomicInteger());
			ai.incrementAndGet();
		}, "Grouped Values Distribution");

		Map<Integer, Integer> ret = map.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			return e.getValue().get();
		}));

		return ret;
	}

	private final static SimpleLockableConcurrentCache<UuidStringId, GroupedValues> cache = new SimpleLockableConcurrentCache<>();

	public GroupedValues getGroupedValues(FusionConfig fusionConfig, TableIdsSet tableIdsSet) {

		return cache.getOrCreate(tableIdsSet.getId(), () -> {
			MatcherOptionsBuilder matcherOptions;
			Map<TripleId, MatchedValues> groupedValues;

			matcherOptions = new MatcherOptionsBuilder().withNewMatcher(groupMatcher, ValueType.QUANTITY, ValueType.STRING, ValueType.TIME, ValueType.URL,
					ValueType.INSTANCE_REFERENCE, ValueType.CATEGORICAL_INT);//
			// .withNewReferenceMatcher(columnMatchesGrouper);
			groupedValues = matcherService.matchAndGroup(matcherOptions, fusionConfig, tableIdsSet.getTableIds());

			return new GroupedValues(groupedValues);

		});

	}

	public GroupedValues getGroupedValues(AnnotatedTablesRepository tableRepository, Set<AnnotatedTableId> tableIdsSet, FusionConfig fusionConfig) {
		BasicMatcherService matcherService = new BasicMatcherService(tableRepository, ontologyRepository);
		MatcherOptionsBuilder matcherOptions;
		Map<TripleId, MatchedValues> groupedValues;
		BasicGroupMatcher groupMatcher = new BasicGroupMatcher(tableRepository, knowledgeBase, ontologyRepository);

		matcherOptions = new MatcherOptionsBuilder().withNewMatcher(groupMatcher, ValueType.QUANTITY, ValueType.STRING, ValueType.TIME, ValueType.URL,
				ValueType.INSTANCE_REFERENCE, ValueType.CATEGORICAL_INT);//
		// .withNewReferenceMatcher(columnMatchesGrouper);
		groupedValues = matcherService.matchAndGroup(matcherOptions, fusionConfig, tableIdsSet);

		return new GroupedValues(groupedValues);
	}

}
