package de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.stockcrow;

import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;

import au.com.bytecode.opencsv.CSVReader;
import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.CrawlerTools;

public class StockCrowCrawler {
	private final static String outputFile = "data.csv";
	private final static String baseUrl = "https://stockrow.com/";
	private final static String incomeStatementSuffix = "/financials?d=y&s=income";
	private final static String balanceSheetSuffix = "/financials?d=y&s=balance";

	public static void main(String[] args) {

		CrawlerTools.noSSLCheckInit();
		// StockCrowCrawler importer = new StockCrowCrawler();
		// Map<FinancialYear, FinancialData> data = importer.getCompanyData(new
		// TrackerId("GOOGL"));

		String outputPath = "C:/Users/yaser/SampleRuns/Wikidata/datasets/FINANCE/";

		StockCrowCrawler importer = new StockCrowCrawler();
		int unavailable = 0;
		int available = 0;
		try {
			Map<TrackerId, InstanceId> companies = importer.getCompanies(outputPath + "companies.csv");
			Map<TrackerId, Map<FinancialYear, FinancialData>> dataMap = new HashMap<>();

			for (Entry<TrackerId, InstanceId> company : companies.entrySet()) {
				Map<FinancialYear, FinancialData> data = importer.getCompanyData(company.getKey());
				if (data != null) {
					dataMap.put(company.getKey(), data);
					available++;
				} else {
					unavailable++;
				}

			}

			writeFile(outputPath + outputFile, dataMap);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		System.out.println("available: " + available);
		System.out.println("unavailable: " + unavailable);
	}

	private static void writeFile(String file, Map<TrackerId, Map<FinancialYear, FinancialData>> dataMap) throws IOException {
		CSVWriter writer = new CSVWriter(new FileWriter(file));
		try {

			int length = 7;
			{
				String[] headerLine = new String[length];

				headerLine[0] = "Tracker";
				headerLine[1] = "Year";
				headerLine[2] = "Revenue";
				headerLine[3] = "Assets";
				headerLine[4] = "NetIncome";
				headerLine[5] = "EBIT";
				headerLine[6] = "Equity";

				writer.writeNext(headerLine);
			}

			for (Entry<TrackerId, Map<FinancialYear, FinancialData>> company : dataMap.entrySet()) {
				String[] line = new String[length];
				line[0] = company.getKey().getTrackerId();
				for (Entry<FinancialYear, FinancialData> year : company.getValue().entrySet()) {
					line[1] = String.valueOf(year.getKey().getYear());
					line[2] = year.getValue().getRevenue() != null ? String.valueOf(year.getValue().getRevenue()) : "";
					line[3] = year.getValue().getAssets() != null ? String.valueOf(year.getValue().getAssets()) : "";
					line[4] = year.getValue().getNetIncome() != null ? String.valueOf(year.getValue().getNetIncome()) : "";
					line[5] = year.getValue().getEbit() != null ? String.valueOf(year.getValue().getEbit()) : "";
					line[6] = year.getValue().getEquity() != null ? String.valueOf(year.getValue().getEquity()) : "";

					writer.writeNext(line);
				}
			}

		} finally {
			writer.close();
		}

	}

	private Map<FinancialYear, FinancialData> getCompanyData(TrackerId value) {

		Map<FinancialYear, FinancialData> retMap = new HashMap<>();

		Pair<Integer, String> incomeStatementPage = CrawlerTools.getDocument(baseUrl + value.getTrackerId() + incomeStatementSuffix);
		if (incomeStatementPage.getFirst().equals(404)) {
			return null;
		}
		Pair<Integer, String> balanceSheetPage = CrawlerTools.getDocument(baseUrl + value.getTrackerId() + balanceSheetSuffix);
		if (balanceSheetPage.getFirst().equals(404)) {
			return null;
		}

		{

			Document incomeStatementDoc = Jsoup.parse(incomeStatementPage.getSecond());
			Elements incomeStatementRows = incomeStatementDoc.select("table.full-width.financial-table tr");

			Elements years = incomeStatementRows.get(0).select(".financial-table__caption_right_9");
			if (years.size() == 0)
				return null;
			FinancialYear[] incomeStatementYears = new FinancialYear[years.size()];
			for (int i = 0; i < years.size(); i++) {
				Element year = years.get(i);
				String[] fullYearString = CrawlerTools.getCleanHtml(year.html()).split("-");
				int yearInt = Integer.parseInt(fullYearString[0]);
				incomeStatementYears[i] = new FinancialYear(yearInt);
			}
			{

				Elements revenues = getRow(incomeStatementRows, "Revenues");
				for (int i = 0; i < revenues.size(); i++) {
					Element revenue = revenues.get(i);
					String fullYearString = CrawlerTools.getCleanHtml(revenue.html());

					double number = getNumber(fullYearString);

					FinancialData oldData = CHMHelper.getFromMap(retMap, incomeStatementYears[i], () -> new FinancialData());
					FinancialData newData = FinancialData.withRevenue(number);

					retMap.put(incomeStatementYears[i], new FinancialData(oldData, newData));
				}
			}
			{
				// Elements ebit =
				// incomeStatementRows.get(6).select(".financial-table__value
				// span");
				Elements ebit = getRow(incomeStatementRows, "Earning Before Interest &amp; Taxes (EBIT)");
				for (int i = 0; i < ebit.size(); i++) {
					Element num = ebit.get(i);
					String fullYearString = CrawlerTools.getCleanHtml(num.html());

					double number = getNumber(fullYearString);

					FinancialData oldData = CHMHelper.getFromMap(retMap, incomeStatementYears[i], () -> new FinancialData());
					FinancialData newData = FinancialData.withEbit(number);

					retMap.put(incomeStatementYears[i], new FinancialData(oldData, newData));
				}
			}

			{

				// Elements netIncome =
				// incomeStatementRows.get(9).select(".financial-table__value
				// span");
				Elements netIncome = getRow(incomeStatementRows, "Net Income");
				for (int i = 0; i < netIncome.size(); i++) {
					Element num = netIncome.get(i);
					String fullYearString = CrawlerTools.getCleanHtml(num.html());

					double number = getNumber(fullYearString);

					FinancialData oldData = CHMHelper.getFromMap(retMap, incomeStatementYears[i], () -> new FinancialData());
					FinancialData newData = FinancialData.withNetIncome(number);

					retMap.put(incomeStatementYears[i], new FinancialData(oldData, newData));
				}
			}

		}

		{
			Document balanceSheetDoc = Jsoup.parse(balanceSheetPage.getSecond());
			Elements balanceSheetRows = balanceSheetDoc.select("table.full-width.financial-table tr");

			Elements years = balanceSheetRows.get(0).select(".financial-table__caption_right_9");
			FinancialYear[] balancesheetYears = new FinancialYear[years.size()];
			for (int i = 0; i < years.size(); i++) {
				Element year = years.get(i);
				String[] fullYearString = CrawlerTools.getCleanHtml(year.html()).split("-");
				int yearInt = Integer.parseInt(fullYearString[0]);
				balancesheetYears[i] = new FinancialYear(yearInt);
			}

			{

				// Elements assets =
				// balanceSheetRows.get(7).select(".financial-table__value
				// span");
				Elements assets = getRow(balanceSheetRows, "Total Assets");
				for (int i = 0; i < assets.size(); i++) {
					Element num = assets.get(i);
					String fullYearString = CrawlerTools.getCleanHtml(num.html());

					double number = getNumber(fullYearString);

					FinancialData oldData = CHMHelper.getFromMap(retMap, balancesheetYears[i], () -> new FinancialData());
					FinancialData newData = FinancialData.withAssets(number);

					retMap.put(balancesheetYears[i], new FinancialData(oldData, newData));
				}
			}

			{

				// Elements equity =
				// balanceSheetRows.get(16).select(".financial-table__value
				// span");
				Elements equity = getRow(balanceSheetRows, "Shareholders Equity (USD)");
				for (int i = 0; i < equity.size(); i++) {
					Element num = equity.get(i);
					String fullYearString = CrawlerTools.getCleanHtml(num.html());

					double number = getNumber(fullYearString);

					FinancialData oldData = CHMHelper.getFromMap(retMap, balancesheetYears[i], () -> new FinancialData());
					FinancialData newData = FinancialData.withEquity(number);

					retMap.put(balancesheetYears[i], new FinancialData(oldData, newData));
				}
			}

		}

		return retMap;
	}

	private Elements getRow(Elements rows, String string) {

		for (int i = 0; i < rows.size(); i++) {
			String header = CrawlerTools.getCleanHtml(rows.get(i).select(".financial-table__caption_normal").html());
			if (string.equals(header)) {
				return rows.get(i).select(".financial-table__value span");
			}
		}

		return null;
	}

	double getNumber(String fullYearString) {
		return Double.parseDouble(fullYearString);

		// String finalCharachter =
		// fullYearString.substring(fullYearString.length() - 1,
		// fullYearString.length());
		// double multiplier = 1.0;
		// boolean remove = false;
		// if (finalCharachter.equals("m")) {
		// multiplier = 1000000;
		// remove = true;
		// }
		// if (finalCharachter.equals("b")) {
		// multiplier = 1000000000;
		// remove = true;
		// }
		// if (finalCharachter.equals("k")) {
		//
		// multiplier = 1000;
		// remove = true;
		// }
		//
		// if (remove) {
		// fullYearString = fullYearString.substring(0, fullYearString.length()
		// - 1);
		// }
		//
		// double number = Double.parseDouble(fullYearString) * multiplier;
		//
		// return number;

	}

	public Map<TrackerId, InstanceId> getCompanies(String file) throws IOException {

		Map<TrackerId, InstanceId> retMap = new HashMap<>();
		CSVReader reader = new CSVReader(new FileReader(file));
		try {

			List<String[]> allLines = reader.readAll();
			for (int i = 1; i < allLines.size(); i++) {
				String[] line = allLines.get(i);
				String[] idSplit = line[0].trim().split("/");
				InstanceId instanceId = InstanceId.simple(idSplit[idSplit.length - 1]);
				TrackerId trackerId = new TrackerId(line[2].trim());
				retMap.put(trackerId, instanceId);
			}

			return retMap;
		} finally {
			reader.close();
		}
	}

	public static class TrackerId {

		private final String trackerId;

		public TrackerId(String trackerId) {
			super();
			this.trackerId = trackerId;
		}

		public String getTrackerId() {
			return trackerId;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((trackerId == null) ? 0 : trackerId.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			TrackerId other = (TrackerId) obj;
			if (trackerId == null) {
				if (other.trackerId != null)
					return false;
			} else if (!trackerId.equals(other.trackerId))
				return false;
			return true;
		}

		@Override
		public String toString() {
			return trackerId;
		}

	}

	public static class FinancialData {
		private final Double revenue;
		private final Double assets;
		private final Double netIncome;
		private final Double ebit;
		private final Double equity;

		public FinancialData() {
			this.revenue = null;
			this.assets = null;
			this.netIncome = null;
			this.ebit = null;
			this.equity = null;
		}

		public static FinancialData withRevenue(Double revenue) {
			return new FinancialData(revenue, null, null, null, null);
		}

		public static FinancialData withAssets(Double assets) {
			return new FinancialData(null, assets, null, null, null);
		}

		public static FinancialData withNetIncome(Double netIncome) {
			return new FinancialData(null, null, netIncome, null, null);
		}

		public static FinancialData withEbit(Double ebit) {
			return new FinancialData(null, null, null, ebit, null);
		}

		public static FinancialData withEquity(Double equity) {
			return new FinancialData(null, null, null, null, equity);
		}

		public FinancialData(FinancialData oldData, FinancialData newData) {
			super();
			this.revenue = newData.getRevenue() != null ? newData.getRevenue() : oldData.getRevenue();
			this.assets = newData.getAssets() != null ? newData.getAssets() : oldData.getAssets();
			this.netIncome = newData.getNetIncome() != null ? newData.getNetIncome() : oldData.getNetIncome();
			this.ebit = newData.getEbit() != null ? newData.getEbit() : oldData.getEbit();
			this.equity = newData.getEquity() != null ? newData.getEquity() : oldData.getEquity();
		}

		public FinancialData(Double revenue, Double assets, Double netIncome, Double ebit, Double equity) {
			super();
			this.revenue = revenue;
			this.assets = assets;
			this.netIncome = netIncome;
			this.ebit = ebit;
			this.equity = equity;
		}

		public Double getRevenue() {
			return revenue;
		}

		public Double getAssets() {
			return assets;
		}

		public Double getNetIncome() {
			return netIncome;
		}

		public Double getEbit() {
			return ebit;
		}

		public Double getEquity() {
			return equity;
		}

		@Override
		public String toString() {
			return "FinancialData [revenue=" + revenue + ", assets=" + assets + ", netIncome=" + netIncome + ", ebit=" + ebit + ", equity=" + equity + "]";
		}

	}

	public static class FinancialYear {
		private final int year;

		public FinancialYear(int year) {
			super();
			this.year = year;
		}

		public int getYear() {
			return year;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + year;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			FinancialYear other = (FinancialYear) obj;
			if (year != other.year)
				return false;
			return true;
		}

		@Override
		public String toString() {
			return String.valueOf(year);
		}

	}

}
