package de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.worldbank;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Consumer;

import com.google.common.collect.BiMap;

import au.com.bytecode.opencsv.CSVReader;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.Qualifier;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class WorldBankConverter implements KBValueCreator {

	private final String path;

	public WorldBankConverter(String path) {
		this.path = path;
	}

	private static final String isoCodesFile = "countrycodes.txt";
	private static final String POP = "POP.csv";
	private static final String GDP_PCAP = "GDP_PCAP.csv";
	private static final String GDP = "GDP.csv";
	private static final String POP_DENS = "POP_DENS.csv";

	private static final PropertyId POP_PROPERTY_ID = Stats.population;
	private static final PropertyId GDP_PCAP_PROPERTY_ID = Stats.nominalGDPperCapita;
	private static final PropertyId GDP_PROPERTY_ID = Stats.nominalGDP;
	private static final PropertyId POP_DENS_PROPERTY_ID = Stats.populationDensity;

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {
		try {
			Map<ValueId, QualifiedValueList> retMap = new ConcurrentHashMap<ValueId, QualifiedValueList>();
			Set<ValueId> remove = new HashSet<>();
			Set<InstanceId> removeInstances = new HashSet<>();

			Set<String> isoCodes = getIsoCodes(path + isoCodesFile);
			Map<String, InstanceId> isoCodeToIntanceIdMap = getIsoToInstanceMap(isoCodes, kb);

			retMap.putAll(readFromWorldBankCSV(path + POP, isoCodeToIntanceIdMap, POP_PROPERTY_ID));
			retMap.putAll(readFromWorldBankCSV(path + POP_DENS, isoCodeToIntanceIdMap, POP_DENS_PROPERTY_ID));
			retMap.putAll(readFromWorldBankCSV(path + GDP, isoCodeToIntanceIdMap, GDP_PROPERTY_ID));
			retMap.putAll(readFromWorldBankCSV(path + GDP_PCAP, isoCodeToIntanceIdMap, GDP_PCAP_PROPERTY_ID));

			Set<InstanceId> contains = new HashSet<InstanceId>(isoCodeToIntanceIdMap.values());

			for (Entry<InstanceId, Integer> instance : kb.getInstances(WikidataClasses.wikidataCountry).entrySet()) {

				Consumer<PropertyId> removeF = (propertyId) -> {
					ValueId valueId = new ValueId(instance.getKey(), propertyId);
					if (!retMap.containsKey(valueId)) {
						remove.add(valueId);
					}
				};

				removeF.accept(GDP_PCAP_PROPERTY_ID);
				removeF.accept(GDP_PROPERTY_ID);
				removeF.accept(POP_DENS_PROPERTY_ID);
				removeF.accept(POP_PROPERTY_ID);

				if (!contains.contains(instance.getKey())) {
					removeInstances.add(instance.getKey());
				}

			}

			return new KBValueContainer(retMap, removeInstances, remove);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private Map<String, InstanceId> getIsoToInstanceMap(Set<String> isoCodes, KnowledgeBase kb) {
		Map<String, InstanceId> retMap = new HashMap<>();
		BiMap<InstanceId, Integer> instances = kb.getInstances(WikidataClasses.wikidataCountry);

		for (InstanceId instanceId : instances.keySet()) {
			ValueId valueId = new ValueId(instanceId, Stats.isoAlpha3Codes);
			QualifiedValueList qvl = kb.getQualifiedValueList(valueId);
			if (qvl != null) {
				QualifiedValueHolder qvh = qvl.getClaim(0);
				if (qvh != null) {
					AbstractObject ao = qvh.getValue();
					if (ao instanceof StringObject) {
						StringObject so = (StringObject) ao;
						String code = so.getString().trim();
						if (isoCodes.contains(code)) {
							retMap.put(code, instanceId);
						}
					}
				}
			} else {
				System.out.println();
			}
		}

		for (String code : isoCodes) {
			if (!retMap.containsKey(code)) {
				System.out.println(code);
			}
		}

		return retMap;
	}

	private Set<String> getIsoCodes(String file) throws IOException {
		return getIsoCodesStatic(file);
	}

	public static Set<String> getIsoCodesStatic(String file) throws IOException {
		BufferedReader br = new BufferedReader(new FileReader(file));
		try {
			Set<String> isoCodes = new HashSet<>();
			for (String line = br.readLine(); line != null; line = br.readLine()) {
				isoCodes.add(line.trim());
			}

			return isoCodes;
		} finally {
			br.close();
		}
	}

	private Map<ValueId, QualifiedValueList> readFromWorldBankCSV(String file, Map<String, InstanceId> isoCodeMappings, PropertyId propertyId)
			throws IOException {
		Map<ValueId, QualifiedValueList> retMap = new HashMap<>();

		CSVReader csv = new CSVReader(new FileReader(file), ',', '"', 4);
		try {
			List<String[]> lines = csv.readAll();

			Map<Integer, List<Qualifier>> columnToYearMap = new HashMap<>();
			String[] headerLine = lines.get(0);
			for (int i = 4; i < headerLine.length; i++) {
				String header = headerLine[i].trim();
				if (header.length() > 0) {
					List<Qualifier> qualifierList = new ArrayList<>();
					int year = Integer.valueOf(headerLine[i].trim());
					Qualifier qualifier = new Qualifier(Stats.pointInTime, TimeObject.fromYear(year));
					qualifierList.add(qualifier);
					columnToYearMap.put(i, qualifierList);
				}
			}

			for (int r = 1; r < lines.size(); r++) {
				String[] line = lines.get(r);
				String key = line[1].trim();
				if (isoCodeMappings.containsKey(key)) {
					InstanceId instanceid = isoCodeMappings.get(key);
					List<QualifiedValueHolder> claims = new ArrayList<>();
					for (int i = 4; i < line.length; i++) {
						String value = line[i].trim();

						if (value.length() > 0) {
							Double doubleV = Double.valueOf(value);

							QuantityObject val = new QuantityObject(doubleV, doubleV, doubleV, "number");

							claims.add(new QualifiedAbstractObjectHolder(val, columnToYearMap.get(i)));
						}

					}

					QualifiedValueList qvl = new QualifiedValueList(claims);

					retMap.put(new ValueId(instanceid, propertyId), qvl);
				}
			}

			return retMap;
		} finally {
			csv.close();
		}

	}

}
