package de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.genetic;

import java.math.BigInteger;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.Random;
import java.util.TreeMap;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.utils.MathsUtil;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.OptimizableRun;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsCollectorData;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsRepositoryCollector;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.random.WeightsRandomRunner;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.random.WeightsRandomRunner.WeightsCurrentRun;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;

public class WeightsGeneticParallelOptimizer implements WeightsOptimizer {

	public static final float STEP = 0.02f;
	public static final int STEPS = Math.round((1.0f / STEP));
	public static final int CREEP_MUTATION = STEPS / 10;
	public static final int TRIAL_PER_RUN = 100;

	public static final int FIRST_RUN_MULTIPLIER = 10;

	public static final Map<Integer, Double> SUM_OF_WEIGHTS_MAP = getSumOfWeights(TRIAL_PER_RUN, FIRST_RUN_MULTIPLIER);

	public static final NavigableMap<Double, Integer> lookupMap = getLookUpMap(TRIAL_PER_RUN);

	@Inject
	public WeightsGeneticParallelOptimizer() {

	}

	private static NavigableMap<Double, Integer> getLookUpMap(int trialPerRun) {
		NavigableMap<Double, Integer> ret = new TreeMap<>();

		double currentScore = 0.0;
		for (int i = 0; i < trialPerRun; i++) {
			ret.put(currentScore, i);
			currentScore += (1.0 / (i + 1));
		}

		return Collections.unmodifiableNavigableMap(ret);
	}

	private static Map<Integer, Double> getSumOfWeights(int trialPerRun, int firstRunMultiplier) {
		Map<Integer, Double> ret = new HashMap<>();
		ret.put(trialPerRun, getSumOfWeights(trialPerRun));
		ret.put(trialPerRun * firstRunMultiplier, getSumOfWeights(trialPerRun));
		return Collections.unmodifiableMap(ret);
	}

	private static double getSumOfWeights(int trialPerRun) {
		double ret = 0.0;

		for (int i = 0; i < trialPerRun; i++) {
			ret += (1.0 / (i + 1));
		}

		return ret;
	}

	public static int totalBreakage = 0;

	public WeightsRepository optimize(OptimizableRun run) {


		WeightsRepositoryCollector collector = new WeightsRepositoryCollector(STEPS, STEP);

		run.run(collector);

		WeightsCollectorData collectorData = collector.getWeightsCollectorData();

		Map<OptimizableWeightKey, Double> weights;
		if (collector.getWorkingClass() == null) {
			weights = new HashMap<>();
		} else {
			
			BigInteger numberOfOutcomes = MathsUtil.outcomes(STEPS + 1, collectorData.getNames().size());

			if (numberOfOutcomes.compareTo(BigInteger.valueOf(1300)) <= 0) {
				Map<WeightsCurrentRun, Map<String, Integer>> currentRuns = WeightsRandomRunner.getAllRuns(collectorData);
				WeightsPopulationRunner population = new WeightsPopulationRunner(currentRuns, collectorData);

				PersistentParallel6.runAndFinish(population.getRuns(), (currentRun) -> {
					WeightsRepository currentWeightsRepository = population.getRun(currentRun);
					double performance = run.run(currentWeightsRepository);

					population.addPerformance(currentRun, performance);
				}, this.getClass());

				weights = population.getLearnedWeights();
			} else {

				Random random = new Random(123456789L);

				Map<String, Integer> lastRun = null;
				int sameRun = 0;

				WeightsPopulationRunner population = getFirstPopulation(collectorData, random);
				for (int i = 0; i < 100; i++) {

					WeightsPopulationRunner finalPopulation = population;
					PersistentParallel6.runAndFinish(finalPopulation.getRuns(), (currentRun) -> {
						WeightsRepository currentWeightsRepository = finalPopulation.getRun(currentRun);
						double performance = run.run(currentWeightsRepository);

						finalPopulation.addPerformance(currentRun, performance);
					}, this.getClass());

					if (i + 1 < 100) {
						if (lastRun == null)
							lastRun = population.getOptimizedRun();
						else if (lastRun.equals(population.getOptimizedRun())) {
							sameRun++;
						} else {
							sameRun = 0;
							lastRun = population.getOptimizedRun();
						}

						if (sameRun == 10) {
							System.out.println("BREAKING AFTER " + i + " runs");
							totalBreakage += i;
							 break;
						}

						population = getNextPopulation(collectorData, population, random);
					}
					else {
						// totalBreakage += i;
					}
				}
				weights = population.getLearnedWeights();
			}

		}
		return new WeightsRepositoryImpl(weights);
	}

	WeightsPopulationRunner getFirstPopulation(WeightsCollectorData collectorData, Random random) {

		Map<WeightsCurrentRun, Map<String, Integer>> allRuns = WeightsRandomRunner
				.createRuns(collectorData.getSTEPS(), collectorData.getNames(), TRIAL_PER_RUN * FIRST_RUN_MULTIPLIER, random);

		return new WeightsPopulationRunner(allRuns, collectorData);
	}

	WeightsPopulationRunner getNextPopulation(WeightsCollectorData collector, WeightsPopulationRunner population, Random random) {


		NavigableMap<WeightsCurrentRun, Double> rankedRuns = population.getRankedRuns();
		Double SUM_OF_WEIGHTS = SUM_OF_WEIGHTS_MAP.get(rankedRuns.size());

		WeightsCurrentRun[] listRun = new WeightsCurrentRun[rankedRuns.size()];
		int currentIndex = 0;
		for (Entry<WeightsCurrentRun, Double> entry : rankedRuns.descendingMap().entrySet()) {
			listRun[currentIndex] = entry.getKey();

			currentIndex++;
		}

		Map<WeightsCurrentRun, Map<String, Integer>> newRuns = new HashMap<>();

		MutableInt counter = new MutableInt(0);
		{
			int i = 0;
			for (WeightsCurrentRun run : listRun) {
				Map<String, Integer> rawRun = new HashMap<>(population.getRawRun(run));
				newRuns.put(new WeightsCurrentRun(counter.intValue()), rawRun);

				counter.increment();
				i++;
				if (i == 1)
					break;
			}
		}

		for (int i = 0; i < 99; i++) {
			double weight1 = random.nextDouble() * SUM_OF_WEIGHTS;
			double weight2 = random.nextDouble() * SUM_OF_WEIGHTS;

			WeightsCurrentRun parent1 = getParent(listRun, weight1);
			WeightsCurrentRun parent2 = getParent(listRun, weight2);
			while (parent1.equals(parent2)) {
				weight1 = random.nextDouble() * SUM_OF_WEIGHTS;
				weight2 = random.nextDouble() * SUM_OF_WEIGHTS;

				parent1 = getParent(listRun, weight1);
				parent2 = getParent(listRun, weight2);
			}
			//


			if (true) {
			boolean recombinationMethod = random.nextBoolean();
				double mutationProb = random.nextDouble();

			if (recombinationMethod) {

				Map<String, Integer> recombined = averageTwoEntries(population.getRawRun(parent1), population.getRawRun(parent2));

					if (mutationProb < 0.5) {
						recombined = mutatedWithAndWithoutCreep(recombined, random, collector.getNamesIndexReverse(), STEPS, CREEP_MUTATION);
					}
					newRuns.put(new WeightsCurrentRun(counter.intValue()), recombined);
				counter.increment();

			} else {
				Map<String, Integer> recombined = randomizeTwoValues(population.getRawRun(parent1), population.getRawRun(parent2), random);
					if (mutationProb < 0.5) {
						recombined = mutatedWithAndWithoutCreep(recombined, random, collector.getNamesIndexReverse(), STEPS, CREEP_MUTATION);
					}
					newRuns.put(new WeightsCurrentRun(counter.intValue()), recombined);
				counter.increment();

			}
			} else {

				int typeOfChild = random.nextInt(4);

				switch (typeOfChild) {
				case 0:
					newRuns.put(new WeightsCurrentRun(counter.intValue()), averageTwoEntries(population.getRawRun(parent1), population.getRawRun(parent2)));
					counter.increment();
					break;
				case 1:
					newRuns.put(new WeightsCurrentRun(counter.intValue()),
							randomizeTwoValues(population.getRawRun(parent1), population.getRawRun(parent2), random));
					counter.increment();
					break;
				case 2:
					newRuns.put(new WeightsCurrentRun(counter.intValue()),
							randomizeTwoValuesWithMutation(population.getRawRun(parent1), population.getRawRun(parent2), random, collector.getSTEPS()));
					counter.increment();
					break;
				case 3:
					Map<String, Integer> averaged = averageTwoEntries(population.getRawRun(parent1), population.getRawRun(parent2));
					Map<String, Integer> creepMutaed = creepMutated(averageTwoEntries(population.getRawRun(parent1), population.getRawRun(parent2)), random,
							STEPS, CREEP_MUTATION);
					newRuns.put(new WeightsCurrentRun(counter.intValue()), creepMutaed);
					counter.increment();
					break;
				}
			}




		}

		int size = newRuns.size();

		return new WeightsPopulationRunner(newRuns, collector);

	}

	private Map<String, Integer> mutatedWithAndWithoutCreep(Map<String, Integer> input, Random random, Map<Integer, String> variableIndeces, int STEPS,
			int CREEP_MUTATION) {
		input = new HashMap<>(input);
		int variableIndex = random.nextInt(variableIndeces.size());
		String variable = variableIndeces.get(variableIndex);
		Integer oldValue = input.get(variable);

		boolean creep = random.nextBoolean();
		if (creep) {
			int change = random.nextInt(CREEP_MUTATION * 2) - CREEP_MUTATION;

			int newValue = Math.min(STEPS, Math.max(0, oldValue + change));
			input.put(variable, newValue);
		} else {
			input.put(variable, random.nextInt(STEPS + 1));
		}

		return input;
	}

	private Map<String, Integer> creepMutated(Map<String, Integer> averageTwoEntries, Random random, int STEPS, int CREEP_MUTATION) {
		averageTwoEntries = new HashMap<>(averageTwoEntries);

		for (Entry<String, Integer> entry : averageTwoEntries.entrySet()) {
			boolean choice = random.nextBoolean();
			if (choice) {
				int change = random.nextInt(CREEP_MUTATION * 2) - CREEP_MUTATION;

				int newValue = Math.min(STEPS, Math.max(0, entry.getValue() + change));
				entry.setValue(newValue);
			}
		}

		return averageTwoEntries;
	}

	private Map<String, Integer> randomizeTwoValuesWithMutation(Map<String, Integer> parent1, Map<String, Integer> parent2, Random random, int STEPS) {
		Map<String, Integer> ret = new HashMap<>();

		for (Entry<String, Integer> parent1Entry : parent1.entrySet()) {
			Integer parent1Value = parent1Entry.getValue();
			Integer parent2Value = parent2.get(parent1Entry.getKey());

			boolean choice = random.nextBoolean();
			if (choice) {
				ret.put(parent1Entry.getKey(), random.nextInt(STEPS + 1));
			} else {
				boolean secondChoice = random.nextBoolean();
				if (secondChoice) {
					ret.put(parent1Entry.getKey(), parent1Value);
				} else {
					ret.put(parent1Entry.getKey(), parent2Value);
				}
			}

		}

		return ret;
	}

	private Map<String, Integer> randomizeTwoValues(Map<String, Integer> parent1, Map<String, Integer> parent2, Random random) {
		Map<String, Integer> ret = new HashMap<>();

		for (Entry<String, Integer> parent1Entry : parent1.entrySet()) {
			Integer parent1Value = parent1Entry.getValue();
			Integer parent2Value = parent2.get(parent1Entry.getKey());

			boolean choice = random.nextBoolean();
			if (choice) {
				ret.put(parent1Entry.getKey(), parent1Value);
			} else {
				ret.put(parent1Entry.getKey(), parent2Value);
			}

		}

		return ret;
	}

	private Map<String, Integer> averageTwoEntries(Map<String, Integer> parent1, Map<String, Integer> parent2) {
		Map<String, Integer> ret = new HashMap<>();

		for (Entry<String, Integer> parent1Entry : parent1.entrySet()) {
			Integer parent1Value = parent1Entry.getValue();
			Integer parent2Value = parent2.get(parent1Entry.getKey());

			Integer newValue = (parent1Value + parent2Value) / 2;

			ret.put(parent1Entry.getKey(), newValue);
		}

		return ret;
	}

	private WeightsCurrentRun getParent(WeightsCurrentRun[] listRun, double weight) {
		return listRun[lookupMap.floorEntry(weight).getValue()];
	}
}
