package de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.genetic;

import java.math.BigInteger;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.utils.MathsUtil;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsRepositoryCollector;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeight;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class WeightsGeneticRunner {

	public static final int MAX_TRIALS_PER_RUN = 100;
	
	public static final int MAX_CHANGE = 2;
	private Map<WeightsCurrentRun, Double> results = new ConcurrentHashMap<>();

	private final Map<WeightsCurrentRun, Map<String, Integer>> runs;

	private final OptimizableClass workingClass;

	private final float STEP;
	private final int STEPS;

	private final Set<Map<String, Integer>> previousRuns = new LinkedHashSet<>();

	private Map<String, Integer> bestRun = null;

	Optional<Boolean> isDone = Optional.of(false);

	private final SortedSet<String> names;
	private final Map<String, Integer> namesIndex;
	private final Map<Integer, String> namesIndexReverse;

 
	private final int trials;
	private final Random random;
	public WeightsGeneticRunner(WeightsRepositoryCollector collector ) {
		this(collector, new Random(collector.getWorkingClass().getName().hashCode()), MAX_TRIALS_PER_RUN);
	}
	public WeightsGeneticRunner(WeightsRepositoryCollector collector, Random rnd ) {
		this(collector, rnd, MAX_TRIALS_PER_RUN);
	}
	public WeightsGeneticRunner(WeightsRepositoryCollector collector, Random rnd, int trials) {

		this.STEP = collector.getSTEP();
		this.STEPS = collector.getSTEPS();
		this.workingClass = collector.getWorkingClass();
		this.names = new TreeSet<>(collector.getCollectorIndex().keySet()); 
		namesIndex = new HashMap<>();
		namesIndexReverse= new HashMap<>();
		int i = 0;
		for (String name : names) { 
			namesIndex.put(name, i);
			namesIndexReverse.put(i, name);
			i++;
		} 
		 
		this.random = rnd; 
		this.trials = trials;
		this.runs = Collections.unmodifiableMap(createRuns());
	}

 

	 

	 

	public boolean isDone() {
		if (results.size() != runs.size()) {
			return false;
		}
		else {
			double maxPerformance = 0.0;
			WeightsCurrentRun bestRun = null;
			 
			for(Entry<WeightsCurrentRun, Double> entry : results.entrySet()) {
				if(bestRun == null || entry.getValue() > maxPerformance) {
					maxPerformance = entry.getValue();
					bestRun = entry.getKey();
				}
			}
			
			
			this.bestRun = runs.get(bestRun);
			
			return true;
		}
	}

	private Map<WeightsCurrentRun, Map<String, Integer>> getAllRuns() {
		Map<WeightsCurrentRun, Map<String, Integer>> allRuns = new HashMap<>();

		NavigableMap<String, Integer> lastRun = null;

		{
			NavigableMap<String, Integer> currentRun = new TreeMap<>();
			for (String index : this.names) {
				currentRun.put(index, STEPS);
			}
			allRuns.put(new WeightsCurrentRun(allRuns.size()), currentRun);
			lastRun = currentRun;
		}

		boolean isDone = false;
		while (!isDone) {

			NavigableMap<String, Integer> currentRun = new TreeMap<>(lastRun);

			int changeIndex = -1;
			for (Entry<String, Integer> entry : currentRun.descendingMap().entrySet()) {
				if (entry.getValue().equals(0)) {
					changeIndex = namesIndex.get(entry.getKey()) - 1;
				} else {
					changeIndex = namesIndex.get(entry.getKey());
					break;
				}
			}

			if (changeIndex == -1) {
				isDone = true;
			} else {
				{
					Integer step = currentRun.get(namesIndexReverse.get(changeIndex));
					step--;
					currentRun.put(namesIndexReverse.get(changeIndex), step);
				}
				for (int i = changeIndex + 1; i < currentRun.size(); i++) {
					currentRun.put(namesIndexReverse.get(i), STEPS);
				}
				allRuns.put(new WeightsCurrentRun( allRuns.size()), currentRun);
				lastRun = currentRun;
			}

		}

		return allRuns;

	} 

	
	private Map<WeightsCurrentRun, Map<String, Integer>> createRuns() {
		Map<WeightsCurrentRun, Map<String, Integer>> newCurrentRuns = new HashMap<>();
		Set< Map<String, Integer>> runsSet = new HashSet<>();
		
		BigInteger numberOfOutcomes = MathsUtil.outcomes(STEPS +1, names.size());
		
		if (numberOfOutcomes.compareTo(BigInteger.valueOf(trials)) <= 0) {
			newCurrentRuns = getAllRuns();
		}
		else {
			for(int i = 0; i < trials; i++) {
				Map<String, Integer> run = new HashMap<>();
				WeightsCurrentRun key = new WeightsCurrentRun(i);
				do {
					for(String name : names) {
						run.put(name, random.nextInt(STEPS + 1));
					}
				} while(!runsSet.add(run));
				
				newCurrentRuns.put(key, Collections.unmodifiableMap(run));
			}
		}
		return newCurrentRuns;
	}
	public Set<WeightsCurrentRun> getRuns() {
	 return runs.keySet();

	}
	void addPerformance(WeightsCurrentRun run, double performance) {
		if(runs.containsKey(run))
			results.put(run, performance);
		else
			throw new IllegalStateException();
	}



	WeightsRepository getRun(WeightsCurrentRun run) {
		return new WeightsRepositorySingleRun(run);
	}

	public class WeightsRepositorySingleRun implements WeightsRepository {

		private final WeightsCurrentRun run;

		private WeightsRepositorySingleRun(WeightsCurrentRun run) {
			super();
			this.run = run;
		}

		@Override
		public double getOrDefault(OptimizableWeight weight) {
			if (workingClass.equals(weight.getKey().getOptimizableClazz())) {
				int currentStep = runs.get(run).get(weight.getKey().getName());

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;

				return value;
			}

			throw new IllegalStateException();
		}

		@Override
		public Optional<Double> get(OptimizableWeight weight) {
			return Optional.of(getOrDefault(weight));
		}

		@Override
		public boolean containsWeightsForClass(OptimizableClass clazz) {
			if (workingClass != null && workingClass.equals(clazz)) {
				return true;
			}
			return false;
		}

		@Override
		public Map<OptimizableWeightKey, Double> getWeights() {
			Map<OptimizableWeightKey, Double> ret = new HashMap<>();

			Map<String, Integer> currentRun = runs.get(run);
			for (Entry<String, Integer> weightEntry : currentRun.entrySet()) {
				int currentStep = weightEntry.getValue();

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;
				ret.put(new OptimizableWeightKey(workingClass, weightEntry.getKey()), value);
			}
			return ret;
		}

		@Override
		public Map<String, Double> getWeights(OptimizableClass optimizableClass) {
			Map<String, Double> ret = new HashMap<>();

			Map<String, Integer> currentRun = runs.get(run);
			for (Entry<String, Integer> weightEntry : currentRun.entrySet()) {
				int currentStep = weightEntry.getValue();

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;
				ret.put(weightEntry.getKey(), value);
			}
			return ret;
		}

	}

	public static class WeightsCurrentRun implements Comparable<WeightsCurrentRun> {
		private final int run;

		private WeightsCurrentRun(  int run) { 
			this.run = run;
		}

	 

		private int getRun() {
			return run;
		}

		@Override
		public int compareTo(WeightsCurrentRun o) {
			return Integer.compare(run, o.run);
		}

	}

	public Map<OptimizableWeightKey, Double> getLearnedWeights() {
		if (!isDone()) {
			throw new IllegalStateException("This optimizing run is not done, why are you calling this method");
		}

		Map<OptimizableWeightKey, Double> weights = bestRun.entrySet().stream().collect(Collectors.toMap((e) -> {
			return new OptimizableWeightKey(workingClass, e.getKey());
		}, (e) -> {
			return Double.valueOf(Math.round(100.0 * ((e.getValue() * STEP))) / 100.0);
		}));
		return weights;
	}

	public Map<String, Integer> getOptimizedRun() {
		return bestRun;
	}

 
}
