package de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.parallel;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsRepositoryCollector;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeight;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class WeightsRepositoryRunner {

	private final float STEP;
	private final OptimizableClass workingClass;
	private final NavigableMap<Integer, NavigableMap<Integer, Integer>> allRuns;
	private final Map<String, Integer> collectorIndex;
	private final Map<Integer, String> collector;
	private final Map<Integer, Double> results = new ConcurrentHashMap<>();

	public WeightsRepositoryRunner(WeightsRepositoryCollector collector) {
		this.workingClass = collector.getWorkingClass();
		this.STEP = collector.getSTEP();
		this.collector = collector.getCollector();
		this.collectorIndex = collector.getCollectorIndex();
		this.allRuns = getAllRuns(collector.getSTEPS());
	}

	private NavigableMap<Integer, NavigableMap<Integer, Integer>> getAllRuns(int STEPS) {
		TreeMap<Integer, NavigableMap<Integer, Integer>> allRuns = new TreeMap<>();

		NavigableMap<Integer, Integer> lastRun = null;

		{
			NavigableMap<Integer, Integer> currentRun = new TreeMap<>();
			for (Integer index : this.collector.keySet()) {
				currentRun.put(index, STEPS);
			}
			allRuns.put(allRuns.size(), currentRun);
			lastRun = currentRun;
		}

		boolean isDone = false;
		while (!isDone) {

			NavigableMap<Integer, Integer> currentRun = new TreeMap<>(lastRun);

			int changeIndex = -1;
			for (Entry<Integer, Integer> entry : currentRun.descendingMap().entrySet()) {
				if (entry.getValue().equals(0)) {
					changeIndex = entry.getKey() - 1;
				} else {
					changeIndex = entry.getKey();
					break;
				}
			}

			if (changeIndex == -1) {
				isDone = true;
			} else {
				{
					Integer step = currentRun.get(changeIndex);
					step--;
					currentRun.put(changeIndex, step);
				}
				for (int i = changeIndex + 1; i < currentRun.size(); i++) {
					currentRun.put(i, STEPS);
				}
				allRuns.put(allRuns.size(), currentRun);
				lastRun = currentRun;
			}

		}

		return allRuns;

	}

	public Set<Integer> getAllRuns() {
		return Collections.unmodifiableSet(allRuns.keySet());
	}

	WeightsRepository getRun(Integer run) {
		return new WeightsRepositorySingleRun(run);
	}

	void addPerformance(Integer run, double performance) {
		results.put(run, performance);
	}

	Map<OptimizableWeightKey, Double> getLearnedWeights() {
		if (results.size() != allRuns.size()) {
			throw new IllegalStateException("We did not receive performance for all runs");
		}

		Integer maxRun = results.entrySet().stream().max((e1, e2) -> {
			int compare = Double.compare(e1.getValue(), e2.getValue());
			if (compare != 0) {
				return compare;
			} else {
				compare = Integer.compare(allRuns.get(e1.getKey()).values().stream().mapToInt(Integer::intValue).sum(),
						allRuns.get(e2.getKey()).values().stream().mapToInt(Integer::intValue).sum());
				if (compare != 0) {
					return compare;
				} else {
					return Integer.compare(e1.getKey(), e2.getKey());
				}
			}
		}).get().getKey();

		Map<OptimizableWeightKey, Double> weights = allRuns.get(maxRun).entrySet().stream().collect(Collectors.toMap((e) -> {
			return new OptimizableWeightKey(workingClass, collector.get(e.getKey()));
		}, (e) -> {
			return Double.valueOf(Math.round(100.0 * ((e.getValue() * STEP))) / 100.0);
		}));

		return weights;
	}
	public class WeightsRepositorySingleRun implements WeightsRepository {

		private final Integer run;

		private WeightsRepositorySingleRun(Integer run) {
			super();
			this.run = run;
		}

		@Override
		public double getOrDefault(OptimizableWeight weight) {
			if (workingClass.equals(weight.getKey().getOptimizableClazz())) {
				int currentStep = allRuns.get(run).get(collectorIndex.get(weight.getKey().getName()));

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;

				return value;
			}

			throw new IllegalStateException();
		}

		@Override
		public Optional<Double> get(OptimizableWeight weight) {
			return Optional.of(getOrDefault(weight));
		}

		@Override
		public boolean containsWeightsForClass(OptimizableClass clazz) {
			if (workingClass != null && workingClass.equals(clazz)) {
				return true;
			}
			return false;
		}

		@Override
		public Map<OptimizableWeightKey, Double> getWeights() {
			Map<OptimizableWeightKey, Double> ret = new HashMap<>();

			NavigableMap<Integer, Integer> currentRun = allRuns.get(run);
			for (Entry<String, Integer> weightEntry : collectorIndex.entrySet()) {
				int currentStep = currentRun.get(weightEntry.getValue());

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;
				ret.put(new OptimizableWeightKey(workingClass, weightEntry.getKey()), value);
			}
			return ret;
		}

		@Override
		public Map<String, Double> getWeights(OptimizableClass optimizableClass) {
			throw new IllegalStateException();
		}

	}
}
