package de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.simple;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.Optional;
import java.util.TreeMap;
import java.util.stream.Collectors;

import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeight;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class WeightsOptimizingRepository implements WeightsRepository {

	public static float STEP = 0.05f;
	public static int STEPS = Math.round((1.0f / STEP));

	private OptimizableClass workingClass;

	private Map<String, Integer> collectorIndex = new HashMap<>();
	private NavigableMap<Integer, String> collector = new TreeMap<>();

	private NavigableMap<Integer, Double> results = new TreeMap<>();

	private NavigableMap<Integer, NavigableMap<Integer, Integer>> runs = new TreeMap<>();

	private NavigableMap<Integer, Integer> currentRun = new TreeMap<>();

	private boolean done = true;

	private boolean locked = false;


	public void addWeight(OptimizableWeight weight) {
		if (locked) {
			if (workingClass == null) {
				throw new IllegalStateException("This repository is locked. How is working class null?");
			}
			if (!workingClass.equals(weight.getKey().getOptimizableClazz())) {
				throw new IllegalStateException("This repository is locked. How is there a different optimizable class?");
			}
			if (!collectorIndex.containsKey(weight.getKey().getName())) {
				throw new IllegalStateException("This repository is locked. How is there an additional weight that needs to be optimized.");
			}
		}
		else {
			if (workingClass == null) {
				workingClass = weight.getKey().getOptimizableClazz();
			} else {
				if (!workingClass.equals(weight.getKey().getOptimizableClazz())) {
					throw new IllegalStateException("You can learn weights only for one optimizable class at a time.");
				}
			}

			String name = weight.getKey().getName();
			if (!collectorIndex.containsKey(name)) {
				int index = collector.size();
				collector.put(index, name);
				collectorIndex.put(name, index);
				currentRun.put(index, STEPS);
			}

			done = false;

		}
	}

	@Override
	public double getOrDefault(OptimizableWeight weight) {
		addWeight(weight);

		Integer index = collectorIndex.get(weight.getKey().getName());

		double value = Math.round(100.0 * ((currentRun.get(index) * STEP))) / 100.0;

		return value;
	}

	@Override
	public Optional<Double> get(OptimizableWeight weight) {
		return Optional.of(getOrDefault(weight));
	}

	public boolean isDoneOptimizing() {
		return done;
	}

	public void lastRunPerformance(double d) {
		this.locked = true;
		int index = results.size();
		results.put(index, d);

		runs.put(index, currentRun);

		NavigableMap<Integer, Integer> newRun = new TreeMap<>(currentRun);

		int changeIndex = -1;
		boolean allZeros = true;
		for (Entry<Integer, Integer> entry : newRun.descendingMap().entrySet()) {
			if (allZeros && entry.getValue().equals(0)) {
				changeIndex = entry.getKey() - 1;
			}
			else {
				allZeros = false;
				changeIndex = entry.getKey();
				break;
			}
		}

		if (changeIndex == -1) {
			this.done = true;
			this.currentRun = null;
		} else {
			{
				Integer step = newRun.get(changeIndex);
				step--;
				newRun.put(changeIndex, step);
			}
			for (int i = changeIndex + 1; i < newRun.size(); i++) {
				newRun.put(i, STEPS);
			}
			this.currentRun = newRun;
		}



	}

	@Override
	public boolean containsWeightsForClass(OptimizableClass clazz) {
		if (workingClass != null && workingClass.equals(clazz)) {
			return true;
		}
		return false;
	}

	@Override
	public Map<OptimizableWeightKey, Double> getWeights() {
		Integer maxRun = results.entrySet().stream().max((e1, e2) -> Double.compare(e1.getValue(), e2.getValue())).get().getKey();

		Map<OptimizableWeightKey, Double> weights = runs.get(maxRun).entrySet().stream().collect(Collectors.toMap((e) -> {
			return new OptimizableWeightKey(workingClass, collector.get(e.getKey()));
		}, (e) -> {
			return Double.valueOf(Math.round(100.0 * ((e.getValue() * STEP))) / 100.0);
		}));

		return weights;
	}

	@Override
	public Map<String, Double> getWeights(OptimizableClass optimizableClass) {
		throw new IllegalStateException();
	}

}
