package de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights;

import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.stream.Collectors;

public class WeightsRepositoryImpl implements WeightsRepository, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private final Map<OptimizableWeightKey, Double> weights;
	private final Set<OptimizableClass> classes;

	public WeightsRepositoryImpl() {
		this(new HashMap<>());
	}

	public WeightsRepositoryImpl(Map<OptimizableWeightKey, Double> weights) {
		this.weights = Collections.unmodifiableMap(new TreeMap<>(weights));
		this.classes = Collections.unmodifiableSet(this.weights.keySet().stream().map(OptimizableWeightKey::getOptimizableClazz).collect(Collectors.toSet()));
	}

	public WeightsRepositoryImpl(WeightsRepository weights) {
		this(weights.getWeights());
	}

	public WeightsRepositoryImpl combineWith(WeightsRepository other) {
		Map<OptimizableWeightKey, Double> map = new HashMap<>();
		map.putAll(this.weights);
		map.putAll(other.getWeights());

		return new WeightsRepositoryImpl(map);

	}

	@Override
	public double getOrDefault(OptimizableWeight weight) {
		Double weightValue = weights.get(weight.getKey());
		if (weightValue == null) {
			return weight.getDefaultValue();
		} else {
			return weightValue;
		}
	}


	@Override
	public Optional<Double> get(OptimizableWeight weight) {
		Double weightValue = weights.get(weight.getKey());
		if (weightValue == null) {
			return Optional.empty();
		} else {
			return Optional.of(weightValue);
		}
	}

	@Override
	public boolean containsWeightsForClass(OptimizableClass clazz) {
		return classes.contains(clazz);
	}

	@Override
	public Map<OptimizableWeightKey, Double> getWeights() {
		return weights;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((classes == null) ? 0 : classes.hashCode());
		result = prime * result + ((weights == null) ? 0 : weights.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof WeightsRepositoryImpl))
			return false;
		WeightsRepositoryImpl other = (WeightsRepositoryImpl) obj;
		if (classes == null) {
			if (other.classes != null)
				return false;
		} else if (!classes.equals(other.classes))
			return false;
		if (weights == null) {
			if (other.weights != null)
				return false;
		} else if (!weights.equals(other.weights))
			return false;
		return true;
	}

	@Override
	public Map<String, Double> getWeights(OptimizableClass optimizableClass) {
		Map<String, Double> ret = weights.entrySet().stream().filter((e) -> e.getKey().getOptimizableClazz().equals(optimizableClass))
				.collect(Collectors.toMap((e) -> e.getKey().getName(), Entry::getValue));

		return ret;
	}

	@Override
	public String toString() {
		return "WeightsRepositoryImpl [weights=" + weights + "]";
	}

}
