package de.mannheim.uni.expansion.utils;

import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NavigableSet;
import java.util.TreeSet;

import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.containers.Quad;

public class QuartileData<E> {

	public static enum Quartile {
		Q1, Q2, Q3, Q4;
	}

	private final int q1IndexIncluded;
	private final Double q1Value;
	private final int q2IndexIncluded;
	private final Double q2Value;
	private final int q3IndexIncluded;
	private final Double q3Value;
	private final Double q4Value;
	private final Map<Quartile, NavigableSet<E>> selectedPerQuartile;

	private final Comparator<E> comparator;

	public QuartileData(List<E> list, Map<E, Number> degreeMapping) {

		Map<E, Integer> sortingIndex = new HashMap<>();
		for (ListEntry<E> entry : ListIterable.iterable(list)) {
			sortingIndex.put(entry.getValue(), entry.getIndex());
		}

		Quad<Double, Integer, Integer, Integer> q2 = getMedianIncludedIndexAndSubListIndeces(0, list, degreeMapping);
		q2Value = q2.getA();
		q2IndexIncluded = q2.getB();

		Quad<Double, Integer, Integer, Integer> q1 = getMedianIncludedIndexAndSubListIndeces(0, list.subList(0, q2.getC() + 1), degreeMapping);
		q1Value = q1.getA();
		q1IndexIncluded = q1.getB();

		Quad<Double, Integer, Integer, Integer> q3 = getMedianIncludedIndexAndSubListIndeces(q2.getD(), list.subList(q2.getD(), list.size()), degreeMapping);
		q3Value = q3.getA();
		q3IndexIncluded = q3.getB();

		q4Value = degreeMapping.get(list.get(list.size() - 1)).doubleValue();

		Map<Quartile, NavigableSet<E>> selectedPerQuartile = new HashMap<>();

		Quartile quartile = Quartile.Q1;
		int indexIncluded = q1IndexIncluded;

		this.comparator = (e1, e2) -> {
			// if (e1 == null || e2 == null || sortingIndex.get(e1) == null ||
			// sortingIndex.get(e2) == null) {
			// int sdf = 123;
			// }
			return Integer.compare(sortingIndex.get(e1), sortingIndex.get(e2));
		};

		NavigableSet<E> currentQuartile = new TreeSet<>(comparator);
		for (ListEntry<E> entry : ListIterable.iterable(list)) {

			if (entry.getIndex() < indexIncluded) {
				currentQuartile.add(entry.getValue());
			} else {
				selectedPerQuartile.put(quartile, currentQuartile);
				currentQuartile = new TreeSet<>(comparator);
				currentQuartile.add(entry.getValue());
				switch (quartile) {
				case Q1:
					quartile = Quartile.Q2;
					indexIncluded = q2IndexIncluded;
					break;
				case Q2:
					quartile = Quartile.Q3;
					indexIncluded = q3IndexIncluded;
					break;
				case Q3:
					quartile = Quartile.Q4;
					indexIncluded = list.size();
					break;
				case Q4:
					throw new IllegalStateException("We should not reach the last quartile in the loop");

				}
			}

		}
		selectedPerQuartile.put(quartile, currentQuartile);

		this.selectedPerQuartile = ImmuMaker.mapWithNavigableSet(selectedPerQuartile);
	}

	private Quad<Double, Integer, Integer, Integer> getMedianIncludedIndexAndSubListIndeces(int startIndex, List<E> list, Map<E, Number> degreeMapping) {
		if (list.size() % 2 == 0) {
			int index2 = list.size() / 2;
			int index1 = index2 - 1;
			double median = (degreeMapping.get(list.get(index1)).doubleValue() + degreeMapping.get(list.get(index2)).doubleValue()) / 2.0;

			return new Quad<>(median, startIndex + index2, startIndex + index1, startIndex + index2);
		} else {
			int medianIndex = (list.size() - 1) / 2;
			double median = degreeMapping.get(list.get(medianIndex)).doubleValue();

			return new Quad<>(median, startIndex + medianIndex, startIndex + medianIndex - 1, startIndex + medianIndex + 1);
		}
	}

	public int getQ1IndexIncluded() {
		return q1IndexIncluded;
	}

	public Double getQ1Value() {
		return q1Value;
	}

	public int getQ2IndexIncluded() {
		return q2IndexIncluded;
	}

	public Double getQ2Value() {
		return q2Value;
	}

	public int getQ3IndexIncluded() {
		return q3IndexIncluded;
	}

	public Double getQ3Value() {
		return q3Value;
	}

	public Map<Quartile, NavigableSet<E>> getSelectedPerQuartile() {
		return selectedPerQuartile;
	}

	public Double getBeginning(Quartile q) {
		switch (q) {
		case Q1:
			return 0.0;
		case Q2:
			return q1Value;
		case Q3:
			return q2Value;
		case Q4:
			return q3Value;

		}

		throw new IllegalStateException();
	}

	public Double getEnd(Quartile q) {
		switch (q) {
		case Q1:
			return q1Value;
		case Q2:
			return q2Value;
		case Q3:
			return q3Value;
		case Q4:
			return q4Value;

		}

		throw new IllegalStateException();
	}

	public Double getQ4Value() {
		return q4Value;
	}

	public Comparator<E> getComparator() {
		return comparator;
	}

}
