package de.mannheim.uni.expansion.utils;

import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.function.Supplier;

public class SimpleLockableConcurrentCache<E, F> {

	private final static LinkedInheritableThreadLocal alreadyInLock = new LinkedInheritableThreadLocal();

	private final ConcurrentHashMap<E, CountDownLatch> locks = new ConcurrentHashMap<>();

	private final ConcurrentHashMap<E, F> cache = new ConcurrentHashMap<>();

	private final static ConcurrentLinkedQueue<SimpleLockableConcurrentCache<?, ?>> allCaches = new ConcurrentLinkedQueue<>();

	public final static void clearAllCaches() {
		for (SimpleLockableConcurrentCache<?, ?> cache : allCaches) {
			cache.clear();
		}
	}

	public SimpleLockableConcurrentCache() {
		allCaches.add(this);
	}

	public boolean hasCacheFor(E key) {
		return cache.containsKey(key);
	}

	public void add(E key, F value) {
		cache.put(key, value);
	}

	public F get(E key) {
		return cache.get(key);
	}

	public F poll(E key) {
		if (hasCacheFor(key)) {
			return get(key);
		} else {
			CountDownLatch latch = locks.get(key);
			if (latch == null) {
				return null;
			} else {
				try {
					PersistentParallel6.setThreadPaused();
					latch.await();
					PersistentParallel6.setThreadUnpaused();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				return cache.get(key);
			}
		}
	}

	public Optional<Consumer<F>> getLock(E key) {
		if (cache.containsKey(key) || locks.containsKey(key)) {
			return Optional.empty();
		} else {
			CountDownLatch latch = new CountDownLatch(1);
			CountDownLatch previousLatch = locks.putIfAbsent(key, latch);
			if (previousLatch != null) {
				return Optional.empty();
			} else {

				Consumer<F> consumer = (value) -> {
					cache.put(key, value);
					latch.countDown();
				};
				return Optional.of(consumer);
			}

		}
	}

	public F getOrCreate(E key, Supplier<F> creator) {
		// this.cache.clear();
		if (hasCacheFor(key)) {
			return get(key);
		} else {

			Optional<Consumer<F>> lock = getLock(key);
			if (lock.isPresent()) {
				alreadyInLock.get().setHasLock();
				F val = creator.get();
				lock.get().accept(val);

				alreadyInLock.get().removeHasLock();
				return val;
			} else {

				if (alreadyInLock.get().hasLock()) {
					F val = creator.get();
					cache.put(key, val);
					return val;
				} else {
					F val = poll(key);
					return val;
				}

			}
		}

	}

	public F getOrCreateNoBlocking(E key, Supplier<F> creator) {
		return CHMHelper.getFromCHM(cache, key, () -> creator.get());

	}

	public void clear() {
		this.cache.clear();
		this.locks.clear();
	}

	public void remove(E key) {
		this.cache.get(key);
	}

	private static class LinkedInheritableThreadLocal extends InheritableThreadLocal<LockLink> {

		protected LockLink childValue(LockLink parentValue) {
			return new LockLink(parentValue, new AtomicBoolean(false));
		}

		public void set(LockLink value) {
			throw new UnsupportedOperationException();
		}

		protected LockLink initialValue() {
			return new LockLink(null, new AtomicBoolean(false));
		}
	}

	private static class LockLink {

		private final LockLink parent;
		private final AtomicBoolean hasLock;

		public LockLink(LockLink parent, AtomicBoolean hasLock) {
			super();
			this.parent = parent;
			this.hasLock = hasLock;
		}

		boolean hasLock() {
			if (hasLock.get())
				return true;
			else if (parent == null) {
				return false;
			} else {
				return parent.hasLock();
			}
		}

		void setHasLock() {
			hasLock.set(true);
		}

		void removeHasLock() {
			hasLock.set(false);
		}

	}
}
