package de.mannheim.uni.expansion.utils;

import java.io.Serializable;
import java.util.AbstractCollection;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;

public class SparseArrayMap<T> implements Map<Integer, T>, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private SparseArray<T> data;


	public SparseArrayMap() {
		this.data = new SparseArray<T>();
	}

	public SparseArrayMap(int initialSize) {
		this.data = new SparseArray<T>(initialSize);
	}

	private int getIndex(Object key) {
		if (key instanceof Integer && key != null) {
			int index = (int) key;
			if (index >= 0) {
				return index;
			} else {
				throw new IllegalStateException("Key must be of higher or equal to 0 to");
			}
		} else {
			throw new IllegalStateException("Key must be of type Integer!");
		}
	}

	@Override
	public int size() {
		return data.size();
	}

	@Override
	public boolean isEmpty() {
		if (data.size() > 0)
			return false;
		else
			return true;
	}

	@Override
	public boolean containsKey(Object key) {
		int index = getIndex(key);
		if (data.get(index) != null) {
			return true;
		}
		return false;
	}

	@Override
	public boolean containsValue(Object value) {
		throw new UnsupportedOperationException();
	}

	@Override
	public T get(Object key) {
		int index = getIndex(key);
		return data.get(index);
	}

	public T get(int index) {
		return data.get(index);
	}

	@Override
	public T put(Integer key, T value) {
		int index = getIndex(key);
		T oldValue =  data.get(index);
		data.append(index, value);
		return oldValue;
	}

	@Override
	public T remove(Object key) {
		int index = getIndex(key);
		T oldValue = data.get(index);
		data.remove(index);
		return oldValue;
	}

	@Override
	public void putAll(Map<? extends Integer, ? extends T> m) {
		for (java.util.Map.Entry<? extends Integer, ? extends T> e : m.entrySet()) {
			this.put(e.getKey(), e.getValue());
		}

	}

	@Override
	public void clear() {
		data.clear();

	}


	@Override
	public Set<Integer> keySet() {
		return new KeySet();

	}

	class KeySet extends AbstractSet<Integer> {
		private final int size;

		public KeySet() {
			this.size = data.size();
		}

			@Override
			public Iterator<Integer> iterator() {
			return new IteratorImpl<Integer>(size, SparseArrayMap.this::getKeyAt);
			}

			@Override
			public int size() {
				if (size != data.size()) {
					throw new ConcurrentModificationException();
				}
				return size;
			}
		};

	@Override
	public Collection<T> values() {
		return new ValueCollection();
	}

	class ValueCollection extends AbstractCollection<T> {

		private final int size;
		
		public ValueCollection() {
			this.size = data.size();
		}

		@Override
		public Iterator<T> iterator() {
			return new IteratorImpl<T>(size, SparseArrayMap.this::getValueAt);
		}

		@Override
		public int size() {
			if (size != data.size()) {
				throw new ConcurrentModificationException();
			}
			return size;
		}
	};

	@Override
	public Set<Entry<Integer, T>> entrySet() {

		return new EntrySet();
	}



	class EntrySet extends AbstractSet<Entry<Integer, T>> {
		private final int size;

		public EntrySet() {
			this.size = data.size();
		}

		@Override
		public Iterator<Entry<Integer, T>> iterator() {

			return new IteratorImpl<Entry<Integer, T>>(size, SparseArrayMap.this::getEntryAt);

		}

		@Override
		public int size() {
			if (size != data.size()) {
				throw new ConcurrentModificationException();
			}
			return size;
		}

	};

	class EntryImpl implements Entry<Integer, T> {


		public EntryImpl(int i) {
			key = data.keyAt(i);
			value = data.valueAt(i);
		}

		Integer key;
		T value;

		@Override
		public Integer getKey() {
			return key;
		}

		@Override
		public T getValue() {
			return value;
		}

		@Override
		public T setValue(T newValue) {
			data.append(key, newValue);
			T oldValue = value;
			value = newValue;
			return oldValue;
		}
	}

	private Integer getKeyAt(Integer index) {
		return data.keyAt(index);
	}
	
	
	private T getValueAt(Integer index) {
		return data.valueAt(index);
	}
	
	private Entry<Integer, T> getEntryAt(Integer index) {
		return new EntryImpl(index);
	}


	class IteratorImpl<X> implements Iterator<X> {

		private int i = -1;
		private int size;
		private final Function<Integer, X> getter;

		public IteratorImpl(int size, Function<Integer, X> getter) {
			this.size = size;
			this.getter = getter;
		}

		@Override
		public boolean hasNext() {
			if (size != data.size()) {
				throw new ConcurrentModificationException();
			}
			if ((i + 1) < size) {
				return true;
			} else {
				return false;
			}
		}

		@Override
		public X next() {
			if (size != data.size()) {
				throw new ConcurrentModificationException();
			}
			i++;

			X x = getter.apply(i);

			return x;
		}

		@Override
		public void remove() {
			data.removeAt(i);
			i--;
			size--;
		}
	};
}
