package de.mannheim.uni.expansion.utils.timestamps;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.uima.UIMAFramework;
import org.apache.uima.cas.CASException;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.resource.metadata.TypeSystemDescription;
import org.apache.uima.util.InvalidXMLException;
import org.apache.uima.util.XMLInputSource;

import de.unihd.dbs.heideltime.standalone.CLISwitch;
import de.unihd.dbs.heideltime.standalone.Config;
import de.unihd.dbs.heideltime.standalone.DocumentType;
import de.unihd.dbs.heideltime.standalone.HeidelTimeStandalone;
import de.unihd.dbs.heideltime.standalone.POSTagger;
import de.unihd.dbs.heideltime.standalone.components.JCasFactory;
import de.unihd.dbs.heideltime.standalone.components.PartOfSpeechTagger;
import de.unihd.dbs.heideltime.standalone.components.impl.AllLanguagesTokenizerWrapper;
import de.unihd.dbs.heideltime.standalone.components.impl.HunPosTaggerWrapper;
import de.unihd.dbs.heideltime.standalone.components.impl.IntervalTaggerWrapper;
import de.unihd.dbs.heideltime.standalone.components.impl.JCasFactoryImpl;
import de.unihd.dbs.heideltime.standalone.components.impl.JVnTextProWrapper;
import de.unihd.dbs.heideltime.standalone.components.impl.StanfordPOSTaggerWrapper;
import de.unihd.dbs.heideltime.standalone.components.impl.TreeTaggerWrapper;
import de.unihd.dbs.heideltime.standalone.components.impl.UimaContextImpl;
import de.unihd.dbs.uima.annotator.heideltime.HeidelTime;
import de.unihd.dbs.uima.annotator.heideltime.resources.Language;
import de.unihd.dbs.uima.annotator.intervaltagger.IntervalTagger;
import de.unihd.dbs.uima.types.heideltime.Sentence;
import de.unihd.dbs.uima.types.heideltime.Timex3;

public class HeidelTimeWrapper {

	private static Logger logger = Logger.getLogger("HeidelTimeStandalone");

	private final Language language = Language.ENGLISH;
	private final POSTagger posTagger = POSTagger.NO;
	private final DocumentType typeToProcess = DocumentType.NARRATIVES;

	private final ThreadLocal<HeidelTime> heidelTime = new ThreadLocal<HeidelTime>();

	private HeidelTime getThreadLocalHeidelTime() throws ResourceInitializationException {

		HeidelTime heidelTimeLocal = this.heidelTime.get();
		if (heidelTimeLocal == null) {
			heidelTimeLocal = new HeidelTime();
			UimaContextImpl uimaContext = new UimaContextImpl(language, typeToProcess, CLISwitch.VERBOSITY2.getIsActive());
			String contextName = uimaContext.getQualifiedContextName();
			uimaContext.getConfigurationManager().setConfigParameterValue(contextName + "Time", Boolean.FALSE);
			uimaContext.getConfigurationManager().setConfigParameterValue(contextName + "Duration", Boolean.FALSE);
			uimaContext.getConfigurationManager().setConfigParameterValue(contextName + "Set", Boolean.FALSE);
			uimaContext.getConfigurationManager().setConfigParameterValue(contextName + "ConvertDurations", Boolean.FALSE);

			heidelTimeLocal.initialize(uimaContext);
			this.heidelTime.set(heidelTimeLocal);
		}
		return heidelTimeLocal;
	}

	private final ThreadLocal<JCas> jcasPool = new ThreadLocal<JCas>();

	private JCas getThreadLocalJCas() throws CASException, ResourceInitializationException {

		JCas jcasLocal = this.jcasPool.get();
		if (jcasLocal == null) {
			jcasLocal = jcasFactory.createJCas();
			this.jcasPool.set(jcasLocal);
		} else {
			jcasLocal.reset();
		}
		return jcasLocal;
	}

	private final ThreadLocal<PartOfSpeechTagger> posTaggerPool = new ThreadLocal<>();

	private PartOfSpeechTagger getThreadLocalPOSTagger() {

		PartOfSpeechTagger partOfSpeechTagger = posTaggerPool.get();
		if (partOfSpeechTagger == null) {
			Properties settings = new Properties();
			switch (language) {
			case ARABIC:
				partOfSpeechTagger = new StanfordPOSTaggerWrapper();
				settings.put(PartOfSpeechTagger.STANFORDPOSTAGGER_ANNOTATE_TOKENS, true);
				settings.put(PartOfSpeechTagger.STANFORDPOSTAGGER_ANNOTATE_SENTENCES, true);
				settings.put(PartOfSpeechTagger.STANFORDPOSTAGGER_ANNOTATE_POS, true);
				settings.put(PartOfSpeechTagger.STANFORDPOSTAGGER_MODEL_PATH, Config.get(Config.STANFORDPOSTAGGER_MODEL_PATH));
				settings.put(PartOfSpeechTagger.STANFORDPOSTAGGER_CONFIG_PATH, Config.get(Config.STANFORDPOSTAGGER_CONFIG_PATH));
				break;
			case VIETNAMESE:
				partOfSpeechTagger = new JVnTextProWrapper();
				settings.put(PartOfSpeechTagger.JVNTEXTPRO_ANNOTATE_TOKENS, true);
				settings.put(PartOfSpeechTagger.JVNTEXTPRO_ANNOTATE_SENTENCES, true);
				settings.put(PartOfSpeechTagger.JVNTEXTPRO_ANNOTATE_POS, true);
				settings.put(PartOfSpeechTagger.JVNTEXTPRO_WORD_MODEL_PATH, Config.get(Config.JVNTEXTPRO_WORD_MODEL_PATH));
				settings.put(PartOfSpeechTagger.JVNTEXTPRO_SENT_MODEL_PATH, Config.get(Config.JVNTEXTPRO_SENT_MODEL_PATH));
				settings.put(PartOfSpeechTagger.JVNTEXTPRO_POS_MODEL_PATH, Config.get(Config.JVNTEXTPRO_POS_MODEL_PATH));
				break;
			case CROATIAN:
				partOfSpeechTagger = new HunPosTaggerWrapper();
				settings.put(PartOfSpeechTagger.HUNPOS_LANGUAGE, language);
				settings.put(PartOfSpeechTagger.HUNPOS_ANNOTATE_TOKENS, true);
				settings.put(PartOfSpeechTagger.HUNPOS_ANNOTATE_POS, true);
				settings.put(PartOfSpeechTagger.HUNPOS_ANNOTATE_SENTENCES, true);
				settings.put(PartOfSpeechTagger.HUNPOS_MODEL_PATH, Config.get(Config.HUNPOS_MODEL_PATH));
				break;
			default:
				if (POSTagger.STANFORDPOSTAGGER.equals(posTagger)) {
					partOfSpeechTagger = new StanfordPOSTaggerWrapper();
					settings.put(PartOfSpeechTagger.STANFORDPOSTAGGER_ANNOTATE_TOKENS, true);
					settings.put(PartOfSpeechTagger.STANFORDPOSTAGGER_ANNOTATE_SENTENCES, true);
					settings.put(PartOfSpeechTagger.STANFORDPOSTAGGER_ANNOTATE_POS, true);
					settings.put(PartOfSpeechTagger.STANFORDPOSTAGGER_MODEL_PATH, Config.get(Config.STANFORDPOSTAGGER_MODEL_PATH));
					settings.put(PartOfSpeechTagger.STANFORDPOSTAGGER_CONFIG_PATH, Config.get(Config.STANFORDPOSTAGGER_CONFIG_PATH));
				} else if (POSTagger.TREETAGGER.equals(posTagger)) {
					partOfSpeechTagger = new TreeTaggerWrapper();
					settings.put(PartOfSpeechTagger.TREETAGGER_LANGUAGE, language);
					settings.put(PartOfSpeechTagger.TREETAGGER_ANNOTATE_TOKENS, true);
					settings.put(PartOfSpeechTagger.TREETAGGER_ANNOTATE_SENTENCES, true);
					settings.put(PartOfSpeechTagger.TREETAGGER_ANNOTATE_POS, true);
					settings.put(PartOfSpeechTagger.TREETAGGER_IMPROVE_GERMAN_SENTENCES, (language == Language.GERMAN));
					settings.put(PartOfSpeechTagger.TREETAGGER_CHINESE_TOKENIZER_PATH, Config.get(Config.CHINESE_TOKENIZER_PATH));
				} else if (POSTagger.HUNPOS.equals(posTagger)) {
					partOfSpeechTagger = new HunPosTaggerWrapper();
					settings.put(PartOfSpeechTagger.HUNPOS_LANGUAGE, language);
					settings.put(PartOfSpeechTagger.HUNPOS_ANNOTATE_TOKENS, true);
					settings.put(PartOfSpeechTagger.HUNPOS_ANNOTATE_POS, true);
					settings.put(PartOfSpeechTagger.HUNPOS_ANNOTATE_SENTENCES, true);
					settings.put(PartOfSpeechTagger.HUNPOS_MODEL_PATH, Config.get(Config.HUNPOS_MODEL_PATH));
				} else if (POSTagger.NO.equals(posTagger)) {
					partOfSpeechTagger = new AllLanguagesTokenizerWrapper();
				} else {

				}
			}
			partOfSpeechTagger.initialize(settings);

			posTaggerPool.set(partOfSpeechTagger);
		}

		return partOfSpeechTagger;

	}



	private final JCasFactory jcasFactory;

	public HeidelTimeWrapper(String configPath) throws ResourceInitializationException, InvalidXMLException, IOException {

		System.out.println(configPath);
		if (!Config.isInitialized()) {

			readConfigFile(configPath);
		}

		jcasFactory = getjcasFactory();

	}

	ThreadLocal<SimpleDateFormat> sdfPool = new ThreadLocal<>();

	SimpleDateFormat getThreadLocalSDF() {
		SimpleDateFormat sdf = sdfPool.get();
		if (sdf == null) {
			sdf = new SimpleDateFormat("yyyy");
			sdfPool.set(sdf);
		}
		return sdf;

	}

	ThreadLocal<ExecutorService> executorServicePool = new ThreadLocal<>();

	ExecutorService getExecutorService() {
		ExecutorService service = executorServicePool.get();
		if (service == null) {
			service = Executors.newSingleThreadExecutor();
			executorServicePool.set(service);
		}
		return service;
	}

	public List<Integer> getYears(String string) {

		if (string == null || string.trim().equals("")) {
			return null;
		}
		try {

			List<Integer> retSet = new ArrayList<>();

			SimpleDateFormat sdf = getThreadLocalSDF();

			JCas jcas = getThreadLocalJCas();
			jcas.setDocumentText(string);

			PartOfSpeechTagger posTagger = getThreadLocalPOSTagger();

			posTagger.process(jcas);

			HeidelTime ht = getThreadLocalHeidelTime();

			FSIterator sentIter = jcas.getAnnotationIndex(Sentence.type).iterator();
			/*
			 * check if the pipeline has annotated any sentences. if not,
			 * heideltime can't do any work, will return from process() with a
			 * warning message.
			 */
			if (!sentIter.hasNext()) {
				System.out.println("SKIPPED OVER FOLLOWING STRING: " + string);
			} else {

				ht.process(jcas);
				// runIntervalTagger(jcas,language);

				FSIterator<Annotation> iterTimex = jcas.getAnnotationIndex(Timex3.type).iterator();
				while (iterTimex.hasNext()) {
					String date = ((Timex3) iterTimex.next()).getTimexValue();

					try {
						Date d = sdf.parse(date);
						Calendar c = Calendar.getInstance();
						c.setTime(d);
						int year = c.get(Calendar.YEAR);

						retSet.add(year);
					} catch (ParseException e) {
						// System.out.println(date);
					}

				}
				// TimeMLResultFormatter formatter = new
				// TimeMLResultFormatter();
				// System.out.println(formatter.format(jcas));

			}

			if (retSet.isEmpty()) {
				return null;
			} else {
				return retSet;
			}

		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}

	}

	private JCasFactory getjcasFactory() throws InvalidXMLException, IOException {
		TypeSystemDescription[] descriptions = new TypeSystemDescription[] { UIMAFramework.getXMLParser()
				.parseTypeSystemDescription(new XMLInputSource(HeidelTimeStandalone.class.getClassLoader().getResource(Config.get(Config.TYPESYSTEMHOME)))) };
		return new JCasFactoryImpl(descriptions);
	}

	private void runIntervalTagger(JCas jcas, Language language) {
		Integer beforeAnnotations = jcas.getAnnotationIndex().size();

		// Prepare the options for IntervalTagger's execution
		Properties settings = new Properties();
		settings.put(IntervalTagger.PARAM_LANGUAGE, language.getResourceFolder());
		settings.put(IntervalTagger.PARAM_INTERVALS, true);
		settings.put(IntervalTagger.PARAM_INTERVAL_CANDIDATES, false);

		// Instantiate and process with IntervalTagger
		IntervalTaggerWrapper iTagger = new IntervalTaggerWrapper();
		iTagger.initialize(settings);
		iTagger.process(jcas);

		// debug output
		Integer afterAnnotations = jcas.getAnnotationIndex().size();
	}

	public static void readConfigFile(String configPath) {
		InputStream configStream = null;
		try {
			logger.log(Level.INFO, "trying to read in file " + configPath);
			configStream = new FileInputStream(configPath);

			Properties props = new Properties();
			props.load(configStream);

			Config.setProps(props);

			configStream.close();
		} catch (FileNotFoundException e) {
			logger.log(Level.WARNING, "couldn't open configuration file \"" + configPath + "\". quitting.");
			System.exit(-1);
		} catch (IOException e) {
			logger.log(Level.WARNING, "couldn't close config file handle");
			e.printStackTrace();
		}
	}
}
