package de.uni_mannheim.informatik.dws.expansion.wikidata.adapter;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Named;

import org.json.JSONObject;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseImpl;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedProperty;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataType;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataProperties;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataHierarchy;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataHierarchy.WikidataHierarchyFactory;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataLabels;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataLabels.WikidataLabelsFactory;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataJSONParser;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataJSONReader;

public class WikidataImporter {
	private final WikidataJSONReader reader;
	private final WikidataJSONParser parser;
	private final String dumpFile;
	private final ThreadSafeKryo threadSafeKryo;

	@Inject
	public WikidataImporter(WikidataJSONReader reader, WikidataJSONParser parser, @Named(WikidataProperties.WIKIDATA_DUMP) String dumpFile,
			ThreadSafeKryo threadSafeKryo) {
		this.reader = reader;
		this.parser = parser;
		this.dumpFile = dumpFile;
		this.threadSafeKryo = threadSafeKryo;

	}

	public Pair<KnowledgeBaseImpl, Map<InstanceId, InstanceId>> read(List<ImporterFactory> importerFactories)
			throws FileNotFoundException, IOException, InterruptedException {
		String cacheFile = dumpFile + ".hierarchycache";
		Path path = Paths.get(cacheFile);
		// Log.TRACE();
		WikidataHierarchy hierarchy = null;
		if (Files.exists(path)) {
			hierarchy = (WikidataHierarchy) threadSafeKryo.read(WikidataHierarchy.class, cacheFile);
		} else {
			final WikidataHierarchyFactory factory = new WikidataHierarchy.WikidataHierarchyFactory();
			reader.read(dumpFile, (parse) -> {
				JSONObject object = new JSONObject(parse);

				Pair<WikidataType, WikidataParsedObject> ite = parser.parse(object);

				if (ite.getFirst().equals(WikidataType.ITEM)) {
					factory.add((WikidataParsedItem) ite.getSecond());
				}
			});
			hierarchy = factory.create();
			factory.destroy();
			threadSafeKryo.writeWithoutClass(hierarchy, cacheFile);
		}

		String labelsFile = dumpFile + ".labelsCache";
		Path labelsPath = Paths.get(labelsFile);
		WikidataLabels labels = null;
		if (Files.exists(labelsPath)) {
			labels = (WikidataLabels) threadSafeKryo.read(WikidataLabels.class, labelsFile);
		} else {
			final WikidataLabelsFactory factory = new WikidataLabels.WikidataLabelsFactory();
			reader.read(dumpFile, (parse) -> {
				JSONObject object = new JSONObject(parse);

				Pair<WikidataType, WikidataParsedObject> ite = parser.parse(object);

				if (ite.getFirst().equals(WikidataType.ITEM)) {
					factory.add((WikidataParsedItem) ite.getSecond());
				}
			});
			labels = factory.create();
			factory.destroy();
			threadSafeKryo.writeWithoutClass(labels, labelsFile);
		}

		parser.setLabels(labels);
		ReadingState rs = new ReadingState(importerFactories, hierarchy, labels);
		reader.read(dumpFile, rs::parse);

		rs.addAll();

		System.out.println("WikidataImporter-CreatingKB");
		KnowledgeBaseImpl knowledgeBase = KnowledgeBaseImpl.create(rs.values, rs.setMappings, new HashMap<>(), new HashMap<>(), new HashMap<>());

		return new Pair<KnowledgeBaseImpl, Map<InstanceId, InstanceId>>(knowledgeBase, rs.dbpediaMappings);

	}

	private class ReadingState {
		private final ConcurrentHashMap<ValueId, QualifiedValueList> values = new ConcurrentHashMap<>();
		private final ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings = new ConcurrentHashMap<>();
		private final ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings = new ConcurrentHashMap<>();
		private final WikidataHierarchy hierarchy;
		private final WikidataLabels labels;

		List<ImporterParser> importers;

		public ReadingState(List<ImporterFactory> importerFactories, WikidataHierarchy hierarchy, WikidataLabels labels) {
			importers = importerFactories.stream().map(ImporterFactory::get).collect(Collectors.toList());
			this.hierarchy = hierarchy;
			this.labels = labels;
		}

		public void addAll() {
			for (

			ImporterParser importer : importers)

			{
				importer.add(values, setMappings, dbpediaMappings);

			}

		}

		private final void parse(String parse) {
			JSONObject object = new JSONObject(parse);

			Pair<WikidataType, WikidataParsedObject> ite = parser.parse(object);

			if (ite.getFirst().equals(WikidataType.ITEM)) {

				Optional<InstanceId> dbpedialink = parser.getEnglishDBPediaLink(object);
				Optional<InstanceId> dbpedialinkGerman = parser.getGermanDBPediaLink(object);

				if (dbpedialink.isPresent()) {
					dbpediaMappings.put(dbpedialink.get(), parser.getInstanceId(object));

				}

				if (dbpedialinkGerman.isPresent()) {
					dbpediaMappings.put(dbpedialinkGerman.get(), parser.getInstanceId(object));

				}

				for (ImporterParser importer : importers) {
					importer.run((WikidataParsedItem) ite.getSecond(), object, parser, dbpediaMappings, hierarchy, labels);
				}
			} else if (ite.getFirst().equals(WikidataType.PROPERTY)) {
				for (ImporterParser importer : importers) {
					importer.runProperty((WikidataParsedProperty) ite.getSecond(), object, parser, dbpediaMappings);
				}
			}

		}
	}

	public static interface ImporterParser {
		//
		// void run(WikidataParsedItem item, JSONObject object,
		// WikidataJSONParser parser, ConcurrentHashMap<ValueId,
		// QualifiedValueList> values,
		// ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings,
		// ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings);

		void add(ConcurrentHashMap<ValueId, QualifiedValueList> values, ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings,
				ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings);

		void runProperty(WikidataParsedProperty second, JSONObject object, WikidataJSONParser parser,
				ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings);

		void run(WikidataParsedItem item, JSONObject object, WikidataJSONParser parser, Map<InstanceId, InstanceId> dbpediaMappings,
				WikidataHierarchy hierarchy, WikidataLabels labels);

	}

	public static interface ImporterFactory {

		ImporterParser get();

	}

}
