package de.uni_mannheim.informatik.dws.expansion.wikidata.adapter;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;
import javax.inject.Named;

import org.json.JSONObject;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseImpl;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataType;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataProperties;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataJSONParser;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataJSONReader;

public class WikidataTransformer {
	private final WikidataJSONReader reader;
	private final WikidataJSONParser parser;
	private final String dumpFile;

	Set<InstanceId> classesToImport = new HashSet<>();
	Set<InstanceId> occupationsToImport = new HashSet<>();

	@Inject
	public WikidataTransformer(WikidataJSONReader reader, WikidataJSONParser parser, @Named(WikidataProperties.WIKIDATA_DUMP) String dumpFile) {
		this.reader = reader;
		this.parser = parser;
		this.dumpFile = dumpFile;

		classesToImport.add(WikidataClasses.aktiengesellschaft);
		classesToImport.add(WikidataClasses.country);
		classesToImport.add(WikidataClasses.publicCompany);
		classesToImport.add(WikidataClasses.publicLimitedCompany);
		classesToImport.add(WikidataClasses.aktiengesellschaft);
		classesToImport.add(WikidataClasses.city);
		classesToImport.add(WikidataClasses.bigCity);

		occupationsToImport.add(WikidataClasses.americalFootballPlayer);
		occupationsToImport.add(WikidataClasses.gridIronFootballPlayer);
		occupationsToImport.add(WikidataClasses.basketballPlayer);
		occupationsToImport.add(WikidataClasses.associationFootballer);
	}

	public Pair<KnowledgeBaseImpl, Map<InstanceId, InstanceId>> read() throws FileNotFoundException, IOException, InterruptedException {

		ReadingState rs = new ReadingState();
		reader.read(dumpFile, rs::parse);

		KnowledgeBaseImpl knowledgeBase = KnowledgeBaseImpl.create(rs.values, rs.setMappings, new HashMap<>(), new HashMap<>(), new HashMap<>());

		return new Pair<KnowledgeBaseImpl, Map<InstanceId, InstanceId>>(knowledgeBase, rs.dbpediaMappings);

	}

	private class ReadingState {
		private final ConcurrentHashMap<ValueId, QualifiedValueList> values = new ConcurrentHashMap<>();
		private final ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings = new ConcurrentHashMap<>();
		private final ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings = new ConcurrentHashMap<>();

		private final void parse(String parse) {
			JSONObject object = new JSONObject(parse);

			Pair<WikidataType, WikidataParsedObject> ite = parser.parse(object);

			if (ite.getFirst().equals(WikidataType.ITEM)) {

				Optional<InstanceId> dbpedialink = parser.getEnglishDBPediaLink(object);
				Optional<InstanceId> dbpedialinkGerman = parser.getGermanDBPediaLink(object);
				InstanceId instanceId = parser.getInstanceId(object);
				if (dbpedialink.isPresent()) {
					dbpediaMappings.put(dbpedialink.get(), parser.getInstanceId(object));

				}

				if (dbpedialinkGerman.isPresent()) {
					dbpediaMappings.put(dbpedialinkGerman.get(), parser.getInstanceId(object));

				}

				ClassId setId = getSet((WikidataParsedItem) ite.getSecond());
				if (setId != null) {

					if (dbpedialink.isPresent() || dbpedialinkGerman.isPresent()) {

						for (Entry<PropertyId, QualifiedValueList> e : ite.getSecond().getValuesMap().entrySet()) {
							QualifiedValueList existingValue = values.putIfAbsent(new ValueId(instanceId, e.getKey()), e.getValue());
							if (existingValue != null) {
								throw new IllegalStateException("WHY DO WE HAVE THE SAME VALUEID TWICE?");
							}
						}

						QualifiedValueList labelQvl = new QualifiedValueList(new StringObject(ite.getSecond().getLabel()));

						values.putIfAbsent(new ValueId(instanceId, Stats.label), labelQvl);

						CHMHelper.getFromCHM(setMappings, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(setId);
					}
				}
			}

		}

		private ClassId getSet(WikidataParsedItem item) {
			QualifiedValueList instanceOf = item.getValues(WikidataClasses.instanceOf);

			if (instanceOf.hasValue(WikidataClasses.human)) {
				QualifiedValueList occupation = item.getValues(WikidataClasses.occuptation);
				for (InstanceId occupationE : occupationsToImport) {
					if (occupation.hasValue(occupationE))
						return WikidataClasses.getSet(occupationE);
				}

				if (occupation.hasValue(WikidataClasses.politician) && item.getValues(Stats.positionHeld) != null) {
					return WikidataClasses.getSet(WikidataClasses.politician);
				}
			}

			for (InstanceId clazz : classesToImport) {
				if (instanceOf.hasValue(clazz))
					return WikidataClasses.getSet(clazz);
			}

			return null;
		}

	}

}
