package de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies;

import java.io.Serializable;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;

public class WikidataHierarchy implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private final ImmutableMap<InstanceId, ImmutableSet<InstanceId>> instanceOf;
	private final ImmutableMap<InstanceId, ImmutableSet<InstanceId>> subClassOf;
	private final ImmutableMap<InstanceId, ImmutableSet<InstanceId>> occupation;

	private WikidataHierarchy(ConcurrentHashMap<InstanceId, Set<InstanceId>> instanceOf, ConcurrentHashMap<InstanceId, Set<InstanceId>> subClassOf,
			ConcurrentHashMap<InstanceId, Set<InstanceId>> occupation) {
		this.instanceOf = ImmutableMap.copyOf(instanceOf.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			return ImmutableSet.copyOf(e.getValue());
		})));
		this.subClassOf = ImmutableMap.copyOf(subClassOf.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			return ImmutableSet.copyOf(e.getValue());
		})));
		this.occupation = ImmutableMap.copyOf(occupation.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			return ImmutableSet.copyOf(e.getValue());
		})));
	}

	public static class WikidataHierarchyFactory {
		private Set<InstanceId> isInstanceOf = Collections.newSetFromMap(new ConcurrentHashMap<>());
		private Set<InstanceId> isSubclassOf = Collections.newSetFromMap(new ConcurrentHashMap<>());
		private Set<InstanceId> isOccupation = Collections.newSetFromMap(new ConcurrentHashMap<>());
		private ConcurrentHashMap<InstanceId, Set<InstanceId>> instanceOf = new ConcurrentHashMap<>();
		private ConcurrentHashMap<InstanceId, Set<InstanceId>> subClassOf = new ConcurrentHashMap<>();
		private ConcurrentHashMap<InstanceId, Set<InstanceId>> occupation = new ConcurrentHashMap<>();
		private boolean destroyed = false;

		public WikidataHierarchyFactory() {

		}

		public void add(WikidataParsedItem item) {
			if (!destroyed) {
				InstanceId instanceId = item.getInstanceId();
				if (item.getLabel() == null) {
					return;
				} else {

					for (QualifiedValueHolder entry : item.getValues(WikidataClasses.instanceOf).getClaims()) {
						InstanceId target = (InstanceId) entry.getValue();
						isInstanceOf.add(target);
						CHMHelper.getFromCHM(instanceOf, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(target);
					}

					for (QualifiedValueHolder entry : item.getValues(WikidataClasses.subClass).getClaims()) {
						InstanceId target = (InstanceId) entry.getValue();
						isSubclassOf.add(target);
						CHMHelper.getFromCHM(subClassOf, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(target);
					}

					for (QualifiedValueHolder entry : item.getValues(WikidataClasses.occuptation).getClaims()) {
						InstanceId target = (InstanceId) entry.getValue();
						isOccupation.add(target);
						CHMHelper.getFromCHM(occupation, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(target);
					}
				}
			} else {
				throw new IllegalStateException("THIS FACTORY HAS ALREADY BEEN DESTROYED");
			}
		}

		public WikidataHierarchy create() {
			if (!destroyed) {
				final ConcurrentHashMap<InstanceId, Set<InstanceId>> finalInstanceOf = new ConcurrentHashMap<>();
				final ConcurrentHashMap<InstanceId, Set<InstanceId>> finalSubClassOf = new ConcurrentHashMap<>();
				final ConcurrentHashMap<InstanceId, Set<InstanceId>> finalOccupation = new ConcurrentHashMap<>();

				PersistentParallel4.runAndFinish(this.isInstanceOf, (e) -> {
					Set<InstanceId> set = new HashSet<>();
					fillSetRecursive(instanceOf, set, e);
					fillSetRecursive(subClassOf, set, e);
					if (set.size() > 0) {
						for (InstanceId instanceId : set) {
							CHMHelper.getFromCHM(finalInstanceOf, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e);
						}
					}

				} , "");

				PersistentParallel4.runAndFinish(this.isSubclassOf, (e) -> {
					Set<InstanceId> set = new HashSet<>();
					fillSetRecursive(subClassOf, set, e);
					fillSetRecursive(instanceOf, set, e);
					if (set.size() > 0) {
						for (InstanceId instanceId : set) {
							CHMHelper.getFromCHM(finalSubClassOf, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e);
						}
					}

				} , "");

				PersistentParallel4.runAndFinish(this.isOccupation, (e) -> {
					Set<InstanceId> set = new HashSet<>();
					fillSetRecursive(occupation, set, e);

					fillSetRecursive(subClassOf, set, e);
					fillSetRecursive(instanceOf, set, e);
					if (set.size() > 0) {
						for (InstanceId instanceId : set) {
							CHMHelper.getFromCHM(finalOccupation, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e);
						}
					}

				} , "");

				return new WikidataHierarchy(finalInstanceOf, finalSubClassOf, finalOccupation);
			} else {
				throw new IllegalStateException("THIS FACTORY HAS ALREADY BEEN DESTROYED");
			}
		}

		private void fillSetRecursive(Map<InstanceId, Set<InstanceId>> fullMap, Set<InstanceId> setToFill, InstanceId currentId) {

			if (fullMap.containsKey(currentId)) {
				for (InstanceId id : fullMap.get(currentId)) {
					if (!setToFill.contains(id)) {
						setToFill.add(id);
						fillSetRecursive(fullMap, setToFill, id);
					}
				}

			}

		}

		public void destroy() {
			this.instanceOf = null;
			this.subClassOf = null;
			this.occupation = null;
			this.isInstanceOf = null;
			this.isOccupation = null;
			this.isSubclassOf = null;
			this.destroyed = true;

		}
	}

	public boolean hasInstanceOf(WikidataParsedItem item, InstanceId target) {

		Set<InstanceId> possibilities = instanceOf.get(target);
		for (QualifiedValueHolder entry : item.getValues(WikidataClasses.instanceOf).getClaims()) {
			InstanceId instanceId = (InstanceId) entry.getValue();

			if (target.equals(instanceId) || (possibilities != null && possibilities.contains(instanceId))) {
				return true;
			}
		}

		return false;
	}

	public boolean hasSubclassOf(WikidataParsedItem item, InstanceId target) {

		Set<InstanceId> possibilities = subClassOf.get(target);
		for (QualifiedValueHolder entry : item.getValues(WikidataClasses.subClass).getClaims()) {
			InstanceId instanceId = (InstanceId) entry.getValue();

			if (target.equals(instanceId) || (possibilities != null && possibilities.contains(instanceId))) {
				return true;
			}
		}

		// for (QualifiedValueHolder entry :
		// item.getValues(WikidataClasses.subClass).getClaims()) {
		// InstanceId candidate = (InstanceId) entry.getValue();
		// if (subClassOf.containsKey(candidate)) {
		// if (subClassOf.get(candidate).contains(target)) {
		// return true;
		// }
		// }
		//
		// }

		return false;
	}

	public boolean hasOccupation(WikidataParsedItem item, InstanceId target) {

		Set<InstanceId> possibilities = occupation.get(target);
		for (QualifiedValueHolder entry : item.getValues(WikidataClasses.occuptation).getClaims()) {
			InstanceId instanceId = (InstanceId) entry.getValue();

			if (target.equals(instanceId) || (possibilities != null && possibilities.contains(instanceId))) {
				return true;
			}
		}

		// for (QualifiedValueHolder entry :
		// item.getValues(WikidataClasses.occuptation).getClaims()) {
		// InstanceId candidate = (InstanceId) entry.getValue();
		// if (occupation.containsKey(candidate)) {
		// if (occupation.get(candidate).contains(target)) {
		// return true;
		// }
		// }
		//
		// }

		return false;
	}

}
