package de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.importers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import org.json.JSONObject;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedProperty;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.Qualifier;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.WikidataImporter.ImporterFactory;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.WikidataImporter.ImporterParser;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataHierarchy;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataLabels;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataJSONParser;

public class FixedClassImporterFactory implements ImporterFactory {

	private final FixedClassGetter fixedClassGetter = new FixedClassGetter();
	private final Map<PropertyId, ValueType> propertyDataTypes = new HashMap<>();

	@Inject
	public FixedClassImporterFactory() {
	}

	@Override
	public FixedClassImporter get() {
		return new FixedClassImporter(fixedClassGetter);
	}

	public Map<PropertyId, ValueType> getPropertyDataTypes() {
		return Collections.unmodifiableMap(propertyDataTypes);
	}

	public class FixedClassImporter implements ImporterParser {
		ConcurrentHashMap<ValueId, QualifiedValueList> values = new ConcurrentHashMap<>();
		ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings = new ConcurrentHashMap<>();
		private final FixedClassGetter fixedClassGetter;

		public FixedClassImporter(FixedClassGetter fixedClassGetter) {
			this.fixedClassGetter = fixedClassGetter;
		}

		@Override
		public void run(WikidataParsedItem item, JSONObject object, WikidataJSONParser parser, Map<InstanceId, InstanceId> dbpediaMappings,
				WikidataHierarchy hierarchy, WikidataLabels labels) {

			InstanceId instanceId = parser.getInstanceId(object);

			Optional<InstanceId> dbpedialink = parser.getEnglishDBPediaLink(object);
			Optional<InstanceId> dbpedialinkGerman = parser.getGermanDBPediaLink(object);

			if (dbpedialink.isPresent() || dbpedialinkGerman.isPresent()) {

				ClassId setId = fixedClassGetter.getSet(item, hierarchy);
				if (setId != null) {

					for (Entry<PropertyId, QualifiedValueList> e : item.getValuesMap().entrySet()) {
						PropertyId propertyId = Stats.getExistingPropertyId(e.getKey());
						if (propertyId != null) {
							ValueId valueId = new ValueId(instanceId, propertyId);
							QualifiedValueList existingValue = values.putIfAbsent(valueId, e.getValue());
							if (existingValue != null) {
								throw new IllegalStateException("WHY DO WE HAVE THE SAME VALUEID TWICE?");
							}
						}
					}

					QualifiedValueList labelQvl = new QualifiedValueList(new StringObject(item.getLabel()));

					List<QualifiedValueHolder> translatedLabels = new ArrayList<>();
					for (Entry<String, String> entry : item.getLabels().entrySet()) {
						Qualifier q = new Qualifier(PropertiesExpansionData.LANG_QUALIFIER, new StringObject(entry.getKey()));
						List<Qualifier> qualifierList = new ArrayList<>();
						qualifierList.add(q);
						translatedLabels.add(new QualifiedAbstractObjectHolder(new StringObject(entry.getValue()), qualifierList));
					}
					QualifiedValueList translatedLabelsQvl = new QualifiedValueList(translatedLabels);

					List<QualifiedValueHolder> translatedAliases = new ArrayList<>();
					for (Entry<String, Set<String>> entry : item.getAliases().entrySet()) {
						Qualifier q = new Qualifier(PropertiesExpansionData.LANG_QUALIFIER, new StringObject(entry.getKey()));
						List<Qualifier> qualifierList = new ArrayList<>();
						qualifierList.add(q);
						for (String value : entry.getValue()) {

							translatedAliases.add(new QualifiedAbstractObjectHolder(new StringObject(value), qualifierList));
						}
					}
					QualifiedValueList translatedAliasesQvl = new QualifiedValueList(translatedAliases);

					values.putIfAbsent(new ValueId(instanceId, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY), translatedLabelsQvl);
					values.putIfAbsent(new ValueId(instanceId, PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY), translatedAliasesQvl);
					values.putIfAbsent(new ValueId(instanceId, PropertiesExpansionData.LABEL_PROPERTY), labelQvl);

					// if (setId.equals(WikidataClasses.wikidataCountry)) {
					// int sdf = 123;
					// }
					CHMHelper.getFromCHM(setMappings, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(setId);
				}
			}

		}

		@Override
		public void add(ConcurrentHashMap<ValueId, QualifiedValueList> values, ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings,
				ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings) {
			values.putAll(this.values);
			setMappings.putAll(this.setMappings);

		}

		@Override
		public void runProperty(WikidataParsedProperty property, JSONObject object, WikidataJSONParser parser,
				ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings) {
			PropertyId propertyId = Stats.getExistingPropertyId(property.getPropertyId());
			if (propertyId != null) {
				propertyDataTypes.put(propertyId, property.getValueType());
			}

		}
	}
}
