package de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.importers;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.json.JSONObject;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.consumers.FourConsumer;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedProperty;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.Qualifier;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.WikidataImporter.ImporterFactory;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.WikidataImporter.ImporterParser;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataHierarchy;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataLabels;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataJSONParser;

public class LeaderImporterFactory implements ImporterFactory {

	private enum LeaderType {
		HEAD_OF_STATE, HEAD_OF_GOVERNMENT;
	}

	public ImporterParser get() {
		return new LeaderImporter();
	}

	public static class LeaderImporter implements ImporterParser {
		ConcurrentLinkedQueue<Pair<InstanceId, QualifiedValueHolder>> person = new ConcurrentLinkedQueue<>();
		ConcurrentHashMap<InstanceId, InstanceId> headOfGovernmentsOfficeMap = new ConcurrentHashMap<>();
		ConcurrentHashMap<InstanceId, InstanceId> headOfStatesOfficeMap = new ConcurrentHashMap<>();

		ConcurrentLinkedQueue<Pair<InstanceId, WikidataParsedItem>> positionInstances = new ConcurrentLinkedQueue<>();

		@Override
		public void run(WikidataParsedItem item, JSONObject object, WikidataJSONParser parser, Map<InstanceId, InstanceId> dbpediaMappings,
				WikidataHierarchy wikidataHierarchy, WikidataLabels labels) {

			InstanceId instanceId = parser.getInstanceId(object);

			// QualifiedValueList instanceOf =
			// item.getValues(WikidataClasses.instanceOf);
			// QualifiedValueList subClassOf =
			// item.getValues(WikidataClasses.subClass);

			if (wikidataHierarchy.hasInstanceOf(item, WikidataClasses.human)) {
				QualifiedValueList positionHeld = item.getValues(Stats.positionHeld);
				if (positionHeld != null) {
					for (QualifiedValueHolder claim : positionHeld.getClaims()) {
						person.add(new Pair<>(instanceId, claim));
					}
				}

			}

			else if (wikidataHierarchy.hasInstanceOf(item, WikidataClasses.country) || wikidataHierarchy.hasInstanceOf(item, WikidataClasses.city)
					|| wikidataHierarchy.hasInstanceOf(item, WikidataClasses.bigCity)) {

				QualifiedValueList headOfGovernmentOffice = item.getValues(Stats.headOfGovernmentOffice);

				if (headOfGovernmentOffice != null && headOfGovernmentOffice.size() > 0) {
					headOfGovernmentsOfficeMap.put((InstanceId) headOfGovernmentOffice.getClaims().get(0).getValue(), instanceId);
				}

				QualifiedValueList headOfStateOffice = item.getValues(Stats.headOfStateOffice);

				if (headOfStateOffice != null && headOfStateOffice.size() > 0) {
					headOfStatesOfficeMap.put((InstanceId) headOfStateOffice.getClaims().get(0).getValue(), instanceId);
				}

			}

			else if (wikidataHierarchy.hasInstanceOf(item, WikidataClasses.position) || wikidataHierarchy.hasInstanceOf(item, WikidataClasses.publicOffice)
					|| wikidataHierarchy.hasSubclassOf(item, WikidataClasses.headOfStateSubclass)
					|| wikidataHierarchy.hasSubclassOf(item, WikidataClasses.headOfGovernmentSubclass)

			) {
				positionInstances.add(new Pair<>(instanceId, item));

				if (wikidataHierarchy.hasSubclassOf(item, WikidataClasses.headOfGovernmentSubclass)) {
					if (item.hasValues(Stats.appliesToTerritorialJurisdiction)) {
						headOfGovernmentsOfficeMap.put(instanceId, (InstanceId) item.getValues(Stats.appliesToTerritorialJurisdiction).getClaim(0).getValue());
					}
					if (item.hasValues(Stats.country)) {
						headOfGovernmentsOfficeMap.put(instanceId, (InstanceId) item.getValues(Stats.country).getClaim(0).getValue());

					}
				}

				if (wikidataHierarchy.hasSubclassOf(item, WikidataClasses.headOfStateSubclass)) {
					if (item.hasValues(Stats.appliesToTerritorialJurisdiction)) {
						headOfStatesOfficeMap.put(instanceId, (InstanceId) item.getValues(Stats.appliesToTerritorialJurisdiction).getClaim(0).getValue());
					}
					if (item.hasValues(Stats.country)) {
						headOfStatesOfficeMap.put(instanceId, (InstanceId) item.getValues(Stats.country).getClaim(0).getValue());

					}
				}
			}

		}

		@Override
		public void add(ConcurrentHashMap<ValueId, QualifiedValueList> values, ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings,
				ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings) {

			FourConsumer<LeaderType, InstanceId, InstanceId, List<Qualifier>> addValue = (leaderType, mainInstanceId, targetId, qualifiers) -> {
				if (!setMappings.containsKey(mainInstanceId)) {
					return;
				}
				ValueId valueId = null;
				if (leaderType == LeaderType.HEAD_OF_STATE) {
					valueId = new ValueId(mainInstanceId, Stats.headOfGovernemnt);
				} else {
					valueId = new ValueId(mainInstanceId, Stats.headOfState);
				}
				QualifiedValueList valueList = values.get(valueId);
				if (valueList == null) {
					List<QualifiedValueHolder> claims = new ArrayList<>();

					claims.add(new QualifiedAbstractObjectHolder(targetId, qualifiers));
					valueList = new QualifiedValueList(claims);
					values.put(valueId, valueList);
				} else {
					List<QualifiedValueHolder> claims = new ArrayList<>(valueList.getClaims());

					if (valueList.hasValue(targetId)) {
						boolean done = false;

						for (Iterator<QualifiedValueHolder> it = claims.iterator(); !done && it.hasNext();) {
							QualifiedValueHolder claim = it.next();
							if (claim.getValue().equals(targetId)) {
								boolean replace = false;
								boolean oneMustBeMissing = false;
								boolean oneMissing = false;
								boolean noQualifiers = true;
								boolean noQualifierMatch = true;

								for (Qualifier qualifier : qualifiers) {
									Qualifier ao = claim.getQualifierValue(qualifier.getQualifier());
									if (ao != null && ao.getQualifierValue() != null) {
										noQualifiers = false;
										TimeObject kbTO = (TimeObject) ao.getQualifierValue();
										TimeObject newTO = (TimeObject) qualifier.getQualifierValue();
										if (TimeObject.compare(kbTO, newTO) == 0) {
											noQualifierMatch = false;
											if (kbTO.getPrecision().getPosition() > newTO.getPrecision().getPosition()) {
												replace = true;
											} else {
												oneMustBeMissing = true;
											}
										}
									} else {
										oneMissing = true;
									}

								}

								if (replace) {
									it.remove();
									claims.add(new QualifiedAbstractObjectHolder(targetId, qualifiers));
									done = true;
									break;
								} else if (oneMustBeMissing) {
									if (oneMissing) {
										it.remove();
										claims.add(new QualifiedAbstractObjectHolder(targetId, qualifiers));
										done = true;
									} else if (!noQualifierMatch) {
										done = true;
									}
								} else if (noQualifiers) {
									it.remove();
									claims.add(new QualifiedAbstractObjectHolder(targetId, qualifiers));
									done = true;
								} else if (!noQualifiers && noQualifierMatch) {
									// ADD WHEN ITERATED THROUGH ALL
								}
							}
						}

						if (!done) {
							claims.add(new QualifiedAbstractObjectHolder(targetId, qualifiers));
						}
					} else {

						claims.add(new QualifiedAbstractObjectHolder(targetId, qualifiers));
					}

					valueList = new QualifiedValueList(claims);
					values.put(valueId, valueList);
				}
			};

			PersistentParallel4.runAndFinish(this.person, (e) -> {

				LeaderType leaderType = null;

				InstanceId mainInstanceId = null;

				Qualifier startTimeQualifier = e.getSecond().getQualifierValue(Stats.startTimeQualifier);
				// Qualifier endTimeQualifier =
				// e.getSecond().getQualifierValue(Stats.endTimeQualifier);
				//
				// if(startTimeQualifier == null ||
				// startTimeQualifier.getQualifierValue() == null) {
				//
				// }

				if (startTimeQualifier != null && startTimeQualifier.getQualifierValue() != null) {
					Qualifier endTimeQualifier = e.getSecond().getQualifierValue(Stats.endTimeQualifier);

					List<Qualifier> qualifiers = new ArrayList<>();
					qualifiers.add(startTimeQualifier);
					if (endTimeQualifier != null && endTimeQualifier.getQualifierValue() != null) {
						qualifiers.add(endTimeQualifier);
					}

					Qualifier headOfGovernmentQualifier = e.getSecond().getQualifierValue(Stats.headOfGovernemnt);
					if (headOfGovernmentQualifier != null) {
						mainInstanceId = (InstanceId) e.getSecond().getQualifierValue(Stats.headOfGovernemnt).getQualifierValue();
						leaderType = LeaderType.HEAD_OF_GOVERNMENT;
					}

					Qualifier ofQualifier = e.getSecond().getQualifierValue(Stats.of);
					if (ofQualifier != null
							&& (e.getSecond().getValue().equals(WikidataClasses.mayor) && e.getSecond().getValue().equals(WikidataClasses.lordMayor))) {
						mainInstanceId = (InstanceId) e.getSecond().getQualifierValue(Stats.of).getQualifierValue();
						leaderType = LeaderType.HEAD_OF_GOVERNMENT;
					}

					if (headOfGovernmentsOfficeMap.containsKey(e.getSecond().getValue())) {
						mainInstanceId = headOfGovernmentsOfficeMap.get(e.getSecond().getValue());
						leaderType = LeaderType.HEAD_OF_GOVERNMENT;
					}

					if (headOfStatesOfficeMap.containsKey(e.getSecond().getValue())) {
						mainInstanceId = headOfStatesOfficeMap.get(e.getSecond().getValue());
						leaderType = LeaderType.HEAD_OF_STATE;
					}

					if (mainInstanceId != null & leaderType != null) {
						addValue.accept(leaderType, mainInstanceId, e.getFirst(), qualifiers);
					} else {
						int df = 123;
					}

				}
			} , "Person");

			PersistentParallel4.runAndFinish(this.positionInstances, (e) -> {
				LeaderType leaderType = null;
				if (headOfStatesOfficeMap.containsKey(e.getFirst())) {
					leaderType = LeaderType.HEAD_OF_STATE;
				} else if (headOfGovernmentsOfficeMap.containsKey(e.getFirst())) {
					leaderType = LeaderType.HEAD_OF_GOVERNMENT;
				}

				if (leaderType != null) {
					QualifiedValueList positionValues = e.getSecond().getValues(Stats.appointedBy);
					InstanceId mainInstanceId = null;
					QualifiedValueList territory = e.getSecond().getValues(Stats.appliesToTerritorialJurisdiction);
					if (territory != null && territory.size() > 0) {
						mainInstanceId = (InstanceId) territory.getClaims().get(0).getValue();
					}
					if (mainInstanceId == null) {
						QualifiedValueList countryValue = e.getSecond().getValues(Stats.country);
						if (countryValue != null && countryValue.size() > 0) {
							mainInstanceId = (InstanceId) countryValue.getClaims().get(0).getValue();
						}

					}

					if (mainInstanceId != null) {
						for (QualifiedValueHolder claim : positionValues.getClaims()) {
							if (claim.getQualifiers() != null) {
								addValue.accept(leaderType, mainInstanceId, (InstanceId) claim.getValue(), claim.getQualifiers());
							}
						}
					}

				} else {
					int sdf = 123;
				}

			} , "PositionInstances");

		}

		@Override
		public void runProperty(WikidataParsedProperty second, JSONObject object, WikidataJSONParser parser,
				ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings) {
			// TODO Auto-generated method stub

		}
	}

	// private static class LeaderObject {
	//
	// private static final InstanceId ;
	// private static final InstanceId position;
	// }

}
