package de.uni_mannheim.informatik.dws.expansion.wikidata.reader;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import javax.inject.Inject;

import org.json.JSONArray;
import org.json.JSONObject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedProperty;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataType;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.Qualifier;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataLabels;
import de.uni_mannheim.informatik.dws.expansion.wikidata.parser.WikidataJSONParserCoordinate;
import de.uni_mannheim.informatik.dws.expansion.wikidata.parser.WikidataJSONParserEntity;
import de.uni_mannheim.informatik.dws.expansion.wikidata.parser.WikidataJSONParserExternalId;
import de.uni_mannheim.informatik.dws.expansion.wikidata.parser.WikidataJSONParserQuantity;
import de.uni_mannheim.informatik.dws.expansion.wikidata.parser.WikidataJSONParserTime;
import de.uni_mannheim.informatik.dws.expansion.wikidata.parser.WikidataJSONParserUrl;

public class WikidataJSONParser {

	// ThreadLocal<JSONParser> parsers = new ThreadLocal<>();

	// private JSONParser getParser() {
	// JSONParser parser = parsers.get();
	// if (parsers.get() == null) {
	// parser = new JSONParser();
	// parsers.set(parser);
	// }
	// return parser;
	// }

	private final WikidataJSONParserTime timeParser;
	private final WikidataJSONParserCoordinate coordinateParser;
	private final WikidataJSONParserEntity entityParser;
	private final WikidataJSONParserExternalId externalIdParser;
	private final WikidataJSONParserUrl urlParser;
	private final WikidataJSONParserQuantity quantityParser;
	// private WikidataLabels labels;

	@Inject
	public WikidataJSONParser(WikidataJSONParserTime timeParser, WikidataJSONParserCoordinate coordinateParser, WikidataJSONParserEntity entityParser,
			WikidataJSONParserExternalId externalIdParser, WikidataJSONParserUrl urlParser, WikidataJSONParserQuantity quantityParser) {
		this.timeParser = timeParser;
		this.coordinateParser = coordinateParser;
		this.entityParser = entityParser;
		this.externalIdParser = externalIdParser;
		this.urlParser = urlParser;
		this.quantityParser = quantityParser;
	}

	public void setLabels(WikidataLabels labels) {
		// this.labels = labels;
		this.entityParser.setLabels(labels);
	}

	public Pair<WikidataType, WikidataParsedObject> parse(JSONObject object) {

		String type = object.getString("type");
		switch (type) {
		case "item":
			return new Pair<WikidataType, WikidataParsedObject>(WikidataType.ITEM, parseItem(object));

		case "property":
			return new Pair<WikidataType, WikidataParsedObject>(WikidataType.PROPERTY, parseProperty(object));
		}
		throw new IllegalStateException("THIS ITEM IS UNKOWN: " + type);

	}

	public InstanceId getInstanceId(JSONObject object) {
		String id = getId(object);
		String label = getLabelOr(object, id);
		return new InstanceId(id, label);
	}

	public PropertyId getPropertyId(JSONObject object) {
		String id = getId(object);
		return new PropertyId(id);
	}

	private WikidataParsedItem parseItem(JSONObject object) {

		Map<PropertyId, QualifiedValueList> claims = parseClaims(object);

		InstanceId instanceId = getInstanceId(object);

		Map<String, String> labels = getLabels(object);
		Map<String, Set<String>> aliases = getAliases(object);

		String label = getLabelOr(object, instanceId.getInstanceId());

		return new WikidataParsedItem(claims, label, labels, aliases, instanceId);

	}

	private WikidataParsedProperty parseProperty(JSONObject object) {

		Map<PropertyId, QualifiedValueList> claims = parseClaims(object);

		PropertyId propertyId = getPropertyId(object);

		Map<String, String> labels = getLabels(object);
		Map<String, Set<String>> aliases = getAliases(object);

		String label = getLabelOr(object, propertyId.toString());

		ValueType dataType = getDataType(object);

		return new WikidataParsedProperty(claims, label, labels, aliases, propertyId, dataType);

	}

	private Map<String, String> getLabels(JSONObject object) {
		Map<String, String> labelsMap = new HashMap<>();
		JSONObject labels = object.getJSONObject("labels");

		for (Iterator<String> keys = labels.keys(); keys.hasNext();) {
			String key = keys.next();
			String label = labels.getJSONObject(key).getString("value");
			labelsMap.put(key, label);
		}

		return labelsMap;
	}

	private Map<String, Set<String>> getAliases(JSONObject object) {
		Map<String, Set<String>> aliasesMap = new HashMap<>();
		JSONObject labels = object.getJSONObject("aliases");

		for (Iterator<String> keys = labels.keys(); keys.hasNext();) {
			String key = keys.next();
			JSONArray aliasArray = labels.getJSONArray(key);

			if (aliasArray.length() > 0) {
				Set<String> aliasSet = new HashSet<>();
				for (int i = 0; i < aliasArray.length(); i++) {
					String alias = aliasArray.getJSONObject(i).getString("value");
					aliasSet.add(alias);
				}

				aliasesMap.put(key, aliasSet);
			}
		}

		return aliasesMap;
	}

	private ValueType getDataType(JSONObject object) {

		String dataType = object.getString("datatype");
		ValueType valueType = getDataType(dataType);

		return valueType;
	}

	public String getId(JSONObject object) {
		return object.getString("id");

	}

	public String getLabelOr(JSONObject object, String alternative) {
		String label = null;
		JSONObject labels = object.getJSONObject("labels");
		if (object.getJSONObject("labels").has("en")) {
			label = labels.getJSONObject("en").getString("value");
		}
		if (label == null) {
			label = alternative;
		}
		return label;
	}

	public Optional<InstanceId> getEnglishDBPediaLink(JSONObject object) {
		if (object.has("sitelinks")) {
			JSONObject sitelinks = object.getJSONObject("sitelinks");
			if (sitelinks.has("enwiki")) {
				JSONObject enWiki = sitelinks.getJSONObject("enwiki");
				if (enWiki.has("title")) {
					InstanceId id = InstanceId.simple(enWiki.getString("title"));
					return Optional.of(id);
				}

			}
		}
		return Optional.empty();
	}

	public Optional<InstanceId> getGermanDBPediaLink(JSONObject object) {
		if (object.has("sitelinks")) {
			JSONObject sitelinks = object.getJSONObject("sitelinks");
			if (sitelinks.has("dewiki")) {
				JSONObject enWiki = sitelinks.getJSONObject("dewiki");
				if (enWiki.has("title")) {
					InstanceId id = InstanceId.simple(enWiki.getString("title"));
					return Optional.of(id);
				}

			}
		}
		return Optional.empty();
	}

	private Map<PropertyId, QualifiedValueList> parseClaims(JSONObject object) {

		JSONObject claims = object.getJSONObject("claims");
		String[] properties = JSONObject.getNames(claims);

		Map<PropertyId, QualifiedValueList> propertyMap = new HashMap<>();

		if (properties != null) {
			for (String property : properties) {
				if (property.equals("P298")) {
					int sdf = 123;
				}
				JSONArray claimArray = claims.getJSONArray(property);
				List<QualifiedValueHolder> claimsList = new ArrayList<>();
				for (int i = 0; i < claimArray.length(); i++) {
					JSONObject claim = claimArray.getJSONObject(i);
					if (claim.has("mainsnak")) {
						AbstractObject value = getValueFromSnak(claim.getJSONObject("mainsnak"));
						if (value != null) {

							if (claim.has("qualifiers")) {
								List<Qualifier> qualifiers = getQualifiers(claim.getJSONObject("qualifiers"));
								claimsList.add(new QualifiedAbstractObjectHolder(value, qualifiers));
							} else {
								claimsList.add(new QualifiedAbstractObjectHolder(value));
							}
						}

					} else {
						System.out.println("NO MAIN SNAK???");
					}
				}

				if (claimsList.size() > 0) {
					QualifiedValueList valueList = new QualifiedValueList(claimsList);
					PropertyId propertyId = new PropertyId(property);
					propertyMap.put(propertyId, valueList);
				}

			}
		}
		return propertyMap;
	}

	private List<Qualifier> getQualifiers(JSONObject jsonObject) {

		List<Qualifier> qualifierList = new ArrayList<>();
		String[] qualifierNames = JSONObject.getNames(jsonObject);

		for (String qualifier : qualifierNames) {

			List<AbstractObject> list = getSimpleValueObjects(jsonObject.getJSONArray(qualifier));

			for (AbstractObject ao : list) {
				qualifierList.add(new Qualifier(new PropertyId(qualifier), ao));
			}
		}

		return qualifierList;
	}

	private List<AbstractObject> getSimpleValueObjects(JSONArray values) {
		List<AbstractObject> valueList = new ArrayList<>();

		for (int i = 0; i < values.length(); i++) {
			valueList.add(getValueFromSnak(values.getJSONObject(i)));
		}

		return valueList;
	}

	private AbstractObject getValueFromSnak(JSONObject snak) {
		if (snak.getString("snaktype").equals("value")) {
			if (!snak.has("datatype")) {
				System.out.println("Missing Datatype value");
				return null;
			}
			String dataType = snak.getString("datatype");
			switch (dataType) {
			case "string":
				return new StringObject(snak.getJSONObject("datavalue").getString("value"));
			case "wikibase-entityid":
				return entityParser.parse(snak.getJSONObject("datavalue"));
			case "wikibase-item":
				return entityParser.parse(snak.getJSONObject("datavalue"));
			case "wikibase-property":
				return entityParser.parse(snak.getJSONObject("datavalue"));
			case "external-id":
				return externalIdParser.parse(snak.getJSONObject("datavalue"));
			case "globe-coordinate":
				return coordinateParser.parse(snak.getJSONObject("datavalue"));
			case "time":
				return timeParser.parse(snak.getJSONObject("datavalue"));
			case "url":
				return urlParser.parse(snak.getJSONObject("datavalue"));
			case "quantity":
				return quantityParser.parse(snak.getJSONObject("datavalue"));
			case "commonsMedia":
				return null;
			case "monolingualtext":
				return null;
			case "math":
				return null;
			default:
				System.out.println("NOT DATATYPE?: " + dataType);
				return null;
			}
		} else {
			// TODO somevalue and novalue
			return null;
		}
	}

	private ValueType getDataType(String dataTypeName) {

		switch (dataTypeName) {
		case "string":
			return ValueType.STRING;
		case "wikibase-entityid":
			return ValueType.INSTANCE_REFERENCE;
		case "wikibase-item":
			return ValueType.INSTANCE_REFERENCE;
		case "wikibase-property":
			return ValueType.PROPERTY_REFERENCE;
		case "external-id":
			return ValueType.STRING;
		case "globe-coordinate":
			return ValueType.COORDINATE;
		case "time":
			return ValueType.TIME;
		case "url":
			return ValueType.URL;
		case "quantity":
			return ValueType.QUANTITY;
		case "commonsMedia":
			return null;
		case "monolingualtext":
			return null;
		case "math":
			return null;
		default:
			System.out.println("NOT DATATYPE?: " + dataTypeName);
			return null;
		}

	}

}
