package de.mannheim.uni.index.dbpedia;

import java.io.IOException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;

import org.apache.lucene.document.Document;
import org.apache.lucene.index.Term;
import org.apache.lucene.queryparser.classic.ParseException;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.queryparser.classic.QueryParserBase;
import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.BooleanQuery.Builder;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.search.TopDocs;
import org.apache.lucene.search.BooleanClause.Occur;

import com.github.benmanes.caffeine.cache.Cache;
import com.github.benmanes.caffeine.cache.Caffeine;
 
import com.wcohen.ss.Levenstein;

import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.index.IIndex;
import de.mannheim.uni.index.management.IndexManagerBase;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;

public class DBPediaInstanceIndex extends IndexManagerBase {

    private boolean verbose = false;
    public boolean isVerbose() {
        return verbose;
    }
    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }
    
	public DBPediaInstanceIndex(IIndex index, String defaultField) {
		super(index, defaultField);
	}

	private boolean removeBrackets = false;
	public boolean isRemoveBrackets() {
        return removeBrackets;
    }
	public void setRemoveBrackets(boolean removeBrackets) {
        this.removeBrackets = removeBrackets;
    }
	
	   public List<DBpediaIndexEntry> searchMany(Collection labels) {
	        long start, setup=0, search=0, result=0;
	        start = System.currentTimeMillis();
	        List<DBpediaIndexEntry> results = new LinkedList<DBpediaIndexEntry>();

	        IndexSearcher indexSearcher = getIndex().getIndexSearcher();

	        QueryParser queryParser = getQueryParserFromCache();
	        
	        try {           
	            BooleanQuery q = null;
	            
	            Builder qBuilder = new BooleanQuery.Builder();
	            
//	            q = new BooleanQuery();
	            
	            for(Object o : labels) {
	                Query q0 = null;
	                String value = QueryParserBase.escape((String)o);
    	            if(!isSearchExactMatches()) {
    	                
    	                //value = StringNormaliser.normalise(value, isRemoveBrackets());
    	                
    	                //String[] tokens = value.split("\\s+");
    	                
    	                List<String> tokens = StringNormalizer.tokenise(value, isRemoveBrackets());
    	                
    	                StringBuilder sb = new StringBuilder();
    	                
    	                for(String token : tokens) {
    	                    sb.append(token);
    	                    
    	                    if(getMaxEditDistance()>0) {
    	                        sb.append("~");
    	                        sb.append(getMaxEditDistance());
    	                    }
    	                    sb.append(" ");
    	                }
    	                
    	                value = sb.toString();
    	                
    	                if(value.trim().length()>0) {
    	                    q0 = queryParser.parse(value);
    	                }
    	            } else {
    	                if(value.trim().length()>0) {
    	                    q0 = new TermQuery(new Term(getDefaultField(), value));
    	                }
    	            }
    	            
    	            if(q0!=null) {
    	                ( qBuilder).add(q0, Occur.SHOULD);
    	            }
	            }
	            q = qBuilder.build();
	            
	            if(q!=null) {
	                if(getFilterValues()!=null && getFilterValues().size()>0) {
//	                    BooleanQuery filter = new BooleanQuery();
	                    Builder filterBuilder =  new BooleanQuery.Builder();
	                    
	                    for(String s : getFilterValues()) {
	                    	filterBuilder.add(new TermQuery(new Term(getFilterField(), s)), Occur.SHOULD);
	                    }
	                    
	                    BooleanQuery all = new BooleanQuery.Builder().add(q, Occur.MUST).add(filterBuilder.build(), Occur.MUST).build();
	                    
	                    q = all;
	                }
	                
	                if(isVerbose()) {
	                    System.out.println("Query: \n" + labels.toString() + "\n" + q.toString());
	                }
	                
	                setup = System.currentTimeMillis() - start;
	                start = System.currentTimeMillis();
	                
	                int numResults = getNumRetrievedDocsFromIndex();
	                ScoreDoc[] hits = indexSearcher.search(q, numResults).scoreDocs;
	                
	                search = System.currentTimeMillis() - start;
	                start = System.currentTimeMillis();
	                
	                if(hits != null)
	                {
	                    if(isVerbose()) {
	                        System.out.println(" found " + hits.length + " hits");
	                    }
	                    for (int i = 0; i < hits.length; i++) {
	                        
	                        Document doc = indexSearcher.doc(hits[i].doc);
	        
	                        DBpediaIndexEntry e = DBpediaIndexEntry.fromDocument(doc);

	                        if(isVerbose()) {
	                            System.out.println(e.getClass_label() + ": " + e.getLabel());
	                        }
	                        
	                        results.add(e);
	                    }
	                }
	            } else {
	                if(isVerbose()) {
	                    System.out.println(String.format("Empty query for '%s'", labels));
	                }
	            }
	            
	            result = System.currentTimeMillis() - start;

	        } catch (IOException e) {
	            e.printStackTrace();
	        } catch (ParseException e1) {
	            System.err.println(String.format("Parse exception for label '%s'", labels));
	            e1.printStackTrace();
	        }
	        if(isVerbose()) {
	            System.out.println(" returning " + results.size() + " documents");
	            System.out.println(String.format("setup: %d\tsearch: %d\tload: %d", setup, search, result));
	        }
	        return results;
	    }
	   
	   
//	   private static final Cache<String, List<DBpediaIndexEntry>> cache = Caffeine.newBuilder().maximumSize(2000000)
//	 	       .expireAfterWrite(10, TimeUnit.MINUTES).build();
	   
	   private static final ConcurrentHashMap<String, List<DBpediaIndexEntry>> chmCache = new ConcurrentHashMap<>();
//	 	       .build(
//	 	           new CacheLoader<Pair<String, String>, Double>() {
//	 	             public Double load(Pair<String, String> key)   {
//	 	            	Levenstein l = getLevenstein();
//	 	               double score = l.score(key.getFirst(), key.getSecond());
//	 	               score = score / Math.max(key.getFirst().length(), key.getSecond().length());
//	 	               if (score < 0) {
//	 	                   score = score * -1;
//	 	               }
//	 	               score = 1 - score;
//	 	               return score;
//	 	             }
//	 	           });
	
	public List<DBpediaIndexEntry> search(String label) {
		 
	    long start, setup=0, search=0, result=0;
	    start = System.currentTimeMillis();
		List<DBpediaIndexEntry> results = new LinkedList<DBpediaIndexEntry>();

		IndexSearcher indexSearcher = getIndex().getIndexSearcher();

		QueryParser queryParser = getQueryParserFromCache();
		
		try {			
			String value = QueryParserBase.escape(label); 
			Query q = null;
			
			if(!isSearchExactMatches()) {
    			
			    //value = StringNormaliser.normalise(value, isRemoveBrackets());
			    
			    //String[] tokens = value.split("\\s+");
			    
			    //List<String> tokens = StringNormaliser.tokenise(value, isRemoveBrackets());
			    List<String> tokens = StringNormalizer.tokenise(value, false, false);
			    
			    StringBuilder sb = new StringBuilder();
			    
			    boolean notFirst = false;
                for(String token : tokens) {
                	if(notFirst) 
                		sb.append(" ");
                	 
                    sb.append(token);
                    
                    if(getMaxEditDistance()>0) {
                    	if(token.length() < 3) {
                    		//sb.append("~");
		                    //sb.append(maxEditDistance);
                    	}
                    	else if(token.length() < 5){
                    		sb.append("~");
		                    sb.append(1);
                    	}
                    	else  {
                    		 sb.append("~");
		                     sb.append(getMaxEditDistance());
                    	}
                       
                    } 
                   
                    
                    notFirst = true;
                }
                
                
 
			    value = sb.toString();
			    
			    
    			
    			if(value.trim().length()>0) {
    				
//    				List<DBpediaIndexEntry> ret = cache.getIfPresent(value);
//    				if(ret != null) {
//    					return ret;
//    				}
    			    q = queryParser.parse(value);
    			}
			} else {
			    if(value.trim().length()>0) {
			        q = new TermQuery(new Term(getDefaultField(), value));
			    }
			}
			
			if(q!=null) {
    			if(getFilterValues()!=null && getFilterValues().size()>0) {
    			    Builder filterBuilder =  new BooleanQuery.Builder();
    			    
    			    for(String s : getFilterValues()) {
    			    	filterBuilder.add(new TermQuery(new Term(getFilterField(), s)), Occur.SHOULD);
    			    }
    			    
    			    BooleanQuery all = new BooleanQuery.Builder().add(q, Occur.MUST).add(filterBuilder.build(), Occur.MUST).build();
    		 
    			    
    			    q = all;
    			}
    			
    			if(isVerbose()) {
    			    System.out.println("Query: \n" + value + "\n" + q.toString());
    			}
    			
    			setup = System.currentTimeMillis() - start;
    			start = System.currentTimeMillis();
    			
    			int numResults = getNumRetrievedDocsFromIndex();
    			TopDocs topDocs = indexSearcher.search(q, numResults);
    			float maxScore = topDocs.getMaxScore();
    			
    			ScoreDoc[] hits = topDocs.scoreDocs;
    			
    			search = System.currentTimeMillis() - start;
    			start = System.currentTimeMillis();
    			
    			if(hits != null)
    			{
    			    if(isVerbose()) {
    			        System.out.println(" found " + hits.length + " hits");
    			    }
    				for (int i = 0; i < hits.length; i++) {
    					
    					Document doc = getIndex().getDocument(hits[i].doc);
    	
    					DBpediaIndexEntry e = DBpediaIndexEntry.fromDocument(doc);
    					
    					if(e.getUri().equals("http://dbpedia.org/resource/B._J._Symons")) {
    						int sdsdf = 123;
    					}

    					if(isVerbose()) {
    					    System.out.println(e.getClass_label() + ": " + e.getLabel());
    					}
    					
    					results.add(e);
    				}
    			}
    			
//    			cache.put(value, results);
			} else {
			    if(isVerbose()) {
			        System.out.println(String.format("Empty query for '%s'", label));
			    }
			}
			
			result = System.currentTimeMillis() - start;

		} catch (IOException e) {
			e.printStackTrace();
		} catch (ParseException e1) {
		    System.err.println(String.format("Parse exception for label '%s'", label));
			e1.printStackTrace();
		}
		if(isVerbose()) {
		    System.out.println(" returning " + results.size() + " documents");
		    System.out.println(String.format("setup: %d\tsearch: %d\tload: %d", setup, search, result));
		}
		return results;
	}

}
