package de.mannheim.uni.matching.firstline;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import de.mannheim.uni.matching.MatchingAdapter;
import de.mannheim.uni.matching.MatchingPair;
import de.mannheim.uni.similarity.matrix.SimilarityMatrix;
import de.mannheim.uni.utils.concurrent.Consumer;
import de.mannheim.uni.utils.concurrent.Parallel;
import de.mannheim.uni.utils.data.Pair;
import de.mannheim.uni.utils.timer.Timer;

/**
 * Generic first-level matcher class
 * generates one similarity value per instance pair
 * @author Oliver
 *
 */
public class LabelBasedMatcherWithFiltering<T> extends FirstLineMatcher<T> {

    private StringBuilder matchingLog;
    public StringBuilder getMatchingLog() {
        return matchingLog;
    }

	/**
	 * runs the matching on the provided instance sets
	 * for all instances in instancesToMatch, candidate instances from candidates are selected by the blocking instance 
	 * and then the similarity scores are calculated by the similarity measure instance  
	 */
	public SimilarityMatrix<T> match(Collection<T> instancesToMatch,
			final Collection<T> candidates, final MatchingAdapter<T> adapter) {
		
	    matchingLog = new StringBuilder();
	    
	    if(candidates == null || instancesToMatch == null) {
			int sdf = 123;
		}
	    
		// create similarity matrix
		final SimilarityMatrix<T> sim = getSimilarityMatrixFactory().createSimilarityMatrix(instancesToMatch.size(), candidates.size());
		
		try {
            new Parallel<T>(1).foreach(instancesToMatch, new Consumer<T>() {

                public void execute(T parameter) {
                    Collection<T> blockedCandidates = null;
                    
                    Timer t = null;
                    
                    if(getParentTimer()!=null) {
                        t = Timer.getNamed("LabelBasedMatcher: Blocking", getParentTimer(), true);
                    }
                    blockedCandidates = getBlocking().getCandidates(parameter, candidates);
                    if(t!=null) {
                        t.stop();
                    }
                    List<T> valList = new ArrayList<>(1);
                    valList.add(parameter);
                    Collection<Pair<T, T>> filtered =new ArrayList<Pair<T,T>>();             
                    
                    for(T o : blockedCandidates) {
                        filtered.add(new Pair<>(parameter, o));
                    }
                    Double maxSimilarity = 0.0;
                    T maxMatch = null;
                    // iterate over all candidates (all instances that could be matched to the current instance)
                    for(Pair<T, T> p : filtered)
                    {
                        // calculate similarity
                        if(getParentTimer()!=null) {
                            t = Timer.getNamed("LabelBasedMatcher: Similarity Calculation", getParentTimer(), true);
                        }                        
                        Double similarity = getSimilarityMeasure().calculate(p.getFirst(), p.getSecond(), adapter, null);
                        
                        if(similarity != null && similarity > maxSimilarity) {
                        	maxSimilarity = similarity;
                        	maxMatch = p.getSecond();
                        }
                        if(t!=null) {
                            t.stop();
                        }
                        
                        if(isCollectMatchingInfo()) {
                            getPairs().add(new MatchingPair<T>(p.getFirst(), p.getSecond()));
                            synchronized (matchingLog) {
                                matchingLog.append(String.format("%s <> %s (%.2f)", adapter.getLabel(p.getFirst()), adapter.getLabel(p.getSecond()), similarity));
                            }
                        }
                        
                        // only add if greater than zero
                        if(similarity!=null && similarity>0.0)
                        {
                            synchronized (sim) {                       
//                                System.out.println("set in matrix: " +p.getFirst() + " --- " + p.getSecond() + "---" + similarity);
                                sim.set(p.getFirst(), p.getSecond(), similarity);
                            }
                        }
                    }
                    
                    int sdf = 123;
                }
                
            }, "LabelBasedMatcher");
        } catch (Exception e) {
            e.printStackTrace();
        }
//		// iterate over all instances to match
//		for(T instance : instancesToMatch) {
//			
//			// iterate over all candidates (all instances that could be matched to the current instance)
//			for(T candidate : getBlocking().getCandidates(instance, candidates))
//			{
//				// calculate similarity
//				Double similarity = getSimilarityMeasure().calculate(instance, candidate, adapter);
//				
//				// only add if greater than zero
//				if(similarity!=null)
//				{
//					sim.set(instance, candidate, similarity);
//				}
//			}
//			
//		}
		
		return sim;
	}

}
